#! /usr/bin/python
# -*- coding: utf-8 -*-

"""	
	SPLIT TO MODULES
	
	This module defines classes to load data of different formats.
	
	Version 3.7								(02-Jun-2015 to 04-Jun-2015)
	- ConstParseLoader and DepParseLoader are moved to constparse.py and
	  depparse.py respectively.
	- PARC700 classes are moved to parse/parc700.py.
	
	Version 3.6												(25-May-2015)
	- All CoNLL classes are moved to parse/conll.py
	
	Version 3.5												(07-May-2015)
	- POSTagLoader is moved to parse/pos.py.
	
	Version 3.4												(13-Feb-2015)
	- getParses() is added to ConstParseLoader.
	
	Version 3.3												(08-Dec-2014)
	- loadLorgParses() is updated to replace the no-parse sentences with 
	  the parse trees from an alternative data set.
	- loadFromPTB() is added to POSTagLoader.
	
	Version 3.2												(04-Nov-2014)
	- ConstPArseLoader.loadLorgParses() is updated to support older version
	  of Lorg as well as non-verbos output.
	
	Version 3.1												(10-Jul-2014)
	- CoNLLSentence.getCoNLLTokens() is added.
	- getTokenID() is added to CoNLLToken.
	- isPredicate(self), getPredicateLabel() added to CoNLLToken2009.
	
	Version 3.0												(28-Apr-2014)
	- toEngMaltInput() is moved down to CoNLL2008Sentence and CoNLL2009Sentence.
	- toEngMaltInput() methods are renamed to toCoNLLParserInput().
	- toCoNLLParserInput() is added to CoNLL2009Token.
	- CoNLLParserInputSentence and CoNLLParserInputToken are added.
	- CoNLLDataLoader.createCoNLLSentence() is upgraded to support 
	  CoNLLParserInputSentence.
	
	Version 2.9												(24-Apr-2014)
	- POSTagLoader is changed to support extension to other formats. Now,
	  SlashTag format loader should be explicitly called instead of passing
	  arguments to the constructor.
	- mapPOSTagging() is added to CoNLLSentence.
	- mapPOSTags() is added to CoNLLToken.
	
	Version 2.8												(10-Apr-2014)
	- onSRLConstraintViolation of CoNLL2009Sentence is deprecated. Instead,
	  it is handled via of getSRL() argument for annotation option.
	- setLemmasFrom() is added to CoNLL2009Sentence.
	
	Version 2.7												(03-Apr-2014)
	- CoNLLSentence.predicateCount() is changed to compute the predicate
	  count by checking all tokens for being a predicate.
	- loadFromDepTree() is added to CoNLL2009Sentence.
	- loadFromDepNode() is added to CoNLL2009Token.
	
	Version 2.6												(01-Apr-2014)
	- toCoNLL2006() is added to PARC700Sentence and PARC700Token.
	- Language is added to PARC700Sentence attributes.
	
	Version 2.5												(25-Mar-2014)
	- findSegment(), getSrcFile() are added to EuroparlLoader.
	- ifContains() is added to EuroparlParagraph.
	- EuroparlSrcFile.name is now the name of file excluding the path.
	- EuroparlParagraph now stores the original text.
	- getChapter() is added to EuroparlSrcFile.
	- Position is added to EuroparlParagraph and its constructor.
	- Chapter lines (thus paragraphs) are stored without the end-of-line.
	
	Version 2.4												(14-Mar-2014)
	- getCoNLLLine() is added to CoNLL2006Token.
	- POSTagLoader is added.
	
	Version 2.3												(03-Feb-2014)
	- DepParseLoader is moved from parse.
	- version() is added to CoNLLxxxxSentence classes.
	
	Version 2.2												(27-Jan-2014)
	- Language is added to CoNLLDataLoader and CoNLL sentence classes. 
	  This may introduce bug in scripts used the classed before as the 
	  constructor needs a compulsory argument.
	
	Version 2.1												(21-Jan-2014)
	- ConstParseLoader is moved over from parse.py.
	
	Version 2.0												(12-Jan-2014)
	- Methods to return gold and predicted POS tags, head and dependency 
	  relations are added to CoNLLToken, CoNLL2008Token and CoNLL2009Token().
	- generatePropositionTrees() and _generatePropositionTree() of CoNLL2009Sentence
	  are depracated. PropositionTree class in tk.py should be used with
	  dependency trees loaded from CoNLLSentence.
	- CoNLLSentence.depTree is changed from property to method to add
	  pdAnnotationOptions.
	  
	Version 1.9												(07-Jan-2014)
	- A bug fixed in CoNLL2009Sentence._generatePropositionTree() regarding 
	  predicates locating after all arguments or with no arguments.
	- CoNLLSentence.getSpanCoNLLTokens() is added.
	
	Version 1.8												(27-Dec-2013)
	- CoNLLSentence.generatePropositionTrees() is added.
	- CoNLLSentence.getSpanSurface() is added.
	
	Version 1.7												(23-Dec-2013)
	- CoNLLSentence is broken down into subclasses for each format version.
	- CoNLLDataLoader.createCoNLLSentence() is added. 
	
	Version 1.6												(20-Dec-2013)
	- toCoNLL2009 is added to PARC700 classes.
	
	Version 1.5												(17-Dec-2013)
	- Bug was fixed in CoNLLDataLoader.selectSentences() with list option
	  complement selection.
	
	Version 1.4												(13-Dec-2013)
	- PARC700Sentence.hasMISSINGHEAD() is added.
	- toMaltInput() is added to PARC700Sentence and PARC700Token.
	
	Version 1.3												(10-Dec-2013)
	- EuroparlLoader and dependent classes are added.
	- Bug was fixed in CoNLLDataLoader.selectSentences() with list option.
	
	Version 1.2												(27-Nov-2013)
	- __init__() methods of CoNLLToken, CoNLL2006Token, CoNLL2008Token and
	  CoNLL2009Token are improved to support empty token creation.
	- toCoNLL2009() is added to CoNLL2006Token and CoNLL2008Token and 
	  CoNLLSentence.
	- CoNLL2008.__init__() is edited to support test data format.
	- toEngMaltInput() is added to CoNLLSentence, CoNLL2008Token and CoNLL2009Token.
	- CoNLLLoader.write() is added.
	
	Version 1.1												(24-Nov-2013)
	- CoNLLDataLoader.selectedSentences() is improved to set length filter
	  and return indexes.
	
	Version 1.0												(22-Nov-2013)
	- CoNLL loader and PARC700 loader are moved from parse.py.
	- CoNLL 2008 and 2009 formats are supported.
	- CoNLLToken is divided into format version classes.
	- CoNLLSentence.__str__() is modified to print the original format of
	  the sentence instead of surface form.
	
"""

import sys, random, os, re, copy
from nlp import nlp

## Start of Europarl ###################################################

class EuroparlLoader:
	'''
	Class for loading Europarl data for a language from source release.
	
	See http://www.statmt.org/europarl/ and source release text files
	'''
	
	def __init__(self, pDataDirname, pFileExtension = "txt", pLanguage = "en"):
		'''
		Creates object and loads data from test files located in pDataDirname
		with the specified extension.
		'''
		
		self.srcFiles = []
		self.language = pLanguage
		
		for vName in os .listdir(pDataDirname):
			if os.path.isfile(os.path.join(pDataDirname, vName)):
				vFilename = os.path.join(pDataDirname, vName)
				if vFilename.endswith('.' + pFileExtension):
					self.srcFiles.append(EuroparlSrcFile(vFilename, self.language))
		
	
	
	def findSegment(self, pSegment):
		'''
		Looks for a given segment in the entire data and returns file, 
		chapter and paragraph if found together with sentence position 
		if the segment turns out to be an exact sentence (-1 otherwise).
		Otherwise, None will be returned.
		
		It halts the search when it meets the first occurrence of the segment.
		
		If the segment is tokenized, it is possible that it cannot be found
		as the source data are normally not tokenized.
		'''
		
		for vFile in self.srcFiles:
			for vChapter in vFile.chapters:
				for vParagraph in [p for p in vChapter.paragraphs if p.speaker != None and p.speaker.srcLanguage == 'EN']:
					for i, vSent in enumerate(vParagraph.sentences):
						if pSegment == vSent:
							return vFile, vChapter, vParagraph, -1
						elif vParagraph.ifContains(pSegment):
							return vFile, vChapter, vParagraph, None
		
		return None
		
	
	
	def getSrcFile(self, pSrcFileName):
		'''
		Returns the source file with the given name
		'''
		
		for vFile in self.srcFiles:
			if vFile.name == pSrcFileName:
				return vFile
		
		return None
		
	
	


class EuroparlSrcFile:
	'''
	Class for Europarl source file.
	
	See http://www.statmt.org/europarl/ and source release text files
	
	A EuroparlSrcFile consist of EuroparlChapters.
	'''
	
	def __init__(self, pSrcFilename, pLanguage):
		'''
		Creates a Europarl source file object from pSrcFilename.
		'''
		
		self.name = os.path.basename(pSrcFilename)
		self.language = pLanguage
		self.chapters = []
		
		vlChapterLines = []
		for vLine in open(pSrcFilename):
			if vLine.startswith("<CHAPTER ID"):
				# loading the collected content into the current chapter
				if len(self.chapters) > 0:
					self.chapters[-1].loadContent(vlChapterLines)
				# initializing a new chapter
				vID = vLine[12: vLine.find('>')]
				self.chapters.append(EuroparlChapter(vID, self.language))
				vlChapterLines = []
			else:
				vlChapterLines.append(vLine.strip())
		
		# loading the content of the last chapter
		self.chapters[-1].loadContent(vlChapterLines)
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		return self.name
		
	
	
	def getChapter(self, pChapterID):
		'''
		Returns the chapter with the given ID
		'''
		
		for vChapter in self.chapters:
			if vChapter.id == pChapterID:
				return vChapter
		
		return None
		
	
	


class EuroparlChapter:
	'''
	Class for Europarl chapter.
	
	A EuroparlChapter consists of EuroparlParagraphs.
	'''
	
	def __init__(self, pID, pLanguage):
		'''
		Creates a chapter object and load content and attributes.
		'''
		
		self.id = pID
		self.language = pLanguage
		self.paragraphs = []
		
	
	
	def loadContent(self, plChapterLines):
		'''
		Loads the chapter content (paragraphs)
		'''
		
		vCurrentSpeaker = None
		vParagraphCntr = 0
		for vLine in plChapterLines:
			if vLine.startswith("<SPEAKER ID"):
				vID = int(vLine[12: vLine.find(' ', 12)].strip('"'))
				
				vSrcLangPos = vLine.find("LANGUAGE=")
				if vSrcLangPos == -1:
					vSrcLanguage = ""
				else:
					vSrcLanguage = vLine[vSrcLangPos + 10: vLine.find('"', vSrcLangPos + 10)]
				
				vNamePos = vLine.find("NAME=")
				if vNamePos == -1:
					vSpeakerName = ""
				else:
					vSpeakerName = vLine[vNamePos + 6: vLine.find('"', vNamePos + 6)]
				
				vAffPos = vLine.find("AFFILIATION=")
				if vAffPos == -1:
					vAffiliation = ""
				else:
					vAffiliation = vLine[vAffPos + 13: vLine.find('"', vAffPos + 13)]
				
				vCurrentSpeaker = EuroparlSpeaker(vID, vSrcLanguage, vSpeakerName, vAffiliation)
			elif vLine.startswith("<P>") or vLine.strip() == '':
				continue
			else:
				vParagraphCntr += 1
				self.paragraphs.append(EuroparlParagraph(vLine, vParagraphCntr, vCurrentSpeaker, self.language))
		
	
	
	def getParagraph(self, pParagraphPos):
		'''
		Returns the paragraph at the given position in the chapter
		'''
		
		for vParagraph in self.paragraphs:
			if vParagraph.position == pParagraphPos:
				return vParagraph
		
		return None
		
	
	


class EuroparlParagraph:
	'''
	Class for Europarl paragraphs
	
	A EuroparlParagraph can be a speaker speech or an adjunct paragraph 
	such as "Resumption of the session". Adjunct paragraphs are identified
	only by having no speaker assigned.
	
	NOTE: Identifying adjuncts is done trivially here and can be improved. 
	'''
	
	def __init__(self, pParagraph, pPosition, pSpeaker, pLanguage):
		'''
		Creates a paragraph object and loads content and attributes.
		'''
		
		self.language = pLanguage
		self.text = pParagraph
		self.position = pPosition
		self.sentences = nlp.segToSentNLTK(pParagraph.strip(), self.language)
		
		vSpeaker = copy.deepcopy(pSpeaker)
		
		if pParagraph.startswith('('):
			## Sometimes language is specified in this format followed
			## by the speech line.
			if re.match("^\([A-Z]{2}\)", pParagraph):
				if vSpeaker != None:
					vSpeaker.srcLanguage == pParagraph[1:3]
			# whole line in parenthesis is considered as adjunct
			elif re.match("^\([^)]+\)$", pParagraph):
				vSpeaker = None
		
		self.speaker = vSpeaker
		
	
	
	def isSpeech(self):
		'''
		Returns true is the paragraph is a speech by an speaker
		'''
		
		return (self.speaker != None)
		
	
	
	def isAdjunct(self):
		'''
		Returns true is the paragraph is adjunct (not a speech)
		'''
		
		return (self.speaker == None)
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		return self.content
		
	
	
	def ifContains(self, pSegment):
		'''
		Returns true if the paragraph contains a given segment
		'''
		
		return pSegment in self.text
		
	
	


class EuroparlSpeaker:
	'''
	Class for the identity of speaker
	'''
	
	def __init__(self, pID, pSrcLanguage, pName, pAffiliation):
		'''
		Creates a speaker object and load the attributes
		'''
		
		self.id = pID
		self.srcLanguage = pSrcLanguage
		self.name = pName
		self.affiliation = pAffiliation
		
	


## End of Europarl #####################################################

