#include <stdio.h>
#include <stdlib.h>
#include <math.h>

// ToDo: make lambda a parameter
double LAMBDA = 0.4;

double *compKernelMatrixSM(double *, int, int, double *, int, int, double *, double *, int, int, int);
double *compKernelMatrixWEST(double *, int, int, double *, int, int, double *, double *, int, double *, int, int, double, int, int);
double *compKernelMatrixWESS(double *, int, int, double *, int, int, double *, double *, int, double *, int, int, int, int);
double *_compKernelMatrix(double *, int, int, double *, int, int, double *, double *, int, double *, int, int, double, int, int);
double _compKernel(double *, int, double *, int, double *, double *, int, double *, double, int, double);
double _calcCosSim(double *, double *, int, double *, double *, int);
int _isAuxEqual(double *, double *, int);




/*
 * Computes anygram kernel matrix for the given samples using string match
 * (SM) method and returns the result in an 1D array
 * 
 * In string match method, each pair of tokens from a pair of instances 
 * are compared against each other. Mathing tokens get a score of 1 and 
 * different tokens get 0.
 * 
 * See _compKernelMatrix() for more details.
 */
double *compKernelMatrixSM(double *pX, int pXShape1, int pXShape2,
                           double *pTrainX, int pTXShape1, int pTXShape2,
                           double *pXAux, double *pTrainXAux, int pAuxDim,
                           int pPhase, int pflgNormalize)
{
	return _compKernelMatrix(pX, pXShape1, pXShape2, 
	                         pTrainX, pTXShape1, pTXShape2,
	                         pXAux, pTrainXAux, pAuxDim,
	                         NULL, 0, 0, 0,
	                         pPhase, pflgNormalize);
}



/*
 * Computes anygram kernel matrix for the given samples using word embedding
 * similarity threshold (WEST) method and returns the result in an 1D array
 * 
 * In word embedding similarity threshold method, the similarity of the word
 * embedding vectors of each token in a pair of tokens from a pair of 
 * instances is compared to a threshold. If it is above the threshold, the
 * tokens are considered similar and get a score of 1. Otherwise, they get
 * a score of 0.
 * 
 * See _compKernelMatrix() for more details.
 */
double *compKernelMatrixWEST(double *pX, int pXShape1, int pXShape2,
                             double *pTrainX, int pTXShape1, int pTXShape2,
                             double *pXAux, double *pTrainXAux, int pAuxDim,
                             double *pEmbeddings, int pEmbedVocabSize, int pEmbedDim, double pThreshold,
                             int pPhase, int pflgNormalize)
{
	return _compKernelMatrix(pX, pXShape1, pXShape2, 
	                         pTrainX, pTXShape1, pTXShape2,
	                         pXAux, pTrainXAux, pAuxDim,
	                         pEmbeddings, pEmbedVocabSize, pEmbedDim, pThreshold,
	                         pPhase, pflgNormalize);
}



/*
 * Computes anygram kernel matrix for the given samples using word embedding
 * similarity score (WESS) method and returns the result in an 1D array
 * 
 * In word embedding similarity score method, the similarity of the word
 * embedding vectors of each token in a pair of tokens from a pair of 
 * instances is used as the score.
 * 
 * See _compKernelMatrix() for more details.
 */
double *compKernelMatrixWESS(double *pX, int pXShape1, int pXShape2,
                             double *pTrainX, int pTXShape1, int pTXShape2,
                             double *pXAux, double *pTrainXAux, int pAuxDim,
                             double *pEmbeddings, int pEmbedVocabSize, int pEmbedDim,
                             int pPhase, int pflgNormalize)
{
	return _compKernelMatrix(pX, pXShape1, pXShape2, 
	                         pTrainX, pTXShape1, pTXShape2,
	                         pXAux, pTrainXAux, pAuxDim,
	                         pEmbeddings, pEmbedVocabSize, pEmbedDim, 0,
	                         pPhase, pflgNormalize);
}



/*
 * Computes kernel matrix for the given samples and returns the result
 * in an 1D array
 * 
 * 1D arrays are used instead of 2D ones to make it easier to work with 
 * by external callers such as ctypes in Pyhon.
 * 
 * pPhase determines the training/test phase. The purpose is to avoid
 * computing the norm when it's the training phase and the values are
 * already calculated during kernel computation (because in training phase
 * pX and pTrainX are the same things). Values are:
 * - 0: training
 * - 1: test
 * 
 * pflgNormalize determines whether the kernel values must be normalized:
 * - 0: do not normalize 
 * - 1: normalize
 * 
 */
double *_compKernelMatrix(double *pX, int pXShape1, int pXShape2,
                          double *pTrainX, int pTXShape1, int pTXShape2,
                          double *pXAux, double *pTrainXAux, int pAuxDim,
                          double *pEmbeddings, int pEmbedVocabSize, int pEmbedDim, double pThreshold, 
                          int pPhase, int pflgNormalize)
{
	double *aXL2 = (double *)malloc(pXShape1 * sizeof(double));
	double *aTXL2 = (double *)malloc(pTXShape1 * sizeof(double));
	double *aK = (double *)malloc(pXShape1 * pTXShape1 * sizeof(double));
	
	int i, j;
	
	double vEmbedSize = pEmbedVocabSize * pEmbedDim;                    // size of embeddings in terms of number of array elements.
	
	// 1. computing norms for normalization if this is not the training phase
	
	if (pflgNormalize == 1 && pPhase != 0)
	{
		// L2 norm for pX
		
		for (i = 0; i < pXShape1; i++)
		{
			int vXOffset = i * pXShape2;                                // offset in input X matrix to substitute the second dimension
			int vAuxXOffset = i * pXShape2 * pAuxDim;                  // offset in auxiliary X matrix to substitute the second dimension
			
			aXL2[i] = _compKernel(pX + vXOffset, pXShape2,
			                      pX + vXOffset, pXShape2,
			                      pXAux + vAuxXOffset, pXAux + vAuxXOffset, pAuxDim,
			                      pEmbeddings, vEmbedSize, pEmbedDim, pThreshold);
		}
		
		
		// L2 norm for pTrainX
		
		for (i = 0; i < pTXShape1; i++)
		{
			int vTXOffset = i * pTXShape2;                              // offset in input TrainX matrix to substitute the second dimension
			int vTAuxXOffset = i * pTXShape2 * pAuxDim;           // offset in auxiliary TrainX matrix to substitute the second dimension
			
			aTXL2[i] = _compKernel(pTrainX + vTXOffset, pTXShape2,
			                       pTrainX + vTXOffset, pTXShape2,
			                       pTrainXAux + vTAuxXOffset, pTrainXAux + vTAuxXOffset, pAuxDim,
			                       pEmbeddings, vEmbedSize, pEmbedDim, pThreshold);
		}
	}
	
	
	// 2. computing kernel values
	
	for (i = 0; i < pXShape1; i++)
	{
		int vKOffset = i * pTXShape1;                                   // offset in output kernel matrix to substitute the second dimension
		
		int vXOffset = i * pXShape2;                                    // offset in input X matrix to substitute the second dimension
		int vAuxXOffset = i * pXShape2 * pAuxDim;                      // offset in auxiliary X matrix to substitute the second dimension
		
		for (j = 0; j < pTXShape1; j++)
		{
			int vTXOffset = j * pTXShape2;                              // offset in input TrainX matrix to substitute the second dimension
			int vTAuxXOffset = i * pTXShape2 * pAuxDim;           // offset in auxiliary TrainX matrix to substitute the second dimension
			
			aK[vKOffset + j] = _compKernel(pX + vXOffset, pXShape2, 
			                               pTrainX + vTXOffset, pTXShape2, 
			                               pXAux + vAuxXOffset, pTrainXAux + vTAuxXOffset, pAuxDim,
			                               pEmbeddings, vEmbedSize, pEmbedDim, pThreshold);
			
			// norms when this is training phase
			if (pflgNormalize == 1 && pPhase == 0 && i == j)
			{
				aXL2[i] = aK[vKOffset + j];
				aTXL2[j] = aK[vKOffset + j];
			}
			//printf("%f,", aK[vKOffset + j]);fflush(stdout);
		}
		
		printf(".");
		fflush(stdout);
	}
	
	
	// 3. normalizing kernel values
	
	if (pflgNormalize == 1)
	{
		for (i = 0; i < pXShape1; i++)
		{
			int vKOffset = i * pTXShape1;                               // offset in output kernel matrix to substitute the second dimension
			
			for (j = 0; j < pTXShape1; j++)
			{
				int vTXOffset = j * pTXShape2;                          // offset in input TrainX matrix to substitute the second dimension
				
				if (pPhase == 0 && i == j)
					aK[vKOffset + j] = 1;                               // NOTE: this can be done in kernel computation to save computation
				else
					aK[vKOffset + j] = aK[vKOffset + j] / sqrt(aXL2[i] * aTXL2[j]);
				//printf("%f,", aK[vKOffset + j]);
			}
			
			//printf("\n");
			//fflush(stdout);
		}
		
		// freeing memory for norm arrays 
		free(aXL2);
		free(aTXL2);
	}
	
	
	return aK;
	
}



/*
 * Computes kernel value for a pair of samples
 * 
 * If the word embeddings is null, the string match (SM) method will be 
 * used. Otherwise, the word embedding similarity score method will be used.
 * In the latter, if a token does not have any embedding vector in the given
 * embeddings, the it will fall back string match for its similarity scores.
 */
double _compKernel(double *pX, int pXLen, 
                   double *pTrainX, int pTXLen,
                   double *pXAux, double *pTrainXAux, int pAuxDim,
                   double *pEmbeddings, double pEmbedSize, int pEmbedDim, double pThreshold)
{
	// figuring out the method based on evidences (can be eliminated if an argument is passed specifying the method)
	char vMethod;
	if (pEmbeddings == NULL || pEmbedSize == 0)
		vMethod = 'm';                                                  // SM
	else if (pThreshold != 0)
		vMethod = 't';                                                  // WEST
	else
		vMethod = 's';                                                  // WESS
	
	
	double **aDelta = (double **)malloc(pXLen * sizeof(double *));      // Delta table: rows=length(pX), columns=length(pTrainX)
	double vK = 0;                                                      // kernel value
	double vSim = 0;
	int i, j;
	
	// filling in the Delta table: bottom-up/left-right
	for (i = pXLen - 1; i >= 0; i--)
	{
		aDelta[i] = (double *)malloc(pTXLen * sizeof(double));
		
		for (j = 0; j < pTXLen; j++)
		{
			// ignoring the padding
			
			if (pX[i] == 0 || pTrainX[j] == 0)
				break;
			
			// calculating the similarity score
			
			if (vMethod == 'm')
			{
				vSim = (pX[i] == pTrainX[j]) && _isAuxEqual(pXAux + i * pAuxDim, pTrainXAux + j * pAuxDim, pAuxDim);
				
				if (vSim == 0)                                          // if not equal, ignore the rest of the n-gram (equivalent to children in tree kernel)
				{
					aDelta[i][j] = 0;
					continue;
				}
			}
			else if (vMethod == 't' || vMethod == 's')
			{
				/*
				int s = _isAuxEqual(pXAux + i * pAuxDim, pTrainXAux + j * pAuxDim, pAuxDim);
				vSim = (pX[i] == pTrainX[j]) && s;
				if ((pX[i] == pTrainX[j]) == 1 && s == 0)
				{
					printf("(%f %f)\n", pX[i], pTrainX[j]);
					printf("----------\n");
				}*/
				
				// comparing for exact match to save efforts of 
				vSim = (pX[i] == pTrainX[j]) && _isAuxEqual(pXAux + i * pAuxDim, pTrainXAux + j * pAuxDim, pAuxDim);
				
				if (vSim == 0)
				{
					int vXEmbedOffset = (int)pX[i] * pEmbedDim;         // offset of pX[i] in the embedding array
					int vTXEmbedOffset = (int)pTrainX[j] * pEmbedDim;   // offset of pTrainX[i] in the embedding array
					
					// falling back to string match if either of the tokens does not have an embedding vector
					if (vXEmbedOffset >= pEmbedSize || vTXEmbedOffset >= pEmbedSize)
					{
						// already know they don't match (vSim is 0 here)
						aDelta[i][j] = 0;
						continue;
					}
					else
					{
						vSim = _calcCosSim(pEmbeddings + vXEmbedOffset, pEmbeddings + vTXEmbedOffset, pEmbedDim, pXAux + i * pAuxDim, pTrainXAux + j * pAuxDim, pAuxDim);
						
						if (vMethod == 't')                             // WEST method
						{
							if (vSim >= pThreshold)
								vSim = 1; 
							else
							{
								aDelta[i][j] = 0; 
								continue;
							}
						}
						/*else if (vMethod == 's')
						{
							if (vSim <= 0)
							{
								aDelta[i][j] = 0; 
								continue;
							}
						}*/
					}
				}
			}
			// else raise an error on the input provided
			
			// calculating delta value
			if (i != pXLen - 1  && j != pTXLen -1 && pX[i + 1] != 0 && pTrainX[j + 1] != 0)      // if not edge cell or padding
				aDelta[i][j] = LAMBDA * (vSim + aDelta[i + 1][j + 1]);
			else
				aDelta[i][j] = LAMBDA * vSim;
			
			vK += aDelta[i][j];
		}
	}
	
	
	// freeing memory for Delta 
	for (i=0; i < pXLen; i++)
		free(aDelta[i]);
	free(aDelta);
	
	return vK;
}



/*
 * Calculates and returns the cosine similarity of the given vectors
 * 
 * Vectors can be provided in two parts: main and auxiliary. The auxiliary
 * one will be treated as the concatenation to their corresponding main 
 * vectors.
 */
double _calcCosSim(double *pVector1, double *pVector2, int pDim, double *pAuxVector1, double *pAuxVector2, int pAuxDim)
{
	double vDot = 0.0;
	double vMag1 = 0.0, vMag2 = 0.0;
	short i;
	
	for (i = 0; i < pDim; i++) 
	{
		vDot += pVector1[i] * pVector2[i];
		vMag1 += pVector1[i] * pVector1[i];
		vMag2 += pVector2[i] * pVector2[i];
	}
	
	for (i = 0; i < pAuxDim; i++) 
	{
		vDot += pAuxVector1[i] * pAuxVector2[i];
		vMag1 += pAuxVector1[i] * pAuxVector1[i];
		vMag2 += pAuxVector2[i] * pAuxVector2[i];
	}
	
	if (vMag1 == 0 || vMag2 == 0)
		return 0;
	else
		return vDot / (sqrt(vMag1) * sqrt(vMag2));
}



/*
 * Compares the auxiliary input for the given tokens and returns 1 if equal, 0 otherwise
 */
int _isAuxEqual(double *pXAux, double *pTrainXAux, int pAuxDim)
{
	int i;
	
	for (i = 0; i < pAuxDim; i++)
	{
		//printf("(%f %f)\n", pXAux[i], pTrainXAux[i]);
		if (pXAux[i] != pTrainXAux[i])
			return 0;
	}
	return 1;
}
