#! /usr/bin/python
# -*- coding: utf-8 -*-

"""
	NOTE: CONVERT LISTS TO NUMPY ARRAY FOR EFFICIENT COMPUTING. IT HAS
	      ALREADY STARTED FROM V 0.2.
	
	This module provides classes to load and manipulate machine learning
	data in various formats.
	

	Version 0.5											   (17-Aug-2016)
	- loadFromArray() is added to Dataset.

	Version 0.4								               (24-Feb-2016)
	- An exception is added to Feature.addValue() to convert integer data
	  to float if the feature type is set to float and the given data is 
	  integer.
	
	Version 0.3								(08-Feb-2016 to 10-Feb-2016)
	- loadValues() is added to Feature.
	- loadTargetValues() in Dataset is changed to load from list instead
	  of file.
	
	Version 0.2                             (08-Dec-2015 to 10-Dec-2015)
	- getFeatureMatrix() and getFeatureMatrixT() are added to Dataset.
	- plot() is added to Dataset.
	
	Version 0.1								(28-Jul-2015 to 16-Oct-2015)
	- Dataset, Features, Feature, IDataInstances, DataInstance, 
	  FeatureVector, InstanceFeature are added.
	
"""


import numpy as np
from utils import util


class Dataset:
	'''
	Class for storing and manipulating dataset
	
	At the implementation level, dataset can be seen as either a set of
	instances each containing a set of features and target values or a 
	set of features and targets each containing values for each instance.
	At the usage level, however, both should be accessible. We choose the
	latter for implementation provide the former via an interface. One 
	reason is that in this way there are (roughly) as many objects to store
	as there are features, whereas in the other way there are #instances*
	#features feature objects to store.
	'''
	
	
	def __init__(self, pTargetType = "numerical"):
		'''
		Constructor
		
		pTargetType specifies the data type in which the target values 
		must be stored. See the target type normalization method or details.
		'''
		
		self._targetType = self._normalizeTargetType(pTargetType)
		
		# features of Features class type
		self._features = Features()
		# target can be multi-value
		self._targets = []
		
		# instances using an interface object of IDataInstances type
		#self._instances = IDataInstances(pDataset = self)
		
		# for future where the feature structure can be set beforehand
		self._fixedFS = False
		
	
	
	@property
	def features(self):
		'''
		Returns the features vector which is a structure enabling access to
		features by their names or indexes
		
		Each feature an object of type Feature.
		'''
		
		return self._features
		
	
	
	@property
	def dimension(self):
		'''
		Returns the number of features as the dimension of the data
		'''
		
		return len(self._features)
		
	
	
	@property
	def instances(self):
		'''
		Creates and returns the data set instances which is an IDataInstances
		object interfacing the data set instances as a list
		'''
		
		return IDataInstances(pDataset = self)
		
	
	
	def _normalizeTargetType(self, pTargetType):
		'''
		Normalizes and returns the input for target type
		
		The possible input values are:
		- nominal/string/str (saved as python str type)
		- numerical/number/float (saved as python float type)
		- integer/int (saved as python integer type)
		- python type str, int, float
		- [x1, x2, ...] (for multi-target problems)
		
		For the multi-target types, the type of individual targets are
		given in a list.
		'''
		
		if pTargetType in [str, int, float]:
			return pTargetType
		elif type(pTargetType) == str:
			vTargetType = pTargetType.lower()
			
			if vTargetType[:3] in ["nom", "str"]:
				return str
			elif vTargetType[:3] in ["num", "flo"]:
				return float
			elif vTargetType[:3] == "int":
				return int
			else:
				raise Exception("Unknown type for target: %s" % pTargetType)
		elif type(pTargetType) == list:
			vlTypes = []
			
			for vType in pTargetType:
				vlTypes.append(self._normalizeTargetType(vType))
			
			return vlTypes
		else:
			raise Exception("Unknown type for target: %s" % pTargetType)
		
	
	
	def getTargetType(self):
		'''
		Returns the target data type
		
		The target type can be any of python types str, int, float or a 
		list of these types.
		'''
		
		return self._targetType
		
	
	
	def isFixedFS(self):
		'''
		Returns true if the dataset has a fixed (readonly) feature structure
		'''
		
		return self._fixedFS
		
	
	
	def getFeatureNames(self):
		'''
		Returns feature names
		'''
		
		return [f.name for f in self.features]
		
	
	
	def __add__(self, vOtherDataset):
		'''
		Addition operator: creates and returns a dataset whose instances
		are the instances of this and another given dataset
		'''
		###################################
		vNewDataset = Dataset()
		vNewDataset.instances = self.instances + vOtherDataset.instances
		
		return vNewDataset
		
	
	
	def __radd__(self, vOtherDataset):
		'''
		List sum operator
		'''
		#########################################
		if vOtherDataset == 0:
			return self
		else:
			return self.__add__(vOtherDataset)
		
	
	
	def _createDataset(self):
		'''
		Creates and returns a dataset object
		'''
		
		return Dataset()
		
	
	
	@property
	def size(self):
		'''
		The size of the dataset in terms if the number of data instances
		'''
		
		return len(self.instances)
		
	
	
	def wipe(self):
		'''
		Removes all data and meta-data from the dataset
		'''
		
		self._features = Features()
		self._targets = []
		
	

	def loadFromArray(self, paFeatureValues, paLabels):
		'''
		Loads data from numpy arrays containing feature values and labels

		paFeatureValues: 2D numpy aray containing feature values
		paLabels: 2D numpy array containing labels

		NOTE: The number of rows in the two input arrays must be equal.
		'''

		self.wipe()

		if len(paFeatureValues) != len(paLabels):
			raise Exception("Feature and label arrays don't match: %s vs. % rows" % (len(paFeatureValues), len(paLabels)))

		for vIFeatureValues, vILabel in zip(paFeatureValues, paLabels):
			self._addInstance(pTarget=vILabel, plFeatureValuePairs=[(str(i), v) for i, v in enumerate(vIFeatureValues, start=1)])



	def loadFromFile(self, pFilename, pFormat):
		'''
		Loads the dataset from one-instance-per-line file in the specified
		format
		
		All the existing data and structure are deleted. To keep them, use
		addFromFile().
		
		If no structure has set up beforehand (NOT IMPLEMENTED YET), the
		feature data types are determined automatically. If digit-value
		features are meant to be strings, use quotations as in python strings.
		
		See addLine() for pFormat.
		'''
		
		self.loadFromLines(plLines = open(pFilename).read().strip().split('\n'), pFormat = pFormat)
		
	
	
	def loadFromLines(self, plLines, pFormat):
		'''
		Loads the dataset from lines in the specified format
		
		All the existing data and structure are deleted. To keep them, use
		addFromLines().
		
		If no structure has set up beforehand (NOT IMPLEMENTED YET), the
		feature data types are determined automatically. If digit-value
		features are meant to be strings, use quotations as in python strings.
		
		See addLine() for pFormat.
		'''
		
		self.wipe()
		
		self.addFromLines(plLines = plLines, pFormat = pFormat)
		
	
	
	def addFromFile(self, pFilename, pFormat):
		'''
		Adds the data into the dataset from one-instance-per-line file in
		the specified format
		
		All the existing data and structure are retained. To delete them,
		use loadFromFile().
		
		If no structure has set up beforehand (NOT IMPLEMENTED YET), the
		feature data types are determined automatically. If digit-value
		features are meant to be strings, use quotations as in python strings.
		
		See addLine() for pFormat.
		'''
		
		self.addFromLines(pDatasetLines = open(pFilename).read().strip().split('\n'), pFormat = pFormat)
		
	
	
	def addFromLines(self, plLines, pFormat):
		'''
		Adds the data into the dataset from lines in the specified format
		
		All the existing data and structure are retained. To delete them,
		use loadFromLines().
		
		If no structure has set up beforehand (NOT IMPLEMENTED YET), the
		feature data types are determined automatically. If digit-value
		features are meant to be strings, use quotations as in python strings.
		
		See addLine() for pFormat.
		'''
		
		vFormat = pFormat.lower()
		
		for vLine in plLines:
			self.addLine(pLine = vLine, pFormat = vFormat)
		
	
	
	def addLine(self, pLine, pFormat):
		'''
		Adds the data in the input line which is in the specified format
		
		pFormat specifies the format of the each instance line. Possible
		values are:
		- t,v,		<target>,<feature value>[,<feature value>[..]]
		- ti:v		<target> <feature index>:<feature value> [<feature index>:<feature value>[..]]
		- v,t		<feature value>[,<feature value>[..]]<target>
		- TO BE EXTENDED
		
		The following should be noted:
		- When no separator is used in the format string between target 
		  and feature/values and feature/values themselves, it means the
		  separator is a space or tab (e.g. in ti:v). 
		- Tab can alternatively be used instead of space. 
		- In all i:v formats, no space is allowed between : and v.
		
		If no structure has set up beforehand (NOT IMPLEMENTED YET), the
		feature data types are determined automatically. If digit-value
		features are meant to be strings, use quotations as in python strings.
		'''
		
		vFormat = pFormat.lower()
		
		if not pLine.startswith('#'):
			if vFormat == "t,v,":
				self._addLineFormat1(pLine)
			
			if vFormat == "ti:v":
				self._addLineFormat2(pLine)
			
			if vFormat == "v,t":
				self._addLineFormat3(pLine)
		
			if vFormat == "v,?":
				self._addLineFormat4(pLine)
		
	
	
	def _addLineFormat1(self, pLine):
		'''
		Adds the data in the input line which is in "t,v," format
		
		<target>,<feature value>[,<feature value>[..]]
		
		The position of each feature in the vector is used as its name,
		since no name is given in this format.
		'''
		
		vlLineSplit = pLine.split(',')
		
		self._addInstance(pTarget = vlLineSplit[0],
		                  plFeatureValuePairs = [(str(i), v) for i, v in enumerate(vlLineSplit[1:], start=1)])
		
	
	
	def _addLineFormat2(self, pLine):
		'''
		Adds the data in the input line which is in "ti:v" format
		
		<target> <feature index>:<feature value> [<feature index>:<feature value>[..]]
		
		The indexes will be used as feature names.
		'''
		
		vlLineSplit = pLine.split()
		
		## To make sure that each feature can be accessed correctly by index, 
		## the feature/value pairs are sorted based on the feature indexes
		## and features with values of None are inserted for missing indexes.
		
		vlRawFeatureValuePairs = sorted([fv.split(':') for fv in vlLineSplit[1:]], key = lambda x: int(x[0]))
		
		vlFeatureValuePairs = []
		vFeatureCounter = 1
		
		for vIndex, vValue in vlRawFeatureValuePairs:
			# missing indexes
			if int(vIndex) > vFeatureCounter:
				# adding as many None-value feature as missing indexes
				for i in range(vFeatureCounter, int(vIndex)):
					vlFeatureValuePairs.append((str(i), None))
					vFeatureCounter += 1
			
			vlFeatureValuePairs.append((vIndex, vValue))
			vFeatureCounter += 1
		
		self._addInstance(pTarget = vlLineSplit[0], plFeatureValuePairs = vlFeatureValuePairs)
		
	
	
	def _addLineFormat3(self, pLine):
		'''
		Adds the data in the input line which is in "v,t" format
		
		<feature value>[,<feature value>[..]],<target>
		
		The position of each feature in the vector is used as its name,
		since no name is given in this format.
		'''
		
		vlLineSplit = pLine.split(',')
		
		self._addInstance(pTarget = vlLineSplit[-1],
		                  plFeatureValuePairs = [(str(i), v) for i, v in enumerate(vlLineSplit[:-1], start=1)])
		
	
	
	def _addLineFormat4(self, pLine):
		'''
		Adds the data in the input line which is in "v,?" format where the target is place-held by a ? mark.
		
		<feature value>[,<feature value>[..]],<target place holder>
		
		The position of each feature in the vector is used as its name,
		since no name is given in this format.
		'''
		
		vlLineSplit = pLine.split(',')
		
		self._addInstance(pTarget = None,
		                  plFeatureValuePairs = [(str(i), v) for i, v in enumerate(vlLineSplit[:-1], start=1)])
		
	
	
	def loadTargets(self, plValues, pflgMatchDatasetSize = False):
		'''
		Loads the target values from a list
		
		The number of values in the list must match the dataset size if asked for checking the same.
		'''
		
		# sanity check
		if pflgMatchDatasetSize:
			if len(plValues) != self.size:
				raise Exception("Target values are not aligned with the data instances!")
		else:
			self._targets = plValues[:]
			# why was this done like thid???
			#for vInstance, vTarget in zip(self.instances, vlTargets):
			#	vInstance.setTarget(self.getTargetType()(vTarget))
		
	
	
	def _addInstance(self, pTarget, plFeatureValuePairs):
		'''
		Adds a new data instance with given target value(s) and feature-value pairs
		'''
		
		# creating a default feature vector
		vFeatureVector = FeatureVector(plFeatureValuePairs = plFeatureValuePairs, pflgAutoDataType = True)
		
		# creating data instance

		vInstance = DataInstance(pTarget, vFeatureVector)
		
		# adding data instance (will adapt the feature vector to the dataset if eligible and necessary)
		self.instances.addInstance(vInstance)
		
	
	
	def addNewFeature(self, pFeatureName, pDataType = type(None)):
		'''
		Adds a new feature with the given name
		'''
		
		vNewFeature = Feature(pName = pFeatureName, pDataType = pDataType)
		self.features.append(vNewFeature)
		
	
	
	def getTargets(self):
		'''
		Returns the target values of the dataset instances in a list in 
		the same order as the instances
		'''
		
		return self._targets
		
	
	
	def getTargetSet(self):
		'''
		Returns the target value set of the dataset in a dictionary with 
		entries being the target values and their counts
		'''
		
		return util.groupBy(self.getTargets())
		
	
	
	def getFeatureMatrix(self):
		'''
		Returns feature values in a matrix the rows of which represent data instances and the columns features
		'''
		
		return np.transpose(self.getFeatureMatrixT())
		
	
	
	def getFeatureMatrixT(self):
		'''
		Returns transposition of the feature values matrix where columns represent data instances and rows features
		
		Building a transposed matrix is more efficient since the dataset is already constructed based on features rather
		than instances (see class documentation above). 
		'''
		
		#### see if this can be made more efficient
		return np.array([f.values for f in self.features])
		
	
	
	def writeToFile(self, pOutFilename, pFormat = "tn:v"):
		'''
		Writes the dataset into file in the specified format
		
		The format string uses the following coding:
		- t: target value
		- n: feature name
		- i: feature index
		- v: feature value
		
		All other characters (e.g. delimiters) are written verbatim in the output. Note that, if no delimiter is specified
		in the format between target value and feature vector, tab is used. Also, if no delimiter is specified between
		features in the feature vector, single space is used.
		'''
		
		open(pOutFilename, 'w').write('\n'.join([i.getStr(pFormat) for i in self.instances]) + '\n')
		
	
	
	def plotFeature(self, pFeature, pOutFilename = None):
		'''
		Plots the given feature (name or index) and the target values
		
		The plot will be saved into the given output file or be shown on display if no file is given. The extension of
		the file determines the format which must be one of the following according to matplotlib:
		   eps, jpeg, jpg, pdf, pgf, png, ps, raw, rgba, svg, svgz, tif, tiff
		'''
		
		import matplotlib.pyplot as plt
		
		vFig = plt.figure()
		vAxe =vFig.add_subplot(111)
		vAxe.plot(self.features[pFeature].getValues(), self.getTargets(), 'bo')
		
		if pOutFilename:
			vFig.savefig(pOutFilename)
		else:
			vFig.show()
		
	
	
	def shuffled(self):
		'''
		Returns a new dataset with shuffled instances
		'''
		
	
	
	def chunk(self, pStart = None, pEnd = None):
		'''
		Returns a chunk of the dataset from the specified range
		'''
		
	
	


class Features:
	'''
	Class to implement and interface the dataset features (Feature objects)
	
	This class enables accessing the features by name as in dictionaries
	and	by index as in lists, but starting with 1.
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		# list of features of type Feature
		self._features = []
		
	
	
	def _getFeatureByName(self, pFeatureName):
		'''
		Returns the feature having the given name
		'''
		
		for vFeature in self._features:
			if vFeature.name == pFeatureName:
				return vFeature
		
	
	
	def __len__(self):
		'''
		Returns the length of feature vector
		'''
		
		return len(self._features)
		
	
	
	def __getitem__(self, pKey):
		'''
		Returns the feature associated with the given key
		
		If the key is integer, the pKeyth feature is returned. If the key 
		is string, the feature with the name pKey is returned.
		'''
		
		if type(pKey) == int:
			if pKey == 0 or pKey > len(self):
				raise IndexError("Index out of range! Valid indexes are in [1, %s] range." % len(self))
			else:
				return self._features[pKey - 1]
		elif type(pKey) == str:
			return self._getFeatureByName(pKey)
		
	
	
	## We don't implement __setitem__() because it is hardly needed to
	## assign a feature object at a specific location.
	##def __setitem__(self, pKey, pValue):
	
	
	def append(self, pFeature):
		'''
		Appends a new feature to the feature list
		
		pFeature is a Feature object.
		'''
		
		# sanity check: seems to conflict when the feature values are not given yet; most probbaly not needed and caution should be take elsewhere.
		#if len(self._features) > 0 and pFeature.length != self._features[0].length:
		#	raise Exception("New feature length does not match the current dataset size: %s vs. %s" % (pFeature.length, self._features[0].length))
		#else:

		self._features.append(pFeature)
		
	
	
	def add(self, plFeatures):
		'''
		Adds a list of new features
		'''
		
		for vFeature in plFeatures:
			self.append(vFeature)
		
	
	
	def __delitem__(self, pKey):
		'''
		Deletes the feature associated with the given key
		
		If the key is integer, the pKeyth feature is deleted. If the key 
		is string, the feature with the name pKey is deleted.
		'''
		
		if type(pKey) == int:
			del self._features[pKey - 1]
		elif type(pKey) == str:
			for i, vFeature in enumerate(self._features):
				if vFeature.name == pKey:
					del self._features[i]
					break
		
	
	
	def __iter__(self):
		'''
		Returns the feature list as feature vector iterator
		'''
		
		return iter(self._features)
		
	
	
	def __contains__(self, pFeatureName):
		'''
		Returns true if a feature with the given name is found in the vector
		'''
		
		for i, vFeature in enumerate(self._features):
			if vFeature.name == pFeatureName:
				return True
		
		return False
		
	
	


class Feature:
	'''
	Class for storing feature and its values
	
	A feature object contains the values of all instances for the corresponding
	feature.
	'''
	
	
	def __init__(self, pName = '', pDataType = type(None)):
		'''
		Constructor
		
		pType is a python data type which specifies the data type expected.
		If not given, it will be set once the first value is added.
		'''
		
		self.name = pName
		self.dataType = pDataType
		self.values = []
		
	
	
	def setDataType(self, pType):
		'''
		Sets the data type of the feature
		'''
		
		self.dataType = pType
		
	
	
	def loadValues(self, plValues):
		'''
		Loads given values into the feature data
		
		All previous values wll be removed.
		'''
		
		self.values = []
		
		for vValue in plValues:
			self.addValue(vValue)
		
	
	
	def getValues(self):
		'''
		Returns the feature values in a list
		'''
		
		return self.values
		
	
	
	@property
	def length(self):
		'''
		Returns the number of values for the feature as its length
		'''
		
		return len(self.values)
		
	
	
	def addValue(self, pValue):
		'''
		Adds the given value to the feature values
		'''
		
		if self.dataType == type(None):       # equivalent to length == 0
			self.setDataType(pType = type(pValue))
			self.values.append(pValue)
		else:
			if type(pValue) != self.dataType:
				if self.dataType is float and type(pValue) is int:             # an exception for convenience
					self.values.append(float(pValue))
				else:
					raise Exception("Invalid feature value type: %s is expected but %s is given!" % (self.dataType, type(pValue)))
			else:
				self.values.append(pValue)
		
	
	
	def addRawValue(self, pRawValue):
		'''
		Adds a raw value to the feature values after adapting to the proper
		type
		'''
		
		if pRawValue == None:
			self.addNullValue()
		else:
			if self.dataType:
				try:
					self.addValue(self.dataType(pRawValue))
				except:
					raise Exception("Invalid feature value type: %s is expected but %s is given!" % (self.dataType, type(pRawValue)))
			else:
				self.addValue(pValue = self._adaptFeatureValue(pRawValue = pRawValue))
		
	
	
	def addNullValue(self, pDataType = type(None)):
		'''
		Add a null value to the feature values
		
		pDataType specifies the type of null value to be used which will
		only be used if the feature type has not already been set. For 
		string features, a null value is an empty string and for numerical
		features it is 0.
		
		If none of the feature type or the data type parameters have been
		set, 0 will be used.
		'''
		
		if self.dataType == int or pDataType == int:
			self.addValue(0)
		elif self.dataType == float or pDataType == float:
			self.addValue(0.0)
		elif self.dataType == str or pDataType == str:
			self.addValue('')
		elif self.dataType == type(None) and pDataType == type(None):
			self.addValue(0)
		
	
	
	def addNullValues(self, pCount, pDataType):
		'''
		Add null values to the feature values as many as requested
		
		pDataType specifies the type of null value to be used which will
		only be used if the feature type has not already been set. For 
		string features, a null value is an empty string and for numerical
		features it is 0.
		'''
		for i in range(pCount):
			self.addNullValue(pDataType = pDataType)
		
	
	
	def _adaptFeatureValue(self, pRawValue):
		'''
		Converts the raw feature value to the correct data type and returns 
		the result
		
		The data type is determined based on the representation of the raw
		feature values.
		'''
		
		return util.findDataType(pRawValue)(pRawValue)
		
	
	
	def delValue(self, pIndex):
		'''
		Deletes the pIndexth feature value
		'''
		
		del self.values[pIndex - 1]
		
	
	


class IDataInstances:
	'''
	Class to implement an interface to data instances as a list
	'''
	
	
	def __init__(self, pDataset):
		'''
		Constructor
		'''
		
		self._dataset = pDataset
		
		self._currentIdx = 0      # iterator index track
		
	
	
	def __len__(self):
		'''
		Returns the number of instances
		
		If there is a discrepancy of 1 between the number of targets and
		the size of any feature, the smaller number will be returned assuming
		that the last instance is in process of being added. If there is 
		a discrepancy of bigger than 1, it will be assumed as inconsistency
		and cause an exception. 
		'''
		
		vTargetCounts = len(self._dataset.getTargets())
		
		vlFeatureLens = [f.length for f in self._dataset.features]
		if vlFeatureLens == []:
			vMinFLen = 0
			vMaxFLen = 0
		else:
			vMinFLen = min(vlFeatureLens)
			vMaxFLen = max(vlFeatureLens)
		
		if vMinFLen < vMaxFLen - 1:
			raise Exception("Bug suspicion: inconsistency in the feature lengths while calculating dataset size!")
		elif vMinFLen == vMaxFLen:
			if not (vMinFLen - 1 <= vTargetCounts <= vMaxFLen + 1):     # e.g. vMinFLen = 3, vMaxFLen = 3 -> vTargetCounts must be in [2,3,4]
				raise Exception("Bug suspicion: inconsistency in the feature lengths and target numbers while calculating dataset size!")
		elif vMinFLen == vMaxFLen - 1:
			if not (vTargetCounts == vMinFLen or vTargetCounts == vMaxFLen):     # e.g. vMinFLen = 2, vMaxFLen = 3 -> vTargetCounts must be 2 or 3
				raise Exception("Bug suspicion: inconsistency in the feature lengths and target numbers while calculating dataset size!")
		
		return min(vTargetCounts, vMinFLen, vMaxFLen)
		
	
	
	def __getitem__(self, pIndex):
		'''
		Returns the pIndexth instance in the data set
		'''
		
		if pIndex == 0 or pIndex > len(self):
			raise IndexError("Index out of range! Valid indexes are in [1, %s] range." % len(self))
		else:
			return DataInstance(self._getTargetAt(pIndex), self._getFVAt(pIndex))
		
	
	
	def _getTargetAt(self, pIndex):
		'''
		Returns the target value of the pIndexth instance
		'''
		
		return self._dataset.getTargets()[pIndex - 1]
		
	
	
	def _getFVAt(self, pIndex):
		'''
		Return the feature vector of the pIndexth instance
		'''
		
		vFV = FeatureVector()
		vFV.loadFromDataset(self._dataset, pIndex)
		
		return vFV
		
	
	
	def __setitem__(self, pIndex, pDataInstance):
		'''
		Sets the pIndexth data instance to the given data instance object
		'''
		
		if pIndex == 0 or pIndex > len(self):
			raise IndexError("Index out of range! Valid indexes are in [1, %s] range." % len(self))
		else:
			self.replaceInstance(pIndex = pIndex, pNewInstance = pDataInstance)
		
	
	
	def replaceInstance(self, pIndex, pNewInstance):
		'''
		Replaces the pIndexth instance with a new one
		
		pInstance is a DataInstance object, which should match the dataset
		structure. 
		'''
		
		if self._dataset.isFixedFS():
			if not self._ifFVMatch(pNewInstance.fv):
				raise Exception("The feature vector in the given data instance does not match the dataset feature structure!")
		
		# 1. replacing the features
		self._replaceFeatureVector(pIndex, pInstance.fv)
		
		# 2. replacing the target(s)
		self._repalceTarget(pIndex, pInstance.target)
		
	
	
	def _replaceFeatureVector(self, pIndex, pFV):
		'''
		Replaces the feature vector of the pIndexth instance with the given
		feature vector
		'''
		
		for vFeature in pFV:
			self._dataset.features[vFeature.name].values[pIndex] = vFeature.value
		
	
	
	def _repalceTarget(self, pIndex, pTarget):
		'''
		Replaces the target value of the pIndexth instance with the given
		feature vector
		'''
		
		self._targets[pIndex] = pTarget
		
	
	
	def append(self, pInstance):
		'''
		Appends a new instance
		
		pInstance is a DataInstance object, which should match the dataset
		structure if this is not the first instance or the dataset has a
		fixed structure. 
		
		NOTE: If the given instance does not neccessarily match the current
		structure, addInstance() must be used, which adapts the structure.
		'''
		
		if len(self) == 0:
			self.addInstance(pInstance)
		else:
			if not self._ifFVMatch(pInstance.fv):
				raise Exception("The feature vector in the given data instance does not match the dataset feature structure!")
			else:
				self.addInstance(pInstance)
		
	
	
	def addInstance(self, pInstance):
		'''
		Adds a new instance to dataset
		
		If the dataset does not have a fixed structure, it tries to adapt
		the feature structure of the instance to that of the dataset. However,
		the target type must match.
		'''
		
		if self._dataset.isFixedFS():
			if not self._ifFVMatch(pInstance.fv):
				raise Exception("The feature vector in the given data instance does not match the dataset feature structure!")

		# 1. adding the features
		self._appendFeatureVector(pInstance.fv)
		
		# 2. adding the target(s)
		self._appendTarget(pInstance.target)
		
	
	
	def _ifFVMatch(self, pFV):
		'''
		Returns true if the input feature vector matches the feature strcuture 
		of the dataset
		
		Number of features, their order and name are checked to be identical.
		'''
		
		if len(pFV) == len(self._dataset.features):
			for vFVFeaure, vDatasetFeature in zip(pFV, self._dataset.features):
				if vFVFeaure.name != vDatasetFeature.name or vFVFeaure.type != vDatasetFeature.type:
					return False
			
			return True
		else:
			return False
		
	
	
	def _appendFeatureVector(self, pFV):
		'''
		Adds the given feature vector to the dataset
		
		The feature values in the vector will be assumed raw and processed
		to match the data type of the corresponding features in the data
		set. If a feature in the data set has no data type assigned yet,
		the data type of the given feature will be used. 
		'''
		
		vDatasetSize = len(self)

		for vFeature in pFV:
			# if the feature exists
			if vFeature.name in self._dataset.getFeatureNames():
				self._dataset.features[vFeature.name].addRawValue(vFeature.value)
				
			# if the feature does not exist yet
			else:
				self._dataset.addNewFeature(pFeatureName = vFeature.name, pDataType = vFeature.dataType)
				if vDatasetSize > 0:
					self._dataset.features[vFeature.name].addNullValues(pCount = vDatasetSize, pDataType = vFeature.dataType)
				self._dataset.features[vFeature.name].addRawValue(vFeature.value)
		
		# if no value is given for a feature 
		for vFeature in self._dataset.features:
			if vFeature.name not in pFV.getFeatureNames():
				vFeature.addNullValue()
		
		# sanity check
		for vFeature in self._dataset.features:
			if vFeature.length != vDatasetSize + 1:
				raise Exception("The number of values in feature %s does not seem to be correct!" % vFeature.name)
		
	
	
	def _appendTarget(self, pTarget):
		'''
		Adds the target to the dataset
		
		NOTE: Only for use within _addInstance() and where a whole instance
		is added including feature values.
		
		It processes the raw target value(s) and converts them to proper
		type based on dataset settings.
		
		Multiple target values per instance for multi-label problems is
		supported. The value passed in pTarget must be a list in such cases.
		'''
		
		# adapting the target value type(s)
		vTarget = self._adaptTargetValue(pTarget)
		
		# adding
		self._dataset._targets.append(vTarget)
		
	
	
	def _adaptTargetValue(self, pValue):
		'''
		Converts and returns the given target value to the correct type
		
		If the input value is a list, the dataset target type must also 
		be a list, or an exception will be raised.
		'''
		
		# None value is not adapted 
		if pValue is None:
			return None
		
		if type(pValue) == type(self._dataset.getTargetType()) is list:
			vlAdaptedValues = []
			
			if len(pValue) != len(self._dataset.getTargetType()):
				raise Expection("The number of multi-value target values does not match the dataset target type: %s vs. %s" % (pValue, self._dataset.getTargetType()))
			
			for vValue, vType in zip(pValue, self._dataset.getTargetType()):
				try:
					vAdaptedValue = vType(vValue)
				except ValueError:
					raise Exception("Cannot convert %s to %s!" % (type(vValue), vType))
				
				vlAdaptedValues.append(vAdaptedValue)
			
			return vlAdaptedValues
			
		elif type(pValue) is list and type(self._dataset.getTargetType()) is not list:
			raise Exception("The target type of the dataset has not been set to be a list; the given target values are a list!")
		elif type(pValue) is not list and type(self._dataset.getTargetType()) is list:
			raise Exception("The target type of the dataset has been set to be a list; the given value is %s!" % type(pValue))
		else:
			try:
				vAdaptedValue = self._dataset.getTargetType()(pValue)
			except ValueError:
				raise Exception("Cannot convert %s to %s!" % (type(pValue), self._dataset.getTargetType()))
			
			return vAdaptedValue
		
	
	
	def __delitem__(self, pIndex):
		'''
		Deletes the pIndexth instance
		'''
		
		for vFeature in self._dataset.features:
			vFeature.delValue(pIndex)
		
	
	
	def __iter__(self):
		'''
		Iterator method
		
		See python iterator objects
		'''
		
		return self
		
	
	
	def next(self):
		'''
		Next method
		
		See python iterator objects
		'''
		
		if self._currentIdx < len(self):
			self._currentIdx += 1
			return DataInstance(self._getTargetAt(self._currentIdx), self._getFVAt(self._currentIdx))
		else:
			raise StopIteration
		
	
	
	def __contains__(self, pInstance):
		'''
		Returns true if an exact instance is found in the dataset
		'''
		
		# first, collecting indexes of instances with matching targets
		
		vlTMIIdxs = []
		
		for i, vTarget in enumerate(self._dataset.getTargets(), start = 1):
			if vTarget == pInstance.target:
				vlTMIIdxs.append(i)
		
		# second, going through target-matched instances for feature vector comparison
		
		for i in vlTMIIdxs:
			if pInstance.fv.isEqual(self[i].fv):
				return True
		
		return False
		
	
	


class DataInstance:
	'''
	Class for data instance
	'''
	
	
	def __init__(self, pTarget = None, pFV = None):
		'''
		Constructor
		'''
		
		self._target = pTarget
		# feature vector of type FeatureVector
		self._fv = pFV
		
	
	
	def setTarget(self, pTarget):
		'''
		Sets the value of the target
		'''
		
		self._target = pTarget
		
	
	
	def setFeatureVector(self, pFeatureVector):
		'''
		Sets the value of the feature vector to the given FeatureVector
		object
		'''
		
		self._fv = pFeatureVector
		
	
	
	@property
	def target(self):
		'''
		Returns the target string of the instance
		'''
		
		return self._target
		
	
	
	@property
	def fv(self):
		'''
		Returns the instance feature vector which is a FeatureVector object
		'''
		
		return self._fv
		
	
	
	@property
	def features(self):
		'''
		Returns the instance feature vector which is a FeatureVector object
		'''
		
		return self._fv
		
	
	
	def __str__(self):
		'''
		String representation of the instance
		'''
		
		return self.getStr()
		
	
	
	def getStr(self, pFormat = "tn:v"):
		'''
		Returns the string representation of the data instance in the specified format
		
		The format string uses the following coding:
		- t: target value
		- n: feature name
		- i: feature index
		- v: feature value
		
		All other characters (e.g. delimiters) are written verbatim in the output. Note that, if no delimiter is specified
		in the format between target value and feature vector, tab is used. Also, if no delimiter is specified between
		features in the feature vector, single space is used.
		'''
		
		vFormat = pFormat.lower()
		
		if vFormat in ["tn:v"]:
			return "%s\t%s" % (self.target, self.fv.getStr("n:v"))
		elif vFormat in ["ti:v"]:
			return "%s\t%s" % (self.target, self.fv.getStr("i:v"))
		
	
	
	def getFeatureValueVector(self):
		'''
		Returns the features value vector in a list in the original order
		'''
		
		return [f.value for f in self.fv]
		
	
	


class FeatureVector:
	'''
	Class to implement data instance feature vector
	
	This class enables accessing the features by name as in dictionaries
	and by index as in lists, but starting with 1.
	'''
	
	
	def __init__(self, plFeatureValuePairs = [], pflgAutoDataType = True):
		'''
		Constructor
		
		plFeatureValuePairs is a list of feature name-value pairs. Though
		a dictionary may appear more suitable, a list is used to retain
		the feature order.
		'''
		
		# list of InstanceFeature objects
		self._features = self.load(plFeatureValuePairs, pflgAutoDataType)
		
	
	
	def load(self, plFeatureValuePairs, pflgAutoDataType = True):
		'''
		Loads the feature vector from a list of feature name-value pairs
		
		Though a dictionary may appear more suitable, a list is used to 
		retain the feature order.
		
		The feature data type can optionally be identified automatically,
		if the feature values are raw string values. For example, for the
		feature value of "123" and "test", an integer and an string type
		will be used respectively. Otherwise, the data type of the value
		will be used. For example, for the feature value of 123 and "123",
		an integer and an string type will be used respectively.
		
		If a feature value is None, 0 and integer will be used for the value
		and data type respectively.
		'''
		
		vlFeatures = []
		
		for vName, vRawValue in plFeatureValuePairs:
			if vRawValue is None:
				vValue = 0
				vDataType = int
			elif pflgAutoDataType:
				vDataType = util.findDataType(vRawValue)
				vValue = vDataType(vRawValue)
			else:
				vDataType = type(vRawValue)
				vValue = vRawValue
			
			vFeature = InstanceFeature(pName = vName, pDataType = vDataType, pValue = vValue)
			
			vlFeatures.append(vFeature)
		
		return vlFeatures
		
	
	
	def loadFromDataset(self, pDataset, pIndex):
		'''
		Loads the feature vector at the given index in the given dataset
		
		pIndex is the 1-based index of the data instance in the dataset
		'''
		
		self._features = []
		
		for vFeature in pDataset.features:
			self._features.append(InstanceFeature(pName = vFeature.name, pDataType = vFeature.dataType, pValue = vFeature.values[pIndex - 1]))
		
	
	
	def __str__(self):
		'''
		String representation of the feature vector
		'''
		
		return self.getStr()
		
	
	
	def getStr(self, pFormat = "n:v"):
		'''
		Returns the string representation of the feature vector in the specified format
		
		The format string uses the following coding:
		- n: feature name
		- i: feature index
		- v: feature value
		
		All other characters (e.g. delimiters) are written verbatim in the output.
		'''
		
		vFormat = pFormat.lower()
		
		if vFormat == "n:v":
			return ' '.join(["%s:%s" % (f.name, f.value) for f in self])
		elif vFormat == "i:v":
			return ' '.join(["%s:%s" % (i, f.value) for i, f in enumerate(self, start  = 1)])
		
	
	
	def _getFeatureByName(self, pFeatureName):
		'''
		Returns the feature having the given name
		'''
		
		for vFeature in self._features:
			if vFeature.name == pFeatureName:
				return vFeature
		
	
	
	def __len__(self):
		'''
		Returns the length of feature vector
		'''
		
		return len(self._features)
		
	
	
	def __getitem__(self, pKey):
		'''
		Returns the feature associated with the given key
		
		If the key is integer, the pKeyth feature is returned. If the key 
		is string, the feature with the name pKey is returned.
		'''
		
		if type(pKey) == int:
			if pKey == 0 or pKey > len(self):
				raise IndexError("Index out of range! Valid indexes are in [1, %s] range." % len(self))
			else:
				return self._features[pKey - 1]
		elif type(pKey) == str:
			return self._getFeatureByName(pKey)
		
	
	
	## We don't implement __setitem__() because it is hardly needed to
	## assign a feature object at a specific location.
	##def __setitem__(self, pKey, pValue):
	
	
	def append(self, pFeature):
		'''
		Appends a new feature to the feature list
		
		pFeature is a Feature object.
		'''
		
		self._features.append(pFeature)
		
	
	
	def __delitem__(self, pKey):
		'''
		Deletes the feature associated with the given key
		
		If the key is integer, the pKeyth feature is deleted. If the key 
		is string, the feature with the name pKey is deleted.
		'''
		
		if type(pKey) == int:
			del self._features[pKey - 1]
		elif type(pKey) == str:
			for i, vFeature in enumerate(self._features):
				if vFeature.name == pKey:
					del self._features[i]
					break
		
	
	
	def __iter__(self):
		'''
		Returns the feature list as feature vector iterator
		'''
		
		return iter(self._features)
		
	
	
	def __contains__(self, pFeatureName):
		'''
		Returns true if a feature with the given name is found in the vector
		'''
		
		for i, vFeature in enumerate(self._features):
			if vFeature.name == pFeatureName:
				return True
		
		return False
		
	
	
	def getFeatureNames(self):
		'''
		Returns names of the feature in the vector
		'''
		
		return [f.name for f in self._features]
		
	
	
	def isEqual(self, pOtherFV):
		'''
		Compares the feature vector with another one and returns true if
		the structure and feature values are exactly the same
		
		The order of the features in the vector does not matter
		'''
		
		if len(pOtherFV) == len(self):
			for vFeature in pOtherFV:
				if not (vFeature.name in self and vFeature.dataType == self[vFeature.name].dataType and vFeature.value == self[vFeature.name].value):
					return False
			
			return True
		else:
			return False
		
	
	


class InstanceFeature:
	'''
	Class for storing a feature of a data instance feature vector and its value
	'''
	
	
	def __init__(self, pName = '', pDataType = type(None), pValue = None):
		'''
		Constructor
		
		pDataType is a python data type which specifies the data type expected.
		'''
		
		self.name = pName
		self.setDataType(pDataType)
		self.setValue(pValue)
		
	
	
	def setDataType(self, pType):
		'''
		Sets the data type of the feature
		'''
		
		self.dataType = pType
		
	
	
	def setValue(self, pValue):
		'''
		Sets the value of the feature to the given value
		'''
		
		if type(pValue) != self.dataType:
			raise Exception("Invalid feature value type: %s is expected but %s is given!" % (self.dataType, type(pValue)))
		else:
			self.value = pValue
		
	
	
	def __str__(self):
		'''
		String representation of the feature
		'''
		
		return self.getStr()
		
	
	
	def getStr(self, pFormat = 'n:v'):
		'''
		Returns the string representation of the feature in the specified format
		'''
		
		vFormat = pFormat.lower()
		
		if vFormat == "n:v":
			return "%s:%s" % (self.name, self.value)
		
	
	

