#! /usr/bin/python
# -*- coding: utf-8 -*-

"""	
	This module defines general-purpose utility functions.
	
	Version 2.2												(08-Jan-2017)
	- IOEval is added.
	
	Version 2.1												(11-Jul-2017)
	- extractExactMismatches() os added to BIOEval.
	
	Version 2.0												(27-Mar-2017)
	- bioTagMap is added to BIOEval.
	
	Version 1.9												(15-Nov-2016)
	- _fixPredSpanStart() is added to BIOEval.
	
	Version 1.8												(07-Apr-2016)
	- PredComp.report() now reports the oracle combination performance. 
	
	Version 1.7												(19-Jan-2016)
	- confusionCases() is added to PredEval.
	
	Version 1.6												(19-Jan-2016)
	- BIOEval is added.
	
	Version 1.5                             (16-Nov-2015 to 18-Nov-2015)
	- plot() is added to PredEval.
	
	Version 1.4								(01-Oct-2015 to 30-Oct-2015)
	- PredComp is added for comparing predictions of systems.
	
	Version 1.3												(21-Jul-2015)
	- Bug in precision() and fscore() is fixed to avoid division by zero.
	
	Version 1.2												(14-Jul-2015)
	- precision(), recall() and fscore() are added to PredEval.
	
	Version 1.1												(26-Jun-2015)
	- confusionMatrix() and confusionMatrixToText() are added to PredEval.
	
	Version 1.0												(15-May-2015)
	- PredEval, PredStatSig, TTest, IAA are moved from util/util.py.
	
"""

import math

#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# PredEval
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class PredEval:
	'''
	Class for scoring a prediction using various metrics
	'''
	
	
	def __init__(self, plPrediction = None, plReference = None):
		'''
		Constructor
		'''
		
		if plPrediction is None:
			self.prediction = []
		else:
			self.prediction = plPrediction
		
		if plReference is None:
			self.reference = []
		else:
			self.reference = plReference
		
	
	
	def acc(self, plPrediction = None, plReference = None, pType = "exact", pRDigits = 0, pErrMargin = 0, pErrMarginRatio = 0):
		'''
		Computes accuracy based on the options specified
		
		Accuracy is the percentage of the correctly predicted instances.
		However, the correctness is defined by pType:
		- exact: predicted value must be equal to reference value
		- round: predicted and reference values rounded to pRDigits must be equal
		- absolute-err-margin: predicted value must be between 
		                       reference value - pErrMargin and 
		                       reference value + pErrMargin
		- ratio-err-margin: predicted value must be between 
		                    reference value - reference value * pErrMarginRatio and
		                    reference value + reference value * pErrMarginRatio
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		vType = pType.lower()
		
		vCorrectNum = 0
		for vPredScore, vRefScore in zip(vlPrediction, vlReference):
			# exact match
			if vType == "exact":
				if vPredScore == vRefScore:
					vCorrectNum += 1
			# round match
			elif vType == "round":
				if round(vPredScore, pRDigits) == round(vRefScore, pRDigits):
					vCorrectNum += 1
			# absolute error margin match 
			elif vType.startswith("abs"):
				if vRefScore - pErrMargin <= vPredScore <= vRefScore + pErrMargin:
					vCorrectNum += 1
			# ratio error margin match 
			elif vType.startswith("ratio"):
				vErrMargin = vRefScore * pErrMarginRatio
				if vRefScore - vErrMargin <= vPredScore <= vRefScore + vErrMargin:
					vCorrectNum += 1
		
		return (vCorrectNum * 100.00) / len(vlPrediction)
		
	
	
	def mae(self, plPrediction = None, plReference = None):
		'''
		Computes MAE
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		vABSDiffSum = 0.0
		for vPredScore, vRefScore in zip(vlPrediction, vlReference):
			vABSDiffSum += abs(vPredScore - vRefScore)
		
		return vABSDiffSum / len(vlPrediction)
		
	
	
	def rmse(self, plPrediction = None, plReference = None):
		'''
		Computes RMSE
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		vDiff2Sum = 0.0
		for vPredScore, vRefScore in zip(vlPrediction, vlReference):
			vDiff2Sum += ((vPredScore - vRefScore) ** 2)
		
		return math.sqrt(vDiff2Sum / len(vlPrediction))
		
	
	
	def r(self, plPrediction = None, plReference = None):
		'''
		Computes Pearson correlation coefficient r
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		vXSum = vYSum = vX2Sum = vY2Sum = vXYSum = 0.0
		vN = len(vlPrediction)
		for vPredScore, vRefScore in zip(vlPrediction, vlReference):
			vXSum += vPredScore
			vYSum += vRefScore
			vX2Sum += vPredScore**2
			vY2Sum += vRefScore**2
			vXYSum += vPredScore * vRefScore
		
		# Just as an example sometimes in Python: 54756.0 - 54756.0 = -2.1027517505e-09.
		## That's why sqrt operand is absolute.
		return ((vN * vXYSum) - (vXSum * vYSum)) / math.sqrt(abs(vN * vX2Sum - vXSum**2) * abs(vN * vY2Sum - vYSum**2))
		
	
	
	def range(self, plPrediction = None, plReference = None):
		'''
		Computes the range of the prediction and its ratio to the range 
		of the reference scores.
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		vPredRange = (min(vlPrediction), max(vlPrediction))
		vRefRange = (min(vlReference), max(vlReference))
		
		vRangeRatio = (vPredRange[1] - vPredRange[0]) * 1.0 / (vRefRange[1] - vRefRange[0])
		
		return vRangeRatio, vPredRange, vRefRange
		
	
	
	def sd(self, plPrediction = None, plReference = None):
		'''
		Computes the standard deviation of the prediction and its ratio
		to that of the reference scores.
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		vDataSize = len(vlPrediction)
		
		if vDataSize != len(vlReference):
			raise Exception("Prediction and reference data set size mismatch: %s vs. %s" % (vDataSize, len(vlReference)))
		
		vPredAvg = sum(vlPrediction) * 1.0 / vDataSize
		vPredVar = sum([(score - vPredAvg)**2 for score in vlPrediction]) / vDataSize
		vPredSD = math.sqrt(vPredVar)
		
		vRefAvg = sum(vlReference) * 1.0 / vDataSize
		vRefVar = sum([(score - vRefAvg)**2 for score in vlReference]) / vDataSize
		vRefSD = math.sqrt(vRefVar)
		
		return vPredSD / vRefSD, vPredSD, vRefSD
		
	
	
	def precision(self, plPrediction = None, plReference = None):
		'''
		Calculates the precision of the predictions for each class as well
		as macro and micro average precision
		
		The returning values are in two dictionaries one for per-class 
		scores and one for average scores.
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		# The calculation is done with the help of confusion matrix
		vdConfMatrix = self.confusionMatrix(plPrediction = vlPrediction, plReference = vlReference)
		
		# initializing returning dictionaries
		vdClassPrec = {c: None for c in vdConfMatrix}
		vdAvgPrec = {"macro": None, "micro": None}
		
		# calculating class precisions and taking note for micro-average 
		vTPSum = 0
		vFPSum = 0
		for vClass in vdClassPrec:
			vTP = vdConfMatrix[vClass][vClass]
			vFP = sum([dic[vClass] for c, dic in vdConfMatrix.iteritems() if c != vClass])
			
			vAllP = vTP + vFP
			if vAllP != 0:
				vdClassPrec[vClass] = (100.0 * vTP) / (vAllP)
			else:
				vdClassPrec[vClass] = 0.0
			
			vTPSum += vTP
			vFPSum += vFP
		
		# calculating micro precision
		vAllPSum = vTPSum + vFPSum
		if vAllPSum != 0:
			vdAvgPrec["micro"] = (100.0 * vTPSum) / (vAllPSum)
		else:
			vdAvgPrec["micro"] = 0.0
		
		# calculating macro precision
		vdAvgPrec["macro"] = sum([vdClassPrec[c] for c in vdClassPrec]) / len(vdClassPrec)
		
		return vdClassPrec, vdAvgPrec
		
	
	
	def recall(self, plPrediction = None, plReference = None):
		'''
		Calculates the recall of the predictions for each class as well
		as macro and micro average recall
		
		The returning values are in two dictionaries one for per-class 
		scores and one for average scores.
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		# The calculation is done with the help of confusion matrix
		vdConfMatrix = self.confusionMatrix(plPrediction = vlPrediction, plReference = vlReference)
		
		# initializing returning dictionaries
		vdClassRec = {c: None for c in vdConfMatrix}
		vdAvgRec = {"macro": None, "micro": None}
		
		# calculating class recalls and taking note for micro-average 
		vTPSum = 0
		vFNSum = 0
		for vClass in vdClassRec:
			vTP = vdConfMatrix[vClass][vClass]
			vFN = sum([count for c, count in vdConfMatrix[vClass].iteritems() if c != vClass])
			vdClassRec[vClass] = (100.0 * vTP) / (vTP + vFN)
			
			vTPSum += vTP
			vFNSum += vFN
		
		# calculating micro recall
		vdAvgRec["micro"] = (100.0 * vTPSum) / (vTPSum + vFNSum)
		
		# calculating macro recall
		vdAvgRec["macro"] = sum([vdClassRec[c] for c in vdClassRec]) / len(vdClassRec)
		
		return vdClassRec, vdAvgRec
		
	
	
	def fscore(self, plPrediction = None, plReference = None, pBeta = 1):
		'''
		Calculates the F-score of the predictions for each class as well
		as macro and micro average score
		
		The pBeta values specifies the weight assigned to recall in the
		general F-score formula:
		
		F-score = (1 + Beta^2).(Precision.Recall) / ((Beta^2).Precision + Recall)
		
		The returning values are in two dictionaries one for per-class 
		scores and one for average scores.
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		# calculating precision and recall scores
		vdClassPrec, vdAvgPrec = self.precision(plPrediction = vlPrediction, plReference = vlReference)
		vdClassRec,  vdAvgRec  = self.recall(plPrediction = vlPrediction, plReference = vlReference)
		
		# initializing returning dictionaries
		vdClassFscore = {c: None for c in vdClassPrec}
		vdAvgFscore = {"macro": None, "micro": None}
		
		# calculating class F-scores
		for vClass in vdClassFscore:
			if vdClassPrec[vClass] + vdClassRec[vClass] != 0:
				vdClassFscore[vClass] = (1 + pBeta**2) * (vdClassPrec[vClass] * vdClassRec[vClass]) / ((pBeta**2) * vdClassPrec[vClass] + vdClassRec[vClass])
			else:
				vdClassFscore[vClass] = 0
		
		# calculating micro F-scores
		vdAvgFscore["micro"] = (1 + pBeta**2) * (vdAvgPrec["micro"] * vdAvgRec["micro"]) / ((pBeta**2) * vdAvgPrec["micro"] + vdAvgRec["micro"])
		
		# calculating macro F-scores
		vdAvgFscore["macro"] = sum([vdClassFscore[c] for c in vdClassFscore]) / len(vdClassFscore)
		
		return vdClassFscore, vdAvgFscore
		
	
	
	def confusionMatrix(self, plPrediction = None, plReference = None):
		'''
		Calculates the confusion matrix based on the predictions and the
		gold (reference) class labels
		
		The matrix is returned in a two-dimentional dictionary where the
		rows represent the gold classes and the columns the predicted classes.
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		vsRefClasses = set(vlReference)
		
		vdConfMatrix = {}
		
		## initializing the matrix:
		## only relying on the classes labels in the gold data because a
		## class may have never been predicted so missing from the predicted 
		## classes
		for vRefClassInRow in vsRefClasses:
			vdConfMatrix[vRefClassInRow] = {}
			for vRefClassInCol in vsRefClasses:
				vdConfMatrix[vRefClassInRow][vRefClassInCol] = 0
		
		# calculating the matrix
		for vRefClass, vPredClass in zip(vlReference, vlPrediction):
			vdConfMatrix[vRefClass][vPredClass] += 1
		
		return vdConfMatrix
		
	
	
	def confusionCases(self, plPrediction = None, plReference = None):
		'''
		Extracts the confusion cases based on the predictions and the
		gold (reference) class labels
		
		The matrix is returned in a tree-dimentional dictionary where the
		rows represent the gold classes and the columns the predicted classes,
		while the third dimension contains the list of positions of confusion 
		cases.
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		vsRefClasses = set(vlReference)
		
		vdConfMatrix = {}
		
		## initializing the matrix:
		## only relying on the classes labels in the gold data because a
		## class may have never been predicted so missing from the predicted 
		## classes
		for vRefClassInRow in vsRefClasses:
			vdConfMatrix[vRefClassInRow] = {}
			for vRefClassInCol in vsRefClasses:
				vdConfMatrix[vRefClassInRow][vRefClassInCol] = []
		
		# calculating the matrix
		for i, (vRefClass, vPredClass) in enumerate(zip(vlReference, vlPrediction), start=1):
			vdConfMatrix[vRefClass][vPredClass].append(i)
		
		return vdConfMatrix
		
	
	
	def confusionMatrixToText(self, pdConfusionMatrix, plCLOrder, pdCLAliases):
		'''
		Converts the confusion matrix to a table-like text
		
		The matrix is supposed to be a two dimentional dictionary output 
		by confusionMatrix().
		
		The order in which the class labels in the matrix will be displayed
		can be provided via pCLOrder.
		
		Class label aliases can be provided to this method to be used instead
		of the original class labels.
		'''
		
		# extracting class labels first
		if plCLOrder != None:
			vlClassLabels = plCLOrder
		else:
			vlClassLabels = [l for l in pdConfusionMatrix.iterkeys()]
		
		# setting up aliases
		if pdCLAliases is None:
			for vCL in vlClassLabels:
				pdCLAliases[vCL] = vCL
		
		# finding the length of longest printable label
		vMaxCLLen = max([len(l) for l in pdCLAliases.itervalues()])
		
		vlTableRows = []
		
		# table header
		vlTableRows.append("%s | %s | SUM " % (''.ljust(vMaxCLLen), 
		                                       ' '.join([pdCLAliases[l] for l in vlClassLabels])))
		vlTableRows.append('-'*(len(vlTableRows[0]) + 3))
		
		# table content
		for vRefCL in vlClassLabels:
			vlTableRows.append("%s | %s | %s" % (pdCLAliases[vRefCL].ljust(vMaxCLLen), 
			                                     ' '.join([str(pdConfusionMatrix[vRefCL][l]).rjust(len(pdCLAliases[l])) for l in vlClassLabels]),
			                                     str(sum([v for v in pdConfusionMatrix[vRefCL].itervalues()])).rjust(5)))
		
		# table footer (sums)
		vlTableRows.append('-'*(len(vlTableRows[1])))
		vlTableRows.append("%s | %s | %s" % ('SUM'.ljust(vMaxCLLen),
		                                     ' '.join([str(sum([dic[pl] for rl, dic in pdConfusionMatrix.iteritems()])).rjust(len(pdCLAliases[pl])) for pl in vlClassLabels]),
		                                     str(sum([sum([v for v in dic.itervalues()]) for dic in pdConfusionMatrix.itervalues()])).rjust(5)))
		
		return '\n'.join(vlTableRows)
		
	
	
	def plot(self, plPrediction = None, plReference = None, pOutFilename = None):
		'''
		Plots the predictions and the gold
		
		The plot will be saved into the given output file or be shown on display if no file is given. The extension of
		the file determines the format which must be one of the following according to matplotlib:
		   eps, jpeg, jpg, pdf, pgf, png, ps, raw, rgba, svg, svgz, tif, tiff
		'''
		
		
		import matplotlib.pyplot as plt
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		vFig = plt.figure()
		vAxe =vFig.add_subplot(111)
		vAxe.plot(range(1, len(vlPrediction) + 1), vlPrediction, 'bo', range(1, len(vlReference) + 1), vlReference, 'gs')
		
		if pOutFilename:
			vFig.savefig(pOutFilename)
		else:
			vFig.show()
		
	
	
	def getCrrPreds(self, plPrediction = None, plReference = None):
		'''
		Returns the (1-based) indexes of correctly predicted instances
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		vlIdxs = []
		for i, (vPredScore, vRefScore) in enumerate(zip(vlPrediction, vlReference), start = 1):
			if vPredScore == vRefScore:
				vlIdxs.append(i)
		
		return vlIdxs
		
	
	
	def getWrongPreds(self, plPrediction = None, plReference = None):
		'''
		Returns the (1-based) indexes of wrongly predicted instances
		'''
		
		if plPrediction is None:
			vlPrediction = self.prediction
		else:
			vlPrediction = plPrediction
		
		if plReference is None:
			vlReference = self.reference
		else:
			vlReference = plReference
		
		vlIdxs = []
		for i, (vPredScore, vRefScore) in enumerate(zip(vlPrediction, vlReference), start = 1):
			if vPredScore != vRefScore:
				vlIdxs.append(i)
		
		return vlIdxs
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# PredComp
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class PredComp:
	'''
	Class for comparing the predictions of systems
	
	NOTE: this class uses "gold" instead of "reference" for variable names
	and is newer than those using the latter. 
	'''
	
	
	def __init__(self, plGold = None):
		'''
		Constructor
		'''
		
		self.gold = plGold
		
		## In the following lists, each element is a dictionary representing the (1-based) index of the instances, the 
		## prediction of system 1 and the prediction of system 2.   
		
		# list of cases predicted correctly by both systems 
		self.bothCorrect = []
		
		# list of cases predicted wrongly by both systems 
		self.bothWrong = []                                                     # both wrong either equally or differently
		self.bothEqualWrong = []                                                # both wrong and equal
		self.bothDiffWrong = []                                                 # both wrong but different
		
		# list of cases predicted equally by both systems (either correct or wrong)
		self.bothEqual = []
		
		# list of cases predicted correctly by system 1 but wrongly by system 2
		self.only1Correct = []
		# list of cases predicted correctly by system 2 but wrongly by system 1
		self.only2Correct = []
		
	
	
	@property
	def sampleSize(self):
		'''
		Returns the size of the sample
		'''
		
		return len(self.gold)
		
	
	
	def compare(self, plPrediction1, plPrediction2):
		'''
		Compares the predictions of two systems with each other and with the gold standard in various aspects
		
		The results of each comparison aspect is stored in its corresponding class attributes in terms of the 1-based 
		indexes of the instances matching the condition of that aspect.
		'''
		
		for i, (vGold, vPred1, vPred2) in enumerate(zip(self.gold, plPrediction1, plPrediction2), start=1):
			if vGold == vPred1 == vPred2:
				self.bothCorrect.append({"idx": i, "sys1": vPred1, "sys2": vPred2})
				self.bothEqual.append({"idx": i, "sys1": vPred1, "sys2": vPred2})
			elif vGold != vPred1 and vGold != vPred2:
				self.bothWrong.append({"idx": i, "sys1": vPred1, "sys2": vPred2})
				if vPred1 == vPred2:
					self.bothEqualWrong.append({"idx": i, "sys1": vPred1, "sys2": vPred2})
					self.bothEqual.append({"idx": i, "sys1": vPred1, "sys2": vPred2})
				else:
					self.bothDiffWrong.append({"idx": i, "sys1": vPred1, "sys2": vPred2})
			elif vGold == vPred1 and vGold != vPred2:
				self.only1Correct.append({"idx": i, "sys1": vPred1, "sys2": vPred2})
			elif vGold != vPred1 and vGold == vPred2:
				self.only2Correct.append({"idx": i, "sys1": vPred1, "sys2": vPred2})
		
	
	
	def report(self):
		'''
		Prints a report on the comparison results
		'''
		
		print "Number of instances            : %3s" % self.sampleSize
		print "System 1\'s accuracy            : %3s (%s%%)" % (len(self.bothCorrect) + len(self.only1Correct), round((len(self.bothCorrect) + len(self.only1Correct))* 100.0 / self.sampleSize, 2))
		print "System 2\'s accuracy            : %3s (%s%%)" % (len(self.bothCorrect) + len(self.only2Correct), round((len(self.bothCorrect) + len(self.only2Correct))* 100.0 / self.sampleSize, 2))
		print ".............................................."
		print "Both systems correct           : %3s (%s%%)" % (len(self.bothCorrect), round(len(self.bothCorrect) * 100.0 / self.sampleSize, 2))
		print "Both systems wrong             : %3s (%s%%)" % (len(self.bothWrong), round(len(self.bothWrong) * 100.0 / self.sampleSize, 2))
		print "Both systems equally wrong     : %3s (%s%%)" % (len(self.bothEqualWrong), round(len(self.bothEqualWrong) * 100.0 / self.sampleSize, 2))
		print "Both systems differently wrong : %3s (%s%%)" % (len(self.bothDiffWrong), round(len(self.bothDiffWrong) * 100.0 / self.sampleSize, 2))
		print "Both systems equal             : %3s (%s%%)" % (len(self.bothEqual), round(len(self.bothEqual) * 100.0 / self.sampleSize, 2))
		print "Only system 1 correct          : %3s (%s%%)" % (len(self.only1Correct), round(len(self.only1Correct) * 100.0 / self.sampleSize, 2))
		print "Only system 2 correct          : %3s (%s%%)" % (len(self.only2Correct), round(len(self.only2Correct) * 100.0 / self.sampleSize, 2))
		print "Oracle combination correct     : %3s (%s%%)" % (self.sampleSize - len(self.bothWrong), round((self.sampleSize - len(self.bothWrong)) * 100.0 / self.sampleSize, 2))
		
	
	
	def getBothCorrectIdxs(self):
		'''
		Returns the indexes of instances predicted correctly by both systems
		'''
		
		return [i["idx"] for i in self.bothCorrect]
		
	
	
	def getBothWrongIdxs(self):
		'''
		Returns the indexes of instances predicted wrongly by both systems
		'''
		
		return [i["idx"] for i in self.bothWrong]
		
	
	
	def getBothEqualWrongIdxs(self):
		'''
		Returns the indexes of instances predicted equally wrong by both systems
		'''
		
		return [i["idx"] for i in self.bothEqualWrong]
		
	
	
	def getBothDiffWrongIdxs(self):
		'''
		Returns the indexes of instances predicted wrongly but in different ways by both systems
		'''
		
		return [i["idx"] for i in self.bothDiffWrong]
		
	
	
	def getBothEqualIdxs(self):
		'''
		Returns the indexes of instances predicted equally by both systems (either correct or wrong)
		'''
		
		return [i["idx"] for i in self.bothEqual]
		
	
	
	def getOnly1CorrectIdxs(self):
		'''
		Returns the indexes of instances predicted correctly by systems 1 and wrongly by system 2
		'''
		
		return [i["idx"] for i in self.only1Correct]
		
	
	
	def getOnly2CorrectIdxs(self):
		'''
		Returns the indexes of instances predicted correctly by systems 2 and wrongly by system 1
		'''
		
		return [i["idx"] for i in self.only2Correct]
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# BIOEval
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class BIOEval:
	'''
	Class for evaluation of BIO sequence labelling 
	
	Note on score calculation method:
	
	The span-based sequence labelling scores can be computed in various ways, such as exact match, where a predicted span
	has to exactly match the gold span, or overlap, where a predicted span only needs to overlap the gold span. Johannson
	& Moschitii (CL, 2013) propose another method to rectify the score inflation caused by the overlap method. Their method
	takes the Cartesian product of the gold and predicted spans (i.e. each element of the product is a tuple of one gold
	and one predicted span, e.g. {((1,4), (3,4)), ((5,7), (3,4))} is the Cartesian product of two gold and one predicted
	span), calculates the span coverage of the gold span by the predicted span in each element (e.g. 2/2 = 1 for ((1,4), (3,4)))
	and sums them. To calculate the recall, this sum is divided by the number of gold spans (e.g. 1 + 0 / 2 = 0.5 for
	the case of above example). To calculated the precision, a similar approach is used, the difference being that the
	span coverage of the predicted span by the gold span is calculated instead of the opposite and the sum is divided at 
	the end by the number of predicted spans rather than the gold spans.
	
	J&M method, however, can also overscore the precision in some circumstances. For one case, if a gold span is overlapped
	by two different predicted spans, the matching labels of both predicted spans will be rewarded. For example, consider
	the following example:
	
	#	G	P
	----------
	1	B	B
	2	I	I
	3	I	B
	4	O	I
	5	O	I
	
	The upper bound precision score will be less than when we consider the prediction of element 3 as I (i.e. correct)
	which leads to precision of 3/5 = 0.6. With J&M method, the precision (and recall for record) will be calculated as:
	
	span coverage of predicted by gold = (2/2) + (1/3) = 4/3
	precision = (4/3) / 2 = 4/6 = 0.66
	
	span coverage of gold by prediction = (2/3) + (1/3) = 1
	recall = 1 / 1 = 1
	
	This precision is bigger than even when the 3rd element's prediction is corrected because its prediction is taken into 
	account twice and the division by 2 at the end could not balance it. One way to overcome this is another method which
	is even more efficient. In this method, the components of precision and recall (TP, FP, TN, FN) are calculated based
	on the prediction of each element, and the precision and recall are calculated based on their outcome values. For the
	case of the example above, we have:
	
	#	G	P		
	-----------------
	1	B	B		TP
	2	I	I		TP
	3	I	B		FN (?)
	4	O	I		FP
	5	O	I		FP
	
	Precision would be: 2 / (2 + 2) = 0.5 and the recall would be: 2 / (2 + 1) = 0.66.
	
	However, there is a complication with this method as well. The score of element 3 is FN in this case because it is
	considered as part of the span [1,3], so I would be the correct label but B is assigned. Therefore, to avoid the same
	pitfall of counting labels of two different predicted span for one gold span as in J&M, it is considered incorrect.
	Now, the problem is that if the prediction was as follows:
	
	#	G	P		
	-----------------
	1	B	O		FN
	2	I	O		FN
	3	I	B		TP (?)
	4	O	I		FP
	5	O	I		FP
	
	Since the gold span has not been taken by any predicted span until element 3, it can be considered a match for the 
	predicted span [3, 5], thus TP. Therefore, this method need to keep track of the matches while scoring each element
	separately.
	
	In this class, we start by implementing J&M to have the means to compare to state-of-the-art. Other methods, including
	our own one described above, may be added.
	'''
	
	
	def __init__(self, plPredSeqs = None, plGoldSeqs = None, pdBIOTagMap = {'B': 'B', 'I': 'I', 'O': 'O'}):
		'''
		Constructor
		
		NOTE: prediction and gold sequence tags are provided in 2D lists, where the first dimension represents sequences
		and the second represents sequence tokens.
		
		The BIO tag map is a dictionary which maps BIO tags to any tag set which is used bu the input data. For example, 
		if the input data (predictions and gold) is using an index set of {0, 1 and 2} instead of "BIO", this argument will 
		look like: {'B': 0, 'I': 1, 'O': 2}
		'''
		
		self.bioTagMap = dict(pdBIOTagMap)
		
		if plPredSeqs is None:
			self.predSeqs = []
		else:
			# fixing the possible problems with predicted span starts (span without B, i.e. I instead of B) 
			self.predSeqs = self._fixPredSpanStart(plPredSeqs)
		
		if plGoldSeqs is None:
			self.goldSeqs = []
		else:
			self.goldSeqs = plGoldSeqs
		
	
	
	def spanPrecision(self, plPredSeqs = None, plGoldSeqs = None, pMethod = "J&M"):
		'''
		Calculates and returns the sequence labelling (micro) precision based on labelled spans
		'''
		
		if plPredSeqs != None:
			self.predSeqs = plPredSeqs 
		
		if plGoldSeqs != None:
			self.goldSeqs = plGoldSeqs 
		
		# sanity check
		if len(self.predSeqs) != len(self.goldSeqs):
			raise Exception("Number of predicted and gold-standard sequences do not match: %s vs. %s" % (len(self.predSeqs), len(self.goldSeqs)))
		
		if pMethod.lower() == "j&m":
			return self._calcPrecJM2013CL()
		
	
	
	def _calcPrecJM2013CL(self):
		'''
		Calculates the (micro) precision using the method described in Johannson & Moschitti (CL, 2013)
		
		For more details, see the documentation of the class above.
		'''
		
		vGSpanSetCoverage = 0
		vGPredSpanCount = 0 
		
		for i, (vGoldSeq, vPredSeq) in enumerate(zip(self.goldSeqs, self.predSeqs), start=1):
			# sanity check
			if len(vGoldSeq) != len(vPredSeq):
				raise Exception("Gold sequence length does not match the predicted sequence length: %s vs. %s" % (len(vGoldSeq), len(vPredSeq)))
			
			vSpanSetCoverage, dummy, vPredSpanCount = self._calcSeqSpanSetCoverage(plSeq1 = vGoldSeq, plSeq2 = vPredSeq)
			
			vGSpanSetCoverage += vSpanSetCoverage
			vGPredSpanCount += vPredSpanCount
		
		if vGPredSpanCount == 0:
			return 0
		else:
			return vGSpanSetCoverage * 1.0 / vGPredSpanCount
		
	
	
	def spanRecall(self, plPredSeqs = None, plGoldSeqs = None, pMethod = "J&M"):
		'''
		Calculates and returns the sequence labelling (micro) recall based on labelled spans
		'''
		
		if plPredSeqs != None:
			self.predSeqs = plPredSeqs 
		
		if plGoldSeqs != None:
			self.goldSeqs = plGoldSeqs 
		
		# sanity check
		if len(self.predSeqs) != len(self.goldSeqs):
			raise Exception("Number of predicted and gold-standard sequences do not match: %s vs. %s" % (len(vllPredSeqs), len(vllGoldSeqs)))
		
		if pMethod.lower() == "j&m":
			return self._calcRecallJM2013CL()
		
	
	
	def _calcRecallJM2013CL(self):
		'''
		Calculates the (micro) recall using the method described in Johannson & Moschitti (CL, 2013)
		
		For more details, see the documentation of the class above.
		'''
		
		vGSpanSetCoverage = 0
		vGGoldSpanCount = 0 
		
		for vGoldSeq, vPredSeq in zip(self.goldSeqs, self.predSeqs):
			# sanity check
			if len(vGoldSeq) != len(vPredSeq):
				raise Exception("Gold sequence length does not match the predicted sequence length: %s vs. %s" % (len(vGoldSeq), len(vPredSeq)))
			
			vSpanSetCoverage, dummy, vGoldSpanCount= self._calcSeqSpanSetCoverage(plSeq1 = vPredSeq, plSeq2 = vGoldSeq)
			
			vGSpanSetCoverage += vSpanSetCoverage
			vGGoldSpanCount += vGoldSpanCount
		
		if vGGoldSpanCount == 0:
			return 0
		else:
			return vGSpanSetCoverage * 1.0 / vGGoldSpanCount
		
	
	
	def spanFScore(self, plPredSeqs = None, plGoldSeqs = None, pMethod = "J&M", pBeta = 1):
		'''
		Calculates and returns the sequence labelling (micro) f-score based on labelled spans
		
		The pBeta values specifies the weight assigned to recall in the
		general F-score formula:
		
		F-score = (1 + Beta^2).(Precision.Recall) / ((Beta^2).Precision + Recall)
		'''
		
		vPrecision = self.spanPrecision(plPredSeqs, plGoldSeqs, pMethod = pMethod)
		vRecall = self.spanRecall(plPredSeqs, plGoldSeqs, pMethod = pMethod)
		
		if (pBeta**2) * vPrecision + vRecall == 0:
			return 0
		else:
			return (1 + pBeta**2) * (vPrecision * vRecall) / ((pBeta**2) * vPrecision + vRecall)
		
	
	
	def _calcSeqSpanSetCoverage(self, plSeq1, plSeq2):
		'''
		Calculates the span set coverage of sequence 1 with respect to sequence 2 
		
		The sequences are two matching lists of BIO labels. Span set coverage is the sum of span coverages of each span 
		in the sequence 1 with respect to each span of the sequence 2 (i.e. sum over a Cartesian product). Span coverage
		of a span with respect to another span is the number element indexes they have in common (intersect) divided by
		the length of the second span (Johannson & Moschitti, CL, 2013).   
		'''
		
		vSpanSetCoverage = 0
		
		vlSpanSetIdxs1 = self._getSpanSetIdxs(plSeq1)
		vlSpanSetIdxs2 = self._getSpanSetIdxs(plSeq2)
		
		for vlSpanIdxs1 in vlSpanSetIdxs1:
			for vlSpanIdxs2 in vlSpanSetIdxs2:
				vSpanSetCoverage += self._calcSpanCoverage(vlSpanIdxs1, vlSpanIdxs2)
		
		return vSpanSetCoverage, len(vlSpanSetIdxs1), len(vlSpanSetIdxs2)
		
	
	
	def _calcSpanCoverage(self, plSpan1Idxs, plSpan2Idxs):
		'''
		Calculates and returns span coverage of span 1 with respect to span 2
		
		Span coverage of a span with respect to another span is the number element indexes they have in common (intersect)
		divided by the length of the second span (Johannson & Moschitti, CL, 2013).
		'''
		
		return len(set(plSpan1Idxs) & set(plSpan2Idxs)) * 1.0 / len(plSpan2Idxs)
		
	
	
	def _getSpanSetIdxs(self, plSequence):
		'''
		Given a sequence of BIO labels, extracts and returns list of spans, each represented by list of its element indexes
		'''
		
		#print plSequence
		vlSpanIdxs = []
		
		for i, vElement in enumerate(plSequence, start = 1):
			if vElement == self.bioTagMap['B']:
				vlSpanIdxs.append([i])
			elif vElement == self.bioTagMap['I']:
				vlSpanIdxs[-1].append(i)
		
		return vlSpanIdxs
		
	
	
	def _fixPredSpanStart(self, plPredSeqs):
		'''
		Fixes the possible problems with predicted span starts (span without B, i.e. I instead of B)
		'''
		
		vllFixedSeqs = []
		
		for vSeq in plPredSeqs:
			vllFixedSeqs.append([])
			for vTag in vSeq:
				if vTag == self.bioTagMap['I']:
					if len(vllFixedSeqs[-1]) == 0:
						vllFixedSeqs[-1].append(self.bioTagMap['B'])
					elif vllFixedSeqs[-1][-1] == self.bioTagMap['O']:
						vllFixedSeqs[-1].append(self.bioTagMap['B'])
					else:
						vllFixedSeqs[-1].append(vTag)
				else:
					vllFixedSeqs[-1].append(vTag)
			
		return vllFixedSeqs
		
	
	
	def extractExactMismatches(self):
		'''
		Extracts cases where prediction and gold boundaries do not exactly match.
		 
		The returned is a list of tuples of mismatched cases in the format of:
		 (instance no., gold sequence, predicted sequence)
		'''
		
		vlMismatches = []
		
		for i, (vGoldSeq, vPredSeq) in enumerate(zip(self.goldSeqs, self.predSeqs), start = 1):
			# sanity check
			if len(vGoldSeq) != len(vPredSeq):
				raise Exception("Gold sequence length does not match the predicted sequence length: %s vs. %s" % (len(vGoldSeq), len(vPredSeq)))
			
			for vGoldTok, vPredTok in zip(vGoldSeq, vPredSeq):
				if vGoldTok != vPredTok:
					vlMismatches.append((i, vGoldSeq, vPredSeq))
					break
		
		return vlMismatches
		
	
	


class IOEval(BIOEval):
	'''
	Class for evaluation of binary (IO) sequence labelling
	
	Binary or IO sequence tagging is similar to BIO taging except that the beginning (B) and inside (I) are not differntiated.
	This can be used when every instance can only have one type of sequence.
	'''
	
	
	def __init__(self, plPredSeqs=None, plGoldSeqs=None, pdIOTagMap={'I': 'I', 'O': 'O'}):
		'''
		Constructor
		
		NOTE: prediction and gold sequence tags are provided in 2D lists, where the first dimension represents sequences
		and the second represents sequence tokens.
		
		The IO tag map is a dictionary which maps IO tags to any tag set which is used by the input data. For example, 
		if the input data (predictions and gold) is using an index set of {0, 1} instead of "IO", this argument will 
		look like: {'I': 0, 'O': 1}
		'''
		
		self.ioTagMap = dict(pdIOTagMap)
		
		if plPredSeqs is None:
			self.predSeqs = []
		else:
			self.predSeqs = plPredSeqs
		
		if plGoldSeqs is None:
			self.goldSeqs = []
		else:
			self.goldSeqs = plGoldSeqs
		
		
		# if there are other things added to the __init__of the BIOEval which are common, call super.__init__) here. 
		
	
	
	def _getSpanSetIdxs(self, plSequence):
		'''
		Given a sequence of IO labels, extracts and returns list of spans, each represented by list of its element indexes
		'''
		
		# print plSequence
		vlSpanIdxs = []
		
		vIsInside = False
		
		for i, vElement in enumerate(plSequence, start=1):
			if not vIsInside and vElement == self.ioTagMap['I']:                        # the beginning
				vlSpanIdxs.append([i])
				vIsInside = True 
			elif vIsInside and vElement == self.ioTagMap['I']:                          # continuing inside
				vlSpanIdxs[-1].append(i)
			elif vIsInside and vElement == self.ioTagMap['O']:                          # out of inside
				vIsInside = False
		
		return vlSpanIdxs
		
	
	



#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# PredStatSig
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class PredStatSig:
	'''
	Class for testing statistical significance of the difference between 
	two predictions
	
	Supported tests:
	- t-test
	- win ratio 
	
	Paired bootstrap resampling:
	
	Since some metrics like RMSE are not defined at instance level (sample
	level instead), paired bootstrap resampling is used here: first a number
	of paired resamples are randomly drawn from the two predictions and 
	the reference and the metric for each of them is computed. We then have
	two samples of the metric and can perform the statistical significance
	test like t-test. 
	'''
	
	
	def __init__(self, plPrediction1, plPrediction2, plReference):
		'''
		Constructor
		'''
		
		self.prediction1 = plPrediction1
		self.prediction2 = plPrediction2
		self.reference = plReference
		
		if len(self.prediction1) != len(self.reference) and len(self.prediction2) != len(self.reference):
			sys.exit("Length mismatch!")
		
		self.pred1Resamples = []
		self.pred2Resamples = []
		self.refResamples = []
		
	
	
	def pairedResample(self, pResampleCount):
		'''
		Generates and loads pResampleCount number of paired bootstrapp
		resamples with replacement.
		
		Note that resampling is done with replacement so the resample size
		is the same as sample size, because replacement means that a data
		point can be duplicate inside a sample.
		'''
		
		self.pred1Resamples = []
		self.pred2Resamples = []
		self.refResamples = []
		vResampleSize = len(self.reference)
		
		# iterating to generate resamples
		for i in range(pResampleCount):
			# resampling
			vlResampleIdxs = []
			for i in range(vResampleSize):
				vlResampleIdxs.append(random.randint(0, vResampleSize - 1))
			
			# sorting resample's indexes for fun
			vlResampleIdxs.sort()
			
			self.pred1Resamples.append([self.prediction1[i] for i in vlResampleIdxs])
			self.pred2Resamples.append([self.prediction2[i] for i in vlResampleIdxs])
			self.refResamples.append([self.reference[i] for i in vlResampleIdxs])
		
	
	
	def tTestRMSE(self):
		'''
		Performs a t-test for statistical significance of the difference
		between the RMSE of two predictions.
		
		Paired bootstrap resampling is used for RMSE (see class documentation)
		
		If paired resamples are already not loaded, it done first. In this
		case sample size will be used as resample count 
		'''
		
		if self.pred1Resamples == [] or self.pred2Resamples == []:
			self.pairedResample(len(self.reference))
		
		vlResampleRMSEs1 = []
		vlResampleRMSEs2 = []
		vResampleSize = len(self.reference)
		
		# iterating through resamples to compute RMSE for each
		for vPred1Resample, vPred2Resample, vRefResample in zip(self.pred1Resamples, self.pred2Resamples, self.refResamples):
			vPredEval = PredEval(plReference = vRefResample)
			vlResampleRMSEs1.append(vPredEval.rmse(vPred1Resample))
			vlResampleRMSEs2.append(vPredEval.rmse(vPred2Resample))
			
		# computing t for the resample RMSEs
		vT = TTest().t(vlResampleRMSEs1, vlResampleRMSEs2, pType = 3)
		
		return vT
		
	
	
	def tTestR(self):
		'''
		Performs a t-test for statistical significance of the difference
		between the Pearson r of two predictions.
		
		Paired bootstrap resampling is used for r (see class documentation)
		
		If paired resamples are already not loaded, it done first. In this
		case sample size will be used as resample count 
		'''
		
		if self.pred1Resamples == [] or self.pred2Resamples == []:
			self.pairedResample(len(self.reference))
		
		vlResampleRs1 = []
		vlResampleRs2 = []
		vResampleSize = len(self.reference)
		
		# iterating through resamples to compute r for each
		for vPred1Resample, vPred2Resample, vRefResample in zip(self.pred1Resamples, self.pred2Resamples, self.refResamples):
			vPredEval = PredEval(plReference = vRefResample)
			vlResampleRs1.append(vPredEval.r(vPred1Resample))
			vlResampleRs2.append(vPredEval.r(vPred2Resample))
			
		# computing t for the resample rs
		vT = TTest().t(vlResampleRs1, vlResampleRs2, pType = 3)
		
		return vT
		
	
	
	def winRatioTestAcc(self, pType = "exact", pRDigits = 0, pErrMargin = 0, pErrMarginRatio = 0):
		'''
		Performs a statistical significance test on the difference between
		the accuracy of of two predictions based on win ratio of predictions on
		paired bootstrap resamples.
		
		See PredEval.acc() for what accuracy is.
		
		If paired resamples are already not loaded, it done first. In this
		case sample size will be used as resample count 
		'''
		
		if self.pred1Resamples == [] or self.pred2Resamples == []:
			self.pairedResample(len(self.reference))
		
		vlResampleAccs1 = []
		vlResampleAccs2 = []
		vResampleSize = len(self.reference)
		
		vPredict1Wins = vPredict2Wins = 0
		
		for vPred1Resample, vPred2Resample, vRefResample in zip(self.pred1Resamples, self.pred2Resamples, self.refResamples):
			vPredEval = PredEval(plReference = vRefResample)
			vlResampleAccs1.append(vPredEval.acc(plPrediction = vPred1Resample, pType = pType, pRDigits = pRDigits, pErrMargin = pErrMargin, pErrMarginRatio = pErrMarginRatio))
			vlResampleAccs2.append(vPredEval.acc(plPrediction = vPred2Resample, pType = pType, pRDigits = pRDigits, pErrMargin = pErrMargin, pErrMarginRatio = pErrMarginRatio))
			
			if vlResampleAccs1[-1] < vlResampleAccs2[-1]:
				vPredict1Wins += 1
			elif vlResampleAccs1[-1] > vlResampleAccs2[-1]:
				vPredict2Wins += 1
		
		return vPredict1Wins, vPredict2Wins
		
	
	
	def winRatioTestMAE(self):
		'''
		Performs a statistical significance test on the difference between
		the MAE of two predictions based on win ratio of predictions on
		paired bootstrap resamples.
		
		If paired resamples are already not loaded, it is done first. In
		this case sample size will be used as resample count.
		'''
		
		if self.pred1Resamples == [] or self.pred2Resamples == []:
			self.pairedResample(len(self.reference))
		
		vlResampleMAEs1 = []
		vlResampleMAEs2 = []
		vResampleSize = len(self.reference)
		
		vPredict1Wins = vPredict2Wins = 0
		
		for vPred1Resample, vPred2Resample, vRefResample in zip(self.pred1Resamples, self.pred2Resamples, self.refResamples):
			vPredEval = PredEval(plReference = vRefResample)
			vlResampleMAEs1.append(vPredEval.mae(vPred1Resample))
			vlResampleMAEs2.append(vPredEval.mae(vPred2Resample))
			
			if vlResampleMAEs1[-1] < vlResampleMAEs2[-1]:
				vPredict1Wins += 1
			elif vlResampleMAEs1[-1] > vlResampleMAEs2[-1]:
				vPredict2Wins += 1
		
		return vPredict1Wins, vPredict2Wins
		
	
	
	def winRatioTestRMSE(self):
		'''
		Performs a statistical significance test on the difference between
		the RMSE of two predictions based on win ratio of predictions on
		paired bootstrap resamples.
		
		Paired bootstrap resampling is used for RMSE (see class documentation)
		
		If paired resamples are already not loaded, it done first. In this
		case sample size will be used as resample count 
		'''
		
		if self.pred1Resamples == [] or self.pred2Resamples == []:
			self.pairedResample(len(self.reference))
		
		vlResampleRMSEs1 = []
		vlResampleRMSEs2 = []
		vResampleSize = len(self.reference)
		
		vPredict1Wins = vPredict2Wins = 0
		
		for vPred1Resample, vPred2Resample, vRefResample in zip(self.pred1Resamples, self.pred2Resamples, self.refResamples):
			vPredEval = PredEval(plReference = vRefResample)
			vlResampleRMSEs1.append(vPredEval.rmse(vPred1Resample))
			vlResampleRMSEs2.append(vPredEval.rmse(vPred2Resample))
			
			if vlResampleRMSEs1[-1] < vlResampleRMSEs2[-1]:
				vPredict1Wins += 1
			elif vlResampleRMSEs1[-1] > vlResampleRMSEs2[-1]:
				vPredict2Wins += 1
		
		return vPredict1Wins, vPredict2Wins
		
	
	
	def winRatioTestR(self):
		'''
		Performs a statistical significance test on the difference between
		Pearson r of two predictions based on win ratio of predictions on
		paired bootstrap resamples.
		
		Paired bootstrap resampling is used for r (see class documentation)
		
		If paired resamples are already not loaded, it done first. In this
		case sample size will be used as resample count 
		'''
		
		if self.pred1Resamples == [] or self.pred2Resamples == []:
			self.pairedResample(len(self.reference))
		
		vlResampleRs1 = []
		vlResampleRs2 = []
		vResampleSize = len(self.reference)
		
		vPredict1Wins = vPredict2Wins = 0
		
		for vPred1Resample, vPred2Resample, vRefResample in zip(self.pred1Resamples, self.pred2Resamples, self.refResamples):
			vPredEval = PredEval(plReference = vRefResample)
			vlResampleRs1.append(vPredEval.r(vPred1Resample))
			vlResampleRs2.append(vPredEval.r(vPred2Resample))
			
			if vlResampleRs1[-1] > vlResampleRs2[-1]:
				vPredict1Wins += 1
			elif vlResampleRs1[-1] < vlResampleRs2[-1]:
				vPredict2Wins += 1
		
		return vPredict1Wins, vPredict2Wins
		
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# TTest
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class TTest:
	'''
	Class for testing statistical significance with t-test
	'''
	
	
	def __init__(self, plSample1 = None, plSample2 = None):
		'''
		Constructor
		'''
		
		# both or neither samples must be provided 
		if plSample1 is None or plSample2 is None:
			self.sample1 = []
			self.sample2 = []
		else:
			self.sample1 = plSample1[:]
			self.sample2 = plSample2[:]
		
	
	
	def t(self, plSample1 = None, plSample2 = None, pType = 1, pMu = 0):
		'''
		Computes t value for two samples.
		
		pType specifies the type of t-test:
		- 1: one-sample
		- 2: two-sample
		- 3: paired
		'''
		
		if self.sample1 == []:
			self.sample1 = plSample1[:]
		if self.sample2 == []:
			self.sample2 = plSample2[:]
		
		if pType == 2:
			return self._twoSampleT()
		elif pType == 3:
			return self._pairedT(pMu)
		
	
	
	def _twoSampleT(self):
		'''
		Computes two-sample T
		
		Currently only equal-size samples case is computed.
		
		t = (MEAN(sample1) - MEAN(sample2)) / 
		    SQRT((VARIANCE(sample1) / COUNT(sample1)) + (VARIANCE(sample2) / COUNT(sample2)))
		'''
		
		vCount1 = len(self.sample1)
		vCount2 = len(self.sample2)
		
		vMean1 = sum(self.sample1) / vCount1
		vMean2 = sum(self.sample2) / vCount2
		
		# sample variances
		vVar1 = sum(map(lambda x: (x - vMean1)**2, self.sample1)) / (vCount1 - 1)
		vVar2 = sum(map(lambda x: (x - vMean2)**2, self.sample2)) / (vCount2 - 1)
		
		vNumerator = vMean1 - vMean2
		vDenominator = math.sqrt((vVar1 / vCount1) + (vVar2 / vCount2))
		
		# t value 
		return vNumerator / vDenominator
		
	
	
	def _pairedT(self, pMu = 0):
		'''
		Computes paired samples T
		
		t = (MEAN(sample1i - sample2i) - mu ) / 
		    (STDV(sample1i - sample2i)) / SQRT(COUNT(sample1)))
		'''
		
		vCount = len(self.sample1)
		if vCount != len(self.sample2):
			sys.exit("Two samples are not the same size: %s vs. %s" % (vCount, len(self.sample2)))
		elif vCount == 2:
			sys.exit("At least two samples are needed!")
		
		vlDiffs = map(lambda x, y : x - y, self.sample1, self.sample2)
		vDiffMean = sum(vlDiffs) / vCount
		
		# sample variances
		vDiffSD = math.sqrt(sum(map(lambda x: (x - vDiffMean)**2, vlDiffs)) / (vCount - 1))
		
		vNumerator = vDiffMean - pMu
		vDenominator = vDiffSD / math.sqrt(vCount)
		
		# t value 
		return vNumerator / vDenominator
		
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# IAA
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class IAA:
	'''
	Class for calculating inter annotator agreements
	
	The class is supposed to support various settings and various methods.
	'''
	
	def __init__(self, pdAnnotationData = {}):
		'''
		Constructor
		
		psAnnotationData is a dictionary containing annotations where
		the keys are the annotator IDs and values are list of labels by
		that annotator.
		'''
		
		## self.annotationData is the dictionary containing annotations where
		## the keys are the annotator IDs and values are lists containing
		## annotations, one label per array element.
		self.annotationData = {}
		
		# loading annotation data
		for vAnnotatorID, vLabels in pdAnnotationData.iteritems():
			self.annotationData[vAnnotatorID] = vLabels
		
	
	
	def toNLTKAgreementData(self):
		'''
		Converts annotation data into NLTK AgreementTask data format (list
		of triples of annotator ID, item and label)
		'''
		
		# items are simply the positions in the list
		
		vlNLTKData = []
		for vAnnotatorID, vlAnnotations in self.annotationData.iteritems():
			for i, vLabel in enumerate(vlAnnotations, start = 1):
				if type(vLabel) is not str:
					vLabel = str(vLabel)
				vlNLTKData.append((vAnnotatorID, i, vLabel))
		
		return vlNLTKData
		
	
	
	def nltkKappa(self, pDistanceFunc = None):
		'''
		Computes Kappa using NLTK library
		'''
		
		from nltk.metrics.agreement import AnnotationTask
		
		vlNLTKData = self.toNLTKAgreementData()
		
		if pDistanceFunc:
			return AnnotationTask(data = vlNLTKData, distance = pDistanceFunc).kappa()
		else:
			return AnnotationTask(data = vlNLTKData).kappa()
		
	
	
	def nltkConusionMatrix(self, pRefAnnotatorID, pTestAnnotatorID):
		'''
		Computes confusion matrix using NLTK library
		'''
		
		from nltk.metrics import ConfusionMatrix
		
		return ConfusionMatrix(self.annotationData[pRefAnnotatorID], self.annotationData[pTestAnnotatorID])
		
	

