#! /usr/bin/python

## This script prepares a Weka ARFF file for the problem of computing 
## information gain for any word in a input file with respect to a set 
## of sentences classified into good and bad (e.g. based on the quality 
## of parse tree produced for them). This file can be used by 
## weka.attributeSelection.InfoGainAttributeEval to rank features.
##
## Current version: 1.0
##

import sys, optparse

##----------------------------------------------------------------------
## extracts features and writes into output ARFF file
def extractFeatures(pOutFile, psetWords, pGoodFile, pBadFile):
 ## extracting features for good sentences
 for sentence in pGoodFile.read().splitlines():
  for word in psetWords: 
   if sentence.find(word) != -1: 
    pOutFile.write("TRUE,")
   else:
    pOutFile.write("FALSE,")
  
  ## writing the class value for bad sentence
  pOutFile.write('GOOD\n')

 ## extracting features for bad sentences
 for sentence in pBadFile.read().splitlines():
  for word in psetWords: 
   if sentence.find(word) != -1: 
    pOutFile.write("TRUE,")
   else:
    pOutFile.write("FALSE,")
  
  ## writing the class value for bad sentence
  pOutFile.write('BAD\n')

 pOutFile.write('\n')
 return


##----------------------------------------------------------------------
## writes the header of the output file according to the header file provided
## ignores any attribute provided in the header file
## also writes the @DATA at the end
def writeHeader(pOutFile, pHeaderFile, psetWords):
 for line in pHeaderFile:
  if line.upper().startswith("@RELATION"):
   pOutFile.write(line)
   break
  else:
   pOutFile.write(line)
 pOutFile.write('\n')
 
 ## adding attribute declarations for features: one feature for each uniquie word
 for word in psetWords:
  if '"' in word:
   pOutFile.write("@ATTRIBUTE\t'" + word + "'\t\t\t{TRUE,FALSE}\n")
  else:
   pOutFile.write("@ATTRIBUTE\t\"" + word + "\"\t\t\t{TRUE,FALSE}\n")
  
 ## adding attribute declaration for the class 
 ## Note that the attribute name has been put wclassw so that it does'nt 
 ## conflict with any possible attribute (words here) named class, because
 ## Weka is sensitive to it.
 pOutFile.write("@ATTRIBUTE\twclassw\t\t{GOOD,BAD}\n")
   
 pOutFile.write('\n')
 pOutFile.write("@DATA\n")



##======================================================================
## main
def main(argv=None):
 if argv is None:
  argv = sys.argv
 
 parser = optparse.OptionParser(usage="%prog [-w TOKENIZED INPUT TEXT] [-g TOKENIZED GOOD SENTENCES] [-b TOKENIZED BAD SENTENCES]" + 
                                      "\nPrepares ARFF file to be used in ranking words based on their contribution to the classification of a set of sentences to good and bad.", version="%prog 1.0")

 parser.add_option("-w", "--words", help="tokenized input text file", metavar="TOKENIZED INPUT TEXT", dest="inputTextFileName", action="store")
 parser.add_option("-g", "--good", help="tokenized good sentences file", metavar="TOKENIZED GOOD SENTENCES", dest="goodFileName", action="store")
 parser.add_option("-b", "--bad", help="tokenized bad sentences file", metavar="TOKENIZED BAD SENTENCES", dest="badFileName", action="store")
 parser.add_option("-H", "--header", help="output ARFF file header (attributes will be ignored)", metavar="OUTPUT ARFF FILE HEADER", dest="headerFileName", action="store")
 parser.add_option("-o", "--output", help="output ARFF file (don't include .arff suffix)", metavar="OUTPUT FILE NAME", dest="outFileName", action="store")

 (opts, posArgs) = parser.parse_args()

 ## opening tokenized input text file
 if opts.inputTextFileName == None:
  print 'Text input file is not provided!'
  sys.exit(2)
 else:
  try:
   inputTextFile  = open(opts.inputTextFileName, 'r')
  except IOError:
   print 'Can\'t open text input file: ' + opts.inputTextFileName
   sys.exit(2)

 ## opening tokenized good sentences file
 if opts.goodFileName == None:
  print 'Tokenized good sentences file is not provided!'
  sys.exit(2)
 else:
  try:
   goodFile  = open(opts.goodFileName, 'r')
  except IOError:
   print 'Can\'t open tokenized good sentences file: ' + opts.goodFileName
   sys.exit(2)

 ## tokenized bad sentences file
 if opts.badFileName == None:
  print 'Tokenized good sentences file is not provided!'
  sys.exit(2)
 else:
  try:
   badFile  = open(opts.badFileName, 'r')
  except IOError:
   print 'Can\'t open tokenized good sentences file: ' + opts.badFileName
   sys.exit(2)

 ## opening output header file
 if opts.headerFileName == None:
  print 'Output header file is not provided!'
  sys.exit(2)
 else:
  try:
   headerFile  = open(opts.headerFileName, 'r')
  except IOError:
   print 'Can\'t open output header file: ' + opts.headerFileName
   sys.exit(2)

 ## creating output file
 if opts.outFileName == None:
  print 'Output file name is not provided!'
  sys.exit(2)
 else:
  try:
   outFile  = open(opts.outFileName + '.arff', 'w')
  except IOError:
   print 'Can\'t create output file: ' + opts.outFileName
   sys.exit(2)

 ## loading the input words into memory (in set data struture to create a list (set) of uniqure words)
 vsetWords = set([])
 for line in inputTextFile.read().splitlines():
  vsetLineWords = set(line.split())
  vsetWords = vsetWords.union(vsetLineWords) 
	
 ## writing output file header
 writeHeader(outFile, headerFile, vsetWords)
 
 ## extracting and writing features
 ## remember to add new feature here
 extractFeatures(outFile, vsetWords, goodFile, badFile)

 inputTextFile.close()
 goodFile.close()
 badFile.close()
 outFile.close()
 headerFile.close()

 
##======================================================================
## calling main
if __name__ == "__main__":
 sys.exit(main())


