#! /usr/bin/python

"""
	This module provides classes for feature extraction configuration and processing. 
	
	This module should be used instead of deprecated fvg.py.
	
	
	Version 0.3                                                                        	    (02-May-2016 to 02-May-2016)
	- In FSConfig.createFeature(), default data type for scalable features is set to float.  
	
	Version 0.2                                                                        	    (25-Feb-2016 to 01-Mar-2016)
	- Standardization method is added to scaling and the settings for original method which is rescaling is adjusted.
	- Expanding of list or array type features is added to FSConfig and FProc.
	
	Version 0.1                                                                        	    (04-Feb-2016 to 12-Feb-2016)
	- FSConfig, FConfig, FProc are created (based on fvg.FVGen and fvg.Feature).
	
"""


import data
import numpy as np
from utils import util
import yaml


class FSConfig:
	'''
	Class for processing feature set configuration
	
	This class is based on fvg.FVGen, where the focus has changed to be only on feature configuration.
	'''
	
	
	def __init__(self, pConfig):
		'''
		Constructor 
		'''
		
		# list of FConfig objects
		self.fConfigs = []
		
		# global configuration 
		self.gNomNormMethod = None
		self.gNomBinFreqCutoff = None
		self.scale = None
		self.scaleMethod = None
		self.gRescaleTrgRange = None
		self.gRoundDigits = None
		
		# loading configuration
		self.loadConfig(pConfig)
		
	
	
	def loadConfig(self, pConfig):
		'''
		Loads feature set configuration
		
		The configuration is at global (see _extractGlobalConfig()) and 
		feature levels. 
		
		At feature level: the format is:
		
		     features:
		         <Feature Name>:
		             <Feature Setting>
		             ...
		         ...
		
		See FConfig for description of <Feature Setting>
		'''
		
		# loading configuration from YAML format
		vConfig = yaml.load(pConfig)
		
		# first loading global settings
		self._extractGlobalConfig(vConfig)
		
		# now loading feature configurations and their individual settings
		if "features" not in vConfig:
			raise Exception("No feature is specified in the configuration!")
		else:
			for vFeatureName, vdFeatureConfig in vConfig["features"].iteritems():
				vFConfig = FConfig(vFeatureName, vdFeatureConfig, self)
				self.fConfigs.append(vFConfig)
		
	
	
	def _extractGlobalConfig(self, pConfig):
		'''
		Extract global settings from pConfig dictionary.
		
		Global settings include:
		
		- Global Nominal Normalization Method: this is the global nominal
		   feature normalization method used in the absence of this setting
		   for an individual feature.
		   
		- Global Nominal Binarization Frequency Cutoff: this is the global 
		   nominal binarization frequency cutoff used in the absence of this
		   setting for an individual feature.
		   
		- Scale: this specifies whether feature values should be scaled
		   before generating vectors. The format is:
		   
		     Scale: <True|False>
		   
		   where the default value is false.
		- Scaling Method: specifies the method to be used for scaling. Implemented methods are (names from wikipedia):
		   - rescaling: original method which maps values to a range [a, b]
		   - standardization: which subtracts feature value from mean of all values for the features and divides it by
		     the standard deviation of those values.
		
		- Global Scaling Target Range: this is the global scaling target 
		   range used in the absence of this setting for an individual 
		   feature.
		'''
		
		if "global nominal normalization method" in pConfig:
			self.gNomNormMethod = pConfig["global nominal normalization method"]
		else:
			self.gNomNormMethod = "bin"
		
		if "global nominal binarization frequency cutoff" in pConfig:
			self.gNomBinFreqCutoff = pConfig["global nominal binarization frequency cutoff"]
		else:
			self.gNomBinFreqCutoff = 0
		
		if "scale" in pConfig:
			self.scale = pConfig["scale"]
		else:
			self.scale = False
		
		if "scaling method" in pConfig:
			self.scaleMethod = pConfig["scaling method"]
		else:
			self.scaleMethod = "rescaling"
		
		if "global rescaling target range" in pConfig or "global scaling target range" in pConfig:     # 2nd for BW-compatibility
			self.gRescaleTrgRange = (pConfig["global rescaling target range"][0], pConfig["global rescaling target range"][1])
		else:
			self.gRescaleTrgRange = (0,1)
		
		if "rounding feature values" in pConfig:
			self.gRoundDigits = pConfig["rounding feature values"]
		else:
			# do not round
			self.gRoundDigits = -1
		
	
	
	def getFConfigs(self):
		'''
		Returns list of feature configuration objects sorted by index
		'''
		
		# sorting based on the first index in the index list of the feature config 
		return sorted(self.fConfigs, key = lambda x: x.indexes[0] if len(x.indexes) > 0 else x.name)
		
	
	
	def createFeature(self, pFConfig, plValues):
		'''
		Creates and returns new feature(s) based on the given feature configuration and values as well as the global 
		configuration
		
		Created features are data.Feature objects. The feature configuration is updated based on the results of the 
		feature creation (e.g. binarization).
		 
		Feature values can be numbers (numerical), strings (nominal) and lists (multi-valued). Numerical features are 
		created normally. Nominal features are normalized if normalization is specified in feature configuration. Multi-
		valued features (e.g. n-grams) have a list of string values per instance. Therefore, they are normalized as well.
		In these cases, more than one feature will be created depending on the output of normalization or list size.
		'''
		
		# setting data type
		if not pFConfig.dataType:
			pFConfig.dataType = type(plValues[0])
		
		vFProc = FProc(pFConfig)
		
		if pFConfig.normNominal:
			return vFProc.normalizeNominal(plValues)
		elif pFConfig.expandable:
			return vFProc.expandListValues(plValues)
		else:
			if self.scale and pFConfig.scalable:
				pFConfig.dataType = float
				vlValues = vFProc.scale(plValues)
			else:
				vlValues = plValues
			
			vFeature = data.Feature(pName = pFConfig.name, pDataType = pFConfig.dataType)
			vFeature.loadValues([self._round(v) for v in vlValues])
			
			return vFeature
		
	
	
	def dumpConfig(self):
		'''
		Saves the configuration used in generating feature vectors
		
		If the configuration is saved before loading feature values, it 
		will not contain the right configuration in case there is a nominal
		feature or feature values should be scaled.
		
		The configuration is saved in YAML format. It may differs from the
		input configuration loaded by loadConfig() if there is at least
		one feature normalized or scaled. In this case, normalization
		setting such as binarized features, or scaling range will be added
		to the configuration. The purpose is the exact replicability of the
		feature vector generation. This is particularly useful for having 
		exact same feature setting for training and test purposes. The reason
		is that	normalization and scaling are dependent on the data, and same
		setting can end up in different outcome for different data. For	example,
		scaling depends on the minimum and maximum of a feature value which can
		be different for two data set. Therefore, a same value in training and
		test set can be scaled to different values which is not desired. 
		'''
		
		# global config
		vdConfig = self._dumpGlobalConfig()
		
		# individual feature config
		
		vdFeatureConfigs = {}
		
		for vFConfig in self.fConfigs:
			vdFeatureConfigs[vFConfig.name] = vFConfig.dumpConfig()
		
		vdConfig["features"] = vdFeatureConfigs
		
		return yaml.dump(vdConfig)
		
	
	
	def _dumpGlobalConfig(self):
		'''
		Saves global configuration used in generating feature vectors)
		'''
		
		vdConfig = {}
		
		vdConfig["global nominal normalization method"] = self.gNomNormMethod
		vdConfig["global nominal binarization frequency cutoff"] = self.gNomBinFreqCutoff
		vdConfig["scale"] = self.scale
		vdConfig["scaling method"] = self.scaleMethod
		vdConfig["global rescaling target range"] = [self.gRescaleTrgRange[0], self.gRescaleTrgRange[1]]
		vdConfig["rounding feature values"] = self.gRoundDigits
		
		return vdConfig
		
	
	
	def _round(self, pValue):
		'''
		Rounds value based on the rounding setting in the configuration.
		'''
		
		if self.gRoundDigits > -1:
			return round(pValue, self.gRoundDigits)
		else:
			return pValue
		
	
	



class FConfig:
	'''
	Feature configuration class
	'''
	
	
	def __init__(self, pName, pdFeatureConfig, pFSConfig):
		'''
		Constructor
		
		pdFeatureConfig is the feature configuration dictionary extracted from feature set configuration. pFSConfig is 
		the feature set config (FSConfig) object to which the feature belongs. The main purpose of the latter is to 
		access the global feature configuration stored in that object.
		'''
		
		self.name = pName
		
		## Feature indexes are stored (in a list due to binarization possibility) because configuration is in a 
		## dictionary format which is order-free.
		self.indexes = None
		
		# featureParams used only in dumping configuration
		self.featureParams = {}
		
		self.disable = False
		
		self.dataType = None
		
		self.expandable = None
		
		self.normNominal = None
		self.nomNormMethod = None
		self.nomBinFeatures = None
		self.nomBinFreqCutoff = None
		self.nomBinFeatureNames = None
		self.binarized = None
		
		self.scalable = None
		self.scaleMethod = None
		self.srcRange = None
		self.trgRange = None
		
		# extracting configuration
		self._extractConfig(pdFeatureConfig, pFSConfig)
		
	
	
	def _extractConfig(self, pdFeatureConfig, pFSConfig):
		'''
		Extract feature settings from given feature configuration dictionary and global configuration of given feature 
		set configuration object (to which the feature belongs)
		
		If a setting in the feature configuration is not given, global setting will be used.
		
		Each feature includes the following settings:
		    a) Disable: this specifies whether the feature should be ignored
		       and not extracted.
		       
		    b) Parameters: this is feature-specific parameters used in 
		       extracting features. Here, it is only used to be included
		       in configuration dumps.
		
		    c) Normalize Nominal: this specifies whether the nominal feature
		       must be normalized
		       normalization method at individual feature level.
		       See Feature constructor
		       
		    d) Nominal Normalization Method: this is the nominal feature
		       normalization method at individual feature level.
		       See Feature.normalizeNominal()
		       
		    e) Nominal Binarization Frequency Cutoff: this is the nominal 
		       binarization frequency cutoff at individual feature level. 
		       See Feature.binarizeNominal()
		       
		    f) Binarized Features: this is the list of binarized feature 
		       names derived from this feature
		       See Feature.binarizeNominal()
		       
		    g) Scalable: this specifies whether the feature should be scaled
		       when scaling is done.
		    
		    h) Rescaling Source Range: this is the range from which the  
		       feature has been previously rescaled.
		       See FProc.
		       
		    i) Rescaling Target Range: this is the range to which the feature 
		       should be rescaled.
		       See FProc.
		
		A feature has also an index list which specifies its index or the
		indexes of its binary features in the vector if it is binarized. 
		'''
		
		if pdFeatureConfig is None:
			vFeatureConfig = {}
		else:
			vFeatureConfig = pdFeatureConfig
		
		vdNomNormConfig = {}
		vdScaleConfig = {}
		
		if "disable" in vFeatureConfig and vFeatureConfig["disable"] == True:
			self.disable = True
		
		# Unlike in fvg, even if feature is disabled, other settings will be extracted.
		
		if "parameters" in vFeatureConfig:
			self.featureParams = vFeatureConfig["parameters"]
		
		if "data type" in vFeatureConfig:
			self.dataType = vFeatureConfig["data type"]
		
		if "expandable" in vFeatureConfig:
			self.expandable = vFeatureConfig["expandable"]
		else:
			self.expandable = False
		
		if "normalize nominal" in vFeatureConfig:
			vdNomNormConfig["normalize"] = vFeatureConfig["normalize nominal"]
		else:
			## If not specified, it will be automatically set when values
			## are loaded: true for string values and false otherwise.
			vdNomNormConfig["normalize"] = None
		
		if "nominal normalization method" in vFeatureConfig:
			vdNomNormConfig["method"] = vFeatureConfig["nominal normalization method"]
		else:
			vdNomNormConfig["method"] = pFSConfig.gNomNormMethod
		
		if "nominal binarization frequency cutoff" in vFeatureConfig:
			vdNomNormConfig["bin freq cutoff"] = vFeatureConfig["nominal binarization frequency cutoff"]
		else:
			vdNomNormConfig["bin freq cutoff"] = pFSConfig.gNomBinFreqCutoff
		
		if "binarized features" in vFeatureConfig:
			vdNomNormConfig["bin features"] = vFeatureConfig["binarized features"]
		else:
			vdNomNormConfig["bin features"] = []
		
		if "scalable" in vFeatureConfig:
			vdScaleConfig["scalable"] = vFeatureConfig["scalable"]
		else:
			vdScaleConfig["scalable"] = False
		
		# scaling method is not individualized and is only globally available 
		vdScaleConfig["scaling method"] = pFSConfig.scaleMethod
		
		if "rescaling source range" in vFeatureConfig:
			vdScaleConfig["source range"] = (vFeatureConfig["rescaling source range"][0], vFeatureConfig["rescaling source range"][1])
		else:
			vdScaleConfig["source range"] = None
		
		if "rescaling target range" in vFeatureConfig:
			vdScaleConfig["target range"] = (vFeatureConfig["rescaling target range"][0], vFeatureConfig["rescaling target range"][1])
		else:
			vdScaleConfig["target range"] = pFSConfig.gRescaleTrgRange
		
		## settings: scaling setting should be extracted before normalization
		## settings, because the latter may used the former in binarization
		 
		self.setScaling(vdScaleConfig)
		self.setNomNorm(vdNomNormConfig)
		
		
		if "index" in vFeatureConfig:
			self.indexes = vFeatureConfig["index"]
		else:
			self.indexes = []
		
	
	
	def setScaling(self, pdConfig):
		'''
		Sets scaling setting
		
		pdConfig includes:
		  - "scalable": 
		     - True: scalable
		     - False: not scalable (default)
		  - "source range": see FProc: self.srcRange
		  - "target range": see FProc: self.trgRange
		                    Default is [0,1].
		'''
		
		if pdConfig is None:
			vdConfig = {}
		else:
			vdConfig = pdConfig
		
		if "scalable" in vdConfig:
			self.scalable = vdConfig["scalable"]
		else:
			self.scalable = False
		
		if "scaling method" in vdConfig:
			self.scaleMethod = vdConfig["scaling method"]
		else:
			self.scaleMethod = "rescaling"
		
		if self.scalable and self.scaleMethod == "rescaling":
			if "source range" in vdConfig:
				self.srcRange = vdConfig["source range"]
			else:
				# will be set in scale() based on values
				self.srcRange = None
			if "target range" in vdConfig:
				self.trgRange = vdConfig["target range"]
			else:
				self.trgRange = (0, 1)
		else:
			self.srcRange = None
			self.trgRange = None
		
	
	
	def setNomNorm(self, pdConfig):
		'''
		Sets nominal normalization setting
		
		pdConfig includes:
		  - "method": nominal feature normalization method; 
		     - bin: binarization in which each nominal value is converted
		            to a binary feature.
		  - "bin freq cutoff": Frequency cutoff  for binarization
		                        Default is 0.
		                        See Feature.binarizeNominal()
		  - "bin features": list of binarized feature names
		                    See Feature.binarizeNominal()
		'''
		
		if pdConfig is None:
			vdConfig = {}
		else:
			vdConfig = pdConfig
		
		if "normalize" in vdConfig:
			self.normNominal = vdConfig["normalize"]
		else:
			self.normNominal = False
		
		if "method" in vdConfig and vdConfig["method"] == "num":
			self.nomNormMethod = "num"
		else:
			self.nomNormMethod = "bin"
		
		if "bin freq cutoff" in vdConfig:
			self.nomBinFreqCutoff = vdConfig["bin freq cutoff"]
		else:
			self.nomBinFreqCutoff = 0
		
		if "bin features" in vdConfig:
			self.nomBinFeatureNames = vdConfig["bin features"]
		else:
			self.nomBinFeatureNames = []
		
	
	
	def setIndexes(self, pStart, pCount = 1):
		'''
		Sets the value of feature index(es)
		
		pStart is the starting index and pCount is the number of indexes (for the case of binarized features). 
		'''
		
		self.indexes = [i for i in range(pStart, pStart + pCount)]
		
	
	
	def dumpConfig(self):
		'''
		Saves the feature configuration into a configuration dictionary
		'''
		
		vdConfig = {}
		
		vdConfig["parameters"] = self.featureParams
		
		vdConfig["data type"] = self.dataType
		
		vdConfig["expandable"] = self.expandable
		
		vdConfig["normalize nominal"] = self.normNominal
		if self.normNominal:
			vdConfig["nominal normalization method"] = self.nomNormMethod
			if self.nomNormMethod == "bin":
				vdConfig["binarized features"] = self.nomBinFeatureNames
				vdConfig["nominal binarization frequency cutoff"] = self.nomBinFreqCutoff
		
		vdConfig["scalable"] = self.scalable
		if self.scalable and self.scaleMethod == "rescaling":
			if self.srcRange != None:
				vdConfig["rescaling source range"] = [self.srcRange[0], self.srcRange[1]]
			vdConfig["rescaling target range"] = [self.trgRange[0], self.trgRange[1]]
		
		vdConfig["index"] = self.indexes
		
		return vdConfig
		
	
	



class FProc:
	'''
	Class for processing feature based on feature configuration
	'''
	
	
	def __init__(self, pFConfig):
		'''
		Constructor
		
		pFConfig is an FConfig object containing feature configuration.
		'''
		
		self.fConfig = pFConfig
		
	
	
	def normalizeNominal(self, plValues):
		'''
		Normalizes nominal feature values and returns resulting features
		
		The returned features of data.Feature objects. The feature config associated with the processor is updated based
		on the outcome.
		'''
		
		if self.fConfig.nomNormMethod == "bin":
			return self.binarizeNominal(plValues)
		else:
			raise Exception("Unknown nominal normalization method!")
		
	
	
	def binarizeNominal(self, plValues):
		'''
		Binzarizes nominal feature values based on its setting in feature configuration and creates and returns binarized
		features
		
		The returned features of data.Feature objects. The feature config associated with the processor is updated based
		on the outcome.
		
		Setting includes:
		  - "bin freq cutoff": if the number of instances for which a derived binary feature fires is less than this 
		                       fraction of the entire data set, the feature will be ignored.
		     - <Freq>: a value in [0, 1). For example, if it is set to 0.01 and the feature fires for less than 1 percent
		               of the instances, it will be ignored.
		  - "bin features": list of binarized feature names which are in fact a seb set of nominal values. If this 
		                    parameter is provided, the above parameters will be ignored. This is useful when exact same 
		                    set of binarized features to an existing setting is needed.
		'''
		
		## Feature value of an instance can be a list (multi-valued features) or non-list. However, they are all 
		## converted to lists for a unified binarization process.
		if self.fConfig.dataType == list:
			vlValueLists = plValues
		else:
			vlValueLists = [[i] for i in plValues]
		
		# extracting/generating binary feature names if not given in the configuration
		if self.fConfig.nomBinFeatureNames == []: 
			# If binary feature list is not given, extracted features should meet the cutoff threshold.
			vCutoffThreshold = self.fConfig.nomBinFreqCutoff * len(vlValueLists)
			self.fConfig.nomBinFeatureNames = [name for name, count in util.groupBy([v for l in vlValueLists for v in set(l)]).iteritems() if count >= vCutoffThreshold]
			if len(self.fConfig.nomBinFeatureNames) == 0:
				raise Exception("The frequency cutoff %s for feature binarization is too restrictive in %s !" % (self.fConfig.nomBinFreqCutoff, self.fConfig.name))
		
		# extracting binary feature values
		vdBinFeatureValues = {}
		for vlValues in vlValueLists:
			for vName in self.fConfig.nomBinFeatureNames:
				if vName not in vdBinFeatureValues:
					vdBinFeatureValues[vName] = []
				
				if vName in vlValues:
					vdBinFeatureValues[vName].append(1)
				else:
					vdBinFeatureValues[vName].append(0)
		
		# binarized features derived from this feature: a list of data.Feature objects
		vlBinFeatures = []
		
		## creating binary features and loading their values
		## Note: the names of binary features are formed by concatenating original feature name and the binary feature 
		##       name with a dash, e.g. root-S.
		## Note: the binary features are not normalizable but inherit the scaling setting.
		for vCBFName, vlCBFValues in vdBinFeatureValues.iteritems():
			vBinFeatureName = self.fConfig.name + '-' + vCBFName
			vBinFeature = data.Feature(pName = vBinFeatureName, pDataType = int)
			vBinFeature.loadValues(vlCBFValues)
			vlBinFeatures.append(vBinFeature)
		
		# updating feature settings
		self.fConfig.binarized = True
		
		return vlBinFeatures
		
	
	
	def expandListValues(self, pllValues):
		'''
		Expands feature to multiple feature if each feature value is a list of values (e.g. word vectors) 
		
		The returned features of data.Feature objects.
		'''
		
		# expanded features derived from this feature: a list of data.Feature objects
		vlExpandedFeatures = []
		
		## creating expanded features and loading their values
		## Note: the names of binary features are formed by concatenating original feature name and the expanded feature 
		##       index with a dash, e.g. wv-1.
		## Note: the expanded features inherit the scaling setting.
		for i, vlEFValues in enumerate([list(fvs) for fvs in zip(*pllValues)], start = 1):
			vEFName = self.fConfig.name + '-' + str(i)
			vEF = data.Feature(pName = vEFName, pDataType = type(vlEFValues[0]))
			vEF.loadValues(vlEFValues)
			vlExpandedFeatures.append(vEF)
		
		self.fConfig.expandable = True             # in case not set by user and got as far 
		
		return vlExpandedFeatures
		
	
	
	def scale(self, plValues):
		'''
		Scales given values based on the method specified in the configuration
		'''
		
		# sanity check
		if not self.fConfig.scalable:
			raise Exception("Invalid request! Feature is not scalable.")
		
		if self.fConfig.scaleMethod == "rescaling":
			return self.rescale(plValues)
		elif self.fConfig.scaleMethod == "standardization":
			return self.standardize(plValues)
		
	
	
	def rescale(self, plValues):
		'''
		Rescales given values into the given range
		
		The rescaling is based on the rescaling attributes:
		  - self.srcRange: the range to scale from. If not given, it will
		                   be set to the range of feature values (min and 
		                   max). If given (usually when we want to have 
		                   exact same scaling to a previous one, e.g. test
		                   like training), but the range of feature values 
		                   does not fall inside this range, it and the 
		                   target range will be adjusted accordingly. 
		                   For example, if this feature had values in range
		                   [10, 100] in  training data and scaled to [0, 1],
		                   these are given as source and target ranges in
		                   the configuration. Then, if in test set the range
		                   is [11, 120] or [5, 100], the source range will
		                   bw adjusted to [10, 120] or [5, 100] and the 
		                   target range to [0, 1.2] or [-0.05, 1].
		                   The format is:
		                     - <[X, Y]>: X is the lower and Y is the upper
		                                 bound of the range (inclusive).
		  - self.trgRange: the range to scale to. In the
		                   latter case, this range will be adjusted and 
		                   scaled to.
		                   The format is:
		                     - <[X, Y]>: X is the lower and Y is the upper
		                                 bound of the range (inclusive).
		'''
		
		vlScaledValues = []
		
		# range of values
		vNewSrcRange = (min(plValues), max(plValues))
		
		if self.fConfig.srcRange is None:
			self.fConfig.srcRange = vNewSrcRange
		else:
			# adjusting the target range if range of values is not fully covered by given source range
			
			if vNewSrcRange[0] < self.fConfig.srcRange[0]:
				vDiff = self.fConfig.srcRange[0] - vNewSrcRange[0]
				vScale = 1.0 * (self.fConfig.trgRange[1] - self.fConfig.trgRange[0]) / (self.fConfig.srcRange[1] - self.fConfig.srcRange[0])
				self.fConfig.srcRange = (vNewSrcRange[0], self.fConfig.srcRange[1])
				self.fConfig.trgRange = (self.fConfig.trgRange[0] - vDiff * vScale, self.fConfig.trgRange[1])
			
			if vNewSrcRange[1] > self.fConfig.srcRange[1]: 
				vDiff = vNewSrcRange[1] - self.fConfig.srcRange[1]
				vScale = 1.0 * (self.fConfig.trgRange[1] - self.fConfig.trgRange[0]) / (self.fConfig.srcRange[1] - self.fConfig.srcRange[0])
				self.fConfig.srcRange = (self.fConfig.srcRange[0], vNewSrcRange[1])
				self.fConfig.trgRange = (self.fConfig.trgRange[0], self.fConfig.trgRange[1] + vDiff * vScale)
		
		for vValue in plValues:
			vlScaledValues.append(util.scaleNumber(vValue, self.fConfig.srcRange, self.fConfig.trgRange))
		
		return vlScaledValues
		
	
	
	def standardize(self, plValues):
		'''
		Standardize the given values and returns the results  
		'''
		
		
		# sanity check
		if len(plValues) == 0:
			return plValues
		else:
			vMean = np.mean(plValues)
			vSD = np.std(plValues)
			return [float((v - vMean) * 1.0 / vSD) for v in plValues]   # feature type is ignored and float is used
		
	
	
