#! /usr/bin/python

"""
	This module provides classes to generate features vectors for machine
	learning algorithms.
	
	
	DEPRECATED: USE fecp.py INSTEAD							(04-Feb-2016)
	
	Version 0.1								(11-May-2015 to 14-May-2015)
	- FVGen and Feature are moved in from svmfg.py.
	- getRawFVectors() is added to FVGen.
	- Feature.numericalizeNominal() is now removed as numericalization
	  does not seem to be correct way of normalization. Also, everything 
	  about numericalization are removed.
	
"""

from utils import util
import sys, yaml, random


class FVGen:
	'''
	Feature vector generator
	'''
	
	
	def __init__(self, pConfig):
		'''
		Constructor
		'''
		
		## Features are stored in a dictionary: keys are the feature names
		## and values are tuples of feature index list and feature object
		## of type Feature. Note that the feature indexes are stored in 
		## a list because it may be binarized and need the indexes of binary
		## features in the feature vectors.
		self.features = {}
		
		# will be initialized with size of first feature added
		self.datasetSize = 0
		
		# loading configuration
		self.loadConfig(pConfig)
		
	
	
	def loadConfig(self, pConfig):
		'''
		Loads feature configuration
		
		The configuration is at global (see _extractGlobalConfig()) and 
		feature levels. 
		
		At feature level: the format is:
		
		     features:
		         <Feature Name>:
		             <Feature Setting>
		             ...
		         ...
		
		See _extractFeatureConfig() for description of <Feature Setting>
		'''
		
		# loading configuration from YAML format
		vConfig = yaml.load(pConfig)
		
		# first loading global settings
		self._extractGlobalConfig(vConfig)
		
		# now loading features and their individual settings
		if "features" not in vConfig:
			raise Exception("No feature is specified in configuration!")
		else:
			for vFeatureName, vFeatureConfig in vConfig["features"].iteritems():
				# loading feature setting
				vDisable, vlIndex, vdFeatureParams, vdNomNormConfig, vdScaleConfig = self._extractFeatureConfig(vFeatureConfig)
				
				# creating feature
				if vDisable == False:
					self.features[vFeatureName] = (vlIndex, Feature(vFeatureName, vdFeatureParams, vdNomNormConfig, vdScaleConfig))
		
	
	
	def _extractGlobalConfig(self, pConfig):
		'''
		Extract global settings from pConfig dictionary.
		
		Global settings include:
		
		- Global Nominal Normalization Method: this is the global nominal
		   feature normalization method used in the absence of this setting
		   for an individual feature.
		   
		- Global Nominal Binarization Frequency Cutoff: this is the global 
		   nominal binarization frequency cutoff used in the absence of this
		   setting for an individual feature.
		   
		- Scale: this specifies whether feature values should be scaled
		   before generating vectors. The format is:
		   
		     Scale: <True|False>
		   
		   where the default value is false.
		   
		- Global Scaling Target Range: this is the global scaling target 
		   range used in the absense of this setting for an individual 
		   feature.
		'''
		
		if "global nominal normalization method" in pConfig:
			self.gNomNormMethod = pConfig["global nominal normalization method"]
		else:
			self.gNomNormMethod = "bin"
		
		if "global nominal binarization frequency cutoff" in pConfig:
			self.gNomBinFreqCutoff = pConfig["global nominal binarization frequency cutoff"]
		else:
			self.gNomBinFreqCutoff = 0
		
		if "scale" in pConfig:
			self.scale = pConfig["scale"]
		else:
			self.scale = False
		
		if "global scaling target range" in pConfig:
			self.gScaleTrgRange = (pConfig["global scaling target range"][0], pConfig["global scaling target range"][1])
		else:
			self.gScaleTrgRange = (0,1)
		
		if "rounding feature values" in pConfig:
			self.gRoundDigits = pConfig["rounding feature values"]
		else:
			# do not round
			self.gRoundDigits = -1
		
	
	
	def _extractFeatureConfig(self, pFeatureConfig):
		'''
		Extract feature settings from pConfig dictionary and global 
		configuration already extracted.
		
		If a setting is not given, global setting will be used. This is
		the reason why this method is not moved to Feature class, as it
		is not supposed to have access to global setting.
		
		Each feature includes the following settings:
		    a) Disable: this specifies whether the feature should be ignored
		       and not extracted.
		       
		    b) Parameters: this is feature-specific parameters used in 
		       extracting features. Here, it is only used to be included
		       in configuration dumps.
		
		    c) Normalize Nominal: this specifies whether the nominal feature
		       must be normalized
		       normalization method at individual feature level.
		       See Feature constructor
		       
		    d) Nominal Normalization Method: this is the nominal feature
		       normalization method at individual feature level.
		       See Feature.normalizeNominal()
		       
		    e) Nominal Binarization Frequency Cutoff: this is the nominal 
		       binarization frequency cutoff at individual feature level. 
		       See Feature.binarizeNominal()
		       
		    f) Binarized Features: this is the list of binarized feature 
		       names derived from this feature
		       see Feature.binarizeNominal()
		       
		    g) Scalable: this specifies whether the feature should be scaled
		       when scaling is done.
		       
		    h) Scaling Source Range: this is the range from which the  
		       feature has been previously scaled.
		       see Feature.scale()
		       
		    i) Scaling Target Range: this is the range to which the feature 
		       should be scaled.
		       see Feature.scale()
		
		A feature has also an index list which specifies its index or the
		indexes of its binary features in the vector if it is binarized. 
		'''
		
		if pFeatureConfig == None:
			vFeatureConfig = {}
		else:
			vFeatureConfig = pFeatureConfig
		
		vdFeatureParams = {}
		vdNomNormConfig = {}
		vdScaleConfig = {}
		
		if "disable" in vFeatureConfig:
			# if disable is true, ignore the feature
			if vFeatureConfig["disable"] == True:
				return True, [], {}, {}, {}
		
		if "parameters" in vFeatureConfig:
			vdFeatureParams = vFeatureConfig["parameters"]
		
		if "normalize nominal" in vFeatureConfig:
			vdNomNormConfig["normalize"] = vFeatureConfig["normalize nominal"]
		else:
			## If not specified, it will be automatically set when values
			## are loaded: true for string values and false otherwise.
			vdNomNormConfig["normalize"] = None
		
		if "nominal normalization method" in vFeatureConfig:
			vdNomNormConfig["method"] = vFeatureConfig["nominal normalization method"]
		else:
			vdNomNormConfig["method"] = self.gNomNormMethod
		
		if "nominal binarization frequency cutoff" in vFeatureConfig:
			vdNomNormConfig["bin freq cutoff"] = vFeatureConfig["nominal binarization frequency cutoff"]
		else:
			vdNomNormConfig["bin freq cutoff"] = self.gNomBinFreqCutoff
		
		if "binarized features" in vFeatureConfig:
			vdNomNormConfig["bin features"] = vFeatureConfig["binarized features"]
		else:
			vdNomNormConfig["bin features"] = []
		
		if "scalable" in vFeatureConfig:
			vdScaleConfig["scalable"] = vFeatureConfig["scalable"]
		else:
			vdScaleConfig["scalable"] = True
		
		if "scaling source range" in vFeatureConfig:
			vdScaleConfig["source range"] = (vFeatureConfig["scaling source range"][0], vFeatureConfig["scaling source range"][1])
		else:
			vdScaleConfig["source range"] = None
		
		if "scaling target range" in vFeatureConfig:
			vdScaleConfig["target range"] = (vFeatureConfig["scaling target range"][0], vFeatureConfig["scaling target range"][1])
		else:
			vdScaleConfig["target range"] = self.gScaleTrgRange
		
		if "index" in vFeatureConfig:
			vlIndex = vFeatureConfig["index"]
		else:
			vlIndex = []
		
		return False, vlIndex, vdFeatureParams, vdNomNormConfig, vdScaleConfig
		
	
	
	def loadFeatureValues(self, pFeatureName, plValues):
		'''
		Loads feature values.
		'''
		
		if self.datasetSize == 0:
			self.datasetSize = len(plValues)
		else:
			if self.datasetSize != len(plValues):
				raise Exception("Data size mismatch for %s: %s vs. %s" % (pFeatureName, len(plValues), self.datasetSize))
		
		self.features[pFeatureName][1].loadValues(plValues)
		
	
	
	def getFeatures(self):
		'''
		Returns list of features created.
		'''
		
		return [item[1] for item in self.features.itervalues()]
		
	
	
	def genFeatureVectors(self):
		'''
		Generates and returns feature vectors and feature configuration.
		
		Feature indexes (see self.features description in self.features)
		are set here.
		
		A feature vector is a list of dictionaries with keys being the 
		feature index and the value a tuple in (feature name, feature value)
		format.
		
		NOTE: This method can be overridden by inherited classes to suit 
		the vector format to the learning algorithm/tool (e.g. svmfg). 
		'''
		
		## list of dictionaries with key being the feature index and values
		## tuples of form (feature name, features value)
		vldFVectors = []
		for i in range(self.datasetSize):
			vldFVectors.append({})
		
		## forming feature vectors
		## If the feature has no index yet, it will be assigned an index
		## list (with only one element if normal and several if binarized
		## one for each derived binary feature).
		
		vFCntr = 0                      # feature counter
		for vlIndex, vFeature in self.features.itervalues():
			if vFeature.binarized:
				# indexes
				if vlIndex == []:
					vFromIndex = vFCntr + 1
					vFCntr = vFromIndex + len(vFeature.nomBinFeatureNames) - 1
					vlIndex = [i for i in range(vFromIndex, vFCntr + 1)]
					self.features[vFeature.name] = (vlIndex, vFeature)
				## (name, value) pairs 
				for vIndex, vBinFeatureName in zip(vlIndex, vFeature.nomBinFeatureNames):
					vlValues = vFeature.nomBinFeatures[vBinFeatureName].getValues(self.scale)
					for i, vValue in enumerate(vlValues):
						## note the name format: main feature name concatenated with the binary feature name
						vldFVectors[i][vIndex] = ("%s=%s" % (vFeature.name, vBinFeatureName), vValue)
			else:
				# indexes
				if vlIndex == []:
					vFCntr += 1
					vlIndex = [vFCntr]
					self.features[vFeature.name] = (vlIndex, vFeature)
				# values
				vlValues = vFeature.getValues(self.scale)
				for i, vValue in enumerate(vlValues):
					vldFVectors[i][vlIndex[0]] = (vFeature.name, vValue)
		
		vConfig = self.dumpConfig()
		
		return vldFVectors, vConfig
		
	
	
	def getRawFVectors(self):
		'''
		Returns the raw feature vectors which are the original features
		and values before normalization and scaling
		'''
		
		## list of dictionaries with key being the feature name and values
		## the feature values
		vldRawFVectors = []
		for i in range(self.datasetSize):
			vldRawFVectors.append({})
		
		for vIdx, vFeature in self.features.iteritems():
			for i, vOrgValue in enumerate(vFeature.getOriginalValues()):
				vldRawFVectors[i][vFeature.name] = vOrgValue
		
		return vldRawFVectors
		
	
	
	def _round(self, pValue):
		'''
		Rounds value based on the rounding setting in the configuration.
		'''
		
		if self.gRoundDigits > -1:
			return round(pValue, self.gRoundDigits)
		else:
			return pValue
		
	
	
	def dumpConfig(self):
		'''
		Saves the configuration used in generating feature vectors.
		
		I the configuration is saved before loading feature values, it 
		will not contain the right configuration if there is a nominal
		feature or feature values should be scaled.
		
		The configuration is saved in YAML format. It may differs from the
		input configuration loaded by loadConfig() if there is at least
		one feature normalized or scaled. In this case, normalization
		setting such as binarized features, or scaling range will be added
		to the configuration. 
		
		The purpose is the exact replicability of the feature vector
		generation. This is particularly useful for having exact same 
		feature setting for training and test purposes. The reason is that
		normalization and scaling are dependent on the data, and same
		setting can end up in different outcome for different data. For
		example, scaling depends on the minimum and maximum of a feature 
		value which can be different for two data set. Therefore, a same
		value in training and test set can be scaled to different values
		which is not desired. 
		'''
		
		vdConfig = {}
		
		# global config
		vdConfig = self._dumpGlobalConfig()
		
		# individual feature config
		
		vdFeatureConfigs = {}
		
		for vlIndex, vFeature in self.features.itervalues():
			vdFeatureConfigs[vFeature.name] = vFeature.dumpConfig()
			# adding indexes
			vdFeatureConfigs[vFeature.name]["index"] = vlIndex
		
		vdConfig["features"] = vdFeatureConfigs
		
		return yaml.dump(vdConfig)
		
	
	
	def _dumpGlobalConfig(self):
		'''
		Saves global configuration used in generating feature vectors)
		'''
		
		vdConfig = {}
		
		vdConfig["global nominal normalization method"] = self.gNomNormMethod
		vdConfig["global nominal binarization frequency cutoff"] = self.gNomBinFreqCutoff
		vdConfig["scale"] = self.scale
		vdConfig["global scaling target range"] = [self.gScaleTrgRange[0], self.gScaleTrgRange[1]]
		vdConfig["rounding feature values"] = self.gRoundDigits
		
		return vdConfig
		
	
	


class Feature:
	'''
	Class for feature
	
	A feature includes its values for a set of instances and methods to 
	do processes on the values. If normalized, it also includes normalization
	setting.
	'''
	
	
	def __init__(self, pName, pdFeatureParams, pdNomNormConfig = None, pdScaleConfig = None):
		'''
		Constructor
		
		pdFeatureParams: feature-specific setting
		pdNomNormConfig: nominal normalization setting
		                 see setNomNorm()
		pdScaleConfig: feature value scaling seeting
		               see setScaling()
		'''
		
		# initialization: defaults are set inside setting functions
		
		self.name = pName
		self._values = []
		
		# featureParams used only in dumping configuration
		self.featureParams = pdFeatureParams
		
		self.nomNormMethod = None
		self.nomBinFeatures = None
		self.nomBinFreqCutoff = None
		self.nomBinFeatureNames = None
		self.binarized = None
		
		self.scalable = None
		self.srcRange = None
		self.trgRange = None
		
		## settings: scaling setting should be extracted before normalization
		## settings, because the latter may used the former in binarization
		 
		self.setScaling(pdScaleConfig)
		self.setNomNorm(pdNomNormConfig)
		
	
	
	def setScaling(self, pdConfig):
		'''
		Sets scaling setting
		
		pdConfig includes:
		  - "scalable": 
		     - True: scalable (default)
		     - False: not scalable
		  - "source range": see Feature.scale(): self.srcRange
		  - "target range": see Feature.scale(): self.trgRange
		                    Default is [0,1].
		'''
		
		if pdConfig == None:
			vdConfig = {}
		else:
			vdConfig = pdConfig
		
		if "scalable" in vdConfig:
			self.scalable = vdConfig["scalable"]
		else:
			self.scalable = True
		
		if self.scalable:
			if "source range" in vdConfig:
				self.srcRange = vdConfig["source range"]
			else:
				# will be set in scale() based on values
				self.srcRange = None
			if "target range" in vdConfig:
				self.trgRange = vdConfig["target range"]
			else:
				self.trgRange = (0, 1)
		else:
			self.srcRange = None
			self.trgRange = None
		
	
	
	def setNomNorm(self, pdConfig):
		'''
		Sets nominal normalization setting
		
		pdConfig includes:
		  - "method": nominal feature normalization method; 
		     - bin: binarization in which each nominal value is converted
		            to a binary feature.
		  - "bin freq cutoff": Frequency cutoff  for binarization
		                        Default is 0.
		                        See Feature.binarizeNominal()
		  - "bin features": list of binarized feature names
		                    See Feature.binarizeNominal()
		'''
		
		if pdConfig == None:
			vdConfig = {}
		else:
			vdConfig = pdConfig
		
		if "normalize" in vdConfig:
			self.normNominal = vdConfig["normalize"]
		else:
			## If not specified, it will be automatically set when values
			## are loaded: true for string values and false otherwise.
			self.normNominal = None
		
		if "method" in vdConfig and vdConfig["method"] == "num":
			self.nomNormMethod = "num"
		else:
			self.nomNormMethod = "bin"
		
		if "bin freq cutoff" in vdConfig:
			self.nomBinFreqCutoff = vdConfig["bin freq cutoff"]
		else:
			self.nomBinFreqCutoff = 0
		
		if "bin features" in vdConfig:
			self.nomBinFeatureNames = vdConfig["bin features"]
		else:
			self.nomBinFeatureNames = []
		
	
	
	def loadValues(self, plValues):
		'''
		Loads feature values.
		
		Feature values can be numbers (numerical), strings (nominal) and
		lists (multi-valued). Numerical features are loaded normally. Nominal
		features are normalized, even if normalization is not specified in
		feature configuration. Multi-valued features (e.g. n-grams) have
		a list of string values per instance. Therefore, they are normalized
		as well.
		'''
		
		self._values = plValues[:]
		
		## If normalization is not specified in the configuration when
		## creating the feature, it is automatically set here: true for
		## string and list values and false otherwise.
		if self.normNominal == None:
			if isinstance(self._values[0], (str, unicode, list)):
				self.normNominal = True
			else:
				self.normNominal = False
		
		if self.normNominal:
			self.normalizeNominal()
		
	
	
	def normalizeNominal(self):
		'''
		Normalizes nominal feature.
		'''
		
		if self.nomNormMethod == "bin":
			self.binarizeNominal()
		else:
			raise Exception("Unknown nominal normalization method!")
		
	
	
	def binarizeNominal(self):
		'''
		Binzarizes nominal feature based on pConfig.
		
		pConfig includes:
		  - "bin freq cutoff": if the number of instances for which a 
		                       derived binary feature fires is less than
		                       this fraction of the entire data set, the
		                       feature will be ignored.
		     - <Freq>: a value in [0, 1). For example, if it is set to 
		               0.01 and the feature fires for less than 1 percent
		               of the instances, it will be ignored.
		  - "bin features": list of binarized feature names which are in 
		                    fact a seb set of nominal values. If this 
		                    parameter is provided, the above parameters
		                    will be ignored. This is useful when exact 
		                    same set of binarized features to an existing
		                    setting is needed.
		'''
		
		## binarized features derived from this feature: a dictionary with
		## keys having binarized feature names and values having them as 
		## objects of type Feature
		self.nomBinFeatures = {}
		
		## Feature value of an instance can be a list (multi-valued features)
		## or non-list. However, they are all converted to lists for a unified
		## binarization process.
		if isinstance(self._values[0], list):
			vlValueLists = self._values
		else:
			vlValueLists = [[i] for i in self._values]
		
		## extracting/generating binary feature names if not given in the 
		## configuration
		if self.nomBinFeatureNames == []: 
			## If binary feature list is not given, extracted features 
			## should meet cutoff threshold.
			vCutoffThreshold = self.nomBinFreqCutoff * len(vlValueLists)
			self.nomBinFeatureNames = [name for name, count in util.groupBy([v for l in vlValueLists for v in set(l)]).iteritems() if count >= vCutoffThreshold]
			if len(self.nomBinFeatureNames) == 0:
				raise Exception("The frequency cutoff %s for feature binarization is too restrictive in %s !" % (self.nomBinFreqCutoff, self.name))
		
		# extracting binary feature values
		
		vdBinFeatureValues = {}
		for vlValues in vlValueLists:
			for vName in self.nomBinFeatureNames:
				if vName not in vdBinFeatureValues:
					vdBinFeatureValues[vName] = []
				
				if vName in vlValues:
					vdBinFeatureValues[vName].append(1)
				else:
					vdBinFeatureValues[vName].append(0)
		
		## creating binary features and loading their values
		## Note: the names of binary features are formed by concatenating
		##       original feature name and the binary feature name with 
		##       a dash. 
		##           e.g. root-S
		## Note: the binary features are not normalizable but inherit the
		##       scaling setting.
		for vCBFName, vlCBFValues in vdBinFeatureValues.iteritems():
			vBinFeatureName = self.name + '-' + vCBFName
			self.nomBinFeatures[vCBFName] = Feature(vBinFeatureName, 
			                                        pdFeatureParams = {},
			                                        pdNomNormConfig = {"normalize": False},
			                                        pdScaleConfig = {"scalable": self.scalable, "source range": self.srcRange, "target range": self.trgRange})
			self.nomBinFeatures[vCBFName].loadValues(vlCBFValues)
		
		self.binarized = True
		
	
	
	def scale(self):
		'''
		Scales feature values if the feature is scalable.
		
		The scaling is based on the scaling attributes:
		  - self.srcRange: the range to scale from. If not given, it will
		                   be set to the range of feature values (min and 
		                   max). If given (usually when we want to have 
		                   exact same scaling to a previous one, e.g. test
		                   like training), but the range of feature values 
		                   does not fall inside this range, it and the 
		                   target range will be adjusted accordingly. 
		                   For example, if this feature had values in range
		                   [10, 100] in  training data and scaled to [0, 1],
		                   these are given as source and target ranges in
		                   the configuration. Then, if in test set the range
		                   is [11, 120] or [5, 100], the source range will
		                   bw adjusted to [10, 120] or [5, 100] and the 
		                   target range to [0, 1.2] or [-0.05, 1].
		                   The format is:
		                     - <[X, Y]>: X is the lower and Y is the upper
		                                 bound of the range (inclusive).
		  - self.trgRange: the range to scale to. In the
		                   latter case, this range will be adjusted and 
		                   scaled to.
		                   The format is:
		                     - <[X, Y]>: X is the lower and Y is the upper
		                                 bound of the range (inclusive).
		'''
		
		vlValues = self._values
		
		if not self.scalable:
			return vlValues
		
		vlScaledValues = []
		
		# range of values
		vNewSrcRange = (min(vlValues), max(vlValues))
		
		if self.srcRange == None:
			self.srcRange = vNewSrcRange
		else:
			## adjusting the target range if range of values is not fully 
			## covered by given source range
			
			if vNewSrcRange[0] < self.srcRange[0]:
				vDiff = self.srcRange[0] - vNewSrcRange[0]
				vScale = 1.0 * (self.trgRange[1] - self.trgRange[0]) / (self.srcRange[1] - self.srcRange[0])
				self.srcRange = (vNewSrcRange[0], self.srcRange[1])
				self.trgRange = (self.trgRange[0] - vDiff * vScale, self.trgRange[1])
			
			if vNewSrcRange[1] > self.srcRange[1]: 
				vDiff = vNewSrcRange[1] - self.srcRange[1]
				vScale = 1.0 * (self.trgRange[1] - self.trgRange[0]) / (self.srcRange[1] - self.srcRange[0])
				self.srcRange = (self.srcRange[0], vNewSrcRange[1])
				self.trgRange = (self.trgRange[0], self.trgRange[1] + vDiff * vScale)
		
		for vValue in vlValues:
			vlScaledValues.append(util.scaleNumber(vValue, self.srcRange, self.trgRange))
		
		return vlScaledValues
		
	
	
	def getValues(self, pflgScale = False):
		'''
		Returns feature values.
		
		Values are optionally scaled before returning.  
		'''
		
		if pflgScale:
			return self.scale()
		else:
			return self._values
		
	
	
	def getOriginalValues(self):
		'''
		Returns original feature values (before normalization).
		'''
		
		return self._values
		
	
	
	def dumpConfig(self):
		'''
		Saves the feature configuration used in generating feature vectors
		'''
		
		vdConfig = {}
		
		vdConfig["parameters"] = self.featureParams
		vdConfig["normalize nominal"] = self.normNominal
		if self.normNominal:
			vdConfig["nominal normalization method"] = self.nomNormMethod
			if self.nomNormMethod == "bin":
				vdConfig["binarized features"] = self.nomBinFeatureNames
				vdConfig["nominal binarization frequency cutoff"] = self.nomBinFreqCutoff
		
		vdConfig["scalable"] = self.scalable
		if self.scalable:
			if self.binarized:
				# if binarized, get its binary features' setting for scaling
				vFeature = self.nomBinFeatures[self.nomBinFeatureNames[0]]
			else:
				vFeature = self
			
			if vFeature.srcRange != None:
				vdConfig["scaling source range"] = [vFeature.srcRange[0], vFeature.srcRange[1]]
			vdConfig["scaling target range"] = [vFeature.trgRange[0], vFeature.trgRange[1]]
		
		return vdConfig
		
	
	
