#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""	
	This module defines functions and classes to perform grid search for
	finding optimum hyper-parameters for machine learning algorithms.
	
	Version 0.2												(10-Mar-2017)
	- Parameter specifications can now simply be a list of search values,
	 in addition to a specification dictionary.
	- An arbitrary number of return values are allowed now but the parameter
	 evaluator function, which are added into the returned best grid structure.
	
	Version 0.1								                (22-Oct-2015)
	- _generateParamSearchSpace() is changed to consider base 0 as no 
	  base.
	- A redundant line in GridSearch.__init__() which set the pfParamEvaluator
	  is removed as it is already done by setParamEvaluator().
	  
	Version 0.1								(10-Jul-2015 to 22-Jul-2015)
	- GridSearch and Grid are added.
	
"""

import itertools, os, sys, multiprocessing, math, shutil


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# GridSearch
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class GridSearch:
	'''
	Class for searching though machine learning hyper-parameters through
	a grid
	'''
	
	
	def __init__(self, pdParamSpec, pfParamEvaluator, pParamEvalArgs, pID = ''):
		'''
		Constructor
		
		pdParamSpec describes the parameters and their search specifications
		(see Grid documentation).
		'''
		
		## ID can help to identify the output in case of running for
		## different data sets at the same time (e.g. if the jobs crash 
		## log files will be identifiable).
		self.id = pID
		
		# creating the grid
		self.grid = Grid(pdParams = pdParamSpec).grid
		
		# setting the evaluator function and arguments (see setParamEvaluator())
		self.setParamEvaluator(pfParamEvaluator = pfParamEvaluator, pParamEvalArgs = pParamEvalArgs)
		
	
	
	@property
	def gridSize(self):
		'''
		Returns the grid size which is the number of rows
		'''
		
		return len(self.grid)
		
	
	
	def setParamEvaluator(self, pfParamEvaluator, pParamEvalArgs):
		'''
		Sets the evaluator function to be used at each search step to
		evaluate the parameter set
		
		Evaluator function is used to evaluate the parameter set at each
		step. The evaluator function helps customize the grid search for
		any arbitrary machine learning tool and algorithm. Its alternative 
		could be to make GridSearch and abstract class and derive from it
		for every machine learning tool/algorithm to customize the evaluation
		of the parameter sets. 
		
		The evaluator function must accept 4 arguments, one to pass the
		parameters of the search step, 2 for output and error logging devices, 
		and finally one for passing its own specific arguments. The last
		argument can be in a form of dictionary or list to contain multiple
		arguments if required.
		
		The evalutor function can return an arbitrary number of values, but 
		the first one mush be the evaluation scores.
		'''
		
		self.paramEvaluator = pfParamEvaluator
		self.paramEvalArgs = pParamEvalArgs
		
	
	
	def search(self, pProcNum = 1, pflgDelTempDir = False, pSearchLogFilename = ''):
		'''
		Searches for the optimum set of parameters in the grid and returns
		the best as well as the worst parameter setting per evaluation
		metric
		
		It also returns the resulting grid.
		
		The evaluation of the parameter set at each search step is done by
		the function assigned for this purpose as a class member (see the
		constructor or setParamEvaluator()).
		'''
		
		# creating a temporary directory for logs and other temporary files
		
		vTempDir = "_%s.%s" % (self.id, str(os.getpid()))
		if not os.path.exists(vTempDir):
			os.makedirs(vTempDir)
		
		
		# multiprocessing
		
		## If the number of processes requested is larger than the number 
		## of steps, it will be overridden
		vStepNum = self.gridSize
		if pProcNum > vStepNum:
			vProcNum = vStepNum
		else:
			vProcNum = pProcNum
		
		# multiprocessing manager
		vMPManager = multiprocessing.Manager()
		
		# queue for output with multiprocessing
		vqOut = vMPManager.Queue()
		
		# list of processes
		vlProcs = []
		# number of steps to be processed by each process
		vStepsPerProc = int(math.ceil(vStepNum / float(vProcNum)))
		# submitting processes
		for vProcNo in range(vProcNum):
			vProc = multiprocessing.Process(target=self._search, args=(vProcNo + 1, self.grid[vProcNo * vStepsPerProc : (vProcNo + 1) * vStepsPerProc], vqOut, vTempDir))
			vlProcs.append(vProc)
			vProc.start()
		
		# collecting results
		vlResultGrid = []
		vlProcOutLog = vlProcErrLog = []
		for vProcNo in range(vProcNum):
			vlResultGrid += vqOut.get()
		
		# waiting for all worker processes to finish
		for vProc in vlProcs:
			vProc.join()
		
		if pflgDelTempDir:
			shutil.rmtree(vTempDir)
		
		# sorting the steps based on step number
		vlResultGrid.sort(key=lambda item: item[0])
		
		# finding and returning the best/worst parameters per metric
		
		vdBestWorstParams = self._extractBestWorstParams(plProcessedGrid = vlResultGrid)
		
		# logging search if requested
		if pSearchLogFilename != '':
			self.logSearch(plProcessedGrid = vlResultGrid, pdBestWorstParams = vdBestWorstParams, pLogFilename = pSearchLogFilename)
		
		return vdBestWorstParams, vlResultGrid
		
	
	
	def _search(self, pProcID, plParamValues, pqOutput, pTempDir):
		'''
		Worker process for searching through the specified parameter-value
		set
		'''
		
		# setting up log files for the process
		
		vfOutLog = open("%s/p%s.log" % (pTempDir, pProcID), "w")
		vfErrLog = open("%s/p%s.err" % (pTempDir, pProcID), "w")
		
		# search through the space 
		
		vlSearchSteps = []
		
		for vStep, vdParamValueSet in plParamValues:
			# logging the search setting for the step
			vfOutLog.write("Step %s:\n\n" % vStep)
			vfOutLog.write("%s\n\n" % '\t'.join(["-%s=%s" % (str(param), str(value)) for param, value in vdParamValueSet.iteritems()]))
			
			# doing the evaluation for the step
			vdScores, vRest = self.paramEvaluator(vdParamValueSet, vfOutLog, vfErrLog, self.paramEvalArgs)
			
			# collecting the scores of the evaluation for the step
			vlSearchSteps.append((vStep,
								  vdParamValueSet,
								  vdScores,
								  vRest))
		
		vfOutLog.flush()
		vfErrLog.flush()
		
		pqOutput.put(vlSearchSteps)
		
	
	
	def _extractBestWorstParams(self, plProcessedGrid):
		'''
		Extracts and returns the best and worst parameter sets based on 
		each metric
		
		It has a prior knowledge about how to sort various metrics. For
		example, it knows that MAE and RMSE are sorted in the ascending 
		order meaning that the best value is the smallest of them.
		
		TODO: add more metrics with ascending order sort
		
		The results are returned in a dictionary where entries are metrics.
		For each metric, a dictionary is given in turn where the entries 
		are best and worst grid rows (search step) for the metric. The 
		following is an example:
		
		{"Accuracy": {"best": (34, {'a': 0.5}, {"Accuracy": 88.5, "F1": 75.3})}}
		'''
		
		vlMetrics = plProcessedGrid[0][2]
		vdBestWorstParams = {}
		
		for vMetric in vlMetrics:
			if vMetric.lower() in ["mae", "rmse"]:
				vdBestWorstParams[vMetric] = {"best" : min(plProcessedGrid, key = lambda gr: gr[2][vMetric]),
				                              "worst": max(plProcessedGrid, key = lambda gr: gr[2][vMetric])}
			else:
				vdBestWorstParams[vMetric] = {"best" : max(plProcessedGrid, key = lambda gr: gr[2][vMetric]),
				                              "worst": min(plProcessedGrid, key = lambda gr: gr[2][vMetric])}
		
		return vdBestWorstParams
		
	
	
	def logSearch(self, plProcessedGrid, pdBestWorstParams, pLogFilename, pScoreRoundDigits = 2):
		'''
		Logs the search steps and the best settings into file
		'''
		
		vfLog = open(pLogFilename, 'w')
		
		# logging best/worst
		
		for vMetric in sorted(pdBestWorstParams):
			vfLog.write("%s:\tBest : %-5s (step %s) with %s\t(other scores: %s)\n" % 
			            (str(vMetric).ljust(16), 
			             round(pdBestWorstParams[vMetric]["best"][2][vMetric], pScoreRoundDigits), 
			             pdBestWorstParams[vMetric]["best"][0], 
			             '\t'.join(["%s: %s" % (p, pdBestWorstParams[vMetric]["best"][1][p]) for p in sorted(pdBestWorstParams[vMetric]["best"][1])]), 
			             '\t'.join(["%s: %s" % (m, round(pdBestWorstParams[vMetric]["best"][2][m], pScoreRoundDigits)) for m in sorted(pdBestWorstParams[vMetric]["best"][2]) if m != vMetric])))
			vfLog.write("%s\tWorst: %-5s (step %s) with %s\t(other scores: %s)\n\n" % 
			            (''.ljust(16),
			             round(pdBestWorstParams[vMetric]["worst"][2][vMetric], pScoreRoundDigits), 
			             pdBestWorstParams[vMetric]["worst"][0], 
			             '\t'.join(["%s: %s" % (p, pdBestWorstParams[vMetric]["worst"][1][p]) for p in sorted(pdBestWorstParams[vMetric]["worst"][1])]), 
			             '\t'.join(["%s: %s" % (m, round(pdBestWorstParams[vMetric]["worst"][2][m], pScoreRoundDigits)) for m in sorted(pdBestWorstParams[vMetric]["worst"][2]) if m != vMetric])))
		
		vfLog.write("---------------------------------------------------\n\n")
		
		# logging grid
		
		for vStep in plProcessedGrid:
			vfLog.write("Step %s: %s\t%s\n" %
			            (str(vStep[0]).ljust(len(str(len(plProcessedGrid)))),
			             '\t'.join(["%s: %-10s" % (p, vStep[1][p]) for p in sorted(vStep[1])]), 
			             '\t'.join(["%s: %-5s"    % (m, round(vStep[2][m], pScoreRoundDigits)) for m in sorted(vStep[2])])))
		
		vfLog.write("\n")
		vfLog.close()
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# Grid
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class Grid():
	'''
	Class for parameter search grid
	'''
	
	
	def __init__(self, pdParams):
		'''
		Constructor
		'''
		
		self.grid = self._generateGrid(pdParams = pdParams)
		
	
	
	def _generateGrid(self, pdParams):
		'''
		Generates and returns the search grid based on parameter specifications
		
		The grid is a multidimensional list which is the Cartesian product 
		of sets of values for each parameter. 
		
		The parameter specifications are provided through list which contains parameter
		values or a dictionary where the keys are the parameter names as recognized by 
		the machine learning tool and the values are dictionaries of search specifications 
		for the parameter (see _generateParamSearchSpace() documentation). 
		'''
		
		# dictionary of list of search values for each parameter
		vdlParamSearchSpace = {}
		
		# 1. generating search space for each parameter
		
		for vParam, vPSpec in pdParams.iteritems():
			if isinstance(vPSpec, list):
				vdlParamSearchSpace[vParam] = vPSpec
			elif isinstance(vPSpec, dict):
				vdlParamSearchSpace[vParam] = self._generateParamSearchSpace(vPSpec)
		
		
		# 2. generating grid by Cartesian product of the search spaces
		
		## The parameter search spaces are in dictionary format, e.g.
		## 
		##  {'c': [1, 2], 
		##   'g': [10, 20]}
		##
		## The grid for this example needs to look like:
		##   
		##  [{'c': 1, 'g': 10},
		##   {'c': 1, 'g': 20},
		##   {'c': 2, 'g': 10},
		##   {'c': 2, 'g': 20}]
		##
		## To use Python Cartesian product for generating the grid, we need 
		## to first:
		
		## 2.1. convert the parameter search space dictionary to list of 
		##      list:
		##  
		##  [[{'c':  1}, {'c':  2}],
		##   [{'g': 10}, {'g': 20}]]
		
		vlldParamSearchSpace = []
		
		for vParam, vlSpace in vdlParamSearchSpace.iteritems():
			vldParamSearchSpace = []
			
			for vPoint in vlSpace:
				vldParamSearchSpace.append({vParam: vPoint})
			
			vlldParamSearchSpace.append(vldParamSearchSpace)
		
		## 2.3. do the Cartesian product on this list of list to get:
		##
		##  [({'c': 1}, {'g': 10}),
		##   ({'c': 1}, {'g': 20}),
		##   ({'c': 2}, {'g': 10}),
		##   ({'c': 2}, {'g': 20})]
		
		vldInitialGrid = itertools.product(*vlldParamSearchSpace)
		
		## 2.4. merge individual dictionaries of search points to finally 
		##      get:
		##
		##  [(1, {'c': 1, 'g': 10}),
		##   (2, {'c': 1, 'g': 20}),
		##   (3, {'c': 2, 'g': 10}),
		##   (4, {'c': 2, 'g': 20})]
		
		vldGrid = []
		
		for vRow, vGridRow in enumerate(vldInitialGrid, start = 1):
			vldGrid.append((vRow, {p: v for vdPoint in vGridRow for p, v in vdPoint.iteritems()}))
		
		return vldGrid
		
	
	
	def _generateParamSearchSpace(self, pdSearchSpec):
		'''
		Generates and returns the search space for the paramter based on 
		the search specification dictionary
		
		The search specification dictionary determines the search range 
		for the parameter (inclusive), the step and optionally a base 
		value if the growth is to be exponential instead of linear.
		
		For example, the following search specification means tha the growth
		is exponential with the base of 2 (2^-10, 2^-9.9, ..., 2^10).
		If the base was not specified, the growth would be linear.
		(-10, -9.9, ..., 10).
		
		{"range": (-10, 10),
		 "step": 0.1,
		 "base": 2}}
		'''
		
		if pdSearchSpec["range"][0] > pdSearchSpec["range"][1]:
			raise Exception("Range is not valid!") 
		
		vPoint = round(pdSearchSpec["range"][0], 5)
		vEnd = round(pdSearchSpec["range"][1], 5)
		vStep = pdSearchSpec["step"]
		
		vlSearchSpace = []
		
		while vPoint <= vEnd:
			# for exponential growth
			if "base" in pdSearchSpec and pdSearchSpec["base"] != 0:
				vlSearchSpace.append(round(pdSearchSpec["base"]**vPoint, 5))
			# for linear growth
			else:
				vlSearchSpace.append(round(vPoint, 5))
			
			vPoint = round(vPoint + vStep, 5)
		
		return vlSearchSpace
		
	
	
	@property
	def size(self):
		'''
		Returns the grid size which is the number of rows
		'''
		
		return len(self.grid)
		
	
	
