#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""	
	This module defines functions and classes for k-Nearest Neighbor algorithm.
	
	Version 0.1								(13-Jul-2015 to 05-Oct-2015)
	- kNNClassifier is added.
	
"""

import math
from utils import util



class kNNClassifier:
	'''
	Class for kNN classifier
	
	Based on the following desciption from MachineLearningMastery.com:
	
	The model for a kNN is the entire training dataset. When a prediction 
	is required for an unseen data instance, the kNN algorithm will search
	through the training dataset for the k most similar instances. The 
	prediction attribute of the most similar instances is summarized and 
	returned as the prediction for the unseen instance. The similarity 
	measure is dependent on the type of data. For real-valued data, the 
	Euclidean distance can be used. Other types of data such as categorical
	or binary data, Hamming distance can be used.
	
	In the case of regression problems, the average of the predicted attribute
	may be returned. In the case of classification, the most prevalent 
	class may be returned.
	
	The kNN algorithm belongs to the family of instance-based, competitive
	learning, lazy, non-parametric (non-linear) learning algorithms.
	'''
	
	
	def __init__(self, pTrainSet = None):
		'''
		Constructor
		
		The input training set is a data.Dataset object.
		'''
		
		self.trainSet = pTrainSet
		
	
	
	def loadTrainSet(self, pTrainSet):
		'''
		Loads training data set which is a data.Dataset object
		'''
		
		self.trainSet = pTrainSet
		
	
	
	def classify(self, pInstance, pK = 1, pAlgorithm = "euclidean"):
		'''
		Classifies the given instance using pK nearest neighbors identified
		by the selected similarity algorithm
		
		The instance is a data.DataInstance object.
		'''
		
		# 1. finding the nearest neighbors
		
		vlNeighbors = self.findNeighbors(pInstance = pInstance, pK = pK, pAlgorithm = pAlgorithm)
		
		# 2. finding the most common nearest neighbor class (voting)
		
		vdNeighborClasses = util.groupBy([n.target for n in vlNeighbors])
		vSortedClasses = sorted(vdNeighborClasses, key = lambda k: vdNeighborClasses[k], reverse=True)
		
		return vSortedClasses[0]
		
	
	
	def findNeighbors(self, pInstance, pK, pAlgorithm):
		'''
		Finds and returns the pK nearest neighbors in the training set to
		a given instance
		
		The instance is a data.DataInstance object.
		'''
		
		# 1. calculating distances
		
		vlDistances = []
		
		for vTrainInstance in self.trainSet.instances:
			vDistance = self._distance(pInstance1 = pInstance, pInstance2 = vTrainInstance, pAlgorithm = pAlgorithm)
			vlDistances.append((vTrainInstance, vDistance))
		
		vlDistances.sort(key = lambda x: x[1])
		
		# 2. finding neighbors
		
		vlNeighbors = []
		
		for i in range(pK):
			vlNeighbors.append(vlDistances[i][0])
		
		return vlNeighbors
		
	
	
	def _distance(self, pInstance1, pInstance2, pAlgorithm = "euclidean"):
		'''
		Calculates and returns the distance between two instances using 
		the selected algorithms
		
		Supported algorithms are:
		- (euc)lidean: Euclidean distance
		'''
		
		if pAlgorithm.lower().startswith("euc"):
			return self._euclideanDistance(pInstance1 = pInstance1, pInstance2 = pInstance2)
		
	
	
	def _euclideanDistance(self, pInstance1, pInstance2):
		'''
		Calculates and returns the Euclidean distance between two instances
		
		The instances are data.DataInstance objects.
		'''
		
		vDistance = 0
		
		for vF1, vF2 in zip(pInstance1.fv, pInstance2.fv):
			vDistance += pow((vF1.value - vF2.value), 2)
		
		return math.sqrt(vDistance)
		
	
	
