"""
	This module defines classes to convert multiclass classification data
	sets to binary ones.
	
	Version 0.5												(25-May-2017 to 31-May-2017)
	- Changes are made to accept lists as input in addition to file names.
	- Output format is added to OVA and OVO dataset generators.
	- mergeSVMOVAPreds() is changed to accept predictions in lists in addition
	  to files. 
	
	Version 0.4												(23-Jul-2015)
	- OVO.generateOVODatasets() is modified to make class order optional.
	- Bug is fixed in OVA.generateOVODatasets() which over-generated class
	  label pairs due to error in inner loop.
	
	Version 0.3												(06-Jul-2015)
	- OVO.mergeSVMOVOPreds() and OVA.mergeSVMOVAPreds() are modified to 
	  return ties.
	
	Version 0.2								(29-Jun-2015 to 30-Jun-2015)
	- MC2BC is added, which abstracts away loading data.
	- OVA is now inherited from MC2BC
	- OVO is added
	
	Version 0.1								(23-Jun-2015 to 25-Jun-2015)
	- OVA is added
	
"""

import re


class MC2BC:
	'''
	Abstract class for converting multiclass to binary classification
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		pass;
		
	
	
	def _loadMCDataset(self, pMCDataset, pInputFormat):
		'''
		Loads and returns multiclass dataset in the given format
		
		The input format can be one the following:
		- SVMLight: one instance per line in <class label> <feature vector> format 
		- anygram-tuple: one instance per line in a (class label, text) tuple format
		- More formats to do
		
		The data is returned in a list of original instance tuples in 
		(class label, feature vector) format
		'''
		
		if isinstance(pMCDataset, str):
			vlInstances = open(pMCDataset).read().strip().split('\n')
		elif isinstance(pMCDataset, list):
			vlInstances = pMCDataset
		
		if pInputFormat.lower().startswith("svmlight"):
			return self._loadFromCLFVDataset(vlInstances)
		elif pInputFormat.lower().startswith("anygram-tuple"):
			return vlInstances
		
	
	
	def _loadFromCLFVDataset(self, plInstances):
		'''
		Loads and returns dataset in <class label> <feature vector> format
		
		The data is returned in a list of original instance tuples in 
		(class label, feature vector) format
		'''
		
		vlInstances = []
		
		for vILine in plInstances:
			if vILine.startswith('#'):
				continue
			
			vlInstances.append(re.search("^([^ \t]+)[ \t]+(.+)", vILine).group(1, 2))
		
		return vlInstances
		
	
	


class OVA(MC2BC):
	'''
	One-vs-all multiclass to binary converter
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		MC2BC.__init__(self);
		
	
	
	def generateOVADatasets(self, pMCDataset, pInputFormat, pOutputFormat):
		'''
		Generates and returns the OVA datasets by converting the given 
		multiclass dataset
		
		The input format can be one the following:
		- SVMLight: one instance per line in <class label> <feature vector>
		            format 
		- More formats to do
		
		The return value is a dictionary the keys of which are the original
		class names and the values tuples of corresponding datasets (depending 
		on putput format specified) and the number of positive and negative 
		examples in each dataset.
		
		The output formats are:
		- string: class and feature vector are merged in a string
		- tuple: class and feature vector are in a tuple
		
		'''
		
		vlInstances = self._loadMCDataset(pMCDataset, pInputFormat)
		
		vsClasses = set([c for c, fv in vlInstances])
		
		vdDatasets = {}
		
		for vClass in vsClasses:
			
			vlOVAInstances = []
			vPosCount = 0
			vNegCount = 0
			
			for vCL, vFV in vlInstances: 
				if vCL == vClass:
					if pOutputFormat.lower().startswith("str"):
						vlOVAInstances.append("+1\t%s" % vFV)
					elif pOutputFormat.lower() == "tuple":
						vlOVAInstances.append(("+1", vFV))
					vPosCount += 1
				else:
					if pOutputFormat.lower().startswith("str"):
						vlOVAInstances.append("-1\t%s" % vFV)
					elif pOutputFormat.lower() == "tuple":
						vlOVAInstances.append(("-1", vFV))
					vNegCount += 1
			
			vdDatasets[vClass] = (vlOVAInstances, vPosCount, vNegCount)
			
		
		return vdDatasets
		
	
	
	def mergeSVMOVAPreds(self, pdPreds):
		'''
		Merges the predictions of OVA binary SVM classifiers
		
		The choice of the classifier is based on the magnitude of the score
		given to an instance by a classifier. Ties are not currently handles,
		so they are up to the decision made by python in sorting the dictionary.
		The method also returns these ties.
		
		The predictions of binary classifiers are given in a dictionary
		where the keys are the multiclass class labels and the values 
		their corresponding prediction files/lists.
		'''
		###################################
		# change the pdPredsFilenames to pdPreds and exchange the dictionary
		# key/value to comply with OVO merging
		vldPreds = []
		
		for i, (vCL, vPredFilename) in enumerate(pdPreds.iteritems()):
			for j, vMargin in enumerate(open(vPredFilename).read().strip().split('\n')):
				if i == 0:
					vldPreds.append({vCL: vMargin})
				else:
					vldPreds[j][vCL] = vMargin
		####################
		# change: both vMargin above to float(vMargin)
		
		vldMergedPredictions = []
		vldTies = []
		
		for vdPreds in vldPreds:
			## We sort the classes in the descending order of their prediction
			## margins. If there are class with positive predicted margin, 
			## the one with the biggest margin will rank first and selected.
			## If there are no class with a positive margin, i.e. the instance
			## is a negative one by all classifiers, the one with the smallest
			## margin (absolute predicted margin value) will be the selected
			## class for the instance because the corresponding classier
			## has been least sure that that instance is not a positive 
			## sample. Since the smallest absolue value is the biggest
			## negative value among all-negative predicted margin, sorting
			## them in decending order will still give us the selected 
			## class from the top of the list. So, the following works.
			vlClassOrder = sorted(vdPreds, key = lambda k: float(vdPreds[k]), reverse=True)
			
			vldMergedPredictions.append({"class": vlClassOrder[0], "margin": vdPreds[vlClassOrder[0]]})
			
			# extracting ties
			if vdPreds[vlClassOrder[0]] == vdPreds[vlClassOrder[1]]:
				vldTies.append(vdPreds)
		
		return vldMergedPredictions, vldTies
		
	
	


class OVO(MC2BC):
	'''
	One-vs-one multiclass to binary converter
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		MC2BC.__init__(self);
		
	
	
	def generateOVODatasets(self, pMCDatasetFilename, pInputFormat, pOutputFormat, plClassOrder = None):
		'''
		Generates and returns the OVA datasets by converting the given 
		multiclass dataset
		
		The class order list specifies the order in which the classes are 
		processed. For example, the first class in this list will be positive
		example in all the combinations with the rest. If not given, an
		ad-hoc order will be used.
		
		The input format can be one of the following:
		- SVMLight: one instance per line in <class label> <feature vector>
		            format 
		- More formats to do
		
		The constructed OVO data sets are returned in a dictionary with 
		the keys being the original class names. The values are dictionaries 
		in turn where the keys include the rest of the classes and the values
		are tuples of corresponding datasets (i.e. key1 vs. key2 classifier
		data as list of instances in string format) and the number of positive
		(key 1 class) and negative (key 2 class) examples in each dataset.
		'''
		
		vlInstances = self._loadMCDataset(pMCDatasetFilename, pInputFormat)
		
		# setting class order
		
		if plClassOrder == None or plClassOrder == []:
			vlClassOrder = list(set([i[0] for i in vlInstances]))
		else:
			vlClassOrder = plClassOrder
		
		vdDatasets = {}
		
		for i, vClass1 in enumerate(vlClassOrder):
			vdDatasets[vClass1] = {}
			
			for vClass2 in vlClassOrder[i+1 : ]:
				if vClass2 == vClass1:
					continue
				
				vlOVOInstances = []
				vPosCount = 0
				vNegCount = 0
				
				for vCL, vFV in vlInstances: 
					if vCL == vClass1:
						if pOutputFormat.lower().startswith("str"):
							vlOVOInstances.append("+1\t%s" % vFV)
						elif pOutputFormat.lower() == "tuple":
							vlOVOInstances.append(("+1", vFV))
						vPosCount += 1
					elif vCL == vClass2:
						if pOutputFormat.lower().startswith("str"):
							vlOVOInstances.append("-1\t%s" % vFV)
						elif pOutputFormat.lower() == "tuple":
							vlOVOInstances.append(("-1", vFV))
						vNegCount += 1
				
				vdDatasets[vClass1][vClass2] = (vlOVOInstances, vPosCount, vNegCount)
		
		return vdDatasets
		
	
	
	def mergeSVMOVOPreds(self, pdPreds):
		'''
		Merges the predictions of OVO binary SVM classifiers
		
		The choice of the classifier is based on voting by the number of 
		times an instance falls into a class. For ties, the sum of absolute 
		values of scores returned by the binary classifiers voting for the 
		winning classes are used. The method also returns these ties.
		
		The predictions of binary classifiers are given in a dictionary
		where the keys are the multiclass class label pairs and the 
		values their corresponding prediction files/lists.
		'''
		
		vldPreds = []
		vsClasses = set()
		
		for i, (vCLs, vPreds) in enumerate(pdPreds.iteritems()):
			if isinstance(vPreds, str):
				vlPreds = open(vPreds).read().strip().split('\n')
			else:
				vlPreds = vPreds
			
			for j, vMargin in enumerate(vlPreds):
				if i == 0:
					vldPreds.append({vCLs: float(vMargin)})
				else:
					vldPreds[j][vCLs] = float(vMargin)
			
			# collecting class labels
			vsClasses.add(vCLs[0])
			vsClasses.add(vCLs[1])
		
		vldMergedPredictions = []
		vldTies = []
		
		for vdPreds in vldPreds:
			## For the current instance, the number of times a class label 
			## classified as positive is counted (by adding the margin value
			## to a bin belonging to that class). At the end, the class 
			## with the maximum number is selected as the final predicted
			## class.
			
			# initializing the bins
			vdBins = {c: [] for c in list(vsClasses)}
			
			# filling bins
			for vCLs, vMargin in vdPreds.iteritems():
				if vMargin > 0:
					vdBins[vCLs[0]].append(vMargin)
				else:
					vdBins[vCLs[1]].append(vMargin)
			
			vlClassOrder = sorted(vdBins, key = lambda k: (len(vdBins[k]), sum([abs(float(s)) for s in vdBins[k]])), reverse=True)
			vldMergedPredictions.append({"class": vlClassOrder[0], "scores": vdBins[vlClassOrder[0]]})
			
			# extracting ties:
			if len(vdBins[vlClassOrder[0]]) == len(vdBins[vlClassOrder[1]]):
				vldTies.append(vdBins)
		
		return vldMergedPredictions, vldTies
		
	
