import tensorflow as tf
import numpy as np
import eval

class RNNSeqTagger:
	'''
	Class for sequence tagging with recurrent neural nets
	'''
	
	def __init__(self, pVerbosity=1):
		'''
		Constructor
		'''
		
		self.verbosity = pVerbosity
		
		self.weights = None                                # output weight vector (NOTE: other weights are handled internally by TF cells)
		self.biases = None                                 # output bias vector (NOTE: other biases are handled internally by TF cells)
		
		self.cellType = None                               # RNN cell type (a lower case string, e.g. basicrnncell)
		self.hiddenSize = None                             # hidden layer size
		self.activation = None                             # activation function (a lower case string, e.g. tanh)
		self.dropoutRate = None                            # input and output dropout rate to be used by tf.nn.rnn_cell.DropoutWrapper
		
		self.lRate = None                                  # learning rate
		self.momentum = None                               # momentum for SGD
		
		self.batchSize = None                              # training batch size
		self._batchCursor = 0                              # current cursor position in the data to generate the next batch
		
		self.epochNum = None                               # number of training epoches
		self._logEpochCount = 50                           # number of training epoches for which a log is printed
		self._logEpochInterval = None                      # the training epoch interval in which the log will be printed, i.e. every _logEpochInterval epoches
		self._validEpochCount = 50                         # number of training epoches for which the model is validated on the validation (dev) set
		self._validEpochInterval = None                    # the epoch interval in which the model will be validated, i.e. every _validEpochInterval epoches
		
		# data attributes
		
		self.trainSet = None                               # training dataset: SeqTagSet object (ml.seqtag)
		self.validationSet = None                          # validation dataset: SeqTagSet object (ml.seqtag)
		
		self.validationOutput = None                       # predicted validation dataset tagging (2D list of sequence tags)
		
		self.tokenIDs = None                               # vocab tokens and their IDs
		self.idTokens = None                               # vocab IDs and their tokens
		self.tagIDs = None                                 # tags and their IDs
		self.idTags = None                                 # tag IDs and their tags
		
		self.tagWeights  = None                            # loss calculation weights for each tag to handle tag distribution imbalance in the data 
		
		self._tokRepDim = None                             # dimension of the token representation vetcor (e.g. one-hot vector)
		self._tokRepGetter = None                          # a getter function placeholder to fetch the token representation: the first argument 
		                                                   # of this function should be the token form string.
		
	
	
	def setEpochNum(self, pEpochNum):
		'''
		Sets number of training epoches and related attributes 
		'''
		
		self.epochNum = pEpochNum
		
		if self.epochNum <= self._logEpochCount:
			self._logEpochInterval = 1
		else:
			self._logEpochInterval = self.epochNum / self._logEpochCount
		
		if self.epochNum <= self._validEpochCount:
			self._validEpochInterval = 1
		else:
			self._validEpochInterval = self.epochNum / self._validEpochCount
		
	
	
	def loadTrainData(self, pSeqTagSet, pflgPrintStat=False):
		'''
		Loads training data set which is a SeqTagSet object
		
		For text data, make sure the data is in the required case. 
		If lowercase is needed, it can be done via SeqTagSet.toLower()
		before loading.
		'''
		
		print "Loading training data ...\n"
		
		self.trainSet = pSeqTagSet
		
		# statistics
		if pflgPrintStat:
			print "  Number of sequences (e.g. sentences)  : %s" % self.trainSet.seqSize
			print "  Number of all token                   : %s" % self.trainSet.tokSize
			print "  Number of token types (unique tokens) : %s" % self.trainSet.vocabSize
			vlTokenFreqList = self.trainSet.getTokenFormFreqList()
			print "  Most frequent token                   : %s (%s times)" % (
			vlTokenFreqList[0][0], vlTokenFreqList[0][1])
			vHapaxCount = len([t for t in vlTokenFreqList if t[1] == 1])
			print "  Number of hapax tokens                : %s (%s%%)" % (
			vHapaxCount, vHapaxCount * 100 / self.trainSet.vocabSize)
			
			print
			
			print "  Tag set size : %s" % len(self.trainSet.getTagSet())
			for vTag, vFreq in self.trainSet.getTagFreqList():
				print "    %s: %s times (%s%%)" % (vTag, vFreq, vFreq * 100 / self.trainSet.tokSize)
		
		# translating data to indexes identifying the vocabulary and tags (NOTE: only for trianing data)
		self.indexData()
		
	
	
	def indexData(self):
		'''
		Indexes data including tokens and tags
		
		Indexing is done by assigning numerical IDs, starting from 0, to tokens and labels.
		
		NOTE: We start the tag IDs .from 1 instead if 0 to avoid conflict with padding zero.
		'''
		
		# indexing tokens
		
		self.tokenIDs = {}
		self.idTokens = {}
		for i, (token, count) in enumerate(self.trainSet.getTokenFormFreqList()):
			self.tokenIDs[token] = i
			self.idTokens[i] = token
		
		# indexing tags
		
		self.tagIDs = {}
		self.idTags = {}
		for i, (tag, count) in enumerate(self.trainSet.getTagFreqList()):
			self.tagIDs[tag] = i
			self.idTags[i] = tag
		
	
	
	def loadValidationData(self, pSeqTagSet, pflgPrintStat=False):
		'''
		Loads validation data set which is a SeqTagSet object
		
		For text data, make sure the data is in the required case. 
		If lowercase is needed, it can be done via SeqTagSet.toLower()
		before loading.
		'''
		
		print "Loading validation data ...\n"
		
		self.validationSet = pSeqTagSet
		
		# statistics
		if pflgPrintStat:
			print "  Number of sequences (e.g. sentences)  : %s" % self.validationSet.seqSize
			print "  Number of all token                   : %s" % self.validationSet.tokSize
			print "  Number of token types (unique tokens) : %s" % self.validationSet.vocabSize
			vlTokenFreqList = self.validationSet.getTokenFormFreqList()
			print "  Most frequent token                   : %s (%s times)" % (vlTokenFreqList[0][0], vlTokenFreqList[0][1])
			vHapaxCount = len([t for t in vlTokenFreqList if t[1] == 1])
			print "  Number of hapax tokens                : %s (%s%%)" % (vHapaxCount, vHapaxCount * 100 / self.validationSet.vocabSize)
			
			print
			
			print "  Tag set size : %s" % len(self.validationSet.getTagSet())
			for vTag, vFreq in self.validationSet.getTagFreqList():
				print "    %s: %s times (%s%%)" % (vTag, vFreq, vFreq * 100 / self.validationSet.tokSize)
			
			# also calculate OOV for validation set
		
	
	
	@property
	def tagSet(self):
		'''
		Returns the tag set in a list
		'''
		
		return list(self.tagIDs)
		
	
	
	def _getClassNum(self):
		'''
		Returns the number of classes
		'''
		
		return len(self.tagIDs)
		
	
	
	def getSequenceStr(self, pSequence, pTags):
		'''
		Returns a string representation of the given sequence in a columnar
		format where tokens are represented as token\ttag
		'''
		
		return '\n'.join(["    %s\t%s" % (tok, tag) if tok != 0 else '    #'
		                  for tok, tag in zip(list(pSequence), list(pTags))]) + '\n'
		
	
	
	def _getTokRep(self, pTokenFormStr):
		'''
		Returns token representation using the getter function in self._tokRepGetter
		'''
		
		vRep = self._tokRepGetter(pTokenFormStr)
		
		if len(vRep) == 0:
			return np.zeros(self._tokRepDim)
		else:
			return vRep
		
	
	
	def _getTokOneHotRep(self, pTokenFormStr):
		'''
		Returns the one-hot representation of the given token form string
		
		_onehotDim is used to set the dimension of the returned vector, which
		is the number of most frequent tokens to be used. If the given token
		is less frequent to fall in this range or it is not in the entire 
		vocabulary a zero vector is returned.
		'''
		
		aOneHot = np.zeros(self._tokRepDim)
		try:
			aOneHot[self.tokenIDs[pTokenFormStr]] = 1
		except (KeyError, IndexError):
			pass
		
		return aOneHot
		
	
	
	def train(self, pRNNCellType = "BasicRNN", pflgBidirectional= False, pHiddenSize = 64, pLayerNum = 1, 
			  pdTokRep = None, pdTagWeights = None,
		      pActivation = "tanh", pDropoutRate = 0.5, pLearningRate = 0.01, pMomentum = 0.5,
		      pBatchSize = 128, pEpochNum = 1000, pDevice = "/cpu:0",
		      pflgValidate = False):
		'''
		Trains the model
		
		pdTokRep is a dictionary which specifies the way tokens are representated in to the training algorithms:
		
		- {"rep": "one-hot", "dim": <dimension>} (default): one-hot vectors of given dimension are used. The dimension 
		   is specified using the dim key which in fact determines the maximum vocabulary size to be used. If not given,
		   the entire vocabulary is used which can be inefficient for large vocabularies.
		
		 - {"rep": "embedding", "dim": <dimension>, "fetcher": <fetcher_function>}: embedding vectors will be used to 
		   represent tokens. The dimension of the ebmeddings must be specified using the dim key. In addition, a function
		   which can be used to fetch the embedding vector of a token passed to it is required to be provided via
		   "fetcher" key of the dictionary. The function should accept the token form string as its first argument.
		   
		pdTagWeights is a dictionary assigning weights to each tag which is used in calculating loss. This is good when
		the tag distribution is highly imbalanaced and the contribution of the majority classes to the loss should be
		downgraded. Therefore, a smaller weight should be set to these classes proportional to their ratio on the data 
		set. 
		'''
		
		if self.trainSet is None:
			raise Exception("No training data loaded!")
		
		if pflgValidate:
			if self.validationSet is None:
				raise Exception("No validation data loaded!")
		
		# setting token representation
		
		if pdTokRep == None:
			pdTokRep = {}
		
		if pdTokRep == {} or pdTokRep["rep"].lower() in ["one-hot", "1-hot"]:
			if "dim" in pdTokRep:
				self._tokRepDim = pdTokRep["dim"]
			else:
				self._tokRepDim = self.trainSet.vocabSize
			
			self._tokRepGetter = self._getTokOneHotRep
		elif pdTokRep["rep"].lower() in ["embedding", "embeddings"]:
			if "dim" in pdTokRep:
				self._tokRepDim = pdTokRep["dim"]
			else:
				raise Exception("Dimension of the embeddings must be provided (using dim key)!")
			
			if "fetcher" in pdTokRep:
				self._tokRepGetter = pdTokRep["fetcher"]
			else:
				raise Exception("No embedding fetcher function is provided!")
		
		
		# setting tag weights
		
		if pdTagWeights is not None or pdTagWeights is not {}:
			self.tagWeights = pdTagWeights
		
		
		# setting RNN setting
		
		self.cellType = pRNNCellType.lower()
		self.bidirectional = pflgBidirectional
		self.hiddenSize = pHiddenSize
		self.activation = pActivation.lower()
		self.lRate = pLearningRate
		self.momentum = pMomentum
		self.dropoutRate = pDropoutRate
		
		
		# setting training setp settings
		
		self.batchSize = pBatchSize
		self.setEpochNum(pEpochNum)
		
		
		# creating the computation graph
		
		tgGraph = self._createGraph(pLayerNum = pLayerNum, pDevice = pDevice)
		
		
		# running the training procedure
		
		self._runTrain(tgGraph, pflgValidate = pflgValidate)
		
	
	
	def _createGraph(self, pLayerNum, pDevice):
		'''
		Creates the computation graph for trianing
		'''
		
		tgGraph = tf.Graph()
		
		with tgGraph.as_default(), tf.device(pDevice):
			tf.set_random_seed(7)
			
			# batch input data for training: (batch_size, batch_sequence_length, token_representation_dimension) shaped
			self._tphBatchInput = tf.placeholder(tf.float32, shape=[self.batchSize, None, self._tokRepDim])
			# flat unpadded batch labels: (batch_token_size) shaped
			self._tphBatchLabels = tf.placeholder(tf.int32, shape=[None])
			# lengths of sequences in the batch for input with variable sequence lengths
			self._tphBatchSeqLens = tf.placeholder(tf.int32, shape=[self.batchSize])
			# flat unpadded batch label weights for loss calculation
			self._tphBatchLabelWeights = tf.placeholder(tf.float32, shape=[None])
			
			# validation input data
			self._tphValidInput = tf.placeholder(tf.float32, shape=[None, None, self._tokRepDim])
			# lengths of sequences in the validaion data with variable sequence lengths
			self._tphValidSeqLens = tf.placeholder(tf.int32, shape=[None])
			
			# output weights and biases
			self.weights = tf.Variable(tf.random_uniform([self.hiddenSize, self._getClassNum()]))
			self.biases = tf.Variable(tf.random_uniform([self._getClassNum()]))
			
			# setting activation function
			if self.activation == "tanh":
				vActivation = tf.nn.tanh
			elif self.activation == "relu":
				vActivation = tf.nn.relu
			elif self.activation == "sigmoid":
				vActivation = tf.nn.sigmoid
			
			# creating RNN cell
			if self.cellType.startswith("basicrnn"):
				tCell = tf.nn.rnn_cell.BasicRNNCell(num_units=self.hiddenSize, activation=vActivation)
			elif self.cellType.startswith("gru"):
				tCell = tf.nn.rnn_cell.GRUCell(num_units=self.hiddenSize, activation=vActivation)
			elif self.cellType.startswith("basiclstm"):
				### handle other constructor parameters
				tCell = tf.nn.rnn_cell.BasicLSTMCell(num_units=self.hiddenSize, activation=vActivation)
			elif self.cellType.startswith("lstm"):
				### handle other constructor parameters
				tCell = tf.nn.rnn_cell.LSTMCell(num_units=self.hiddenSize, activation=vActivation)
			else:
				raise Exception("Cell type not recognized: %s" % self.cellType)
			
			
			# regularized cell                                    ToDo: make regularization method specified as option
			tRegCell = tf.nn.rnn_cell.DropoutWrapper(tCell, output_keep_prob = self.dropoutRate)
			
			# multi-layer RNN
			if pLayerNum > 1:
				tRegCell = tf.nn.rnn_cell.MultiRNNCell([tRegCell] * pLayerNum, state_is_tuple = True)
			
			
			# -- training -------------------------
			
			# running the RNN
			tBatchLogits = self._rnn(ptCell = tRegCell, pflgBidirectional = self.bidirectional, ptInputs = self._tphBatchInput, ptInputSeqLens = self._tphBatchSeqLens)
			
			# unpadded logits (to avoid problems with sparse_softmax_cross_entropy_with_logits)
			tBooleanMask = tf.sequence_mask(self._tphBatchSeqLens, tf.shape(self._tphBatchInput)[1])            # shape: (batch_size, batch_sequence_length)
			tBooleanMask = tf.reshape(tBooleanMask, [-1])
			tUPBatchLogits = tf.boolean_mask(tBatchLogits, tBooleanMask)
			
			# individual token losses
			self._tBatchTokLosses = tf.nn.sparse_softmax_cross_entropy_with_logits(tUPBatchLogits, self._tphBatchLabels)
			#tOneHotBatchLabels = tf.one_hot(indices = self._tphBatchLabels, depth = self._getClassNum()) 
			#self._tBatchTokLosses = tf.nn.weighted_cross_entropy_with_logits(tUPBatchLogits, tOneHotBatchLabels, pos_weight=0.15)
			
			# ToDo: apply weights to token losses to downgrade O class
			if self.tagWeights is not None:
				self._temp1 = self._tphBatchLabels
				self._temp2 = self._tBatchTokLosses
				self._temp3 = self._tphBatchLabelWeights
				self._tBatchTokLosses = self._tBatchTokLosses * self._tphBatchLabelWeights
			
			# masked losses to neutralize padded elements
			#tMask = tf.sign(tf.to_float(tBatchLabels))
			#self._tBatchTokLosses = tMask * tBatchTokLosses
			
			# indvidual sequence losses
			#tBatchSeqLosses = tf.reshape(self._tBatchTokLosses, [self.batchSize, -1])
			#tBatchSeqLosses = tf.reduce_sum(tBatchSeqLosses, reduction_indices = 1) / tf.to_float(self._tphBatchSeqLens)
			
			# average batch loss
			self._tBatchLoss = tf.reduce_mean(self._tBatchTokLosses)
			
			# optimizer: add support for other algorithms         ToDo: make optimization algorithm used specified as option
			self._toOptimizer = tf.train.MomentumOptimizer(learning_rate = self.lRate,
			                                               momentum = self.momentum).minimize(self._tBatchLoss)
			#self._toOptimizer = tf.train.AdamOptimizer(learning_rate = self.lRate).minimize(self._tBatchLoss)
			
			# predictions: reshaped back to (batch_size, batch_seq_length, class_num)
			############
			self._tOutput = tf.reshape(tBatchLogits, [self.batchSize, -1, self._getClassNum()])
			############
			#self._tBatchPreds = tf.nn.softmax(tBatchLogits)
			self._tBatchPreds = tf.reshape(tBatchLogits, [self.batchSize, -1, self._getClassNum()])
			
			
			# -- validation -------------------------
			#tf.get_variable_scope().reuse_variables()          # ToDo: remove or finalize; experimental to fix bugs 
			
			# running the RNN                                     ToDo: do we need unregularied cell for prediction below
			#tValidLogits = self._rnn(ptCell = tCell, ptInputs = self._tphValidInput, ptInputSeqLens = self._tphValidSeqLens)
			
			# predictions: reshaped back to (validation_set_size, validation_set_seq_length, class_num)
			#self._tValidPreds = tf.nn.softmax(tValidLogits)
			#self._tValidPreds = tf.reshape(self._tValidPreds, [tf.shape(self._tphValidInput)[0], -1, self._getClassNum()])
		
		return tgGraph
		
	
	
	def _rnn(self, ptCell, pflgBidirectional, ptInputs, ptInputSeqLens):
		'''
		Creates and runs a RNN network specfied by the given cell on the given input and returns the logits
		
		NOTE: To be able to multiply the output tensors of shape (batch_size, batch_seq_length, hidden_size) to the weights
		using matmul, the tensors are reshaped to rank 2. Consequently, the output is also of rank two and must be reshaped
		back latter in the calling procedure if needed.
		'''
		
		if pflgBidirectional:
			tOutput, _ = tf.nn.bidirectional_dynamic_rnn(cell_fw = ptCell,
		                                                 cell_bw = ptCell,
		                                                 inputs = ptInputs,
		                                                 sequence_length = ptInputSeqLens,
		                                                 dtype = tf.float32)
		else:
			tOutput, _ = tf.nn.dynamic_rnn(cell = ptCell,
		                                   inputs = ptInputs,
		                                   sequence_length = ptInputSeqLens,
		                                   dtype = tf.float32)
		
		# converting output tensor to matrix to use matmul: (batch_size * batch_seq_length, hidden_size)
		tOutput = tf.reshape(tOutput, [-1, self.hiddenSize])
		
		# logits
		return tf.matmul(tOutput, self.weights) + self.biases
		
	
	
	def _runTrain(self, ptgGraph, pflgValidate = False):
		'''
		Runs the training graph
		'''
		
		print "\n\nTraining ...\n"
		
		with tf.Session(graph = ptgGraph) as session:
			tf.initialize_all_variables().run()
			
			self._resetBatchGenerator()
			vLossSum = 0
			for epoch in range(1, self.epochNum + 1):
				# training (including validation settings) ----------------------
				
				############### ToDo we have to iterate here as many times as [train set size] / [batch size] 
				
				# generating training batch
				vaBatchInput, vaBatchLabels, vaBatchSeqLens, vllUPBatchLabels, vaBatchLabelWeights = self._generateBatch(pBatchSize = self.batchSize)
				
				#running the session based on whether validation should be performed or not
				if pflgValidate and (epoch == 1 or epoch % self._validEpochInterval == 0):
					# genertaing validation data set
					vaValidInput, vaValidLabels, vaValidSeqLen = self._generateValidSet()
					
					# preparing feeding dictionary to operations
					vdFeed = {self._tphBatchInput: vaBatchInput,
					          self._tphBatchLabels: vaBatchLabels,
					          self._tphBatchSeqLens: vaBatchSeqLens,
					          self._tphBatchLabelWeights: vaBatchLabelWeights,
					          self._tphValidInput: vaValidInput,
					          self._tphValidSeqLens: vaValidSeqLen}
					
					# running the optimizer and loss function (loss value is also needed)
					_, vBatchLoss, vaBatchPreds, vaValidPreds = session.run([self._toOptimizer, self._tBatchLoss, self._tBatchPreds, self._tValidPreds, self._temp1, self._temp2],
					                                                        feed_dict = vdFeed)
				else:
					# preparing feeding dictionary to operations
					vdFeed = {self._tphBatchInput: vaBatchInput,
					          self._tphBatchLabels: vaBatchLabels,
					          self._tphBatchSeqLens: vaBatchSeqLens,
					          self._tphBatchLabelWeights: vaBatchLabelWeights}
					
					# running the optimizer and loss function (loss value is also needed)
					_, vBatchLoss, vaBatchPreds, _aBatchTokLosses = session.run([self._toOptimizer, self._tBatchLoss, self._tBatchPreds, self._tBatchTokLosses],
					                                          feed_dict = vdFeed)
					
					#########################################
					#for t, l, w, wl  in zip(_temp1, _temp2, _temp3, _aBatchTokLosses):
					#	print self.idTags[t], l, w, wl
					#########################################
				
				# evaluating the batch
				if self.verbosity > 1 or pflgValidate:
					# adding up to epotch loss
					vLossSum += vBatchLoss
					
					if epoch == 1:
						print("Epoch 1: Average loss: %s" % vLossSum)
					elif epoch % self._logEpochInterval == 0:
						print("Epoch %s: Average loss: %s" % (epoch, vLossSum / self._logEpochInterval))
						vLossSum = 0
					
					if epoch == 1 or epoch % self._validEpochInterval == 0:
						# deleting the sequence padding from predictions and gold labels
						vllBatchPreds = self._delPadding(np.argmax(vaBatchPreds, 2), vaBatchSeqLens)
						
						# encoding the tag indexes to tag strings (e.g. BIO)
						vllBatchLabels = self._encodeTagging(vllUPBatchLabels)
						vllBatchPreds = self._encodeTagging(vllBatchPreds)
						
						vBatchPrec, vBatchRec, vBatchF1 = self._eval(vllBatchLabels, vllBatchPreds)
						#print "         Batch evaluation (precision, recall, F1): %s, %s, %s         " % (round(vBatchPrec, 2), round(vBatchRec, 2), round(vBatchF1, 2))
						
				
				# validation -----------------------------------------------------
				
				if pflgValidate:
					if epoch == 1 or epoch % self._validEpochInterval == 0:
						# deleting the sequence padding from predictions and gold labels
						vllValidLabels = self._delPadding(vaValidLabels, vaValidSeqLen)
						vllValidPreds = self._delPadding(np.argmax(vaValidPreds, 2), vaValidSeqLen)
						
						# encoding the tag indexes to tag strings (e.g. BIO)
						vllValidLabels = self._encodeTagging(vllValidLabels)
						self.validationOutput = self._encodeTagging(vllValidPreds)
						
						vValidPrec, vValidRec, vValidF1 = self._eval(vllValidLabels, self.validationOutput)
						print "         Validation set evaluation (precision, recall, F1): %s, %s, %s" % (round(vValidPrec, 2), round(vValidRec, 2), round(vValidF1, 2))
			
			#########################################
			#for ot, op in zip(_output[-1, :5], vaBatchPreds[-1, :5]):
			#	print '>', ot, np.argmax(ot)
			#	print '>>', op, np.argmax(op)
			
			#print '>', _aBatchTokLosses.tolist()
			
			print "         Batch evaluation (precision, recall, F1): %s, %s, %s         " % (round(vBatchPrec, 2), round(vBatchRec, 2), round(vBatchF1, 2))
			self.t = self._generateOutput(vllBatchLabels, vllBatchPreds)
			#########################################
			
			if self.verbosity == 1 and not pflgValidate:
				print("Average loss at last epoch: %s" % vBatchLoss)
				
				# deleting the sequence padding from predictions and gold labels
				vllBatchPreds = self._delPadding(np.argmax(vaBatchPreds, 2), vaBatchSeqLens)
				
				# encoding the tag indexes to tag strings (e.g. BIO)
				vllBatchLabels = self._encodeTagging(vllUPBatchLabels)
				self.validationOutput = self._encodeTagging(vllBatchPreds)
				
				vBatchPrec, vBatchRec, vBatchF1 = self._eval(vllBatchLabels, self.validationOutput)
				print "Batch evaluation (precision, recall, F1): %s, %s, %s" % (round(vBatchPrec, 2), round(vBatchRec, 2), round(vBatchF1, 2))
		
		
		print("\nDone.")
		
	
	
	def _resetBatchGenerator(self):
		'''
		Resets the batch generator cursor to start from the begining of the data
		'''
		
		self._batchCursor = 0
		
	
	
	def _generateBatch(self, pBatchSize):
		'''
		Generates and returns a batch of given size from the training data
		
		The returned values are:
		- a 2D numpy arrays of zero padded token IDs
		- a flat numpy array of tag IDs
		- an array containing original sequence lengths in the batch
		- a 2D list of tag IDs (unpadded)
		- a flat numpy array of tag weights (if weights are assigned to tags)
		
		The numpy array of tag IDs will be used in loss computation and 2D list of tag IDs for evaluating the training 
		batch predictions. The numpy array of tag weights will be used in computing the loss. It is None of no weights
		are set for tags.
		'''
		
		# Note that it's possible that there is not enough data after the cursor 
		# position to fill the batch, so it should go back to the begining of the 
		# data set. Therefore, the end of range is calculated by modulo.
		
		
		# collecting lengths of sequences falling in the batch
		
		vTempBatchCursor = self._batchCursor                                    # temporary cursor for length collection       
		vaBatchSeqLens = np.zeros((pBatchSize), dtype=np.int32)
		for i in range(pBatchSize):
			vaBatchSeqLens[i] = self.trainSet.sequences[vTempBatchCursor].length
			vTempBatchCursor = (vTempBatchCursor + 1) % self.trainSet.seqSize
		
		
		# finding the maximum batch length for padding
		vMaxBatchLen = np.amax(vaBatchSeqLens)
		
		
		# collecting the batch data while 1) changing the representation (e,g, to one-hot) and 2) padding
		
		# padded input
		vaBatchInput = np.ndarray(shape=(pBatchSize, vMaxBatchLen, self._tokRepDim), dtype=np.float32)
		# unpadded batch labels for loss calculation in training 
		vllUPBatchLabels = []
		# unpadded loss calculation weights for batch labels (can be merged with vllUPBatchLabels)
		vllUPBatchLabelWeights = []
		
		for i in range(pBatchSize):
			# real data part
			for j in range(self.trainSet.sequences[self._batchCursor].length):
				vaBatchInput[i, j] = self._getTokRep(self.trainSet.sequences[self._batchCursor].getTokenForms()[j])
			
			# padding part
			for j in range(self.trainSet.sequences[self._batchCursor].length, vMaxBatchLen):
				vaBatchInput[i, j] = np.zeros(self._tokRepDim)
			
			# unpadded labels
			vllUPBatchLabels.append([self.tagIDs[tag] for tag in self.trainSet.sequences[self._batchCursor].getTags()])
			
			# loss weights
			if self.tagWeights is not None:
				vllUPBatchLabelWeights.append([self.tagWeights[tag] for tag in self.trainSet.sequences[self._batchCursor].getTags()])
			
			# setting the cursor
			self._batchCursor = (self._batchCursor + 1) % self.trainSet.seqSize
		
		# flat array of labels (tag IDs) for loss calcualtion
		vaBatchLabels = np.asarray([l for s in vllUPBatchLabels for l in s], dtype = np.int32)
		
		# flat array of loss weights for labels
		if self.tagWeights is not None:
			vaBatchLabelWeights = np.asarray([l for s in vllUPBatchLabelWeights for l in s], dtype = np.float32)
		else:
			vaBatchLabelWeights = np.ones(vaBatchLabels.shape)                              # just to make tensorflow shut up (othewise None)
		
		return vaBatchInput, vaBatchLabels, vaBatchSeqLens, vllUPBatchLabels, vaBatchLabelWeights
		
	
	
	def _generateValidSet(self):
		'''
		Generates and returns validation data
		
		The returned values are 2D numpy arrays of token IDs and tag IDs 
		and an array containing original sequence lengths in the dataset as
		sequence input array is zero-padded to have equal length arrays 
		in the batch.
		'''
		
		# collecting lengths of sequences
		vaDatasetSeqLens = np.array([s.length for s in self.validationSet.sequences])
		
		# finding the maximum sequence length for padding
		vMaxSeqLen = np.amax(vaDatasetSeqLens)
		
		# collecting the data while 1) changing the representation (e,g, to one-hot) and 2) padding
		vaInput = np.ndarray(shape = (self.validationSet.seqSize, vMaxSeqLen, self._tokRepDim), dtype = np.float32)
		vaLabels = np.ndarray(shape = (self.validationSet.seqSize, vMaxSeqLen), dtype = np.int32)
		
		for i in range(self.validationSet.seqSize):
			# real data part
			for j in range(self.validationSet.sequences[i].length):
				vaInput[i, j] = self._getTokRep(self.validationSet.sequences[i].getTokenForms()[j])
				vaLabels[i, j] = self.tagIDs[self.validationSet.sequences[i].getTags()[j]]
			
			# padding part
			for j in range(self.validationSet.sequences[i].length, vMaxSeqLen):
				vaInput[i, j] = np.zeros(self._tokRepDim)
				vaLabels[i, j] = 0
		
		return vaInput, vaLabels, vaDatasetSeqLens
		
	
	
	def _delPadding(self, paInput, paSeqLens):
		'''
		Given a 2D numpy array of sequences padded for equal lengths and the actual sequence lengths, removes the
		paddings and returns the results in a 2D list
		'''
		
		# sanity check
		if len(paInput) != len(paSeqLens):
			raise Exception("Different lengths for input and sequence lengths: %s vs. %s" % (len(paInput), len(paSeqLens)))
		
		vllResult = []
		
		for vSeq, vLen in zip(paInput, paSeqLens):
			vllResult.append(list(vSeq[ : vLen]))
		
		return vllResult
		
	
	
	def _encodeTagging(self, pllSeqTagging):
		'''
		Encode the sequence tags passed in a 2D list of sequence token tags by converting tag indexes to tags 
		'''
		
		return [[self.idTags[t] for t in s] for s in pllSeqTagging]
		
	
	
	def _eval(self, paGoldLabels, paPredLabels):
		'''
		Evaluates the predicted labels versus the gold labels 
		'''
		
		# ToDo: this class is not specific to BIO tagging; make the evaluation general 
		vEval = eval.BIOEval(plPredSeqs = paPredLabels, plGoldSeqs = paGoldLabels)
		
		return vEval.spanPrecision(), vEval.spanRecall(), vEval.spanFScore()
		
	
	
	def getValidationOutput(self):
		'''
		Returns validation set tagging in a 2D list of tuples of token forms and tags
		'''
		
		return self._generateOutput(self.validationSet.getSequenceTokenForms(), self.validationOutput)
		
	
	
	def _generateOutput(self, pllTokenForms, pllTags):
		'''
		Generates and returns tagging output which is a 2D list of tuples of token forms and tags each passed in
		2D lists themselvs
		'''
		
		# sanity checks
		if len(pllTokenForms) != len(pllTags):
			raise Exception("Oops! sequence number mismatch found while generating output: %s vs. %s" % (len(pllTokenForms), len(pllTags)))
		
		vllOutput = []
		
		for vlTokens, vlTags in zip(pllTokenForms, pllTags):
			# sanity check
			if len(vlTokens) != len(vlTags):
				raise Exception("Oops! sequence size mismatch found while generating output: %s vs. %s" % (len(vlTokens), len(vlTags)))
			
			vllOutput.append(zip(vlTokens, vlTags))
		
		return vllOutput
		
	
	