#! /usr/bin/env python
# coding: utf-8


"""	
    This module defines classes for sequence-tagged data.
    

    Version 1.0                                                                              (28-Oct-2016 to 22-Nov-2016)
    - SeqTagSet, SeqTagSequence and SeqTagToken are added. 
    
"""

from utils import util


class SeqTagSet:
    '''
    Class for sequence-tagged data set
    '''
        
    def __init__(self):
        '''
        Constructor
        '''
        
        # list of SeqTagSequence objects
        self.sequences = None
        
    
    
    
    @property
    def seqSize(self):
        '''
        Returns the number of sequences in the data set
        '''
        
        return len(self.sequences)
        
    
    
    @property
    def tokSize(self):
        '''
        Returns the number of tokens in the data set
        '''
        
        return len(self.getAllTokenForms())
        
    
    
    @property
    def vocabSize(self):
        '''
        Returns the vocabulary size of the data set, i.e. the number of uniqure tokens
        '''
        
        return len(set(self.getAllTokenForms()))
        
    
    
    def load(self, pFilename, pColumns = "token tag", pSampleSize = -1):
        '''
        Loads data from input file in the specified columnar format
        
        See SeqTagToken for supported columnar formats.
        
        pSampleSize specifies the number of sequences to be loaded from the top of the file. 
        It is useful when the dataset is big and not all of it is needed. With -1, the entire
        data will be loaded
        '''
        
        if pSampleSize == -1:
            self.sequences = [SeqTagSequence().load(s, pColumns = pColumns) for s in open(pFilename).read().strip().split("\n\n")]
        else:
            self.sequences = [SeqTagSequence().load(s, pColumns = pColumns) for s in open(pFilename).read().strip().split("\n\n")[: pSampleSize]]
        
    
    
    def addSequences(self, plSequences):
        '''
        Adds the given list of sequence objects to the dataset 
        '''
        
        if self.sequences is None:
            self.sequences = plSequences
        else:
            self.sequences += plSequences
        
    
    
    def toLower(self):
        '''
        Converts token forms to lowercase
        
        This is useful for textual data.
        '''
        
        for vSeq in self.sequences:
            vSeq.toLower()
        
    
    
    def getAllTokenForms(self):
        '''
        Returns all token froms in the data set in a list
        '''
    
        return [t.form for s in self.sequences for t in s.tokens]
        
    

    def getSequenceTokenForms(self):
        '''
        Returns a 2D list containing tokens of sequences
        
        List shape: (#sequence, #squence tokens)
        '''
        
        return [[t.form for t in s.tokens] for s in self.sequences]
            
    
    
    def getTokenFormFreqs(self):
        '''
        Returns token frequencies in a dictionary
        '''
        
        return util.groupBy(self.getAllTokenForms()).iteritems()
        
    
    
    def getTokenFormFreqList(self, pSortOrder = "desc"):
        '''
        Returns sorted token frequencies in a list of tuples of form (token, freq)  
        '''
        
        return sorted([(k, v) for k, v in self.getTokenFormFreqs()], key = lambda x: x[1], reverse = (pSortOrder.lower() == "desc"))
        
    
    
    def getAllTags(self):
        '''
        Returns all token tags in the data set in a list
        '''
        
        return [t.tag for s in self.sequences for t in s.tokens]
        
    
    
    def getSequenceTags(self):
        '''
        Returns a 2D list containing token tags of sequences

        List shape: (#sequence, #squence tokens)
        '''
    
        return [[t.tag for t in s.tokens] for s in self.sequences]
        
    
    
    def getTagFreqs(self):
        '''
        Returns token tag frequencies in a dictionary
        '''
    
        return util.groupBy(self.getAllTags()).iteritems()
        
    
    
    def getTagFreqList(self, pSortOrder="desc"):
        '''
        Returns sorted token tag frequencies in a list of tuples of form (token, freq)  
        '''
    
        return sorted([(k, v) for k, v in self.getTagFreqs()], key=lambda x: x[1], reverse=(pSortOrder.lower() == "desc"))
        
    
    
    def getTagSet(self):
        '''
        Returns the tag set of the entire data setin a list 
        '''
        
        return [t for t in self.getTagFreqList()]
        
    
    


class SeqTagSequence:
    '''
    Class for sequence-tagged sequence
    '''
    
    def __init__(self):
        '''
        Constructor
        '''
        
        # list of SeqTagToken objects
        self.tokens = None
        
    
    
    def load(self, pInput, pColumns = "token tag"):
        '''
        Loads sequence from input string in the specified columnar format
        
        For convenience, it returns the sequence object after loading.
        '''
        
        self.tokens = [SeqTagToken().load(s, pColumns = pColumns) for s in  pInput.split("\n")]
        
        return self
        
    
    
    @property
    def length(self):
        '''
        Returns the sequence length, i.e. number if its tokens 
        '''
        
        return len(self.tokens)
        
    
    
    def toLower(self):
        '''
        Converts token forms to lowercase
        
        This is useful for textual data.
        '''
        
        for vTok in self.tokens:
            vTok.toLower()
        
    
    
    def getTokenForms(self):
        '''
        Returns token forms of the sequence
        '''
        
        return [t.form for t in self.tokens]
        
    
    
    def getTags(self):
        '''
        Returns token tags of the sequence
        '''
        
        return [t.tag for t in self.tokens]
        
    
    


class SeqTagToken:
    '''
    Class for sequence-tagged token
    '''
    
    def __init__(self, pForm = None, pTag = None):
        '''
        Constructor
        '''
        
        self.form = pForm
        self.tag = pTag
        
    
    
    def load(self, pInput, pColumns = "token tag"):
        '''
        Loads token from input line in the specified columnar format
        
        For convenience, it returns the token object after loading.
        '''
        
        # setting up columns
        vdColumns = {c.lower(): i for i, c in enumerate(pColumns.split())}
        
        vlColumns = pInput.split()
        
        # compulsary columns
        self.form = vlColumns[vdColumns["token"]]
        self.tag = vlColumns[vdColumns["tag"]]
        
        return self
        
    
    
    def toLower(self):
        '''
        Converts token forms to lowercase
        
        This is useful for textual data.
        '''
        
        try:
            self.form = self.form.lower()
        except:
            pass
        
    
    



class BIOSeqTagSet(SeqTagSet):
    '''
    Class for BIO-tag data set
    '''
    
    
    def __init__(self):
        '''
        Constructor
        '''

        SeqTagSet.__init__(self)
        
    
    
    def filteredSubSet(self):
        '''
        Returns a subset of the data in which the sequences with no BI tags are filtered 
        '''
        
        # returning selection as a set
        vNewSet = SeqTagSet()
        
        vNewSet.addSequences([s for s in self.sequences if 'B' in s.getTags() or 'I' in s.getTags()])
        
        return vNewSet
        
    
    
