#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""	
	This module defines functions and classes in relation to SVMLight and
	SVMLight-TK tool.
	
	THIS IS THE FROM-SCRATCH REWRITING OF THE NOW ABSOLUTE SVMLIGHT.PY.
	
	Version 0.4                                             (14-Jun-2017)
	- Changed in mc2bc.py are reflected.
	
	Version 0.3                                             (06-Nov-2015)
	- In SVMLightMC.loadDataset(), the exception on having fewer than two class for multi-class problem has been changed
	  to a warning because a test set may legitimately have two or fewer class. The exception can be moved to where the
	  train set is loaded.
	
	Version 0.2												(21-Oct-2015)
	- In cross-validation and also parameter search (inside the evaluator function) of SVMLight and SVMLightMC, the model
	  files were deleted after being used in test.
	
	Version 0.1								(08-Jul-2015 to 23-Jul-2015)
	- SVMLight, SVMLightDataset, SVMLightDataInstance, SVMLightTK, 
	  SVMLightTKDataset, SVMLightTKDataInstance are added.
"""

from ConfigParser import SafeConfigParser
from decimal import *
from collections import namedtuple
import sys, subprocess, tempfile, os, random, re, warnings
import eval, gridsearch, mc2bc
from utils import util

#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# SVMLight
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class SVMLight:
	'''
	Class for SVMLight
	'''
	
	
	def __init__(self, pLearnerPath = '', pClassifierPath = '', pID = '', pLogFilePrefix = ''):
		'''
		Constructor
		'''
		
		## Name can help to identify the output in case of running for
		## different data sets at the same time (e.g. if the jobs crash 
		## log files will be identifiable).
		self.id = pID
		
		# loading config file if either of paths are not provided 
		if pLearnerPath == '' or pClassifierPath == '':
			vAppPath = os.path.dirname(__file__)
			vConfigParser = SafeConfigParser()
			vConfigParser.read(os.path.join(vAppPath, "svmlight.ini"))
			self.learner = vConfigParser.get("SVMLIGHT","learner")
			self.classifier = vConfigParser.get("SVMLIGHT","classifier")
		
		# overriding the config file if they are passed as arguments
		
		if pLearnerPath != '':
			self.learner = pLearnerPath
		
		if pClassifierPath != '':
			self.classifier = pClassifierPath
		
		# training, test sets of type SVMLightDataset
		
		self.trainSet = None
		self.testSet = None
		
		# setting the log devices
		
		if pLogFilePrefix == '':
			self.fOutLog = sys.stdout
			self.fErrLog = sys.stderr
		else:
			self.fOutLog = open(pLogFilePrefix + ".out", 'w')
			self.fErrLog = open(pLogFilePrefix + ".err", 'w')
		
	
	
	def _createDataset(self):
		'''
		Creates and returns a dataset object
		'''
		
		return SVMLightDataset()
		
	
	
	def loadDataset(self, pDatasetFilename):
		'''
		Loads and returns the dataset from file in SVMLight format
		'''
		
		vDataset = self._createDataset()
		vDataset.loadFromFile(pDatasetFilename)
		
		return vDataset
		
	
	
	def loadTrainSet(self, pDatasetFilename):
		'''
		Loads the training data from SVMLight file
		'''
		
		self.trainSet = self.loadDataset(pDatasetFilename)
		
	
	
	def loadTestSet(self, pDatasetFilename):
		'''
		Loads the test data from SVMLight file
		'''
		
		self.testSet = self.loadDataset(pDatasetFilename)
		
	
	
	def train(self, pDataset = None, pModelOutFilename = '', pdLearnerOptions = {}, pfOutLog = None, pfErrLog = None):
		'''
		Trains a SVM model on a dataset if given or the training set otherwise 
		'''
		
		# setting logging devices
		
		if pfOutLog != None:
			vfOutLog = pfOutLog
		else:
			vfOutLog = self.fOutLog
		
		if pfErrLog != None:
			vfErrLog = pfErrLog
		else:
			vfErrLog = self.fErrLog
		
		vfOutLog.write("Training ...\n\n")
		
		# setting dataset
		
		if pDataset == None:
			vTrainSet = self.trainSet
		else:
			vTrainSet = pDataset
		
		if vTrainSet == None or vTrainSet.size == 0:
			raise Exception("Training dataset is empty!")
		
		if len(vTrainSet.getTargetSet()) > 2:
			raise Exception("There are more than 2 classes in the training dataset. Use SVMLightMC instead!")
		
		# generating option list
		
		vlOptions = []
		for vOption, vValue in pdLearnerOptions.iteritems():
			vlOptions.append(vOption)
			vlOptions.append(vValue)
		
		# generating input file
		
		vTempTrainFilename = tempfile.mktemp(suffix="train")
		vTrainSet.writeToFile(pOutFilename = vTempTrainFilename)
		
		# setting model file name
		
		if pModelOutFilename != '':
			vModelFilename = pModelOutFilename
		else:
			vModelFilename = tempfile.mktemp(suffix="model")
		
		# doing the training
		
		vfOutLog.write("-- SVMLight learner log start ----------\n\n")
		vfOutLog.flush()
		subprocess.call([self.learner] + vlOptions + [vTempTrainFilename, vModelFilename], stdout=vfOutLog, stderr=vfErrLog)
		vfOutLog.write("\n-- SVMLight learner log end ------------\n\n")
		
		os.remove(vTempTrainFilename)
		
		return vModelFilename
		
	
	
	def predict(self, pModelFilename, pDataset = None, pdPredictorOptions = {}, pfOutLog = None, pfErrLog = None):
		'''
		Predicts the target values of a dataset if given, with a trained
		SVM model, or the test set otherwise, and returns predictions
		'''
		
		# setting logging devices
		
		if pfOutLog != None:
			vfOutLog = pfOutLog
		else:
			vfOutLog = self.fOutLog
		
		if pfErrLog != None:
			vfErrLog = pfErrLog
		else:
			vfErrLog = self.fErrLog
		
		vfOutLog.write("Predicting ...\n\n")
		
		# setting datasets
		
		if pModelFilename == '':
			raise Exception("No model found!")
		
		if pDataset == None:
			vTestSet = self.testSet
		else:
			vTestSet = pDataset
		
		if vTestSet == None or vTestSet.size == 0:
			raise Exception("Test dataset is empty!")
		
		if len(vTestSet.getTargetSet()) > 2:
			raise Exception("There are more than 2 classes in the input dataset. Use SVMLightMC instead!")
		
		# generating option list
		
		vlOptions = []
		for vOption, vValue in pdPredictorOptions.iteritems():
			vlOptions.append(vOption)
			vlOptions.append(vValue)
		
		# generating input file
		
		vTempEvalFilename = tempfile.mktemp(suffix="eval")
		vTestSet.writeToFile(pOutFilename = vTempEvalFilename)
		
		# setting prediction output file name
		
		vOutFilename = tempfile.mktemp(suffix="out")
		
		# doing the prediction
		
		vfOutLog.write("-- SVMLight classifier log start ----------\n\n")
		vfOutLog.flush()
		subprocess.call([self.classifier] + vlOptions + [vTempEvalFilename, pModelFilename, vOutFilename], stdout=vfOutLog, stderr=vfErrLog)
		vfOutLog.write("\n-- SVMLight classifier log end ------------\n\n")
		
		# loading output predictions
		
		vlPredictions = []
		for vLine in open(vOutFilename):
			vlPredictions.append(float(vLine.strip()))
		
		# cleanup
		
		os.remove(vOutFilename)
		
		return vlPredictions
		
	
	
	def test(self, pModelFilename, pModelType, pDataset = None, pdPredictorOptions = {}, pdScoringOptions = {"acc": {}}, pfOutLog = None, pfErrLog = None):
		'''
		Tests a trained SVM model on a dataset if given or the test set 
		otherwise and returns predictions and scores
		
		pModelType specifies whether the model is a classification or
		regression model. This will be used for scoring.
		
		pdScoringOptions specifies the metrics to be used and their settings
		(see score() for details).
		'''
		
		# -- predicting ----
		
		vlPredictions = self.predict(pModelFilename = pModelFilename, pDataset = pDataset, pdPredictorOptions = pdPredictorOptions, pfOutLog = pfOutLog, pfErrLog = pfErrLog)
		
		# -- scoring -------
		
		# setting logging devices
		
		if pfOutLog != None:
			vfOutLog = pfOutLog
		else:
			vfOutLog = self.fOutLog
		
		if pfErrLog != None:
			vfErrLog = pfErrLog
		else:
			vfErrLog = self.fErrLog
		
		vfOutLog.write("Scoring ...\n\n")
		
		# extracting gold targets
		
		if pDataset == None:
			vlGoldTargets = self.testSet.getTargetValues()
		else:
			vlGoldTargets = pDataset.getTargetValues()
		
		if pModelType.lower()[0] == 'c':
			vlPredictions = [1.0 if m > 0 else -1.0 for m in vlPredictions]
		
		# doing the scoring
		
		vResults, vdConfMatrix = self._score(plPredictions = vlPredictions, plGoldTargets = vlGoldTargets, pModelType = pModelType, pdOptions = pdScoringOptions)
		
		return vlPredictions, vResults, vdConfMatrix
		
	
	
	def _score(self, plPredictions, plGoldTargets, pModelType, pdOptions):
		'''
		Scores the predictions against the gold target values and returns
		the scores
		
		pModelType specifies whether the model is a classification or
		regression model.
		
		Scoring options are provided via a dictionary with keys being the
		metrics and the values being the settings. The supported metrics 
		are grouped into classification and regression metrics. See 
		_scoreClassification() and _scoreRegression() for more details.
		
		'''
		
		if len(plGoldTargets) != len(plPredictions):
			raise Exception("The number of predictions does not match the number of gold target values: %s vs. %s" % (len(plPredictions), len(plGoldTargets)))
		
		if pModelType.lower()[0] == 'c':
			return self._scoreClassification(plPredictions = plPredictions, plGoldTargets = plGoldTargets, pdOptions = pdOptions)
		elif pModelType.lower()[0] == 'r':
			return self._scoreRegression(plPredictions = plPredictions, plGoldTargets = plGoldTargets, pdOptions = pdOptions)
		
	
	
	def _scoreClassification(self, plPredictions, plGoldTargets, pdOptions):
		'''
		Scores the classifications against the gold target values and 
		returns the scores
		
		Each metric may be accompanied with a setting which is again provided
		via a dictionary and the value of the element whose key is the metric
		name itself. Supported metrics are (case insensitive):
		
		- (Acc)uracy : {}
		- (Prec)ision: {}
		- (Rec)all   : {}
		- (F)-Score  : {beta: recall weight (default: 1)}
		
		Additional settings for precision, recall and F-score:
		 
		- {type: class/macro/micro (default: class)}
		- {labels: dictionary of labels for numerical class values / only for type = class}
		'''
		
		vPredEval = eval.PredEval(plPrediction = plPredictions, plReference = plGoldTargets)
		
		vdScores = {}
		
		for vKey in pdOptions:
			# -- accuracy -----
			if vKey.lower().startswith("acc"):
				vdScores["Accuracy"] = vPredEval.acc(pType = "exact")
				
			# -- precision ----
			elif vKey.lower().startswith("prec"):
				vPrecision = vPredEval.precision()
				
				# class precisions as default
				
				if "labels" in [k.lower() for k in pdOptions[vKey]]:
					vdScores["Precision"] = {pdOptions[vKey]["labels"][c]: s for c, s in vPrecision[0].iteritems()}
				else:
					vdScores["Precision"] = vPrecision[0]
				
				# looking if other type than per-class is requested
				
				if "type" in [k.lower() for k in pdOptions[vKey]]:
					if pdOptions[vKey]["type"].lower() == "macro":
						vdScores["Precision"] = vPrecision[1]["macro"]
					elif pdOptions[vKey]["type"].lower() == "micro":
						vdScores["Precision"] = vPrecision[1]["micro"]
				
			# -- recall--------
			elif vKey.lower().startswith("rec"):
				vRecall = vPredEval.recall()
				
				# class recalls as default
				
				if "labels" in [k.lower() for k in pdOptions[vKey]]:
					vdScores["Recall"] = {pdOptions[vKey]["labels"][c]: s for c, s in vRecall[0].iteritems()}
				else:
					vdScores["Recall"] = vRecall[0]
				
				# looking if other type than per-class is requested
				
				if "type" in [k.lower() for k in pdOptions[vKey]]:
					if pdOptions[vKey]["type"].lower() == "macro" :
						vdScores["Recall"] = vRecall[1]["macro"]
					elif pdOptions[vKey]["type"].lower() == "micro" :
						vdScores["Recall"] = vRecall[1]["micro"]
				
			# -- f-score ------
			elif vKey.lower().startswith('f'):
				vBeta = 1
				if "beta" in [k.lower() for k in pdOptions[vKey]]:
					vBeta = pdOptions[vKey]["beta"]
				
				vFScore = vPredEval.fscore(pBeta = vBeta)
				
				# class F-scores as default
				
				if "labels" in [k.lower() for k in pdOptions[vKey]]:
					vdScores["F%s" % vBeta] = {pdOptions[vKey]["labels"][c]: s for c, s in vFScore[0].iteritems()}
				else:
					vdScores["F%s" % vBeta] = vFScore[0]
				
				# looking if other type than per-class is requested
				
				if "type" in [k.lower() for k in pdOptions[vKey]]:
					if pdOptions[vKey]["type"].lower() == "macro":
						vdScores["F%s" % vBeta] = vFScore[1]["macro"]
					elif pdOptions[vKey]["type"].lower() == "micro":
						vdScores["F%s" % vBeta] = vFScore[1]["micro"]
		
		vdConfMatrix = vPredEval.confusionMatrix()
		
		return vdScores, vdConfMatrix
		
	
	
	def _scoreRegression(self, plPredictions, plGoldTargets, pdOptions):
		'''
		Scores the regression results against the gold target values and 
		returns the scores
		
		Supported metrics are (case insensitive):
		
		- (Acc)uracy      : {type: exact/round/absolute-err-margin/ratio-err-margin, RDigits: 0 (d), ErrMargin: 0 (d), ErrMarginRatio: 0 (d)} (see eval.PredEval.acc())
		- MAE             : {} 
		- RMSE            : {}
		- ((Pearson) (r)) : {}
		'''
		
		vPredEval = eval.PredEval(plPrediction = plPredictions, plReference = plGoldTargets)
		
		vdScores = {}
		
		for vKey in pdOptions:
			if vKey.lower().startswith('acc'):
				# accuracy
				
				vType = "exact"
				vRDigits = 0
				vErrMargin = 0
				vErrMarginRatio = 0
				
				for vAKey in pdOptions[vKey]:
					if vAKey.lower() == "type":
						vType = pdOptions[vKey][vAKey]
					elif vAKey.lower() == "rdigits":
						vRDigits = pdOptions[vKey][vAKey]
					elif vAKey.lower() == "errmargin":
						vErrMargin = pdOptions[vKey][vAKey]
					elif vAKey.lower() == "errmarginratio":
						vErrMarginRatio = pdOptions[vKey][vAKey]
				
				vdScores["Accuracy"] = vPredEval.acc(pType = vType, pRDigits = vRDigits, pErrMargin = vErrMargin, pErrMarginRatio = vErrMarginRatio)
				
			elif vKey.lower() == "mae":
				# MAE
				
				vdScores["MAE"] = vPredEval.mae()
				
			elif vKey.lower() == "rmse":
				# RMSE
				
				vdScores["RMSE"] = vPredEval.rmse()
				
			elif vKey.lower().startswith('pearson') or vKey.lower() == 'r':
				# Pearson r
				
				vdScores["Pearson r"] = vPredEval.r()
		
		## NOTE: this will be too huge a matric for regression. It should 
		## be reimplemented for regression to consider ranges instead of
		## exact values, i.e. how many times the values in range x are 
		## confused with those in range y.
		vdConfMatrix = vPredEval.confusionMatrix()
		
		return vdScores, vdConfMatrix
		
	
	
	def flattenClassScores(self, pdScores):
		'''
		Flattens the per-class scores in the given score dictionary by
		merging the metric name and the class name and returns the new
		dictionary
		
		Per-class scores such as per-class precision are in the form of 
		a dictionary with entries of per-class scores. These dictionaries
		in the given score dictionary are converted to multiple scores, 
		one per class for each metric.
		'''
		
		vdFlatScores = dict(pdScores)
		
		for vMetric in pdScores:
			if isinstance(vdFlatScores[vMetric], dict):
				for vClass in vdFlatScores[vMetric]:
					vdFlatScores["%s (%s)" % (vMetric, vClass)] = vdFlatScores[vMetric][vClass]
				
				del vdFlatScores[vMetric]
		
		return vdFlatScores
		
	
	
	def crossValidate(self, pFoldNum, pDataset = None, pflgShuffle = True, pdLearnerOptions = {}, pdPredictorOptions = {}, pdScoringOptions = {}, pfOutLog = None, pfErrLog = None):
		'''
		NOTE: better to change the pflgShuffle to another option which can
		also load from a saved split so that the train/test datasets are
		the same for all settings (sync with other crossValidate methods
		in the module).
		
		Runs a pFoldNum fold cross validation over dataset if given, or the
		training set otherwise, and returns the average scores
		
		Training (learner) and classification (predictor) options should 
		be provided via dictionaries with keys being parameters (including 
		dashes) and values being the parameter values, e.g. {"-t":11, "-c":1}.
		
		Scoring options are also provided via a dictionary. For details,
		follow the documentation of test().
		'''
		
		# setting logging devices
		
		if pfOutLog != None:
			vfOutLog = pfOutLog
		else:
			vfOutLog = self.fOutLog
		
		if pfErrLog != None:
			vfErrLog = pfErrLog
		else:
			vfErrLog = self.fErrLog
		
		# determining and shuffling dataset
		
		if pDataset == None:
			vDataset = self.trainSet.shuffled()
		else:
			vDataset = pDataset.shuffled()
		
		if vDataset == None or vDataset.size == 0:
			raise Exception("Dataset is empty!")
		
		vChunckSize = vDataset.size / pFoldNum
		
		## Note: if pFoldNum is not a factor of the size of the cross-validion
		## data, the remainder will never be used as test set.
		
		# list of list of predictions per fold: fold numbers are keys
		vllPredictions = []
		# list of dictionaries of scores per fold: fold numbers are keys
		vldScores = []
		
		for i in range(pFoldNum):
			vfOutLog.write("Cross-validation %d/%d ...\n\n" % (i + 1, pFoldNum))
			
			# splitting train/test
			
			vEvalSetRange = ((i * vChunckSize) + 1, (i + 1) * vChunckSize)
			
			vEvalSet = vDataset.chunk(vEvalSetRange[0], vEvalSetRange[1])
			vTrainSet = vDataset.chunk(1, vEvalSetRange[0] - 1) + vDataset.chunk(vEvalSetRange[1] + 1)
			
			# sanity check
			if vEvalSet.size != vChunckSize or vTrainSet.size != (vDataset.size - vEvalSet.size):
				raise Exception("There is a problem with splitting the CV dataset: dataset size: %s, chunk size: %s, training set size: %s, evaluation set size: %s" % (vDataset.size, vChunckSize, vTrainSet.size, vEvalSet.size))
			
			# training
			vModelFilename = self.train(pDataset = vTrainSet, pdLearnerOptions = pdLearnerOptions, pfOutLog = pfOutLog, pfErrLog = pfErrLog)
			
			# testing
			vlPredictions, vdScores = self.test(pModelFilename = vModelFilename, pModelType = pdLearnerOptions["-z"], pDataset = vEvalSet, 
			                                    pdPredictorOptions = pdPredictorOptions, pdScoringOptions = pdScoringOptions, 
			                                    pfOutLog = pfOutLog, pfErrLog = pfErrLog)
			
			vllPredictions.append(vlPredictions)
			vldScores.append(vdScores)
			
			# cleaning up model file (added as a patch to disk space problem on 21/10/2015)
			os.remove(vModelFilename)
		
		# generating and returning the scores
		
		return self._avgCVScores(pldScores = vldScores)
		
	
	
	def _avgCVScores(self, pldScores):
		'''
		Averages and returns the cross-validation scores
		
		The scores are given in a list, the elements of which are dictionaries
		of scores per CV fold
		'''
		
		vdAvgScores = {}
		
		vFoldNum = len(pldScores)
		
		# initializing the average scores dictionary
		
		for vdFoldScores in pldScores:
			for vMetric, vScore in vdFoldScores.iteritems():
				if isinstance(vScore, dict):          # per-class scores
					vdAvgScores[vMetric] = {cs: 0 for cs in vScore}
				else:                                 # scalar scores
					vdAvgScores[vMetric] = 0
		
		# calculating the average scores
		
		for vdFoldScores in pldScores:
			for vMetric, vScore in vdFoldScores.iteritems():
				if isinstance(vScore, dict):          # per-class scores
					vdAvgScores[vMetric] = {c: vdAvgScores[vMetric][c] + (vScore[c] / vFoldNum) for c in vScore}
				else:                                 # scalar scores
					vdAvgScores[vMetric] += (vScore / vFoldNum)
		
		return vdAvgScores
		
	
	
	def gridSearch(self, pdParamSpec, pdLearnerOptions = {}, pdPredictorOptions = {}, pdScoringOptions = {}, pCVFoldNum = 0, pProcNum = 1, pflgDelTempDir = False, pSearchLogFilename = ''):
		'''
		Performs a grid search for the optimum SVM parameters using the 
		specified range of parameter values and returns the best as well
		as the worst parameter sets for each evaluation metric
		
		It also returns the whole processed grid.
		
		For performing a cross-validation over the training set for the
		evaluation of parameters at each step, pCVFoldNum must be an integer
		bigger than 0 (and ideally 1), otherwise a normal training/testing
		process will be done on the training and test datasets.
		'''
		
		# setting up train/test options (excluding the search parameters)
		
		vParamEvalArgs = {"CVFoldNum": pCVFoldNum}
		vParamEvalArgs["learnerOptions"] = pdLearnerOptions
		vParamEvalArgs["predictorOptions"] = pdPredictorOptions
		vParamEvalArgs["scoringOptions"] = pdScoringOptions
		
		# doing the search
		
		vGridSearch = gridsearch.GridSearch(pdParamSpec = pdParamSpec, pfParamEvaluator = self.gsEvaluateParams, pParamEvalArgs = vParamEvalArgs, pID = self.id)
		vdBestWorst, vlGrid = vGridSearch.search(pProcNum = pProcNum, pflgDelTempDir = pflgDelTempDir, pSearchLogFilename = pSearchLogFilename)
		
		return vdBestWorst, vlGrid
		
	
	
	def gsEvaluateParams(self, pdParamValueSet, pfOutLog, pfErrLog, pdArgs):
		'''
		Evaluates training parameters based on the specified options and
		returns the scores
		
		This method is passed as evaluator function to GridSearch class.
		For details, see gridsearch.GridSearch.setParamEvaluator().
		'''
		
		# extracting and setting arguments
		
		vCVFoldNum = 0
		vdLearnerOptions = {}
		vdPredictorOptions = {}
		vdScoringOptions = {}
		
		for vArg in pdArgs:
			if vArg.lower() == "cvfoldnum":
				vCVFoldNum = pdArgs[vArg]
			elif vArg.lower() == "learneroptions":
				vdLearnerOptions = pdArgs[vArg]
			elif vArg.lower() == "predictoroptions":
				vdPredictorOptions = pdArgs[vArg]
			elif vArg.lower() == "scoringoptions":
				vdScoringOptions = pdArgs[vArg]
		
		# setting the training hyper-parameters
		
		for vParam, vValue in pdParamValueSet.iteritems():
			vdLearnerOptions["-" + vParam] = str(vValue)
		
		# doing the evaluation
		
		if vCVFoldNum == 0:
			vModelFilename = self.train(pdLearnerOptions = vdLearnerOptions, pfOutLog = pfOutLog, pfErrLog = pfErrLog)
			vlPredictions, vdScores, dummy = self.test(pModelFilename = vModelFilename, pModelType = vdLearnerOptions["-z"], pdPredictorOptions = vdPredictorOptions, pdScoringOptions = vdScoringOptions, pfOutLog = pfOutLog, pfErrLog = pfErrLog)
		else:
			vdScores = self.crossValidate(pFoldNum = vCVFoldNum, pdLearnerOptions = vdLearnerOptions, pdPredictorOptions = vdPredictorOptions, pdScoringOptions = vdScoringOptions, pfOutLog = pfOutLog, pfErrLog = pfErrLog)
		
		# cleaning up model file (added as a patch to disk space problem on 21/10/2015)
		os.remove(vModelFilename)
		
		return self.flattenClassScores(vdScores), vlPredictions
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# SVMLightMC
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class SVMLightMC(SVMLight):
	'''
	Class for multiclassed SVMLight
	
	This class is designed for treating the problem as multiclass. Although
	it includes multi-class to binary methods such as OVA and OVO, training,
	testing, tuning and other processes are multiclass from usage point 
	of view. If the binary classifiers need to be treated individually,
	SVMLight class can be used for each binary classifier separatelly, 
	where the multiclass dataset has been converted to binary datasets
	using mc2bc module or SVMLightDataset methodsbeforehand. For example,
	training with this class outputs separate models for each binary 
	classifier, but the options used for training, such as hyper-parameters
	are the same for all of them and cannot be specified individually. As
	another example, the grid search in this class evaluates the parameters
	based on the merged predictions of the binary classifiers. If binary
	classifiers have to be tuned individually, one SVMLight per classifier
	should be used. For testing, if the predictions of the individual classes
	are needed, the same approach should be taken. However, if the multiclass
	results are needed, the test method can be used as it accepts mutiple
	models and merges the prediction and scores at the end.
	'''
	
	
	def __init__(self, pLearnerPath = '', pClassifierPath = '', pID = '', pLogFilePrefix = ''):
		'''
		Constructor
		'''
		
		SVMLight.__init__(self, pLearnerPath = pLearnerPath, pClassifierPath = pClassifierPath, pID = pID, pLogFilePrefix = pLogFilePrefix)
		
	
	
	def loadDataset(self, pDatasetFilename):
		'''
		Loads and returns the dataset from file in SVMLight format
		'''
		
		vDataset = SVMLight.loadDataset(self, pDatasetFilename = pDatasetFilename)
		
		if len(vDataset.getTargetSet()) <= 2:
			warnings.warn("There are fewer than 3 classes in the dataset. Use SVMLight instead!", stacklevel = 2)
		
		return vDataset
		
	
	
	def train(self, pMCMethod, pDataset = None, pdModelOutFilenames = {}, pdLearnerOptions = {}, pfOutLog = None, pfErrLog = None):
		'''
		Trains binary SVM models on a multiclass dataset if given or the 
		training set otherwise using the specified method
		
		The given dataset or the loaded training set must be multiclass.
		
		pMCMethod specifies the method to be used for multiclass classification.
		The following methods are supported (case insensitive):
		- OVA (or OVR): one-versus-all (one-versus-rest)
		- OVO: one-versus-one
		
		NOTE: all models are trained using the same learner options. For
		training OVA models with different options, first generate OVA
		datasets using SVMLightDataset.generateOVADatasets() and then 
		train models individually using train().
		
		Model file names can be given to store the models. They are given
		in a dictionary, the keys of which are class names for OVA and 
		class name pair tuples for OVO, corresponding to the models.
		'''
		
		# setting logging devices
		
		if pfOutLog != None:
			vfOutLog = pfOutLog
		else:
			vfOutLog = self.fOutLog
		
		if pfErrLog != None:
			vfErrLog = pfErrLog
		else:
			vfErrLog = self.fErrLog
		
		vfOutLog.write("Training ...\n\n")
		
		# generating option list
		
		vlOptions = []
		for vOption, vValue in pdLearnerOptions.iteritems():
			vlOptions.append(vOption)
			vlOptions.append(vValue)
		
		# setting datasets
		
		if pDataset == None:
			vTrainSet = self.trainSet
		else:
			vTrainSet = pDataset
		
		if vTrainSet == None or vTrainSet.size == 0:
			raise Exception("Training dataset is empty!")
		
		if pMCMethod.lower() == "ova":
			vdBinDatasets = vTrainSet.generateOVADatasets()
		elif pMCMethod.lower() == "ovo":
			vdBinDatasets = vTrainSet.generateOVODatasets(plClassPairs = [cp for cp in pdModelOutFilenames])
		else:
			raise Exception("%s is not a supported multiclass classification method!" % pMCMethod)
		
		# generating input files
		
		## NOTE: vKey in the rest will be class name for OVA and class
		## names pair tuple for OVO
		
		vdTempTrainFilenames = {}
		
		for vKey, vBinDataset in vdBinDatasets.iteritems():
			vTempTrainFilename = tempfile.mktemp(suffix="train%s" % str(vKey))
			vdTempTrainFilenames[vKey] = vTempTrainFilename
			vBinDataset.writeToFile(pOutFilename = vTempTrainFilename)
		
		# setting model file name
		
		vdModelFilenames = {}
		
		for vKey in vdBinDatasets:
			if pdModelOutFilenames != {}:
				vModelFilename = pdModelOutFilenames[vKey]
			else:
				vModelFilename = tempfile.mktemp(suffix="model.%s" % str(vKey))
			
			vdModelFilenames[vKey] = vModelFilename
		
		# doing the training
		
		for vKey, vTempTrainFilename in vdTempTrainFilenames.iteritems():
			vfOutLog.write("-- SVMLight learner log start (model: %s) ----------\n\n" % str(vKey))
			vfOutLog.flush()
			subprocess.call([self.learner] + vlOptions + [vTempTrainFilename, vdModelFilenames[vKey]], stdout=vfOutLog, stderr=vfErrLog)
			vfOutLog.write("\n-- SVMLight learner log end ------------------------\n\n")
			
			os.remove(vTempTrainFilename)
		
		return vdModelFilenames
		
	
	
	def predict(self, pMCMethod, pdModelFilenames, pDataset = None, pdPredictorOptions = {}, pfOutLog = None, pfErrLog = None):
		'''
		Classifies the instances of a dataset if given, with trained binary
		SVM models, or of the test set otherwise, and returns predictions
		
		The given dataset or the loaded training set must be multiclass.
		
		pMCMethod specifies the method to be used for multiclass classification.
		The following methods are supported (case insensitive):
		- OVA (or OVR): one-versus-all (one-versus-rest)
		- OVO: one-versus-one
		
		NOTE: all classifications are done using the same predictor options.
		
		Model file names are given in a dictionary, the keys of which are
		class names for OVA and class name pair tuples for OVO, corresponding 
		to the models.
		'''
		
		# setting logging devices
		
		if pfOutLog != None:
			vfOutLog = pfOutLog
		else:
			vfOutLog = self.fOutLog
		
		if pfErrLog != None:
			vfErrLog = pfErrLog
		else:
			vfErrLog = self.fErrLog
		
		vfOutLog.write("Predicting ...\n\n")
		
		# generating option list
		
		vlOptions = []
		for vOption, vValue in pdPredictorOptions.iteritems():
			vlOptions.append(vOption)
			vlOptions.append(vValue)
		
		# setting datasets
		
		if pdModelFilenames == {}:
			raise Exception("No model found!")
		
		if pDataset == None:
			vTestSet = self.testSet
		else:
			vTestSet = pDataset
		
		if vTestSet == None or vTestSet.size == 0:
			raise Exception("Test dataset is empty!")
		
		# generating input file
		
		## NOTE: if evaluation of each binary OVO classifier is aimed, the
		## test set should be binarized using generateOVODataset().
		
		vTempEvalFilename = tempfile.mktemp(suffix="eval")
		vTestSet.writeToFile(pOutFilename = vTempEvalFilename)
		
		# setting prediction output file names
		
		vdOutFilenames = {}
		
		## NOTE: vKey in the rest will be class name for OVA and class
		## names pair tuple for OVO.
		
		for vKey in pdModelFilenames.iterkeys():
			vOutFilename = tempfile.mktemp(suffix="out%s" % str(vKey))
			vdOutFilenames[vKey] = vOutFilename
		
		# doing the prediction
		
		for vKey in pdModelFilenames.iterkeys():
			vfOutLog.write("-- SVMLight classifier log start (class: %s) ----------\n\n" % str(vKey))
			vfOutLog.flush()
			subprocess.call([self.classifier] + vlOptions + [vTempEvalFilename, pdModelFilenames[vKey], vdOutFilenames[vKey]], stdout=vfOutLog, stderr=vfErrLog)
			vfOutLog.write("\n-- SVMLight classifier log end ------------\n\n")
		
		# merging binary predictions into multiclass
		
		vlPredictions = self.mergePredictions(pMCMethod = pMCMethod, pdPredFilenames = vdOutFilenames)
		
		# clean up
		
		os.remove(vTempEvalFilename)
		
		for vOutFilename in vdOutFilenames.itervalues():
			os.remove(vOutFilename)
		
		return vlPredictions
		
	
	
	def mergePredictions(self, pMCMethod, pdPredFilenames):
		'''
		Merges the predictions of binary SVM classifiers into multiclass
		
		Binary predictions are given in files whose names are passed via
		a dictionary whose entries are the class labels and the corresponding
		prediction files.
		
		pMCMethod specifies the method used for multiclass classification.
		The following methods are supported (case insensitive):
		- OVA (or OVR): one-versus-all (one-versus-rest)
		- OVO: one-versus-one
		
		'''
		
		if pMCMethod.lower() == "ova":
			vOVA = mc2bc.OVA()
			vldMergedPreds, vldTies = vOVA.mergeSVMOVAPreds(pdPredFilenames)
		elif pMCMethod.lower() == "ovo":
			vOVO = mc2bc.OVO()
			vldMergedPreds, vldTies = vOVO.mergeSVMOVOPreds(pdPredFilenames)
		else:
			raise Exception("%s is not a supported multiclass classification method!" % pMCMethod)
		
		return [i["class"] for i in vldMergedPreds]
		
	
	
	def test(self, pMCMethod, pdModelFilenames, pDataset = None, pdPredictorOptions = {}, pdScoringOptions = {"acc": {}}, pfOutLog = None, pfErrLog = None):
		'''
		Tests trained binary SVM models on a dataset if given or on the test set 
		otherwise and returns predictions and scores
		
		pMCMethod specifies the method to be used for multiclass classification.
		The following methods are supported (case insensitive):
		- OVA (or OVR): one-versus-all (one-versus-rest)
		- OVO: one-versus-one
		
		pdScoringOptions specifies the metrics to be used and their settings
		(see score() for details).
		'''
		
		# -- predicting ----
		
		vlPredictions = self.predict(pMCMethod = pMCMethod, pdModelFilenames = pdModelFilenames, pDataset = pDataset, pdPredictorOptions = pdPredictorOptions, pfOutLog = pfOutLog, pfErrLog = pfErrLog)
		
		# -- scoring -------
		
		# setting logging devices
		
		if pfOutLog != None:
			vfOutLog = pfOutLog
		else:
			vfOutLog = self.fOutLog
		
		if pfErrLog != None:
			vfErrLog = pfErrLog
		else:
			vfErrLog = self.fErrLog
		
		vfOutLog.write("Scoring ...\n\n")
		
		# extracting gold targets
		
		if pDataset == None:
			vlGoldTargets = self.testSet.getTargetValues()
		else:
			vlGoldTargets = pDataset.getTargetValues()
		
		# doing the scoring
		
		vdScores, vdConfMatrix = self._score(plPredictions = [float(p) for p in vlPredictions], plGoldTargets = vlGoldTargets, pModelType = 'c', pdOptions = pdScoringOptions)
		
		return vlPredictions, vdScores, vdConfMatrix
		
	
	
	def crossValidate(self, pMCMethod, pFoldNum, pDataset = None, pflgShuffle = True, pdLearnerOptions = {}, pdPredictorOptions = {}, pdScoringOptions = {}, pfOutLog = None, pfErrLog = None):
		'''
		NOTE: better to change the pflgShuffle to another option which can
		also load from a saved split so that the train/test datasets are
		the same for all settings (sync with other crossValidate methods
		in the module).
		
		Runs a pFoldNum fold cross validation over a dataset if given,
		or over the training set otherwise, and returns the average scores.
		
		pMCMethod specifies the method to be used for multiclass classification.
		The following methods are supported (case insensitive):
		- OVA (or OVR): one-versus-all (one-versus-rest)
		- OVO: one-versus-one
		
		Training (learner) and classification (predictor) options should 
		be provided via dictionaries with keys being parameters (including 
		dashes) and values being the parameter values, e.g. {"-t":11, "-c":1}.
		
		Scoring options are also provided via a dictionary. For details,
		follow the documentation of test().
		'''
		
		# setting logging devices
		
		if pfOutLog != None:
			vfOutLog = pfOutLog
		else:
			vfOutLog = self.fOutLog
		
		if pfErrLog != None:
			vfErrLog = pfErrLog
		else:
			vfErrLog = self.fErrLog
		
		# determining and shuffling dataset
		
		if pDataset == None:
			vDataset = self.trainSet.shuffled()
		else:
			vDataset = pDataset.shuffled()
		
		if vDataset == None or vDataset.size == 0:
			raise Exception("Dataset is empty!")
		
		vChunckSize = vDataset.size / pFoldNum
		
		## Note: if pFoldNum is not a factor of the size of the cross-validion
		## data, the remainder will never be used as test set.
		
		# list of list of predictions per fold: fold numbers are keys
		vllPredictions = []
		# list of dictionaries of scores per fold: fold numbers are keys
		vldScores = []
		
		for i in range(pFoldNum):
			vfOutLog.write("Cross-validation %d/%d ...\n\n" % (i + 1, pFoldNum))
			
			# splitting train/test
			
			vEvalSetRange = ((i * vChunckSize) + 1, (i + 1) * vChunckSize)
			
			vEvalSet = vDataset.chunk(vEvalSetRange[0], vEvalSetRange[1])
			vTrainSet = vDataset.chunk(1, vEvalSetRange[0] - 1) + vDataset.chunk(vEvalSetRange[1] + 1)
			
			# sanity check
			if vEvalSet.size != vChunckSize or vTrainSet.size != (vDataset.size - vEvalSet.size):
				raise Exception("There is a problem with splitting the CV dataset: dataset size: %s, chunk size: %s, training set size: %s, evaluation set size: %s" % (vDataset.size, vChunckSize, vTrainSet.size, vEvalSet.size))
			
			# training
			vdModelFilenames = self.train(pMCMethod = pMCMethod, pDataset = vTrainSet, pdLearnerOptions = pdLearnerOptions, pfOutLog = pfOutLog, pfErrLog = pfErrLog)
			
			# testing
			vlPredictions, vdScores, dummy = self.test(pMCMethod = pMCMethod, pdModelFilenames = vdModelFilenames, pDataset = vEvalSet, 
			                                           pdPredictorOptions = pdPredictorOptions, pdScoringOptions = pdScoringOptions, 
			                                           pfOutLog = pfOutLog, pfErrLog = pfErrLog)
			
			vllPredictions.append(vlPredictions)
			vldScores.append(vdScores)
			
			# cleaning up model files (added as a patch to disk space problem on 21/10/2015)
			for vModelFilename in vdModelFilenames.itervalues():
				os.remove(vModelFilename)
		
		# generating and returning the scores
		
		return self._avgCVScores(pldScores = vldScores)
		
	
	
	def gridSearch(self, pMCMethod, pdParamSpec, pdLearnerOptions = {}, pdPredictorOptions = {}, pdScoringOptions = {}, pCVFoldNum = 0, pProcNum = 1, pflgDelTempDir = False, pSearchLogFilename = ''):
		'''
		Performs a grid search for the optimum SVM parameters for binarized
		multiclass classification problem using the specified range of 
		parameter values and returns the best as well as the worst parameter
		sets for each evaluation metric
		
		It also returns the whole processed grid.
		
		pMCMethod specifies the method to be used for multiclass classification.
		The following methods are supported (case insensitive):
		- OVA (or OVR): one-versus-all (one-versus-rest)
		- OVO: one-versus-one
		
		The evaluation of parameters is based on the multiclass output 
		rather than the output of binary classifiers separately. For the
		latter, use SVMLight and individual grid searches for each binary
		classifier. For more information, see the documentation section 
		of this class.
		
		For other details, see SVMLight.gridSearch().
		'''
		
		# setting up train/test options (excluding the search parameters)
		
		vParamEvalArgs = {"CVFoldNum": pCVFoldNum}
		vParamEvalArgs["mcMethod"] = pMCMethod
		vParamEvalArgs["learnerOptions"] = pdLearnerOptions
		vParamEvalArgs["predictorOptions"] = pdPredictorOptions
		vParamEvalArgs["scoringOptions"] = pdScoringOptions
		
		# doing the search
		
		vGridSearch = gridsearch.GridSearch(pdParamSpec = pdParamSpec, pfParamEvaluator = self.gsEvaluateParams, pParamEvalArgs = vParamEvalArgs, pID = self.id)
		vdBestWorst, vlGrid = vGridSearch.search(pProcNum = pProcNum, pflgDelTempDir = pflgDelTempDir, pSearchLogFilename = pSearchLogFilename)
		
		return vdBestWorst, vlGrid
		
	
	
	def gsEvaluateParams(self, pdParamValueSet, pfOutLog, pfErrLog, pdArgs):
		'''
		Evaluates training parameters for multiclass classfication based
		on the specified options and returns the scores
		
		pdArgs["mcMethod"] specifies the method to be used for multiclass 
		classification. The following methods are supported (case insensitive):
		- OVA (or OVR): one-versus-all (one-versus-rest)
		- OVO: one-versus-one
		
		This method is passed as evaluator function to GridSearch class.
		For details, see gridsearch.GridSearch.setParamEvaluator().
		'''
		
		# extracting and setting arguments
		
		vCVFoldNum = 0
		vMCMethod = ''
		vdLearnerOptions = {}
		vdPredictorOptions = {}
		vdScoringOptions = {}
		
		for vArg in pdArgs:
			if vArg.lower() == "cvfoldnum":
				vCVFoldNum = pdArgs[vArg]
			elif vArg.lower() == "mcmethod":
				vMCMethod = pdArgs[vArg]
			elif vArg.lower() == "learneroptions":
				vdLearnerOptions = pdArgs[vArg]
			elif vArg.lower() == "predictoroptions":
				vdPredictorOptions = pdArgs[vArg]
			elif vArg.lower() == "scoringoptions":
				vdScoringOptions = pdArgs[vArg]
		
		# setting the training hyper-parameters
		
		for vParam, vValue in pdParamValueSet.iteritems():
			vdLearnerOptions["-" + vParam] = str(vValue)
		
		# doing the evaluation
		
		if vCVFoldNum == 0:
			vdModelFilenames = self.train(pMCMethod = vMCMethod, pdLearnerOptions = vdLearnerOptions, pfOutLog = pfOutLog, pfErrLog = pfErrLog)
			vlPredictions, vdScores, dummy2 = self.test(pMCMethod = vMCMethod, pdModelFilenames = vdModelFilenames, pdPredictorOptions = vdPredictorOptions, pdScoringOptions = vdScoringOptions, pfOutLog = pfOutLog, pfErrLog = pfErrLog)
			
			# cleaning up model files (added as a patch to disk space problem on 21/10/2015)
			for vModelFilename in vdModelFilenames.itervalues():
				os.remove(vModelFilename)
		
		else:
			vdScores = self.crossValidate(pMCMethod = vMCMethod, pFoldNum = vCVFoldNum, pdLearnerOptions = vdLearnerOptions, pdPredictorOptions = vdPredictorOptions, pdScoringOptions = vdScoringOptions, pfOutLog = pfOutLog, pfErrLog = pfErrLog)
		
		return self.flattenClassScores(vdScores), vlPredictions
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# SVMLightDataset
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class SVMLightDataset:
	'''
	Class for SVMLight dataset
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		# list of data instances of a type inherited from SVMLightDataInstance
		self.instances = []
		
	
	
	def __add__(self, vOtherDataset):
		'''
		Addition operator: creates and returns a dataset whose instances
		are the instances of this and another given dataset
		'''
		
		vNewDataset = SVMLightDataset()
		vNewDataset.instances = self.instances + vOtherDataset.instances
		
		return vNewDataset
		
	
	def __radd__(self, vOtherDataset):
		'''
		List sum operator
		'''
		
		if vOtherDataset == 0:
			return self
		else:
			return self.__add__(vOtherDataset)
		
	
	
	def _createDataset(self):
		'''
		Creates and returns a dataset object
		'''
		
		return SVMLightDataset()
		
	
	
	def _createDataInstance(self):
		'''
		Creates and returns a data instance object
		'''
		
		return SVMLightDataInstance()
		
	
	
	@property
	def size(self):
		'''
		The size of the dataset in terms if the number of data instances
		'''
		
		return len(self.instances)
		
	
	
	def writeToFile(self, pOutFilename):
		'''
		Writes the dataset into file in SVMLight input format
		'''
		
		open(pOutFilename, 'w').write('\n'.join([i.getStr() for i in self.instances]) + '\n')
		
	
	
	def loadFromLines(self, pDatasetLines):
		'''
		Loads the dataset from lines in SVMLight format
		'''
		
		for vILine in pDatasetLines:
			if vILine.startswith('#'):
				continue
			
			self.instances.append(self._createDataInstance())
			self.instances[-1].loadFromLine(vILine)
		
	
	
	def loadFromFile(self, pDatasetFilename):
		'''
		Loads the dataset from file in SVMLight format
		'''
		
		for vILine in open(pDatasetFilename).read().strip().split('\n'):
			if vILine.startswith('#'):
				continue
			
			self.instances.append(self._createDataInstance())
			self.instances[-1].loadFromLine(vILine)
		
	
	
	def shuffled(self):
		'''
		Returns a new dataset with shuffled instances
		'''
		
		vShDataset = self._createDataset()
		vShDataset.instances = self.instances[:]
		random.shuffle(vShDataset.instances)
		
		return vShDataset
		
	
	
	def chunk(self, pStart = None, pEnd = None):
		'''
		Returns a chunck of the dataset from the specified range
		
		Start and end specify the position of the first and last instances
		to be included respectively. If no range parameter is set, the 
		whole dataset will be returned.
		'''
		
		if pStart != None:
			vStart = pStart - 1
		
		vChDataset = self._createDataset()
		vChDataset.instances = self.instances[vStart : pEnd]
		
		return vChDataset
		
	
	
	def getTargets(self):
		'''
		Returns the target strings of the dataset instances in a list in 
		the same order as the instances
		'''
		
		return [i.getTarget() for i in self.instances]
		
	
	
	def getTargetValues(self):
		'''
		Returns the float target values of the dataset instances in a list 
		in the same order as the instances
		
		Target values are returned in float type because in reality SVM
		work with real values.
		'''
		
		return [i.getTargetValue() for i in self.instances]
		
	
	
	def getTargetSet(self):
		'''
		Returns the target value set of the dataset in a dictionary with 
		entries being the target values and their counts
		'''
		
		return util.groupBy(self.getTargets())
		
	
	
	def generateOVADatasets(self):
		'''
		Generates and returns one-versus-all data sets from the dataset
		if there are more than two target values in the dataset
		'''
		
		if len(self.getTargetSet()) <= 2:
			raise Exception("OVA can only be applied to multiclass problems!")
		
		# writing the dataset into file temporarily to be used by OVA
		
		vTempFilename = tempfile.mktemp(suffix="ova")
		self.writeToFile(pOutFilename = vTempFilename)
		
		# generating OVA datasets
		
		vOVA = mc2bc.OVA()
		vdOrgOVADatasets = vOVA.generateOVADatasets(pMCDatasetFilename = vTempFilename, pInputFormat = "svmlight", pOutputFormat = "str", )
		
		# cleanup
		
		os.remove(vTempFilename)
		
		# converting the original OVA datasets to SVMLightDatasets
		
		vdSVMLightOVADatasets = {}
		
		for vClass in vdOrgOVADatasets:
			vNewDataset = self._createDataset()
			vNewDataset.loadFromLines(vdOrgOVADatasets[vClass][0])
			vdSVMLightOVADatasets[vClass] = vNewDataset
		
		return vdSVMLightOVADatasets
		
	
	
	def generateOVODatasets(self, plClassPairs):
		'''
		Generates and returns one-versus-one data sets from the dataset
		if there are more than two target values in the dataset
		
		Class pairs are given in a list of pair tuples and indicate the 
		way the one-vs-one classes are paired in each binary dataset. For
		example (-1, 0) indicates that the positive class must be assigned 
		to -1 instances and the negative to 0 ones. If not given, an ad-hoc 
		order will be used.
		'''
		
		if len(self.getTargetSet()) <= 2:
			raise Exception("OVO can only be applied to multiclass problems!")
		
		# writing the dataset into file temporarily to be used by OVO
		
		vTempFilename = tempfile.mktemp(suffix="ova")
		self.writeToFile(pOutFilename = vTempFilename)
		
		## extracting class order from class pair list because the OVO 
		## dataset generator accepts the order list of classes instead.
		
		## finding the counts of classes being the first element of the pairs 
		## (The first argument of map finds the counts per class and the
		##  second extracts the class labels from pairs; yes a bit of clever
		##  coding!)
		vl1stClassCounts = map(lambda c: (c, 
		                                  len([p[0] for p in plClassPairs if p[0] == c])),
		                       list(set([c for p in plClassPairs for c in p])))
		# sorting the list of first classes in the pairs based on their counts
		vd1stClassCounts = {i[0]: i[1] for i in vl1stClassCounts}
		vlClassOrder = sorted(vd1stClassCounts, key = lambda x: vd1stClassCounts[x], reverse = True)
		
		# generating OVO datasets
		
		vOVO = mc2bc.OVO()
		vdOrgOVODatasets = vOVO.generateOVODatasets(pMCDatasetFilename = vTempFilename, pInputFormat = "svmlight", pOutputFormat = "str", plClassOrder = vlClassOrder)
		
		# cleanup
		
		os.remove(vTempFilename)
		
		# converting the original OVO datasets to SVMLightDatasets
		
		vdSVMLightOVODatasets = {}
		
		for vClass1 in vdOrgOVODatasets:
			for vClass2 in vdOrgOVODatasets[vClass1]:
				vNewDataset = self._createDataset()
				vNewDataset.loadFromLines(vdOrgOVODatasets[vClass1][vClass2][0])
				vdSVMLightOVODatasets[(vClass1, vClass2)] = vNewDataset
		
		return vdSVMLightOVODatasets
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# SVMLightDataInstance
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class SVMLightDataInstance:
	'''
	Class for SVMLight data instance
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		# feature vector string (e.g. "1:x 2:y 3:z")
		self.fv = ''
		# the target value to be learnt/predicted
		self.target = None
		
	
	
	def __str__(self):
		'''
		String representation of the instance in SVMLight format
		'''
		
		return self.getStr()
		
	
	
	def getStr(self):
		'''
		Returns the string representation of the data instance in SVMLight
		format
		'''
		
		return "%s\t%s" % (self.target, self.fv)
		
	
	
	def loadFromLine(self, pLine):
		'''
		Loads the data instance from a line in SVMLight format
		'''
		
		vlTargetFVSplit = re.search("^([^ \t]+)[ \t]+(.+)", pLine).group(1, 2)
		
		#self.target = float(vlTargetFVSplit[0])
		self.target = vlTargetFVSplit[0]
		self.fv = vlTargetFVSplit[1]
		
	
	
	def getTarget(self):
		'''
		Returns the target string of the instance
		'''
		
		return self.target
		
	
	
	def getTargetValue(self):
		'''
		Returns the float target value of the instance
		
		Target value is returned in float type because in reality SVM work
		with real values.
		'''
		
		return float(self.target)
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# SVMLightTK
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class SVMLightTK(SVMLight):
	'''
	Class for tree kernel version of SVMLight
	'''
	
	
	def __init__(self, pLearnerPath = '', pClassifierPath = '', pID = '', pLogFilePrefix = ''):
		'''
		Constructor
		'''
		
		SVMLight.__init__(self, pLearnerPath, pClassifierPath, pID, pLogFilePrefix)
		
		# loading config file if either of paths are not provided 
		if pLearnerPath == '' or pClassifierPath == '':
			vAppPath = os.path.dirname(__file__)
			vConfigParser = SafeConfigParser()
			vConfigParser.read(os.path.join(vAppPath, "svmlight.ini"))
			self.learner = vConfigParser.get("SVMLIGHT-TK","learner")
			self.classifier = vConfigParser.get("SVMLIGHT-TK","classifier")
		
	
	
	def _createDataset(self):
		'''
		Creates and returns a dataset object
		'''
		
		return SVMLightTKDataset()
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# SVMLightTKMC
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class SVMLightTKMC(SVMLightMC):
	'''
	Class for tree kernel version of SVMLight with multiclass support
	'''
	
	
	def __init__(self, pLearnerPath = '', pClassifierPath = '', pID = '', pLogFilePrefix = ''):
		'''
		Constructor
		'''
		
		SVMLightMC.__init__(self, pLearnerPath, pClassifierPath, pID, pLogFilePrefix)
		
		# loading config file if either of paths are not provided 
		if pLearnerPath == '' or pClassifierPath == '':
			vAppPath = os.path.dirname(__file__)
			vConfigParser = SafeConfigParser()
			vConfigParser.read(os.path.join(vAppPath, "svmlight.ini"))
			self.learner = vConfigParser.get("SVMLIGHT-TK","learner")
			self.classifier = vConfigParser.get("SVMLIGHT-TK","classifier")
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# SVMLightTKDataset
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class SVMLightTKDataset(SVMLightDataset):
	'''
	Class for SVMLightTK dataset
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		SVMLightDataset.__init__(self)
		
	
	
	def _createDataset(self):
		'''
		Creates and returns a dataset object
		'''
		
		return SVMLightTKDataset()
		
	
	
	def _createDataInstance(self):
		'''
		Creates and returns a data instance object
		'''
		
		return SVMLightTKDataInstance()
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# SVMLightDataInstance
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class SVMLightTKDataInstance(SVMLightDataInstance):
	'''
	Class for SVMLightTK data instance
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		SVMLightDataInstance.__init__(self)
		
		# list of tree representations
		self.trees = []
		
	
	
	def getStr(self):
		'''
		Returns the string representation of the data instance in SVMLightTK
		format
		'''
		
		if self.fv != '':
			return "%s\t|BT| %s |ET| %s |EV|" % (self.target, ' |BT| '.join([t.strip() for t in self.trees]), self.fv)
		else:
			return "%s\t|BT| %s |ET|" % (self.target, ' |BT| '.join([t.strip() for t in self.trees]))
		
	
	
	def loadFromLine(self, pLine):
		'''
		Loads the data instance from a line in SVMLightTK format
		'''
		
		vlTargetTFVSplit = re.search("^([^ \t]+)[ \t]+(.+)", pLine).group(1, 2)
		
		#self.target = float(vlTargetTFVSplit[0])
		self.target = vlTargetTFVSplit[0]
		
		vlTreeFVSplit = vlTargetTFVSplit[1].split('|ET|')
		
		# extracting trees
		for vTree in vlTreeFVSplit[0].strip().split('|BT|')[1 : ]:
			self.trees.append(vTree.strip())
		
		# extracting feature/values
		self.fv = vlTreeFVSplit[1].strip()
		
	
	


