#! /usr/bin/python
# -*- coding: utf-8 -*-

"""	
	This module defines classes for tree kernels.
	
	Version 0.8                                             (16-May-2016)
	- Unary formats in NGramKTree is updated.
	
	Version 0.7                              (04-Apr-2016 to 11-Apr-2016)
	- Bigram and binary formats are added to NGramKTree.
	
	Version 0.6                              (17-Nov-2015 to 24-Nov-2015)
	- PropositionKTrees.generatePropTrees() is edited to work for all tree
	  classes derived from DepTree and ConstTree.
	- The format numbers in PropositionKTree.generateFromConstTree() is 
	  converted to strings representing the format of the innermost treelet
	  in the format, which has been the new format definition since v0.4.
	
	Version 0.5								(06-Oct-2015 to 13-Oct-2015)
	- DepKTree.generateDepKTree() supports subtrees under a node provided
	  as a parameter in addition to the whole tree under the root.
	- New formats are added to the DepKTree.
	- ConstDepKTree is added.
	- The format numbers in ConstKTree.generateConstKTree() is converted to 
	  strings representing the format of the innermost treelet in the 
	  format. Necessary arguments are added to specify the format details.
	- NGramTree is added to construct a tree from text token forms.
	
	Version 0.4								(24-Aug-2015 to 25-Aug-2015)
	- The format numbers in DepKTree.generateDepKTree() is converted to 
	  strings representing the format of the innermost treelet in the 
	  format.
	- New formats are added to DepKTree.generateDepKTree().
	
	Version 0.3												(29-Jan-2014)
	- The constructor of PropositionKTrees is edited. It requires the
	  tree as argument. The type of the tree is identified automatically.
	- PropositionKTrees.generateFromDepTrees() is renamed to generatePropTrees()
	  and does not require tree as parameter any longer as the tree is 
	  passed to the constructor.
	- ConstKTree.generateConstKTree() is re-factored to generate proposition-
	  specific annotation.
	
	Version 0.2												(28-Jan-2014)
	- ConsKTree is added.
	- Generating proposition trees from constituency trees is added.
	
	Version 0.1												(15-Jan-2014)
	- PropositionKTree and PropositionKTrees are added.
	- DepKTree is added.
	
"""



class PropositionKTrees:
	'''
	The class for proposition trees of a sentence
	'''
	
	
	def __init__(self, pTree):
		'''
		Creates a proposition tree object
		
		pTree can be a dependency or constituency tree which is identified
		automatically.
		'''
		
		self._sourceTree = pTree
		
		# list of proposition trees of type PropositionKTree
		self._propTrees = None
		
	
	
	@property
	def trees(self):
		'''
		Returns the trees in bracketing format
		'''
		
		return [t.tree for t in self._propTrees]
		
	
	
	def generatePropTrees(self, pFormat):
		'''
		Generates the proposition trees of the propositions of the SRL of
		the given dependency tree
		'''
		
		import inspect
		
		# extracting hte class hierarchy to determine tree type later on
		vlClassHierarchy = [c.__name__ for c in inspect.getmro(self._sourceTree.__class__)]
		
		self._propTrees = []
		for vProp in self._sourceTree.srl.propositions:
			vPropTree = PropositionKTree()
			if "DepTree" in vlClassHierarchy:
				vPropTree.generateFromDepTree(self._sourceTree, vProp, pFormat)
			elif "ConstTree" in vlClassHierarchy:
				vPropTree.generateFromConstTree(self._sourceTree, vProp, pFormat)
			self._propTrees.append(vPropTree)
		
		return self.trees
		
	
	


class PropositionKTree:
	'''
	The class for proposition tree
	
	Various formats of proposition tree are implemented. Refer to the 
	documentation of the methods implementing each format. Formats are 
	numbered:
	0: the basic (default) format: see _generateFromDepTree0()
	1: see _generateFromDepTree1()
	TODO: implement on demand
	
	A PTB bracketing-style tree is generated for the proposition.
	'''
	
	
	def __init__(self):
		'''
		Creates a proposition tree object
		'''
		
		self.tree = ""
		
	
	
	def generateFromDepTree(self, pDepTree, pProposition, pFormat = 0):
		'''
		Generates the proposition tree of the given proposition from a
		dependency tree in the required format
		'''
		
		if pFormat == 0:
			self.tree = self._generateFromDepTree0(pDepTree, pProposition)
		if pFormat == 1:
			self.tree = self._generateFromDepTree1(pDepTree, pProposition)
		if pFormat == 2:
			self.tree = self._generateFromDepTree2(pDepTree, pProposition)
		if pFormat == 3:
			self.tree = self._generateFromDepTree3(pDepTree, pProposition)
		if pFormat == 4:
			self.tree = self._generateFromDepTree4(pDepTree, pProposition)
		if pFormat == 5:
			self.tree = self._generateFromDepTree5(pDepTree, pProposition)
		if pFormat == 6:
			self.tree = self._generateFromDepTree6(pDepTree, pProposition)
		
		return self.tree
		
	
	
	def _generateFromDepTree0(self, pDepTree, pProposition):
		'''
		Generates the proposition tree format 0 of the given proposition from a
		dependency tree
		
		Format 0 proposition tree has a root which is the predicate frameset. 
		Argument labels and predicate lemma come under the root. Each 
		argument label in turn has the token(s) filling the role as its
		children. Similarly, the predicate lemma has the predicate form 
		under it. The order of arguments and predicate corresponds to 
		their positions in the sentence.
		'''
		
		## generating argument nodes while checking for the right place
		## to insert the predicate node
		vlNodes = []
		vflgPredAdded = False
		for vArg in pProposition.arguments:
			if not vflgPredAdded and vArg.span[0] > pProposition.predicate.position:
				vlNodes.append("(%s %s)" % (pProposition.predicate.lemma, pDepTree.getNode(pProposition.predicate.position).getForm()))
				vflgPredAdded = True
			
			vlNodes.append("(%s %s)" % (vArg.label, pDepTree.getNode(vArg.span[0]).getForm()))
			
			# sanity check
			if vArg.span[0] <= pProposition.predicate.position <= vArg.span[1]:
				raise Exception("Problem with predicate and argument positions: %s and (%s, %s) at sentence:\n%s" % (pProposition.predicate.position, vArg.span[0], vArg.span[1], pDepTree))
		
		## in case if the predicate is located after all arguments (or 
		## has no argument)
		if not vflgPredAdded:
			vlNodes.append("(%s %s)" % (pProposition.predicate.lemma, pDepTree.getNode(pProposition.predicate.position).getForm()))
		
		return "(%s %s)" % (pProposition.predicate.frameset, ''.join(vlNodes))
		
	
	
	def _generateFromDepTree1(self, pDepTree, pProposition):
		'''
		Generates the proposition tree format 1 of the given proposition 
		from a dependency tree
		
		Format 1 is a modification of format 0 by adding dependency to 
		leaves (under word forms).
		'''
		
		## generating argument nodes while checking for the right place
		## to insert the predicate node
		vlNodes = []
		vflgPredAdded = False
		for vArg in pProposition.arguments:
			if not vflgPredAdded and vArg.span[0] > pProposition.predicate.position:
				vlNodes.append("(%s (%s %s))" % (pProposition.predicate.lemma, pDepTree.getNode(pProposition.predicate.position).getForm(), pDepTree.getNode(pProposition.predicate.position).getDepRel()))
				vflgPredAdded = True
			
			vlNodes.append("(%s (%s %s))" % (vArg.label, pDepTree.getNode(vArg.span[0]).getForm(), pDepTree.getNode(vArg.span[0]).getDepRel()))
			
			# sanity check
			if vArg.span[0] <= pProposition.predicate.position <= vArg.span[1]:
				raise Exception("Problem with predicate and argument positions: %s and (%s, %s) at sentence:\n%s" % (pProposition.predicate.position, vArg.span[0], vArg.span[1], pDepTree))
		
		## in case if the predicate is located after all arguments (or 
		## has no argument)
		if not vflgPredAdded:
			vlNodes.append("(%s (%s %s))" % (pProposition.predicate.lemma, pDepTree.getNode(pProposition.predicate.position).getForm(), pDepTree.getNode(pProposition.predicate.position).getDepRel()))
		
		return "(%s %s)" % (pProposition.predicate.frameset, ''.join(vlNodes))
		
	
	
	def _generateFromDepTree2(self, pDepTree, pProposition):
		'''
		Generates the proposition tree format 2 of the given proposition 
		from a dependency tree
		
		Format 2 is a modification of format 1 by adding dependency as
		sibling of word forms instead of under them.
		'''
		
		## generating argument nodes while checking for the right place
		## to insert the predicate node
		vlNodes = []
		vflgPredAdded = False
		for vArg in pProposition.arguments:
			if not vflgPredAdded and vArg.span[0] > pProposition.predicate.position:
				vlNodes.append("(%s (%s )(%s ))" % (pProposition.predicate.lemma, pDepTree.getNode(pProposition.predicate.position).getForm(), pDepTree.getNode(pProposition.predicate.position).getDepRel()))
				vflgPredAdded = True
			
			vlNodes.append("(%s (%s )(%s ))" % (vArg.label, pDepTree.getNode(vArg.span[0]).getForm(), pDepTree.getNode(vArg.span[0]).getDepRel()))
			
			# sanity check
			if vArg.span[0] <= pProposition.predicate.position <= vArg.span[1]:
				raise Exception("Problem with predicate and argument positions: %s and (%s, %s) at sentence:\n%s" % (pProposition.predicate.position, vArg.span[0], vArg.span[1], pDepTree))
		
		## in case if the predicate is located after all arguments (or 
		## has no argument)
		if not vflgPredAdded:
			vlNodes.append("(%s (%s )(%s ))" % (pProposition.predicate.lemma, pDepTree.getNode(pProposition.predicate.position).getForm(), pDepTree.getNode(pProposition.predicate.position).getDepRel()))
		
		return "(%s %s)" % (pProposition.predicate.frameset, ''.join(vlNodes))
		
	
	
	def _generateFromDepTree3(self, pDepTree, pProposition):
		'''
		Generates the proposition tree format 3 of the given proposition 
		from a dependency tree
		
		Format 3 is a modification of format 2 by adding POS tag as
		sibling of word form and dependency relation.
		'''
		
		## generating argument nodes while checking for the right place
		## to insert the predicate node
		vlNodes = []
		vflgPredAdded = False
		for vArg in pProposition.arguments:
			if not vflgPredAdded and vArg.span[0] > pProposition.predicate.position:
				vlNodes.append("(%s (%s )(%s )(%s ))" % (pProposition.predicate.lemma, 
				                                         pDepTree.getNode(pProposition.predicate.position).getForm(), 
				                                         pDepTree.getNode(pProposition.predicate.position).getDepRel(), 
				                                         pDepTree.getNode(pProposition.predicate.position).getPOSTag()))
				vflgPredAdded = True
			
			vlNodes.append("(%s (%s )(%s )(%s ))" % (vArg.label, pDepTree.getNode(vArg.span[0]).getForm(), pDepTree.getNode(vArg.span[0]).getDepRel(), pDepTree.getNode(vArg.span[0]).getPOSTag()))
			
			# sanity check
			if vArg.span[0] <= pProposition.predicate.position <= vArg.span[1]:
				raise Exception("Problem with predicate and argument positions: %s and (%s, %s) at sentence:\n%s" % (pProposition.predicate.position, vArg.span[0], vArg.span[1], pDepTree))
		
		## in case if the predicate is located after all arguments (or 
		## has no argument)
		if not vflgPredAdded:
			vlNodes.append("(%s (%s )(%s )(%s ))" % (pProposition.predicate.lemma, 
			                                         pDepTree.getNode(pProposition.predicate.position).getForm(), 
			                                         pDepTree.getNode(pProposition.predicate.position).getDepRel(), 
			                                         pDepTree.getNode(pProposition.predicate.position).getPOSTag()))
		
		return "(%s %s)" % (pProposition.predicate.frameset, ''.join(vlNodes))
		
	
	
	def _generateFromDepTree4(self, pDepTree, pProposition):
		'''
		Generates the proposition tree format 4 of the given proposition 
		from a dependency tree
		
		Format 4 is a modification of format 3 by adding the head (word 
		form) of the node as another child.
		e.g. (Arg1 (book )(dobj )(NN )(read ))
		'''
		
		## generating argument nodes while checking for the right place
		## to insert the predicate node
		vlNodes = []
		vflgPredAdded = False
		for vArg in pProposition.arguments:
			if not vflgPredAdded and vArg.span[0] > pProposition.predicate.position:
				vlNodes.append("(%s (%s )(%s )(%s )(%s ))" % (pProposition.predicate.lemma, 
				                                              pDepTree.getNode(pProposition.predicate.position).getForm(), 
				                                              pDepTree.getNode(pProposition.predicate.position).getDepRel(), 
				                                              pDepTree.getNode(pProposition.predicate.position).getPOSTag(), 
				                                              pDepTree.getNode(pProposition.predicate.position).getHeadNode().getForm() if pDepTree.getNode(pProposition.predicate.position).getHeadNode().getForm() != '' else "root"))
				vflgPredAdded = True
			
			vlNodes.append("(%s (%s )(%s )(%s )(%s ))" % (vArg.label, 
			                                              pDepTree.getNode(vArg.span[0]).getForm(), 
			                                              pDepTree.getNode(vArg.span[0]).getDepRel(), 
			                                              pDepTree.getNode(vArg.span[0]).getPOSTag(), 
			                                              pDepTree.getNode(vArg.span[0]).getHeadNode().getForm() if pDepTree.getNode(vArg.span[0]).getHeadNode().getForm() != '' else "root"))
			
			# sanity check
			if vArg.span[0] <= pProposition.predicate.position <= vArg.span[1]:
				raise Exception("Problem with predicate and argument positions: %s and (%s, %s) at sentence:\n%s" % (pProposition.predicate.position, vArg.span[0], vArg.span[1], pDepTree))
		
		## in case if the predicate is located after all arguments (or 
		## has no argument)
		if not vflgPredAdded:
			vlNodes.append("(%s (%s )(%s )(%s )(%s ))" % (pProposition.predicate.lemma, 
			                                              pDepTree.getNode(pProposition.predicate.position).getForm(), 
			                                              pDepTree.getNode(pProposition.predicate.position).getDepRel(), 
			                                              pDepTree.getNode(pProposition.predicate.position).getPOSTag(), 
			                                              pDepTree.getNode(pProposition.predicate.position).getHeadNode().getForm() if pDepTree.getNode(pProposition.predicate.position).getHeadNode().getForm() != '' else "root"))
		
		return "(%s %s)" % (pProposition.predicate.frameset, ''.join(vlNodes))
		
	
	
	def _generateFromDepTree5(self, pDepTree, pProposition):
		'''
		Generates the proposition tree format 5 of the given proposition 
		from a dependency tree
		
		Format 5 is a modification of format 4 by adding the path of 
		dependency relations and head words to the tree root to the dependency
		relation leaf. 
		e.g. (AM-TMP (Once )(advmod (know root))(RB ))
		'''
		
		## generating argument nodes while checking for the right place
		## to insert the predicate node
		vlNodes = []
		vflgPredAdded = False
		for vArg in pProposition.arguments:
			if not vflgPredAdded and vArg.span[0] > pProposition.predicate.position:
				vlNodes.append("(%s (%s )%s(%s ))" % (pProposition.predicate.lemma, 
				                                              pDepTree.getNode(pProposition.predicate.position).getForm(), 
				                                              self.argPathToDepTreeRoot(pDepTree.getNode(pProposition.predicate.position)), 
				                                              pDepTree.getNode(pProposition.predicate.position).getPOSTag()))
				vflgPredAdded = True
			
			vlNodes.append("(%s (%s )%s(%s ))" % (vArg.label, 
			                                              pDepTree.getNode(vArg.span[0]).getForm(), 
			                                              self.argPathToDepTreeRoot(pDepTree.getNode(vArg.span[0])), 
			                                              pDepTree.getNode(vArg.span[0]).getPOSTag()))
			
			# sanity check
			if vArg.span[0] <= pProposition.predicate.position <= vArg.span[1]:
				raise Exception("Problem with predicate and argument positions: %s and (%s, %s) at sentence:\n%s" % (pProposition.predicate.position, vArg.span[0], vArg.span[1], pDepTree))
		
		## in case if the predicate is located after all arguments (or 
		## has no argument)
		if not vflgPredAdded:
			vlNodes.append("(%s (%s )%s(%s ))" % (pProposition.predicate.lemma, 
			                                              pDepTree.getNode(pProposition.predicate.position).getForm(), 
			                                              self.argPathToDepTreeRoot(pDepTree.getNode(pProposition.predicate.position)), 
			                                              pDepTree.getNode(pProposition.predicate.position).getPOSTag()))
		
		return "(%s %s)" % (pProposition.predicate.frameset, ''.join(vlNodes))
		
	
	
	def _generateFromDepTree6(self, pDepTree, pProposition):
		'''
		Generates the proposition tree format 6 of the given proposition 
		from a dependency tree
		
		Format 6 is a modification of format 5 by marking the head/dependency 
		path with argument role label
		e.g. (AM-TMP (Once )(advmod_AM-TMP (know root_AM-TMP))(RB ))
		'''
		
		## generating argument nodes while checking for the right place
		## to insert the predicate node
		vlNodes = []
		vflgPredAdded = False
		for vArg in pProposition.arguments:
			if not vflgPredAdded and vArg.span[0] > pProposition.predicate.position:
				vlNodes.append("(%s (%s )%s(%s ))" % (pProposition.predicate.lemma, 
				                                              pDepTree.getNode(pProposition.predicate.position).getForm(), 
				                                              self.argPathToDepTreeRoot(pDepTree.getNode(pProposition.predicate.position)), 
				                                              pDepTree.getNode(pProposition.predicate.position).getPOSTag()))
				vflgPredAdded = True
			
			vlNodes.append("(%s (%s )%s(%s ))" % (vArg.label, 
			                                              pDepTree.getNode(vArg.span[0]).getForm(), 
			                                              self.argPathToDepTreeRoot(pDepTree.getNode(vArg.span[0]), pflgMarkDepRelNodes = True, pArgLabel = vArg.label), 
			                                              pDepTree.getNode(vArg.span[0]).getPOSTag()))
			
			# sanity check
			if vArg.span[0] <= pProposition.predicate.position <= vArg.span[1]:
				raise Exception("Problem with predicate and argument positions: %s and (%s, %s) at sentence:\n%s" % (pProposition.predicate.position, vArg.span[0], vArg.span[1], pDepTree))
		
		## in case if the predicate is located after all arguments (or 
		## has no argument)
		if not vflgPredAdded:
			vlNodes.append("(%s (%s )%s(%s ))" % (pProposition.predicate.lemma, 
			                                              pDepTree.getNode(pProposition.predicate.position).getForm(), 
			                                              self.argPathToDepTreeRoot(pDepTree.getNode(pProposition.predicate.position)), 
			                                              pDepTree.getNode(pProposition.predicate.position).getPOSTag()))
		
		return "(%s %s)" % (pProposition.predicate.frameset, ''.join(vlNodes))
		
	
	
	def argPathToDepTreeRoot(self, pDepNode, pflgMarkDepRelNodes = False, pArgLabel = ""):
		'''
		Generates and returns the path from a dependency node to the root
		of the tree it belongs to
		
		pflgMarkDepRelNodes specifies whether the dependency relation in 
		the path must be marked with the argument role label given in pArgLabel.
		'''
		
		vPath = ""
		
		if pflgMarkDepRelNodes:
			vMark = '_' + pArgLabel
		else:
			vMark = ''
		
		if pDepNode.getHeadNode().isRoot():
			vPath = "(root )"
		elif pDepNode.getHeadNode().getHeadNode().isRoot():
			vPath = "(%s (%s root%s))" % (pDepNode.getDepRel() + vMark, pDepNode.getHeadNode().getForm(), vMark)
		else:
			vPath = "(%s (%s %s))" % (pDepNode.getDepRel() + vMark, pDepNode.getHeadNode().getForm(), self.argPathToDepTreeRoot(pDepNode.getHeadNode(), pflgMarkDepRelNodes, pArgLabel))
		
		return vPath
		
	
	
	def generateFromConstTree(self, pConstTree, pProposition, pFormat):
		'''
		Generates the proposition tree of the given proposition from a
		constituency tree in the required format
		'''
		
		if pFormat == "(phrase_args (pos_args form))":
			self.tree = self._generateFromConstTree0(pConstTree, pProposition)[2: -1]
		
		return self.tree
		
	
	
	def _generateFromConstTree0(self, pConstTree, pProposition):
		'''
		Generates the proposition tree format 0 of the given proposition 
		from a constituency tree
		
		Format 0 proposition tree is a proposition subtree (PST) with
		argument role labels of a given proposition concatenated with
		the syntactic label of the argument node:
		  (phrase_args (pos_args form))
		'''
		
		vPST, vPSTPredNode = pConstTree.getPST(pProposition.predicate.getNode())
		vConstKTree = ConstKTree(vPST)
		
		return vConstKTree.generateConstKTree(pPredNode = vPSTPredNode, pFormat = "(phrase_args (pos_args form))")
		
	
	


class DepKTree:
	'''
	The class for dependency tree for use in tree kernels.
	
	For use in tree kernels, the tree is represented in bracketing format.
	'''
	
	
	def __init__(self, pDepTree):
		'''
		Creates a dependency kernel tree object
		'''
		
		self.depTree = pDepTree
		self.kTree = ""
		
	
	
	def generateDepKTree(self, pNode = "root", pFormat = "(rel form)", pdOptions = {}):
		'''
		Generates dependency kernel tree or subtree under a given node in
		the required format 
		
		NOTE: The subtree here should not be confused with the notion of
		subtree as a tree kernel variation used in parallel to subset tree
		kernel.
		
		The returning represenation is a s-expression.
		
		When a subtree under a node is considered, the representation can
		be top-down or bottom-up. However, when the whole tree is considered,
		i.e. the given node is the root (pNode = "root", default), only 
		top-down representation is possible. In a bottom-up format, the
		word form precedes the dependency relation, and in a top-down format,
		the opposite is true.
		
		The default format is (rel form) which is the pure dependency tree
		with only dependency relations and token forms as node labels. In
		general, the format string is the representation of the innermost
		treelet in the format. See each format-specific method for exact
		details.
		
		For each format, a recursive function is defined which starting 
		from the node. In a top-down format, it generates subtrees for each
		node down to leaves. In a bottom-up format, it generates subtrees 
		for each node up to the root.
		
		A leaf node of the dependency tree (node with no dependent in 
		top-down or the root node in bottom-up format) is rewritten in (<1> <2>)
		format. Component <1> is always a single label (e.g. dependency 
		relation). Components <2> can be a single label (e.g. word form)
		or any other format which forms a valid subtree with <1>. 
		
		An internal node of the dependency tree is rewritten in 
		(<1> <2> .. <n>) format. Component <1> is the same as in leaf node.
		Components <2> to <n> are subtrees (under <1>) where at least one
		of them is in (<1> <2>) format where <2> consists of subtrees for
		dependent nodes (top-down format) or head nodes (bottom-up) of 
		this node generated recursively.
		
		pdOptions contains format options for specific formats.
		'''
		
		if type(pNode) is str and pNode.lower() == "root":
			pNode = self.depTree.root
		
		# old format 0
		if pFormat == "(rel form)":
			self.kTree = "( %s)" % ''.join([self._generateDepKSubtree0(self.depTree.getNode(d), 0) for d in pNode.dependents])
		# old format 1
		elif pFormat == "(rel_roles form)":
			self.kTree = "( %s)" % ''.join([self._generateDepKSubtree1(self.depTree.getNode(d), 0) for d in pNode.dependents])
		# old format 2
		elif pFormat == "(rel_roles form)":
			self.kTree = "( %s)" % ''.join([self._generateDepKSubtree2(self.depTree.getNode(d), 0) for d in pNode.dependents])
		# old format 3
		elif pFormat == "(rel (form roles))":
			self.kTree = "( %s)" % ''.join([self._generateDepKSubtree3(self.depTree.getNode(d), 0) for d in pNode.dependents])
		# old format 4
		elif pFormat == "(rel (roles form))":
			self.kTree = "( %s)" % ''.join([self._generateDepKSubtree4(self.depTree.getNode(d), 0) for d in pNode.dependents])
		# old format 5
		elif pFormat == "(rel (roles )(form ()))":
			self.kTree = "( %s)" % ''.join([self._generateDepKSubtree5(self.depTree.getNode(d), 0) for d in pNode.dependents])
		# new format
		elif pFormat == "(rel (pos form))":
			self.kTree = "( %s)" % ''.join([self._generateDepKSubtree6(self.depTree.getNode(d), 0) for d in pNode.dependents])
		# new format
		elif pFormat == "(rel (pos )(form ()))":
			self.kTree = "( %s)" % ''.join([self._generateDepKSubtree7(self.depTree.getNode(d), 0) for d in pNode.dependents])
		# new format
		elif pFormat == "(form (rel head))":
			self.kTree = self._generateDepKSubtree8(pNode, pdOptions = pdOptions)
		
		return self.kTree
		
	
	
	# (rel form) -> old format 0
	def _generateDepKSubtree0(self, pNode, pCurrentHead):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (rel form) format
		
		pCurrentHead identifies which head is calling this method in 
		case the node has multiple heads.
		
		In this format:
		leaf node    : (<dependency relation> <token form>)
		internal node: (<dependency relation> (<token form> <dependent 1> .. <dependent n>))
		'''
		
		vDependents = ''.join([self._generateDepKSubtree0(self.depTree.getNode(d), pNode.position) for d in pNode.dependents])
		
		if len(pNode.dependents) == 0:
			vKSubtree = "(%s %s)" % (pNode.getDepRel(pCurrentHead),
			                         pNode.form)
		else:
			vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
			                              pNode.form,
			                              vDependents)
		
		return vKSubtree
		
	
	
	# (rel_roles form) -> old format 1
	def _generateDepKSubtree1(self, pNode, pCurrentHead):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (rel_roles form) format
		
		pCurrentHead identifies which head is calling this method in 
		case the node has multiple heads.
		
		In this format:
		leaf node    : (<dependency relation + argument role labels> <token form>)
		internal node: (<dependency relation + argument role labels> (<token form> <dependent 1> .. <dependent n>))
		'''
		
		vDependents = ''.join([self._generateDepKSubtree1(self.depTree.getNode(d), pNode.position) for d in pNode.dependents])
		
		if len(pNode.dependents) == 0:
			vKSubtree = "(%s %s)" % ('_'.join([pNode.getDepRel(pCurrentHead)] + pNode.getArgRoles()), 
			                         pNode.form)
		else:
			vKSubtree = "(%s (%s %s))" % ('_'.join([pNode.getDepRel(pCurrentHead)] + pNode.getArgRoles()), 
			                              pNode.form, 
			                              vDependents)
		
		return vKSubtree
		
	
	
	# (rel_roles form) -> old format 2
	def _generateDepKSubtree2(self, pNode, pCurrentHead):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (rel_roles form) format
		
		pCurrentHead identifies which head is calling this method in 
		case the node has multiple heads.
		
		In this format:
		leaf node    : (<dependency relation + argument role labels / predicate frameset> <token form>)
		internal node: (<dependency relation + argument role labels / predicate frameset> (<token form> <dependent 1> .. <dependent n>))
		'''
		
		vPredicateFrameset = pNode.getPredicateFrameset()
		if vPredicateFrameset == '':
			vLabel1 = '_'.join([pNode.getDepRel(pCurrentHead)] + pNode.getArgRoles())
		else:
			vLabel1 = '_'.join([pNode.getDepRel(pCurrentHead)] + pNode.getArgRoles() + [vPredicateFrameset])
		
		vDependents = ''.join([self._generateDepKSubtree2(self.depTree.getNode(d), pNode.position) for d in pNode.dependents])
		
		if len(pNode.dependents) == 0:
			vKSubtree = "(%s %s)" % (vLabel1, 
			                         pNode.form)
		else:
			vKSubtree = "(%s (%s %s))" % (vLabel1, 
			                              pNode.form, 
			                              vDependents)
		
		return vKSubtree
		
	
	
	# (rel (form roles)) -> old format 3
	def _generateDepKSubtree3(self, pNode, pCurrentHead):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (rel (form roles)) format
		
		pCurrentHead identifies which head is calling this method in 
		case the node has multiple heads.
		
		In this format, the annotation components for non-argument and
		predicate nodes are the same as format (rel form). For argument 
		nodes:
		leaf node    : (<dependency relation> (<token form> <argument role labels>))
		internal node: (<dependency relation> (<token form> (<argument role labels> <dependent 1> .. <dependent n>)))
		'''
		
		vDependents = ''.join([self._generateDepKSubtree3(self.depTree.getNode(d), pNode.position) for d in pNode.dependents])
		
		if pNode.isArgument():
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
				                              pNode.form,
				                              '_'.join(pNode.getArgRoles()))
			else:
				vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
				                                   pNode.form,
				                                   '_'.join(pNode.getArgRoles()),
				                                   vDependents)
		else:
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s %s)" % (pNode.getDepRel(pCurrentHead),
				                         pNode.form)
			else:
				vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
				                              pNode.form,
				                              vDependents)
		
		return vKSubtree
		
	
	
	# (rel (roles form)) -> old format 4
	def _generateDepKSubtree4(self, pNode, pCurrentHead):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (rel (roles form)) format
		
		pCurrentHead identifies which head is calling this method in 
		case the node has multiple heads.
		
		In this format, the annotation components for non-argument and
		predicate nodes are the same as format (rel form). For argument 
		nodes:
		leaf node    : (<dependency relation> (<argument role labels> <token form>))
		internal node: (<dependency relation> (<argument role labels> (<token form> <dependent 1> .. <dependent n>)))
		'''
		
		vDependents = ''.join([self._generateDepKSubtree4(self.depTree.getNode(d), pNode.position) for d in pNode.dependents])
		
		if pNode.isArgument():
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead), 
				                              '_'.join(pNode.getArgRoles()), 
				                              pNode.form)
			else:
				vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead), 
				                                   '_'.join(pNode.getArgRoles()), 
				                                   pNode.form, 
				                                   vDependents)
		else:
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s %s)" % (pNode.getDepRel(pCurrentHead), 
				                         pNode.form)
			else:
				vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead), 
				                              pNode.form, 
				                              vDependents)
		
		return vKSubtree
		
	
	
	# (rel (roles )(form ())) -> old format 5
	def _generateDepKSubtree5(self, pNode, pCurrentHead):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (rel (roles )(form ())) format
		
		pCurrentHead identifies which head is calling this method in 
		case the node has multiple heads.
		
		In this format, the annotation components for non-argument and
		predicate nodes are the same as (rel form) format. For argument 
		nodes, argument label is a node as a sibling of the token form:
		leaf node    : (<dependency relation> (<argument role labels> )(<token form> ))
		internal node: (<dependency relation> (<argument role labels> )(<token form> <dependent 1> .. <dependent n>)))
		'''
		
		vDependents = ''.join([self._generateDepKSubtree5(self.depTree.getNode(d), pNode.position) for d in pNode.dependents])
		
		if pNode.isArgument():
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s (%s )(%s ))" % (pNode.getDepRel(pCurrentHead), 
				                              '_'.join(pNode.getArgRoles()), 
				                              pNode.form)
			else:
				vKSubtree = "(%s (%s )(%s %s))" % (pNode.getDepRel(pCurrentHead), 
				                                   '_'.join(pNode.getArgRoles()), 
				                                   pNode.form, 
				                                   vDependents)
		else:
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s %s)" % (pNode.getDepRel(pCurrentHead), 
				                         pNode.form)
			else:
				vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead), 
				                              pNode.form, 
				                              vDependents)
		
		return vKSubtree
		
	
	
	# (rel (pos form))
	def _generateDepKSubtree6(self, pNode, pCurrentHead):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (rel (pos form)) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format:
		leaf node    : (<dependency relation> (<POS tag> <token form>))
		internal node: (<dependency relation> (<POS tag> (<token form> <dependent 1> .. <dependent n>)))
		'''
		
		vDependents = ''.join([self._generateDepKSubtree6(self.depTree.getNode(d), pNode.position) for d in pNode.dependents])
		
		if len(pNode.dependents) == 0:
			vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
			                              pNode.getPOSTag(),
			                              pNode.form)
		else:
			vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
			                                   pNode.getPOSTag(),
			                                   pNode.form,
			                                   vDependents)
		
		return vKSubtree
		
	
	
	# (rel (pos )(form ()))
	def _generateDepKSubtree7(self, pNode, pCurrentHead):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (rel (pos )(form ())) format
		
		pCurrentHead identifies which head is calling this method in 
		case the node has multiple heads.
		
		In this format, pos tag is a sibling node to token form under the
		dependency relation node:
		leaf node    : (<dependency relation> (<POS> )(<token form> ))
		internal node: (<dependency relation> (<POS> )(<token form> <dependent 1> .. <dependent n>)))
		'''
		
		vDependents = ''.join([self._generateDepKSubtree7(self.depTree.getNode(d), pNode.position) for d in pNode.dependents])
		
		if len(pNode.dependents) == 0:
			vKSubtree = "(%s (%s )(%s ))" % (pNode.getDepRel(pCurrentHead),
			                                 pNode.getPOSTag(),
			                                 pNode.form)
		else:
			vKSubtree = "(%s (%s )(%s %s))" % (pNode.getDepRel(pCurrentHead),
			                                   pNode.getPOSTag(),
			                                   pNode.form,
			                                   vDependents)
		
		return vKSubtree
		
	
	
	# (form (rel head))
	def _generateDepKSubtree8(self, pNode, pdOptions = {}):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (form (rel head)) format
		
		The outcome is a bottom-up subtree.
		
		In this format:
		leaf node    : (<token form> <dependency relation>)
		internal node: (<token form> (<dependency relation 1> <head 1>) .. (<dependency relation 1> <head 1>))
		
		pdOptions is a dictionary of the following options:
		- "include root": True for including the "root" for the root node
		                  (default is False)
		- "include subroot": True for including the word form of the node
		                     whose head is the root node (default is True)
		'''
		
		
		if pNode.getHeadNode().isRoot():
			if "include root" in pdOptions and pdOptions["include root"] is True:
					vKSubtree = "(%s root)" % pNode.form
			else:
				if "include subroot" in pdOptions and pdOptions["include subroot"] is False:
					vKSubtree = ''
				else:
					vKSubtree = pNode.form
		else:
			vlHeadsSubTrees = ["(%s %s)" % (r, self._generateDepKSubtree8(self.depTree.getNode(h), pdOptions = pdOptions)) for h, r in pNode.getHeadDeps()]
			vKSubtree = "(%s %s)" % (pNode.form,
			                         ''.join(vlHeadsSubTrees))
		
		return vKSubtree
		
	
	


class ConstKTree:
	'''
	The class for constituency tree for use in tree kernels.
	
	For use in tree kernels, the tree is represented in bracketing format.
	'''
	
	
	def __init__(self, pConstTree):
		'''
		Creates a constituency kernel tree object
		'''
		
		self.constTree = pConstTree
		self.kTree = ""
		
	
	
	def generateConstKTree(self, pFormat = "(phrase (pos form))", pPredNode = None, pflgPredFrameset = False):
		'''
		Generates constituency kernel tree in the required format
		
		The default format 0 is the pure constituency tree without extra
		layer of annotation unless it is already attached in the original
		format.
		
		Some formats include only the annotation specific to a given 
		proposition as argument in pProposition.
		'''
		
		if pFormat == "(phrase (pos form))":
			self.kTree = "( %s)" % self._generateConstKSubtree0(self.constTree.root)
		elif pFormat == "(phrase_args (pos_args form))" and pPredNode is None and pflgPredFrameset == False:
			self.kTree = "( %s)" % self._generateConstKSubtree1(self.constTree.root)
		elif pFormat == "(phrase_args (pos_args form))" and pPredNode is None and pflgPredFrameset == True:
			self.kTree = "( %s)" % self._generateConstKSubtree2(self.constTree.root)
		elif pFormat == "(phrase_args (pos_args form))" and pPredNode is not None:
			self.kTree = "( %s)" % self._generateConstKSubtree3(self.constTree.root, pPredNode)
		
		return self.kTree
		
	
	
	# 0
	def _generateConstKSubtree0(self, pNode):
		'''
		Returns the kernel subtree of the given node in bracketing 
		representation in the basic (phrase (pos form)) format.
		
		This format is the pure constituency tree without extra layer
		of annotation unless it is already attached in the original
		format.
		'''
		
		return pNode.getPTBFormat()
		
	
	
	# 1
	def _generateConstKSubtree1(self, pNode):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (phrase_args (pos_args form)) format
		
		In this format argument role label is concatenated to syntactic
		label.
		'''
		
		if pNode.isPreTerminal() or pNode.isTerminal():
			vKSubtree = "(%s %s)" % ('_'.join([pNode.getSynTag()] + pNode.getArgRoles()),
			                         pNode.getTerminal())
		else:
			vKSubtree = "(%s %s)" % ('_'.join([pNode.getSynTag()] + pNode.getArgRoles()),
			                         ''.join([self._generateConstKSubtree1(n) for n in pNode.children]))
		
		return vKSubtree
		
	
	
	# 2
	def _generateConstKSubtree2(self, pNode):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in format 2
		
		This format is similar to format 1 but the predicate word form
		is replaced with predicate frameset.
		'''
		
		if pNode.isArgument():
			if pNode.isPreTerminal() or pNode.isTerminal():
				vKSubtree = "(%s %s)" % ('_'.join([pNode.getSynTag()] + pNode.getArgRoles()),
				                         pNode.getTerminal())
			else:
				vKSubtree = "(%s %s)" % ('_'.join([pNode.getSynTag()] + pNode.getArgRoles()),
				                         ''.join([self._generateConstKSubtree2(n) for n in pNode.children]))
		elif pNode.isPredicate():
			if pNode.isPreTerminal() or pNode.isTerminal():
				vKSubtree = "(%s %s)" % (pNode.getSynTag(), pNode.getPredicateFrameset())
			else:
				vKSubtree = "(%s %s)" % (pNode.getSynTag(), ''.join([self._generateConstKSubtree2(n) for n in pNode.children]))
		else:
			if pNode.isPreTerminal() or pNode.isTerminal():
				vKSubtree = "(%s %s)" % (pNode.getSynTag(), pNode.getTerminal())
			else:
				vKSubtree = "(%s %s)" % (pNode.getSynTag(), ''.join([self._generateConstKSubtree2(n) for n in pNode.children]))
		
		return vKSubtree
		
	
	
	# 3
	def _generateConstKSubtree3(self, pNode, pPredNode):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in format 3
		
		This format is similar to format 1 but only the given proposition
		(via predicate node) is annotated.
		'''
		
		if pNode.isPreTerminal() or pNode.isTerminal():
			vKSubtree = "(%s %s)" % ('_'.join([pNode.getSynTag()] + pNode.getArgRoleOf(pPredNode)),
			                         pNode.getTerminal())
		else:
			vKSubtree = "(%s %s)" % ('_'.join([pNode.getSynTag()] + pNode.getArgRoleOf(pPredNode)),
			                         ''.join([self._generateConstKSubtree3(n, pPredNode) for n in pNode.children]))
		
		return vKSubtree
		
	
	


class ConstDepKTree:
	'''
	Class for integrating dependency subtrees in constituency trees for 
	tree kernel use
	'''
	
	
	def __init__(self, pConstTree, pDepTree):
		'''
		Constructor
		'''
		
		self.constTree = pConstTree 
		self.depTree = pDepTree
		self.kTree = ""
		
		self.depKTree = self._createDepKTree()
		
	
	
	def _createDepKTree(self):
		'''
		Creates and returns a new DepKTree
		'''
		
		return DepKTree(pDepTree = sef.depTree)
		
	
	
	def generateConstDepKTree(self, pFormat = "(phrase (pos (form (rel head))))", pdOptions = {}):
		'''
		Generates the tree representation in the required format
		
		Basically, the integration of constituency and dependency trees
		is done by extracting for each word a dependency subtree from the
		dependency tree and attaching it under the terminal node of the
		same word in the constituency tree. In the most basic and default
		format, the constituency tree is in its default plain format and
		the dependency structure contains the word and its dependency 
		relation to its head.
		
		pdOptions contains format options for specific formats.
		'''
		 
		if pFormat == "(phrase (pos (form (rel head))))":
			self.kTree = "( %s)" % self._generateCDKSubtree0(self.constTree.root, pdOptions = pdOptions)
		
		return self.kTree
		
	
	
	def _generateCDKSubtree0(self, pNode, pdOptions = {}):
		'''
		Recursively generates the kernel subtree of the given node integrated 
		with the dependency subtrees in (phrase (pos (form (rel head)))) 
		format
		
		This is the basic format where each terminal node is replaces by 
		the dependency subtree of the corresponding word which contains
		the dependency relation label and the head word form up to the root
		node.
		
		See the relevant format function in DepKTree for pdOptions.
		'''
		
		if pNode.isPreTerminal():
			vKSubtree = "(%s %s)" % (pNode.getSynTag(),
			                         self.depKTree.generateDepKTree(pNode = self.depTree.getNode(pNode.getTokenSpan()[0]), pFormat = "(form (rel head))", pdOptions = pdOptions))
		else:
			vKSubtree = "(%s %s)" % (pNode.getSynTag(),
			                         ''.join([self._generateCDKSubtree0(n, pdOptions = pdOptions) for n in pNode.children]))
		
		return vKSubtree
		
	
	


class NGramKTree:
	'''
	Class for representing surface form of a text as a tree
	
	The basic idea is to construct a tree from text tokens.
	'''
	
	
	def __init__(self, plTokens):
		'''
		Constructor
		'''
		
		self.tokens = []
		for i in range(len(plTokens)):
			# loading tokens while masking brackets
			self.tokens.append(plTokens[i].replace('(', "-LRB-").replace(')', "-RRB"))
		
		self.kTree = ''
		
	
	
	def generateNGramKTree(self, pFormat = "unary"):
		'''
		Generates and returns a tree representation of the tokens
		'''
		
		if pFormat == "unary":
			#self.kTree = "( %s)" % self._generateNGramKTree0(self.tokens)
			self.kTree = "( %s)" % self._generateNGramKTree0(["<s>"] + self.tokens + ['X'])
		elif pFormat == "bigram":
			self.kTree = "( %s)" % self._generateNGramKTree1(self.tokens)
		elif pFormat == "binary":
			self.kTree = "( %s)" % self._generateNGramKTree2(["<s>"] + self.tokens)
		
		return self.kTree
		
	
	
	def _generateNGramKTree0(self, plTokens):
		'''
		Recursively generates the tree in the most basic format in which every word is a child of its previous word
		'''
		
		if len(plTokens) == 1:
			return "(%s )" % plTokens[0]
		elif len(plTokens) == 2:
			return "(%s %s)" % (plTokens[0], plTokens[1])
		else:
			return "(%s %s)" % (plTokens[0], self._generateNGramKTree0(plTokens[1:]))
		
	
	
	def _generateNGramKTree1(self, plTokens):
		'''
		Generates the tree in the bigram format
		
		The bigram format is a format designed to capture both unigrams and bigrams.
		
		Example: I go to school.
		         (root (I (I go)) (go (go to)) (to (to school)) (school (school .)) )
		'''
		
		if len(plTokens) == 1:
			return "(%s %s)" % (plTokens[0], plTokens[0])
		else:
			vTree = ''
			
			for i in range(len(plTokens) - 1):
				vTree += "(%s (%s %s))" % (plTokens[i], plTokens[i], plTokens[i+1])
			
			return vTree
		
	
	
	def _generateNGramKTree2(self, plTokens):
		'''
		Recursively generates the tree in the binary format
		
		Example: I go to school.
		         ( (I (go X)(go (to X)(to (school X)(school (. X)(. ))))) )
		'''
		
		if len(plTokens) == 1:
			return "(%s X)" % plTokens[0]
		elif len(plTokens) == 2:
			return "(%s (%s X)(%s ))" % (plTokens[0], plTokens[1], plTokens[1])
		else:
			return "(%s (%s X)%s)" % (plTokens[0], plTokens[1], self._generateNGramKTree2(plTokens[1:]))
		
	
	
