"""	
	This module defines classes for word2vec training.
	
	
	Version 0.1																				(27-Sep-2016 to 10-Oct-2016)
	- W2V is added.
	
"""

import tensorflow as tf
import random, math
from we import *


class W2V(WordEmbedding):
	'''
	Class for building and using word2vec embeddings
	'''
	
	def __init__(self, pVerbosity=1):
		'''
		Constructor
		'''
		
		# (some of) inherited attributes to see here
		
		# self.embeddings = None                            # numpy array containing final word embeddings
		# self.corpus = None                                # original corpus as list of sentences
		# self.wordCounts = None                            # vocab words and their counts
		# self.wordIDs = None                               # vocab words and their IDs (indexes) in frequency order
		# self.IDWords = None                               # vocab IDs and their words
		# self.data = None                                  # the data: list of word IDs (not word forms) in their 
		# original order in the corpus
		
		WordEmbedding.__init__(self)
		
		
		# data attributes
		
		self.validationSet = None  # data set for monitoring the training process
		
		# parameters of the trained embeddings
		
		self.method = None  # cbow (CBOW) or sg (Skip-gram)
		self.batchSize = None  # training batch size
		self.negSampleSize = None  # number of [negative] random examples to sample for
		# sampled_softmax_loss
		self.epochNum = None  # number of training epoches
		self.contextWindowSize = None  # number of words in left/right window of the current word
		self.contextNum = None  # number of words to be used from context windows
		
		# private variables
		
		self._currWordIdx = 0  # keeps track of the index of currently processed word
		
	
	
	def train(self, pEmbedDim=50, pMethod="cbow", pDevice="/cpu:0",
			  pBatchSize=128, pContectWindowSize=1, pContextNum=2, pNegSampleSize=64, pEpochNum=100000,
			  pflgValidate=False, pValidationSetSize=16):
		'''
		Trains the embedding using specified arguments
		'''
		
		if self.data is None:
			raise Exception("No data loaded!")
		
		# setting method
		if pMethod.lower() == "cbow":
			self.method = "cbow"
		elif pMethod.lower().startswith("skip") or pMethod.lower() == "sg":
			self.method = "sg"
		else:
			raise Exception("Unknown method: %s" % pMethod)
		
		# setting batch size
		self.batchSize = pBatchSize
		
		# setting context parameters
		self.contextWindowSize = pContectWindowSize
		self.contextNum = pContextNum
		
		# setting sample size for negative random sampling
		self.negSampleSize = pNegSampleSize
		
		# setting number of training epoch 
		self.epochNum = pEpochNum
		
		# preparing validation data
		if pflgValidate:
			# random sampling validation set from top 100 word IDs
			self.validationSet = np.array(random.sample(range(100), pValidationSetSize))
		
		# creating the computation graph
		tgGraph = self._createGraph(pEmbedDim=pEmbedDim, pDevice=pDevice, pflgValidate=pflgValidate)
		
		self._runTrain(tgGraph, pflgValidate=pflgValidate)
		
	
	
	def _createGraph(self, pEmbedDim, pDevice, pflgValidate=False):
		'''
		Creates the computation graph for trianing embeddings
		'''
		
		tgGraph = tf.Graph()
		
		with tgGraph.as_default(), tf.device(pDevice):
			# batch input data
			self._tphBatchData = tf.placeholder(tf.int32, shape=[self.batchSize])
			self._tphBatchLabels = tf.placeholder(tf.int32, shape=[self.batchSize, 1])
			
			# variables: embeddings and softmax weights
			tvEmbeddings = tf.Variable(tf.random_uniform([self.vocabSize, pEmbedDim], -1.0, 1.0))
			tvSoftmaxWeights = tf.Variable(tf.truncated_normal([self.vocabSize, pEmbedDim],
															   stddev=1.0 / math.sqrt(pEmbedDim)))
			tvSoftmaxBiases = tf.Variable(tf.zeros([self.vocabSize]))
			
			# looking up embeddings of the batch
			tvBatchEmbeddings = tf.nn.embedding_lookup(tvEmbeddings, self._tphBatchData)
			print self.negSampleSize, self.vocabSize
			# calculating the average loss for the sample (average of losses of individual examples)
			self._tBatchLoss = tf.reduce_mean(tf.nn.sampled_softmax_loss(weights=tvSoftmaxWeights,
																		 biases=tvSoftmaxBiases,
																		 inputs=tvBatchEmbeddings,
																		 labels=self._tphBatchLabels,
																		 num_sampled=self.negSampleSize,
																		 num_classes=self.vocabSize))
			
			# optimizing embeddings and softmax weights
			self._toOptimizer = tf.train.AdagradOptimizer(0.1).minimize(self._tBatchLoss)
			
			# normalizing the embeddings
			vNorm = tf.sqrt(tf.reduce_sum(tf.square(tvEmbeddings), 1, keep_dims=True))
			self._tNormalEmbeddings = tvEmbeddings / vNorm
			self.normalized = True
			
			# calculating validation similarities
			if pflgValidate:
				tValidationEmbeddings = tf.nn.embedding_lookup(self._tNormalEmbeddings, self.validationSet)
				self._tValidationSimilarities = tf.matmul(tValidationEmbeddings, tf.transpose(self._tNormalEmbeddings))
		
		return tgGraph
		
	
	
	def _runTrain(self, ptgGraph, pflgValidate=False):
		'''
		Runs the training graph
		'''
		
		with tf.Session(graph=ptgGraph) as session:
			# initializing variables
			tf.initialize_all_variables().run()
			if self.verbosity > 0:
				print("Variables initialized.\n")
			
			if self.verbosity == 1:
				print("Training...")
			
			self._currWordIdx = 0
			vEpochLoss = 0
			for epoch in range(1, self.epochNum + 1):
				# training ----------------------
				
				# generating batvh
				vaBatchData, vaBatchLabels = self._generateBatch()
				
				# preparing feeding dictionary to operations
				vdFeed = {self._tphBatchData: vaBatchData, self._tphBatchLabels: vaBatchLabels}
				
				# running the optimizer and loss function (loss value is also needed)
				_, vBatchLoss = session.run([self._toOptimizer, self._tBatchLoss], feed_dict=vdFeed)
				
				# reflecting average epoch losses
				if self.verbosity > 1 or pflgValidate:
					# adding up to epotch loss
					vEpochLoss += vBatchLoss
					
					if epoch == 1:
						print("Average loss at epoch 1: %s" % vEpochLoss)
					# vEpochLoss = 0        # don't do this because in that case there will be 1999 losses summed at epoch 2000
					elif epoch % 2000 == 0:
						print("Average loss at epoch %s: %s" % (epoch, vEpochLoss / 2000))
						vEpochLoss = 0
				
				# validation ----------------------
				
				if pflgValidate:
					if epoch == 1 or epoch % 10000 == 0:
						vaValidSims = self._tValidationSimilarities.eval()
						
						for i in range(len(self.validationSet)):
							vValidWord = self.IDWords[self.validationSet[i]]
							
							# sorting the similar words (only showing top 8; omitting word itself at [0])
							vlNearstIDs = (-vaValidSims[i, :]).argsort()[1:9]
							
							vLog = "\tNearst to %s: " % vValidWord
							for id in vlNearstIDs:
								vLog += "%s, " % self.IDWords[id]
							
							print(vLog)
			
			if self.verbosity == 1 and not pflgValidate:
				print("Average loss at last epoch: %s" % vBatchLoss)
			
			self.embeddings = self._tNormalEmbeddings.eval()
			
			if self.verbosity > 0:
				print("\nWord embeddings saved.")
		
	
	
	def _generateBatch(self):
		'''
		Generates and returns batch data
		'''
		
		if self.method == "sg":
			return self._generateSGBatch(pBatchSize=self.batchSize,
										 pSkipWindowSize=self.contextWindowSize,
										 pSkipNum=self.contextNum)
		elif self.method == "cbow":
			return self._generateCBOWBatch(pBatchSize=self.batchSize,
										   pContextWindowSize=self.contextWindowSize,
										   pContextNum=self.contextNum)
		
	
	
	def _generateSGBatch(self, pBatchSize, pSkipWindowSize, pSkipNum):
		'''
		Generates and return batch data for skip-gram embedding training
		
		pSkipWidowSize: number of context words in left/right
		pSkipNum      : number of times a word is reused as input to generate a sample, i.e. the number of words from 
						the context of the input word to be used as label
						This should be at most twice as the window size, i.e. once per context word (label). Therefore,
						batch size should be divisable to this number.
		'''
		
		# a few checks
		assert pBatchSize % pSkipNum == 0
		assert pSkipNum <= 2 * pSkipWindowSize
		
		# If this is the first batch, the current word index is pushed forward so that enough context words are left
		# in the left side of the first input word, i.e. as many as skip window size. That is, the first input word 
		# is not the first word in the data but the (pSkipWindowSize + 1)_th.
		if self._currWordIdx == 0:
			self._currWordIdx = self._currWordIdx
		
		vaBatchData = np.ndarray(shape=(pBatchSize), dtype=np.int32)  # batch data (input word IDs)
		vaBatchLabels = np.ndarray(shape=(pBatchSize, 1), dtype=np.int32)  # batch labels (context word IDs)
		
		# iterating as many as the number of required input words to generate a batch (each input word is used 
		# pSkipNum times)
		for i in range(pBatchSize // pSkipNum):
			vlContextPool = []  # a pool of eligible context word for the 
			# current input word
			# pooling left context
			for j in range(1, pSkipWindowSize + 1):
				vlContextPool.append(self.data[self._currWordIdx - j])
			# pooling right context
			for j in range(1, pSkipWindowSize + 1):
				vlContextPool.append(self.data[self._currWordIdx + j])
			
			# shuffling the pool
			random.shuffle(vlContextPool)
			
			# creating the data instances with the current input word: picking context words (labels) as many as pSkipNum;
			# Context words are selected randomly (see shuffling above), but randomness is only meaningful when 
			# pSkipNum < 2 * pSkipWindowSize. Otherwise, all context words will be used.
			for j in range(pSkipNum):
				vaBatchData[i * pSkipNum + j] = self.data[self._currWordIdx]
				vaBatchLabels[i * pSkipNum + j, 0] = vlContextPool[j]
			
			self._currWordIdx = (self._currWordIdx + 1) % len(self.data)  # It's possible that the number of training 
			                                                              # steps is too big and data is finished and has
			                                                              # to start over, so we get the modulo.
		
		return vaBatchData, vaBatchLabels
		
	
	
	def _generateCBOWBatch(self, pBatchSize, pContextWindowSize, pContextNum):
		'''
		Generates and return batch data for CBOW embedding training
		
		pContextWidowSize: number of context words in left/right
		pContextNum      : number of times a word is reused as target to generate a sample, i.e. the number of words
						   from the context of the target word to be used as input
						   This should be at most twice as the window size, i.e. once per context word (input). 
						   Therefore, batch size should be divisable to this number.
		'''
		
		# a few checks
		assert pBatchSize % pContextNum == 0
		assert pContextNum <= 2 * pContextWindowSize
		
		# If this is the first batch, the current word index is pushed forward so that enough context words are left
		# in the left side of the first target word, i.e. as many as context window size. That is, the first target
		# word is not the first word in the data but the (pContextWindowSize + 1)_th.
		if self._currWordIdx == 0:
			self._currWordIdx = self._currWordIdx
		
		vaBatchData = np.ndarray(shape=(pBatchSize), dtype=np.int32)  # batch data (context word IDs)
		vaBatchLabels = np.ndarray(shape=(pBatchSize, 1), dtype=np.int32)  # batch labels (target word IDs)
		
		# iterating as many as the number of required input words to generate a batch (each target word is used 
		# pContextNum times)
		for i in range(pBatchSize // pContextNum):
			vlContextPool = []  # a pool of eligible context word 
			# for the current target word
			# pooling left context
			for j in range(1, pContextWindowSize + 1):
				vlContextPool.append(self.data[self._currWordIdx - j])
			# pooling right context
			for j in range(1, pContextWindowSize + 1):
				vlContextPool.append(self.data[self._currWordIdx + j])
			
			# shuffling the pool
			random.shuffle(vlContextPool)
			
			# creating the data instances with the current target word: picking context words (inputs) as many as 
			# pContextNum; Context words are selected randomly (see shuffling above), but randomness is only 
			# meaningful when pContextNum < 2 * pContextWindowSize. Otherwise, all context words will be used.
			for j in range(pContextNum):
				vaBatchData[i * pContextNum + j] = vlContextPool[j]
				vaBatchLabels[i * pContextNum + j, 0] = self.data[self._currWordIdx]
			
			self._currWordIdx = (self._currWordIdx + 1) % len(self.data)  # It's possible that the number of training 
		# steps is to high and data is finished and 
		# has to start over, so we get the 
		# modulo.
		
		return vaBatchData, vaBatchLabels
		
	
	