#! /usr/bin/python

"""	
	This script combines the n-best output of multiple MT systems and 
	reranks it based on various measures.
	
	- Union-based reranking: the union of n-best lists of multiple systems
	(distinct) is reranked based on new scores for each item which is the
	summation or average of the decoder scores assigned by each system
	to it. When a translation item does not exists in the n-best of a
	system, 0 or the minimum decoder score in the n-best list of that 
	system for the corresponding input sentence is used.
	
	- Agreement-based reranking: the intersection of n-best translations
	of multiple systems for each sentence is extracted and reranked 
	according to various methodes applied on decoder scores.
	
	We define agreement (intersection) degree here as the number of systems
	having an non-empty intersection of n-best translations for a specific
	sentence. Therefore, a degree-5 intersection is a full intersection 
	for a 5-system n-best, but an intersection of degree 4 is not full in
	this case, though it is an intersection.
	
	The reason for using non-full intersections is that most of the 
	sentences do not share any n-best item among all systems. So, this
	method is used as a back off strategy.
	
	- External-score-based reranking: combined n-best list are scored 
	using an external metric and reranked. The purpose is two fold:
	First, scores like TER at n-best item level can be used to analyze the 
	effectiveness of agreement-based combination reranking. Second, estimated
	confidence scores can be used for reranking the combined n-best.
	
	Optionally, it escapes reranking and outputs the merged and scored 
	n-bests in the original order. This is useful for diagnostic and analytic
	purposes. Of cource the selected translation are not considered to be
	useful. Only statistics outputs are useful.
	
	Note that, regardless of what measure is used for reranking, all of 
	them can optionally be computed for each combined item for analysis 
	purpose.
	
	The input is the n-best output paths of each system together with 
	other required data such as reference tranlations.
	
	The output is the best translation set together with some statistics
	and optionally the whole reranked set for each sentence.
	
	
	Version: 1.8											(22-Aug-2012)
	- Sort direction for reranking is now optional. Choosing an ascending
	direction (which is not the default) helps compute the lower bound
	combination --ce-score-sort-dir is dropped and merged by this new more
	general option. NOTE: although TER scores should be sorted in ascending
	order to compute the oracle (actual TERs not estimated ones), it is
	still handled internally by multiplying the scores by -1 and sorting
	them in descending order. Therefor, ascending should not be provided 
	as option when the purpose of reranking is to compute TER-based oracle.
	
	Version: 1.7											(16-Aug-2012)
	- Optionally, the merged set can be output in the original order without
	reranking (only for analytic purposes).
	
	Version: 1.6 											(14-Aug-2012)
	- Estimated confidence scores is implemented as external score. Scores
	are extracted from weka prediction output file for merged n-best.
	
	Version: 1.5 											(12-Aug-2012)
	- Minimum decoder rank is implemented.
	
	Version: 1.4
	- Normalization of decoder scores is added.
	
	Version: 1.3
	- Union-based reranking is added
	- All scores can be computed for each item
	- N-best input changed from per-sentence to per-system
	
	Version: 1.2
	- Reranking is extended to external scores
	- TER is implemented as external score
	- Statisctics and selected translations are redirected to files
	
	Version: 1.1
	- Counting and reporting empty intersections
	
"""  

from collections import namedtuple
from operator import itemgetter
from decimal import *
import sys, optparse, os, mtnbest, mteval, random, weka


def scoreTER(plMergedNBest, pRefSent, pScorerClass):
	'''
	Scores translations in plMergedNBest against plRefSent by calling
	scoreTER from mteval module and extracting the scores from generated 
	files.
	'''
	
	vlTransSents = [vItem.trans for vItem in plMergedNBest]
	vlRefSents = [pRefSent]*len(vlTransSents)
	
	# openning devnull to redirect the log output of TER scorer
	vDevnull = open(os.devnull, 'w')
	
	## generating a random name for temporary TER score file to avoid 
	## conflict between mutiple runs of the scripts
	vTERFilename = "./" + ''.join(random.choice("0123456789") for x in range(5))
	
	vlScores = mteval.scoreTER(vlTransSents, vlRefSents, vTERFilename, pScorerClass, "ter", vDevnull)
	
	## iterating merged n-best items to extract score for each item and
	## constructing the new merged n-best to contain the scores
	vlScoredNBest = []
	for vItem, vScore in zip(plMergedNBest, vlScores):
		# TER scores are multiplied by -1 to be sorted in descending order
		vItem.extscores["ter"] = -vScore
		vlScoredNBest.append(mtnbest.CombinedNBestItem(vItem.trans, vItem.memidxs, vItem.degree, vItem.sysscores, vItem.combscores, vItem.extscores))
	
	os.remove(vTERFilename + ".ter")
	
	return vlScoredNBest
	

def scoreCE(plMergedNBest, plCEScores, pStartIdx, pEndIdx):
	'''
	Assigns estimated confidence scores from plCEScores to sentence n-best
	translations in plMergedNBest. 
	
	pStartIdx: starting index of the CE scores related to the merged n-best
			   of current sentence
			   Note, plCEScores is supposed to be corresponding to the
			   merged n-best items for all sentences.
	pEndIdx: ending index of the CE scores related to the merged n-best
			 of current sentence
	'''
	
	# constructing the new merged n-best to contain the scores
	vlScoredNBest = []
	for vItem, vScore in zip(plMergedNBest, plCEScores[pStartIdx : pEndIdx + 1]):
		vItem.extscores["ce"] = Decimal(vScore)
		vlScoredNBest.append(mtnbest.CombinedNBestItem(vItem.trans, vItem.memidxs, vItem.degree, vItem.sysscores, vItem.combscores, vItem.extscores))
	
	return vlScoredNBest
	



##======================================================================
## main
def main(argv=None):
	if argv is None:
		argv = sys.argv
 
	# data structure for score range
	ScoreRange = namedtuple("ScoreRange", "min max")
	
	
	parser = optparse.OptionParser(usage="%prog <SYSTEM NAMES> <N-BEST NAMES> <SOURCE SET> <COMBINED TRANSLATION OUTPUT> <STATISTICS OUTPUT FILE> [options]" + 
										 "\nCombines the n-best output of multiple MT systems and reranks it based on various measures.", version="%prog 1.7")

	# common options
	parser.add_option("--rerank-measure", "--measure", help="measure for reranking: [union-dec-score, union-dec-rank, agree-dec-score, agree-dec-rank (the original rank of the item in n-best), ext-score-xxx (xxx is the key for the external score)]", dest="rerankMeasure", action="store")
	parser.add_option("--rerank-sort-dir", help="sort direction for reranking: (a)scending,  (d)escending (NOTE, do not use ascending for TER-based reranking for computing oracle combination; it is handled internally) [default: descending]", dest="rerankSortDir", default="descending", action="store")
	parser.add_option("--out-per-sentence-stat", help="output file name for statistics for each sentence", dest="outPerSentStat", action="store_true")
	parser.add_option("--rerank-details-file", help="output file name for reranked list for each sentence", dest="rerankDetailOut", action="store")
	parser.add_option("--calc-union-dec-score", help="calculate union-based-combined decoder score", dest="calcUnionDecScore", action="store_true")
	parser.add_option("--calc-union-dec-rank", help="calculate union-based-combined decoder ranking position", dest="calcUnionDecRank", action="store_true")
	parser.add_option("--calc-agree-dec-score", help="calculate agreement-based-combined decoder score", dest="calcAgreeDecScore", action="store_true")
	parser.add_option("--calc-agree-dec-rank", help="calculate agreement-based-combined decoder ranking position", dest="calcAgreeDecRank", action="store_true")
	parser.add_option("--calc-ter-score", help="calculate ter score", dest="calcTERScore", action="store_true")
	parser.add_option("--calc-ce-score", help="calculate or extract confidence estimation score", dest="calcCEScore", action="store_true")
	parser.add_option("--normalize-dec-scores", help="normalize decoder scores before combination", dest="normDecScores", action="store_true")
	parser.add_option("--norm-trg-range", help="target range to normalize scores to [default: -100,-1]", dest="normTrgRange", default="-100,-1", action="store")
	# union-based options
	parser.add_option("--union-score-comb-method", help="score combination method for union-based raranking: [sum, avg]", dest="unionCombMethod", default="sum", action="store")
	parser.add_option("--union-min-margin", help="percentage of the minimum decoder score to be subtracted from it and used to replace zero score when n-best item translation does not exist in a system output [0: use the minimum itself; if omitted, minimum won't be used at all]", dest="minForZeroMargin", default=-1, action="store")
	parser.add_option("--union-min-rank", help="minimum decoder rank to replace -1 rank when n-best item translation does not exist in a system output [ignore to use -1 itself; use maximum n-best size to substitute --union-min-margin=0 for decoder score; use maximum n-best size minus an integer to substitue --union-min-margin with non-zero value]", dest="minDecRank", default=-1, action="store")
	# agreement-based options
	parser.add_option("--min-intersect-degree", help="minimum agreement (intersection) degree to back off to", dest="minIntersectDegree", default=0, action="store")
	parser.add_option("--empty-intersect-method", help="method to be used to select translation when agreements (intersections) of all degrees higher than --min-intersect-degree are empty (prefsys, confidence)", dest="emptyIntersectMethod", default="prefsys", action="store")
	parser.add_option("--pref-sys-no", help="if --empty-intersect-method is set to prefsys, this is preferred system's number", dest="prefSysNo", default=1, action="store")
	parser.add_option("--agree-score-comb-method", help="score combination method for agreement-based reranking: [muliply, sum]", dest="agreeCombMethod", default="multiply", action="store")
	# external-score-based options
	parser.add_option("--ter-scorer", help="path to TER scorer (tercom) java class", dest="terScorer", action="store")
	parser.add_option("--ref-set", help="reference tranalations (for sentence-level scoring)", dest="refSet", action="store")
	parser.add_option("--ce-weka-out", help="path to estimated confidence scores in weka output", dest="ceWekaOut", action="store")
 
	(opts, posArgs) = parser.parse_args()
 
	if len(posArgs) < 5:
		parser.error("At least 5 argument is required (comma-separated list of system names, comma-separated list of system n-best absolute files, source absolute file, file name for combined translation output, file name for statistics")
	
	# processing arguments
	 
	vlSysNames = posArgs[0].split(',')
	vSysNum = len(vlSysNames)
	vlNBestFilenames = posArgs[1].split(',')
	vSourceFilename = posArgs[2]
	vTransOutFilename = posArgs[3]
	vStatOutFilename = posArgs[4]
	
	# processing options
	
	# no back off intersection degree by default
	if opts.minIntersectDegree < 2:
		vMinIntersectDegree = vSysNum
	else:
		vMinIntersectDegree = int(opts.minIntersectDegree)
	
	if opts.rerankMeasure == "union-dec-score":
		opts.calcUnionDecScore = True
	if opts.rerankMeasure == "union-dec-rank":
		opts.calcUnionDecRank = True
	if opts.rerankMeasure == "agree-dec-score":
		opts.calcAgreeDecScore = True
	if opts.rerankMeasure == "agree-dec-rank":
		opts.calcAgreeDecRank = True
	if opts.rerankMeasure == "ext-score-ter":
		opts.calcTERScore = True
	if opts.rerankMeasure == "ext-score-ce":
		opts.calcCEScore = True
	
	vPrefSysNo = int(opts.prefSysNo)
	
	vNormTrgRange = ScoreRange(0, 0)
	if opts.normDecScores:
		vlNormTrgRange = [Decimal(num) for num in opts.normTrgRange.split(',')]
		vNormTrgRange = ScoreRange(vlNormTrgRange[0], vlNormTrgRange[1])
	
	vMinForZeroMargin = Decimal(opts.minForZeroMargin)
	vMinDecRank = int(opts.minDecRank)
	
	# openning source set
	
	try:
		vfSourceSet = open(vSourceFilename)
	except IOError:
		sys.exit("Cannot open source data set file: " + vSourceFilename)
	
	
	# opening and loading reference file if given
	
	vlRefSents = []
		
	if opts.refSet != None:
		try:
			vfRefSet = open(opts.refSet)
		except IOError:
			sys.exit("Cannot open reference file: " + opts.refSet)
		
		for vLine in vfRefSet:
			vlRefSents.append(vLine.strip(' \n'))
		
		vfRefSet.close()
	
	# opening and loading estimated scores if given
	## Note: the score file is supposed to be correspondent to the merged
	## n-best lists of all sentences. Therefor, that merged n-best might 
	## be previously generated by the same method here (or exacltly with
	## this script), then scored using external any external tools.
	
	if opts.calcCEScore != None:
		if opts.ceWekaOut == None:
			sys.exit("No Weka output file is given!")
		else:
			print "Loading estimated confidence scores ..."
			vlCEScores = weka.extractPredictions(opts.ceWekaOut, False)
	
	# creating output file for selected translations
	
	try:
		vfTransOut = open(vTransOutFilename, 'w')
	except IOError:
		sys.exit("Cannot create translation output file: " + vTransOutFilename)
	
	
	# creating output file for reranking details
	
	if opts.rerankDetailOut != None:
		try:
			vfRerankDetailOut = open(opts.rerankDetailOut, 'w')
		except IOError:
			sys.exit("Cannot create reranking details output file: " + opts.rerankDetailOut)
	
	
	# creating output file for statistics
	
	try:
		vfStatOut = open(vStatOutFilename, 'w')
	except IOError:
		sys.exit("Cannot create statistics output file: " + vStatOutFilename)
	
	
	## extracting the intersection information of systems' n-best items 
	## for each sentence
	
	## list of intersection sizes for various intersection degrees for 
	## each sentence
	# see the script description for definition of intersection degree
	vlIntersectSizes = []
	
	## list of number of intersections of various degrees for the whole
	## data set. Note that degree one, which represents empty intersection
	## is also included (at index 0).
	vlDegreeInterCounts = [0]*vSysNum
	
	## counts of sentences treated with each selection method:
	## 0: highest degree intersection
	## 1: back off intersection
	## 2: back off no-intersection method
	vlSelMethodCounts = [0, 0, 0]
	
	# list of number of top-ranked translations from each agreement degree
	vlTopRankesDegree = [0]*vSysNum
	
	# initializing report output
	if opts.rerankDetailOut != None:
		vfRerankDetailOut.write("System order: " + ' '.join(vlSysNames) + '\n')
		
	## loading n-best lists for all systems and finding the minimum and 
	## maximum decoder scores for each system
	vlNBestLists = []
	vlScoreRanges = []
	vSysCntr = 0
	for vSysNBestFileName in vlNBestFilenames:
		print "Loading n-best of " + vlSysNames[vSysCntr] + "..."
		vlNBestLists.append(mtnbest.extractSentenceNBests(vSysNBestFileName.strip(), False))
		vSysCntr += 1
		
		# sum(vlNBestLists[-1], []) is a clever way for flattening the list
		vlScores = [item.score for item in sum(vlNBestLists[-1], [])]
		vlScoreRanges.append(ScoreRange(min(vlScores), max(vlScores)))
		
	# main loop
	
	vSentenceNo = 0
	# counts the number of merged n-best items up to the current sentence
	vMergedNBestItemCntr = 0
	for vSrcSent in vfSourceSet:
		print "Processing sentence " + str(vSentenceNo + 1) + " ..."
		
		## merging n-bests of systems for the sentence into a distinct 
		## set of translation (union of n-bests)
		vlMergedNBest = []
		vlMergedNBest = mtnbest.mergeNBests([NBest[vSentenceNo] for NBest in vlNBestLists])
		vMergedNBestLen = len(vlMergedNBest)
		vMergedNBestItemCntr += vMergedNBestLen
		
		# calculating requested scores for merged n-best items
		
		vlScoredMergedNBest = vlMergedNBest
		
		if opts.calcUnionDecScore:
			vlScoredMergedNBest = mtnbest.calcUnionCombScores(vlScoredMergedNBest, "decoder-score", opts.unionCombMethod, vMinForZeroMargin, vMinDecRank, opts.normDecScores, vlScoreRanges, vNormTrgRange)
		
		if opts.calcUnionDecRank:
			vlScoredMergedNBest = mtnbest.calcUnionCombScores(vlScoredMergedNBest, "decoder-rank", opts.unionCombMethod, vMinForZeroMargin, vMinDecRank, opts.normDecScores, vlScoreRanges, vNormTrgRange)
		
		if opts.calcAgreeDecScore:
			vlScoredMergedNBest, vSelMethod = mtnbest.calcAgreeCombScores(vlScoredMergedNBest, vMinIntersectDegree, opts.emptyIntersectMethod, vPrefSysNo, "decoder-score", opts.agreeCombMethod, opts.normDecScores, vlScoreRanges, vNormTrgRange)
			# adjusting selection method counts
			vlSelMethodCounts[vSelMethod] += 1
		
		if opts.calcAgreeDecRank:
			vlScoredMergedNBest, vSelMethod = mtnbest.calcAgreeCombScores(vlScoredMergedNBest, vMinIntersectDegree, opts.emptyIntersectMethod, vPrefSysNo, "decoder-rank", opts.agreeCombMethod, opts.normDecScores, vlScoreRanges, vNormTrgRange)
			# adjusting selection method counts
			vlSelMethodCounts[vSelMethod] += 1
		
		if opts.calcTERScore != None:
			if opts.refSet == None:
				sys.exit("No reference translation is given!")
			else:
				vlScoredMergedNBest = scoreTER(vlScoredMergedNBest, vlRefSents[vSentenceNo], opts.terScorer)
		
		if opts.calcCEScore != None:
			vlScoredMergedNBest = scoreCE(vlScoredMergedNBest, vlCEScores, vMergedNBestItemCntr - vMergedNBestLen, vMergedNBestItemCntr)
		
		# reranking merged n-best based on requested measure
		vlRerankedNBest = []
		if opts.rerankMeasure != None:
			vlRerankedNBest = mtnbest.rerankMergedNBest(vlScoredMergedNBest, opts.rerankMeasure, opts.rerankSortDir)
		else:
			vlRerankedNBest = vlScoredMergedNBest
		
		# outputing selected translation for current sentence
		vfTransOut.write(vlRerankedNBest[0].trans + '\n')
		
		
		# collecting and outputing statistics
		
		# calculating intersection sizes of various degrees for the sentence
		vlIntersectSizes.append([0]*vSysNum)
		for vDegree in range(2, vSysNum + 1):
			vSize = len([item.degree for item in vlMergedNBest if item.degree == vDegree])
			vlIntersectSizes[-1][vDegree - 1] = vSize
			
			# adding to the number of non-empty intersections of this degree
			## Note that intersection of degree 1 does not make sense
			if vSize > 0:
				vlDegreeInterCounts[vDegree - 1] += 1
		
		## setting the number of sentences with no n-best intersection,
		## at index 0
		if len([item for item in vlIntersectSizes[-1][1:] if item > 0]) == 0:
			vlDegreeInterCounts[0] += 1
		
		# setting the agreement degree of top-ranked item
		vlTopRankesDegree[vlRerankedNBest[-1].degree - 1] += 1
		
		# outputing sentence statistics
		if opts.outPerSentStat:
			vfStatOut.write(str(vSentenceNo + 1).rjust(4) + ': ')
			vfStatOut.write('[' + ' '.join([str(size) for size in vlIntersectSizes[-1]]) + ']\n')
	
		# outputing the reranked list
		if opts.rerankDetailOut != None:
			vfRerankDetailOut.write('\n' + str(vSentenceNo + 1) + ') ' + vSrcSent)
			if vMinForZeroMargin != -1:
				vlMinSysScores = [] 
				for vSysIdx in range(vSysNum):
					vlMinSysScores.append(str(min(vItem.sysscores[vSysIdx] for vItem in vlRerankedNBest).quantize(Decimal("0.0001"))))
				vfRerankDetailOut.write('\nMinimum system scores: [' + ', '.join(vlMinSysScores) + ']')
				vfRerankDetailOut.write('\n')
			vfRerankDetailOut.write('\n')
			vItemCntr = 1
			for vItem in vlRerankedNBest:
				vfRerankDetailOut.write(str(vItemCntr).rjust(4) + '. ')
				vfRerankDetailOut.write("N-best ranks: " + ' '.join([str(item).rjust(3) for item in vItem.memidxs]) + '\t')
				vfRerankDetailOut.write('Agreement degree: ' + str(vItem.degree) + '\t') 
				vfRerankDetailOut.write("System scores: " + ' '.join([str(score.quantize(Decimal("0.0001"))).rjust(8) for score in vItem.sysscores]) + '\t')
				for vCombScoreType, vScore in vItem.combscores.iteritems():
					if type(vScore) is Decimal:
						vfRerankDetailOut.write(vCombScoreType + ': ' + str(vScore.quantize(Decimal("0.0001"))).rjust(9) + '\t')
					else:
						vfRerankDetailOut.write(vCombScoreType + ': ' + str(vScore).rjust(5) + '\t')
				for vExtMetric, vScore in vItem.extscores.iteritems():
					vfRerankDetailOut.write(vExtMetric + ': ' + str(vScore) + '\t')
				vfRerankDetailOut.write(vItem.trans + '\n')
				vItemCntr += 1
		
		vSentenceNo += 1

		
	vSentenceNum = vSentenceNo
	
	# outputing statistics
	
	vfStatOut.write("\nSummary:\n")
	vfStatOut.write("\nNumber of systems: " + str(vSysNum) + " (" + ' '.join(vlSysNames) + ")")
	for vSysIdx in range(vSysNum):
		vfStatOut.write("\nDecoder score range of " + vlSysNames[vSysIdx] + ": [" + str(vlScoreRanges[vSysIdx].min) + ', ' + str(vlScoreRanges[vSysIdx].max) + ']')
	vfStatOut.write("\n")
	vfStatOut.write("\nReranking measure: " + opts.rerankMeasure)
	vfStatOut.write("\n")
	vfStatOut.write("\nMerged n-best size: " + str(vMergedNBestItemCntr))
	
	# agreement-based-specific statistics
	if opts.calcAgreeDecScore or opts.calcAgreeDecRank:
		vfStatOut.write("\n")
		vfStatOut.write("\nAgreement-based score combination statistics:")
		vfStatOut.write("\n")
		vfStatOut.write("\nMinimum back off agreement degree: " + str(vMinIntersectDegree))
		vfStatOut.write("\nAgreement-based score combination method: " + opts.agreeCombMethod)
		if opts.emptyIntersectMethod == "prefsys":
			vfStatOut.write("\nSelection method for empty-intersection sentences: preferred system's output")
			vfStatOut.write("\nPreferred system: " + vlSysNames[vPrefSysNo - 1])
		elif opts.emptyIntersectMethod == "confidence":
			vfStatOut.write("\nBack off selection method for no-agreement sentences: most confident translation among all n-bests")
		vfStatOut.write("\nDataset size: " + str(vSentenceNum))
		vfStatOut.write("\n")
		vfStatOut.write("\nNumber of translations selected from full agreements: " + str(vlSelMethodCounts[0]))
		vfStatOut.write("\nNumber of translations selected from back off agreements: " + str(vlSelMethodCounts[1]))
		vfStatOut.write("\nNumber of translations selected with back off selection method: " + str(vlSelMethodCounts[2]))
	
	# common statistics
	
	# top-ranked n-best items (selected translations) falling in each agreement degree
	vfStatOut.write("\n")
	vfStatOut.write("\nRanking Statistics:")
	vfStatOut.write("\n")
	for vDegree in range (vSysNum, 0, -1):
		vfStatOut.write("\nNumber of top-ranked translations selected from agreement degree " + str(vDegree) + ": " + str(vlTopRankesDegree[vDegree - 1]))
	
	# outputing agreement (intersection) statistics for all degrees
	vfStatOut.write("\n")
	vfStatOut.write("\nAgreement Statistics:")
	vfStatOut.write("\n")
	for vDegree in range (vSysNum, 1, -1):
		vfStatOut.write("\nNumber of sentences having degree-" + str(vDegree) + " agreements: " + str(vlDegreeInterCounts[vDegree - 1]))
		vfStatOut.write("\nPercentage of sentences having degree-" + str(vDegree) + " agreements: %" + str(vlDegreeInterCounts[vDegree - 1]*100/vSentenceNum))
		vlDegreeInterSizes = [item[vDegree - 1] for item in vlIntersectSizes]
		vfStatOut.write("\nMin/max degree-" + str(vDegree) + " agreement size: " + str(min(vlDegreeInterSizes)) + '/' + str(max(vlDegreeInterSizes)))
		vfStatOut.write("\nAverage degree-" + str(vDegree) + " agreement size: " + str(sum(vlDegreeInterSizes)/vlDegreeInterCounts[vDegree - 1]))
		vfStatOut.write("\n")
	vfStatOut.write("\nNumber of sentences having no agreement: " + str(vlDegreeInterCounts[0]))
	vfStatOut.write("\nPercentage of sentences having no agreement: %" + str(vlDegreeInterCounts[0]*100/vSentenceNum))
	vfStatOut.write("\n")
	
	vfSourceSet.close()
	vfTransOut.close()
	if opts.rerankDetailOut != None:
		vfRerankDetailOut.close()
	vfStatOut.close()

 
##======================================================================
## calling main
if __name__ == "__main__":
 sys.exit(main())

