#! /usr/bin/python

## This script extracts n-best translations of each sentence in the input
## n-best file together with their scores.
##
## Version: 1.2
##
## Added:
## - The default behavior which was outputting only the translation (without
##   scores) is now optional (-s or --sentence). Instead, it outputs
##   all n-bests of a single sentence into a separate file as it is in the
##   input n-best file (i.e. sentence no, sentence, component scores, 
##   decoder score).
##
## Version: 1.1
##
## Added:
## - Option -a is added to mask @ (to prevent Berkeley wrong parsing) 
##
## Changed:
## - When -b and -a options are used, corresponding suffixes are added 
##   to the output file names.

import sys, optparse, codecs


##======================================================================
## main
def main(argv=None):
 if argv is None:
  argv = sys.argv
 
 parser = optparse.OptionParser(usage="%prog <INPUT N-BEST FILE> <SUFFIX LENGTH>" + 
                                      "\nExtracts n-best translations of each sentence in the input n-best file and their scores each into separate files.", version="%prog 1.0")

 parser.add_option("-b", "--maskrbrackets", help="mask round brackets with -LRB- and -RRB-", metavar="MASK RB", dest="maskRBrackets", action="store_true")
 parser.add_option("-a", "--maskat", help="mask @ with -atsymb-", metavar="MASK AT", dest="maskAT", action="store_true")
 parser.add_option("-s", "--sentence", help="output only translation", metavar="SENTENCE ONLY", dest="onlySentence", action="store_true")

 # processing input arguments

 (opts, posArgs) = parser.parse_args()
 
 if len(posArgs) < 2:
  parser.error("At least 2 arguments is required")
  
 # opening input n-best file
 try:
  vfInput  = codecs.open(posArgs[0], encoding='utf-8')
 except IOError:
  sys.exit('Can\'t open input n-best file: ' + posArgs[0])


 ## depending on the options, additional suffixes should be added to the 
 ## output file name
 vExtraSuffix = ''
 if opts.maskAT == True:
  vExtraSuffix += 'at.'

 if opts.maskRBrackets == True:
  vExtraSuffix += 'mrb.'
  
    
  
 ## Suffix length is used to format the name output files. One file is
 ## created for n-best translations of each sentence, with a name formed 
 ## with the name of input file followed by a suffix representing sentence
 ## index (zero-based). E.g., a suffix length of 2 leads to a name like
 ## xxxx.00 for the n-best of first sentence.
 vSuffixLen = int(posArgs[1])
 vCurrSentence = -1
 for line in vfInput:
  vlLineSplit = line.split("|||")
  vSentenceNo = int(vlLineSplit[0].strip())
  vSentence = vlLineSplit[1].strip() + ' '
  vModelScores = vlLineSplit[2].strip()
  vScore = vlLineSplit[3].strip()
  
  print "Extracting for sentence " + str(vSentenceNo + 1) + " ..."
  
  if opts.maskAT == True:
   vSentence = vSentence.replace('@', "-atsymb-")

  if opts.maskRBrackets == True:
   vSentence = vSentence.replace('(', "-LRB-")
   vSentence = vSentence.replace(')', "-RRB-")
  
  # checking if this is the n-best for a new sentence
  if vCurrSentence == vSentenceNo:
   # writing the sentence into a separate file if the option is specified
   if opts.onlySentence:
    vfSentOutput.write(vSentence + '\n')
   else:
    vfOutput.write(str(vSentenceNo) + " ||| " + vSentence + " ||| " + vModelScores + " ||| " + vScore + ' \n')

  else:
   try:
    vfOutput.close()
   except NameError:
    pass

   try:
    vfSentOutput.close()
   except NameError:
    pass

   vCurrSentence += 1
   # creating output files for the new sentence
   vOutFilePrefix = posArgs[0] + '.' + vExtraSuffix + str(vCurrSentence).zfill(vSuffixLen)
   try:
	# writing the sentence into a separate file if the option is specified
    if opts.onlySentence:
     vfSentOutput = codecs.open(vOutFilePrefix, mode='w', encoding='utf-8')
     vfSentOutput.write(vSentence + '\n')
    else:
     vfOutput = codecs.open(vOutFilePrefix, mode='w', encoding='utf-8')
     vfOutput.write(str(vSentenceNo) + " ||| " + vSentence + " ||| " + vModelScores + " ||| " + vScore + ' \n')
     
   except IOError:
    sys.exit('Can\'t create output file: ' + vOutFilePrefix)
    

 try:
  vfOutput.close()
 except NameError:
  pass

 try:
  vfSentOutput.close()
 except NameError:
  pass

 vfInput.close()
 

##======================================================================
## calling main
if __name__ == "__main__":
 sys.exit(main())

