#! /usr/bin/python

## This script extracts features for MT system selection, and
## output it into an ARFF file to be used in Weka machine learning
## toolkit.
## It can also extract features from ARFF files which are already
## prepared for training or test for other related purposes.
## It was originally written to take parser accuracy prediction
## from its ARFF file and add new class for each instance which
## represents an MT system that performs the best among others
## for that instance (based on sentence-level metric like TER).
##
## Current version: 1.0
##

import sys, weka, optparse


##----------------------------------------------------------------------
## extracts classes which are selected MT systems for each instance
def extractClasses(pClassFileName):
 try:
  vfClasses = open(pClassFileName, 'r')
 except IOError:
  sys.exit('Can\'t open class (selected MT system) file: ' + pClassFileName)

 vlClasses = []
 for line in vfClasses:
  ## the format of class file is supposed to be:
  ## TER1 ||| TER2 ||| TER3 ||| CLASS
  vlClasses.append(line.split()[6])

 print str(len(vlClasses)) + " samples extracted."

 return vlClasses



##----------------------------------------------------------------------
## extracts attribute lines from ARFF feature file
## If pflgLastIn is set to true, it considers the last attribute, which
## is usially outcome or class as a non-class attribute, in new feature
## set.
def extractFeatureFileAttrs(pFeatureFileName, pflgLastIn):
 try:
  vfFeatures = open(pFeatureFileName, 'r')
 except IOError:
  sys.exit('Can\'t open input ARFF feature file: ' + pFeatureFileName)

 vlAttrs = []

 # skip until first attribute
 vLine = vfFeatures.readline()
 while not vLine.upper().startswith("@ATTRIBUTE"):
  vLine = vfFeatures.readline()
  continue

 ## now start collecting attributes until they are found
 ## Note that the first attribute line is fetched above
 while vLine.upper().startswith("@ATTRIBUTE"):
  vlAttrs.append(vLine)
  vLine = vfFeatures.readline()

 # deciding on last attribute inclusion
 if not pflgLastIn:
  vlAttrs = vlAttrs[:-1]

 print str(len(vlAttrs)) + " attributes extracted from feature file"

 vfFeatures.close()

 return vlAttrs



##----------------------------------------------------------------------
## extracts instances (data lines) from ARFF feature file
## If pflgLastIn is set to false, it excludes the last feature value of
## al instances.
def extractFeatureFileData(pFeatureFileName, pflgLastIn):
 try:
  vfFeatures = open(pFeatureFileName, 'r')
 except IOError:
  sys.exit('Can\'t open input ARFF feature file: ' + pFeatureFileName)

 vlInstances = []

 # skip until data part
 while not vfFeatures.readline().upper().startswith("@DATA"):
  continue

 # now start collecting attributes until they are found
 vLine = vfFeatures.readline()
 while not vLine.strip() == "":
  # deciding on deleting the last feature value
  if not pflgLastIn:
   vLine = vLine[:vLine.rfind(',')]

  vlInstances.append(vLine)
  vLine = vfFeatures.readline()


 print str(len(vlInstances)) + " instances extracted from feature file"

 vfFeatures.close()

 return vlInstances


##----------------------------------------------------------------------
## extracts features and writes them together with those extracted from
## an input feature file into output ARFF file
def extractFeatures(pClasses, pOutFile, pFeatureFileData):
 vLoopCntr = 0
 for vClass, vFeatureFileInstance in zip(pClasses, pFeatureFileData):
  # writing feature file part
  pOutFile.write(vFeatureFileInstance.strip(" \n"))

  # writing class value
  pOutFile.write(',' + vClass + '\n');

  vLoopCntr += 1

 pOutFile.write('\n')
 return



##----------------------------------------------------------------------
## writes the header of the output file according to the header file provided
## It ignores any attribute provided in the header file, and adds attributes
## together with those provided in pFeatureFileAttrs.
## It also adds class attribute at the end of attribute decalartions and
## writes the @DATA at the end.
def writeHeader(pHeaderFile, pOutFile, pFeatureFileAttrs):
 for line in pHeaderFile:
  if line.upper().startswith("@RELATION"):
   pOutFile.write(line)
   break
  else:
   pOutFile.write(line)
 pOutFile.write('\n')

 # adding attribute declarations in pFeatureFileAttrs
 for vAttr in pFeatureFileAttrs:
  pOutFile.write(vAttr)

 ## adding attribute declarations for class
 ## Note: it's better to extract the classes from class file provided rather
 ## than hard-coding in here.
 pOutFile.write("@ATTRIBUTE\tclass\t\t{basep,basehp,basets}\n")

 pOutFile.write('\n')
 pOutFile.write("@DATA\n")



##======================================================================
## main
def main(argv=None):
 if argv is None:
  argv = sys.argv

 parser = optparse.OptionParser(usage="%prog <ARFF FILE HEADER> <OUTPUT FILE NAME> [OPTIONS]" +
                                      "\nExtracts a set of features for MT system selection, and output it into an ARFF file to be used in Weka machine learning toolkit.", version="%prog 1.0")

 parser.add_option("-c", "--class", help="class (selected MT systems) file ", metavar="CLASS FILE", dest="classFileName", action="store")
 parser.add_option("-n", "--samplesize", help="number of instances if MT systems file is not specified (for unlabeled instances)", metavar="SAMPLE SIZE", dest="sampleSize", action="store")
 parser.add_option("-f", "--featurefile", help="ARFF training file containing features", metavar="ARFF TRAINING FILE", dest="featureFileName", action="store")
 parser.add_option("-l", "--lasin", help="include the last attribute of --featurefile as feature (otherwise leave it out)", metavar="INCLUDE LAST ATTRIBUTE", dest="lastIn", action="store_true")

 (opts, posArgs) = parser.parse_args()


 # processing arguments

 if len(posArgs) < 2:
  parser.error("At least 2 arguments are required!")


 # opening header file
 try:
  vfHeader  = open(posArgs[0], 'r')
 except IOError:
  sys.exit('Can\'t open header file: ' + posArgs[0])

 # creating output file
 try:
  vfOutput  = open(posArgs[1] + '.arff', 'w')
 except IOError:
  sys.exit('Can\'t create output file: ' + posArgs[1])


 # processing options

 ## loading selected MT system (classes) into memory
 ## Note that for unlabeled test data, this is not given. Instead, the
 ## sample set size (-n or --samplesize option) must be provided. In this
 ## case, ? is a class template for each instance.
 vlClasses = []
 if opts.classFileName != None:
  vlClasses = extractClasses(opts.classFileName)
 elif opts.sampleSize != None:
  vlClasses = ['?'] * int(opts.sampleSize)
 else:
  sys.exit('Either sample size or classes (selected MT systems) should be provided!')


 # extracting attributes lines from ARFF feature file
 vlFeatureFileAttrs = extractFeatureFileAttrs(opts.featureFileName, opts.lastIn)

 ## writing output file header
 ## remember to add new feature here
 writeHeader(vfHeader, vfOutput, vlFeatureFileAttrs)

 # extracting instances from ARFF feature file
 vlFeatureFileData = extractFeatureFileData(opts.featureFileName, opts.lastIn)

 # checking whether the number of classes and instance in feature file match
 if len(vlFeatureFileData) != len(vlClasses):
  sys.exit("Number of classes and instance in feature file do not match!")

 ## extracting and writing features
 ## remember to add new feature here
 extractFeatures(vlClasses, vfOutput, vlFeatureFileData)


 vfHeader.close()
 vfOutput.close()


##======================================================================
## calling main
if __name__ == "__main__":
 sys.exit(main())

