#! /usr/bin/python

"""	
	This module defines functions to evaluate machine translation output.
	
	Version: 1.3											(14-Mar-2013)
	- scoreTER() is edited.
	- wrapTER() is added.
	- extractTERScores() is renamed to extractSegTERs()
	- scoreMTEval13a() is improved.
	
	Version: 1.2
	- extractTERScores() is added
	
	Version: 1.1
	- scoreBLEUNIST() is added
	
	Version: 1.0
	- wrapMTEvalXML()
	- scoreTER()
"""

from decimal import *
import sys, tempfile, subprocess, os

def scoreMTEval13a(pTransFile, pRefFile, pSrcFile, 
				   pScorerPath, pScorer="mteval-v13a.pl", pPerlLib="", 
				   pOutFilename="", pOutputType=0, plScorerOptions=[],
				   pflgTestSig=False, pStatSigPath="", pStatSigScorer="mteval-v13a-sig.pl",
				   pDatasetID="Dataset", pDocID="Dataset", pSrcLang="Any", pTrgLang="Any", pSysID="NewSys", pRefID="Dataset", pGenre="", pflgKeepXML=False,
				   pBaseTransFile="", pBaseSysID="BaseLine"):
	'''
	Scores the translation set in pTransFile against references in pRefFile
	with NIST and BLEU using mteval v13a and return scores.
	
	It also optionally calculate the statistical significance of the 
	difference between pTransFile and pBaseTransFile using CMU ARK tool 
	(outputs the results in both directions.) For this case, pScorerPath,
	pPerlLib and pStatSigPath should be properly set.
	
	It save the scores into file and also returns the segment- and document-
	level scores.
	
	pOutputType is the same as -d option:
		- 0 (default) for system-level score only
		- 1 to include document-level scores
		- 2 to include segment-level scores
		- 3 to include ngram-level scores
	'''
	
	# preparing translation file
	
	vTransXML = pTransFile + ".xml"
	try:
		vfTransXML = open(vTransXML, 'w')
	except:
		sys.exit("Cannot create translation XML: " + vTransXML)
	
	wrapMTEvalXML(pTransFile, "tstset", pDatasetID, pDocID, pSrcLang, pTrgLang, pSysID, pRefID, pGenre, vfTransXML)
	vfTransXML.close()
	
	# preparing reference file
	
	vRefXML = pRefFile + ".xml"
	try:
		vfRefXML = open(vRefXML, 'w')
	except:
		sys.exit("Cannot create reference XML: " + vRefXML)
	
	wrapMTEvalXML(pRefFile, "refset", pDatasetID, pDocID, pSrcLang, pTrgLang, pSysID, pRefID, pGenre, vfRefXML)
	vfRefXML.close()
	
	# preparing source file
	
	vSrcXML = pSrcFile + ".xml"
	try:
		vfSrcXML = open(vSrcXML, 'w')
	except:
		sys.exit("Cannot create source XML: " + vSrcXML)
	
	wrapMTEvalXML(pSrcFile, "srcset", pDatasetID, pDocID, pSrcLang, pTrgLang, pSysID, pRefID, pGenre, vfSrcXML)
	vfSrcXML.close()
	
	# creating score output file
	
	if pOutFilename == "":
		vTransEvalFilename = pTransFile + ".mteval13a"
	else:
		vTransEvalFilename = pOutFilename
	
	try:
		vfTransEval = open(vTransEvalFilename, 'w')
	except IOError:
		sys.exit("Cannot create score output file: " + vTransEvalFilename)
	
	# scoring
	subprocess.call(["perl", "-I" + pPerlLib, pScorerPath + "/" + pScorer, "-d", str(pOutputType), "-r", vRefXML, "-t", vTransXML, "-s", vSrcXML] + plScorerOptions, stdout=vfTransEval)
	
	vfTransEval.close()
	
	
	# testing statistical significance
	
	if pflgTestSig:
		print "Testing statistical significance of BLEU difference (CMU ARK tool) ..."
		# preparing baseline translation file
		vBaseTransXML = pBaseTransFile + ".xml"
		try:
			vfBaseTransXML = open(vBaseTransXML, 'w')
		except:
			sys.exit("Cannot create translation XML: " + vBaseTransXML)
		wrapMTEvalXML(pBaseTransFile, "tstset", pDatasetID, pDocID, pSrcLang, pTrgLang, pBaseSysID, pRefID, pGenre, vfBaseTransXML)
		vfBaseTransXML.close()
		
		# generating stat file for the translation
		
		vTransStatFile = pTransFile + ".stat"
		## Stat generator adds to the end of existing stat file instead 
		## of recretaing it. So, it's removed here before.
		if os.path.exists(vTransStatFile):
			os.remove(vTransStatFile)
		
		vDevnull = open(os.devnull, 'w')
		subprocess.call(["perl", "-I" + pPerlLib, pStatSigPath + "/" + pStatSigScorer, "-r", vRefXML, "-t", vTransXML, "-s", vSrcXML, "-f", vTransStatFile], stdout=vDevnull)
		
		# generating stat file for the baseline
		
		vBaseStatFile = pBaseTransFile + ".stat"
		if os.path.exists(vBaseStatFile):
			os.remove(vBaseStatFile)
		
		# creating score output file for baseline
		vBaseEvalFilename = pBaseTransFile + ".mteval13a"
		try:
			vfBaseEval = open(vBaseEvalFilename, 'w')
		except IOError:
			sys.exit("Cannot create score output file: " + vBaseEvalFilename)
		
		subprocess.call(["perl", "-I" + pPerlLib, pStatSigPath + "/" + pStatSigScorer, "-r", vRefXML, "-t", vBaseTransXML, "-s", vSrcXML, "-f", vBaseStatFile], stdout=vfBaseEval)
		
		vfBaseEval.close()
		
		vfStatSigTemp = tempfile.NamedTemporaryFile(delete=False)
		
		# testing significance of translation against baseline
		subprocess.call([pStatSigPath + "/paired_bootstrap_resampling_bleu_v13a.pl", vTransStatFile, vBaseStatFile, "1000", "0.05"], stdout=vfStatSigTemp)
		
		# testing significance of baseline against translation
		subprocess.call([pStatSigPath + "/paired_bootstrap_resampling_bleu_v13a.pl", vBaseStatFile, vTransStatFile, "1000", "0.05"], stdout=vfStatSigTemp)
		
		# reading the results
		
		vStatSigResult = []
		
		vfStatSigTemp.close()
		try:
			vfStatSigTemp = open(vfStatSigTemp.name)
		except IOError:
			sys.exit("Cannot read temporary statistical significance result file")
		
		# assumed format: System 1 BLEU better: 3 / 1000 = 0.003
		vStatSigResult.append(pSysID + " BLEU is better in :" + vfStatSigTemp.readline()[22:])
		vStatSigResult.append(pBaseSysID + " BLEU is better in :" + vfStatSigTemp.readline()[22:])
		vfStatSigTemp.close()
		
		# writing significance test results into output file
		
		try:
			vfStatSig = open(pTransFile + ".mteval13a-sig", 'w')
		except IOError:
			sys.exit("Cannot create statistical significance results file: " + pTransFile + ".mteval13a-sig")
		
		vfStatSig.write(vStatSigResult[0] + '\n' + vStatSigResult[1] + '\n')
		vfStatSig.close()
		
		if os.path.exists(vTransStatFile):
			os.remove(vTransStatFile)
		if os.path.exists(vBaseStatFile):
			os.remove(vBaseStatFile)
		os.remove(vfStatSigTemp.name)
	
	if not pflgKeepXML:
		os.remove(vTransXML)
		if pflgTestSig:
			os.remove(vBaseTransXML)
		os.remove(vRefXML)
		os.remove(vSrcXML)
	
	vNIST, vBLEU, vlSegNISTs, vlSegBLEUs = extractMTEvalScores(vTransEvalFilename, pOutputType)
	
	if pOutputType == 2:
		return vNIST, vBLEU, vlSegNISTs, vlSegBLEUs
	else:
		return vNIST, vBLEU
	

def wrapMTEvalXML(pInputFile, pDatasetType, pDatasetID, pDocID, pSrcLang, pTrgLang, pSysID, pRefID, pGenre, pOutDevice):
	'''
	wraps the input file content in XML complying with NIST mteval p13a 
	format.
	'''
	
		# opening input file
	try:
		vfInput  = open(pInputFile)
	except IOError:
		sys.exit('Can\'t open input file: ' + pInputFile)
	
	pOutDevice.write('<?xml version="1.0" encoding="UTF-8"?>' + '\n')
	pOutDevice.write('<!DOCTYPE mteval SYSTEM "ftp://jaguar.ncsl.nist.gov/mt/resources/mteval-xml-v1.3.dtd">' + '\n')
	pOutDevice.write('<mteval>' + '\n')
	
	if pDatasetType == "srcset":
		pOutDevice.write('<' + pDatasetType + ' setid="' + pDatasetID + '" srclang="' + pSrcLang + '">' + '\n')
	elif pDatasetType == "refset":
		pOutDevice.write('<' + pDatasetType + ' setid="' + pDatasetID + '" srclang="' + pSrcLang + '" trglang="' + pTrgLang + '" refid="' + pRefID + '">' + '\n')
	elif pDatasetType == "tstset":
		pOutDevice.write('<' + pDatasetType + ' setid="' + pDatasetID + '" srclang="' + pSrcLang + '" trglang="' + pTrgLang + '" sysid="' + pSysID + '">' + '\n')
	
	if not pGenre:
		vGenre = "default"
	else:
		vGenre = pGenre
	
	pOutDevice.write('<doc docid="' + pDocID + '" genre="' + vGenre + '">' + '\n')

	
	vLineCntr = 1
	for vLine in vfInput:
		# preparing the line
		vLine = vLine.strip('\n')
		vLine = vLine.replace('<', '-LAB-')
		vLine = vLine.replace('>', '-RAB-')
		vLine = vLine.replace('&', '-AMP-')
		
		pOutDevice.write('<p>' + '\n')
		pOutDevice.write('<seg id="' + str(vLineCntr) + '"> ' +  vLine + ' </seg>' + '\n')
		pOutDevice.write('</p>' + '\n')
		vLineCntr += 1

	pOutDevice.write('</doc>' + '\n')
	pOutDevice.write('</' + pDatasetType+ '>' + '\n')
	pOutDevice.write('</mteval>' + '\n')
	
	
	vfInput.close()
	

def extractMTEvalScores(pScoreFile, pScoreFileType):
	'''
	Extracts BLEU and NIST scores from pScoreFile based on pScoreFileType
	
	pScoreFileType specifies the types of score in the pScoreFile thus to 
	be extracted. It is the same as -d option of the MTEval scorer:
		- 0 (default) for system-level score only
		- 1 to include document-level scores
		- 2 to include segment-level scores
		- 3 to include ngram-level scores
	'''
	
	import re
	
	try:
		vlScoreFileLines = open(pScoreFile).read().split('\n')
	except IOError:
		sys.exit("Cannot open score file %s" % pScoreFile)
	
	vNIST = vBLEU = 0
	vlSegNISTs = []
	vlSegBLEUs = []
	
	if pScoreFileType == 0:
		for vLine in vlScoreFileLines:
			vlScores = re.findall(r"NIST score = ([\d.]+)[ \t]+BLEU score = ([\d.]+) for system", vLine)
			if len(vlScores) > 0:
				vNIST = Decimal(vlScores[0][0])
				vBLEU = Decimal(vlScores[0][1])
				break
	elif pScoreFileType == 2:
		for vLine in vlScoreFileLines:
			vlScores = re.findall(r"NIST score = ([\d.]+)[ \t]+BLEU score = ([\d.]+) for system", vLine)
			if len(vlScores) > 0:
				vNIST = Decimal(vlScores[0][0]) 
				vBLEU = Decimal(vlScores[0][1]) 
				break
			
			vlScores = re.findall(r"NIST score using \d-grams = ([\d.]+) for system .*? on segment (\d+)", vLine)
			if len(vlScores) > 0:
				vlSegNISTs.append(Decimal(vlScores[0][0]))
				# sanity check
				vSegmentNo = int(vlScores[0][1])
				if vSegmentNo != len(vlSegNISTs):
					sys.exit("Some went wrong with segment NISTs! Segment number mismatch: %s vs. %s" % (vSegmentNo, len(vlSegNISTs)))
				else:
					continue
			
			vlScores = re.findall(r"BLEU score using \d-grams = ([\d.]+) for system .*? on segment (\d+)", vLine)
			if len(vlScores) > 0:
				vlSegBLEUs.append(Decimal(vlScores[0][0]))
				# sanity check
				vSegmentNo = int(vlScores[0][1])
				if vSegmentNo != len(vlSegBLEUs):
					sys.exit("Some went wrong with segment BLEUs! Segment number mismatch: %s vs. %s" % (vSegmentNo, len(vlSegBLEUs)))
				else:
					continue
			
	
	return vNIST, vBLEU, vlSegNISTs, vlSegBLEUs
	

def scoreTERcom(pTransFile, pRefFile, pScorerClass, pOutFilePrefix="", plScorerOptions=[], pOutputType="ter", pflgCutoff=False, pOutDevice=None):
	'''
	Scores the translation set in plTransSents against references in 
	plRefSents with TER metric using tercmp scorer.
	
	It save the scores into file and also returns the segment- and document-
	level scores.
	
	If pflgCutoff is set to True, it return another segment score list and 
	document score for which the segment scores higher than 1 (number of 
	errors bigger than sentence length) are cut off (similar to cap option
	in TERp). In this case, cut-off document score are appended to the 
	end of output score file.
	'''
	
	if pOutFilePrefix == "":
		vOutFilePrefix = pTransFile
	else:
		vOutFilePrefix = pOutFilePrefix
	
	# writing the translations into a temp file
	
	vfTransTemp = tempfile.NamedTemporaryFile(delete=False)
	vfRefTemp = tempfile.NamedTemporaryFile(delete=False)
	
	vfTransTemp.write('\n'.join(wrapTER(open(pTransFile).read().strip().split('\n'))) + '\n')
	vfRefTemp.write('\n'.join(wrapTER(open(pRefFile).read().strip().split('\n'))) + '\n')
	
	vfTransTemp.close()
	vfRefTemp.close()
	
	# scoring
	subprocess.call(["java", "-jar", pScorerClass, "-h", vfTransTemp.name, "-r", vfRefTemp.name, "-n", vOutFilePrefix, "-o", pOutputType] + plScorerOptions, stdout=pOutDevice)
	
	os.remove(vfTransTemp.name)
	os.remove(vfRefTemp.name)
	
	# extracting and returning scores from generated score files to return
	
	vScoreFilename = vOutFilePrefix + '.' + pOutputType
	
	if pflgCutoff:
		vDocTER, vlSegTERs, vCODocTER, vlCOSegTERs = extractTERScores(vScoreFilename, pOutputType, True)
		
		# appending cut-off score to the end of score file
		open(vScoreFilename, 'a').write("\nCut-off score: %s\nCut-off 1-TER: %s\n" % (vCODocTER, str(1 - vCODocTER)))
		
		return vDocTER, vlSegTERs, vCODocTER, vlCOSegTERs
	else:
		return extractTERScores(vScoreFilename, pOutputType, False)
	

def wrapTER(plSegments):
	'''
	Wraps segments with an ID as TER scorers require a sentence ID at the
	end of each segment
	'''
	
	vlWrapped = []
	for i, vSeg in enumerate(plSegments):
		vlWrapped.append(vSeg + ' (' + str(i) + ')')
	
	return vlWrapped
	

def extractTERScores(pScoreFile, pScoreFileType, pflgCutoff=False):
	'''
	Extracts TER scores from pScoreFile based on pScoreFileType and returns
	segment and document scores
	
	pScoreFileType is one of the score file format supported by TER scoring
	scheme.
	
	If pflgCutoff is set to True, it return another segment score list and 
	document score for which the segment scores higher than 1 (number of 
	errors bigger than sentence length) are cut off (similar to cap option
	in TERp).
	
	'''
	
	import re
	
	try:
		vlScoreFileLines = open(pScoreFile).read().split('\n')
	except IOError:
		sys.exit("Cannot open score file %s" % pScoreFile)
	
	vlSegTERs = []
	vErrSum = Decimal(0)
	vLenSum = Decimal(0)
	vDocTER = Decimal(0)
	vlCOSegTERs=[]
	vCOErrSum = Decimal(0)
	vCOLenSum = Decimal(0)
	vCODocTER = Decimal(0)
	
	if pScoreFileType == "ter":
		for vLine in vlScoreFileLines:
			vlComponents = re.findall(r"(\d+):(\d+) ([\d.]+) ([\d.]+) ([\d.]+)", vLine)
			## NOTE: the second component seems to be reference ID for 
			## multi-reference which is not yet supported. 
			if len(vlComponents) > 0:
				vErrSum += Decimal(vlComponents[0][2])
				vLenSum += Decimal(vlComponents[0][3])
				vlSegTERs.append(Decimal(vlComponents[0][4]))
				if pflgCutoff:
					vCOLenSum += Decimal(vlComponents[0][3])
					if vlSegTERs[-1] > 1:
						vlCOSegTERs.append(Decimal(1.0))
						vCOErrSum += Decimal(vlComponents[0][3])
					else:
						vlCOSegTERs.append(Decimal(vlComponents[0][4]))
						vCOErrSum += Decimal(vlComponents[0][2])
				
				# sanity check
				if int(vlComponents[0][0]) != len(vlSegTERs) - 1:
					sys.exit("Some went wrong with! Segment number mismatch: %s vs. %s" % (vlComponents[0][0], len(vlSegTERs) - 1))
				continue
			
		vDocTER = vErrSum / vLenSum
		if pflgCutoff:
			vCODocTER = vCOErrSum / vCOLenSum
	elif pScoreFileType == "sum":
		for vLine in vlScoreFileLines:
			vlComponents = re.findall(r"(\d+):(\d+)[ \t|]+([\d.]+)[ \t|]+([\d.]+)[ \t|]+([\d.]+)[ \t|]+([\d.]+)[ \t|]+([\d.]+)[ \t|]+([\d.]+)[ \t|]+([\d.]+)[ \t|]+([\d.]+)", vLine)
			## NOTE: the second component seems to be reference ID for 
			## multi-reference which is not yet supported. 
			if len(vlComponents) > 0:
				vErrSum += Decimal(vlComponents[0][7])
				vLenSum += Decimal(vlComponents[0][8])
				vlSegTERs.append(Decimal(vlComponents[0][9]))
				if pflgCutoff:
					vCOLenSum += Decimal(vlComponents[0][8])
					if vlSegTERs[-1] > 100:
						vlCOSegTERs.append(Decimal(100.0))
						vCOErrSum += Decimal(vlComponents[0][8])
					else:
						vlCOSegTERs.append(Decimal(vlComponents[0][9]))
						vCOErrSum += Decimal(vlComponents[0][7])
				
				# sanity check
				if int(vlComponents[0][0]) != len(vlSegTERs) - 1:
					sys.exit("Some went wrong with! Segment number mismatch: %s vs. %s" % (vlComponents[0][0], len(vlSegTERs) - 1))
				continue
			
		vDocTER = vErrSum / vLenSum
		if pflgCutoff:
			vCODocTER = vCOErrSum / vCOLenSum
	
	
	if pflgCutoff:
		return vDocTER, vlSegTERs, vCODocTER, vlCOSegTERs
	else:
		return vDocTER, vlSegTERs
	

def scoreMeteor14(pTransFile, pRefFile, pLang, pScorerClass, pOutFilePrefix="", pOutputType="normal", pflgOutAlign=False, plScorerOptions=[]):
	'''
	Scores the translation set in plTransSents against references in 
	plRefSents with Meteor metric using version 1.4 of the scorer.
	
	It save the scores into file and also returns the segment- and document-
	level scores.
	'''
	
	if pOutFilePrefix == "":
		vOutFilePrefix = pTransFile
	else:
		vOutFilePrefix = pOutFilePrefix
	
	try:
		vfOut = open(vOutFilePrefix, 'w')
	except IOError:
		sys.exit("Cannot create score output file: " + vOutFilePrefix)
	
	# setting additional options
	
	if pOutputType.lower().startswith("verb"):
		plScorerOptions.append("-vOut")
	
	if pflgOutAlign:
		plScorerOptions.append("-writeAlignments")
	
	# scoring
	subprocess.call(["java", "-jar", pScorerClass, pTransFile, pRefFile, "-f", vOutFilePrefix, "-l", pLang] + plScorerOptions, stdout=vfOut)
	
	# extracting and returning scores from generated score files to return
	return extractMeteorScores(vOutFilePrefix, pOutputType)
	

def extractMeteorScores(pScoreFile, pScoreFileType):
	'''
	Extracts Meteor scores from pScoreFile based on pScoreFileType and 
	returns segment and document scores
	
	pScoreFileType specifies whether the score file is verbose (precision,
	recall, frag, score) or normal (only score).
	'''
	
	import re
	
	try:
		vlScoreFileLines = open(pScoreFile).read().split('\n')
	except IOError:
		sys.exit("Cannot open score file %s" % pScoreFile)
	
	vlScores = []
	vDocScore = Decimal(0)
	
	if pScoreFileType == "normal":
		for vLine in vlScoreFileLines:
			# segment scores
			vlComponents = re.findall(r"Segment (\d+) score:\s+([\d.]+)", vLine)
			if len(vlComponents) > 0:
				vlScores.append(Decimal(vlComponents[0][1]))
				# sanity check
				if int(vlComponents[0][0]) != len(vlScores):
					sys.exit("Something went wrong! Segment number mismatch: %s vs. %s" % (vlComponents[0][0], len(vlScores)))
				continue
			# final score
			vlComponents = re.findall(r"Final score:\s+([\d.]+)", vLine)
			if len(vlComponents) > 0:
				vDocScore = Decimal(vlComponents[0])
	elif pScoreFileType == "verbose":
		for vLine in vlScoreFileLines:
			# segment scores
			vlComponents = re.findall(r"Segment (\d+) score:\s+[\d.]+\s+[\d.]+\s+(?:[\d.]|NaN)+\s+([\d.]+)", vLine)
			if len(vlComponents) > 0:
				vlScores.append(Decimal(vlComponents[0][1]))
				# sanity check
				if int(vlComponents[0][0]) != len(vlScores):
					sys.exit("Something went wrong! Segment number mismatch: %s vs. %s" % (vlComponents[0][0], len(vlScores)))
				continue
			# final score
			vlComponents = re.findall(r"Final score:\s+([\d.]+)", vLine)
			if len(vlComponents) > 0:
				vDocScore = Decimal(vlComponents[0])
	
	return vDocScore, vlScores
	
