#! /usr/bin/python

"""	
	This module defines functions to process MT n-best output.
	
	Version: 1.7											(22-Aug-2012)
	- rerankMergedNBest() and sort functions are changed to support custom
	sort direction.
	
	Version: 1.6											(12-Aug-2012)
	- Minimum decoder rank is implemented (pMinDecScore)
	
	Version: 1.5											(07-Aug-2012)
	- calcUnionCombScores(), calcAgreeCombScores(), computeCombDecScores()
	is changed to add support for normalization.
	
	Version: 1.4 (02-Aug-2012)
	- calcMergedAgreeScore() is renamed to calcAgreeCombScores() and
	changed to fork score combination to decoder score and decorer rank
	- computeCombScores() is split into computeCombDecScores() and
	computeCombDecRanks()
	- combscore in CombinedNBestItem is changed combscores to contain 
	all possible comination scores (e.g. decoder score and decoder rank)
	- calcUnionCombScores() is added
	- Average is added as score combination method and avgScores() is added
	to cacluate it.
	
	Version: 1.3
	- extscores is added to CombinedNBestItem structure.
	- measure parameter values are changed to reflect reranking measure 
	  rather than combination.
	- rerankCombNBest() is split into calcMergedAgreeScore() and 
	  rerankAgreeMergedNBest() in order to combine scores and rerank the
	  list separetely.
	- rerankExtScoreNBest() is added.
	- combNBests() is renamed to mergeNBests().
	- rerankExtScoreMergedNBest() is added.
	- sortByExtScore() is added.
	- extractSentenceNBests() is added.
	- mergeNBests() is changed to accept n-best items in list instead of
	files
	
	Version: 1.2
	- rerankCombNBest() is changed to incorporate combination based on
	  n-best item index instead of decoder score.
	
	Version: 1.1
	- extractIntersect()
	- combineNBests()
	- rerankCombNBest()
	
"""

from collections import namedtuple
from decimal import *
import sys, util

'''
	data structure for n-best item
	Note: more details can be added 
'''
NBestItem = namedtuple('NBestItem', 'trans score')
 
 
'''
data structure for agreement-based-combined n-best items:
	trans: translated n-best sentence
	memidxs: list of membership indexes of the sentence in each system
			 n-best
	degree: degree of agreement (intersection) which is the number of sets
			haveing the n-best translation item in common
	sysscores: list of scores assigned by each system
	combscores: dictionary of combined scores such as agreement-based decoder
				score or rank
	extscores: dictionary of external scores such as TER or confidence
			   estimation
'''
CombinedNBestItem = namedtuple('CombinedNBestItem', 'trans memidxs degree sysscores combscores extscores')


def extractSentenceNBests(plNBestFilename, pflgVerbose):
	'''
	Given a n-best file, extracts and returns n-best items for each 
	sentence.
	'''
	
	# opening input n-best file
	try:
		vfNBest = open(plNBestFilename)
	except IOError:
		sys.exit('Can\'t open input n-best file: ' + plNBestFilename)
	
	vlNBestItems = []
	vlSentNBestItems = []
	vCurrSentence = -1
	for line in vfNBest:
		vlLineSplit = line.split("|||")
		vSentenceNo = int(vlLineSplit[0].strip())
		vTrans = vlLineSplit[1].strip()
		vScore = vlLineSplit[3].strip()
		
		# checking if this is the n-best for a new sentence
		if vCurrSentence == vSentenceNo:
			vlSentNBestItems.append(NBestItem(vTrans, Decimal(vScore)))
		else:
			if len(vlSentNBestItems) > 0:
				vlNBestItems.append(vlSentNBestItems)
				vlSentNBestItems = []
			vlSentNBestItems.append(NBestItem(vTrans, Decimal(vScore)))
			vCurrSentence += 1
			if pflgVerbose:
				print "Sentence " + str(vCurrSentence + 1) + "..."
	
	## collecting the n-best items of last sentence (it cannot be collected
	## inside the loop)
	vlNBestItems.append(vlSentNBestItems)
	
	vfNBest.close()
	
	return vlNBestItems
	

def extractFromSingleNBest(pNBestFileName):
	'''
	Extracts sentence and decoder score from n-best output of a single
	sentence.
	'''
	
	vlNBestItems = []
	
	# opening input n-best file
	try:
		vfInput  = open(pNBestFileName)
	except IOError:
		sys.exit('Can\'t open input n-best file: ' + pNBestFileName)
 
  
	# processing n-best
	for line in vfInput:
		vlLineSplit = line.split("|||")
		vlNBestItems.append(NBestItem(vlLineSplit[1].strip(), Decimal(vlLineSplit[-1].strip())))
  
	vfInput.close()
 
	return vlNBestItems 
 
 
def extractIntersect(plNBestFileNames):
	'''Extracts intersection of n-best translation outputs of multiple
	   systems for a sentence.'''
	
	# data structure for returning data
	IntersectNBestItem = namedtuple('IntersectNBestItem', 'trans indexes scores')
	
	# loading n-best items
	vlNBestLists = []
	for vNBestFileName in plNBestFileNames:
		vlNBestLists.append(extractFromSingleNBest(vNBestFileName))
	
	# extracting intersect
	#
	## Note that Python's set intersection cannot be used since we want
	## to keep the translation order in the original n-best list.
	## To extract intersection we iterate the n-best translations of the
	## first system and take it if it occures in all systems' n-best.
	## We also save the index of each intersect translation in each n-best
	## list.
	
	vlIntersectNBestItems = []
	vCntr1 = 0
	for vNBestList1 in vlNBestLists[0]:
		# index of intersection n-best item in each system n-best
		vlIndexes = []	
		vlScores = []
		vInclude = True
		for vSysNBestItems in vlNBestLists[1:]:
			vIndex = -1
			vScore = 0.0
			vCntrX = 0
			for vNBestListX in vSysNBestItems:
				if vNBestList1.trans == vNBestListX.trans:
					vIndex = vCntrX
					vScore = vNBestListX.score
					break
				vCntrX += 1
			
			if vIndex != -1:
				vlIndexes.append(vIndex)
				vlScores.append(vScore)
			else:
				# if the same transaltion is not found in only one of the 
				# system n-bests, it is not included in the intersection
				vInclude = False
				break
					
		if vInclude:
			vlIndexes.insert(0, vCntr1)
			vlScores.insert(0, vNBestList1.score)
			vlIntersectNBestItems.append(IntersectNBestItem(vNBestList1.trans,
										 vlIndexes, vlScores))
		
		vCntr1 += 1		
	
	return vlIntersectNBestItems


def mergeNBests(plNBestLists):
	'''
	Given multiple sets of n-best lists for a sentence, combines translations 
	in them into a distinct set and identifies the membership of each 
	individual translation in each n-best list. For example, an item with 
	membership index list of [2, 10, 7, -1], given 4 n-best lists, is the
	member of first, second, and third sets in 2nd, 10th, and 7th indexes,
	but is not a member of fourth set. This item is called of agreement
	(intersection) degree of 3, since it is a member of 3 lists.
	'''
	
	# collecting a set of individual n-best translation sentences
	## This will later be attached with membership indexed and score list
	## and returned.
	vlNBestSentences = []
	for vSysNBestItems in plNBestLists:
		for vNBestItem in vSysNBestItems:
			if not vNBestItem.trans in vlNBestSentences:
				vlNBestSentences.append(vNBestItem.trans)
	
	# extracting intersections information between n-best sets
	#
	## To extract intersections we iterate the unique set of n-best sentences
	## extracted above, look for it in each of the n-best sets, and add
	## its index in each set to the membership index list if it is a member
	## or -1 otherwise.
	
	vlMergedNBest = []
	# iterating unique n-best sentences
	for vNBestSentence in vlNBestSentences:
		# iterating systems
		vlIndexes = []
		vlScores = []
		for vSysNBestItems in plNBestLists:
			# iterating system n-best items
			vIdxCntr = 0
			vIndex = -1
			vScore = Decimal(0)
			for vNBestItem in vSysNBestItems:
				if vNBestSentence == vNBestItem.trans:
					vIndex = vIdxCntr
					vScore = vNBestItem.score
					break
				vIdxCntr += 1
				
			vlIndexes.append(vIndex)
			vlScores.append(vScore)
		
		# colculating intersection degree
		vDegree = len([idx for idx in vlIndexes if idx > -1])
		
		vlMergedNBest.append(CombinedNBestItem(vNBestSentence, vlIndexes, vDegree, vlScores, {}, {}))
	
	return vlMergedNBest
	

def calcUnionCombScores(plMergedNBest, pMeasure, pCombMethod, pMinForZeroMargin, pMinDecRank, pflgNormalize, pSysScoreRanges, pNormTrgRange):
	'''
	Given a distinct set of merged n-best lists of multiple systems for 
	a sentence, each with the corresponding decoder rank and score in 
	their original n-bests, combines the decoder scores (original or 
	normalized) or ranks in a variety of configurations.
	
		plMergedNBest: distinct merged n-best list for the sentence
		pMeasure: measure to be combined:
				  - decoder-score
				  - decoder-rank (the original rank of the item in n-best)
		pCombMethod: score combination method:
					 -multiply
					 -sum
		pMinForZeroMargin: percentage of the minimum decoder score to be
						   subtracted from it and used to replace zero
						   score when n-best item translation does not 
						   exist in a system output:
						   -1: don't replace zero at all
			`			    0: use minimum itself (subtract 0)
		pMinDecRank: minimum decoder rank to replace -1 rank when n-best
					 item translation does not exist in a system output 
					 -1: original rank
		pflgNormalize: normalize systems scores before combining based on
					   lower and upper bounds in pSysScoreRanges
		pSysScoreRanges: lower and upperbound (min and max) systems scores
						 extracted from the enitire n-best list for each
						 system.
		pNormTrgRange: target range to normalize scores to
	'''
	
	# extracting the number of systems
	vSysNum = len(plMergedNBest[0].memidxs)
	
	if pMeasure == "decoder-score":
		vlScoredMergedNBest = computeCombDecScores(plMergedNBest, "udecscore", pCombMethod, pMinForZeroMargin, pflgNormalize, pSysScoreRanges, pNormTrgRange)
	elif pMeasure == "decoder-rank":
		vlScoredMergedNBest = computeCombDecRanks(plMergedNBest, "udecrank", pCombMethod, pMinDecRank)
		
	return vlScoredMergedNBest
	

def calcAgreeCombScores(plMergedNBest, pMinIntersectDegree, pEmptyIntersectMethod, pPrefSysNo, pMeasure, pCombMethod, pflgNormalize, pSysScoreRanges, pNormTrgRange):
	'''
	Given a distinct set of merged n-best lists of multiple systems, each
	with the corresponding decoder rank and score in their original n-bests
	and the degree of agreement on each item between by systems, combines 
	the decoder scores (original or normalized) or ranks in a variety of
	configurations.
	    
		plMergedNBest: distinct merged n-best list for the sentence
		pMinIntersectDegree: the minimum intersection degree to back off
							  to
		pEmptyIntersectMethod: method to be used to select translation
							   when intersections of all degrees higher
							   than pMinIntersectDegree are empty:
							   prefsys: select the output of a preferred
										system provided by pPrefSysNo
										(Note: in this case only one
											   item is returned instead
											   of reranking)
								confidence: rerank based on decoder score
		pPrefSysNo: the index (1-based) of system whose score should be
					used when there is no intersection
		pMeasure: measure to be combined (for intersection-based cases):
				  - decoder-score
				  - decoder-rank (the original rank of the item in n-best)
		pCombMethod: score combination method:
					 -multiply
					 -sum
		pflgNormalize: normalize systems scores before combining based on
					   lower and upper bounds in pSysScoreRanges
		pSysScoreRanges: lower and upperbound (min and max) systems scores
						 extracted from the enitire n-best list for each
						 system.
		pNormTrgRange: target range to normalize scores to
	'''
	
	# extracting the number of systems
	vSysNum = len(plMergedNBest[0].memidxs)
	
	# set of items falling in the requested agreement (intersection) degree
	vlIntersect = []
	# rest of the items (not falling in the minimum agreement degree
	vlComplement = []
	
	## finding items with the highest intersection degree by backing off
	## to a lower degree
	for vDegree in range(vSysNum, pMinIntersectDegree - 1, -1):
		vlIntersect = [item for item in plMergedNBest if item.degree == vDegree]
		# if the intersection is not empty, don't back off
		if len(vlIntersect) > 0:
			vlComplement = [item for item in plMergedNBest if item.degree != vDegree]
			# sanity check
			if len(vlIntersect) + len(vlComplement) != len(plMergedNBest):
				sys.exit("Problem with intersection and complement size!")
			break
	
	vIntersectLen = len(vlIntersect)
	
	# setting the combined score dictionary key based on measure type
	if pMeasure == "decoder-score":
		vCombScoreType = "adecscore"
	elif pMeasure == "decoder-rank":
		vCombScoreType = "adecrank"
	
			# if the intersection is empty, use pEmptyIntersectMethod to rerank
	if vIntersectLen == 0:
		if pEmptyIntersectMethod == "prefsys":
			## May be a more efficient way of chosing the output of a 
			## preferred system would be directly getting it from its 
			## n-best output. However, to unify things, everything is 
			## centeralized here.
			vlScoredMergedNBest = setPrefSysScores(plMergedNBest, pPrefSysNo, vCombScoreType)
		elif pEmptyIntersectMethod == "confidence":
			vlScoredMergedNBest = setConfScores(plMergedNBest, vCombScoreType)
	else:
		vlScoredIntersect = []
		if pMeasure == "decoder-score":
			vlScoredIntersect = computeCombDecScores(vlIntersect, vCombScoreType, pCombMethod, False, pflgNormalize, pSysScoreRanges, pNormTrgRange)
		elif pMeasure == "decoder-rank":
			vlScoredIntersect = computeCombDecRanks(vlIntersect, vCombScoreType, pCombMethod, -1)
		
		vlScoredComplement = []
		vlScoredComplement = addCombScore(vlComplement, vCombScoreType)
		vlScoredMergedNBest = vlScoredIntersect + vlScoredComplement
		
	## identifying the selection method used:
	## 0: highest degree intersection
	## 1: back off intersection
	## 2: back off no-intersection method
	vSelMethod = 0
	if vIntersectLen == 0:
		vSelMethod = 2
	else:
		if vlIntersect[0].degree == vSysNum:
			vSelMethod = 0
		else:
			vSelMethod = 1
	
	return vlScoredMergedNBest, vSelMethod
	

def rerankMergedNBest(plMergedNBest, pRerankMeasure, pSortDir):
	'''
	Reranks the merged n-best list based on pRerankMeasure.
	'''
	
	## When combined systems don't agree on any n-best item for a sentence
	## (empty intersection) and a preferred system's output is selected, 
	## items generated by other systems are scored 0. These items (if any
	## are filtered into the following lists. The zero-score items will
	## be merged after reranking
	## Note that the filtered list is only used by agreement-based reranking
	## functions below.
	vNonZeroAgreeMergedNBest = []
	vZeroAgreeMergedNBest = []
	[item for item in plMergedNBest if item.combscores != 0]
	
	if pRerankMeasure == "union-dec-score":
		return sortByCombScore(plMergedNBest, "udecscore", pSortDir)
	if pRerankMeasure == "union-dec-rank":
		return sortByCombScore(plMergedNBest, "udecrank", pSortDir)
	if pRerankMeasure == "agree-dec-score":
		vNonZeroAgreeMergedNBest = [item for item in plMergedNBest if item.combscores["adecscore"] != 0]
		vZeroAgreeMergedNBest = [item for item in plMergedNBest if item.combscores["adecscore"] == 0]
		return sortByCombScore(vNonZeroAgreeMergedNBest, "adecscore", pSortDir) + vZeroAgreeMergedNBest
	if pRerankMeasure == "agree-dec-rank":
		vNonZeroAgreeMergedNBest = [item for item in plMergedNBest if item.combscores["adecrank"] != 0]
		vZeroAgreeMergedNBest = [item for item in plMergedNBest if item.combscores["adecrank"] == 0]
		return sortByCombScore(vNonZeroAgreeMergedNBest, "adecrank", pSortDir) + vZeroAgreeMergedNBest
	if pRerankMeasure == "ext-score-ter":
		return sortByExtScore(plMergedNBest, "ter", pSortDir)
	if pRerankMeasure == "ext-score-ce":
		return sortByExtScore(plMergedNBest, "ce", pSortDir)
	


#-----------------------------------------------------------------------
# Local functions

def setPrefSysScores(plNBestItems, pPrefSysNo, pCombScoreType):
	''' 
	Sets the combination score of each item in the input n-best list to
	the score of system identified pPrefSysNo.
	'''
	vlScoredIntersect = []
	for vItem in plNBestItems:
		vItem.combscores[pCombScoreType] = vItem.sysscores[pPrefSysNo-1]
		vlScoredIntersect.append(CombinedNBestItem(vItem.trans, vItem.memidxs, vItem.degree, vItem.sysscores, vItem.combscores, {}))
	
	return vlScoredIntersect
	

def setConfScores(plNBestItems, pCombScoreType):
	'''
	Sets the combination score of each item in the input n-best list to 
	the maximum of system scores for that item.
	'''
	vlScoredIntersect = []
	for vItem in plNBestItems:
		vItem.combscores[pCombScoreType] = max([score for score in vItem.sysscores if score != 0])
		vlScoredIntersect.append(CombinedNBestItem(vItem.trans, vItem.memidxs, vItem.degree, vItem.sysscores, vItem.combscores, {}))
	
	return vlScoredIntersect
	

def computeCombDecScores(plNBestItems, pCombScoreType, pCombMethod, pMinForZeroMargin, pflgNormalize, pScoreRanges, pNormTrgRange):
	'''
	Sets the combination score of each item in the input n-best list to 
	the interpolation of decoder scores of systems for that item. 
	
	The inpterpolation method is identified by pCombMethod:
		pCombMethod: score combination method:
					 -multiply
					 -sum
		pMinForZeroMargin: percentage of the minimum decoder score to be
						   subtracted from it and used to replace zero
						   score when n-best item translation does not 
						   exist in a system output:
						   -1: don't replace zero at all
			`			    0: use minimum itself (subtract 0)
		pflgNormalize: normalize systems scores before combining based on
					   lower and upper bounds in pSysScoreRanges
		pSysScoreRanges: lower and upper bound (min and max) systems scores
						 extracted from the enitire n-best list for each
						 system.
	'''
	
	# extracting the number of systems
	
	vSysNum = len(plNBestItems[-1].sysscores)
	vlZeroReplacement = [Decimal(0)] * vSysNum
	
	## finding the minimum decoder score of each system in the
	## n-best items (for current sentence)
	if pMinForZeroMargin != -1:
		for vSysIdx in range(vSysNum):
			vMinScore = min(vItem.sysscores[vSysIdx] for vItem in plNBestItems)
			vlZeroReplacement[vSysIdx] =  vMinScore + (vMinScore * pMinForZeroMargin / 100)
	
	vlScoredIntersect = []
	for vItem in plNBestItems:
		vlScores = []
		for vScoreIdx in range(vSysNum):
			if vItem.sysscores[vScoreIdx] == 0:
				vScore = vlZeroReplacement[vScoreIdx]
			else:
				vScore = vItem.sysscores[vScoreIdx]
				
			vScoreRangeMin = pScoreRanges[vScoreIdx].min
			if pflgNormalize:
				## For some sentences for which the minimum is the 
				## global minimum for the system decoder score, 
				## marginalized score will fall out of range of system
				## score range. Therefor, When normalization takes 
				## place as well, error will occur. To work around it,
				## the score range is adjusted accordingly.
				if pMinForZeroMargin != -1:
					vScoreRangeMin = vScoreRangeMin + (vScoreRangeMin / 100)
			
				# if score is 0, do not normalize and pass 0 itself
				if vScore != 0:
					vScore = util.normalizeNumber(vScore, vScoreRangeMin, pScoreRanges[vScoreIdx].max, pNormTrgRange.min, pNormTrgRange.max)
			
			vlScores.append(vScore)
		
		if pCombMethod == "multiply":
			vItem.combscores[pCombScoreType] = multiplyScores(vlScores)
		elif pCombMethod == "sum":
			vItem.combscores[pCombScoreType] = sumScores(vlScores)
		elif pCombMethod == "avg":
			vItem.combscores[pCombScoreType] = avgScores(vlScores)
		
		vlScoredIntersect.append(CombinedNBestItem(vItem.trans, vItem.memidxs, vItem.degree, vItem.sysscores, vItem.combscores, {}))
	
	return vlScoredIntersect
	

def computeCombDecRanks(plNBestItems, pCombScoreType, pCombMethod, pMinDecRank):
	'''
	Sets the combination score of each item in the input n-best list to 
	the interpolation of decoder ranks of systems for that item. 
	
	The inpterpolation method is identified by pCombMethod:
	pCombMethod: score combination method:
			  -multiply
			  -sum
	pMinDecRank: minimum decoder rank to replace -1 rank when n-best
				 item translation does not exist in a system output 
				 -1: original rank

	'''
	
	vlScoredIntersect = []
	for vItem in plNBestItems:
		vlScores = []
		
		# setting new rank for -1 ranks based on pMinDecRank
		if pMinDecRank > 0:
			## making the provided minumum rank 0-based: e.g. 500 is meant
			## to be minimum replacement without small margin (see 
			## pMinForZeroMargin for computeCombDecScores()) so it should 
			## be equal to n-best's minimum possible rank which is 499 
			##(0-based). And 501 is meant to be minimum replacement with
			## a small margin (1) so it should be 500. So, any positive
			## pMinDecRank should be subtracted by 1
			vMinDecRank = pMinDecRank - 1
		else:
			vMinDecRank = pMinDecRank
		vlScores = [vMinDecRank if idx == -1 else idx for idx in vItem.memidxs]
		
		## adjusting ranks: multiplied by -1 to make the sort order uniform
		## with decoder scores and subtracted by 1 to make -1 (missing 
		## translation) 0 and 0 (best translation) -1 so to have uniform
		## combination functions with decoder scores
		vlScores = [-1*idx - 1 for idx in vlScores]
		
		if pCombMethod == "multiply":
			vItem.combscores[pCombScoreType] = multiplyScores(vlScores)
		elif pCombMethod == "sum":
			vItem.combscores[pCombScoreType] = sumScores(vlScores)
		elif pCombMethod == "avg":
			vItem.combscores[pCombScoreType] = avgScores(vlScores)
		
		vlScoredIntersect.append(CombinedNBestItem(vItem.trans, vItem.memidxs, vItem.degree, vItem.sysscores, vItem.combscores, {}))
	
	return vlScoredIntersect
	

def addCombScore(pMergedNBestItems, pCombScoreType):
	'''
	Adds a new combined score type to the merged n-best items
	'''
	vlNewNBestItems = []
	for vItem in pMergedNBestItems:
		vItem.combscores[pCombScoreType] = Decimal(0)
		vlNewNBestItems.append(CombinedNBestItem(vItem.trans, vItem.memidxs, vItem.degree, vItem.sysscores, vItem.combscores, {}))
	
	return vlNewNBestItems
	

def sortByCombScore(pMergedNBestItems, pCombScoreType, pSortDir):
	'''
	Sorts the input n-best items by combination score
	'''
	
	if pSortDir.startswith('d'):
		vReverseSort = True
	else:
		vReverseSort = False
	
	vlSorted = []
	vlSorted = sorted(pMergedNBestItems, key=lambda item: item.combscores[pCombScoreType], reverse=vReverseSort)
	
	return vlSorted
	

def multiplyScores(plScores):
	'''
	compute the multiplication of scores in the input list
	'''
	vResult = 1
	for vScore in plScores:
		if vScore != 0:
			vResult *= vScore
	
	if vResult > 0:
		return -vResult
	else:
		return vResult
	

def sumScores(plScores):
	'''
	compute the sum of scores in the input list
	'''
	vResult = 0
	for vScore in plScores:
		vResult += vScore
	
	return vResult
	

def avgScores(plScores):
	'''
	compute the average of scores in the input list
	
	Note: since the scores of negative (log-probabilties), zero scores,
	which mean a system has not produced the translation, will lead to a 
	higher score. Therefore, the less systems agree on a tranlsation, the
	higher the score will be. Since this is not a desirable effect, we 
	exclude them in averaging.
	'''
	vSum = 0
	for vScore in plScores:
		vSum += vScore
	
	vNum = len([score for score in plScores if score != 0])
	
	return vSum/vNum
	

def sortByExtScore(plNBestItems, pExtMetric, pSortDir):
	'''
	Sorts the input n-best items by external score
	'''
	
	if pSortDir.startswith('d'):
		vReverseSort = True
	else:
		vReverseSort = False
	
	vlSorted = []
	vlSorted = sorted(plNBestItems, key=lambda item: item.extscores[pExtMetric], reverse=vReverseSort)
	
	return vlSorted
	
	

