#! /usr/bin/python

## This module defines functions and data structures to process the output
## of multiple MT systems for comparison and combination.
##
## Current version 1.3
## - getBestScoreIdx() is added
## - formatSysSelection() is added
## - combineOutputs() is added
##
## Version 1.2
## - extractAvgLenRatio() is added
##
## Version: 1.1
## - Writing juxtaposition output is added. It was originally done in 
##   juxtapose-mutli-mt-out.py, but now is done by writeJuxtOut() function
##   and its satellite functions here.
## - loadJuxta() is added.
##   To Do: Convert format to XML.
## - extractSentenceFeaturesJux() is added.
## - extractAvgScorePerSrcLen() is added. 

from collections import namedtuple
from decimal import *
from random import choice
import sys


##----------------------------------------------------------------------
## Types

# data structure for systems and their output files
SysOutput = namedtuple('SysOutput', 'sysName outputFile scoreFile')

# data structure for systems output sentences with their scores
SysOutSntc = namedtuple('SysOutSntc', 'sysName outSntc score')

## data structure for juxtaposed output
## sntcNo: number of the sentence in the document
## src: the source sentence
## ref: reference sentence
## outputs: systems outputs (of SysOutSntc type)
JuxtOutput = namedtuple('JuxtOutput', 'sntcNo src ref outputs')

## data structure for sentence level features extraced from juxstaposition
## sntcNo: number of the sentence in the document
## srcLen: source sentence length
## sysScores: list of system scores
## lenRatio: the ratio of target to source length for reference and all systems
JuxtSntcFeature = namedtuple('JuxtSntcFeature', 'sntcNo srcLen lenRatios sysScores')

##----------------------------------------------------------------------
## Juxtuposes outputs of multiple MT systems together with the source and
## reference, sentence by sentence, for comparsion purposes. In the output,
## the first line is the source sentence, the second line is the reference
## sentence, and the following lines are the system outputs for the sentence.
## Each system output is accomponied by its MT evaluation score (e.g. TER).
def juxtaposeOutput(pSrcFileName, pRefFileName, plSysNames, plSysOutFileNames, plSysOutScoreFileNames, pflgPrint):

 # opening source file
 try:
  vfSrc  = open(pSrcFileName, 'r')
 except IOError:
  sys.exit('Can\'t open source file: ' + pSrcFileName)

 # opening reference file
 try:
  vfRef  = open(pRefFileName, 'r')
 except IOError:
  sys.exit('Can\'t open reference file: ' + pRefFileName)
  
 # opening system output files
 vlSysOutputs = []
 for vSys, vOutput, vScores in zip(plSysNames, plSysOutFileNames, plSysOutScoreFileNames):
  try:
   vlSysOutputs.append(SysOutput(vSys, open(vOutput, 'r'), open(vScores, 'r')))
  except IOError:
   sys.exit('Can\'t open output or score file: ' + vOutput + ', ' + vScores)
  
 
 ## creating juxtaposed output sentence by sentence
 
 vlJuxtOut = []
 vSntcCntr = 1
 for vSrcSntc, vRefSntc in zip(vfSrc, vfRef):
  # extracting system outputs
  vlSysOutSntcs = []
  for vOutput in vlSysOutputs:
   # creating sentence ouput for each system
   (vlSysOutSntcs.append(SysOutSntc(vOutput.sysName, 
                                    vOutput.outputFile.readline().strip(' \n'),
                                    Decimal(vOutput.scoreFile.readline().strip(' \n')))))
  
  # creating juxtaposed output for sentence 
  (vlJuxtOut.append(JuxtOutput(vSntcCntr,
                               vSrcSntc.strip(' \n'),
                               vRefSntc.strip(' \n'),
                                                   vlSysOutSntcs)))
  # printing the juxtaposed output for setences if requested
  if pflgPrint:
   print '[' + str(vSntcCntr) + '][source    ][' + vlJuxtOut[-1].src + ']'
   print '[' + str(vSntcCntr) + '][reference ][' + vlJuxtOut[-1].ref + ']'
   for vOutput in vlJuxtOut[-1].outputs:
    print '[' + str(vSntcCntr) + '][' + vOutput.sysName.ljust(10, ' ') + '][' + vOutput.outSntc + '][' + str(vOutput.score) + ']'
   print                      # empty line
  
  vSntcCntr += 1  
  
 return vlJuxtOut



##----------------------------------------------------------------------
## Writes sentence juxtaposition into the specified output file
def writeSntcJuxta(pfOut, pSntcJuxta, pTabularize):
 if pTabularize:
  pfOut.write(str(pSntcJuxta.sntcNo) + '\tsource    \t' + pSntcJuxta.src + '\n')
  pfOut.write(str(pSntcJuxta.sntcNo) + '\treference \t' + pSntcJuxta.ref + '\n')
  for vOutput in pSntcJuxta.outputs:
   pfOut.write(str(pSntcJuxta.sntcNo) + '\t' + vOutput.sysName.ljust(10, ' ') + '\t' + vOutput.outSntc + '\t' + str(vOutput.score) + '\n')
 else:
  pfOut.write('[' + str(pSntcJuxta.sntcNo) + '][source    ][' + pSntcJuxta.src + ']\n')
  pfOut.write('[' + str(pSntcJuxta.sntcNo) + '][reference ][' + pSntcJuxta.ref + ']\n')
  for vOutput in pSntcJuxta.outputs:
   pfOut.write('[' + str(pSntcJuxta.sntcNo) + '][' + vOutput.sysName.ljust(10, ' ') + '][' + vOutput.outSntc + '][' + str(vOutput.score) + ']\n')
 
 pfOut.write('\n')                                          # empty line


##----------------------------------------------------------------------
## Determines whether there is a score tie between the specified system
## outputs (scores for all outputs are the same)
def isScoreTie(pSysOutputs):
 vCurrentScore = pSysOutputs[0].score
 for vOutput in pSysOutputs[1:]:
  if vCurrentScore != vOutput.score:
   return False
 
 return True
 

##----------------------------------------------------------------------
## Determines whether there is a real (output sentence) tie between the 
## specified system outputs (output sentences by all systems are the same)
## Note that it's assumed that real ties imply a score tie. In other words,
## we assume that the scorer assignes equal scores to exactly same outputs.
def isRealTie(pSysOutputs):
 vCurrentSntc = pSysOutputs[0].outSntc
 for vOutput in pSysOutputs[1:]:
  if vCurrentSntc != vOutput.outSntc:
   return False
 
 return True

 
##----------------------------------------------------------------------
## Finds the best MT score in the specified system output list
## The best score is the maximum of scores if pSortDir is descending and
## minimum if it is ascending (like TER).
def getBestScore(pSysOutputs, pSortDir):
 vBestScore = pSysOutputs[0].score

 for vOutput in pSysOutputs[1:]:
  if pSortDir.startswith('a'):
   if vOutput.score < vBestScore:
    vBestScore = vOutput.score
  elif pSortDir.startswith('d'):
   if vOutput.score > vBestScore:
    vBestScore = vOutput.score
 
 return vBestScore
 
 
##----------------------------------------------------------------------
## Finds whether the is only one best-performing system by looking at the
## outputs and comparing the scores to the specified best-scores.
def isSoloWin(pSysOutputs, pBestScore):
 vBestCnt = 0 
 for vOutput in pSysOutputs:
  if vOutput.score == pBestScore:
   vBestCnt += 1
 
 if vBestCnt == 1:
  return True
 elif vBestCnt > 1:
  return False
 else:
  sys.exit("[isSoloWin()] The best score is not found in the output scores!")
  


##----------------------------------------------------------------------
## Writes specified juxtapositions extracted from pJuxta into output files.
## While full juxtaposition is default, the followings are optional:
## - Score tie split (ties and differences)
## - Real tie split (ties and differences)
## - Any-wins for each system (where the system is winner in any case)
## - Solo-wins for each system (where the system is the only winner)
## It also collect statistics along the way and print them into standard 
## output.
def writeJuxtOut(pJuxta, plSysNames, pFullOutFileName, pOutScoreTieSplit, pScoreTieFileName, pScoreDifFileName,
                 pOutRealTieSplit, pRealTieFileName, pRealDifFileName, pScoreSortDir,
                 pOutWins, plAnyWinFileNames, plSoloWinFileNames,
                 pflgTabularize, pflgPrintStats):
 # creating full output file
 try:
  vfFullOut  = open(pFullOutFileName, 'w')
 except IOError:
  sys.exit('Can\'t create full juxtaposed output file: ' + pFullOutFileName)
 
 # creating score-tie split files if requested
 if pOutScoreTieSplit:
  try:
   vfScoreTieOut = open(pScoreTieFileName, 'w')
  except IOError:
   sys.exit('Can\'t create score-tie juxtaposition file: ' + pScoreTieFileName)
  try:
   vfScoreDifOut = open(pScoreDifFileName, 'w')
  except IOError:
   sys.exit('Can\'t create score-difference juxtaposition file: ' + pScoreDifFileName)
  
 # creating real-tie split files if requested
 if pOutRealTieSplit:
  try:
   vfRealTieOut = open(pRealTieFileName, 'w')
  except IOError:
   sys.exit('Can\'t create real-tie juxtaposition file: ' + pRealTieFileName)
  try:
   vfRealDifOut = open(pRealDifFileName, 'w')
  except IOError:
   sys.exit('Can\'t create real-difference juxtaposition file: ' + pRealDifFileName)
  
 ## creating any and solo wins file for each individual system
 vlfAnySysWins = []
 vlfSoloSysWins = []
 if pOutWins:
  for vAnyWinFileName, vSoloWinFileName in zip(plAnyWinFileNames, plSoloWinFileNames):
   try:
    vlfAnySysWins.append(open(vAnyWinFileName, 'w'))
   except IOError:
    sys.exit('Can\'t create all-wins juxtaposition file: ' + vAnyWinFileName)
   try:
    vlfSoloSysWins.append(open(vSoloWinFileName, 'w'))
   except IOError:
    sys.exit('Can\'t create solo-wins juxtaposition file: ' + vSoloWinFileName)
  
 # initializing statistics
 vScoreTiesCnt = 0
 vRealTieCnt = 0
 vlAnySysWinsCnts = [0] * len(plSysNames)
 vlSoloSysWinsCnts = [0] * len(plSysNames)
 
 ## looping through juxtaposition to write the outputs and collect the
 ## statistics
 for vSntcJuxta in pJuxta:
  # writing full output
  writeSntcJuxta(vfFullOut, vSntcJuxta, pflgTabularize)
  
  vIsScoreTie = isScoreTie(vSntcJuxta.outputs)
  vIsRealTie = isRealTie(vSntcJuxta.outputs)
  
  # writing score tie or difference output
  if pOutScoreTieSplit:
   if vIsScoreTie:
    writeSntcJuxta(vfScoreTieOut, vSntcJuxta, pflgTabularize)
    vScoreTiesCnt += 1
   else:
    writeSntcJuxta(vfScoreDifOut, vSntcJuxta, pflgTabularize)

  # writing real tie or difference output
  if pOutRealTieSplit:
   if vIsRealTie:
    writeSntcJuxta(vfRealTieOut, vSntcJuxta, pflgTabularize)
    vRealTieCnt += 1
   else:
    ## Note that the real difference (in fact non-real-tie) is the case 
    ## where there is a score tie. That is only when scores are the same 
    ## difference in output sentences is considered as non-real-tie. 
    if vIsScoreTie:
     writeSntcJuxta(vfRealDifOut, vSntcJuxta, pflgTabularize)

  ## writing system wins outputs and collecting win statistics
  ## Note that more than one (but less than all) systems can be winner.
  if pOutWins != None:
   if not vIsScoreTie:
    vBestScore = getBestScore(vSntcJuxta.outputs, pScoreSortDir)
    for vSysOut, vfAnySysWin, vfSoloSysWin in zip(vSntcJuxta.outputs, vlfAnySysWins, vlfSoloSysWins):
     if vSysOut.score == vBestScore:
      writeSntcJuxta(vfAnySysWin, vSntcJuxta, pflgTabularize)
      vlAnySysWinsCnts[vSntcJuxta.outputs.index(vSysOut)] += 1
      # writing also into solo-win file, if this system is the only winner
      if isSoloWin(vSntcJuxta.outputs, vBestScore):
       writeSntcJuxta(vfSoloSysWin, vSntcJuxta, pflgTabularize)
       vlSoloSysWinsCnts[vSntcJuxta.outputs.index(vSysOut)] += 1
 
 
 # printing statistics
 if pflgPrintStats:
  print "Number of score ties: " + str(vScoreTiesCnt)
  print "Number of real ties: " + str(vRealTieCnt)
  for vSysName, vAnySysWinsCnt in zip(plSysNames, vlAnySysWinsCnts):
   print "Number of any " + vSysName + " wins: " + str(vAnySysWinsCnt)
  for vSysName, vSoloSysWinsCnt in zip(plSysNames, vlSoloSysWinsCnts):
   print "Number of solo " + vSysName + " wins: " + str(vSoloSysWinsCnt)
 
  
 # closing open files
 
 vfFullOut.close() 
 try:
  vfScoreTieOut.close()
 except NameError:
  pass
 try:
  vfScoreDifOut.close()
 except NameError:
  pass
 try:
  vfRealTieOut.close()
 except NameError:
  pass
 try:
  vfRealDifOut.close()
 except NameError:
  pass
 for vfAnySysWins in vlfAnySysWins:
  try:
   vfAnySysWins.close()
  except NameError:
   pass
 for vfSoloSysWins in vlfSoloSysWins:
  try:
   vfSoloSysWins.close()
  except NameError:
   pass



##----------------------------------------------------------------------
## Loads a juxstaposition from input file to a list of juxtaposition data
## structure type. The assumed format for input file is:
## [<sentence no>][source    ][<source sentence>]
## [<sentence no>][reference ][<reference sentence>]
## [<sentence no>][sys 1 name][sys 1 output][sys 1 score]
## ...
## [<sentence no>][sys n name][sys n output][sys n score]
## <empty line>
## ...
## Note: There must be an empty line at the end of the file, otherwise 
## the last sentence will be missing.
def loadJuxta(pJuxtaFileName):
 # opening juxtaposition file
 try:
  vfJuxta  = open(pJuxtaFileName, 'r')
 except IOError:
  sys.exit('Can\'t open juxtaposition file: ' + pJuxtaFileName)
 
 vlJuxta = []
 vlOutputs = []
 for line in vfJuxta:
  vLine = line.strip(' \n')
  if vLine == '':
   vlJuxta.append(JuxtOutput(vSntcNo, vSrc, vRef, vlOutputs))
   vlOutputs = []
  else:
   ## We don't split sentence based on [] to avoid mistakenly spliting 
   ## the sentences containing [].
   v1stRBraPos = vLine.find(']')
   v2ndLBraPos = vLine.find('[', 2)
   v2ndRBraPos = vLine.find(']', v2ndLBraPos + 1)
   v3rdLBraPos = vLine.find('[', v2ndRBraPos + 1)
   v3rdRBraPos = vLine.find(']', v3rdLBraPos + 1)
   
   vSntcNo = int(vLine[1:v1stRBraPos])
   vTag = vLine[v2ndLBraPos + 1: v2ndRBraPos].strip().lower()   
   if 'source' in vTag:
	vSrc = vLine[v3rdLBraPos + 1: v3rdRBraPos].strip()
   elif 'reference' in vTag:
    vRef = vLine[v3rdLBraPos + 1: v3rdRBraPos].strip()
   else:
    vOutSntc = vLine[v3rdLBraPos + 1: v3rdRBraPos].strip()
    ## In order to avoid conflict caused by [] as part of output sentence,
    ## we look for [] from the end of the line.
    v4thLBraPos = vLine.rfind('[')
    v4thRBraPos = vLine.rfind(']')
    vScore = Decimal(vLine[v4thLBraPos + 1: v4thRBraPos].strip())
    vlOutputs.append(SysOutSntc(vTag, vOutSntc, vScore))
   
 
 vfJuxta.close()
 return vlJuxta



##----------------------------------------------------------------------
## Extract features for each sentence in the specified juxtaposition. For
## the list of features, see the definition of JuxtSntcFeature named tuple
def extractSentenceFeaturesJux(pJuxta, pSrcLen, pLenRatio, pSysScores, pflgPrint, pfSntcFeature):
 
 vlFeatures = []
 
 for vSntcJuxta in pJuxta:
  vSrcLen = -1
  vlLenRatios = []
  vlSysScores = []

  if pSrcLen or pLenRatio:
   # NOTE: the correct way is to tokenize rather than split based on space
   vSrcLen = len(vSntcJuxta.src.split())
  
  if pLenRatio:
   vRefLen = len(vSntcJuxta.ref.split())
   vlLenRatios.append(Decimal(vRefLen) / vSrcLen)
   for vOutput in vSntcJuxta.outputs:
    vlLenRatios.append(Decimal(len(vOutput.outSntc.split())) / vSrcLen)
  
  if pSysScores:
   for vOutput in vSntcJuxta.outputs:
    vlSysScores.append(vOutput.score)
 
  vlFeatures.append(JuxtSntcFeature(vSntcJuxta.sntcNo,
                                    vSrcLen,
                                    vlLenRatios,
                                    vlSysScores))
  if pflgPrint:
   pfSntcFeature.write(str(vlFeatures[-1].sntcNo) + ',' + str(vlFeatures[-1].srcLen))
   for vRatio in vlFeatures[-1].lenRatios:
    pfSntcFeature.write(',' + str(vRatio))
   for vScore in vlFeatures[-1].sysScores:
    pfSntcFeature.write(',' + str(vScore))
   pfSntcFeature.write('\n')


 return vlFeatures
 
 
 
##----------------------------------------------------------------------
## Extract average score per source length for all systems from sentence-
## level feature list extracted by extractSentenceFeaturesJux().
## At the end, only length values with a minimum instance number of 
## pMinInstCnt are kept.
def extractAvgScorePerSrcLen(plFeatures, pMinInstCnt, pflgPrint, pfSntcFeature):
 ## For efficient data access, we create list with a length of maximum 
 ## length value in input feature list for storing new feature values.
 ## So, we can access each specific length value in new feature list via
 ## list index. E.g. to store data for length value of 5, we use 
 ## vlAvgSysScoresPerLen[5]. Note that the length value is still in the
 ## data structure, since the returning list optionally contains only 
 ## length values for which data exist (see pAllIntInRange).
 
 ## data structure to store statistics for each length value in the input
 ## feature list
 AvgSysScoresPerLen = namedtuple('AvgSysScoresPerLen', 'lenValue instanceCnt sysScoreAvgs')
 
 # finding minimum and maximum length value in input feature list
 vlLenValues = [feature.srcLen for feature in plFeatures]
 vMinLenValue = min(vlLenValues)
 vMaxLenValue = max(vlLenValues)
 vlLenValues = []
 
 #finding the number of systems
 vSysCount = len(plFeatures[0].sysScores)
 
 # initializing data structure for collecting sum scores
 vlSumSysScoresPerLen = [AvgSysScoresPerLen(0, 0, [0] * vSysCount)] * (vMaxLenValue + 1)
 
 # looping through sentence-level feature list to compute the feature
 for vFeature in plFeatures:
  vlNewSysScoreSums = []
  for (vNewSysScore, vOldSysScoreSum) in zip(vFeature.sysScores, vlSumSysScoresPerLen[vFeature.srcLen].sysScoreAvgs):
   vlNewSysScoreSums.append(vNewSysScore + vOldSysScoreSum)
  vNewSumSysScoresPerLen = AvgSysScoresPerLen(vFeature.srcLen, vlSumSysScoresPerLen[vFeature.srcLen].instanceCnt + 1, vlNewSysScoreSums)
  vlSumSysScoresPerLen[vFeature.srcLen] = vNewSumSysScoresPerLen
   
 ## looping through the collected feature values to compute the average 
 ## score of each system for each length value
 ## Note: also in this loop, based on the value of pAllIntInRange, it is 
 ##       decided whether the 0-instance length values should be kept or 
 ##       removed. The 0-instance length values from 0 to the minimum 
 ##       non-0-instance value are always ignored.
 vlAvgSysScoresPerLen = []
 for vItem in vlSumSysScoresPerLen[vMinLenValue:]:
  if (vItem.instanceCnt < pMinInstCnt):
   continue
  else:
   if pflgPrint:
    pfSntcFeature.write(str(vItem.lenValue) + ',' + str(vItem.instanceCnt))

   vlSysScoreAvgs = []
   for vSysScoreSum in vItem.sysScoreAvgs:
    if (vItem.instanceCnt == 0):
     vlSysScoreAvgs.append(0)
    else:
     vlSysScoreAvgs.append(vSysScoreSum / vItem.instanceCnt)
    if pflgPrint:
     pfSntcFeature.write(',' + str(vlSysScoreAvgs[-1]))
 
   vNewItem = AvgSysScoresPerLen(vItem.lenValue, vItem.instanceCnt, vlSysScoreAvgs)
   vlAvgSysScoresPerLen.append(vNewItem)
   if pflgPrint:
    pfSntcFeature.write('\n')
 
 return vlAvgSysScoresPerLen



##----------------------------------------------------------------------
## Extract the average ratio of target to source length for reference and
## all systems from sentence-level feature list extracted by 
## extractSentenceFeaturesJux().
def extractAvgLenRatio(plFeatures, plSysNames, pflgPrint, pfSntcFeature):
 vlLenRatioSums = [0] * (len(plFeatures[0].lenRatios))
 # looping through sentence-level feature list to sum length ratios
 for vFeature in plFeatures:
  vItrCntr = 0
  for vLenRatio in vFeature.lenRatios:
   vlLenRatioSums[vItrCntr] += vLenRatio
   vItrCntr += 1
   
 vlAvgLenRatios = []
 vFeatureCnt = len(plFeatures)
 # computing averages from collected sums#
 for vLenRatioSum in vlLenRatioSums:
  vlAvgLenRatios.append(Decimal(vLenRatioSum) / vFeatureCnt)
   
 
 # printing the feature if requested
 if pflgPrint:
  pfSntcFeature.write("Reference " + str(vlAvgLenRatios[0]) + '\n')
  for vSysName, vAvgLenRatio in zip(plSysNames, vlAvgLenRatios[1:]):
   pfSntcFeature.write(vSysName.ljust(10) + str(vAvgLenRatio) + '\n')
 
 
 return vlAvgLenRatios

 
 
 ##----------------------------------------------------------------------
## Returns the index of best score in the provided list of scores.
## For tiebreaking, when more than one scores are equally the best, it 
## optionally chooses either the first best in the order of score list
## or selects one randomly
def getBestScoreIdx(pScores, pTieBreak):
 vlBestScoreIdxs = []
 vBestScore = min(pScores)
 
 # we need to check all scores, since we need to know all best scores
 vScoreIdx = 0
 for vScore in pScores:
  if vScore == vBestScore:
   vlBestScoreIdxs.append(vScoreIdx)
  vScoreIdx += 1

 # now tiebreaking if needed
 if len(vlBestScoreIdxs) > 1:
  if pTieBreak == 'r':              
   return choice(vlBestScoreIdxs)
  else:
   return vlBestScoreIdxs[0]      ## choose the first in ascending order
                                  ## (equal to pTieBreak == 'a')
 else:
  return vlBestScoreIdxs[0]      



##----------------------------------------------------------------------
## Return the systems scores along with the selected system in a tab-
## separeted format.
def formatSysSelection(plScores, pSelSysName):
 vResult = ''
 for vScore in plScores:
  vResult += str(vScore) + '\t'
 
 vResult += pSelSysName
 
 return vResult
 


##----------------------------------------------------------------------
## Combines the outputs of several MT systems by selecting the best output
## based on sentence-by-sentence translation scores. It also outputs the
## scores combination along with the seleted system for each sentence.
def combineOutputs(plSysNames, plTransFileNames, plScoreFileNames, pTieBreak, pSelectionFileName):

 # creating selection output file
 try: 
  vfSelection = open(pSelectionFileName, 'w')
 except IOError:
  sys.exit("Can't create selection output file: " + pSelectionFileName)
 

 ## opening score files 
 ## Note that system names, translation file names and score file names
 ## provided as the script argument must follow the same order.
 vlfScoreFiles = []
 for vScoreFileName in plScoreFileNames:
  try:
   vlfScorFiles.append(open(vScoreFileName, 'r'))
  except IOError:
   sys.exit("Can't open score file: " + vScoreFileName)

 ## loading scores into list of tuples (by iterating through score files 
 ## in parallel)
 vlScores = []
 for vlScoreLines in zip(*vlfScoreFiles):
  vlScores.append(tuple([Decimal(vScore.strip(' \n')) for vScore in vlScoreLines]))


 ## opening translation files 
 ## Note that system names, translation file names and score file names
 ## provided as the script argument must follow the same order.
 vlfTransFiles = []
 for vTransFileName in plTransFileNames:
  try:
   vlfTransFiles.append(open(vTransFileName, 'r'))
  except IOError:
   sys.exit("Can't open translation output file: " + vTransFileName)

 
 # iterating through translations to chose the best 
 vSntcCntr = 0
 for vlTransSntcs in zip(*vlfTransFiles):
  vBestScoreIdx = getBestScoreIdx(vlScores[vSntcCntr], pTieBreak)
  # writing score select
  vfSelection.write(formatSysSelection(vlScores[vSntcCntr], plSysNames[vBestScoreIdx]) + '\n')
  print vlTransSntcs[vBestScoreIdx],
  vSntcCntr += 1 
  
 
 vfSelection.close()
 for vfTransFile in vlfTransFiles:
  vfTransFile.close()
 for vfScoreFile in vlfScoreFiles:
  vfScoreFile.close() 

 


