#! /usr/bin/python

## Note: this script should be merged with renrank-mt-nbest.py and mtnbest.py.

## This script reranks the n-best output of a MT system for a sentence 
## based on parser agreement. 
##
## It takes an evalb file which contains parse scores of one parser 
## against another one (in the absence of gold parses). It then reranks 
## the n-best based on and interpolation of one of the parse scores 
## (which is specified as option to script) and decoder score, each with
## an optional weight.
##
## Current version: 1.1
## 
## Added:
## - Parser score threshold is added as an option.
  

from collections import namedtuple
from operator import itemgetter
import sys, optparse, os, evalbproc, mtnbest



##----------------------------------------------------------------------
## Normalizes the input score using a linear map which is automatically 
## identifed using the bounds provided. In other words, it maps from
## [pLowerBound, pUpperbound] to [0, 100].
def normalizeScore(pScore, pLowerBound, pUpperBound):
 if pLowerBound > pUpperBound:
  sys.exit("Invalid range: [" + str(pLowerBound) + ', ' + str(pUpperBound) + ']')
 elif pLowerBound == pUpperBound:
  if pLowerBound == 0:
   return 0
  else:
   return 50
 else:
  return float(pScore - pLowerBound) / ((pUpperBound - pLowerBound) / 100.0)
 
 

##======================================================================
## main
def main(argv=None):
 if argv is None:
  argv = sys.argv
 
 parser = optparse.OptionParser(usage="%prog <N-BEST FILE> <EVALB SCORE FILE> <OUTPUT DIR> [options]" + 
                                      "\nReranks the n-best output of a MT system for a sentence based on parser agreement.", version="%prog 1.1")

 parser.add_option("-s", "--psortscore", help="parse score to be compared: (r)ecall, (p)recision or (f)-score)", dest="parseSortScore", type='str', default='f', action="store")
 parser.add_option("--psweight", help="parser score weight for interpolation (in the scale of 0 to 100)", dest="parseScoreWeight", default=50, action="store")
 parser.add_option("--psthreshold", help="parser score threshold (in the scale of 0 to 100)", dest="parseScoreThreshold", default=0, action="store")
 parser.add_option("-t", "--outtop", help="output the top-ranked sentence", dest="outTop", action="store_true")
 
 (opts, posArgs) = parser.parse_args()
 
 if len(posArgs) < 3:
  parser.error("At least 3 argument is required")
  
 # processing options
 
 # calculating score weights
 vParseScoreWeight = float(opts.parseScoreWeight)
 if vParseScoreWeight > 100:
  vParseScoreWeight = 100
 elif vParseScoreWeight < 0:
  vParseScoreWeight = 0
 vDecoderScoreWeight = 100 - vParseScoreWeight
 

 # loading sentences and decoder scores from n-best
 vNBests = mtnbest.extractFromSingleNBest(posArgs[0])
 
 
 # loading parse scores from evalb file
 vParseScores = evalbproc.extractSentenceScores(posArgs[1], False, False, False, False)
 
 
 # checking for evalb and n-best line length to be the same
 if len(vParseScores) != len(vNBests):
  sys.exit("Evalb and n-best files do not match in the number of lines")
 
 
 # data structure for reranked list
 RerankedItem = namedtuple('RerankedItem', 'no sentence score')
 vlRerankedItems = []
 
 
 # calculating the interpolated score for each n-best item
 
 for vNBestItem, vPScoreItem in zip(vNBests, vParseScores):
  # fetching the parse score requested for sorting
  if opts.parseSortScore == 'f':
   vParseScore = float(vPScoreItem.fscore)
  elif opts.parseSortScore == 'r':
   vParseScore = float(vPScoreItem.recall)
  elif opts.parseSortScore == 'p':
   vParseScore = float(vPScoreItem.precision)
  
  
  ## We don't involve the parser score (set it to 0) if it's below the 
  ## threshold. Note that, when decoder score weighs 0 and no item scores
  ## above threshold, all interpolated scores are 0. Since the sort 
  ## functoin reserves the original order for equivalent keys, the n-best
  ## list will remain intact and the decoder choice is finally selected.
  if vParseScore < float(opts.parseScoreThreshold):
   vParseScore = 0
  
  ## Since parser scores are in [0,100] and decoder scores are log 
  ## probabilities, the latter should be normalized to [0,100]. This is 
  ## done by a linear map which is automatically identified using the
  ## bounds of decoder scores. We assume that the n-best items are sorted
  ## by decoder score.
  vDecoderScore = normalizeScore(vNBestItem.score, vNBests[-1].score, vNBests[0].score)
  
  vIntScore = (vDecoderScoreWeight / 100.0 *  vDecoderScore + 
               vParseScoreWeight / 100.0 * float(vParseScore))
  
  vlRerankedItems.append(RerankedItem(vPScoreItem.no, vNBestItem.trans, vIntScore)) 


 ## sorting the scores
 ## Note that, when the n-best items are sorted and multiple items have
 ## the same key (interpolated score), i.e. a tie is encountered, their
 ## original order is preserved. Since the original order is based on 
 ## decoder score, it means that from the topmost items with the same 
 ## key, the one with the highest decoder score is chosen.
 vlRerankedItems.sort(key=itemgetter(2), reverse=True)
 

 # writing reranked output
 vRerankedFileName = posArgs[2] + '/' + os.path.basename(posArgs[0]) + '.re'
 try:
  vfReranked  = open(vRerankedFileName, 'w')
 except IOError:
  print 'Can\'t create reranked output file: ' + vRerankedFileName
  sys.exit(2)
 
 for vItem in vlRerankedItems:
  vfReranked.write(str(vItem.no) + " ||| " + vItem.trans + " ||| " + str(vItem.score) + '\n')


 # printing the top reranked sentence
 if opts.outTop:
  print vlRerankedItems[0].trans



##======================================================================
## calling main
if __name__ == "__main__":
 sys.exit(main())

