#! /usr/bin/python

"""	This script reranks the n-best lists of a set of sentences based on
	the selected measure. It generates a translation set based on top-
	ranked items for each sentence and/or the reranked n-best lists for
	each sentence.
	
	Given the n-best file for a set of sentences generated by a system, 
	it extracts the translations for each sentences and scores them using
	requested metric with regards to a given reference translation set. 
	Alternatively, it extract scores already prepared in various formats.
	
	Note: rerank-mt-nbest-pa.py script does the same thing with parser
	accuracy scores in evalb score files. That script should be ported
	into here as a new measure option.
	
	Version: 1.2
	- Sort direction is added								(22-Aug-2012)
	
	Version: 1.1
	- N-best input changed from per-sentence to per-system	(20-Aug-2012)
	
	Version: 1.0
	- TER scoring is implemented (tercom: Snover et al. (2006))
"""  

from decimal import *
import sys, optparse, mteval, mtnbest, os


##======================================================================
## main
def main(argv=None):
	if argv is None:
		argv = sys.argv
 
	parser = optparse.OptionParser(usage="%prog <N-BEST FILE> <REFERENCE FILE> <OUTPUT DIR> [options]" + 
										 "\nReranks the n-best lists of a set of sentences based on the selected measure.", version="%prog 1.1")

	parser.add_option("--rerank-meaure", help="measure used for reranking [default: ter]", dest="rerankMeasure", default="ter", action="store")
	parser.add_option("--output-types", help="list of types of outputs to be produced: reranked-translation (translation set based on top-ranked items for each sentence) [default], reranked-lists (reranked n-best per sentence)", dest="outTypes", default="reranked-translation", action="store")
	parser.add_option("--trans-out-name", help="output file name for selelected translsations", dest="transOutFilename", action="store")
	parser.add_option("--ter", help="path to tercom TER scorer java class", dest="scoreTER", action="store")
	parser.add_option("--tero", help="output type for tercom [default: ter]", dest="terOutType", default="ter", action="store")
	parser.add_option("--sort-dir", help="sort direction (a)scending,  (d)escending (each measure uses it's own default; e.g. ascending for TER; opposite drection can be used for computation of lowerbound)", dest="sortDir", action="store")
	parser.add_option("--del-score-files", help="delete score files", dest="delScoreFiles", action="store_true")
	
 
	(opts, posArgs) = parser.parse_args()
 
	if len(posArgs) < 3:
		parser.error("At least 3 argument is required!")
	
	# processing arguments
	 
	vNBestFilename = posArgs[0]
	vRefFilename = posArgs[1]
	vOutputPath = posArgs[2]
	
	# opening reference translation file
	
	try:
		vfRef  = open(vRefFilename)
	except IOError:
		sys.exit('Can\'t open reference file: ' + vRefFilename)
	
	# list of topmost-ranked translations (e.g. oracle)
	vlTopRerankedTrans = []
	
	
	# loading the entire n-best list 
	vlNBestItems = mtnbest.extractSentenceNBests(vNBestFilename.strip(), False)
	
	
	# iterating through sentences (based of reference sentences)
	vSentCntr = 0
	for vRefSent in vfRef:
		vRefSent = vRefSent.strip()
		
		vlSentNBestItems = vlNBestItems[vSentCntr]

		## extracting translations from n-best items and replicating 
		## reference sentences as many as n-best items
		vlTransSents = []
		vlRefSents = []
		for vItem in vlSentNBestItems:
			vlTransSents.append(vItem.trans)
			vlRefSents.append(vRefSent)
		
		# generating the score file name prefix including the path
		tmp, vPrefix = os.path.split(vNBestFilename)
		vOutFilePrefix = vOutputPath + '/' + vPrefix + '.' + str(vSentCntr)
		
		
		# (scoring and) reranking with the requested measure
		
		vlScores = []
		
		if opts.rerankMeasure.lower() == "ter":
			
			vReverseSort = False
			if opts.sortDir.startswith('d'):
				vReverseSort = True
			
			vDevnull = open(os.devnull, 'w')
			vlScores = mteval.scoreTER(vlTransSents, vlRefSents, vOutFilePrefix, opts.scoreTER, opts.terOutType, vDevnull)
			
			# reranking
			vlRerankedScores = sorted(zip(range(len(vlScores)), vlScores), key=lambda item: item[1], reverse=vReverseSort)
			
			# collecting top-ranked sentence
			vlTopRerankedTrans.append(vlSentNBestItems[vlRerankedScores[0][0]].trans)
			
			# writing reranked n-best to output file
			
			if "reranked-lists" in opts.outTypes:
				try:
					vfRerankedList = open(vOutFilePrefix + ".reranked", 'w')
				except IOError:
					sys.exit("Cannot create reranked n-best file: " + vOutFilePrefix + ".reranked")
				
				for vRerankedItem in vlRerankedScores:
					vfRerankedList.write(str(vRerankedItem[0]).ljust(5) + 
										 "TER: " + str(vRerankedItem[1]).rjust(10) + ' ' +
										 "SYS: " + str(vlSentNBestItems[vRerankedItem[0]].score.quantize(Decimal("0.0001"))).rjust(8) + ' ..... ' +
										 vlSentNBestItems[vRerankedItem[0]].trans)
					vfRerankedList.write('\n')
				
				vfRerankedList.close()
			
			# deleting generated score files
			if opts.delScoreFiles:
				vScoreFilename = vOutFilePrefix + '.' + opts.terOutType
				os.remove(vScoreFilename)
			
		
		vSentCntr += 1
		
	# writing reranked n-best to output file
	
	if "reranked-translation" in opts.outTypes:
		try:
			vfRerankedTrans = open(opts.transOutFilename, 'w')
		except IOError:
			sys.exit("Cannot create selected translation output file: " + opts.transOutFilename)
		
		for vTrans in vlTopRerankedTrans:
			vfRerankedTrans.write(vTrans + '\n')
		
		vfRerankedTrans.close()
		
	vfRef.close()
	
	
 
##======================================================================
## calling main
if __name__ == "__main__":
 sys.exit(main())

