﻿#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""	
	This script masks and cleans input text line-by-line. Mask patterns
	are optional.
	
	NOTE: This script is tuned to Symantec translation memory and forum 
		  text but may work well for other text as well.
	NOTE: until version 1.0, the name of the script was mask.py.
	NOTE: The script support a few languages. While most of the regular 
	      expressions are identical across languages, some are built 
	      multilingual and some are built language-specific (different
	      version for each language).
	
	Version: 1.6											(07-03-2013)
	- --rem-cr is added to remove carriage return.
	
	Version: 1.5											(29-01-2013)
	- Output file naming is changed.
	- Bracket and @ masks were removed since they are not cleaning masks. 
	- --tag is split into --markup and --symctag. The former is revised
	  to completely remove tags instead of masking. The latter fixes 
	  Symantec special tags like menu item wrapper.
	  
	Version: 1.4											(20-12-2012)
	- Cleanings and maskings are moved to nlp.py. Now they are called as
	  functions.
	
	Version: 1.3											(17-12-2012)
	- Garbage cleaning is improved.
	- WFP is improved.
	
	Version: 1.2											(21-09-2012)
	- Multiprocessing support is added.
	- Language support is added.
	- Menu items and HTML/XML tags are merged under --tag and are both 
	  now stripped out instead of masking with -TAG-.
	- Some trivial changes to some regular expressions
	- Date/time masking is considerably shrinked. Based on Treebanks,
	  it is mostly left for parsing and translation. Only formats like
	  12-10-2012 which are fully numbers with separators are masked.
	- --good3 is added.
	- To handle corpora with empty lines, reading from stdin is removed.
	- Instead of good sentences, their indexes (1-based) are output. 
	  Outputing themselves is now optional. Non-good ones are not output.
	- Optionally, a parallel corpus is checked for exact line matchs. Indexes
	  of such lines (1-based) are output.
	- Language detection is added.
	- Statistics are output to a file.
	- And many other changes ..
	
	Version: 1.1
	- Input can optionally be provided in a file.
	- --var, --bslash, --pholder are extended.
	- --longstr, --mergemask, --nogood, --cb, --at are added.
	- garbage cleaning and --lbegin were unified into --garbage.
	- --useless, --number are updated.
	- --good is extended to good1 and good2 (and associated options).
	- --tokenize, --lang is added.
	
"""

from collections import namedtuple
from datetime import *
from decimal import *
from Queue import *
import sys, optparse, re, nlp, multiprocessing, math, time


'''
	data structure for cleaning statistics and properties of a line
'''
CleanStat = namedtuple('CleanStat', 'unmaskSymcCount unmaskSymcTime garbageCount garbageTime hSpecialCount hSpecialTime sCharCount sCharTime crCount crTime sStrCount sStrTime memAddrCount memAddrTime regKeyCount regKeyTime pHolderCount pHolderTime lNumCount lNumTime tagCount tagTime varCount varTime bSlashCount bSlashTime emailCount emailTime urlCount urlTime wPathCount wPathTime uPathCount uPathTime ipvCount ipvTime dtCount dtTime numCount numTime lStrCount lStrTime aMaskCount aMaskTime isForeign foreignTime isUseless uselessTime isGood1 good1Time isGood2 good2Time isGood3 good3Time tokTime totalTime')

def clean(pInput, pLang, pOpts):
	'''
	Cleans text line.
	
	NOTE: in cases where the mask is surrounded by a space, the intention
	      is to treat fusion alongside. E.g. follow:http://. -> follow -URL- .
	      Since this adds an unwanted space for non-fused cases, a tokenization
	      or cleaning of trailing spaces is advised.
	
	pOpts: optparse-based command line options.
	'''
	
	vUnmaskSymcCount = 0
	vUnmaskSymcTime = 0
	vGarbageCount = 0
	vGarbageTime = 0
	vHSpecialCount = 0
	vHSpecialTime = 0
	vSCharCount = 0
	vSCharTime = 0
	vCRCount = 0
	vCRTime = 0
	vSStrCount = 0
	vSStrTime = 0
	vMemAddrCount = 0
	vMemAddrTime = 0
	vRegKeyCount = 0
	vRegKeyTime = 0
	vPHolderCount = 0
	vPHolderTime = 0
	vLNumCount = 0
	vLNumTime = 0
	vTagCount = 0
	vTagTime = 0
	vVarCount = 0
	vVarTime = 0
	vBSlashCount = 0
	vBSlashTime = 0
	vEmailCount = 0
	vEmailTime = 0
	vURLCount = 0
	vURLTime = 0
	vWPathCount = 0
	vWPathTime = 0
	vUPathCount = 0
	vUPathTime = 0
	vIPVCount = 0
	vIPVTime = 0
	vDTCount = 0
	vDTTime = 0
	vNumCount = 0
	vNumTime = 0
	vLStrCount = 0
	vLStrTime = 0
	vAMaskCount = 0
	vAMaskTime = 0
	vIsForeign = False
	vForeignTime = 0
	vIsUseless = False
	vUselessTime = 0
	vIsGood1 = False
	vGood1Time = 0
	vIsGood2 = False
	vGood2Time = 0
	vIsGood3 = False
	vGood3Time = 0
	vTokTime = 0
	vTotalTime = 0
	
	vInput = pInput.decode("utf-8")
	
	vStart = time.clock()
	
	## At the very first external masks present in the input are unmasked.
	## NOTE: this patterns are specific to Symantec data.
	## NOTE: there are some other already-masked patterns which are
	##       treated later in their specific way. For example &ProductName;
	##       is treated in variable masking.
	if pOpts.unmaskSymc:
		vNow = time.clock()
		# V_BAR
		vPattern = re.compile(r"V_BAR", re.IGNORECASE)
		vInput, vNum = vPattern.subn(r"/", vInput)
		vUnmaskSymcCount += vNum
		vUnmaskSymcTime += time.clock() - vNow
	
	# decoding HTML special characters
	if pOpts.decodeHTMLSpecial:
		vNow = time.clock()
		vInput, vHSpecialCount = nlp.decodeHTMLSpecial(vInput)
		vHSpecialTime += time.clock() - vNow
	
	
	# replacing special characters
	if pOpts.replaceSpecial:
		vNow = time.clock()
		vInput, vSCharCount = nlp.replaceSpecialChr(vInput)
		vSCharTime += time.clock() - vNow
	
	# removing carriage-returns
	if pOpts.remCR:
		vNow = time.clock()
		vInput, vCRCount = nlp.removeCR(vInput)
		vCRTime += time.clock() - vNow
	
	# triming spaces
	vInput = vInput.strip(" ")
	
	
	# cleaning known garbages
	if pOpts.cleanGarbage:
		vNow = time.clock()
		vInput, vGarbageCount = nlp.cleanGarbage(vInput)
		vGarbageTime += time.clock() - vNow
	
	# masking special strings
	if pOpts.maskSpecialStr:
		vNow = time.clock()
		vInput, vSStrCount = nlp.maskSpecialStr(vInput)
		vSStrTime += time.clock() - vNow
	
	# masking memory addresses
	if pOpts.maskMemAddr:
		vNow = time.clock()
		vInput, vMemAddrCount = nlp.maskMemAddr(vInput)
		vMemAddrTime += time.clock() - vNow
	
	# masking windows registry keys
	if pOpts.maskRegKey:
		vNow = time.clock()
		vInput, vRegKeyCount = nlp.maskRegKey(vInput)
		vRegKeyTime += time.clock() - vNow
	
	# masking placeholders: \{...\}
	# Note: mask after windows registry keys, before windows file path
	if pOpts.maskPHolder:
		vNow = time.clock()
		vInput, vPHolderCount = nlp.maskPHolder(vInput)
		vPHolderTime += time.clock() - vNow
	
	# masking list numbers
	# Note: mask before angle-bracket-related patterns
	if pOpts.maskListNum:
		vNow = time.clock()
		vInput, vLNumCount = nlp.maskListNum(vInput)
		vLNumTime += time.clock() - vNow
	
	# Removing HTML/XML tags
	# Note: mask before variable
	if pOpts.markup:
		vNow = time.clock()
		vInput, vTagCount = nlp.remHTMLTag(vInput)
		vTagTime += time.clock() - vNow
	
	# Removing/replacing special tags seen in Symantec TM
	# Note: mask before variable
	## For example, in Click  < >Reports < >, the menu item will be
	## replaced with "Reports". If needed, the quotation marks will 
	## be mask later. In fact, this is not masking but editing.
	if pOpts.symcTag:
		vNow = time.clock()
		vInput, vTagCount = nlp.remSYMCTag(vInput)
		vTagTime += time.clock() - vNow
	
	# masking variables
	# Note: mask after menu items above
	if pOpts.maskVar:
		vNow = time.clock()
		vInput, vVarCount = nlp.maskVar(vInput)
		vVarTime += time.clock() - vNow
	
	# replacing \n \r \t with whitespace
	# Note: mask before windows file path
	if pOpts.maskBSlash:
		vNow = time.clock()
		vInput, vBSlashCount = nlp.maskBSlash(vInput)
		vBSlashTime += time.clock() - vNow
	 
	# masking Email
	# Note: Email should be masked before URL
	if pOpts.maskEmail:
		vNow = time.clock()
		vInput, vEmailCount = nlp.maskEmail(vInput)
		vEmailTime += time.clock() - vNow
	
	# masking URL
	if pOpts.maskURL:
		vNow = time.clock()
		vInput, vURLCount = nlp.maskURL(vInput)
		vURLTime += time.clock() - vNow
	
	# masking windows file paths
	if pOpts.maskWPath:
		vNow = time.clock()
		vInput, vWPathCount = nlp.maskWPath(vInput)
		vWPathTime += time.clock() - vNow
	
	# masking Unix-like file paths
	if pOpts.maskUPath:
		vNow = time.clock()
		vInput, vUPathCount = nlp.maskUPath(vInput)
		vUPathTime += time.clock() - vNow
	
	# masking IP addresses and version number
	## Note: Since it is not easy to distinguish between IP and version,
	## they are masked together.
	if pOpts.maskIPV:
		vNow = time.clock()
		vInput, vIPVCount = nlp.maskIPV(vInput, pLang)
		vIPVTime += time.clock() - vNow
	
	# masking date/times
	## NOTE: Date/time masking is considerably shrinked. Based on Treebanks,
	##       it is mostly left for parsing and translation. Only formats
	##       like 12-10-2012 which are fully numbers with separators are
	##        masked.
	## NOTE: Date and time are masked separately. They'll be merged later,
	##       when all neighbouring similar masked are merged in a post-
	##       processing run.
	## NOTE: Some regex requires numbers to be masked afterwards. E.g.
	##       they look for 4 digits while number masks are 3 (999).
	if pOpts.maskDateTime:
		vNow = time.clock()
		vInput, vDTCount = nlp.maskDateTime(vInput)
		vDTTime += time.clock() - vNow
	
	# masking numbers
	## Note: Numbers are masked with a fixed number pattern like 999
	if pOpts.maskNum:
		vNow = time.clock()
		vInput, vNumCount = nlp.maskNum(vInput, pLang)
		vNumTime += time.clock() - vNow
	
	# masking long strings
	## Note: maximum string length for non-long string is optional.
	## However, there is a default set into pOpts.maskLongStr when
	## options are processed.
	if pOpts.maskLongStr != None:
		vNow = time.clock()
		vInput, vLStrCount = nlp.maskLongStr(vInput, pLang, int(pOpts.maskLongStr))
		vLStrTime += time.clock() - vNow
	
	
	# merging adjacent non-identical similar masks
	# Note: do before masking special characters
	if pOpts.mrgMasks:
		vNow = time.clock()
		vInput, vAMaskCount = nlp.mrgMasks(vInput)
		vAMaskTime += time.clock() - vNow
	
	
	# detecting language
	## Due to unreliable behaviour obsereved by both tools used here
	## for language detection, the decision is made based on the 
	## agreement between the two: if one of the guesses is not a 
	## foreign language, the line is considered not foreign.
	if pOpts.cleanForeign:
		vlTokens = vInput.split()
		if len(vlTokens) > pOpts.foreignCutoff:
			vNow = time.clock()
			# detecting language with CCLD
			vCCLD = nlp.ccldDetectSentLang(vInput)
			# detecting language with langid
			vLID = nlp.lidDetectSentLang(vInput)
			# comparison
			if pLang not in (vCCLD.topLangCode,  vLID.langCode):
				vIsForeign = True
			vForeignTime += time.clock() - vNow
	
	# checking for uselessness: not having a non-mask token with at least
	## two contiguous alphabet character
	if not vIsForeign and pOpts.cleanUseless:
		vNow = time.clock()
		vIsUseless = nlp.isUseless(vInput)
		vUselessTime += time.clock() - vNow
	
	
	# tokenization
	## NOTE: eventhough some masks include tokenization, since they may 
	##       be optionally ignored while tokenization is requested, they
	##       are recosidered here again.
	## NOTE: tokenization take place before seleting good sentences.
	if not vIsUseless and not vIsForeign and pOpts.tokenize:
		vNow = time.clock()
		vInput = nlp.tokenizeSegment(vInput, pLang, False)
		vTokTime += time.clock() - vNow
	
	# checking for good line
	## Note: a good line is defined in different ways identified by 
	##       good-1, good-2, ... (cf. nlp.py: isGoodSentence())
	
	if not vIsUseless and not vIsForeign and pOpts.good1:
		vNow = time.clock()
		vIsGood1 = nlp.isGoodSentence(vInput, 1)
		vGood1Time += time.clock() - vNow
	
	if not vIsUseless and not vIsForeign and pOpts.good2:
		vNow = time.clock()
		vIsGood2 = nlp.isGoodSentence(vInput, 2, pIsTokenized=pOpts.tokenize, pLang=pLang, pGood2Ratio=pOpts.good2Ratio)
		vGood2Time += time.clock() - vNow
	
	if not vIsUseless and not vIsForeign and pOpts.good3:
		vNow = time.clock()
		vIsGood3 = nlp.isGoodSentence(vInput, 3, pIsTokenized=pOpts.tokenize, pLang=pLang, pGood3RatioGLB=pOpts.good3RatioGLB, pGood3RatioFLB=pOpts.good3RatioFLB, pGood3RatioF=pOpts.good3RatioF)
		vGood3Time += time.clock() - vNow
	
	# triming spaces
	vInput = vInput.strip(" ")
	
	vTotalTime += time.clock() - vStart
	
	vCleanStat = CleanStat(vUnmaskSymcCount, vUnmaskSymcTime, vGarbageCount, vGarbageTime, vHSpecialCount, vHSpecialTime, vSCharCount, vSCharTime, vCRCount, vCRTime, vSStrCount, vSStrTime, vMemAddrCount, vMemAddrTime, vRegKeyCount, vRegKeyTime, vPHolderCount, vPHolderTime, vLNumCount, vLNumTime, vTagCount, vTagTime, vVarCount, vVarTime, vBSlashCount, vBSlashTime,vEmailCount, vEmailTime, vURLCount, vURLTime, vWPathCount, vWPathTime, vUPathCount, vUPathTime, vIPVCount, vIPVTime, vDTCount, vDTTime, vNumCount, vNumTime, vLStrCount, vLStrTime, vAMaskCount, vAMaskTime, vIsForeign, vForeignTime, vIsUseless, vUselessTime, vIsGood1, vGood1Time, vIsGood2, vGood2Time, vIsGood3, vGood3Time, vTokTime, vTotalTime)
	
	return vInput, vCleanStat
	

def worker(pChunkNo, pChunkSize, plInputChunk, pLang, pOpts, pqOutput):
	'''
	worker process for processing the specified chunk of the input
	'''
	
	vlOutput = []
	vlCleanStat = []
	vlForeignIdxs = []
	vForeignTime = 0
	for vLineCntr, vInput in enumerate(plInputChunk, start=1):
		# cleaning takes place here
		vOutput, vCleanStat = clean(vInput, pLang, pOpts)
		
		vlOutput.append(vOutput)
		vlCleanStat.append(vCleanStat)
		
		if pOpts.outProgress:
			if vLineCntr % (100000 / int(pOpts.procNum)) == 0:
				sys.stdout.write('!')
				sys.stdout.flush()
			elif vLineCntr % (10000 / int(pOpts.procNum)) == 0:
				sys.stdout.write('.')
				sys.stdout.flush()
	
	pqOutput.put((pChunkNo, vlOutput, vlCleanStat)) 
	


##======================================================================
## main
def main(argv=None):
	if argv is None:
		arv = sys.argv
	
	print "Started at: %s" % datetime.now().strftime("%Y-%m-%d %H:%M:%S")
	
	parser = optparse.OptionParser(usage="%prog <CORPUS> [OPTIONS]" + 
										 "\nMasks and cleans input text line-by-line. Mask patterns are optional.", version="%prog 1.5")

	parser.add_option("--procnum", help="number of processes for multiprocessing", dest="procNum", default=1, type="int", action="store")
	parser.add_option("--outprog", help="output progress indicator", dest="outProgress", action="store_true")
	parser.add_option("--lang", help="language of the corpus", dest="lang", default="en", action="store")
	parser.add_option("--outfile", help="name of output file", dest="outFilename", action="store")
	parser.add_option("--unmask-symc", help="unmask patterns already masked in Symantec text by external tools (e.g. V_BAR for /)", dest="unmaskSymc", action="store_true")
	parser.add_option("-a", "--all", help="mask every implemented pattern except @ (masking may include removing)", dest="maskAll", action="store_true")
	parser.add_option("-g", "--garbage", help="clean known garbages", dest="cleanGarbage", action="store_true")
	parser.add_option("-c", "--htmlspecial", help="mask HTML special character(e.g. &amp)", dest="decodeHTMLSpecial", action="store_true")
	parser.add_option("-k", "--specialchr", help="replaces special character (e.g. m-dashes)", dest="replaceSpecial", action="store_true")
	parser.add_option("-m", "--rem-cr", help="remove carriage return", dest="remCR", action="store_true")
	parser.add_option("-s", "--specialstr", help="mask special strings (e.g. C4:8A:1D:)", dest="maskSpecialStr", action="store_true")
	parser.add_option("-o", "--memaddr", help="mask memory addresses", dest="maskMemAddr", action="store_true")
	parser.add_option("-r", "--regkey", help="mask windows registry key", dest="maskRegKey", action="store_true")
	parser.add_option("-p", "--pholder", help="mask placeholders", dest="maskPHolder", action="store_true")
	parser.add_option("-l", "--listnum", help="mask list numbers", dest="maskListNum", action="store_true")
	parser.add_option("-t", "--markup", help="remove HTML/XML tags", dest="markup", action="store_true")
	parser.add_option("-y", "--symctag", help="Removes/replaces Symantec special tags", dest="symcTag", action="store_true")
	parser.add_option("-v", "--var", help="mask variables", dest="maskVar", action="store_true")
	parser.add_option("-b", "--bslash", help="replace \n \r \t with white space", dest="maskBSlash", action="store_true")
	parser.add_option("-e", "--email", help="mask Email", dest="maskEmail", action="store_true")
	parser.add_option("-u", "--url", help="mask URL", dest="maskURL", action="store_true")
	parser.add_option("-w", "--winpath", help="mask Windows file path", dest="maskWPath", action="store_true")
	parser.add_option("-x", "--unixpath", help="mask Unix file path", dest="maskUPath", action="store_true")
	parser.add_option("-i", "--ipver", help="mask IP address and version number", dest="maskIPV", action="store_true")
	parser.add_option("-d", "--datetime", help="mask date/time", dest="maskDateTime", action="store_true")
	parser.add_option("-n", "--number", help="mask numbers", dest="maskNum", action="store_true")
	parser.add_option("--longstr", help="mask string longer than the given value by this option", dest="maskLongStr", action="store")
	parser.add_option("--mergemask", help="merge similar neighbouring masks", dest="mrgMasks", action="store_true")
	parser.add_option("--tokenize", help="tokenize the lines", dest="tokenize", action="store_true")
	parser.add_option("--useless", help="mask useless lines", dest="cleanUseless", action="store_true")
	parser.add_option("--good1", help="select good-1 sentences: lines starting and ending in good letters", dest="good1", action="store_true")
	parser.add_option("--good2", help="select good-2 sentences: lines containing a certain ratio of good words", dest="good2", action="store_true")
	parser.add_option("--good2ratio", help="minimum ratio of good-2 tokens in a line making it a good sentence", dest="good2Ratio", default=70, type="int", action="store")
	parser.add_option("--good3", help="select good-3 sentences: good-3 and a relaxation fallback method (cf. nlp.py)", dest="good3", action="store_true")
	parser.add_option("--good3ratioglb", help="minimum ratio of good-3 tokens in a line making it a good sentence", dest="good3RatioGLB", default=70, type="int", action="store")
	parser.add_option("--good3ratioflb", help="minimum ratio of good-3 tokens in a line for relaxation fallback (cd. nlp.py)", dest="good3RatioFLB", default=50, type="int", action="store")
	parser.add_option("--good3ratiof", help="minimum ratio of fallback tokens in a line making it a good-3 sentence (cd. nlp.py)", dest="good3RatioF", default=80, type="int", action="store")
	parser.add_option("--outgood", help="output good sentences to a file", dest="outGood", action="store_true")
	parser.add_option("--foreign", help="detect and clean foreign lines and output their indexes", dest="cleanForeign", action="store_true")
	parser.add_option("--foreigncut", help="line length longer than which will be checked for foreign language and cleaned", dest="foreignCutoff", default=7, type="int", action="store")
	parser.add_option("--outforeign", help="output foreign lines to a file", dest="outForeign", action="store_true")
	parser.add_option("--paramatch", help="check the parallel corpus for line matches and clean and output match indexes (one-based)", dest="cleanParaMatch", action="store_true")
	parser.add_option("--paramatchcut", help="line length longer than which will be checked matche in parallel and cleaned; NOTE: one-word lines will be cleaned anyway if --paramatch is present", dest="paraMatchCutoff", default=5, type="int", action="store")
	parser.add_option("--paracorp", help="parallel corpus to check for line matches", dest="parallelCorpus", action="store")
	
	(opts, posArgs) = parser.parse_args()
	
	# processing options
	
	# identifying language
	if opts.lang.lower() in ["english", "en"]:
		vLang = "en"
	elif opts.lang.lower() in ["french", "fr"]:
		vLang = "fr"
	elif opts.lang.lower() in ["german", "de"]:
		vLang = "de"
	else:
		sys.stderr.write(opts.lang + " is not identified! English will be used.\n")
		vLang = "en"
	
	## overriding individual pattern options if --all is set 
	### remember to add new feature options here
	if opts.maskAll:
		opts.cleanGarbage = True
		opts.maskNum = True
		opts.decodeHTMLSpecial = True
		opts.maskSpecialStr = True
		opts.maskMemAddr = True
		opts.maskRegKey = True
		opts.maskPHolder = True
		opts.maskListNum = True
		opts.maskTag = True
		opts.maskVar = True
		opts.maskBSlash = True
		opts.maskEmail = True
		opts.maskURL = True
		opts.maskWPath = True
		opts.maskUPath = True
		opts.maskIPV = True
		opts.maskDateTime = True
		if opts.maskLongStr == None:
			if vLang in ("en", "fr"):
				opts.maskLongStr = 30
			elif vLang == "de":
				opts.maskLongStr = 40
	
	vProcNum = int(opts.procNum)
	
	if len(posArgs) < 1:
		parser.error("Input corpus is not provided!")
	
	# opening input corpus file
	vInCorpusFilename = posArgs[0]
	try:
		vfCorpus = open(vInCorpusFilename)
	except IOError:
		sys.exit("Cannot open input corpus file: " + vInCorpusFilename)
	
	# creating output file
	if opts.outFilename == None:
		vOutFilename = vInCorpusFilename + ".cl"
	else:
		vOutFilename = opts.outFilename
	
	try:
		vfOut = open(vOutFilename, 'w')
	except IOError:
		sys.exit("Cannot create output file: " + vOutFilename)
	
	# creating output files for good-1 indexes
	if opts.good1:
		vGood1IdxFilename = vOutFilename + '.g-1.idx'
		try:
			vfGood1Idx = open(vGood1IdxFilename, 'w')
		except IOError:
			sys.exit("Cannot create output file for good-1 indexes: " + vGood1IdxFilename)
		
		# creating output file for good-1 sentences
		if opts.outGood:
			vGood1Filename = vOutFilename + '.g-1'
			try:
				vfGood1 = open(vGood1Filename, 'w')
			except IOError:
				sys.exit("Cannot create output file for good-1 sentence: " + vGood1Filename)
	
	# creating output files for good-2 indexes
	if opts.good2:
		vGood2IdxFilename = vOutFilename + '.g-2.idx'
		try:
			vfGood2Idx = open(vGood2IdxFilename, 'w')
		except IOError:
			sys.exit("Cannot create output file for good-2 indexes: " + vGood2IdxFilename)
		
		# creating output file for good-2 sentences
		if opts.outGood:
			vGood2Filename = vOutFilename + '.g-2'
			try:
				vfGood2 = open(vGood2Filename, 'w')
			except IOError:
				sys.exit("Cannot create output file for good-2 sentence: " + vGood2Filename)
	
	# creating output files for good-3 indexes
	if opts.good3:
		vGood3IdxFilename = vOutFilename + '.g-3.idx'
		try:
			vfGood3Idx = open(vGood3IdxFilename, 'w')
		except IOError:
			sys.exit("Cannot create output file for good-3 indexes: " + vGood3IdxFilename)
		
		# creating output file for good-3 sentences
		if opts.outGood:
			vGood3Filename = vOutFilename + '.g-3'
			try:
				vfGood3 = open(vGood3Filename, 'w')
			except IOError:
				sys.exit("Cannot create output file for good-3 sentence: " + vGood3Filename)
	
	# creating output file for foreign lines indexes
	if opts.cleanForeign:
		vForeignIdxFilename = vOutFilename + '.foreign.idx'
		try:
			vfForeignIdx = open(vForeignIdxFilename, 'w')
		except IOError:
			sys.exit("Cannot create output file for foreign lines indexes: " + vForeignIdxFilename)
		
		# creating output file for good-3 sentences
		if opts.outForeign:
			vForeignFilename = vOutFilename + '.foreign'
			try:
				vfForeign = open(vForeignFilename, 'w')
			except IOError:
				sys.exit("Cannot create output file for foreign lines: " + vForeignFilename)
	
	## opening parallel corpus and creating output file for match indexes
	## if parallel matching is requested
	if opts.cleanParaMatch:
		if opts.parallelCorpus == None:
			sys.exit("Parallel corpus is not provided.")
		try:
			vfParCorp = open(opts.parallelCorpus)
		except IOError:
			sys.exit("Cannot open parallel corpus: " + opts.parallelCorpus)
		
		vfParMatchFilename = vOutFilename + ".parmatch.idx"
		try:
			vfParMatch = open(vfParMatchFilename, 'w')
		except IOError:
			sys.exit("Cannot create output file for parallel match indexes: " + vfParMatchFilename)
	
	# reading input strings from corpus file
	## While reading, the parallel corpus is checked for line matches if
	## requested.
	vStart = time.clock() 
	vlInput = []
	vlParaMatchs = []
	if opts.cleanParaMatch:
		sys.stdout.write("Reading input and parallel-matching ")
		for i, (vLine, vParLine) in enumerate(zip(vfCorpus, vfParCorp), start=1):
			vlTokens = vLine.split()
			## line longer than 1 and shorter than or equal to cut-off are
			## not checked for parallel match
			if 1 < len(vlTokens) <= opts.paraMatchCutoff:
				vlInput.append(vLine)
			else:
				if len(vlTokens) == 0 or ' '.join(vlTokens) != ' '.join(vParLine.split()):
					vlInput.append(vLine)
				else:
					vlInput.append('\n')
					vlParaMatchs.append(str(i))
			
			if opts.outProgress and i % (100000) == 0:
					sys.stdout.write('.')
					sys.stdout.flush()
		
		sys.stdout.write('!')
		vfParMatch.write(','.join(vlParaMatchs) + '\n')
		vfParCorp.close()
		vfParMatch.close()
	else:
		sys.stdout.write("Reading input ")
		for i, vLine in enumerate(vfCorpus, start=1):
			vlInput.append(vLine)
			if opts.outProgress and i % (100000) == 0:
				sys.stdout.write('.')
				sys.stdout.flush()
		
		sys.stdout.write('!')
	sys.stdout.write(" (" + str(time.clock() - vStart) + " sec)\n")
	
	vInputSize = len(vlInput)
	
	
	# processing 
	sys.stdout.write("Processing ")
	
	# queue for output with multiprocessing
	vqOut = multiprocessing.Queue()
	# size of each input chunk to be processed with each process
	vInChunkSize = int(math.ceil(vInputSize / float(opts.procNum)))
	# list of processes
	vlProcs = []
	# submitting processes
	for vProcNo in range(vProcNum):
		vProc = multiprocessing.Process(target=worker, args=(vProcNo, vInChunkSize, vlInput[vProcNo * vInChunkSize : (vProcNo + 1) * vInChunkSize], vLang, opts, vqOut))
		vlProcs.append(vProc)
		vProc.start()
	
	# collecting results
	vlOutput = [""]*vInputSize
	vlCleanStats = [0]*vInputSize
	for vProcNo in range(vProcNum):
		vResult = vqOut.get()
		## the result returned by the worker process is a tuple of:
		## - input chunk index (to retain the order of input)
		## - list of processed lines output in that chunk
		## - cleaning statistics for that chunk 
		vChunkNo = vResult[0]
		vlOutput[vChunkNo * vInChunkSize : (vChunkNo + 1) * vInChunkSize] = vResult[1]
		vlCleanStats[vChunkNo * vInChunkSize : (vChunkNo + 1) * vInChunkSize] = vResult[2]
	
	# waiting for all worker processes to finish
	for vProc in vlProcs:
		vProc.join()
	
	sys.stdout.write("! (" + str(sum([int(item.totalTime) for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	
	
	# writing output
	sys.stdout.write("Writing output ")
	vStart = time.clock() 
	vlForeignIdxs = []
	vlGood1Idxs = []
	vlGood2Idxs = []
	vlGood3Idxs = []
	for i, (vOutput, vCleanStat) in enumerate(zip(vlOutput, vlCleanStats), start=1):
		if isinstance(vOutput, unicode):
			vOutput = vOutput.encode("UTF-8")
		
		if (vCleanStat.isForeign and opts.cleanForeign) or (vCleanStat.isUseless and opts.cleanUseless):
			vfOut.write('\n')
		else:
			vfOut.write(vOutput)
		
		if opts.cleanForeign:
			if vCleanStat.isForeign:
				vlForeignIdxs.append(str(i))
				if opts.outForeign:
					vfForeign.write(vOutput)
		
		if opts.good1:
			if vCleanStat.isGood1:
				vlGood1Idxs.append(str(i))
				if opts.outGood:
					vfGood1.write(vOutput)
		
		if opts.good2:
			if vCleanStat.isGood2:
				vlGood2Idxs.append(str(i))
				if opts.outGood:
					vfGood2.write(vOutput)
		
		if opts.good3:
			if vCleanStat.isGood3:
				vlGood3Idxs.append(str(i))
				if opts.outGood:
					vfGood3.write(vOutput)
		
		if opts.outProgress and i % (100000) == 0:
			sys.stdout.write('.')
			sys.stdout.flush()
	
	if opts.cleanForeign:
		vfForeignIdx.write(','.join(vlForeignIdxs) + '\n')
	if opts.good1:
		vfGood1Idx.write(','.join(vlGood1Idxs) + '\n')
	if opts.good2:
		vfGood2Idx.write(','.join(vlGood2Idxs) + '\n')
	if opts.good3:
		vfGood3Idx.write(','.join(vlGood3Idxs) + '\n')
	sys.stdout.write("! (" + str(time.clock() - vStart) + " sec)\n")
	
	# writing statistics
	try:
		vStatFilename = vOutFilename + '.stat'
		vfStat = open(vStatFilename, 'w')
	except IOError:
		sys.exit("Cannot create statistics file: " + vStatFilename)
	
	vfStat.write(str(vInputSize) + " lines processed\n")
	vfStat.write(str(sum([item.unmaskSymcCount for item in vlCleanStats])) + " Symantec masks in " + str(len([item for item in vlCleanStats if item.unmaskSymcCount > 0])) + " lines unmasked (" + str(sum([item.unmaskSymcTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.garbageCount for item in vlCleanStats])) + " garbage in " + str(len([item for item in vlCleanStats if item.garbageCount > 0])) + " lines cleaned (" + str(sum([item.garbageTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.hSpecialCount for item in vlCleanStats])) + " HTML special characters in " + str(len([item for item in vlCleanStats if item.hSpecialCount > 0])) + " lines masked (" + str(sum([item.hSpecialTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.sStrCount for item in vlCleanStats])) + " special strings in " + str(len([item for item in vlCleanStats if item.sStrCount > 0])) + " lines masked (" + str(sum([item.sStrTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.sCharCount for item in vlCleanStats])) + " special characters in " + str(len([item for item in vlCleanStats if item.sCharCount > 0])) + " lines masked (" + str(sum([item.sCharTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.crCount for item in vlCleanStats])) + " carriage return in " + str(len([item for item in vlCleanStats if item.crCount > 0])) + " lines removed (" + str(sum([item.crTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.memAddrCount for item in vlCleanStats])) + " memory addresses in " + str(len([item for item in vlCleanStats if item.memAddrCount > 0])) + " lines masked (" + str(sum([item.memAddrTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.regKeyCount for item in vlCleanStats])) + " windows registry keys in " + str(len([item for item in vlCleanStats if item.regKeyCount > 0])) + " lines masked (" + str(sum([item.regKeyTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.pHolderCount for item in vlCleanStats])) + " place-holders in " + str(len([item for item in vlCleanStats if item.pHolderCount > 0])) + " lines masked (" + str(sum([item.pHolderTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.lNumCount for item in vlCleanStats])) + " list numbers in " + str(len([item for item in vlCleanStats if item.lNumCount > 0])) + " lines masked (" + str(sum([item.lNumTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.tagCount for item in vlCleanStats])) + " HTML/XML tags in " + str(len([item for item in vlCleanStats if item.tagCount > 0])) + " lines masked (" + str(sum([item.tagTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.varCount for item in vlCleanStats])) + " variables in " + str(len([item for item in vlCleanStats if item.varCount > 0])) + " lines masked (" + str(sum([item.varTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.bSlashCount for item in vlCleanStats])) + " newline/tab/.. in " + str(len([item for item in vlCleanStats if item.bSlashCount > 0])) + " lines masked (" + str(sum([item.bSlashTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.emailCount for item in vlCleanStats])) + " Emails in " + str(len([item for item in vlCleanStats if item.emailCount > 0])) + " lines masked (" + str(sum([item.emailTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.urlCount for item in vlCleanStats])) + " URLs in " + str(len([item for item in vlCleanStats if item.urlCount > 0])) + " lines masked (" + str(sum([item.urlTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.wPathCount for item in vlCleanStats])) + " Windows file paths in " + str(len([item for item in vlCleanStats if item.wPathCount > 0])) + " lines masked (" + str(sum([item.wPathTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.uPathCount for item in vlCleanStats])) + " Unix file paths in " + str(len([item for item in vlCleanStats if item.uPathCount > 0])) + " lines masked (" + str(sum([item.uPathTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.ipvCount for item in vlCleanStats])) + " IP addresses and version numbers in " + str(len([item for item in vlCleanStats if item.ipvCount > 0])) + " lines masked (" + str(sum([item.ipvTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.dtCount for item in vlCleanStats])) + " date/times in " + str(len([item for item in vlCleanStats if item.dtCount > 0])) + " lines masked (" + str(sum([item.dtTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.numCount for item in vlCleanStats])) + " numbers in " + str(len([item for item in vlCleanStats if item.numCount > 0])) + " lines masked (" + str(sum([item.numTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.lStrCount for item in vlCleanStats])) + " long strings in " + str(len([item for item in vlCleanStats if item.lStrCount > 0])) + " lines masked (" + str(sum([item.lStrTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(sum([item.aMaskCount for item in vlCleanStats])) + " adjacent masks in " + str(len([item for item in vlCleanStats if item.aMaskCount > 0])) + " lines masked (" + str(sum([item.aMaskTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
	vfStat.write(str(len([item for item in vlCleanStats if item.isUseless])) + " useless lines cleaned\n")
	vfStat.write("Tokenization took " + str(sum([item.tokTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec\n")
	if opts.good1 != None:
		vfStat.write(str(len(vlGood1Idxs)) + " good-1 sentences extracted (" + str(sum([item.good1Time for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
		vfGood1Idx.close()
		if opts.outGood:
			vfGood1.close()
	if opts.good2 != None:
		vfStat.write(str(len(vlGood2Idxs)) + " good-2 sentences extracted (" + str(sum([item.good2Time for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
		vfGood2Idx.close()
		if opts.outGood:
			vfGood2.close()
	if opts.good3 != None:
		vfStat.write(str(len(vlGood3Idxs)) + " good-3 sentences extracted (" + str(sum([item.good3Time for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
		vfGood3Idx.close()
		if opts.outGood:
			vfGood3.close()
	if opts.cleanForeign:
		vfStat.write(str(len(vlForeignIdxs)) + " foreign lines detected and cleaned (" + str(sum([item.foreignTime for item in vlCleanStats])) + '/' + str(vProcNum) + " sec)\n")
		vfForeignIdx.close()
		if opts.outForeign:
			vfForeign.close()
	if opts.cleanParaMatch:
		vfStat.write(str(len(vlParaMatchs)) + " identical parallel lines detected and cleaned (before cleaning process and after a split/join to strip)\n")
	if vInCorpusFilename != None:
		vfCorpus.close()
	vfOut.close()
	vfStat.close()
	
	print "Finished at: %s" % datetime.now().strftime("%Y-%m-%d %H:%M:%S")
	


##======================================================================
## calling main
if __name__ == "__main__":
	sys.exit(main())

