#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""	
	This module defines functions for tasks related to natural language 
	processing.
	
	
	Version 3.6												(06-Mar-2017)
	- A new argument is added to tokenizeSegment() to set the behaviour
	  of the unicode() function used inside it when facing errors.
	
	Version 3.5												(24-Oct-2017)
	- chNumericalizeCorpus() is added.
	
	Version 3.4												(14-Feb-2017)
	- numericalizeCorpus() is added.
	
	VERSION 3.3												(26-Sep-2016)
	- tokenizeSegment() is edited to handle tweets correctly by not 
	  tokenizing around @ and #. 
	
	Version 3.2												(19-Jun-2015)
	- pflgPrint is removed from tokenizeSegment().
	- pflgTokenizeFSlash is added to tokenizeSegment().
	
	Version 3.1												(10-Feb-2015)
	- tokenizeSegment() is updated.
	- maskEmoticons() and unmaskEmoticons() are added.
	
	Version 3.0												(17-Dec-2013)
	- calcOOV() is added.
	
	Version 2.9												(18-Dec-2013)
	- mapPenn2WordNet() and nltkWNLemmatizer() are added.
	
	Version 2.8												(29-May-2013)
	- extractNGrams() is added.
	
	Version 2.7												(01-May-2013)
	- convertDQuoteWSJ() is moved to parse.py.
	
	Version 2.6												(22-Mar-2013)
	- prepForParsing() and extractPOS() are moved to parse.py
	
	Version 2.5												(18-Mar-2013
	- prepForParsing() is upgraded.
	
	Version 2.4												(14-Mar-2013)
	- tokenizeSegment() is improved to treat French elision.
	
	Version 2.3												(07-Mar-2013)
	- decodeHTMLSpecial() is improved.
	- tokenizeSegment() improved to treat empty segments.
	- removeCR() is added.
	
	Version 2.2												(22-Feb-2013)
	- tokenizeSegment() is improved.
	- permuteLines() is added.
	
	Version 2.1												(20-Feb-2013)
	- extractPOS() is added.
	- extractCorpusStat() is upgraded.
	- replaceLines() is added.
	
	Version 2.0												(12-Feb-2013)
	- tokenizeText() is changed to accept and output string instead of 
	  file.
	- extractCorpusStat() is changed to accept string instead of file.
	- tokenizeSegment() is improved.
	- deduplicateCorpus() is improved to optionally trim empty lines.
	
	Version 1.9												(06-Feb-2013)
	- extractUnseen() is added.
	- deduplicateCorpus() is added.
	- maskHTMLSpecial() is renamed to decodeHTMLSpecial().
	- cleanGarbage() is revised to exclude >.
	- remTag() is split into remHTMLTag() and remSYMCTag(). The former is
	  revised to completely remove tags instead of masking. The latter
	  fixes Symantec special tags like menu item wrapper.
	- cleanHTML() is revised.
	- prepCleanHTML() is added.
	- Masking identical adjacent patterns for Email is removed.
	- fixListItemSplit() is renamed to fixSentSeg()
	- removeHTMLComment() is revised.
	- segToSentNLTK() is added.
	- segToSentence() is added.
	- maskWPath() is revised.
	- maskURL() is revised.
	- prepForParsing() is added.
	- convertDQuoteWSJ() is added.
	
	Version 1.8												(21-Jan-2013)
	- trimText() is added.
	- fixListItemSplit() is added.
	- sys.sys bug caused at an unknown stage by blind replacement is fixed.
	
	Version 1.7												(02-Jan-2013)
	- cleanHTML() is improved.
	- nltkCleanHTML() is removed.
	
	Version 1.6												(28-Dec-2012)
	- HTML cleaning is added.
	- decodeHTMLSpecial() is improved.
	- removeHTMLComment() is added.
	
	Version 1.5												(14-Dec-2012)
	- tokenizeSegment() is improved to tokenize fusions like "now...it",
	  correctly handle German ordinal numbers. It does not try to split
	  fusions like "now.It" any more to avoid tokenizing cases like
	  "Trojan.Agent/Gen-ReLoader".
	- Cleaning and masking functions are moved from clean.py in.
	  
	Version 1.4												(28-Nov-2012)
	- A bug in handling non-breaking prefixes in tokenizeSegment() was
	  fixed and support to the . at the end of abbreviations was added, 
	  i.e., N.I.S. is not tokenized to N.I.S . any more. 
	  
	Version 1.3												(16-Oct-2012)
	- nltkCleanHTML() is added.
	- nltkSplitSent() is added.
	
	Version 1.2												(01-Oct-2012)
	- isGood() is added.
	- extractGoodSentences() is added.
	- ccldDetectSentLang() is added.
	- ccldDetectLinesLang() is added.
	- extractParMatchLines() is added.
	- extractCorpusStat() is moved in from tempfuncs
	- extractCorporaOverlap() is added.
	
	Version 1.1												(28-Sep-2012)
	- All languages are handled inside tokenizeSegment() instead of 
	  language-specific functions.
	
	Version 1.0												(10-Sep-2012)
	- tokenizeSegment()
	- tokenizeSegmentCommon()
	- tokenizeSegmentEn()
	- tokenizeText()
	- loadNonBreakTokens()
	
"""

from decimal import *
from math import *
from collections import namedtuple
from collections import defaultdict
import sys, re, os

from numba.targets.linalg import vdot
from utils import util

'''
	Punctuation marks
'''
gcPuncs = ur".,;:?!'\"`‘’“”«»-"

'''
	Unicode capital letters: French, German
''' 
gcUniCap = u"ÀÂÄÈÉÊËÎÏÔŒÙÛÜŸÖ"

'''
	Unicode capital letters: French, German
''' 
gcUniLow = u"àâäèéêëîïôœùûüÿöß"

'''
	French vowles
''' 
gcFrVowels = u"AÀÂÄEÈÉÊËIÎÏOÔŒUÙÛÜYŸaàâäeèéêëiîïoôœuùûüyÿ"

'''
	data structure for Chromium CLD (language detector) output
'''
CCLD = namedtuple('CCLD', "topLangName topLangCode isReliable textBytesFound details")

'''
	data structure for langid (language detector) output
'''
LID = namedtuple('LID', "langCode confidence")

'''
	Emoticon masks
'''
gcEmoticonMask = [(":\)"   , "-EMTCN01-", ":)" ),
                  (":D\b" , "-EMTCN02-", ":D" ),
                  (":-\)"  , "-EMTCN03-", ":-)"),
                  (":\("   , "-EMTCN04-", ":(" ),
                  (":-\("  , "-EMTCN05-", ":-("),
                  (";\)"   , "-EMTCN06-", ";)" ),
                  (";-\)"  , "-EMTCN07-", ";-)"),
                  (":o\)"  , "-EMTCN08-", ":o)"),
                  (":-D\b", "-EMTCN09-", ":-D"),
                  (":P\b" , "-EMTCN10-", ":P" ),
                  (":\|"   , "-EMTCN11-", ":|" )
                  ]



def loadNonBreakTokens(pLang):
	'''
	Loads the list of non-breaking tokens for pLang from the corresponding 
	file and returns.
	
	It looks for the file in the same directory as this script.
	'''
	
	vNBTFile = os.path.dirname(os.path.realpath(__file__)) + "/nlp.py.nonbrk"
	
	if pLang.lower() in ["english", "en"]:
		vNBTFile += ".en"
	elif pLang.lower() in ["french", "fr"]:
		vNBTFile += ".fr"
	elif pLang.lower() in ["german", "de"]:
		vNBTFile += ".de"
	else:
		sys.stderr.write(pLang + " is not identified! English will be used.\n")
		vNBTFile += ".en"
	
	try:
		vfNBTFile = open(vNBTFile)
	except:
		sys.exit("Cannot open non-breaking tokens file: " + vNBTFile)
	
	
	vlNBT = []
	for vLine in vfNBTFile:
		vLine = vLine.strip()
		if vLine.startswith('#') or (len(vLine) == 0):
			continue
		else:
			vlNBT.append(vLine.decode("UTF-8"))
	
	vfNBTFile.close()
	return vlNBT
	


def tokenizeSegment(pSegment, pLang, pflgTokenizeFSlash = False, pflgTweet = False, pflgUnicodeErrs = "strict"):
	''' 
	Tokenizes a given text segment. A segment concept covers clause, 
	sentence, and paragraph without new line.
	
	Note: In version 3.1, support for emoticons is added. Instead of 
	      treating them among the other rules, they are first masked before
	      any other process and unmasked at the end.
	'''
	
	if pLang.lower() in ["english", "en"]:
		vLang = "en"
	elif pLang.lower() in ["french", "fr"]:
		vLang = "fr"
	elif pLang.lower() in ["german", "de"]:
		vLang = "de"
	else:
		sys.stderr.write(pLang + " is not identified! English will be used.\n")
		vLang = "en"
	
	vSegment = pSegment.strip()
	if vSegment == "":
		return pSegment
	
	if isinstance(vSegment, unicode):
		vWasUnicode = True
	else:
		vSegment = unicode(vSegment, "UTF-8", errors = pflgUnicodeErrs)
		vWasUnicode = False
	
	# masking emoticons
	vSegment = maskEmoticons(vSegment, pflgCommonMask = False)
	
	# replacing all types of spaces by normal space: U+00A0 U+2002 U+2003
	vPattern = re.compile(r"(?: | | )")
	vSegment = vPattern.sub(r" ", vSegment)
	
	# treating brackets
	vPattern = re.compile(r"([\(\)\[\]\{\}<>])")
	vSegment = vPattern.sub(r" \1 ", vSegment)
	
	## NOTE: trailing punctuations wont be split 
	
	## '`"‘“«» as openning (» sometimes (e.g. SYMC) appears as opening)
	vPattern = re.compile(ur"(?:(?<=^)|(?<=[ \t]))((['`\"‘“«»])\2*)")
	vSegment = vPattern.sub(r"\1 ", vSegment)
	# '`"’”«» as closing
	# vPattern = re.compile(ur"((['`\"’”«»])\2*)(?=[\s.,;:?!])", re.UNICODE)		version 2.1
	vPattern = re.compile(ur"((['`\"’”«»])\2*)(?=$|[\s.,;:?!])", re.UNICODE)
	vSegment = vPattern.sub(r" \1 ", vSegment)
	
	# / (new in version 3.2)
	if pflgTokenizeFSlash:
		vPattern = re.compile(r"([/]+)")
		vSegment = vPattern.sub(r" \1 ", vSegment)
		
	
	## French elision: ' surrounded by letters
	if vLang == "fr":
		#vPattern = re.compile(ur"([^\W\d_]')(?=["+gcFrVowels+"])", re.UNICODE)		version 2.3
		vPattern = re.compile(ur"([^\W\d_]')(?=[^\W\d_])", re.UNICODE)
		vSegment = vPattern.sub(r"\1 ", vSegment)
	
	# `'’ as apostrophe in English (all considered to be the same)
	if vLang == "en":
		# X's X'll X've X're X'm X'd
		# vPattern = re.compile(ur"([\w])(?=([`'’])\2*(?:s|ll|ve|re|m|d)[\s.,;:?!])", re.IGNORECASE | re.UNICODE)		version 2.1
		vPattern = re.compile(ur"([\w])(?=([`'’])\2*(?:s|ll|ve|re|m|d)(?:$|[\s.,;:?!]))", re.IGNORECASE | re.UNICODE)
		vSegment = vPattern.sub(r"\1 ", vSegment)
		# Xn't
		# vPattern = re.compile(ur"([a-zA-Z])(?=(?:[n]([`'’])\2*[t])[\s.,;:?!])", re.IGNORECASE | re.UNICODE)			version 2.1
		vPattern = re.compile(ur"([a-zA-Z])(?=(?:[n]([`'’])\2*[t])(?:$|[\s.,;:?!]))", re.IGNORECASE | re.UNICODE)
		vSegment = vPattern.sub(r"\1 ", vSegment)
	
	# cannot
	if vLang == "en":
		vPattern = re.compile(r"cannot", re.IGNORECASE)
		vSegment = vPattern.sub(r"can not", vSegment)
	
	## , :
	## split only if followed by space 
	## NOTE: , : are also split if not surrounded by digits
	vPattern = re.compile(r"(([,:])\2*)(?=$|[\s])", re.UNICODE)
	vSegment = vPattern.sub(r" \1", vSegment)
	vPattern = re.compile(r"(?:(?<=^)|(?<=[^\d]))(([,:])\2*)(?=[^\d])", re.UNICODE)
	vSegment = vPattern.sub(r" \1 ", vSegment)
	
	## ; ! ? … — and | ^ * (non-punctuations; for efficiency)
	## = added in version 3.1
	## always split (like parenthesis)
	vPattern = re.compile(ur"([;!?…—|^*=]+)")
	vSegment = vPattern.sub(ur" \1 ", vSegment)
	
	## trailing dashes
	vPattern = re.compile(ur"([-]{2,})")
	vSegment = vPattern.sub(ur" \1 ", vSegment)
	
	## treating $ % £ €
	## split from any digit preceding or following it
	vPattern = re.compile(ur"(?<=[\d])(([$%£€])\2*)", re.UNICODE)
	vSegment = vPattern.sub(r" \1 ", vSegment)
	vPattern = re.compile(ur"(([$%£€])\2*)(?=[\d])", re.UNICODE)
	vSegment = vPattern.sub(r" \1 ", vSegment)
	
	
	# special characters: ™ ® ©
	vPattern = re.compile(ur"([™®©])")
	vSegment = vPattern.sub(r" \1 ", vSegment)
	
	# @ # at the beginning and end of token					Version 2.0
	if not pflgTweet:
		vPattern = re.compile(r"(?:(?<=\s)|(?<=^))([@#])")
		vSegment = vPattern.sub(r"\1 ", vSegment)
		vPattern = re.compile(r"([@#])(?=$|[\s])")
		vSegment = vPattern.sub(r" \1", vSegment)
	
	# unmasking emoticons
	vSegment = unmaskEmoticons(vSegment)
	
	## . 
	## NOTE: . is treated at the end
	## NOTE: . needs to be done on token level to handle non-breaking 
	##       tokens like Mr.
	vlSegTokens = vSegment.split()
	vTokenizedSeg = ""
	vlNonBreakTokens = loadNonBreakTokens(vLang)
	for vToken in vlSegTokens:
		## German ordinal numbers (e.g. 11. ), though they are handled by
		## non-breaking prefixes to some extent 
		if (not (pLang == "de" and re.match("^\d+[.]$", vToken)) 
			 and 
		    (vToken.lower() not in [token.lower() for token in vlNonBreakTokens])):
			## . at the end of token unless preceded by a dot and capital
			## letter (to avoid splitting abbreviations to some extent) 
			vPattern = re.compile(r"(?<![.][A-Z"+gcUniCap+r"])(([.])\2*)$", re.UNICODE)
			vToken = vPattern.sub(r" \1", vToken)
			## helping sentence split to some extent:  "him.They" (Deactivated
			## in version 1.5)
			##vPattern = re.compile(ur"(?<=[a-z"+gcUniLow+r"])([.])(?=[A-Z"+gcUniCap+"])", re.UNICODE)
			##vToken = vPattern.sub(r" \1 ", vToken)
			# fusion on ellipses is also treated:  "him...they", " ..they"
			vPattern = re.compile(ur"([.]{2,})(?=[\w])", re.UNICODE)
			vToken = vPattern.sub(r" \1 ", vToken)
		
		vTokenizedSeg += ' ' + vToken
	
	vTokenizedSeg = ' '.join(vTokenizedSeg.split())
	if not vWasUnicode:
		vTokenizedSeg = vTokenizedSeg.encode("utf-8")
	
	if pSegment[-1] == '\n':
		return vTokenizedSeg + '\n'
	else:
		return vTokenizedSeg
	


def tokenizeText(pCorpus, pLang):
	''' 
	Tokenized text in the given corpus.
	
	NOTE: Replaces all types of white spaces with normal space.
	'''
	
	if pLang.lower() in ["english", "en"]:
		vLang = "en"
	elif pLang.lower() in ["french", "fr"]:
		vLang = "fr"
	elif pLang.lower() in ["german", "de"]:
		vLang = "de"
	else:
		sys.stderr.write(pLang + " is not identified! English will be used.\n")
		vLang = "en"
	
	vlOutput = []
	for i, vLine in enumerate(pCorpus.strip().split('\n'), start = 1):
		vlOutput.append(tokenizeSegment(vLine, vLang, False))
		
		if i % 100000 == 0:
			sys.stderr.write('!')
		elif i % 10000 == 0:
			sys.stderr.write('.')
	
	if i >= 10000:
		sys.stderr.write('\n')
	
	return '\n'.join(vlOutput) + '\n'
	


def maskEmoticons(pText, pflgCommonMask = True):
	'''
	Masks emoticons based on a list defined globally
	
	The mask can be common (-EMT-) or specific to each emoticon as defined
	in a global list.
	'''
	
	vMasked = pText
	
	for vEmoticon, vMask, dummy in gcEmoticonMask:
		vPattern = re.compile(vEmoticon)
		if pflgCommonMask:
			vMasked = vPattern.sub("-EMT-", vMasked)
		else:
			vMasked = vPattern.sub(vMask, vMasked)
	
	return vMasked
	


def unmaskEmoticons(pText):
	'''
	Unmasks emoticons based on a list defined globally
	'''
	
	vUnmasked = pText
	
	for dummy, vMask, vEmoticon in gcEmoticonMask:
		vUnmasked = vUnmasked.replace(vMask, ' ' + vEmoticon + ' ')
	
	return vUnmasked
	


def isGoodSentence(pInput, pGoodType, pIsTokenized=False, pLang="en", pGood2Ratio=70, pGood3RatioGLB=70, pGood3RatioFLB=50, pGood3RatioF=50):
	'''
	Given pInput, checks whether it is a good sentence according to 
	the definition of pGoodType.
	
	Argumnets other than pGoodType and pInput are only used with some 
	pGoodTypes and can be left empty for others:
		- pIsTokenized: good-2, good-3
		- pLang: good-2, good-3
		- pGood2Ratio: good-2
		- ...
	
	pIsTokenized tells whether the input corpus is already tokenized or
	need to be tokeized if needed (for example for good-2).
	
	pGood3RatioGLB is the lower bound for good token ratio, bigger than
	which the sentence is passed as good.
	
	pGood3RatioFLB is the lower bound for good token ratio, bigger than
	which the fallback is applied (cf. good-3 description below).
	
	pGood3RatioFB is the fallback tokens ratio with respect to the total 
	non-good tokens. 
	
	Note to the code for the definition of possible pGoodType.
	'''
	
	if isinstance(pInput, unicode):
		vInput = pInput
		vWasUnicode = True
	else:
		vInput = unicode(pInput, "UTF-8")
		vWasUnicode = False
	
	vIsGood = False
	
	## good-1: starting with an alphabet letter and ending with a 
	##         punctuation mark of '.', '?' or '!' or ':' or ';' or an 
	##         alphabet letter
	## Note to the algorithm for up-to-date definition
	if pGoodType == 1:
		vlTokens = vInput.split()
		if len(vlTokens) > 0:
			if re.match(ur"[A-z"+gcUniCap+gcUniLow+"].*", vlTokens[0]):
				if re.match(ur".*?[A-z"+gcUniCap+gcUniLow+u".?!:;\"'’”»]", vlTokens[-1]):
					vIsGood = True
	
	## The following definitions of good word and good numeric will be 
	## used by good-2 and -3 heuristics.
	
	# good words: only alphabet, hyphen, and one ' at the end or beginning
	vGoodWordRE = r"^[']?(?:[^\W\d_]+[-])*[^\W\d_]+[']?$"
	## good numerics
	vGoodNumRE = r"^[\d]+[./,]*[\d]+$"
	
	## good-2: containing at least a certain ratio of good words; 
	## Note to the algorithm for up-to-date definition
	if pGoodType == 2:
		if not pIsTokenized:
			vInput = tokenizeSegment(vInput, pLang, False)
		
		vlTokens = vInput.split()
		vLineLen = len(vlTokens)
		if vLineLen > 0:
			vGood2Count = 0
			for vToken in vlTokens:
				# good words or good numerics
				if (re.match(vGoodWordRE, vToken, re.UNICODE) or
				    re.match(vGoodNumRE, vToken, re.UNICODE)):
					vGood2Count += 1
			
			vRatio = Decimal(vGood2Count) / vLineLen
			if (vRatio * 100) >= Decimal(pGood2Ratio):
				vIsGood = True
	
	## good-3: containing at least a certain ratio of good words like good-2,
	##         but as a fallback, for lower ratios than that but higher 
	##         than a lower bound threshold, if a certain ratio of the rest
	##         is punctuation, mask (-XYZ-), etc., it is also considered  
	##         as good sentence.
	##         For example, if at least %70 of the tokens are good words,
	##         the sentence is good, but if less than %50 are good words,
	##         it is not. If %60 (between %50 and %70) are good word, then
	##         if %80 of the remaining %40 is comprised of only punctuation
	##         marks or masks, it is a good sentence. Therefor, this 
	##         heuristic needs 3 ratio thresholds: %50,%70,%80.
	## Note to the algorithm for up-to-date definition
	if pGoodType == 3:
		if not pIsTokenized:
			vInput = tokenizeSegment(vInput, pLang, False)
		
		vlTokens = vInput.split()
		vLineLen = len(vlTokens)
		if vLineLen > 0:
			vGood3Count = 0
			vFallback3Count = 0
			for vToken in vlTokens:
				# good words or good numerics
				if (re.match(vGoodWordRE, vToken, re.UNICODE) or
				    re.match(vGoodNumRE, vToken, re.UNICODE)):
					vGood3Count += 1
				## fallback: 
				## - masks
				## - punctuations
				## - words containing _ or numbers (e.g. Win32, b2b)
				elif (re.match(r"-[A-Z]{3}-", vToken) or
				      re.match("["+gcPuncs+"]", vToken, re.UNICODE) or
				      re.match(r"^[']?(?:[\w]+[-])*[\w]+[']?$", vToken, re.UNICODE)
				      ):
					vFallback3Count += 1
			
			vGood3Ratio = Decimal(vGood3Count) / vLineLen
			if (vGood3Ratio * 100) >= Decimal(pGood3RatioGLB):
				vIsGood = True
			elif (vGood3Ratio * 100) >= Decimal(pGood3RatioFLB):
				# falling back
				vNoGood3Count = vLineLen - vGood3Count
				vFallback3Ratio = Decimal(vFallback3Count) / vNoGood3Count
				if (vFallback3Ratio * 100) >= Decimal(pGood3RatioF):
					vIsGood = True
	
	return vIsGood
	

def extractGoodSentences(pCorpus, pGoodType, pIsTokenized=False, pLang="en", pGood2Ratio=70, pGood3RatioGLB=70, pGood3RatioFLB=50, pGood3RatioF=80):
	'''
	Given a corpus checks each line and extract those falling in pGoodType
	criteria into a file with the same name and a suffix identifying the
	pGoodType. It also outputs the index of those lines into a file with
	the same name and ".idx" suffix.
	
	Argumnets other than pCorpus and pGoodType  are only used with some 
	pGoodTypes and can be left empty for others:
		- pIsTokenized: good-2
		- pLang: good-2
		- pGoodWordRatio: good-2
	
	pIsTokenized tells whether the input corpus is already tokenized or
	need to be tokeized if needed (for example for good-2).
	
	It uses isGoodSentence() function to identify the line being a good 
	sentence or not.
	'''
	
	try:
		vfCorpus = open(pCorpus)
	except IOError:
		sys.exit("Cannot open corpus file: " + pCorpus)
	
	try:
		vOutputFilename = pCorpus + '.good-' + str(pGoodType)
		vfOutput = open(vOutputFilename, 'w')
	except IOError:
		sys.exit("Cannot create output file: " + vOutputFilename)
	
	vLineCntr = 1
	vGoodCntr = 0
	vlGoodIdxs = []
	for vLine in vfCorpus:
		if isGoodSentence(vLine, pGoodType, pIsTokenized, pLang, pGood2Ratio, pGood3RatioGLB, pGood3RatioFLB, pGood3RatioF):
			vfOutput.write(vLine)
			vlGoodIdxs.append(str(vLineCntr))
			vGoodCntr += 1
		
		if vLineCntr % 100000 == 0:
			sys.stderr.write('!')
		elif vLineCntr % 10000 == 0:
			sys.stderr.write('.')
		
		vLineCntr += 1
	
	sys.stderr.write('\n' + str(vLineCntr-1) + " sentences processed\n")
	sys.stderr.write(str(vGoodCntr) + " good-" + str(pGoodType) + " sentences extracted\n") 
	
	try:
		vOutputIdxFilename = vOutputFilename + ".idx"
		vfOutputIdx = open(vOutputIdxFilename, 'w')
		vfOutputIdx.write(','.join(vlGoodIdxs) + '\n')
	except IOError:
		sys.exit("Cannot create output file: " + vOutputIdxFilename)
	
	vfCorpus.close()
	vfOutput.close()
	vfOutputIdx.close()
	

def ccldDetectSentLang(pSegment, pIsPlainTxt=True, pRemoveWeakMatches=False):
	'''
	Given a text, detects it language using Chromium Compact Language
	Detector.
	
	If pIsPlainTxt is set to false, the detector will try to strip HTML/
	XML markups.
	
	if pRemoveWeakMatches is set to false, even low-confidence guesses
	will be kept. Otherwise, unknown will be returned if no relibale guess 
	is made. (When this is not set to false and left as default, the 
	detector performs poorly and returns unknow for a huge number of 
	English sentences.)
	'''
	
	import cld
	
	if isinstance(pSegment, unicode):
		vSegment = pSegment.encode("UTF-8")
	else:
		vSegment = pSegment
	vTopLangName, vTopLangCode, vIsReliable, vTextBytesFound, vDetails = cld.detect(vSegment, isPlainText=pIsPlainTxt, removeWeakMatches=pRemoveWeakMatches)
	
	return CCLD(vTopLangName, vTopLangCode, vIsReliable, vTextBytesFound, vDetails)
	

def ccldDetectLinesLang(pCorpus, pIsPlainTxt=True, pRemoveWeakMatches=False):
	'''
	Given a corpus, detects the language of each line using Chromium 
	Compact Language Detector and writes its code at the beginning of 
	each line separated by pDelim
	'''
	
	import cld
	
	try:
		vfCorpus = open(pCorpus)
	except IOError:
		sys.exit("Cannot open input corpus: " + pCorpus)
	
	try:
		vOutFilename = pCorpus + ".lang"
		vfOutput = open(vOutFilename, 'w')
	except IOError:
		sys.exit("Cannot create output file: " + vOutFilename)
	
	vLineCntr = 1
	for vLine in vfCorpus:
		vResult = ccldDetectSentLang(vLine, pIsPlainTxt)
		vfOutput.write(vResult.topLangCode + '\n')
	
		if vLineCntr % 100000 == 0:
			sys.stderr.write('!')
		elif vLineCntr % 10000 == 0:
			sys.stderr.write('.')
		
		vLineCntr += 1
	
	sys.stderr.write('\n')
	
	vfCorpus.close()
	vfOutput.close()
	

def lidDetectSentLang(pSegment):
	'''
	Given a text, detects it language using Marco Lui's langid tool.
	'''
	
	import langid
	
	vResult = langid.classify(pSegment)
	
	return LID(vResult[0], vResult[1])
	

def lidDetectLinesLang(pCorpus):
	'''
	Given a corpus, detects the language of each line Marco Lui's langid 
	tool and writes its code at the beginning of each line separated by 
	pDelim
	'''
	
	try:
		vfCorpus = open(pCorpus)
	except IOError:
		sys.exit("Cannot open input corpus: " + pCorpus)
	
	try:
		vOutFilename = pCorpus + ".lang"
		vfOutput = open(vOutFilename, 'w')
	except IOError:
		sys.exit("Cannot create output file: " + vOutFilename)
	
	for vLineCntr, vLine in enumerate(vfCorpus, start=1):
		vResult = lidDetectSentLang(vLine)
		vfOutput.write(vResult.langCode + '\n')
	
		if vLineCntr % 100000 == 0:
			sys.stderr.write('!')
		elif vLineCntr % 10000 == 0:
			sys.stderr.write('.')
	
	sys.stderr.write('\n')
	
	vfCorpus.close()
	vfOutput.close()
	

def agreeDetectLinesLang(pCorpus, pHypothesisLang=''):
	'''
	04-Oct-2012
	
	Given a corpus, detects the language of each line based on the agreement
	between available detectors. If they don't agree on the language but
	pHypothesisLang is among the gusses, pHypothesisLang is considered
	the right guess. Otherwise, the language is considered as unknown.
	'''
	
	try:
		vfCorpus = open(pCorpus)
	except IOError:
		sys.exit("Cannot open input corpus file: %s" % pCorpus)
	
	try:
		vOutFilename = pCorpus + ".lang"
		vfOutput = open(vOutFilename, 'w')
	except IOError:
		sys.exit("Cannot create output file: " + vOutFilename)
	
	for vLine in vfCorpus:
		# detecting language with CCLD
		vCCLD = ccldDetectSentLang(vLine)
		# detecting language with langid
		vLID = lidDetectSentLang(vLine)
		# comparison
		if vCCLD.topLangCode == vLID.langCode:
			vfOutput.write(vCCLD.topLangCode + '\n')
		elif pHypothesisLang in (vCCLD.topLangCode,  vLID.langCode):
			vfOutput.write(pHypothesisLang + '\n')
		else:
			vfOutput.write("xx\n")
	
	vfCorpus.close()
	vfOutput.close()
	

def extractParMatchLines(plCorpora, pOutFilename):
	'''
	03-Oct-2012
	
	Given a list of corpora, takes the first as base and compares corresponding
	lines to check if any of them in the other corpora is the same as that
	in the base one. If so (one is enough), writes the lines in all corpora
	in the output file.
	
	This is written primarily for detecting foreign lines in parallel 
	corpora.
	'''
	
	# opening input files
 	vlfCorpora = []
	for vFileName in plCorpora:
		try:
			vlfCorpora.append(open(vFileName))
		except IOError:
			sys.exit("Cannot open input file: " + vFileName)
	
	try:
		vfOutput = open(pOutFilename, 'w')
	except IOError:
		sys.exit("Cannot open input file: " + pOutFilename)
	
	for i, vlParLines in enumerate(zip(*vlfCorpora), start=1):
		for vLine in vlParLines[1:]:
			if ' '.join(vlParLines[0].split()) == ' '.join(vLine.split()):
				vfOutput.write(str(i) + "\n")
				vfOutput.write(''.join(vlParLines))
				vfOutput.write("----------\n")
				break
	
	# closing files
	for vFile in vlfCorpora:
		vFile.close()
	vfOutput.close()
	

def extractCorpusStat(pCorpus, pOutFilePrefix, pflgWordList = False, pflgVocab = False, pflgVocabFreqVS = False, pflgVocabFreqFS = False, pflgSentLenDist = False):
	'''
	Extracts statistics such as line, word, and vocabulary count, etc. It 
	optionally generates the word list and vocabulary. Corpus should be 
	tokenized.
	
	pflgVocabFreqVS: vocabulary frequency sorted by vocabulary
	pflgVocabFreqFS: vocabulary frequency sorted by frequency
	pflgSentLenDist: sentence length distribution
	'''
	
	vlWords = []
	vlWordLens = []
	vlLineLens = []
	
	for vLine in pCorpus.strip('\n').split('\n'):
		vlLineWords = vLine.split()
		vlWords += vlLineWords
		vlLineLens.append(len(vlLineWords))
		
	vLineCount = len(vlLineLens)
	
	vWordCount = len(vlWords)
	vBlankLineCount = len([l for l in vlLineLens if l == 0])
	vMinLineLen = min(vlLineLens)
	vMaxLineLen = max(vlLineLens)
	vAvgLineLen = Decimal(Decimal(vWordCount)/vLineCount).quantize(Decimal('0.01'))
	vlLineLenVars = map(lambda x: (x - vAvgLineLen)**2, vlLineLens)
	vLineLenVar = Decimal(sum(vlLineLenVars))/vLineCount
	vLineLenSD = round(sqrt(vLineLenVar), 2)
	
	vlWordLens = [len(w) for w in vlWords]
	vCharCount = sum(vlWordLens)
	vMinWordLen = min(vlWordLens)
	vMaxWordLen = max(vlWordLens)
	vAvgWordLen = Decimal(Decimal(vCharCount)/vWordCount).quantize(Decimal('0.01'))
	vlWordLenVars = map(lambda x: (x - vAvgWordLen)**2, vlWordLens)
	vWordLenVar = Decimal(sum(vlWordLenVars))/vWordCount
	vWordLenSD = round(sqrt(vWordLenVar), 2)
	
	## generating the vocabulary and calculating the word frequency at the
	## same time 
	vdVocab = defaultdict(int)
	for vWord in vlWords:
		vdVocab[vWord] += 1
	
	vsVocab = sorted(vdVocab.iterkeys())
	vVocabSize = len(vsVocab)
	
	## sorting the vocabulary dictionary by frequency to extract most and 
	## least frequent words
	vlFSortedVocab = sorted(vdVocab.iteritems(), key=lambda (word, freq): (freq, word))
	
	if pflgWordList:
		try:
			vWordListFileName = pOutFilePrefix + ".words"
			vfWordList = open(vWordListFileName, 'w')
		except IOError:
			sys.exit("Cannot create word list output file: " + vWordListFileName)
		
		vfWordList.write('\n'.join(vlWords) + '\n')
		vfWordList.close()
	
	if pflgVocab:
		try:
			vVocabFileName = pOutFilePrefix + ".vocab"
			vfVocab = open(vVocabFileName, 'w')
		except IOError:
			sys.exit("Cannot create vocabulary output file: " + vVocabFileName)
		
		for vWord in vsVocab:
			vfVocab.write(vWord + '\n')
		
		vfVocab.close()
	
	if pflgVocabFreqVS:
		try:
			vVocabFreqVSFileName = pOutFilePrefix + ".vocab.freq.vs"
			vfVocabFreqVS = open(vVocabFreqVSFileName, 'w')
		except IOError:
			sys.exit("Cannot create vocabulary frequency (sorted by vocabulary) output file: " + vVocabFreqVSFileName)
		
		for vWord in vsVocab:
			vfVocabFreqVS.write(str(vdVocab[vWord]).rjust(7) + ' ' + vWord + '\n')
		
		vfVocabFreqVS.close()
	
	if pflgVocabFreqFS:
		try:
			vVocabFreqFSFileName = pOutFilePrefix + ".vocab.freq.fs"
			vfVocabFreqFS = open(vVocabFreqFSFileName, 'w')
		except IOError:
			sys.exit("Cannot create vocabulary frequency (sorted by vocabulary) output file: " + vVocabFreqFSFileName)
		
		for vWord, vFreq in vlFSortedVocab:
			vfVocabFreqFS.write(str(vFreq).rjust(7) + ' ' + vWord + '\n')
		
		vfVocabFreqFS.close()
	
	if pflgSentLenDist:
		try:
			vSentLenDistFilename = pOutFilePrefix + ".len-dist"
			vfSentLenDist = open(vSentLenDistFilename, 'w')
		except IOError:
			sys.exit("Cannot create sentence length distribution file: %s" % vSentLenDistFilename)
		
		vdLenCounts = util.groupBy(vlLineLens)
		for vLen, vCount in vdLenCounts.iteritems():
			vfSentLenDist.write("%s : %s\n" % (str(vLen).rjust(4), str(vCount)))
		
		vfSentLenDist.close()
	
	
	print "Number of lines .......... " + str(vLineCount)
	print "Number of words .......... " + str(vWordCount)
	print "Average line length ...... " + str(vAvgLineLen) + " words"
	print "S.D. of line length ...... " + str(vLineLenSD)
	print "Number of blank lines .... " + str(vBlankLineCount)
	print "Minimum line length ...... " + str(vMinLineLen) + " words"
	print "Maximum line length ...... " + str(vMaxLineLen) + " words"
	print "Size of vocabulary ....... " + str(vVocabSize)
	print "Most frequent word ....... " + vlFSortedVocab[-1][0] + ' (' + str(vlFSortedVocab[-1][1]) + " times)"
	print "Least frequent word ...... " + vlFSortedVocab[0][0] + ' (' + str(vlFSortedVocab[0][1]) + " times)"
	print "Number of unique words ... " + str(len([item for item in vlFSortedVocab if item[1] == 1]))
	print "Number of characters ..... " + str(vCharCount)
	print "Minimum word length ...... " + str(vMinWordLen) + " characters"
	print "Maximum word length ...... " + str(vMaxWordLen) + " characters"
	print "Average word length ...... " + str(vAvgWordLen) + " characters"
	print "S.D. of word length ...... " + str(vWordLenSD)
	


def calcOOV(pSrcCorpus, pTrgCorpus):
	'''
	Calculates the out-of-vocabulary rate of a source corpus with respect 
	to a target corpus at both sentence and corpus level
	'''
	
	# data structure for sentence-level OOV
	SentOOV = namedtuple("SentOOV", "no sent oovcnt oovrate")
	
	# return list
	vlSentOOVs = []
	
	# extracting the vocabulary set of the source corpus
	print "Calculating source vocabulary "
	vsSrcVocab = set(pSrcCorpus.split())
	
	# extracting the vocabulary set of the target corpus
	print "Calculating target vocabulary "
	vsTrgVocab = set(pTrgCorpus.split())
	
	
	# Caclulating sentence level OOV 
	print "Caclulating sentence level OOV ",
	
	vSrcSentCntr = 0
	for vSent in pSrcCorpus.strip('\n').split('\n'):
		if vSrcSentCntr >= 10000:
			if vSrcSentCntr % 10000 == 0:
				print '.',
		
		vOOVCnt = 0
		vsTokens = set(vSent.split())
		for vToken in vsTokens:
			if vToken not in vsTrgVocab:
				vOOVCnt += 1
				
		vlSentOOVs.append(SentOOV(vSrcSentCntr, vSent, vOOVCnt, Decimal(vOOVCnt)/len(vsTokens)))
		
		vSrcSentCntr += 1
	
	print 
	
	# calculating corpus level OOV
	print "Caclulating corpus level OOV ",
	
	vSrcVocabSize = len(vsSrcVocab)
	vCorpusOOVCnt = vSrcVocabSize - len(vsSrcVocab.intersection(vsTrgVocab))
	vCorpusOOVRate = Decimal(vCorpusOOVCnt) * 100 / vSrcVocabSize
	
	print 
	
	
	# returning results
	
	return vlSentOOVs, vCorpusOOVCnt, vSrcVocabSize, vCorpusOOVRate
	


def extractCorporaOverlap(pCorpus1, pCorpus2, pOutFilename, pflgSSM = True):
	'''
	Given two SORTED corpora of DISTINCT lines, finds identical lines and
	output index pairs. If pflgSSM is set to true, it will strip, split 
	and merge both sides before comparison.
	
	Note that it stops search when finds the first occurance of the line
	in the second corpora, they need to be deduplicated first.
	
	The output format is: i1=i2
	                      i1=12
	                      i1=i2
	'''
	
	sys.stdout.write("Reading corpora ...\n")
	sys.stdout.flush()
	
	try:
		vlC1 = open(pCorpus1).readlines()
	except IOError:
		sys.exit("Cannot open corpus 1: %s" % pCorpus1)
	
	try:
		vlC2 = open(pCorpus2).readlines()
	except IOError:
		sys.exit("Cannot open corpus 2: %s" % pCorpus2)
	
	try:
		vfOut = open(pOutFilename, 'w')
	except IOError:
		sys.exit("Cannot create output file: %s" % pOutFilename)
	
	
	sys.stdout.write("Processing ")
	sys.stdout.flush()
	
	## looping through the first corpus and looking for identical lines
	## in the second oen
	## NOTE: since we need indexes, no efficient way in Python is found
	##       other than linear search.
	
	# keeps track of the previous match in the second corpus so that the
	# new search starts from that instead of the beginning.
	vLastIdx2 = 1
	
	for vIdx1, vL1 in enumerate(vlC1, start=1):
		for vIdx2, vL2 in enumerate(vlC2[vLastIdx2:], start=vLastIdx2 + 1):
			if pflgSSM:
				# strip/split/merge to make sure
				vNSSML1 = ' '.join(vL1.strip().split())
				vNSSML2 = ' '.join(vL2.strip().split())
				## to speed up, compare the whole line only if the first
				## letter of the second is biger than that of the first
				## (because the corpora are sorted)
				if vNSSML1[0] < vNSSML2[0]:
					break
				else:
					if vNSSML1 == vNSSML2:
						vfOut.write(str(vIdx1) + '=' + str(vIdx2) + '\n')
						vLastIdx2 = vIdx2
						break
			else:
				## to speed up, compare the whole line only if the first
				## letter of the second is biger than that of the first
				## (because the corpora are sorted)
				if vL1[0] < vL2[0]:
					break
				else:
					if vL1 == vL2:
						vfOut.write(str(vIdx1) + '=' + str(vIdx2) + '\n')
						vLastIdx2 = vIdx2
						break
		
		if vIdx1 % 10000 == 0:
			sys.stdout.write('!')
			sys.stdout.flush()
		elif vIdx1 % 1000 == 0:
			sys.stdout.write('.')
			sys.stdout.flush()
	
	sys.stdout.write('\n')
	
	vfOut.close()
	

def nltkSplitSent(pCorpus, pflgRemoveCR = True):
	'''
	A wrapper around NLTK tokenize.sent_tokenize().
	
	pflgRemoveCR: remove \r
	'''
	
	from nltk import tokenize
	
	try:
		vInputText = open(pCorpus).read()
	except IOError:
		sys.exit("Cannot open input text: %s" % pCorpus)
	
	if pflgRemoveCR:
		vInputText = vInputText.replace("\r\n", "\n")
		
	vSplitText = tokenize.sent_tokenize(vInputText)
	
	try:
		vOutFilename = pCorpus + ".splt"
		vfOut = open(vOutFilename, 'w')
		vfOut.write('\n'.join(vSplitText) + '\n')
		vfOut.close()
	except IOError:
		sys.exit("Cannot create output file: %s" % vOutFilename)
	

'''
Cleaning and masking functions start here

NOTE: When the re concerns unicode, (e.g. \w), three things should be 
      taken care of:
       1) converting the input to unicode (UTF-8)
       2) using re.UNICODE with pattern
       3) converting the re to unicode if it contains unicode
          characters like € or em-dash
      The first is done here. The last two are left for each specific 
      pattern to be done where needed.
NOTE: when the re contains unicode like em-dash but not any of \w
      \s \b (both lower and upper cases), using re.U (2) has no effect.
      However, 1 and 3 are effective: both must be off or on TOGETHER.
NOTE: when the re contains any of \w \s \b (both lower and upper 
      cases), 3 is not compulsory (if re does not contain unicode).
      However, it is the best practice to do 1 and 2 together. I.e.
      sometimes either one is enough but sometimes they must both 
      be set. Note that 2 is more important.
'''

def decodeHTMLSpecial(pInput):
	'''
	Decodes HTML special characters
	
	Before version 1.9, it was maskHTMLSpecial().
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	## Since the number of occurrences is needed, we don't use
	## replace function.
	vHSpecialCount = 0
	
	## &amp; should be treated first for the cases like &amp;trade, 
	## and separtely because iteritems() does not follow the order
	## of the dictionary elements.
	vPattern = re.compile(r"&amp;", re.IGNORECASE)
	vInput, vNum = vPattern.subn(r"&", vInput)
	vHSpecialCount += vNum
	
	## the first line is added in version 1.6
	## the last line is added in version 2.3
	vdHTMSpec = {'&lt;':"<", '&gt;':">", \
	             '&apos;':"'", '&copy;':'(C)', '&dagger;':'†', '&eacute;':'é', '&euml;':'ë', '&euro;':'€', '&hellip;':'...', \
	             '&ldquo;':'``', '&rdquo;':"''", '&lsquo;':'`', '&rsqu?o;':"'", '&[mn]d?ash;':'-', '&oslash;':'ø', '&quot;':'"', \
				 '&laquo;':'«', '&raquo;':"»", '&reg;':'®', '&trade;':'™', '&uuml;':'ü', '(?:&[ \t]*n[ \t]*b[ \t]*s[ \t]*p[ \t]*;|&nbps;|&space;)+':' ', \
				 '&#160;':' ', '&#45;':'-'}
	for p, m in vdHTMSpec.iteritems():
		vPattern = re.compile(p, re.IGNORECASE)
		vInput, vNum = vPattern.subn(m.decode("UTF-8"), vInput)
		vHSpecialCount += vNum
	
	return vInput, vHSpecialCount
	

def removeHTMLComment(pInput):
	'''
	Remove HTML comments (<!-- -->) from pInput by replacing with \n. 
	
	New in version 1.6
	'''
	
	vPattern = re.compile(r"<!--.*?-->", re.DOTALL)
	vOutput , vNum = vPattern.subn(r" ", pInput)
	
	return vOutput, vNum
	

def maskListNum(pInput):
	'''
	Masks enumerated list numbers 
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vLNumCount = 0
	
	# N., N-, N) 
	vPattern = re.compile(r"^\s*\d+\s*[\).-](?=\s+[^\d]|$)", re.UNICODE)
	vInput, vNum = vPattern.subn(r"-LNM- ", vInput)
	vLNumCount += vNum
	# <N> 
	vPattern = re.compile(r"^\s*<\s*\d+\s*>", re.UNICODE)
	vInput, vNum = vPattern.subn(r"-LNM- ", vInput)
	vLNumCount += vNum
	# A., A-, A) 
	vPattern = re.compile(r"^\s*\w\s*[).-](?=\s+[^\d])", re.UNICODE)
	vInput, vNum = vPattern.subn(r"-LNM- ", vInput)
	vLNumCount += vNum
	if vLNumCount > 0:
		# masking adjacent identical masks
		vPattern = re.compile(r"-LNM-(?:[ \t]*-LNM-)+")
		vInput = vPattern.sub(r"-LNM-", vInput)
	
	return vInput, vLNumCount
	

def replaceSpecialChr(pInput):
	'''
	Replaces special characters
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vSCharCount = 0
	
	# spaces: U+00A0 U+2002 U+2003
	## NOTE: when u"" is used for the re, make sure the vInput is decoded
	##       as unicode as well
	vPattern = re.compile(ur"(?: | | )")
	vInput, vNum = vPattern.subn(r" ", vInput)
	vSCharCount += vNum
	## en-dashes, em-dash, horisontal bar (U+2015)
	## NOTE: we tokenize here because these dashes are mostly not for
	##       hyphenation.
	vPattern = re.compile(ur"[ \t]*(?:–|—|―)[ \t]*")
	vInput, vNum = vPattern.subn(r" - ", vInput)
	vSCharCount += vNum
	
	return vInput, vSCharCount
	

def removeCR(pInput):
	'''
	Remove carriage return from the end of input
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vPattern = re.compile(r"\r")
	vOutput , vNum = vPattern.subn(r"", pInput)
	
	return vOutput, vNum
	

def cleanGarbage(pInput):
	'''
	Cleans know grabages
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vGarbageCount = 0
	
	# empty brackets
	vPattern = re.compile(r"[\(\[\{][ \t]*[\)\]\}]")
	vInput, vNum = vPattern.subn(r"", vInput)
	vGarbageCount += vNum
	# line beginnings
	# vPattern = re.compile(ur"^[ \t]*(?:(?://||||•|·|�|§|©|д|Д)[ \t]*)+")		# version 1.2 of clean.py
	# vPattern = re.compile(ur"^(?:(?://|>||||•|·|�|§|©|д|Д)[ \t]*)+")			# version 1.8
	vPattern = re.compile(ur"^(?:(?://||||•|·|�|§|©|д|Д)[ \t]*)+")
	vInput, vNum = vPattern.subn(r"", vInput)
	vGarbageCount += vNum
	# line endings
	vPattern = re.compile(ur"(?:|||•|·|�|§|©|д|Д)+(?=$)")
	vInput, vNum = vPattern.subn(r"", vInput)
	vGarbageCount += vNum
	
	return vInput, vGarbageCount
	

def maskSpecialStr(pInput):
	'''
	Masks special strings
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vSStrCount = 0
	
	# e.g. 60:06:01:60:83:B0:11:00:4A:C4:8A:1D:
	vPattern = re.compile(r"(?:(?<=[\s#])|^)(?:[\d\w]{2,}:){4,}[\d\w]*", re.UNICODE)
	vInput, vNum = vPattern.subn(r" -SPS-", vInput)
	vSStrCount += vNum
	if vSStrCount > 0:
		# masking adjacent identical masks
		vPattern = re.compile(r"-SPS-(?:[ \t-]*-SPS-)+")
		vInput = vPattern.sub(r"-SPS-", vInput)
	
	return vInput, vSStrCount
	

def maskMemAddr(pInput):
	'''
	Masks memory addresses
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vMemAddrCount = 0
	
	# e.g. 0x88888888
	vPattern = re.compile(r"0x[\d\w]+", re.IGNORECASE | re.UNICODE)
	vInput, vNum = vPattern.subn(r"-MEM-", vInput)
	vMemAddrCount += vNum
	# This pattern was found in already filtered test data
	vPattern = re.compile(r"<MEM_ADDR>", re.IGNORECASE)
	vInput, vNum = vPattern.subn(r"-MEM-", vInput)
	vMemAddrCount += vNum
	if vMemAddrCount > 0:
		# masking adjacent identical masks
		vPattern = re.compile(r"-MEM-(?:[ \t-]*-MEM-)+")
		vInput = vPattern.sub(r"-MEM-", vInput)
	
	return vInput, vMemAddrCount
	

def maskRegKey(pInput):
	'''
	Masks windows registry keys
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vRegKeyCount = 0
	
	#vPattern = re.compile(r"(^|[ #(\[])HKEY_.*?([ #).\n\]])")
	vPattern = re.compile(r"(?:HKEY_|HKLM).*?((?:[\s#\)\"',\]])|\.[\s])", re.UNICODE)
	vInput, vNum = vPattern.subn(r" -WRK- \1", vInput)
	vRegKeyCount += vNum
	if vRegKeyCount > 0:
		# masking adjacent identical masks
		vPattern = re.compile(r"-WRK-(?:[ \t]*-WRK-)+")
		vInput = vPattern.sub(r"-WRK-", vInput)
	
	return vInput, vRegKeyCount
	

def maskPHolder(pInput):
	'''
	Masks placeholders: \{...\}
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vPHolderCount = 0
	
	vPattern = re.compile(r"\\\{.*?\\\}")
	vInput, vNum = vPattern.subn(r" -PHL- ", vInput)
	vPHolderCount += vNum
	# %5 , %1E_P_SdE_P_E
	vPattern = re.compile(r"(?:(?<=^)|(?<=[\s]))%[^\s.]*", re.UNICODE)
	vInput, vNum = vPattern.subn(r"-PHL-", vInput)
	vPHolderCount += vNum
	# ________
	vPattern = re.compile(r"\b[_]{2,}\b", re.UNICODE)
	vInput, vNum = vPattern.subn(r"-PHL-", vInput)
	vPHolderCount += vNum
	# xxxxxxx
	vPattern = re.compile(r"\b[x]{2,}\b", re.IGNORECASE | re.UNICODE)
	vInput, vNum = vPattern.subn(r" -PHL- ", vInput)
	vPHolderCount += vNum
	if vPHolderCount > 0:
		# masking adjacent identical masks
		vPattern = re.compile(r"-PHL-(?:[ \t]*-PHL-)+")
		vInput = vPattern.sub(r"-PHL-", vInput)
	
	return vInput, vPHolderCount
	

def remHTMLTag(pInput):
	'''
	Removes HTML/XML tags
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vTagCount = 0
	
	# HTML tags
	vPattern = re.compile(r"(?<!<)<\/?\w[^<>\n]*[>\n]", re.UNICODE)
	vInput, vNum = vPattern.subn(r" ", vInput)
	vTagCount += vNum
	## removed from version 1.9
	## HTML-like
	#vPattern = re.compile(r"<\w+\s*\w+\=.*?[>/].*>", re.UNICODE)
	#vInput, vNum = vPattern.subn(r" ", vInput)
	#vTagCount += vNum
	## XML-like
	#vPattern = re.compile(r"<\w+\:.*?[>/].*>", re.UNICODE)
	#vInput, vNum = vPattern.subn(r" ", vInput)
	#vTagCount += vNum
	
	return vInput, vTagCount
	

def remSYMCTag(pInput):
	'''
	Removes/replaces special tags seen in Symantec TM
	
	For example, in Click  < >Reports < >, the menu item will be replaced
	with "Reports". If needed, the quotation marks will be masked later. 
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vTagCount = 0
	
	# < >Reports < >
	vPattern = re.compile(r"<\s*>\s*([^<>]*?)\s*<\s*>", re.UNICODE)
	vInput, vNum = vPattern.subn(r'"\1"', vInput)
	vTagCount += vNum
	
	return vInput, vTagCount
	

def maskVar(pInput):
	'''
	Masks variables
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vVarCount = 0
	
	# e.g. <token >AssignOrgGroupTask < >
	vPattern = re.compile(r"<[\w-]+\s*>\s*[^<>]*?\s*<\s*>", re.UNICODE)
	vInput, vNum = vPattern.subn(r" -VAR- ", vInput)
	vVarCount += vNum
	# fusions like [..]X...
	# Note: the masked pattern could even be removed!
	vPattern = re.compile(ur"(?:\s|^)\[[^\]]*?\]([A-Z"+gcUniCap+"])", re.UNICODE)
	vInput, vNum = vPattern.subn(r"-VAR- \1", vInput)
	vVarCount += vNum
	# [..], <..> 
	vPattern = re.compile(r"[<\[].*?[>\]]")
	vInput, vNum = vPattern.subn(r"-VAR-", vInput)
	vVarCount += vNum
	# e.g. &ProductName; &ProdNameSNAC ;
	vPattern = re.compile(r"(?:(?<=^)|(?<=[\s]))&[^;\s]+[\s]*[;]", re.IGNORECASE | re.UNICODE)
	vInput, vNum = vPattern.subn(r"-VAR-", vInput)
	vVarCount += vNum
	# e.g. @productName @Reg_WBTDetails@
	vPattern = re.compile(r"(?:(?<=^)|(?<=[\s]))@[^\s.:=]+", re.IGNORECASE | re.UNICODE)
	vInput, vNum = vPattern.subn(r"-VAR-", vInput)
	vVarCount += vNum
	if vVarCount > 0:
		# masking adjacent identical masks
		vPattern = re.compile(r"-VAR-(?:[ \t]*-VAR-)+")
		vInput = vPattern.sub(r"-VAR-", vInput)
	
	return vInput, vVarCount
	

def maskBSlash(pInput):
	'''
	Replaces \n \r \t with whitespace
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vBSlashCount = 0
	
	# There are cases with %n instead of \n 
	vPattern = re.compile(r"%n")
	vInput, vNum = vPattern.subn(r" ", vInput)
	vBSlashCount += vNum
	# There are cases with \tab instead of \t
	vPattern = re.compile(r"\\tab")
	vInput, vNum = vPattern.subn(r" ", vInput)
	vBSlashCount += vNum
	vPattern = re.compile(r"^(?:\\+[ntr])+")
	vInput, vNum = vPattern.subn(r" ", vInput)
	vBSlashCount += vNum
	vPattern = re.compile(r"(?:\s+(?:\\+[ntr])+)+", re.UNICODE)
	vInput, vNum = vPattern.subn(r" ", vInput)
	vBSlashCount += vNum
	vPattern = re.compile(r"(?:([?.,\[\}]+)(?:\\+[ntr])+)+")
	vInput, vNum = vPattern.subn(r"\1 ", vInput)
	vBSlashCount += vNum
	vPattern = re.compile(r"(?:\\+[ntr])+$")
	vInput, vNum = vPattern.subn(r"", vInput)
	vBSlashCount += vNum
	# this should be done last
	vPattern = re.compile(r"(?:\\+[ntr]){2,}")
	vInput, vNum = vPattern.subn(r" ", vInput)
	vBSlashCount += vNum
	
	return vInput, vBSlashCount
	

def maskEmail(pInput):
	'''
	Masks Email addresses
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vEmailCount = 0
	
	# Note: This pattern masks more than emails.
	vPattern = re.compile(r"\b[^\s@]+@[\w-]+(?:[.][\w-]+)+", re.UNICODE)
	vInput, vNum = vPattern.subn(r" -EML- ", vInput)
	vEmailCount += vNum
	# removed in version 1.9
	#if vEmailCount > 0:
	#	# masking adjacent identical masks
	#	vPattern = re.compile(r"-EML-(?:[ \t]*-EML-)+")
	#	vInput = vPattern.sub(r"-EML-", vInput)
	
	return vInput, vEmailCount
	

def maskURL(pInput):
	'''
	Masks URL
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vURLCount = 0
	
	vPattern = re.compile(r"(t?ftp:|https?:).*?(?=[\s)\"',\]\}]|\.[\s])", re.IGNORECASE | re.UNICODE)
	vInput, vNum = vPattern.subn(r" -URL- ", vInput)
	vURLCount += vNum
	#vPattern = re.compile(r"[w]{2,}\..*?\..*?(?=[\s#)\"',\]\}]|\.[\s])", re.IGNORECASE | re.UNICODE)		version 1.8
	# deactivated in version 1.9: to be revised. 
	#vPattern = re.compile(r"[w]{2,}\.[^ .]+\.[^ .]+(?=[\s#)\"',\]\}]|\.[\s])", re.IGNORECASE | re.UNICODE)
	#vInput, vNum = vPattern.subn(r" -URL- ", vInput)
	#vURLCount += vNum
	#vPattern = re.compile(r"\b[\S]{2,}[.](?:com|net|org|co\.uk|ru|ch|de|fr)\b[\S]*?(?=[\s#)\"',\]\}]|\.[\s])", re.IGNORECASE | re.UNICODE)
	#vInput, vNum = vPattern.subn(r" -URL- ", vInput)
	#vURLCount += vNum
	# This pattern was found in already filtered test data
	vPattern = re.compile(r"<url_ph>", re.IGNORECASE)
	vInput, vNum = vPattern.subn(r"-URL-", vInput)
	vURLCount += vNum
	if vURLCount > 0:
		# masking adjacent identical masks
		vPattern = re.compile(r"-URL-(?:[ \t]*-URL-)+")
		vInput = vPattern.sub(r"-URL-", vInput)
	
	return vInput, vURLCount
	

def maskWPath(pInput):
	'''
	Masks windows file paths.
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vWPathCount = 0
	
	## This pattern treats paths containing whitespace. The results
	## will be consumed by the next patterns to fully mask the path.
	## Note that it's not included in counting.
	vPattern = re.compile(r"\\[\w-]+(?:[\s]+[\w-]+){1,3}(?=\\)", re.UNICODE)
	vInput, vNum = vPattern.subn(r"\\WFP", vInput)
	## mathcing some known whitespace issues not captured above
	## NOTE: add to this
	vPattern = re.compile(r"\\enterprise vault", re.IGNORECASE)
	vInput, vNum = vPattern.subn(r"\\WFP", vInput)
	# this pattern utilizes " to handle paths containing whitespace
	vPattern = re.compile(r"[\"'][^\s]*?\\+.*?[\"']", re.UNICODE)
	vInput, vNum = vPattern.subn(r" -WFP- ", vInput)
	vWPathCount += vNum
	vPattern = re.compile(r"(?:\b|\\)[\\]*(?:[^\s/]+?\\+)+[^\s/]*(?:\b|$)", re.UNICODE)
	vInput, vNum = vPattern.subn(r" -WFP- ", vInput)
	vWPathCount += vNum
	## This pattern is meant to match file names. However, it may 
	## also match domain names which are missed above in URL masking
	## as well as some period fusions. Also, it does not cover any 
	## possible file name.
	#vPattern = re.compile(r"(^|[^\w.-])[\w-]{2,}(?:\.[\w-]{2,})+(['\"\)\s.])", re.UNICODE)		verison 1.2
	#vPattern = re.compile(r"(^|[^\w.-])[\w-]{2,}(?:\.[\w-]{2,})+(['\"\)\s.,])", re.UNICODE)	version 1.8
	#vInput, vNum = vPattern.subn(r"\1-WFP-\2", vInput)
	#vWPathCount += vNum
	# This pattern was found in already filtered test data
	vPattern = re.compile(r"<winpath>", re.IGNORECASE)
	vInput, vNum = vPattern.subn(r"-WFP-", vInput)
	vWPathCount += vNum
	if vWPathCount > 0:
		# merging masks with various neighbours (e.g. -WFP- /)
		vPattern = re.compile(r"-WFP-\s*\\+(?=[\s#)\"',\]\}]|\.[\s])", re.UNICODE)
		vInput = vPattern.sub(r" -WFP- ", vInput)
		# masking adjacent identical masks
		vPattern = re.compile(r"-WFP-(?:[ \t]*-WFP-)+")
		vInput = vPattern.sub(r"-WFP-", vInput)
	
	return vInput, vWPathCount
	
 
def maskUPath(pInput):
	'''
	Masks Unix-like file paths
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vUPathCount = 0
	
	## This pattern treats paths containing whitespace. The results
	## will be consumed by the next patterns to fully mask the path.
	## Note that it's not included in counting.
	#vPattern = re.compile(r"/[\w-]{2,}(?:[\s]+[\w-]{2,}){1,3}(?=/)", re.UNICODE)		version 1.8
	#vInput, vNum = vPattern.subn(r"/UFP", vInput)
	# this pattern utilizes "" to handle paths containing whitespace
	vPattern = re.compile(r"([\"'])[^\s]*?/+.*?([\"'])", re.UNICODE)
	#vInput, vNum = vPattern.subn(r" -UFP- ", vInput)									version 1.8
	vInput, vNum = vPattern.subn(r"\1 -UFP- \2", vInput)
	vUPathCount += vNum
	## This pattern matchs string with more than one / to avoid 
	## masking non-path paterns like Users//////Symantec
	vPattern = re.compile(r"(?:\b|/)[/]*(?:[^\s/]+?/+){2,}[^\s/]*\b", re.UNICODE)
	vInput, vNum = vPattern.subn(r" -UFP- ", vInput)
	vUPathCount += vNum
	if vUPathCount > 0:
		# masking adjacent identical masks
		vPattern = re.compile(r"-UFP-(?:[ \t]*-UFP-)+")
		vInput = vPattern.sub(r"-UFP-", vInput)
	
	return vInput, vUPathCount
	

def maskIPV(pInput, pLang):
	'''
	Masks IP addresses and version number
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vIPVCount = 0
	
	vPattern = re.compile(r"\b[\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}(?::[\d]{1,5})?(?:/[\d]{1,2})?(?=[\D])", re.UNICODE)
	vInput, vNum = vPattern.subn(r"-IPV-", vInput)
	vIPVCount += vNum
	# handling decimal point variation
	if pLang == "en":
		vPattern = re.compile(r"\b[vV]?[\d]{1,2}\.[\d]{1,2}(?:\.[\d]{1,5})+(?=[\D])", re.UNICODE)
	elif pLang in ["fr", "de"]:
		vPattern = re.compile(r"\b[vV]?[\d]{1,2},[\d]{1,2}(?:,[\d]{1,5})+(?=[\D])", re.UNICODE)
	vInput, vNum = vPattern.subn(r"-IPV-", vInput)
	vIPVCount += vNum
	# This pattern was found in already-filtered test data and TM
	vPattern = re.compile(r"<(?:ip_address|adresse_ip|ip_ver|&version;)>", re.IGNORECASE)
	vInput, vNum = vPattern.subn(r"-IPV-", vInput)
	vIPVCount += vNum
	if vIPVCount > 0:
		# masking adjacent identical masks
		vPattern = re.compile(r"-IPV-(?:[ \t]*-IPV-)+")
		vInput = vPattern.sub(r"-IPV-", vInput)
	
	return vInput, vIPVCount
	

def maskDateTime(pInput):
	'''
	Masks date/time
	
	NOTE: Date/time masking is considerably shrinked. Based on Treebanks,
	      it is mostly left for parsing and translation. Only formats
	      like 12-10-2012 which are fully numbers with separators are
	      masked.
	NOTE: Date and time are masked separately. They'll be merged later,
	      when all neighbouring similar masked are merged in a post-
	      processing run.
	NOTE: Some regex requires numbers to be masked afterwards. E.g.
	      they look for 4 digits while number masks are 3 (999).
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vDTCount = 0
	
	# dates
	# e.g. 2/1/98, 02-1-1987, 01.12.2002 (German)
	vPattern = re.compile(r"\b\d{1,2}([./-])\d{1,2}\1(?:\d{4}|\d{2})", re.UNICODE)
	vInput, vNum = vPattern.subn(r"-DTM- ", vInput)
	vDTCount += vNum
	# e.g. 2000-01-1, 98/11/10, 2002.12.01
	vPattern = re.compile(r"\b(?:\d{4}|\d{2})([./-])\d{1,2}\1\d{1,2}", re.UNICODE)
	vInput, vNum = vPattern.subn(r" -DTM- ", vInput)
	vDTCount += vNum
	#vMonths = "January|Jan|February|Feb|March|Mar|April|Apr|May|June|Jun|July|Jul|August|Aug|September|Sep|October|Oct|November|Nov|December|Dec"
	#vDays = "Monday|Mon|Tuesday|Tues|Tue|Wednesday|Wed|Thursday|Thurs|Thur|Thu|Friday|Fri|Saturday|Sat|Sunday|Sun"
	# e.g. 2 Jan 98, 02 December 2005
	#vPattern = re.compile(r"\b\d{1,2}(?:th|st|rd|nd)?\s(?:" + vMonths + r")\s*[,]?\s*(?:\d{4}|\d{2})", re.IGNORECASE)
	#vInput, vNum= vPattern.subn(r" -DTM- ", vInput)
	#vDTCount += vNum
	# e.g. December 31th, 2008 - Aug 18 
	#vPattern = re.compile(r"(?:" + vMonths + r")[,]?\s+\d{1,2}(?:th|st|rd|nd)?[ \t]*[,]?[ \t]*(?:\d{4}|\d{2})?", re.IGNORECASE)
	#vInput, vNum = vPattern.subn(r" -DTM- ", vInput)
	#vDTCount += vNum
	# e.g. Thursday 30th September 
	#vPattern = re.compile(r"(?:" + vDays + r")[,]?\s+\d{1,2}(?:th|st|rd|nd)?[ \t]*[,]?[ \t]*(?:" + vMonths + r")?", re.IGNORECASE)
	#vInput, vNum = vPattern.subn(r" -DTM- ", vInput)
	#vDTCount += vNum
	
	# times
	# yyyy-mm-ddThh:mmZ is a frequent pattern in Symantec data
	vPattern = re.compile(r"(?:\b|T)[\d]{1,2}[:][\d]{2}(?:h(?:rs)?)?(?::[\d]{2})?(?:\.[\d]+)?\s*(?:\(mm:ss\))?(?:\b|(?=[apz]))", re.IGNORECASE | re.UNICODE)
	vInput, vNum = vPattern.subn(r"-DTM- ", vInput)
	vDTCount += vNum
	# mainly French
	vPattern = re.compile(r"\b[\d]{1,2}\s*h\s*[\d]{1,2}(?:\s*m(?:in)?\s*[\d]{1,2}\s*s)?", re.UNICODE)
	vInput, vNum = vPattern.subn(r"-DTM- ", vInput)
	vDTCount += vNum
	#vTimeZones = "GMT|UTC|BST|CET|CEST|CST|EDT|ET|EST|EET|PDT|PST|UK"
	#vTimePeriod = "AM|A\.M\.|A\.M|PM|P\.M\.|P\.M"
	#vPattern = re.compile(r"(?:(?<=[^\d:])|^)(?:(?:" + vTimeZones + r")[\+\-]?)?[\d]{1,2}[.:][\d]{2}(?:h(rs)?)?(?::[\d]{2})?(?:\.[\d]+)?\s*(?:" + vTimePeriod + r")?\s*(?:(?:[\(](?:" + vTimeZones + r")[\)])|(?:" + vTimeZones + r"))?(?:\(mm:ss\))?(?=[^\d])", re.IGNORECASE)
	#vInput, vNum = vPattern.subn(r"-DTM- ", vInput)
	#vDTCount += vNum
	#vPattern = re.compile(r"(?:(?<=[^\d:])|^)(?:(?:" + vTimeZones + r")[\+\-]?)?[\d]{1,2}(?:[:.][\d]{1,2})?\s*(?:" + vTimePeriod + r")\s*(?:(?:[\(](?:" + vTimeZones + #r")[\)])|(?:" + vTimeZones + r"))?(?=[^\d\w])", re.IGNORECASE)
	#vInput, vNum = vPattern.subn(r"-DTM- ", vInput)
	#vDTCount += vNum
	
	# This pattern was found in already filtered test data
	vPattern = re.compile(r"<date_and_time>", re.IGNORECASE)
	vInput, vNum = vPattern.subn(r"-DTM-", vInput)
	vDTCount += vNum
	if vDTCount > 0:
		# masking adjacent identical masks
		vPattern = re.compile(r"-DTM-(?:[ \t]*-DTM-)+")
		vInput = vPattern.sub(r"-DTM-", vInput)
	
	# merging neighbouring DTMs
	# e.g. -DTM- 2008
	vPattern = re.compile(r"(?:-DTM-\s*)+\s*\d{4}", re.IGNORECASE | re.UNICODE)
	vInput = vPattern.sub(r"-DTM- ", vInput)
	# e.g. Sunday, -DTM-
	#vPattern = re.compile(r"(?:" + vDays + "),?\s*-DTM-", re.IGNORECASE | re.UNICODE)
	#vInput = vPattern.sub(r" -DTM-", vInput)
	# e.g. 2008--DTM- T-DTM- Z
	vPattern = re.compile(r"(?:\b\d{4}-)?-DTM-\s*T\s*-DTM-[ \t]*(?:Z|\+-DTM-|--DTM-)?", re.IGNORECASE | re.UNICODE)
	vInput = vPattern.sub(r" -DTM- ", vInput)
	# e.g. 27- -DTM-
	vPattern = re.compile(r"\b\d{1,4}-[ \t]*-DTM-", re.IGNORECASE | re.UNICODE)
	vInput = vPattern.sub(r" -DTM-", vInput)
	# e.g. -DTM- -28, 2005
	vPattern = re.compile(r"-DTM-[ \t]*-\d{1,4},?[ \t]*(?:\d{1,4})?", re.IGNORECASE | re.UNICODE)
	vInput = vPattern.sub(r"-DTM- ", vInput)
	
	return vInput, vDTCount
	

def maskNum(pInput, pLang):
	'''
	Masks numbers with 999
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vNumCount = 0
	
	# money
	if pLang in ["en"]:
		vPattern = re.compile(ur"(?P<curr>[£€$]|GBP|AUD|NZD|SGD)[ \t]*\d+(?:,\d+)*(?:\.\d+)?", re.UNICODE)
	elif pLang in ["fr", "de"]:
		vPattern = re.compile(ur"(?P<curr>[£€$])[ \t]*\d+(?:\.\d+)*(?:,\d+)?", re.UNICODE)
	vInput, vNum = vPattern.subn(r"\g<curr>999", vInput)
	vNumCount += vNum
	# any number sequence
	## Note: do this after all number-related masking
	if pLang in ["en"]:
		vPattern = re.compile(r"(?:(?<=^)|(?<=[^\w-]))\d+(?:,\d+)*(?:\.\d+)?(?=[^\w-])", re.UNICODE)
	elif pLang in ["fr", "de"]:
		vPattern = re.compile(r"(?:(?<=^)|(?<=[^\w-]))\d+(?:\.\d+)*(?:,\d+)?(?=[^\w-])", re.UNICODE)
	vInput, vNum = vPattern.subn(r"999", vInput)
	vNumCount += vNum
	
	return vInput, vNumCount
	

def maskLongStr(pInput, pLang, pMaskLen):
	'''
	Masks strings longer than pMaskLen
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vLStrCount = 0
	
	## While the longest French word can be 25 charcters, tokens can
	## be constructed by combining words with underscores. To tackle
	## this issue, we escape tokens without masking if it has an 
	## underscore.
	if not (pLang == "fr" and '_' in vInput):
		vPattern = re.compile(r"[^\s]{" + str(pMaskLen + 1) + ",}", re.UNICODE)
		vInput, vNum = vPattern.subn(r"-LNG-", vInput)
		vLStrCount += vNum
		if vLStrCount > 0:
			# masking adjacent identical masks
			vPattern = re.compile(r"-LNG-(?:[ \t]*-LNG-)+")
			vInput = vPattern.sub(r"-LNG-", vInput)
	
	return vInput, vLStrCount
	

def maskABracket(pInput, pflgTokenize):
	'''
	Masks angle brackets (includes tokenization)
	
	Optionally does the tokenization by adding spaces around the mask.
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vABracketCount = 0
	
	if pflgTokenize:
		vTkzr = ' '
	else:
		vTkzr = ''
	
	vPattern = re.compile("<")
	vInput, vNum = vPattern.subn(vTkzr + "-LAB-" + vTkzr, vInput)
	vABracketCount += vNum
	vPattern = re.compile(">")
	vInput, vNum = vPattern.subn(vTkzr + "-RAB-" + vTkzr, vInput)
	vABracketCount += vNum
	
	return vInput, vABracketCount
	

def maskRBracket(pInput, pflgTokenize):
	'''
	Masks round brackets
	
	Optionally does the tokenization by adding spaces around the mask.
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vRBracketCount = 0
	
	if pflgTokenize:
		vTkzr = ' '
	else:
		vTkzr = ''
	
	vPattern = re.compile("\(")
	vInput, vNum = vPattern.subn(vTkzr + "-LRB-" + vTkzr, vInput)
	vRBracketCount += vNum
	vPattern = re.compile("\)")
	vInput, vNum = vPattern.subn(vTkzr + "-RRB-" + vTkzr, vInput)
	vRBracketCount += vNum
	
	return vInput, vRBracketCount
	

def maskSBracket(pInput, pflgTokenize):
	'''
	Masks square brackets (includes tokenization)
	
	Optionally does the tokenization by adding spaces around the mask.
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vSBracketCount = 0
	
	if pflgTokenize:
		vTkzr = ' '
	else:
		vTkzr = ''
	
	vPattern = re.compile("\[")
	vInput, vNum = vPattern.subn(vTkzr + "-LSB-" + vTkzr, vInput)
	vSBracketCount += vNum
	vPattern = re.compile("\]")
	vInput, vNum = vPattern.subn(vTkzr + "-RSB-" + vTkzr, vInput)
	vSBracketCount += vNum
	
	return vInput, vSBracketCount
	

def maskCBracket(pInput, pflgTokenize):
	'''
	Masks curly brackets (includes tokenization)
	
	Optionally does the tokenization by adding spaces around the mask.
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vCBracketCount = 0
	
	if pflgTokenize:
		vTkzr = ' '
	else:
		vTkzr = ''
	
	vPattern = re.compile("\{")
	vInput, vNum = vPattern.subn(vTkzr + "-LCB-" + vTkzr, vInput)
	vCBracketCount += vNum
	vPattern = re.compile("\}")
	vInput, vNum = vPattern.subn(vTkzr + "-LCB-" + vTkzr, vInput)
	vCBracketCount += vNum
	
	return vInput, vCBracketCount
	

def maskAtSign(pInput):
	'''
	Masks @
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vAtSignCount = 0
	
	vPattern = re.compile("@")
	vInput, vNum = vPattern.subn(r"-ATS-", vInput)
	vAtSignCount += vNum
	
	return vInput, vAtSignCount
	

def mrgMasks(pInput):
	'''
	Merges adjacent non-identical similar masks
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vAMaskCount = 0
	
	vPattern = re.compile(r"(?:(?:-URL-|-EML-|-UFP-|-IPV-|-LNG-)[ \t]*){2,}")
	vInput, vNum = vPattern.subn(r" -URL- ", vInput)
	vAMaskCount += vNum
	vPattern = re.compile(r"(?:(?:-WRK-|-UFP-|-WFP-|-LNG-)[ \t]*){2,}")
	vInput, vNum = vPattern.subn(r" -WRK- ", vInput)
	vAMaskCount += vNum
	vPattern = re.compile(r"(?:(?:-SPS-|-MEM-|-IPV-|-LNG-)[ \t]*){2,}")
	vInput, vNum = vPattern.subn(r" -URL- ", vInput)
	vAMaskCount += vNum
	vPattern = re.compile(r"(?:(?:-VAR-|-PHL-|-LNG-)[ \t]*){2,}")
	vInput, vNum = vPattern.subn(r" -VAR- ", vInput)
	vAMaskCount += vNum
	
	return vInput, vAMaskCount
	

def isUseless(pInput):
	'''
	Checks if pInput is useless accrding to the following definition:
	
	A text is useless if it does not have a non-mask token with at least
	two contiguous alphabet characters
	'''
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	vUseful = False
	vlTokens = vInput.split()
	for vToken in vlTokens:
		## if the token is not a mask, check whether it has at 
		## least 2 alphabet characters
		if re.match("(?:[^\s\w]*-[A-Z]{3}-[^\s\w]*)|(?:[^\s\w]*999[^\s\w]*)", vToken, re.UNICODE) == None:
			vPattern = re.compile(ur"[A-z"+gcUniCap+gcUniLow+"]{2,}", re.UNICODE)
			if len(re.findall(vPattern, vToken)) > 0:
				vUseful = True
				break
	
	return not vUseful
	

'''
Cleaning and masking functions end here
'''

def cleanHTML(pInput, pInsertOLNums, pTool="remTag"):
	'''
	Extract text from HTML file source using pTool. remtag is a very simplistic
	in-house function with the same name.
	
	Since each tool has it own shortcoming, the function performs a pre 
	or post-processing to cover them. For example, NLTK and BeautifulSoup
	do not insert numbers to numbered list (<ol>) items/ If pInsertOLNums
	is set to true, it will be done in a preprocessing step.
	'''
	
	vTool = pTool.lower()
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	# pre-processing
	
	vInput = prepCleanHTML(vInput, pInsertOLNums, pTool)
	
	# main phase
	
	if vTool in ["beautifulsoup", "bsoup"]:
		vOutput = cleanHTMLBSoup(vInput)
	elif vTool == "nltk":
		vOutput = cleanHTMLNLTK(vInput)
	elif vTool == "remTag":
		vOutput, dummy = remTag(vInput)
	
	# post-processing
	
	## 1. decoding HTML special characters.
	if vTool in ["beautifulsoup", "bsoup", "nltk"]:
		vOutput, dummy = decodeHTMLSpecial(vOutput)
	
	## 2. replacing special characters
	vOutput, dummy = replaceSpecialChr(vOutput)
	
	## 3. some HTML tags originally encoded using <code>&amp;</code> are
	##    decoded when cleaning HTML tags by <code>NLTK</code>, but have
	##    not been removed. The remaining tags are removed by in-house
	##    remTag() function.
	##    Note that when the decoding is done beforehand, the cleaning 
	##    process never ends!
	
	vOutput = prepCleanHTML(vOutput, pInsertOLNums, pTool)
	vOutput, dummy = remTag(vOutput) 
	
	return vOutput
	

def prepCleanHTML(pInput, pInsertOLNums, pTool="remTag"):
	'''
	Pre-processes pInput (HTML) to cover the shortcomings of different 
	tools.
	'''
	
	vTool = pTool.lower()
	
	if not isinstance(pInput, unicode):
		vInput = pInput.decode("utf-8")
	else:
		vInput = pInput
	
	## 1. inserting \n before and after text blocks as some tools do not
	## start new paragraphs in new lines.
	if vTool in ["remtag", "beautifulsoup", "bsoup", "nltk"]:
		vPattern = re.compile(r"<(p|h\d|div|ul|ol|li|table|blockquote|iframe)>", re.IGNORECASE)
		vInput = vPattern.sub(r"\n<\1>", vInput)
		# vPattern = re.compile(r"</(p|h\d|div|ul|ol|table|blockquote)>", re.IGNORECASE)	version 1.6
		vPattern = re.compile(r"</(p|h\d|div|ul|ol|li|table|blockquote|iframe)>", re.IGNORECASE)
		vInput = vPattern.sub(r"</\1>\n", vInput)
	
	## 2. Removing HTML comments (replacing with \n)
	if vTool in ["remtag", "beautifulsoup", "bsoup", "nltk"]:
		vInput, dummy = removeHTMLComment(vInput)
	
	## 3. inserting numbers to numbered list items
	## It assumes that both ol and li are upper or lower at the same time.
	if vTool in ["remtag", "beautifulsoup", "bsoup", "nltk"] and pInsertOLNums:
		# uppercase
		vEOLPos = 0
		while True: 
			vBOLPos = vInput.find("<OL>", vEOLPos)
			if vBOLPos == -1:
				break
			vEOLPos = vInput.find("</OL>", vBOLPos + 4)
			
			vNum = 1
			vItems = vInput[vBOLPos + 4 : vEOLPos]
			vELIPos = 0
			while True:
				vBLIPos = vItems.find("<LI>", vELIPos)
				if vBLIPos == -1:
					break
				
				vItems = vItems[ : vBLIPos + 4] + str(vNum) + ". " + vItems[vBLIPos + 4 : ]
				vELIPos = vItems.find("</LI>", vBLIPos + 4)
				
				vNum += 1
			
			vInput = vInput[ : vBOLPos + 4] + vItems + vInput[vEOLPos:]
		
		# lowercase
		vEOLPos = 0
		while True: 
			vBOLPos = vInput.find("<ol>", vEOLPos)
			if vBOLPos == -1:
				break
			vEOLPos = vInput.find("</ol>", vBOLPos + 4)
			
			vNum = 1
			vItems = vInput[vBOLPos + 4 : vEOLPos]
			vELIPos = 0
			while True:
				vBLIPos = vItems.find("<li>", vELIPos)
				if vBLIPos == -1:
					break
				
				vItems = vItems[ : vBLIPos + 4] + str(vNum) + ". " + vItems[vBLIPos + 4 : ]
				vELIPos = vItems.find("</li>", vBLIPos + 4)
				
				vNum += 1
			
			vInput = vInput[ : vBOLPos + 4] + vItems + vInput[vEOLPos:]
	
	return vInput
	

def cleanHTMLBSoup(pInput):
	'''
	Cleans HTML tags from pInput using BeautifulSoup.
	'''
	
	from bs4 import BeautifulSoup
	
	## I found that &# sequence breaks BeautfulSoup. Therefore, it's 
	## converted to &amp;# then back again after cleaning tags.
	
	vInput = re.sub("&#", "&amp;#", pInput)
	
	vInputSoup = BeautifulSoup(vInput)
	vOutput = vInputSoup.get_text()
	vOutput = re.sub("&amp;#", "&#", vOutput)
	
	return vOutput
	

def cleanHTMLNLTK(pInput):
	'''
	Cleans HTML tags from pInput using NLTK.
	'''
	
	import nltk
	
	vOutput = nltk.clean_html(pInput)
	
	return vOutput
	

def trimText(pInput):
	'''
	Given a text as input string, trims spaces and tabs surrounding the 
	lines and removes blank lines.
	'''
	
	vlTrimed = []
	for vLine in pInput.split('\n'):
		vLine = vLine.strip()
		if vLine != "":
			vlTrimed.append(vLine)
	
	return '\n'.join(vlTrimed) + '\n'
	

def fixSentSeg(pInput):
	'''
	Given a corpus in pInput, fixes the known sentence segmentation issues:
	
	- wrongly-split enumerated list items on the period after the enumerator
	 (e.g. 1. or a.)
	'''
	
	# 1. sample
	vPattern = re.compile(r"(\n[0-9]{1,2}\.)[\n]+", re.IGNORECASE)
	vOutput = vPattern.sub(r"\1 ", pInput)
	
	# a. sample
	vPattern = re.compile(r"(\n[a-z]\.)[\n]+", re.IGNORECASE)
	vOutput = vPattern.sub(r"\1 ", vOutput)
	
	return vOutput
	

def extractUnseen(pCorpus1Filename, pCorpus2Filename, pOutFilename, pOutIndexes = False):
	'''
	Extracts lines from pCorpus1Filename which are not part of or include
	any line in pCorpus2Filename.
	
	Optionally, a CSV of on-based indexes of extracted lines in pCorpus1Filename
	is output.
	
	Note that the inclusion criteria is very strict.
	'''
	
	try:
		vlCorpus1 = open(pCorpus1Filename).readlines()
	except IOError:
		sys.exit("Cannot open %s!" % pCorpus1Filename)
	
	try:
		vlCorpus2 = open(pCorpus2Filename).readlines()
	except IOError:
		sys.exit("Cannot open %s!" % pCorpus2Filename)
	
	vlUnseenIdx = []
	for i, vLine1 in enumerate(vlCorpus1, start = 1):
		vSLine1 = vLine1.strip()
		vSeen = False
		
		for vLine2 in vlCorpus2:
			vSLine2 = vLine2.strip()
			if vSLine1 in vSLine2 or vSLine2 in vSLine1:
				vSeen = True
				break
		
		if not vSeen:
			vlUnseenIdx.append(i)
		
		if i % 10000 == 0:
			sys.stdout.write('!')
			sys.stdout.flush()
	
	try:
		open(pOutFilename, 'w').write(''.join([vlCorpus1[idx - 1] for idx in vlUnseenIdx]))
	except IOError:
		sys.exit("Cannot create %s!" % pOutFilename)
	
	if pOutIndexes:
		try:
			open(pOutFilename + ".idx", 'w').write(','.join([str(idx) for idx in vlUnseenIdx]))
		except IOError:
			sys.exit("Cannot create %s.idx!" % pOutFilename)
	

def deduplicateCorpus(pCorpusFilename, pOutFilename, removeBlank = False, pOutIndexes = False):
	'''
	Removes duplicates from pCorpusFileName while keeping the original 
	order intact.
	
	Optionally, one-based indexes of selected sentences are output.
	'''
	
	try:
		vlCorpus = open(pCorpusFilename).readlines()
	except IOError:
		sys.exit("Cannot open %s!" % pCorpusFilename)
	
	vlOutIdx = []
	vlOutput = []
	for i, vLine in enumerate(vlCorpus, start = 1):
		if removeBlank and vLine.strip() == "":
			continue
		elif vLine not in vlOutput:
			vlOutput.append(vLine)
			vlOutIdx.append(i)
	
	try:
		open(pOutFilename, 'w').write(''.join(vlOutput))
	except IOError:
		sys.exit("Cannot create %s!" % pOutFilename)
	
	if pOutIndexes:
		try:
			open(pOutFilename + ".idx", 'w').write(','.join([str(idx) for idx in vlOutIdx]))
		except IOError:
			sys.exit("Cannot create %s.idx!" % pOutFilename)
	

def segToSentence(pInput, pTool, pFixOutput = False, pLang="en"):
	'''
	Segments text in pInput using pTools sentence segmenter.
	Optionally, it fixes the output of tool for some known issues.
	'''
	
	vTool = pTool.lower()
	
	if vTool == "nltk":
		vSents = segToSentNLTK(pInput, pLang)
	
	if pFixOutput:
		vSents = trimText(vSents)
		vSents = fixSentSeg(vSents)
	
	return vSents
	

def segToSentNLTK(pInput, pLang="en"):
	'''
	Segments text into sentences using NLTK and returns as a string with
	new line characters as sentence delimiters.
	'''
	
	import nltk.data
	
	if pLang.lower().startswith("en"):
		vLangName = "english"
	elif pLang.lower().startswith("fr"):
		vLangName = "french"
	
	vSegmenter = nltk.data.load('tokenizers/punkt/%s.pickle' % vLangName)
	
	return '\n'.join(vSegmenter.tokenize(pInput))
	

def replaceLines(pCorpus, pNewLines, pRepIdxs, pDelim=','):
	'''
	Replaces lines in pCorpus with those in pNewLines based on line numbers
	in pRepIdxs.
	'''
	
	vlCorpus = pCorpus.strip('\n').split('\n')
	vlNewLines = pNewLines.strip('\n').split('\n')
	vlRepIdxs = [int(idx) for idx in pRepIdxs.strip().split(pDelim)]
	
	if len(vlNewLines) != len(vlRepIdxs):
		sys.exit("Number of new lines and replacement indexes do not match!")
	
	for vIdx, vNewLine in zip(vlRepIdxs, vlNewLines):
		vlCorpus[vIdx - 1] = vNewLine
	
	return '\n'.join(vlCorpus)
	


def permuteLines(pCorpus, pMethod = "random", pOrder = "", pOrderDelim = ','):
	'''
	Given a corpus, permutes its lines according to pMethod.
	
	pMethod values are:
	- random: shuffles the corpus; other valid values: r, rand
	- order: permutes lines in the order given by pOrder, which is a string
	         of 1-based indexes separated by pOrderDelim; other valid 
	         values: o
	
	If the method is random, it also returns the 1-based indexes of lines
	in the original corpus.
	'''
	
	import random
	
	vlLines = pCorpus.strip('\n').split('\n')
	
	if pMethod.lower() in ["random", "rand", "r"]:
		vlOrder = range(len(vlLines))
		random.shuffle(vlOrder)
		vlPermuted = [vlLines[idx] for idx in vlOrder]
		return vlPermuted, ','.join([str(idx + 1) for idx in vlOrder])
	elif pMethod.lower() in ["order", "o"]:
		if pOrder == "":
			sys.exit("Order is not given")
		vlPermuted = [vlLines[int(idx) - 1] for idx in pOrder.strip().split(pOrderDelim)]
		return vlPermuted
	else:
		sys.exit("Method %s is not supported" % pMethod)
	
	


def extractNGrams(pInput, pN, pflgTokenize = False, pLang = "en"):
	'''
	Extracts pN-grams from the input.
	
	pInput can be a list of tokens or a segment. The latter will be tokenized
	into a list if pflgTokenize is false, and split otherwise. The return
	will be a list of n-grams, which are lists if the input is list and
	strings otherwise.
	
	If the length of input is smaller than pN, the return will be empty.
	'''
	
	if isinstance(pInput, list):
		vlInput = pInput
	else:
		if pflgTokenize:
			vlInput = tokenizeSegment(pInput, pLang, False)
		else:
			vlInput = pInput.split()
	
	vLen = len(vlInput)
	
	vlNGrams = []
	for i in range(vLen - pN + 1):
		vlNGrams.append(vlInput[i : i + pN])
	
	if isinstance(pInput, list):
		return vlNGrams
	else:
		return [' '.join(ngram) for ngram in vlNGrams]
	


def mapPenn2WordNet(pPennPOS):
	'''
	maps Penn POS tag to WordNet POS tag
	'''
	
	from nltk.corpus import wordnet
	
	if pPennPOS.startswith('NN'):
		return wordnet.NOUN
	if pPennPOS.startswith('JJ'):
		return wordnet.ADJ
	if pPennPOS.startswith('VB'):
		return wordnet.VERB
	if pPennPOS.startswith('RB'):
		return wordnet.ADV
	else:
		return ''
	

def nltkWNLemmatizer(pToken, pPOS = ''):
	'''
	Lemmatizes the token with its POS tag as an additional information
	'''
	
	from nltk.stem.wordnet import WordNetLemmatizer
	
	vLMTZR = WordNetLemmatizer()
	
	vWNPOS = mapPenn2WordNet(pPOS)
	
	if vWNPOS == '':
		return vLMTZR.lemmatize(pToken)
	else:
		return vLMTZR.lemmatize(pToken, vWNPOS)
	


def numericalizeCorpus(pCorpus, pVocabSize = None, pSentStartIdx = None, pOOVIdx = 2, pStartingIdx = 3):
	'''
	Converts tokens in the given corpus into numerical indecies generated based on token frequencies
	
	The corpus is expected to in a 2D list of tokens. Both numericalized corpus and token-index map are returned.
	
	If pVocabSize is given, only that number of tokens from the top frequencies will be used. In that case, pOOVIdx is used
	to replace the tokens falling out of this set.
	
	If not none, pSentStartIdx is inserted to the beginning of each line. For example, if it is set to 1, all the lines
	will start with 1 and then the actual token indecies of that line.
	
	The top token will get the index specified in pStartingIdx.
	
	NOTE: 0 is not used to be reserved for padding purposes.
	
	NOTE: this function is inspired by Keras data loading mechanism.
	'''
	
	# flattening corpus
	vlTokens = [t for s in pCorpus for t in s]
	
	# computing frequencies
	vdTokenFreqs = util.groupBy(vlTokens)
	
	# sorting based on token frequency
	vlSortedTokens = [t for t, f in sorted([(t, f) for (t, f) in vdTokenFreqs.iteritems()], key = lambda x: x[1], reverse = True)]
	
	# creating token-index map
	
	vdTokenIdxMap = {}
	
	if pVocabSize is None:
		vVocabSize = len(vlSortedTokens)
	else:
		vVocabSize = pVocabSize
	
	# first part falling in vocabulary
	for idx, vToken in enumerate(vlSortedTokens[ : vVocabSize], start = pStartingIdx):
		vdTokenIdxMap[vToken] = idx
	
	# second part: OOV
	for vToken in vlSortedTokens[vVocabSize : ]:
		vdTokenIdxMap[vToken] = pOOVIdx
	
	# now converting corpus
	
	vllOutput = []
	for vLine in pCorpus:
		vNewLine = [vdTokenIdxMap[t] for t in vLine]
		
		if pSentStartIdx is None:
			vllOutput.append(vNewLine)
		else:
			vllOutput.append([pSentStartIdx] + vNewLine)
	
	return vllOutput, vdTokenIdxMap
	


def chNumericalizeCorpus(pCorpus):
	'''
	Converts characters in the given corpus into numerical indecies
	
	The corpus is expected to in a 2D list of tokens. Both character-numericalized corpus and character-index map are 
	returned.
	
	Indicies are assigned to characters based on their alphabetical order.
	
	NOTE: 0 is not used to be reserved for padding purposes.
	'''
	
	# flattening corpus
	
	vlChars = [c for s in pCorpus for t in s for c in t]
	
	# creating character-index map
	
	vdCharIdxMap = {}
	
	for idx, ch in enumerate(sorted(set(vlChars)), start = 1):
		vdCharIdxMap[ch] = idx
	
	# now converting corpus
	
	vlllOutput = []
	for vLine in pCorpus:
		vllNewLine = []
		for vToken in vLine:
			vlNewToken = [vdCharIdxMap[c] for c in vToken]
			vllNewLine.append(vlNewToken)
		
		vlllOutput.append(vllNewLine)
	
	
	return vlllOutput, vdCharIdxMap
	
	
