#! /usr/bin/python

## This module defines functions to process Berkeley parser output.
##
## Current version: 1.1
##
## Added:
## - mapToCommonTagSet(): maps the syntactic tags in Berkeley parser 
##   output to a common tag set. 
## - To be used by the above function, mappings are defined in the module
##   as lists of pairs (tuples). This mapping are specified using numbers
##   by callers and identified inside the function.
##   Currently, only one mapping has been defined, with mapping number of
##   1, with Berkeley output with a grammar defined on FTB and Stanford 
##   output with its built-in frenchFactored grammar in mind. The mapping
##   is from these two tag sets to a common one consisting of 25 tags.
##   Note that the mapping does not include top node tag. It is treated
##   inside the function.


from collections import namedtuple
import sys, shutil, os, codecs


##----------------------------------------------------------------------
## Mappings
##
## Note that we also include neutral mappings where source and target tags
## are the same. The reason is to catch unexpected tags not included in 
## the table, through checking for all encountered tags in the input.
##
## clMap1: maps from FTB-trained Berkeley tag set to a common tag set 
##         between itself and frenchFactored-trained Stanford: 41 tags 
##         to a set of 25 tags.

clMap1 = [('ADJ', 'A'),
          ('ADJWH', 'A'),
          ('AdP', 'AdP'),
          ('ADV', 'ADV'),
          ('ADVWH', 'ADV'),
          ('AP', 'AP'),
          ('CC', 'C'),
          ('CLO', 'CL'),
          ('CLR', 'CL'),
          ('CLS', 'CL'),
          ('COORD', 'COORD'),
          ('CS', 'C'),
          ('DET', 'D'),
          ('DETWH', 'D'),
          ('ET', 'ET'),
          ('I', 'I'),
          ('NC', 'N'),
          ('NP', 'NP'),
          ('NPP', 'NP'),
          ('P', 'P'),
          ('P+D', 'P'),
          ('PONCT', 'PUNC'),
          ('PP', 'PP'),
          ('P+PRO', 'PRO'),
          ('PREF', 'PREF'),
          ('PRO', 'PRO'),
          ('PROREL', 'PRO'),
          ('PROWH', 'PRO'),
          ('SENT', 'SENT'),
          ('Sint', 'Sint'),
          ('Srel', 'Srel'),
          ('Ssub', 'Ssub'),
          ('V', 'V'),
          ('VIMP', 'V'),
          ('VINF', 'V'),
          ('VN', 'VN'),
          ('VPinf', 'VPinf'),
          ('VPP', 'V'),
          ('VPpart', 'VPpart'),
          ('VPR', 'V'),
          ('VS', 'V')]


##----------------------------------------------------------------------
## Generates a dummy parse which annotates each word of the input sentence
## with label X with the root tag X (e.g. (TOP (X (X it) (X is) (X good)))) 
def generateDummyParse(pSentence):
 vlNodes = ["(X " + token + ")" for token in pSentence.split()]
 return "(X " + ' '.join(vlNodes) + ')'



##----------------------------------------------------------------------
## Extracts the parse trees from Berkeley output, and optionally adds TOP
## tag. Note that depending on the grammar, the output of Berkeley parser
## may contain trees already wrapped into ROOT (as TOP replacement) and 
## together with those without such tag. The function, in case of request
## for adding TOP, replaces ROOT with TOP if it is already present or adds
## TOP otherwise.
## Note that in any case, the output of Berkeley is wrapped into parantheses 
## plus a space, which are dropped by this function.
## It optionally replaces -atsymb- with @.
## It optionally replaces -LRB- and -RRB- with ( and ).
## Berkeley parser outputs (()) for sentences which cannot be parsed. This
## funstion replaces them with a default dummy parse or with one generated
## by tagging each word with lable X, depending on the value of pflgXDummy.
#### Note: the latter treatment has not been implemented yet.
def extractParses(pParseFileName, pEncoding, pflgAddTOP, pflgUnmaskAT, pflgUnmaskRB, pflgXDummy, pflgPrint):
 try:
  vfParse  = codecs.open(pParseFileName, 'r', encoding=pEncoding)
 except IOError:
  sys.exit('Can\'t open parse file: ' + pParseFileName)

 vlParses = []
 for line in vfParse:
  if pflgUnmaskAT:
   line = line.replace("-atsymb-", '@')

  if pflgUnmaskRB:
   line = line.replace("-LRB-", '(')
   line = line.replace("-RRB-", ')')
  
  if line.startswith("(())"):
   if pflgXDummy:
    ### X-labaled dummy parse generation is not implemented yet
    pass
   else:
    vlParses.append("(TOP (UNPARSED UNPARSED))")
  elif line.startswith('( ('):
   if pflgAddTOP:
    if line[3:].startswith("ROOT"):
     vlParses.append("(TOP" + line.strip('\n')[7:-2])
    else:
     vlParses.append("(TOP" + line.strip('\n')[1:-2] + ')')
   else:
    vlParses.append(line.strip('\n'))
  else:
   sys.exit("Unknown line opening: " + line)
   break
  
  print vlParses[-1]
   
 vfParse.close()
 
 return vlParses

 
##----------------------------------------------------------------------
## Changes the syntactic tags according to the specified mapping.
## Note that the mapping does not include top node tag. It is treated 
## based pChangeTopNode (saying that change the top node tag), and 
## pTopRootTag (specifying the top node tag to be used in the output)
def mapToCommonTagSet(pParseTree, pMapping, pflgChangeTopNode, pTopRootTag, pflgPrint):
 # selecting mapping based on pMapping (currently no selection)
 ##if pMapping == 1:
 vlMapping = clMap1
 
 ## removing current top node tagging:
 ## Anything outside second openning bracket and second last closing 
 ## bracket is considered the top node.
 ## For fun, we call top-node-less tree naked tree

 vStripTree = pParseTree.strip('\n ')

 vNakedTreeStart = vStripTree.find('(', 1)
 vNakedTreeEnd = vStripTree.rfind(')', 0, vStripTree.rfind(')'))
 
 vNakedTree = vStripTree[vNakedTreeStart: vNakedTreeEnd + 1]
 
 vOldTopNodeOpening = vStripTree[0:vNakedTreeStart]
 vOldTopNodeClosing = vStripTree[vNakedTreeEnd + 1:]
 
 
 # mapping tags
 
 vMappedTree = ''
 
 vCurrPos = 0
 while True:
  vTagStart = vNakedTree.find('(', vCurrPos)
  if vTagStart == -1:
   vMappedTree = (vMappedTree + 
                  vNakedTree[vCurrPos:]) 
   break
  else:
   vTagEnd = vNakedTree.find(' ', vTagStart) - 1
   
   vTag = vNakedTree[vTagStart + 1: vTagEnd + 1]
   ## The following acts as a saftey guard against non-recognized tag in
   ## in the input. I.e., any tag not found in the mapping table causes
   ## exception.
   vMappedTag = '(' + [mapping for mapping in vlMapping if mapping[0] == vTag][0][1]
   
   vMappedTree = (vMappedTree + 
                  vNakedTree[vCurrPos: vTagStart] + 
                  vMappedTag +
                  ' ')
   vCurrPos = vTagEnd + 2
   
 
 # adding TOP node
 
 if pflgChangeTopNode:
  vMappedTree = '(' + pTopRootTag.strip() + ' ' + vMappedTree + ')'
 else:
  vMappedTree = vOldTopNodeOpening + vMappedTree + vOldTopNodeClosing
 
 if pflgPrint:
  print vMappedTree
 
 return vMappedTree

