#! /usr/bin/python
# -*- coding: utf-8 -*-

"""	
	This module defines classes related to CoNLL format.
	
	Version 1.1												(17-Aug-2015)
	- CoNLLSentence.depTree() is changed to getDepTree().
	- Some of the migration bugs from v1.0 are fixed.
	
	Version 1.0												(25-May-2015)
	- All CoNLL classes are moved from parse/conll.py
	
"""

import sys
import depparse
from srl import srl


class CoNLLDataLoader:
	'''
	Class for loading data in CoNLL format.
	
	The class loads a CoNLL data set in various formats such as 2006, etc. 
	'''
	
	
	def __init__(self, pDataFilename, pLanguage, pFormatVersion = "2006", pflgVerbos = True):
		'''
		Creates object and loads data from pDataFilename in pFormatVersion
		'''
		
		# final data storage: a list of CoNLLSentence 
		self.conllSentences = []
		self.formatVersion = pFormatVersion
		self.language = pLanguage
		
		try:
			vData = open(pDataFilename).read()
		except IOError:
			raise Exception("Cannot open data file: %s" % pDataFilename)
		
		vlSentLines = []
		for vLine in vData.split("\n"):
			if vLine.strip() == "":
				if len(vlSentLines) > 0:       # sentence end signal
					self.conllSentences.append(self.createCoNLLSentence(vlSentLines, pFormatVersion))
					vlSentLines = []
					
					if pflgVerbos:
						sys.stdout.write('.')
						sys.stdout.flush()
			else:
				vlSentLines.append(vLine.strip())
		
		# just in case the file does not end with a blank line
		if len(vlSentLines) > 0:
			self.conllSentences.append(self.createCoNLLSentence(vlSentLines, pFormatVersion))
			vlSentLines = []
			
			if pflgVerbos:
				sys.stdout.write('.')
				sys.stdout.flush()
		
		if pflgVerbos:
			print "\n%s sentences loaded!" % self.size
		
	
	
	def createCoNLLSentence(self, plLines, pFormatVersion):
		'''
		Creates the appropriate CoNLL sentence object based on the format
		version.
		'''
		
		if pFormatVersion.lower().endswith("input"):
			return CoNLLParserInputSentence(plLines, self.language)
		elif pFormatVersion.endswith("2006"):
			return CoNLL2006Sentence(plLines, self.language)
		elif pFormatVersion.endswith("2008"):
			return CoNLL2008Sentence(plLines, self.language)
		elif pFormatVersion.endswith("2009"):
			return CoNLL2009Sentence(plLines, self.language)
		
	
	
	@property
	def size(self):
		'''
		Returns the size of the data set loaded
		'''
		
		return len(self.conllSentences)
		
	
	
	def getText(self):
		'''
		Returns the surface forms of the loaded sentences
		'''
		
		return '\n'.join([sent.surface for sent in self.conllSentences]) + '\n'
		
	
	
	def selectSentences(self, pMethod = "random", pSize = 0, pIndexes = '', pIndexDelim = ',', pflgSortIdxList = False, pflgComplement = False, pdFilter = {}):
		'''
		Selects and returns sentences from the loaded data based on the
		specified method and details.
		
		pMethon is:
		- random: pSize selected are selected  randomly
		- list: selection is done based on a list of indexes
		
		pFilter is a dictionary of filter options:
		- min-length: minimum length of selected sentences
		- max-length: maximum length of selected sentences
		
		If pflgComplement is set to true and the method is "list", the 
		complement of specified indexes is selected.
		
		If pflgSortIdxList is set to true, the selected sentences with 
		index list method will be in their original order in the data set.
		Otherwise, they will be in the order provided in the list, which 
		is the default.
		'''
		
		## setting up positions of eligible sentences according to filtering
		## criteria in pdFilter
		vlIndexes = []
		for i in range(self.size):
			if "min-length" in pdFilter:
				if self.conllSentences[i] < pdFilter["min-length"]:
					continue
			if "max-length" in pdFilter:
				if self.conllSentences[i] > pdFilter["max-length"]:
					continue
			vlIndexes.append(i + 1)
		
		
		# seleting sentences
		
		if pMethod.lower() in ["rand", "random"]:
			vlSelection = random.sample(vlIndexes, pSize)
			vlSelection.sort()
		elif pMethod.lower() == "list":
			vlIndexes = [int(idx) for idx in pIndexes.split(pIndexDelim)]
			if pflgComplement:
				vlSelection = list(set(range(1, self.size + 1)) - set(vlIndexes))
				vlSelection.sort()
			else:
				vlSelection = list(vlIndexes)
				if pflgSortIdxList:
					vlSelection.sort()
		
		vlSelSentences = []
		for i in vlSelection:
			vlSelSentences.append(self.conllSentences[i - 1])
		
		return vlSelSentences, vlSelection
		
	
	
	def write(self, pOutput):
		'''
		Writes the data into output in the original CoNLL format
		'''
		
		open(pOutput, "w").write("\n\n".join([s.getOrgFormat() for s in self.conllSentences]) + "\n\n")
		
	


class CoNLLSentence:
	'''
	Base class for sentence in CoNLL format
	
	Various CoNLL formats are derived form this class.
	'''
	
	def __init__(self, pLanguage):
		'''
		Creates a sentence object from pLines assuming them in pFormatVersion
		format.
		'''
		
		# A sentence is a list of tokens of CoNLLToken type.
		self.conllTokens = []
		self._depTree = None
		self.language = pLanguage
		
	
	
	@property
	def surfaceTokens(self):
		'''
		Surface tokens of the sentence (see class documentation above)
		'''
		
		vlSurfaceTokens = [(token.id, token.form) for token in self.conllTokens]
		
		## discounting possible duplicate CoNLL tokens with the help of 
		## unique sentence position in vlSurfaceTokens tuple
		vsDistinctTokens = sorted(set(vlSurfaceTokens), key = lambda token: token[0])
		
		return [token[1] for token in vsDistinctTokens]
		
	
	
	@property
	def length(self):
		'''
		Length of the sentence in standard sense (number of words and punctuations)
		
		Sentence length is not the number of tokens (see class documentation
		above).
		'''
		
		return len(self.surfaceTokens)
		
	
	
	@property
	def surface(self):
		'''
		Surface form of the sentence
		'''
		
		return ' '.join(self.surfaceTokens)
		
	
	
	def getCoNLLTokens(self):
		'''
		Returns CoNLL tokens of the sentence
		'''
		
		return self.conllTokens
		
	
	
	def getSpanSurface(self, pSpan):
		'''
		Return the token forms in the given span
		
		Span is a tuple indicating the start and end token positions.
		'''
		
		return ' '.join([t.form for t in self.conllTokens[pSpan[0] - 1 : pSpan[1]]])
		
	
	
	def getSpanCoNLLTokens(self, pSpan):
		'''
		Return the CoNLL tokens in the given span
		
		Span is a tuple indicating the start and end token positions.
		'''
		
		return self.conllTokens[pSpan[0] - 1 : pSpan[1]]
		
	
	
	def getOrgFormat(self, pdParams = {}):
		'''
		Returns the sentence in original CoNLL format
		
		pdParams is the format-specific paramter set.
		'''
		
		return '\n'.join([token.getCoNLLLine(pdParams) for token in self.conllTokens])
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		return self.getOrgFormat()
		
	
	
	def getDepTree(self, pdAnnotationOptions):
		'''
		Returns the dependency tree of the sentence as a DepTree object
		in the pTreeFormat representation format.
		'''
		
		if self._depTree == None:
			self._depTree = depparse.DepTree()
			self._depTree.loadCoNLLDependency(self, pdAnnotationOptions)
		
		return self._depTree
		
	
	
	def mapPOSTagging(self, pdTagMap):
		'''
		Replaces POS tags with their mappings using the specified tag map
		'''
		
		for vToken in self.conllTokens:
			vToken.mapPOSTags(pdTagMap)
		
	


class CoNLLParserInputSentence(CoNLLSentence):
	'''
	Class for sentence in CoNLL format for dependency parser input 
	
	This format is based on CoNLL 2006 format but lacking the dependency 
	annotation.
	'''
	
	def __init__(self, plCoNLLLines, pLanguage):
		'''
		Creates a sentence object from given CoNLL lines
		'''
		
		CoNLLSentence.__init__(self, pLanguage)
		
		for vLine in plCoNLLLines:
			self.conllTokens.append(CoNLLParserInputToken(vLine))
		
	


class CoNLL2006Sentence(CoNLLSentence):
	'''
	Class for sentence in CoNLL 2006 format
	'''
	
	def __init__(self, plCoNLLLines, pLanguage):
		'''
		Creates a sentence object from given CoNLL lines
		'''
		
		CoNLLSentence.__init__(self, pLanguage)
		
		for vLine in plCoNLLLines:
			self.conllTokens.append(CoNLL2006Token(vLine))
		
	
	
	@property
	def version(self):
		'''
		Returns the CoNLL version
		'''
		
		return "2006"
		
	
	
	def toCoNLL2009(self):
		'''
		Converts the sentence to CoNLL 2009 format
		'''
		
		v2009Sentence = CoNLL2009Sentence([], pLanguage = self.language)
		for vToken in self.conllTokens:
			v2009Token = vToken.toCoNLL2009()
			if v2009Token != None:
				v2009Sentence.conllTokens.append(v2009Token)
		
		return v2009Sentence
		
	


class CoNLL2008Sentence(CoNLLSentence):
	'''
	Class for sentence in CoNLL 2008 format
	'''
	
	def __init__(self, plCoNLLLines, pLanguage):
		'''
		Creates a sentence object from given CoNLL lines
		'''
		
		CoNLLSentence.__init__(self, pLanguage)
		
		for vLine in plCoNLLLines:
			self.conllTokens.append(CoNLL2008Token(vLine))
		
	
	
	@property
	def version(self):
		'''
		Returns the CoNLL version
		'''
		
		return "2008"
		
	
	
	@property
	def surfaceTokens(self):
		'''
		Surface tokens of the sentence (see class documentation above)
		'''
		
		## CoNLL 2008 splits on hyphens. We ignore the splits and keep 
		## the whole one.
		vlSurfaceTokens = [(token.id, token.form) for token in self.conllTokens if not (token.form == '_' and token.gPOS == '_' and token.pPOS == '_')]
		
		## discounting possible duplicate CoNLL tokens with the help of 
		## unique sentence position in vlSurfaceTokens tuple
		vsDistinctTokens = sorted(set(vlSurfaceTokens), key = lambda token: token[0])
		
		return [token[1] for token in vsDistinctTokens]
		
	
	
	def toCoNLL2009(self):
		'''
		Converts the sentence to CoNLL 2009 format
		'''
		
		v2009Sentence = CoNLLSentence([], pFormatVersion = "2009")
		for vToken in self.conllTokens:
			v2009Token = vToken.toCoNLL2009()
			if v2009Token != None:
				v2009Sentence.conllTokens.append(v2009Token)
		
		return v2009Sentence
		
	
	
	def toCoNLLParserInput(self, pPOSType = "g"):
		'''
		ToDo: upgrade to work with CoNLLParserInputSentence.
		
		Converts to CoNLL format required by the dependency parsers such 
		as Malt and returns the sentence lines.
		'''
		
		return '\n'.join([token.toEngMaltInput(pPOSType) for token in self.conllTokens])
		
	


class CoNLL2009Sentence(CoNLLSentence):
	'''
	Class for sentence in CoNLL 2009 format
	'''
	
	def __init__(self, plCoNLLLines, pLanguage):
		'''
		Creates a sentence object from given CoNLL lines
		'''
		
		CoNLLSentence.__init__(self, pLanguage)
		
		for vLine in plCoNLLLines:
			self.conllTokens.append(CoNLL2009Token(vLine))
		
		self._srl = None
		## DEPRECATED: use getSRL() argument for annotation option .
		## action to perform on a linguistic constraint violation within 
		## the SRL of the sentence
		self.onSRLConstraintViolation = "exception"
		
	
	
	@property
	def version(self):
		'''
		Returns the CoNLL version
		'''
		
		return "2009"
		
	
	
	@property
	def predicateCount(self):
		'''
		Returns the number of predicates
		'''
		
		#return self.conllTokens[0].fieldNum - 14
		
		return len([t for t in self.conllTokens if t.fillPred])
		
	
	
	@property
	def srl(self):
		'''
		DEPRECIATED: use getSRL()
		
		Returns the semantic role labeling of the sentence
		'''
		
		vlErrStrs = []
		
		if self._srl == None:
			self._srl = srl.SRL(pLanguage = self.language)
			vlErrStrs = self._srl.loadFromCoNLL2009(self, self.onSRLConstraintViolation)
		
		if vlErrStrs != []:
			sys.stderr.write('\n'.join(vlErrStrs) + '\n')
		
		return self._srl
		
	
	
	def getSRL(self, pdAnnotationOptions):
		'''
		Returns the semantic role labeling of the sentence
		'''
		
		vlErrStrs = []
		
		if self._srl == None:
			self._srl = srl.SRL(pLanguage = self.language)
			vlErrStrs = self._srl.loadFromCoNLL2009(self, pdAnnotationOptions = pdAnnotationOptions)
		
		if vlErrStrs != []:
			sys.stderr.write('\n'.join(vlErrStrs) + '\n')
		
		return self._srl
		
	
	
	def generatePropositionTrees(self):
		'''
		DEPRECIATED: use PropositionTree in tk.py with dependency tree
		             loaded from CoNLLSentence.
		
		Generates a tree for each proposition.
		
		A proposition tree has a root which is the predicate frameset. 
		Argument labels and predicate lemma come under the root. Each 
		argument label in turn has the token(s) filling the role as its
		children. Similarly, the predicate lemma has the predicate form
		under it. The order of arguments and predicate corresponds to
		their positions in the sentence.
		
		A PTB bracketing-style tree is generated for each proposition.
		
		Primarily developed for tree kernels
		'''
		
		vlTrees = []
		
		for vProp in self.srl.propositions:
			vlTrees.append(self._generatePropositionTree(vProp))
		
		return vlTrees
		
	
	
	def _generatePropositionTree(self, pProposition):
		'''
		DEPRECIATED: use PropositionTree in tk.py with dependency tree
		             loaded from CoNLLSentence.
		
		Generates proposition tree
		'''
		
		## generating argument nodes while checking for the right place
		## to insert the predicate node
		vlNodes = []
		vflgPredAdded = False
		for vArg in pProposition.arguments:
			if not vflgPredAdded and vArg.span[0] > pProposition.predicate.position:
				vlNodes.append("(%s %s)" % (pProposition.predicate.lemma, self.getSpanSurface((pProposition.predicate.position, pProposition.predicate.position))))
				vflgPredAdded = True
			
			vlNodes.append("(%s %s)" % (vArg.label, self.getSpanSurface(vArg.span)))
			
			# sanity check
			if vArg.span[0] <= pProposition.predicate.position <= vArg.span[1]:
				raise Exception("Problem with predicate and argument positions: %s and (%s, %s) at sentence:\n%s" % (pProposition.predicate.position, vArg.span[0], vArg.span[1], self.getOrgFormat()))
		
		## in case if the predicate is located after all arguments (or 
		## has no argument)
		if not vflgPredAdded:
			vlNodes.append("(%s %s)" % (pProposition.predicate.lemma, self.getSpanSurface((pProposition.predicate.position, pProposition.predicate.position))))
		
		return "(%s %s)" % (pProposition.predicate.frameset, ''.join(vlNodes))
		
	
	
	def loadFromDepTree(self, pDepTree):
		'''
		Loads (converts) the sentence from a given dependency tree
		'''
		
		self.conllTokens = []
		
		for vNode in pDepTree.getNodes():
			# multi-head nodes are converted to two tokens
			for i, vHeadDep in enumerate(vNode.getHeadDeps(), start = 1):
				vCoNLL2009Token = CoNLL2009Token()
				vCoNLL2009Token.loadFromDepNode(vNode, i, pDepTree.getSRLPropCount())
				self.conllTokens.append(vCoNLL2009Token)
		
	
	
	def setLemmasFrom(self, pCoNLL2009Sentence):
		'''
		Sets the lemma and pLemma field values from those of another sentence
		'''
		
		# sanity check
		if len(self.conllTokens) != len(pCoNLL2009Sentence.conllTokens):
			raise Exception("Token count mismatch:\n%s\n%s" % (self.getOrgFormat(), pCoNLL2009Sentence.getOrgFormat()))
		
		for vThisToken, vGivenToken in zip(self.conllTokens, pCoNLL2009Sentence.conllTokens):
			vThisToken.lemma = vGivenToken.lemma
			vThisToken.pLemma = vGivenToken.pLemma
		 
	
	
	def toCoNLLParserInput(self, pPOSType = "g"):
		'''
		ToDo: upgrade to work with CoNLLParserInputSentence.
		
		Converts to CoNLL format required by the dependency parsers such 
		as Malt and returns the sentence lines.
		'''
		
		return '\n'.join([token.toEngMaltInput(pPOSType) for token in self.conllTokens])
		
	


class CoNLLToken:
	'''
	Base Class for token in CoNLL sentence
	
	A token corresponds to a line in CoNLLSentence with attributes like
	form, head, dependency relation, etc. List of attributes are customized
	for various CoNLL format versions in the subclasses.
	'''
	
	def __init__(self, pCoNLLLine = ""):
		'''
		Creates a token object
		'''
		
		if pCoNLLLine == "":
			self.id = None
			self.form = None
			self.lemma = None
			self.fieldNum = 0
		else:
			vlTokenSplit = pCoNLLLine.split()
			
			self.id = int(vlTokenSplit[0])
			self.form = vlTokenSplit[1]
			self.lemma = vlTokenSplit[2]
			
			# we also keep the length of token fields as some foramts are dynamic
			self.fieldNum = len(vlTokenSplit)
		
	

	def __str__(self):
		'''
		String representation of the class
		'''
		
		return self.form
		
	
	
	def getTokenID(self):
		'''
		Returns token ID
		'''
		
		return self.id
		
	
	
	def getForm(self):
		'''
		Returns token form
		'''
		
		return self.form
		
	
	
	def getGoldPOS(self):
		'''
		Returns the gold POS tag
		'''
		
		# note that _getGoldPOS is implemented in the derived classes.
		return self._getGoldPOS()
		
	
	
	def getPredPOS(self):
		'''
		Returns the predicted POS tag
		'''
		
		# note that _getPredPOS is implemented in the derived classes.
		return self._getPredPOS()
		
	
	
	def getGoldHead(self):
		'''
		Returns the gold dependency head
		'''
		
		# note that _getGoldHead is implemented in the derived classes.
		return self._getGoldHead()
		
	
	
	def getPredHead(self):
		'''
		Returns the predicted dependency head
		'''
		
		# note that _getPredHead is implemented in the derived classes.
		return self._getPredHead()
		
	
	
	def getGoldDepRel(self):
		'''
		Returns the gold dependency relation
		'''
		
		# note that _getGoldDepRel is implemented in the derived classes.
		return self._getGoldDepRel()
		
	
	
	def getPredDepRel(self):
		'''
		Returns the predicted dependency relation
		'''
		
		# note that _getPredDepRel is implemented in the derived classes.
		return self._getPredDepRel()
		
	


class CoNLLParserInputToken(CoNLLToken):
	'''
	Class for token in CoNLL input format for dependency parsers
	'''
	
	def __init__(self, pCoNLLLine = ""):
		'''
		Creates a token object
		'''
		
		CoNLLToken.__init__(self, pCoNLLLine)
		
		if pCoNLLLine == "":
			self.coarsePOS = None                  # coarse-grained POS tag
			self.finePOS = None                    # fine-grained POS tag
			self.features = None       
		else:
			vlTokenSplit = pCoNLLLine.split()
			
			self.coarsePOS = vlTokenSplit[3]       # coarse-grained POS tag
			self.finePOS = vlTokenSplit[4]         # fine-grained POS tag
			self.features = vlTokenSplit[5]
		
	
	
	def getCoNLLLine(self, pdParams = {}):
		'''
		Returns the token line in its original CoNLL format
		
		pdParams is the format-specific paramter set.
		'''
		
		return '\t'.join([str(self.id), self.form, self.lemma, self.coarsePOS, self.finePOS, self.features])
		
	
	
	def mapPOSTags(self, pdTagMap):
		'''
		Replaces POS tag with its mapping using the specified tag map
		'''
		
		if self.coarsePOS not in pdTagMap:
			raise Exception("%s does not exists in the map!" % self.coarsePOS)
		else:
			self.coarsePOS = pdTagMap[self.coarsePOS]
		
		if self.finePOS not in pdTagMap:
			raise Exception("%s does not exists in the map!" % self.finePOS)
		else:
			self.finePOS = pdTagMap[self.finePOS]
		
	


class CoNLL2006Token(CoNLLToken):
	'''
	Class for token in CoNLL 2006 sentence
	'''
	
	def __init__(self, pCoNLLLine = ""):
		'''
		Creates a token object
		'''
		
		CoNLLToken.__init__(self, pCoNLLLine)
		
		if pCoNLLLine == "":
			self.coarsePOS = None                  # coarse-grained POS tag
			self.finePOS = None                    # fine-grained POS tag
			self.features = None       
			self.head = None
			self.depRel = None
			self.projHead = None                   # projective head
			self.projDepRel = None                 # projective dependency relation
		else:
			vlTokenSplit = pCoNLLLine.split()
			
			self.coarsePOS = vlTokenSplit[3]       # coarse-grained POS tag
			self.finePOS = vlTokenSplit[4]         # fine-grained POS tag
			self.features = vlTokenSplit[5]
			self.head = int(vlTokenSplit[6])
			self.depRel = vlTokenSplit[7]
			if vlTokenSplit[8] == '_':            # projective head
				self.projHead = None
			else:
				self.projHead = int(vlTokenSplit[8])   
			self.projDepRel = vlTokenSplit[9]          # projective dependency relation
		
	
	
	def getCoNLLLine(self, pdParams = {}):
		'''
		Returns the token line in its original CoNLL format
		
		pdParams is the format-specific paramter set.
		'''
		
		return '\t'.join([str(self.id), self.form, self.lemma, self.coarsePOS, self.finePOS, self.features, str(self.head), self.depRel, self.projHead if self.projHead != None else '_', self.projDepRel])
		
	
	
	def toCoNLL2009(self):
		'''
		Converts to CoNLL 2009 format
		'''
		
		v2009Token = CoNLL2009Token()
		
		v2009Token.id = self.id
		v2009Token.form = self.form
		v2009Token.lemma = self.lemma
		v2009Token.pLemma = self.lemma
		v2009Token.POS = self.coarsePOS
		v2009Token.pPOS = self.coarsePOS
		v2009Token.features = self.features
		v2009Token.pFeatures = '_'
		v2009Token.head = self.head
		v2009Token.pHead = self.head
		v2009Token.depRel = self.depRel
		v2009Token.pDepRel = self.depRel
		v2009Token.fillPred = False
		v2009Token.pred = '_'
		v2009Token.args = []
		
		v2009Token.fieldNum = 14
		
		return v2009Token
		
	
	
	def _getGoldPOS(self):
		'''
		Returns the gold POS tag
		'''
		
		return self.finePOS
		
	
	
	def _getPredPOS(self):
		'''
		Returns the predicted POS tag
		'''
		
		# note that _getPredPOS is implemented in the derived classes.
		return self.finePOS
		
	
	
	def _getGoldHead(self):
		'''
		Returns the gold dependency head
		'''
		
		# note that _getGoldHead is implemented in the derived classes.
		return self.head
		
	
	
	def _getPredHead(self):
		'''
		Returns the predicted dependency head
		'''
		
		# note that _getPredHead is implemented in the derived classes.
		return self.head
		
	
	
	def _getGoldDepRel(self):
		'''
		Returns the gold dependency relation
		'''
		
		# note that _getGoldDepRel is implemented in the derived classes.
		return self.depRel
		
	
	
	def _getPredDepRel(self):
		'''
		Returns the predicted dependency relation
		'''
		
		# note that _getPredDepRel is implemented in the derived classes.
		return self.depRel
		
	
	
	def mapPOSTags(self, pdTagMap):
		'''
		Replaces POS tag with its mapping using the specified tag map
		'''
		
		if self.coarsePOS not in pdTagMap:
			raise Exception("%s does not exists in the map!" % self.coarsePOS)
		else:
			self.coarsePOS = pdTagMap[self.coarsePOS]
		
		if self.finePOS not in pdTagMap:
			raise Exception("%s does not exists in the map!" % self.finePOS)
		else:
			self.finePOS = pdTagMap[self.finePOS]
		
	


class CoNLL2008Token(CoNLLToken):
	'''
	Class for token in CoNLL 2008 sentence
	'''
	
	def __init__(self, pCoNLLLine = ""):
		'''
		Creates a token object
		'''
		
		CoNLLToken.__init__(self, pCoNLLLine)
		
		if pCoNLLLine == "":
			self.gPOS = None                        # gold POS tag
			self.pPOS = None                        # predicted POS tag
			self.sForm = None                       # split form 
			self.sLemma = None                      # lemma of split form
			self.psPOS = None                       # predicted POS tags of split form
			self.head = None                        # gold dependency head
			self.depRel = None                      # dependency relation
			self.pred = None                        # predicate and sense
			self.args = None                        # arguments
		else:
			vlTokenSplit = pCoNLLLine.split()
			
			self.gPOS = vlTokenSplit[3]
			self.pPOS = vlTokenSplit[4]
			self.sForm = vlTokenSplit[5]
			self.sLemma = vlTokenSplit[6]
			self.psPOS = vlTokenSplit[7]
			
			## test sets may not contain the remaining fields even as 
			## placeholders
			if self.fieldNum > 8:
				self.head = int(vlTokenSplit[8])
				self.depRel = vlTokenSplit[9]
				self.pred = vlTokenSplit[10]
			else:
				self.head = '_'
				self.depRel = '_'
				self.pred = '_'
			
			## now extracting arguments if they exist:
			## Each argument token is represented by a tupe (argument type, predicate order)
			## e.g. (A1, 2) means the token is the Arg1 of the 2nd predicate in the sentence.
			self.args = []
			if self.fieldNum > 11:
				for i, vArg in enumerate(vlTokenSplit[11:]):
					if vArg != '_':
						self.args.append((vArg, i + 1))
		
	
	
	def getCoNLLLine(self, pdParams = {}):
		'''
		Returns the token line in its original CoNLL format
		'''
		
		# putting up together argument fields
		vlArgs = []
		for i in range(self.fieldNum - 11):
			vlArg = [arg for (arg, pred) in self.args if pred == i + 1]
			if len(vlArg) == 0:
				vlArgs.append('_')
			else:
				vlArgs.append(vlArg[0])
		
		return '\t'.join([str(self.id), self.form, self.lemma, self.gPOS, self.pPOS, self.sForm, self.sLemma, self.psPOS, str(self.head), self.depRel, self.pred] + vlArgs)
		
	
	
	def toCoNLL2009(self):
		'''
		Converts to CoNLL 2009 format
		
		CoNLL 2008 splits on hyphens which is not the case in 2009. We 
		do not address this here as it poses problem with token ID and 
		consequently with heads.
		'''
		
		v2009Token = CoNLL2009Token()
		
		v2009Token.id = self.id
		v2009Token.form = self.form
		v2009Token.lemma = self.lemma
		v2009Token.pLemma = self.lemma
		v2009Token.POS = self.gPOS
		v2009Token.pPOS = self.pPOS
		v2009Token.features = '_'
		v2009Token.pFeatures = '_'
		v2009Token.head = self.head
		v2009Token.pHead = self.head
		v2009Token.depRel = self.depRel
		v2009Token.pDepRel = self.depRel
		if self.pred == '_':
			v2009Token.fillPred = False
		else:
			v2009Token.fillPred = True
		v2009Token.pred = self.pred
		v2009Token.args = self.args
		
		v2009Token.fieldNum = self.fieldNum + 3
		
		return v2009Token
		
	
	
	def toCoNLLParserInput(self, pPOSType = "g"):
		'''
		ToDo: upgrade to work with CoNLLParserInputToken.
		
		Converts to CoNLL format required by the dependency parsers such 
		as Malt and returns the token line
		
		pPOSType specifies whether gold or predicted POS field will be used.
		'''
		
		if pPOSType.lower().startswith('g'):
			vPOS = self.gPOS
		else:
			vPOS = self.pPOS
		
		return '\t'.join([str(self.id), self.form, '_', vPOS, vPOS, '_'])
		
	
	
	def _getGoldPOS(self):
		'''
		Returns the gold POS tag
		'''
		
		# note that _getGoldPOS is implemented in the derived classes.
		return self.gPOS
		
	
	
	def _getPredPOS(self):
		'''
		Returns the predicted POS tag
		'''
		
		# note that _getPredPOS is implemented in the derived classes.
		return self.pPOS
		
	
	
	def _getGoldHead(self):
		'''
		Returns the gold dependency head
		'''
		
		# note that _getGoldHead is implemented in the derived classes.
		return self.head
		
	
	
	def _getPredHead(self):
		'''
		Returns the predicted dependency head
		'''
		
		# note that _getPredHead is implemented in the derived classes.
		return self.head
		
	
	
	def _getGoldDepRel(self):
		'''
		Returns the gold dependency relation
		'''
		
		# note that _getGoldDepRel is implemented in the derived classes.
		return self.depRel
		
	
	
	def _getPredDepRel(self):
		'''
		Returns the predicted dependency relation
		'''
		
		# note that _getPredDepRel is implemented in the derived classes.
		return self.depRel
		
	


class CoNLL2009Token(CoNLLToken):
	'''
	Class for token in CoNLL 2009 sentence
	'''
	
	
	def __init__(self, pCoNLLLine = ""):
		'''
		Creates a token object
		'''
		
		CoNLLToken.__init__(self, pCoNLLLine)
		
		if pCoNLLLine == "":
			self.pLemma = None               # predicted lemma
			self.POS = None                  # gold major (coarse) POS tag
			self.pPOS = None                 # predicted major POS tag
			self.features = None             # gold morphological features
			self.pFeatures = None            # predicted morphological features
			self.head = None                 # gold dependency head
			self.pHead = None                # predicted dependency head
			self.depRel = None               # dependency relation
			self.pDepRel = None              # predicted dependency relation
			self.fillPred = None             # Is the token a predicated?
			self.pred = None                 # predicate and sense
			self.args = None                 # arguments
		else:
			vlTokenSplit = pCoNLLLine.split()
			
			self.pLemma = vlTokenSplit[3]
			self.POS = vlTokenSplit[4]
			self.pPOS = vlTokenSplit[5]
			self.features = vlTokenSplit[6]
			self.pFeatures = vlTokenSplit[7]
			self.head = int(vlTokenSplit[8])
			self.pHead = int(vlTokenSplit[9])
			self.depRel = vlTokenSplit[10]
			self.pDepRel = vlTokenSplit[11]
			if vlTokenSplit[12] == 'Y':
				self.fillPred = True
			elif vlTokenSplit[12] in ['-', '_']:
				self.fillPred = False
			else:
				raise Exception("Unknown symbol in FillPred field: %s" % vlTokenSplit[12])
			self.pred = vlTokenSplit[13]
			# some sanity check
			if self.pred in ['-', '_'] and self.fillPred == True:
				raise Exception("Incompatible FillPred and Pred fields: %s %s" % (vlTokenSplit[12], vlTokenSplit[13]))
			elif self.pred not in ['-', '_'] and self.fillPred == False:
				self.fillPred = True               # particularly for treating Classic manual annotation which misses FillPred values
			
			## now extracting arguments:
			## Each argument token is represented by a tupe (argument type, predicate order)
			## e.g. (A1, 2) means the token is the Arg1 of the 2nd predicate in the sentence.
			self.args = []
			if self.fieldNum > 14:
				for i, vArg in enumerate(vlTokenSplit[14:]):
					if vArg != '_':
						self.args.append((vArg, i + 1))
		
	
	
	def getPredicateLabel(self):
		'''
		Returns the predicate label
		
		If the token is not a predicate, the default value in the field 
		will be returned
		'''
		
		return self.pred
		
	
	
	def getCoNLLLine(self, pdParams = {}):
		'''
		Returns the token line in its original CoNLL format
		
		pdParams is the format-specific paramter set.
		'''
		
		# putting up together argument fields
		vlArgs = []
		for i in range(self.fieldNum - 14):
			vlArg = [arg for (arg, pred) in self.args if pred == i + 1]
			if len(vlArg) == 0:
				vlArgs.append('_')
			else:
				vlArgs.append(vlArg[0])
		
		# setting FillPred field
		if self.fillPred and ("no-pred-fill" not in pdParams or pdParams["no-pred-fill"] == False):
			vFillPred = 'Y'
		else:
			vFillPred = '_'
		
		return '\t'.join([str(self.id), self.form, self.lemma, self.pLemma, self.POS, self.pPOS, self.features, self.pFeatures, str(self.head), str(self.pHead), self.depRel, self.pDepRel, vFillPred, self.pred] + vlArgs)
		
	
	
	def toEngMaltInput(self, pPOSType = "g"):
		'''
		Converts to English Maltparser input format and returns the token 
		line
		
		See http://www.maltparser.org/mco/english_parser/engmalt.html for
		the format.
		
		pPOSType specifies whether gold or predicted POS field will be used.
		'''
		
		if pPOSType.lower().startswith('g'):
			vPOS = self.POS
		else:
			vPOS = self.pPOS
		
		return '\t'.join([str(self.id), self.form, '_', vPOS, vPOS, '_'])
		
	
	
	def _getGoldPOS(self):
		'''
		Returns the gold POS tag
		'''
		
		# note that _getGoldPOS is implemented in the derived classes.
		return self.POS
		
	
	
	def _getPredPOS(self):
		'''
		Returns the predicted POS tag
		'''
		
		# note that _getPredPOS is implemented in the derived classes.
		return self.pPOS
		
	
	
	def _getGoldHead(self):
		'''
		Returns the gold dependency head
		'''
		
		# note that _getGoldHead is implemented in the derived classes.
		return self.head
		
	
	
	def _getPredHead(self):
		'''
		Returns the predicted dependency head
		'''
		
		# note that _getPredHead is implemented in the derived classes.
		return self.pHead
		
	
	
	def _getGoldDepRel(self):
		'''
		Returns the gold dependency relation
		'''
		
		# note that _getGoldDepRel is implemented in the derived classes.
		return self.depRel
		
	
	
	def _getPredDepRel(self):
		'''
		Returns the predicted dependency relation
		'''
		
		# note that _getPredDepRel is implemented in the derived classes.
		return self.pDepRel
		
	
	
	def isPredicate(self):
		'''
		Returns true if the token is a predicate, false otherwise
		'''
		
		return self.fillPred
		
	
	
	def loadFromDepNode(self, pDepNode, pHeadDepIdx, pPropCount):
		'''
		Loads (converts) the token from a given dependency node
		
		pHeadDepIdx specifies which head-dependency relation of the node
		should be used, because nodes may have multiple heads. The index
		starts from 1.
		
		pPropCount helps set the field number which will be used later
		'''
		
		# loading generic fields can be moved to CoNLLToken
		self.id = pDepNode.position
		self.form = pDepNode.form
		self.lemma = '_'
		
		self.pLemma = '_'
		self.POS = pDepNode.posTag
		self.pPOS = pDepNode.posTag
		self.features = '_'
		self.pFeatures = '_'
		
		self.head = pDepNode.headDeps[pHeadDepIdx - 1][0]
		self.pHead = pDepNode.headDeps[pHeadDepIdx - 1][0]
		self.depRel = pDepNode.headDeps[pHeadDepIdx - 1][1]
		self.pDepRel = pDepNode.headDeps[pHeadDepIdx - 1][1]
		
		if pDepNode.isPredicate():
			self.fillPred = True
		else:
			self.fillPred = False
		
		if pDepNode.isPredicate():
			self.pred = pDepNode.getPredicateFrameset()
		else:
			self.pred = '_'
		
		self.args = [(r, n) for r, n in pDepNode.getRolePredOridinals()]
		
		self.fieldNum = 14 + pPropCount
		
	
	
	def toCoNLLParserInput(self, pPOSType = "g"):
		'''
		ToDo: upgrade to work with CoNLLParserInputToken.
		
		Converts to CoNLL format required by the dependency parsers such 
		as Malt and returns the token line
		
		pPOSType specifies whether gold or predicted POS field will be used.
		'''
		
		if pPOSType.lower().startswith('g'):
			vPOS = self.gPOS
		else:
			vPOS = self.pPOS
		
		return '\t'.join([str(self.id), self.form, '_', vPOS, vPOS, '_'])
		
	
