#! /usr/bin/python
# -*- coding: utf-8 -*-

"""	
	This module defines classes related to dependency parsing.
	
	Version 1.7											   (08-07-2016)
	- lowerWords() is added to DepTree and lowerWordForm() to DepNode.
	
	Version 1.6								(04-May-2016 to 04-May-2016)
	- DepNode.getPathTo() is amended to handle root nodes as destination. 
	
	Version 1.5								(01-Apr-2016 to 06-Apr-2016)
	- DepNode.getHeadChain() is added.
	- DepNode.isNegated() is added.
	- DepTree.getPOSs() is added.
	
	Version 1.4												(10-Mar-2016)
	- DepNode.getPathTo() is edited to return the list of nodes from the 
	  source to the lowest common ancestor (LCA) of the source and destination 
	  nodes and the list of nodes from the LCA to the destination node.
	- DepNode.getDepRelPathTo() is added which functions the same as 
	  the previous version of the getPathTo().
	
	Version 1.3                                             (19-Feb-2016)
	- getPathTo() is added to DepNode.
	
	Version 1.2                                             (12-Nov-2015)
	- pdAnnotationOptions is added to DepParseLoader.loadCoNLLDeps().
	- pflgDeepSRL is added to DepTree.deepCopy().
	
	Version 1.1								(12-Aug-2015 to 21-Aug-2015)
	- setForm(), setDepRel(), setPOSTag(), deepCopy() are added to DepNode.
	- deepCopy() is added to DepTree.
	- Error in DepParseLoader.loadCoNLLDeps() is fixed.
	- Fine POS in DepNode._loadCoNLL2006Annotation() is made default.
	- DepTree._createNewTree() and DepNode._createNewNode() are added.
	- DepNode.getPosition() is added.
	
	Version 1.0								(25-May-2015 to 02-Jun-2015)
	- DepTree, DepNode, DepEval, DepParseLoader are moved here from dloader.py.
"""


import conll


class DepTree:
	'''
	Dependency tree
	
	A dependency tree is a list of nodes of type DepNode. At index 0 of 
	the list is the dummy root node.
	
	Note that the dependency tree is implemented independent of any format
	like CoNLLSentence to be used more generally. Otherwise, nodes could 
	simply store links to CoNLLSentence tokens to get details.
	
	The tree can be non-projective with nodes having more than one heads.
	
	There is a method for comparing the tree with another one and computing
	LAS and UAS.
	'''
	
	def __init__(self, pLanguage = ''):
		'''
		Constructor
		'''
		
		self.nodes = []                                # list of nodes
		self.srl = None                                # SRL annotation
		# TODO: take the SRL out and use inherit another class like SRLDepTree.
		
		self.language = pLanguage
		
	
	
	@property
	def root(self):
		'''
		Returns the dummy root node of the tree.
		'''
		
		return self.nodes[0]
		
	
	
	@property
	def topNode(self):
		'''
		Returns the top node of the tree which is the dependent of the
		dummy root node.
		'''
		
		return self.nodes[self.nodes[0].dependents[0]]
		
	
	
	@property
	def height(self):
		'''
		Returns the height of the tree.
		
		The height is the size of longest branch of the tree.
		'''
		
		return self.topNode.getHeight()
		
	
	
	@property
	def surface(self):
		'''
		Returns the surface form of the corresponding sentence
		'''
		
		return ' '.join([n.form for n in self.nodes[1 : ]])
		
	
	
	def _createNewTree(self, pLanguage = ''):
		'''
		Creates and returns a new tree
		'''
		
		return DepTree(pLanguage = pLanguage)
		
	
	
	def getNodes(self):
		'''
		Returns nodes
		'''
		
		return self.nodes[1 : ]
		
	
	
	def getPOSs(self):
		'''
		Returns the list of POS tags of the tree which matches the list of tokens
		'''
		
		return [n.getPOSTag() for n in self.getNodes()]
		
	
	
	def __str__(self):
		'''
		String representation of the class
		
		By default it's the PTB bracketing representation.
		'''
		
		return "\n\n".join([n.__str__() for n in self.nodes])
		
	
	
	def loadCoNLLDependency(self, pCoNLLSentence, pdAnnotationOptions):
		'''
		Creates the dependency tree by loading the dependency structure 
		from a CoNLLSentence object to a tree represented by a list of 
		nodes of DepNode type.
		
		pdAnnotationOptions is a dictionary which specifies various options
		in loading annotation. It is format-dependent. 
		
		One possible option explicitly used in this method is 
		on-srl-constraint-violation which specifies the action to be 
		taken when a constraint found violated during loading SRL. For 
		possible values see, dloader.CoNLL2009Sentence.
		
		For other options follow each method call using this option.
		
		'''
		
		vlCoNLLTokens = pCoNLLSentence.conllTokens
		
		# creating root head (dummy root) node
		self.nodes.append(DepNode(pDepTree = self))
		
		## looping through tokens to create nodes for each but without 
		## children 
		for vToken in vlCoNLLTokens:
			## If the token position (id) in sentence is the same as previous
			## node's token's, it is a duplicate (token with multiple head)
			## and the new head and dependency should be added to the previous 
			## node.
			if vToken.id == self.nodes[-1].position:
				self.nodes[-1]._loadCoNLLAnnotation(vToken, pdAnnotationOptions)
			else:
				vForm = vToken.form
				vPosition = vToken.id
				
				vNode = DepNode(pDepTree = self, pForm = vForm, pPosition = vPosition)
				vNode._loadCoNLLAnnotation(vToken, pdAnnotationOptions)
				
				self.nodes.append(vNode)
		
		## looping through nodes to add them to the children (dependents)
		## of their parent (head)
		for i in range(1, len(self.nodes)):
			# looping though heads (parents)
			for vHead, vDepRel in self.nodes[i].headDeps:
				self.nodes[vHead].dependents.append(i) 
		
		# loading SRL annotation if the format supports
		if pCoNLLSentence.__class__.__name__ in ["CoNLL2008Sentence", "CoNLL2009Sentence"]:
			self.loadSRL(pCoNLLSentence.getSRL(pdAnnotationOptions = pdAnnotationOptions))
		
	
	
	def loadSRL(self, pSRL):
		'''
		Loads semantic role labeling from SRL object
		
		NOTE: it currently assumes that the input SRL is of dependency type
		      (see SRL class). A procedure is needed to check its type and
		      adapt to dependency if it's in other formalisms (see 
		      ContTree.loadSRL()). 
		
		It both loads the whole annotation into the srl attribute (of type
		SRL) and the token annotations into nodes.
		'''
		
		# 1. loading the whole SRL
		
		self.srl = pSRL
		
		# 2. removing the current annotation
		
		for vNode in self.getNodes():
			vNode.remSRL()
		
		# 3. loading token annotaions
		
		for vProp in pSRL.propositions:
			for vArg in vProp.arguments:
				self.nodes[vArg.span[1]].predRoles.append((vProp.predicate.position, vArg.label))
			
			# NOTE: a predicate node itself is assigned (0, frameset) 
			self.nodes[vProp.predicate.position].predRoles.append((0, vProp.predicate.frameset))
		
	
	
	def loadPARC700Dependency(self, pPARC700Sentence):
		'''
		Creates the dependency tree by loading the dependency structure 
		from a PARC700Sentence object to a tree represented by a list of 
		nodes of DepTree type.
		
		It handles MISSINGHEAD as a normal node and duplicate tokens as
		nodes with multiple head.
		'''
		
		## NOTE: The position of MISSINGHEAD is -1 in the PARC700Sentence.
		##       Since it is not a valid list index, the token is moved 
		##       to the end of sentence and its dependents head to its
		##       new position.
		## NOTE: This may not be efficient to loop the sentence separately 
		##       to fix MISSINGHEAD's dependents, but it's clearer.
		if pPARC700Sentence.parcTokens[0].position == -1:
			# moving
			vlParcTokens = pPARC700Sentence.parcTokens[1:] + pPARC700Sentence.parcTokens[0:1]
			# setting new position
			vlParcTokens[-1].position = vlParcTokens[-2].position + 1
			# adjusting dependents' heads
			for i in range(len(vlParcTokens)):
				if vlParcTokens[i].head == -1:
					vlParcTokens[i].head = vlParcTokens[-1].position
		else:
			vlParcTokens = pPARC700Sentence.parcTokens
		
		# creating root head (dummy root) node first
		self.nodes.append(DepNode(pDepTree = self))
		
		## looping through tokens to create nodes for each (even for MISSINGHEAD)
		## but without children 
		for vToken in vlParcTokens:
			## If the token position in sentence is the same as previous
			## node's token's, it is a duplicate (token with multiple head)
			## and the only job is to add new head and dependency to the 
			## previous node.
			if vToken.position == self.nodes[-1].position:
				self.nodes[-1].headDeps.append((vToken.head, vToken.depRel))
			else:
				vForm = vToken.form
				vPosition = vToken.position
				vlHeadDeps = [(vToken.head, vToken.depRel)]
				vPOSTag = vToken.posTag
				self.nodes.append(DepNode(pDepTree = self, pForm = vForm, pPosition = vPosition, plHeadDeps = vlHeadDeps, pPOSTag = vPOSTag))
		
		## looping through nodes to add them to the children (dependents)
		## of their parent (head)
		for i in range(1, len(self.nodes)):
			# looping though heads (parents)
			for vHead, vDepRel in self.nodes[i].headDeps:
				self.nodes[vHead].dependents.append(i) 
		
	
	
	def deepCopy(self, pflgDeepSRL = False):
		'''
		NOTE: it seems python deepcopy() works better. Try the idea before
		using this method.
		
		Creates and returns a new dependency tree which is a deep copy of 
		the current tree
		
		The SRL component of the tree will be optionally deep copied. Otherwise
		the SRL of the current tree will be assigned to the new tree, so must be 
		used with CARE.
		'''
		
		vCopyTree = self._createNewTree(pLanguage = self.language)
		
		if self.srl and pflgDeepSRL:
			vCopyTree.srl = self.srl.deepCopy()
		else:
			vCopyTree.srl = self.srl
		
		for vNode in self.nodes:
			vCopyTree.nodes.append(vNode.deepCopy(self))
		
		return vCopyTree
		
	
	
	def getPTBTree(self):
		'''
		DEPRECATED: USE tk.DepKTree
		
		Returns the tree in PTB bracketing format.
		
		NOTE: for nodes with multiple head, only the first one is considered.
		
		An internal recursive function is used.
		'''
		
		# defining the function
		def _writePTBTree(pNode, pCurrentHead):
			'''
			Recursively writes the PTB bracketing representation of the 
			given node.
			
			pCurrentHead identifies which head is calling this method in 
			case the node has multiple heads.
			'''
			
			## In sum, a bracket is opened for a node followed by its 
			## dependency relation to currently calling head (pCurrentHead).
			## Then if it has children, a bracket is opened and followed 
			## by its token form followed by the representation of the 
			## children closed by two bracket. Otherwise the token form 
			## follows and the bracket is closed.
			if len(pNode.headDeps) == 0:
				vPTBTree = "( "
			else:
				vPTBTree = "(" + [headDep[1] for headDep in pNode.headDeps if headDep[0] == pCurrentHead][0] + ' '
			
			if len(pNode.dependents) == 0:
				vPTBTree += pNode.form + ')'
			else:
				vPTBTree += '(' + pNode.form + ' ' 
				for vDependent in pNode.dependents:
					vPTBTree += _writePTBTree(self.nodes[vDependent], pNode.position)
				vPTBTree += '))'
			
			return vPTBTree
			
		
		
		# writing the tree
		vPTBTree = _writePTBTree(self.root, 0)
		
		# stripping the outter brackets and returning
		return vPTBTree[2:-1]
		
	
	
	def getTopNodePOS(self):
		'''
		Returns the POS tag of the top node (the dependent of the dummy 
		root node). 
		'''
		
		return self.topNode.posTag
		
	
	
	def getDepRelTagset(self):
		'''
		Returns the dependency relation tag set of the tree.
		
		Tag set is a dictionary of dependency relations and their counts. 
		'''
		
		vdDepRelTags = {}
		for vNode in self.nodes:
			for head, rel in vNode.headDeps:
				if rel in vdDepRelTags:
					vdDepRelTags[rel] += 1
				else:
					vdDepRelTags[rel] = 1
		
		return vdDepRelTags
		
	
	
	def getFormRelPairs(self):
		'''
		Returns the pairs of surface form of each node and its dependency
		relation to its head.
		
		Since a node may have multiple heads, for each head of a node a
		separate pair will be returned.
		'''
		
		vlFormRelPairs = []
		
		for vNode in self.nodes:
			vlFormRelPairs += vNode.getFormRelPairs()
		
		return vlFormRelPairs
		
	
	
	def compareWith(self, pAnotherDepTree, pflgIgnorePunc = False, pflgCaseSensitive = False, pflgCheckWordMismatch = True):
		'''
		Compares the tree with pAnotherDepTree and returns labeled attachment
		score (LAS), unlabeled attachment score (UAS), label accuracy score
		(LAC) together with the counts associated with these scores (to
		be used in computing the scores for a whole document) and fine-grained
		match counts for each dependency relation type.
		
		Match count for a dependency relation includes number of peer nodes
		with that relation to the same heads (true positives), number of
		nodes with that relations in this tree without matchs in pAnotherDepTree
		(fp), number of nodes with that relations in pAnotherDepTree 
		without matchs in this tree (fn), and number of peer nodes with
		that relation (acc) without taking heads into account.
		
		Among output, it returns a string which describes any error seen
		in the trees.
		'''
		
		vLA = 0                        # overall LAS
		vdFLA = {}                     # fine-grained LAS
		vUA = 0                        # overall UAS
		
		
		## French trees produced by Const2Dep may have an extra MISSINGHEAD
		## node, which is ignored here
		vThisTreeNodes = [n for n in self.nodes[1:] if not n.isMISSINGHEAD()]
		vAnotherTreeNodes = [n for n in pAnotherDepTree.nodes[1:] if not n.isMISSINGHEAD()]
		
		vLength1 = len(vThisTreeNodes)
		vLength2 = len(vAnotherTreeNodes)
		if vLength1 != vLength2:
			return "Length mismatch: %s vs. %s" % (vLength1, vLength2), None, None, None, None, None, None, None, None
		
		## Note that since a node can have multiple heads, the overall scores
		## cannot be computed by divding to the lengths. So we count the
		## number of comparisons
		vCount = 0
		
		# dummy root node (index 0) is skipped
		for vNode1, vNode2 in zip(vThisTreeNodes, vAnotherTreeNodes):
			if pflgCheckWordMismatch and vNode1.form != vNode2.form:
				return "Word mismatch: %s vs. %s" % (vNode1.form, vNode2.form), vCount, None, None, None, None, None, None, None
			# ignoring punctuations
			if pflgIgnorePunc:
				## NOTE: defunct to be implemented later
				## Idea: convert to a list of relations to be ignored instead
				## of only punctuation which can differ depending on tag 
				## set thus hard to handle.
				raise Exception("Not implemented yet!")
			
			for vNode1HeadDep in vNode1.headDeps:
				if pflgCaseSensitive:
					vNode1Rel = vNode1HeadDep[1]
				else:
					vNode1Rel = vNode1HeadDep[1].lower()
				
				## adding relation to dictionary to collect counts if already
				## not there
				if vNode1Rel not in vdFLA:
					vdFLA[vNode1Rel] = {"tp":0, "fp": 0, "fn": 0, "acc": 0}
				
				for vNode2HeadDep in vNode2.headDeps:
					if pflgCaseSensitive:
						vNode2Rel = vNode2HeadDep[1]
					else:
						vNode2Rel = vNode2HeadDep[1].lower()
					
					## adding relation to dictionary to collect counts if
					## already not there
					if vNode2Rel not in vdFLA:
						vdFLA[vNode2Rel] = {"tp":0, "fp": 0, "fn": 0, "acc": 0}
					
					vCount += 1
					
					if vNode1HeadDep[0] == vNode2HeadDep[0] and vNode1Rel == vNode2Rel:
						vdFLA[vNode1Rel]["tp"] += 1
						vdFLA[vNode1Rel]["acc"] += 1
						vUA += 1
					else:
						if vNode1HeadDep[0] == vNode2HeadDep[0]:
							vUA += 1
						if vNode1Rel == vNode2Rel:
							vdFLA[vNode1Rel]["acc"] += 1
						
						## False positive and negatives are counted independent 
						## of heads
						vdFLA[vNode1Rel]["fp"] += 1
						vdFLA[vNode2Rel]["fn"] += 1
		
		# calculating fine-grained labeled attachment scores
		for vTag in vdFLA.iterkeys():
			if vdFLA[vTag]["tp"] + vdFLA[vTag]["fp"] == 0:
				# when the tag in the other tree is not in this tree
				vdFLA[vTag]["p"] = 0
			else:
				vdFLA[vTag]["p"] = vdFLA[vTag]["tp"] * 100.0 / (vdFLA[vTag]["tp"] + vdFLA[vTag]["fp"])
			
			if vdFLA[vTag]["tp"] + vdFLA[vTag]["fn"] == 0:
				# when the tag is not in the other tree
				vdFLA[vTag]["r"] = 0     
			else:
				vdFLA[vTag]["r"] = vdFLA[vTag]["tp"] * 100.0 / (vdFLA[vTag]["tp"] + vdFLA[vTag]["fn"])
			
			if vdFLA[vTag]["p"] + vdFLA[vTag]["r"] == 0:
				vdFLA[vTag]["f1"] = 0
			else:
				vdFLA[vTag]["f1"] = 2 * vdFLA[vTag]["p"] * vdFLA[vTag]["r"] / (vdFLA[vTag]["p"] + vdFLA[vTag]["r"])
		
		vLA = sum([s["tp"] for s in vdFLA.itervalues()])
		vLAS = vLA / float(vCount)
		vUAS = vUA / float(vCount)
		vLAcc = sum([s["acc"] for s in vdFLA.itervalues()])
		vLAC = vLAcc / float(vCount)
		
		return "", vCount, vLA, vUA, vLAcc, vdFLA, vLAS, vUAS, vLAC
		
	
	
	def getSpanSurface(self, pSpan):
		'''
		Returns the token forms in the given span
		
		Span is a tuple indicating the start and end token positions.
		'''
		
		return ' '.join([n.form for n in self.nodes[pSpan[0] : pSpan[1] + 1]])
		
	
	
	def getNode(self, pPosition):
		'''
		Returns the node at the given position
		
		pPosition is the position of the token in the sentence (1-based)
		corresponding to the node.
		'''
		
		if pPosition < len(self.nodes):
			return self.nodes[pPosition]
		else:
			raise Exception("Invalid dependency node position: %s" % pPosition)
		
	
	
	def getSentLength(self):
		'''
		Returns the length of the corresponding sentence
		'''
		
		return len(self.nodes) - 1
		
	
	
	def toCoNLL2009Sentence(self):
		'''
		Converts the tree to a CoNLL2009Sentence object and returns
		'''
		
		vCoNLL2009Sent = dloader.CoNLL2009Sentence([], self.language)
		
		vCoNLL2009Sent.loadFromDepTree(self)
		
		return vCoNLL2009Sent
		
	
	
	def getSRLPropCount(self):
		'''
		Returns the number of propositions in the SRL of the tree if it 
		is already loaded
		'''
		
		return len([n for n in self.getNodes() if n.isPredicate()])
		
	
	
	def getPredOrdinal(self, pPredPosition):
		'''
		Returns the chronological number of the predicate at a given position 
		among all predicates of tree's SRL.
		'''
		
		vOrdinal = 0
		
		for i, vNode in enumerate(self.getNodes(), start = 1):
			if vNode.isPredicate():
				vOrdinal += 1
				if i == pPredPosition:
					return vOrdinal
		
		# sanity check
		raise Exception("No predicate found at position %s!" % pPredPosition)
		
	
	
	def lowerWords(self):
		'''
		Converts the word forms to lowercase 
		'''
		
		for vNode in self.getNodes():
			vNode.lowerWordForm()
		
	
	


class DepNode:
	'''
	Dependency Tree
	
	A node contains list of positions of its parents (the heads) and children
	(the dependents) in the tree list. It also has other attributes like
	dependency relation to its head, word from, etc.
	'''
	
	
	def __init__(self, pDepTree = None, pForm = "", pPosition = 0, plHeadDeps = None, plDependents = None, pPOSTag = "", plPredRoles = None):
		'''
		Constructor
		
		plHeadDeps is list of tuples of heads of the node and the dependency
		relation to each head. For example [(2, OBJ), (10, MOD)] means
		that the node modifies two heads in positions 2 and 10 in the sentence
		and thus the tree list as their OBJ and MOD.
		
		Defaults are for root head (the dummy root) node.
		'''
		
		self.depTree = pDepTree                    # dependency tree the node belongs to
		
		self.form = pForm                          # token surface form
		self.position = pPosition                  # token position in the sentence
		
		if plHeadDeps == None:                    # list of head and dependency tuples
			self.headDeps = []
		else:
			self.headDeps = plHeadDeps
		
		if plDependents == None:                  # children
			self.dependents = []
		else:
			self.dependents = plDependents             
		
		self.posTag = pPOSTag                      # POS tag
		
		if plPredRoles == None:                   # list of (predicate position, semantic role) tuples
			self.predRoles = []                    ## NOTE: a predicate node itself is assigned (0, frameset)
		else:
			self.predRoles = plPredRoles
		
	
	
	def _createNewNode(self, pDepTree = None, pForm = "", pPosition = 0, plHeadDeps = None, plDependents = None, pPOSTag = "", plPredRoles = None):
		'''
		Creates and returns an new node
		'''
		
		return DepNode(pDepTree = pDepTree,
		               pForm = self.form,
		               pPosition = self.position,
		               plHeadDeps = self.headDeps[:],
		               plDependents = self.dependents[:],
		               pPOSTag = self.posTag,
		               plPredRoles = self.predRoles[:])
		
	
	
	def deepCopy(self, pDepTree):
		'''
		NOTE: it seems python deepcopy() works better. Try the idea before
		using this method.
		
		Creates and returns a new dependency node which is a deep copy of 
		the current node
		'''
		
		return self._createNewNode(pDepTree = pDepTree,
		                           pForm = self.form,
		                           pPosition = self.position,
		                           plHeadDeps = self.headDeps[:],
		                           plDependents = self.dependents[:],
		                           pPOSTag = self.posTag,
		                           plPredRoles = self.predRoles[:])
		
	
	
	def _loadCoNLLAnnotation(self, pToken, pdAnnotationOptions):
		'''
		Loades node annotation by calling appropriate method based on token
		format version.
		'''
		
		if pToken.__class__.__name__ == "CoNLL2006Token":
			self._loadCoNLL2006Annotation(pToken, pdAnnotationOptions)
		elif pToken.__class__.__name__ == "CoNLL2008Token":
			self._loadCoNLL2008Annotation(pToken, pdAnnotationOptions)
		elif pToken.__class__.__name__ == "CoNLL2009Token":
			self._loadCoNLL2009Annotation(pToken, pdAnnotationOptions)
		
	
	
	def _loadCoNLL2006Annotation(self, pToken, pdAnnotationOptions):
		'''
		Loades node annotation from CoNLL 2006 token.
		
		pdAnnotationOptions specifies annotation options:
		- fine-or-coarse: whether fine-grained or coarse-grained POS tags
		                  should be loaded (default: fine)
		'''
		
		if pdAnnotationOptions is not None and "dep-rel-case" in pdAnnotationOptions:
			if pdAnnotationOptions["dep-rel-case"].lower().startswith("lower"):
				vDepRel = pToken.depRel.lower()
			elif pdAnnotationOptions["dep-rel-case"].lower().startswith("upper"):
				vDepRel = pToken.depRel.upper()
			elif pdAnnotationOptions["dep-rel-case"].lower() in ["org", "original"]:
				vDepRel = pToken.depRel
			else:
				raise Exception("% is not a valid value for dependency relation case!" % pdAnnotationOptions["dep-rel-case"])
		else:
			vDepRel = pToken.depRel
		
		self.headDeps.append((pToken.head, vDepRel))
		
		if pdAnnotationOptions is not None:
			if "fine-or-coarse" not in pdAnnotationOptions:
				self.posTag = pToken.finePOS
			elif pdAnnotationOptions["fine-or-coarse"].lower().startswith("coarse"):
				self.posTag = pToken.coarsePOS
			else:
				self.posTag = pToken.finePOS
		else:
			self.posTag = pToken.finePOS
		
	
	
	def _loadCoNLL2008Annotation(self, pToken, pdAnnotationOptions):
		'''
		Loades node annotation from CoNLL 2008 token.
		
		pdAnnotationOptions specifies annotation options:
		- gold-or-predicted: whether gold or predicted annotaions (e.g. 
		                     POS, ... ) should be loaded.
		'''
		
		if "dep-rel-case" in pdAnnotationOptions:
			if pdAnnotationOptions["dep-rel-case"].lower().startswith("lower"):
				vDepRel = pToken.depRel.lower()
			elif pdAnnotationOptions["dep-rel-case"].lower().startswith("upper"):
				vDepRel = pToken.depRel.upper()
			elif pdAnnotationOptions["dep-rel-case"].lower() in ["org", "original"]:
				vDepRel = pToken.depRel
			else:
				raise Exception("% is not a valid value for dependency relation case!" % pdAnnotationOptions["dep-rel-case"])
		else:
			vDepRel = pToken.depRel
		
		self.headDeps.append((pToken.pHead, vDepRel))
		
		if pdAnnotationOptions["gold-or-predicted"].lower().startswith("gold"):
			self.posTag = pToken.gPOS
		elif pdAnnotationOptions["gold-or-predicted"].lower().startswith("pred"):
			self.posTag = pToken.pPOS
		
	
	
	def _loadCoNLL2009Annotation(self, pToken, pdAnnotationOptions):
		'''
		Loades node annotation from CoNLL 2009 token.
		
		pdAnnotationOptions specifies annotation options:
		- gold-or-predicted: whether gold or predicted annotaions (e.g. 
		                     head, ... ) should be loaded.
		'''
		
		if "dep-rel-case" in pdAnnotationOptions:
			if pdAnnotationOptions["dep-rel-case"].lower().startswith("lower"):
				vDepRel = pToken.depRel.lower()
			elif pdAnnotationOptions["dep-rel-case"].lower().startswith("upper"):
				vDepRel = pToken.depRel.upper()
			elif pdAnnotationOptions["dep-rel-case"].lower() in ["org", "original"]:
				vDepRel = pToken.depRel
			else:
				raise Exception("% is not a valid value for dependency relation case!" % pdAnnotationOptions["dep-rel-case"])
		else:
			vDepRel = pToken.depRel
		
		if pdAnnotationOptions["gold-or-predicted"].lower().startswith("gold"):
			self.headDeps.append((pToken.head, vDepRel))
			self.posTag = pToken.POS
		elif pdAnnotationOptions["gold-or-predicted"].lower().startswith("pred"):
			self.headDeps.append((pToken.pHead, vDepRel))
			self.posTag = pToken.pPOS
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		vOutput  = "Form .................... %s\n" % self.form
		vOutput += "Position ................ %s\n" % self.position
		vOutput += "Head and dependencies ... %s\n" % ','.join(["[%s: %s]" % (headDep[0], headDep[1]) for headDep in self.headDeps])
		vOutput += "POS Tag ................. %s\n" % self.posTag
		vOutput += "Dependents .............. %s\n" % ','.join([str(i) for i in self.dependents])
		vOutput += "Predicate and roles ..... %s"   % ','.join(["[%s: %s]" % (predRole[0], predRole[1]) for predRole in self.predRoles])
		
		return vOutput
		
	
	
	def getDepCount(self):
		'''
		Returns the number of dependents of the node.
		'''
		
		return len(self.dependents)
		
	
	
	def getDependentNodes(self):
		'''
		Returns the dependent nodes of the node.
		'''
		
		return [self.depTree.nodes[dependent] for dependent in self.dependents]
		
	
	
	def getDepPositions(self):
		'''
		Returns the positions of the dependents of the node
		'''
		
		return self.dependents
		
	
	
	def getDepDepRels(self):
		'''
		Returns the dependency relations of the node to its dependents.
		
		Note that this is not the dependency relation(s) of the node to
		its head(s)
		'''
		
		vlDepDepRels = []
		for vDependent in self.dependents:
			# dependency relations of dependent to its heads
			vDepRels = self.depTree.nodes[vDependent].headDeps
			# selecting only relations to this node
			vlDepDepRels.append([rel for head, rel in vDepRels if head == self.position][0])
		
		return vlDepDepRels
		
	
	
	def getHeight(self):
		'''
		Returns the height of the subtree under this node.
		
		The height of a tree with only one node is 0.
		
		The height is the size of longest branch of the tree.
		'''
		
		if len(self.dependents) == 0:
			return 0
		else:
			vMaxDepHeight = 0
			for vDepNode in self.getDependentNodes():
				vHeight = vDepNode.getHeight()
				if vHeight > vMaxDepHeight:
					vMaxDepHeight = vHeight
			
			return vMaxDepHeight + 1
		
	
	
	def getFormRelPairs(self):
		'''
		Returns the pair of surface form of the node and its dependency
		relation to its head.
		
		Since a node may have multiple heads, for each head a separate 
		pair will be returned. Therefore, the returned value is a list of
		tuples.
		'''
		
		vlFormRelPairs = []
		
		for vHeadDep in self.headDeps:
			vlFormRelPairs.append((self.form, vHeadDep[1]))
		
		return vlFormRelPairs
		
	
	
	def getPosition(self):
		'''
		Returns the position of the node in the sentence
		'''
		
		return self.position
		
	
	
	def getForm(self):
		'''
		Returns the form of the node token
		'''
		
		return self.form
		
	
	
	def setForm(self, pValue):
		'''
		Sets the value of the form of the node token
		'''
		
		self.form = pValue
		
	
	
	def getHeadPos(self):
		'''
		Returns the positions of heads of this node
		
		All the heads of multihead nodes are returned.
		'''
		
		return [headDep[0] for headDep in self.headDeps]
		
	
	
	def getHeadDeps(self):
		'''
		Returns head-dependency list
		'''
		
		return self.headDeps
		
	
	
	def getDepRel(self, pHead = None):
		'''
		Returns the dependency relation of the node to its head.
		
		For multihead nodes, the first head is returned unless pHead is 
		given.
		'''
		
		if pHead:
			return [headDep[1] for headDep in self.headDeps if headDep[0] == pHead][0]
		else:
			return self.headDeps[0][1]
		
	
	
	def setDepRel(self, pRel, pHead = None):
		'''
		Sets the dependency relation of the node to its given head
		
		If no head is given, it assumes the first head
		'''
		
		if pHead:
			for i, vHeadDep in enumerate(self.headDeps):
				if vHeadDep[0] == pHead:
					self.headDeps[i] = (self.headDeps[i][0], pRel)
		else:
			self.headDeps[0] = (self.headDeps[0][0], pRel)
		
	
	
	def getHeadNode(self):
		'''
		Returns the head node of the node (the first head in case of multiple
		heads)
		'''
		
		return self.depTree.nodes[self.headDeps[0][0]]
		
	
	
	def getHeadChain(self):
		'''
		Returns the list of nodes in the head chain of this node up to the root (inclusive)
		
		The first head is used in case of multiple heads.
		'''
		
		if self.isRoot():
			return []
		else:
			return [self.getHeadNode()] + self.getHeadNode().getHeadChain() 
		
	
	
	def getPOSTag(self):
		'''
		Returns the POS tag of the node
		'''
		
		return self.posTag
		
	
	
	def setPOSTag(self, pValue):
		'''
		Sets the value of the POS tag of the node
		'''
		
		self.posTag = pValue
		
	
	
	def isRoot(self):
		'''
		Returns true if the node is root node of the tree
		'''
		
		return (self.position == 0)
		
	
	
	def isPredicate(self):
		'''
		Returns true if the node is a predicate
		'''
		
		return len([p for p, r in self.predRoles if p == 0]) > 0
		
	
	
	def isArgument(self):
		'''
		Returns true if the node is a an argument for any predicate in the
		SRL of the sentence.
		'''
		
		return len([r for p, r in self.predRoles if p != 0]) > 0
		
	
	
	def getPredicateFrameset(self):
		'''
		Returns the predicate frameset from the semantic role labeling 
		of the node
		'''
		
		if self.isPredicate():
			return [r for p, r in self.predRoles if p == 0][0]
		else:
			return ''
		
	
	
	def getPredArgRoles(self):
		'''
		Returns list of (predicte position, sematic role) tuples of the 
		token
		'''
		
		return [(p, r) for p, r in self.predRoles if p != 0]
		
	
	
	def getRolePredOridinals(self):
		'''
		Returns list of (semantic role, predicte oridinal number) tuples 
		of the token
		
		Predicate ordinal number specifies the chronogical number of the 
		predicate among the predicates of the whole SRL
		'''
		
		return [(r, self.depTree.getPredOrdinal(p)) for p, r in self.getPredArgRoles()]
		
	
	
	def getArgRoles(self):
		'''
		Returns the list of argument role labels from the semantic role 
		labeling of the node
		'''
		
		return [r for p, r in self.predRoles if p != 0]
		
	
	
	def isVerb(self):
		'''
		Returns true if node is a verb
		
		It assumes that all verbs have a POS tag starting with V independent
		of tag set and language.
		ToDo: improve to make use of universal POS tags  
		'''
		
		return self.getPOSTag().lower().startswith('v')
		
	
	
	def isMISSINGHEAD(self):
		'''
		Returns true if the node is a MISSINGHEAD node found in the output
		of Const2Dep French dependency converter
		'''
		
		return self.form == "MISSINGHEAD"
		
	
	
	def remSRL(self):
		'''
		Remove SRL annotation
		'''
		
		self.predRoles = []
		
	
	
	def getPathTo(self, pNode):
		'''
		Extracts and returns the path of nodes from this node to the given node
		
		Two lists are returned, one containing nodes from the source to the lowest common ancestor (LCA) of the source 
		and destination nodes and the other containing nodes from the LCA to the destination node.
		
		NOTE: this method is not recursive. It has been designed for efficiency.
		'''
		
		# 1. stacking this node and all its heads up until root (only first head of multiple head cases)
		vlNodeHeads = [self]
		while not vlNodeHeads[-1].getHeadNode().isRoot():
			vlNodeHeads.append(vlNodeHeads[-1].getHeadNode())
		
		# 2. stacking destination node and all its heads up until root
		vlDestHeads = [pNode]
		while not vlDestHeads[-1].getHeadNode().isRoot():
			vlDestHeads.append(vlDestHeads[-1].getHeadNode())
		
		# 3. finding lowest common ancestor (the u-turn point)
		
		# if there is more than one node in either list, match the two list to find the LCA and cut the rest
		if len(vlNodeHeads) > 1 and len(vlDestHeads) > 1:
			for vNPIdx, vDPIdx in zip(reversed(range(len(vlNodeHeads))), reversed(range(len(vlDestHeads)))):
				# if the nodes before these two are the same, this node (at the current index) cannot be LCA
				if vlNodeHeads[vNPIdx - 1] == vlDestHeads[vDPIdx - 1]:
					del vlNodeHeads[vNPIdx]
					del vlDestHeads[vDPIdx]
				# if the nodes before these two are not the same, this node (at the current index) is LCA
				else:
					break
		
		return vlNodeHeads, vlDestHeads
		
	
	
	def getDepRelPathTo(self, pNode):
		'''
		Extracts and returns the path of dependency relations from this node to the given node
		
		It uses getPathTo() to extract the nodes in the path first.
		
		The path is a string consisting of the syntactic label of the node and the directions represented by /, \, > and
		< (up, down, right, left in order). Left and right are used when descending from the lowest common ancestor down
		to the destination node. If the path turns right, \> is used and if left, \<.
		'''
		
		vlSrcToLCA, vlLCAToDest = self.getPathTo(pNode)
		
		# 1. this node to LCA 
		vPath = '/'.join([n.getDepRel() for n in vlSrcToLCA[:-1]])
		# 2. u-turn (if vLCA is not the start or the end)
		if len(vlSrcToLCA) > 1 and len(vlLCAToDest) > 1:
			if vlSrcToLCA[-2].getPosition() < vlLCAToDest[-2].getPosition():
				vPath += '\>'
			else:
				vPath += '\<'
		# 3. LCA to destination node
		vPath += '\\'.join([n.getDepRel() for n in reversed(vlLCAToDest[:-1])])
		
		return vPath
		
	
	
	def getWordsInPathTo(self, pNode):
		'''
		Extracts and returns the word froms in the dependency path from this node to the given node
		
		It used getPathTo() to extract the nodes in the path first.
		'''
		
		vlSrcToLCA, vlLCAToDest = self.getPathTo(pNode)
		
		# 1. this node to LCA 
		vlPathWords = [n.getForm() for n in vlSrcToLCA[:-1]]
		# 2. LCA to destination node
		vlPathWords += [n.getForm() for n in reversed(vlLCAToDest[:-1])]
		
		return vlPathWords
		
	
	
	def isNegated(self, pNegationDepRel = "neg"):
		'''
		Returns ture if the node has a dependent with relation representing the negation
		
		The default negation dependency relationship is 'neg'
		'''
		
		if pNegationDepRel in [c.getDepRel(pHead = self.getPosition()) for c in self.getDependentNodes()]:
			return True
		else:
			return False
		
	
	
	def lowerWordForm(self):
		'''
		Converts the word form to lowercase 
		'''
		
		self.setForm(self.getForm().lower())
		
	
	


class DepEval:
	'''
	Dependency parser evaluation
	'''
	
	def __init__(self, pLanguage):
		'''
		Constructor
		'''
		
		self.golds = []
		self.parses = []
		
		# sentence level counts and scores
		self.sentenceScores = []
		
		self.nodeCount = 0
		self.LAS = 0
		self.UAS = 0
		self.LAC = 0
		self.FLA = {}
		
		self.language = pLanguage
		
	
	def loadGold(self, pGoldsFile, pFormat, pdAnnotationOptions):
		'''
		Loads gold trees
		'''
		
		self.golds = self._loadTrees(pGoldsFile, pFormat, pdAnnotationOptions)
		
	
	def loadParsed(self, pParseFile, pFormat, pdAnnotationOptions):
		'''
		Loads parsed trees
		'''
		
		self.parses = self._loadTrees(pParseFile, pFormat, pdAnnotationOptions)
		
	
	def _loadTrees(self, pTreeFile, pFormat, pdAnnotationOptions):
		'''
		Loads trees
		
		pFormat can be:
		- conllxxxx: when xxxx is the version (year): e.g. conll2006
		- parc700
		'''
		
		if pFormat.lower().startswith("conll"):
			vlTrees = [s.depTree(pdAnnotationOptions) for s in dloader.CoNLLDataLoader(pTreeFile, pLanguage = self.language, pFormatVersion = pFormat[5:9], pflgVerbos = False).conllSentences]
		elif pFormat.lower() == "parc700":
			vlTrees = [s.depTree for s in dloader.PARC700DataLoader(pTreeFile).parcSentences]
		
		return vlTrees
		
	
	def score(self, pflgIgnorePunc = False, pflgCaseSensitive = False, pflgCheckWordMismatch = True):
		'''
		Scores loaded parsed trees with respect to gold trees in terms of 
		LAS, UAS, LAC (label accuracy), and fine-grained tag scores (precision,
		recall, F1) at both document and sentence level.
		'''
		
		vGoldLen = len(self.golds)
		vLen = len(self.parses)
		
		if vGoldLen != vLen:
			sys.exit("Gold length does not match with parsed length: %s vs. %s" % (vGoldLen, vLen))
		
		# sentence level scores
		self.sentenceScores = []
		for i, (vParseTree, vGoldTree) in enumerate(zip(self.parses, self.golds), start = 1):
			vErr, vNCount, vLA, vUA, vLAcc, vdFLA, vLAS, vUAS, vLAC = vParseTree.compareWith(vGoldTree, 
			                                                                                 pflgIgnorePunc = pflgIgnorePunc,
			                                                                                 pflgCaseSensitive = pflgCaseSensitive,
			                                                                                 pflgCheckWordMismatch = pflgCheckWordMismatch)
			
			if vErr != "":
				sys.exit("Tree #%s: %s" % (i, vErr))
			else:
				self.sentenceScores.append({"nodeCount": vNCount,
				                            "LA": vLA,
				                            "UA": vUA,
				                            "LAcc": vLAcc,
				                            "FLA": vdFLA,
				                            "LAS": vLAS,
				                            "UAS": vUAS,
				                            "LAC": vLAC})
		
		# document-level scores
		self._calcDocScores()
		
	
	def	_calcDocScores(self):
		'''
		Calculates document-level LAS, UAS, LAC (label accuracy), and 
		fine-grained labeled attachment scores (precision, recall, F1) 
		from sentence-level scores already calculated.
		'''
		
		self.nodeCount = sum([s["nodeCount"] for s in self.sentenceScores])
		self.LAS = sum([s["LA"] for s in self.sentenceScores]) / float(self.nodeCount)
		self.UAS = sum([s["UA"] for s in self.sentenceScores]) / float(self.nodeCount)
		self.LAC = sum([s["LAcc"] for s in self.sentenceScores]) / float(self.nodeCount)
		
		vlTagSet = self.getTagset()
		
		for vTag in vlTagSet:
			self.FLA[vTag] = {"gfreq": 0, "pfreq": 0, "correct": 0, "precision": 0, "recall": 0, "f1": 0}
			# tag frequency in gold set
			self.FLA[vTag]["gfreq"] = sum([fla["tp"] + fla["fn"] for s in self.sentenceScores for tag, fla in s["FLA"].iteritems() if tag == vTag])
			# tag frequency in parsed set
			self.FLA[vTag]["pfreq"] = sum([fla["tp"] + fla["fp"] for s in self.sentenceScores for tag, fla in s["FLA"].iteritems() if tag == vTag])
			# correct labels (true positives)
			self.FLA[vTag]["correct"] = sum([fla["tp"] for s in self.sentenceScores for tag, fla in s["FLA"].iteritems() if tag == vTag])
			# tag precision
			if self.FLA[vTag]["pfreq"] == 0:
				self.FLA[vTag]["precision"] = 0
			else:
				self.FLA[vTag]["precision"] = self.FLA[vTag]["correct"] * 100.0 / self.FLA[vTag]["pfreq"]
			# tag recall
			if self.FLA[vTag]["gfreq"] == 0:
				self.FLA[vTag]["recall"] = 0
			else:
				self.FLA[vTag]["recall"] = self.FLA[vTag]["correct"] * 100.0 / self.FLA[vTag]["gfreq"]
			# tag F1
			if (self.FLA[vTag]["precision"] + self.FLA[vTag]["recall"]) == 0:
				self.FLA[vTag]["f1"] = 0
			else:
				self.FLA[vTag]["f1"] = 2 * self.FLA[vTag]["precision"] * self.FLA[vTag]["recall"] / (self.FLA[vTag]["precision"] + self.FLA[vTag]["recall"])
			# tag accuracy
			if (self.FLA[vTag]["precision"] + self.FLA[vTag]["recall"]) == 0:
				self.FLA[vTag]["f1"] = 0
			else:
				self.FLA[vTag]["f1"] = 2 * self.FLA[vTag]["precision"] * self.FLA[vTag]["recall"] / (self.FLA[vTag]["precision"] + self.FLA[vTag]["recall"])
		
	
	def getTagset(self):
		'''
		Extracts and returns the set of dependency tags used in the two
		data set (gold and parsed).
		'''
		
		return sorted(list(set([tag for s in self.sentenceScores for tag in s["FLA"]])))
		
	
	def getSentenceLASs(self):
		'''
		Returns list of sentence-level LAS scores
		'''
		
		return [score["LAS"] for score in self.sentenceScores]
		
	
	def getSentenceUASs(self):
		'''
		Returns list of sentence-level UAS scores
		'''
		
		return [score["UAS"] for score in self.sentenceScores]
		
	
	def getSentenceLACs(self):
		'''
		Returns list of sentence-level LAC scores
		'''
		
		return [score["LAC"] for score in self.sentenceScores]
		
	
	def writeScores(self, pOutFilename):
		'''
		Write scores into file.
		'''
		
		vfOut = open(pOutFilename, 'w')
		
		vfOut.write("No.\tLAS\tUAS\tLAC\n")
		vfOut.write("================================\n")
		for i, vSentScores in enumerate(self.sentenceScores, start=1):
			vfOut.write("%s\t%s\t%s\t%s\n" % (i, round(vSentScores["LAS"], 4), round(vSentScores["UAS"], 4), round(vSentScores["LAC"], 4)))
		vfOut.write("================================\n")
		vfOut.write("Sum\t%s\t%s\t%s\n" % (round(self.LAS, 4), round(self.UAS, 4), round(self.LAC, 4)))
		
		vfOut.close()
		
	
	def writeFLAScores(self, pOutFilename, pFormat = "sent"):
		'''
		Write fine-grained labeled attachment scores into file.
		
		pFormat:
		- doc: only document-level scores
		- sent: sentence level scores with their summary
		'''
		
		vfOut = open(pOutFilename, 'w')
		
		if pFormat.lower() in ["sent", "sentence"]:
			self._writeSentFLAScores(vfOut)
		elif pFormat.lower() in ["doc", "document"]:
			self._writeDocFLAScores(vfOut)
		
		vfOut.close()
		
	
	def _writeSentFLAScores(self, pfOutFile):
		'''
		Write sentence-level fine-grained labeled attachment scores into 
		file.
		'''
		
		vlTagSet = self.getTagset()
		
		vHeader = "%s\n" % ' '.join(["%6s" % t for t in vlTagSet])
		vLineLen = 8 + len(vHeader.expandtabs())
		
		## precision #######
		
		pfOutFile.write('=' * vLineLen + '\n')
		pfOutFile.write("Precision\n")
		pfOutFile.write('-' * vLineLen + '\n')
		pfOutFile.write("No.\t" + vHeader)
		pfOutFile.write('-' * vLineLen + '\n')
		
		# sentences
		for i, vSentScores in enumerate(self.sentenceScores, start=1):
			vlFLAs = []
			for vTag in vlTagSet:
				if vTag in vSentScores["FLA"]:
					vScoreStr = str(util.padDecimalZero(round(vSentScores["FLA"][vTag]["p"], 2), 2))
				else:
					vScoreStr = "-1"
				vTagLen = len(vTag)
				vlFLAs.append(vScoreStr.rjust(max(vTagLen, 6)))
			pfOutFile.write("%s\t%s\n" % (i, ' '.join(vlFLAs)))
		
		# sum
		pfOutFile.write('-' * vLineLen + '\n')
		vlFLAs = []
		for vTag in vlTagSet:
			vScoreStr = str(util.padDecimalZero(round(self.FLA[vTag]["precision"], 2), 2))
			vTagLen = len(vTag)
			vlFLAs.append(vScoreStr.rjust(max(vTagLen, 6)))
		pfOutFile.write("Sum\t%s\n" % ' '.join(vlFLAs))
		pfOutFile.write('-' * vLineLen + '\n')
		pfOutFile.write("\t" + vHeader)
		pfOutFile.write('=' * vLineLen + '\n')
		
		## recall #######
		
		pfOutFile.write("Recall\n")
		pfOutFile.write('-' * vLineLen + '\n')
		pfOutFile.write("No.\t" + vHeader)
		pfOutFile.write('-' * vLineLen + '\n')
		
		# sentences
		for i, vSentScores in enumerate(self.sentenceScores, start=1):
			vlFLAs = []
			for vTag in vlTagSet:
				if vTag in vSentScores["FLA"]:
					vScoreStr = str(util.padDecimalZero(round(vSentScores["FLA"][vTag]["r"], 2), 2))
				else:
					vScoreStr = "-1"
				vTagLen = len(vTag)
				vlFLAs.append(vScoreStr.rjust(max(vTagLen, 6)))
			pfOutFile.write("%s\t%s\n" % (i, ' '.join(vlFLAs)))
		
		# sum
		pfOutFile.write('-' * vLineLen + '\n')
		vlFLAs = []
		for vTag in vlTagSet:
			vScoreStr = str(util.padDecimalZero(round(self.FLA[vTag]["recall"], 2), 2))
			vTagLen = len(vTag)
			vlFLAs.append(vScoreStr.rjust(max(vTagLen, 6)))
		pfOutFile.write("Sum\t%s\n" % ' '.join(vlFLAs))
		pfOutFile.write('-' * vLineLen + '\n')
		pfOutFile.write("\t" + vHeader)
		pfOutFile.write('=' * vLineLen + '\n')
		
		## F1 #######
		
		#vfOut.write('=' * vLineLen + '\n')
		pfOutFile.write("F1\n")
		pfOutFile.write('-' * vLineLen + '\n')
		pfOutFile.write("No.\t" + vHeader)
		pfOutFile.write('-' * vLineLen + '\n')
		
		# sentences
		for i, vSentScores in enumerate(self.sentenceScores, start=1):
			vlFLAs = []
			for vTag in vlTagSet:
				if vTag in vSentScores["FLA"]:
					vScoreStr = str(util.padDecimalZero(round(vSentScores["FLA"][vTag]["f1"], 2), 2))
				else:
					vScoreStr = "-1"
				vTagLen = len(vTag)
				vlFLAs.append(vScoreStr.rjust(max(vTagLen, 6)))
			pfOutFile.write("%s\t%s\n" % (i, ' '.join(vlFLAs)))
		
		# sum
		pfOutFile.write('-' * vLineLen + '\n')
		vlFLAs = []
		for vTag in vlTagSet:
			vScoreStr = str(util.padDecimalZero(round(self.FLA[vTag]["f1"], 2), 2))
			vTagLen = len(vTag)
			vlFLAs.append(vScoreStr.rjust(max(vTagLen, 6)))
		pfOutFile.write("Sum\t%s\n" % ' '.join(vlFLAs))
		pfOutFile.write('-' * vLineLen + '\n')
		pfOutFile.write("\t" + vHeader)
		pfOutFile.write('=' * vLineLen + '\n')
		
	
	def _writeDocFLAScores(self, pfOutFile):
		'''
		Write document-level fine-grained labeled attachment scores into 
		file.
		'''
		
		vlTagSet = self.getTagset()
		
		pfOutFile.write("Label         Gold  System  Correct  Precision  Recall      F1\n")
		pfOutFile.write("---------------------------------------------------------------\n")
		for vTag in vlTagSet:
			pfOutFile.write("%-11s%7s%8s%9s%11s%8s%8s\n" % (vTag, 
			                                                self.FLA[vTag]["gfreq"],
			                                                self.FLA[vTag]["pfreq"],
			                                                self.FLA[vTag]["correct"],
			                                                util.padDecimalZero(round(self.FLA[vTag]["precision"], 2), 2),
			                                                util.padDecimalZero(round(self.FLA[vTag]["recall"], 2), 2),
			                                                util.padDecimalZero(round(self.FLA[vTag]["f1"], 2), 2)))
		
	
	


class DepParseLoader:
	'''
	Dependency parse loader
	
	NOTE: MOVE TO dloader.py
	
	Loads dependency parse trees into DepTree objects from various input 
	formats. 
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		# list of DepTree
		self.depTrees = None
		
	
	
	def loadCoNLLDeps(self, pCoNLLFile, pLanguage, pVersion = "2006", pdAnnotationOptions = None):
		'''
		Loads the dependency trees in CoNLL format of pVersion.
		'''
		
		vCoNLLData = conll.CoNLLDataLoader(pCoNLLFile, pLanguage = pLanguage, pFormatVersion = pVersion, pflgVerbos = False)
		
		self.depTrees = []
		
		for vSent in vCoNLLData.conllSentences:
			self.depTrees.append(vSent.getDepTree(pdAnnotationOptions = pdAnnotationOptions))
		
	
	
	def loadPARC700Deps(self, pPARC700File):
		'''
		Loads the dependency trees in PARC700 format of pVersion.
		'''
		
		vPARC700Data = PARC700DataLoader(pPARC700File, pflgVerbos = False)
		
		self.depTrees = []
		
		for vSent in vPARC700Data.parcSentences:
			self.depTrees.append(vSent.depTree)
		
	
	
	def getDepTrees(self):
		'''
		Returns the list of dependency trees
		'''
		
		return self.depTrees
		
	
	
	def getDepRelTagset(self):
		'''
		Returns dependency relation tag set of the loaded data.
		
		Tag set is a dictionary of dependency relations and their counts. 
		'''
		
		vdDepRelTags = {}
		for vTree in self.depTrees:
			for tag, count in vTree.getDepRelTagset().iteritems():
				if tag in vdDepRelTags:
					vdDepRelTags[tag] += count
				else:
					vdDepRelTags[tag] = count
		
		return vdDepRelTags
		
	
	
