#! /usr/bin/python

"""	
	This module defines functions to process evalb parse score file.
	
	Version: 1.1											(15-09-2012)
	- Score data structures is renamed to SentScore and globalized together
	  with ScoreLine.
	- extractDocumentScores() is added
	
	Version: 1.0
	- extractSentenceScores()
	- extractSentenceScoreLines()
	
"""

from __future__ import division
from decimal import *
from collections import namedtuple
import sys


'''
	data structure for n-best item
'''
# Document-level scores
DocScores = namedtuple('DocScores', 'recall precision fscore tagacc')
# sentence-level scores
SentScores = namedtuple('SentScores', 'no recall precision fscore tagacc')
# sentence score line
ScoreLine = namedtuple('ScoreLine', 'no scores')


def extractDocumentScores(pEvalbFileName):
	'''
	Extracts document-level scores from pEvalbFileName. 
	'''
	
	# opening evalb score file
	try:
		vfEvalb  = open(pEvalbFileName, 'r')
	except IOError:
		sys.stderr.write('Can\'t open evalb file: ' + pEvalbFileName)
		sys.exit(2)
	
	# ignoring until "Bracketing Recall"
	vLine=vfEvalb.readline()
	while vLine:
		if vLine.startswith("Bracketing Recal"):
			break
		else:
			vLine=vfEvalb.readline()
	
	vLineSplit = vLine.split()
	vRecallStr = vLineSplit[3] 
	vRecall = Decimal(vRecallStr)
		
	vLine=vfEvalb.readline()
	vLineSplit = vLine.split()
	vPrecisionStr = vLineSplit[3] 
	vPrecision = Decimal(vPrecisionStr)
	
	vLine=vfEvalb.readline()
	vLineSplit = vLine.split()
	vFScoreStr = vLineSplit[3] 
	vFScore = Decimal(vFScoreStr)
	
	# ignoring until "Tagging accuracy"
	vLine=vfEvalb.readline()
	while vLine:
		if vLine.startswith("Tagging accuracy"):
			break
		else:
			vLine=vfEvalb.readline()
	vLineSplit = vLine.split()
	vFTagAccStr = vLineSplit[3] 
	vFTagAcc = Decimal(vFTagAccStr)
	
	
	vScores = DocScores(vRecall, vPrecision, vFScore, vFTagAcc)
	
	vfEvalb.close() 
	return vScores
	

def extractSentenceScores(pEvalbFileName, pprnSentenceNo, pprnPrecision, pprnRecall, pprnFScore):
	'''
	Extracts scores from pEvalbFileName sentence by sentence and prints 
	any of them to stdout if requested. 
	'''
	
	vScoresList = []
 
	# opening evalb score file
	try:
		vfEvalb  = open(pEvalbFileName, 'r')
	except IOError:
		sys.stderr.write('Can\'t open evalb file: ' + pEvalbFileName)
		sys.exit(2)
	
	# ignoring the evalb header
	vLine=vfEvalb.readline()
	while vLine:
		# we only check for the first 10 '=' characters to decide on the table header seporator 
		if vLine[:10] == "==========":
			break
		else:
			vLine=vfEvalb.readline()
	
	# processing scores section of evalb
	vLine=vfEvalb.readline()
	while vLine:
		# we only check for the first 10 '=' characters to decide on the table footer seporator 
		if vLine[:10] == "==========":
			break
		else:
			vLineSplit = vLine.split()
		
		# the 1st column is supposed to be the sentence number
		vSentenceNoStr = vLineSplit[0]
		if pprnSentenceNo:
			print vSentenceNoStr + '\t',
		
		# the 4th column is supposed to be the recall
		vRecallStr = vLineSplit[3] 
		vRecall = Decimal(vRecallStr)
		if pprnRecall:
			print vRecallStr + '\t',
		
		# the 5th column is supposed to be the precesion
		vPrecisionStr = vLineSplit[4] 
		vPrecision = Decimal(vPrecisionStr)
		if pprnPrecision:
			print vPrecisionStr + '\t',
		
		# F-Score need to be calculated
		if (vPrecision + vRecall) != 0:
			vFScore = (2 * vPrecision * vRecall) / (vPrecision + vRecall) 
			vFScore = vFScore.quantize(Decimal('0.01'))
		else:
			vFScore = Decimal(0).quantize(Decimal('0.01'))
		vFScoreStr = str(vFScore)
		if pprnFScore:
			print vFScoreStr + '\t',
		
		# putting the values into the return data structure
		vScoresList.append(SentScores(int(vSentenceNoStr), vRecall, vPrecision, vFScore, 0.0)) 
	
		# ending the line
		if pprnFScore or pprnPrecision or pprnRecall or pprnSentenceNo:
			print
		
		# reading the next line
		vLine=vfEvalb.readline()
	
	vfEvalb.close() 
	return vScoresList
	

def extractSentenceScoreLines(pEvalbFileName, pflgPrint):
	'''
	Extracts sentence score lines without caring about score components.
	Each line is returned in two parts: sentence number and a string of 
	scores as is read form file.
	'''
	
	vlScoreLines = []
 
	# opening evalb score file
	try:
		vfEvalb  = open(pEvalbFileName, 'r')
	except IOError:
		sys.stderr.write('Can\'t open evalb file: ' + pEvalbFileName)
		sys.exit(2)
   
	# ignoring the evalb header
	vLine=vfEvalb.readline()
	while vLine:
		# we only check for the first 10 '=' characters to decide on the table header seporator 
		if vLine[:10] == "==========":
			break
		else:
			vLine=vfEvalb.readline()
	
	# processing scores section of evalb
	vLine=vfEvalb.readline()
	while vLine:
		# we only check for the first 10 '=' characters to decide on the table footer seporator 
		if vLine[:10] == "==========":
			break
		else:
			vLineSplit = vLine.strip().split(' ', 1)
	
		# the 1st column is supposed to be the sentence number
		vSentenceNoStr = vLineSplit[0]
		if pflgPrint:
			print vSentenceNoStr + '\t',
		
		# the rest is fethced altogether
		vScoresStr = vLineSplit[1]
		if pflgPrint:
			print vScoresStr
		
		# putting the values into the return data structure
		vlScoreLines.append(ScoreLine(int(vSentenceNoStr), vScoresStr))
		# reading the next line
		vLine=vfEvalb.readline()
	
	vfEvalb.close() 
	return vlScoreLines
	
