#! /usr/bin/python
# -*- coding: utf-8 -*-

"""	
	This module defines classes for Foreebank annotation data.
	
	
	Version 1.5								(08-Jun-2015 to 10-Jun-2015)
	- Bug is fixed in ForeeAnnSentence.checkSanity() where code 232 was 
	  returned without checking also for grammatical correction to be empty
	  in addition to the surface correction to compare the corrected parse
	  tree's surface with the original sentence.
	- Sanity error codes 241, 251, 252, 253 are added.
	- extractUnaryNodes() and extractBadNPStucts() are added to ForeeConstTree 
	  and ForeeConstNode.
	
	Version 1.4								(03-Jun-2015 to 04-Jun-2015)
	- Bug is fixed in ForeeAnnSentence.checkSanity() where code 221 was 
	  returned instead of 223 which was not defined to be handled in 
	  ForeeAnn.checkSanity().
	- extractSuffixes() is added to ForeeConstTree and ForeeConstNode.
	
	Version 1.3								(26-May-2015 to 28-May-2015)
	- ForeeAnn.loadForeeAnn() is changed to optionally add new data to the
	  existing ones instead of overwriting them. ForeeAnnSentence.loadData()
	  is accordingly changed to set new sentence ID as provided instead 
	  of using the ones in the data block.
	- Error 213 is added to sanity check.
	- ForeeConstTree and ForeeConstNode are added.
	- removeSuffix() and hasSuffixedLabel() are moved from ConstTree and
	  ConstNode in constparse to ForeeConstTree and ForeeConstNode here.
	  All ConstTree instances are changed to ForeeConstTree.
	- mergeBrokenTokenNodes() is added to ForeeConstTree and ForeeConstNode.
	- getSuffixes() is added to ForeeConstNode.
	
	Version 1.2								(12-May-2015 to 25-May-2015)
	- ForeePOSTagging and ForeePOSTaggedToken are added.
	- replaceSurfaceTokens(), replaceGramFormTokens(), replaceCrrParseTerminalTokens(),
	  isGramCorrected() are added to ForeeAnnSentence.
	
	Version 1.1												(05-May-2015)
	- Bug in ForeebankSentence.size is fixed. It returned the character
	  size before instead of token size.
	- getSuffixSeqList() is added to ForeebankSentence.
	
	Version 1.0												(30-Dec-2014)
	- getCorrectGrammars() is added to ForeeAnn.
	- getGrammaticalForm() and getCorrectGrammar() are added to ForeeAnnSentence.
	
	Version 0.9												(27-Nov-2014)
	- applySurfaceCorrection() is added to ForeeAnnSentence.
	- applySurfaceCorrections() and replaceAnnotations() are added to ForeeAnn.
	
	Version 0.8												(03-Nov-2014)
	- ForeeAnnIAA is added.
	- ed(), getSurfaceLen() and getGrammaticalFormLen() are added to the 
	  ForeeAnnSentence.
	- Typo in all getCorrectedParsesSurface() name (Sufrace instead of Surface)
	  is fixed.
	
	Version 0.7												(23-Sep-2014)
	- getCorrectedParse() and getCorrectedConstTree() are edited to optionally
	  delete pro-drop nodes (e.g. *PRO* in PTB).
	- getCorrectedParsesSurface(), getCorrectForms() are added to ForeeAnn.
	- getCorrectedParseSurface(), getCorrectForms(), getSurfaceForm() 
	  are added to ForeeAnnSentence.
	
	Version 0.6												(18-Sep-2014)
	- getCorrectedForms() and replaceAutoParses() are added ForeeAnn.
	- setSurfaceForm() and setAutoParse() are added to ForeeAnnSentence.
	- setSurfaceStatus(), setComprehensibility(), setConfidence() are 
	  changed to assign values themselvs instead of returning it.
	- ForeeAnnSentence.loadData() is changed to have the values assigned
	  to the attributes in the set methods.
	- Sentence ID is added to ForeeAnnSentence and the sentence no is loaded
	  from the data set.
	- getStr() and __str__() are added to ForeeAnn.
	
	Version 0.5												(03-Sep-2014)
	- ForeeAnnSentence.getStr() is updated to write multiline comments in
	  one line.
	
	Version 0.4												(15-Aug-2014)
	- addSentence(), extractSubset(), getSentences(), replaceSentences(),
	  checkSanity() are added to ForeeAnn.
	- checkSanity(), getAutoConstTree(), getCorrectedConstTree() are added 
	  to ForeeAnnSentence.
	
	Version 0.3												(07-Aug-2014)
	- getCorrectedParseTagset(), getAutoParseTagset(), getSuffixTags(),
	  extractSuffixStat() are added to ForeeAnn and ForeeAnnSentence.
	- size(), getAutoParse(), getCorrectedParse() are added to ForeeAnnSentence.
	- getTokenSize(), getAutoParses(), getCorrectedParses() are added to
	  ForeeAnn.
	
	Version 0.2												(18-Jul-2014)
	- Foreebank and ForeebankSentence are renamed to ForeeAnn and ForeeAnnSentence.
	
	Version 0.1												(23-Jan-2014)
	- Foreebank and ForeebankSentence are added.
	
"""


import re
import constparse, pos
from utils import util


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# ForeeAnn
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class ForeeAnn:
	'''
	Class for Foreebank annotation data
	
	A Foreebank annotation object represents a data set containing Foreebank 
	sentences with annotations.
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		self.sentences = []
		
	
	
	def getSentences(self):
		'''
		Returns all the sentences in the dataset
		'''
		
		return self.sentences
		
	
	
	@property
	def size(self):
		'''
		Returns the size of the annotation data set which is the number 
		of the sentences
		'''
		
		return len(self.sentences)
		 
	
	
	def getTokenSize(self):
		'''
		Returns the token size of the annotation data set which is the 
		number of tokens is the surface form of the sentences, i.e. the 
		sum of the sentence sizes
		'''
		
		return sum([s.size for s in self.sentences])
		 
	
	
	def createForeeAnn(pTextFilename, pParseFilename, pContextIdxFilename, pOrgFormFilename, pOutFilename):
		'''
		Given required resources such as source sentences and their automatic 
		parse trees, creates a Foreebank annotation data set.
		
		pContextIdxFilename contains the line position of each sentence in the
		original corpus from which the sample is taken (where the original 
		context is in place and no line is dropped yet).
		
		pOrgFormFilename is the parallel corpus to pTextFilename where each
		sentence in the latter is the original form (right after stripping of
		HTML thus before tokenizing and cleaning) of its peer in the former.
		
		It overwrites current content.
		'''
		
		self.sentences = []
		
		vlSents = open(pTextFilename).read().strip().split('\n')
		vlParses = open(pParseFilename).read().strip().split('\n')
		vlContexIdxs = open(pContextIdxFilename).read().strip().split('\n')
		vlOrgForms = open(pOrgFormFilename).read().strip().split('\n')
		
		for vSent, vParse, vIdx , vOrg in zip(vlSents, vlParses, vlContexIdxs, vlOrgForms):
			self.sentences.append(ForeebankSentence())
			self.sentences[-1].setSurface(vSent)
			self.sentences[-1].setOrgForm(vOrg)
			self.sentences[-1].setContextIdx(vIdx)
			self.sentences[-1].setAutomaticParse(vParse)
		
	
	
	def loadForeeAnn(self, pAnnFilename, pflgOverwrite = True, pflgNewIDs = False):
		'''
		Loads Foreebank data from annotation (.ann) file
		
		It overwrites current content if the pflgOverwrite is true. It also
		generates new sentence IDs (No) if pflgNewIDs is true which is a
		good thing when overwrite is not true and loading is done multiple 
		times.
		'''
		
		if pflgOverwrite:
			self.sentences = []
		
		vfAnn = open(pAnnFilename)
		
		vlBlock = []
		for vLine in vfAnn:
			if re.match("No\t+\d+$", vLine) and len(vlBlock) > 0:
				self.sentences.append(ForeeAnnSentence())
				if pflgNewIDs:
					self.sentences[-1].loadData(''.join(vlBlock), pNewID = str(len(self.sentences)))
				else:
					self.sentences[-1].loadData(''.join(vlBlock))
				vlBlock = [vLine]
			else:
				vlBlock.append(vLine)
		
		# the last block
		self.sentences.append(ForeeAnnSentence())
		self.sentences[-1].loadData(''.join(vlBlock))
		
		vfAnn.close()
		
	
	
	def __str__(self):
		'''
		String representation of the sentence
		'''
		
		return self.getStr(pflgGenerateNewIds = False)
		
	
	
	def getStr(self, pflgGenerateNewIds = True):
		'''
		Writes the data into a string in Foreebank annotation format
		If pflgGenerateNewIds is set to true, new IDs will be genrated
		based on the position of the sentence in the data set and the 
		original IDs will be used otherwise. For backward-compatibility,
		the default is true.
		'''
		
		vStr = ''
		
		for i, vSent in enumerate(self.sentences, start = 1):
			if pflgGenerateNewIds:
				vSStr = "No\t\t\t%s\n" % i
				vSStr += vSent.getStr(pflgWriteID = False)
			else:
				vSStr = vSent.getStr(pflgWriteID = True)
			
			vStr += (vSStr + '\n')
		
		return vStr
		
	
	
	def writeForeeAnn(self, pOutFilename, pflgGenerateNewIds = True):
		'''
		Writes data into Foreebank annotation file
		
		If pflgGenerateNewIds is set to true, new IDs will be generated
		based on the position of the sentence in the data set and the 
		original IDs will be used otherwise. For backward-compatibility,
		the default is true.
		'''
		
		open(pOutFilename, 'w').write(self.getStr(pflgGenerateNewIds = pflgGenerateNewIds))
		
	
	
	def surfaceStatusStat(self):
		'''
		Returns the statistics of surface status in the Foreebank
		'''
		
		vdStat = {"ok": 0,
		          "unclean": 0,
		          "tokenisation": 0,
		          "segmentation": 0,
		          "removed": 0}
		
		for vSent in self.sentences:
			vdStat[vSent.getSurfaceStatusStr()] += 1
		
		return vdStat
		
	
	
	def getSurfaceCorrectedSentences(self):
		'''
		Returns surface-corrected sentences together with their indexes 
		in a list of tuples
		'''
		
		return [(i, s) for i, s in enumerate(self.sentences, start = 1) if s.getCorrectedForm() != '']
		
	
	
	def getCorrectedForms(self):
		'''
		Returns the surface form of the sentences which have been corrected
		together with their indexes in a list of tuples
		'''
		
		return [(i, s.getCorrectedForm()) for i, s in enumerate(self.sentences, start = 1) if s.getCorrectedForm() != '']
		
	
	
	def comprehensStat(self):
		'''
		Returns the statistics of comprehensibility in the Foreebank
		'''
		
		vdStat = {"comprehensible": 0,
		          "context required": 0,
		          "incomprehensible": 0}
		
		for vSent in self.sentences:
			vdStat[vSent.getComprehensibilityStr()] += 1
		
		return vdStat
		
	
	
	def getGramCorrectedSentences(self):
		'''
		Returns grammatically-corrected sentences together with their 
		indexes in a list of tuples
		'''
		
		return [(i, s) for i, s in enumerate(self.sentences, start = 1) if s.grammaticalForm != '']
		
	
	
	def getParseCorrectedSentences(self):
		'''
		Returns parse-corrected sentences together with their indexes in
		a list of tuples
		'''
		
		return [(i, s) for i, s in enumerate(self.sentences, start = 1) if s.automaticParse != s.correctedParse]
		
	
	
	def getParseCorrectSentences(self):
		'''
		Returns sentences with correct automatic parses together with their
		indexes in a list of tuples
		'''
		
		return [(i, s) for i, s in enumerate(self.sentences, start = 1) if s.automaticParse == s.correctedParse]
		
	
	
	def confidenceStat(self):
		'''
		Returns the statistics of annotator confidence in the Foreebank
		'''
		
		vdStat = {"sure": 0,
		          "fairly sure": 0,
		          "half sure": 0,
		          "highly uncertain": 0,
		          "impossible": 0}
		
		for vSent in self.sentences:
			vdStat[vSent.getConfidenceStr()] += 1
		
		return vdStat
		
	
	
	def getCommentedSentences(self):
		'''
		Returns sentences with comment field filled together with indexes
		in a list of tuples
		'''
		
		return [(i, s) for i, s in enumerate(self.sentences, start = 1) if s.comment.strip() != '']
		
	
	
	def select(self, pdCriteria):
		'''
		Selects and returns list of sentences, together with their indexes,
		whose field value match ANY of the specified criteria.
		
		The criteria is specified in a dictionary with keys containing 
		sentence annotation fields and values contain list of eligible 
		values for that field.
		'''
		
		vlSelection = []
		
		for i, vSent in enumerate(self.sentences, start = 1):
			for vCField in pdCriteria:
				if vCField.lower().startswith("num") or vCField.lower() == "no":
					if i in pdCriteria[vCField]:
						if i not in [j for j, s in vlSelection]:
							vlSelection.append((i, vSent))
				elif vCField.lower().startswith("sentence"):
					if vSent.surface in pdCriteria[vCField]:
						if i not in [j for j, s in vlSelection]:
							vlSelection.append((i, vSent))
				elif vCField.lower().startswith("original"):
					if vSent.orgForm in pdCriteria[vCField]:
						if i not in [j for j, s in vlSelection]:
							vlSelection.append((i, vSent))
				elif vCField.lower().startswith("context"):
					if vSent.contextIdx in pdCriteria[vCField]:
						if i not in [j for j, s in vlSelection]:
							vlSelection.append((i, vSent))
				elif vCField.lower().startswith("surface"):
					if vSent.getSurfaceStatusStr() in pdCriteria[vCField]:
						if i not in [j for j, s in vlSelection]:
							vlSelection.append((i, vSent))
				elif vCField.lower().startswith("corrected"):
					if vSent.correctedForm in pdCriteria[vCField]:
						if i not in [j for j, s in vlSelection]:
							vlSelection.append((i, vSent))
				elif vCField.lower().startswith("comprehensibility"):
					if vSent.getComprehensibilityStr() in pdCriteria[vCField]:
						if i not in [j for j, s in vlSelection]:
							vlSelection.append((i, vSent))
				elif vCField.lower().startswith("grammatical"):
					if vSent.grammaticalForm in pdCriteria[vCField]:
						if i not in [j for j, s in vlSelection]:
							vlSelection.append((i, vSent))
				elif vCField.lower().startswith("automatic"):
					if vSent.automaticParse in pdCriteria[vCField]:
						if i not in [j for j, s in vlSelection]:
							vlSelection.append((i, vSent))
				elif vCField.lower().startswith("manual"):
					if vSent.correctedParse in pdCriteria[vCField]:
						if i not in [j for j, s in vlSelection]:
							vlSelection.append((i, vSent))
				elif vCField.lower().startswith("confidence"):
					if vSent.getConfidenceStr() in pdCriteria[vCField]:
						if i not in [j for j, s in vlSelection]:
							vlSelection.append((i, vSent))
				elif vCField.lower().startswith("comment"):
					if vSent.comment in pdCriteria[vCField]:
						if i not in [j for j, s in vlSelection]:
							vlSelection.append((i, vSent))
		
		return vlSelection
		
	
	
	def getFieldValues(self, pFieldName):
		'''
		Returns automatic parses in a list
		'''
		
		vlValues = []
		
		for vSent in self.sentences:
			if pFieldName.lower().startswith("sentence"):
				vlValues.append(vSent.surface)
			elif pFieldName.lower().startswith("original"):
				vlValues.append(vSent.orgForm)
			elif pFieldName.lower().startswith("context"):
				vlValues.append(vSent.contextIdx)
			elif pFieldName.lower().startswith("surface"):
				vlValues.append(vSent.getSurfaceStatusStr())
			elif pFieldName.lower().startswith("corrected"):
				vlValues.append(vSent.correctedForm)
			elif pFieldName.lower().startswith("comprehensibility"):
				vlValues.append(vSent.getComprehensibilityStr())
			elif pFieldName.lower().startswith("grammatical"):
				vlValues.append(vSent.grammaticalForm)
			elif pFieldName.lower().startswith("automatic"):
				vlValues.append(vSent.automaticParse)
			elif pFieldName.lower().startswith("manual"):
				vlValues.append(vSent.correctedParse)
			elif pFieldName.lower().startswith("confidence"):
				vlValues.append(vSent.getConfidenceStr())
			elif pFieldName.lower().startswith("comment"):
				vlValues.append(vSent.comment)
		
		return vlValues
		
	
	
	def replaceFieldValues(self, plIdxValueTuples):
		'''
		Replaces the value of the given field in the sentences identified 
		by their index in the data set
		
		The indexes and replacement values are supplied in a list of tuples.
		'''
		
		vlRepIdxs = [idx for idx, value in plIdxValueTuples]
		for i, vSent in enumerate(self.sentences, start = 1):
			if i not in vlRepIdxs:
				continue
			
			vValue = [value for idx, value in plIdxValueTuples if idx == i][0]
			
			if pFieldName.lower().startswith("sentence"):
				vSent.setForm(vValue)
			elif pFieldName.lower().startswith("original"):
				vSent.setOrgForm(vValue)
			elif pFieldName.lower().startswith("context"):
				vSent.setContextIdx(vValue)
			elif pFieldName.lower().startswith("surface"):
				vSent.setSurfaceStatus(vValue)
			elif pFieldName.lower().startswith("corrected"):
				vSent.setCorrectedForm(vValue)
			elif pFieldName.lower().startswith("comprehensibility"):
				vSent.setComprehensibility(vValue)
			elif pFieldName.lower().startswith("grammatical"):
				vSent.setGrammaticalForm(vValue)
			elif pFieldName.lower().startswith("automatic"):
				vSent.setAutoParse(vValue)
			elif pFieldName.lower().startswith("manual"):
				vSent.setCorrectedParse(vValue)
			elif pFieldName.lower().startswith("confidence"):
				vSent.setConfidence(vValue)
			elif pFieldName.lower().startswith("comment"):
				vSent.setComment(vValue)
		
	
	
	def getAutoParseTagset(self):
		'''
		Returns the tag set of the automatic parses of all sentence
		'''
		
		vdTagset = {}
		
		for vSent in self.sentences:
			for vTag, vCount in vSent.getAutoParseTagset().iteritems():
				if vTag in vdTagset:
					vdTagset[vTag] += vCount
				else:
					vdTagset[vTag] = vCount
		
		return vdTagset
		
	
	
	def getCorrectedParseTagset(self, pflgRemSuffixs = False):
		'''
		Returns the tag set of the corrected parses of all sentence
		'''
		
		vdTagset = {}
		
		for vSent in self.sentences:
			for vTag, vCount in vSent.getCorrectedParseTagset(pflgRemSuffixs = pflgRemSuffixs).iteritems():
				if vTag in vdTagset:
					vdTagset[vTag] += vCount
				else:
					vdTagset[vTag] = vCount
		
		return vdTagset
		
	
	
	def getSuffixTags(self):
		'''
		Returns the tags in corrected parses containing suffixes
		'''
		
		vdTagset = {}
		
		for vSent in self.sentences:
			for vTag, vCount in vSent.getSuffixTags().iteritems():
				if vTag in vdTagset:
					vdTagset[vTag] += vCount
				else:
					vdTagset[vTag] = vCount
		
		return vdTagset
		
	
	
	def extractSuffixStat(self):
		'''
		Extracts statistics about suffix tags
		'''
		
		vdSuffixCounts = {}
		vdSuffixedTagCounts = {}
		
		# extracting counts
		
		for vSent in self.sentences:
			vdSentSuffixs, vdSentSuffixedTags = vSent.extractSuffixStat()
			
			for vSuffix, (vCount, vRatio) in vdSentSuffixs.iteritems():
				if vSuffix in vdSuffixCounts:
					vdSuffixCounts[vSuffix] += vCount
				else:
					vdSuffixCounts[vSuffix] = vCount
			
			for vTag, (vCount, vRatio) in vdSentSuffixedTags.iteritems():
				if vTag in vdSuffixedTagCounts:
					vdSuffixedTagCounts[vTag] += vCount
				else:
					vdSuffixedTagCounts[vTag] = vCount
		
		# computing ratios
		
		vdSuffixs = {}
		vdSuffixedTags = {}
		
		vCorpusTokenSize = self.getTokenSize()
		vTagset = self.getCorrectedParseTagset(pflgRemSuffixs = True)
		
		for vTag, vCount in vdSuffixedTagCounts.iteritems():
			vdSuffixedTags[vTag] = (vCount, vCount * 1.0 / vTagset[vTag])
		
		for vSuffix, vCount in vdSuffixCounts.iteritems():
			if vSuffix.endswith('M'):
				vdSuffixs[vSuffix] = (vCount, vCount * 1.0 / self.size)
			else:
				vdSuffixs[vSuffix] = (vCount, vCount * 1.0 / vCorpusTokenSize)
		
		return vdSuffixs, vdSuffixedTags
		
	
	
	def getAutoParses(self):
		'''
		Returns the corrected parses in a list
		'''
		
		vlAutoParses = []
		
		for vSent in self.sentences:
			vlAutoParses.append(vSent.getAutoParse())
		
		return vlAutoParses
		
	
	
	def getCorrectedParses(self, pflgRemSuffixs = False, pflgDelDNodes = False, pflgDelXNodes = False, pflgDelPRONodes = False, pflgMergeBrokenNodes = False):
		'''
		Returns the corrected parses in a list
		
		The tag suffixes can optionally be removed. The nodes inserted 
		manually to account for dropped tokens (_D suffixes) can also 
		optionally be deleted.
		'''
		
		vlCorrectedParses = []
		
		for vSent in self.sentences:
			vlCorrectedParses.append(vSent.getCorrectedParse(pflgRemSuffixs = pflgRemSuffixs, pflgDelDNodes = pflgDelDNodes, pflgDelXNodes = pflgDelXNodes, pflgDelPRONodes = pflgDelPRONodes, pflgMergeBrokenNodes = pflgMergeBrokenNodes))
		
		return vlCorrectedParses
		
	
	
	def getCorrectedParsesSurface(self, pflgRemSuffixs = False, pflgDelDNodes = False, pflgDelXNodes = False, pflgDelPRONodes = False, pflgMergeBrokenNodes = False):
		'''
		Returns the  surface forms of the corrected parses in a list
		'''
		
		vlSurfaces = []
		
		for vSent in self.sentences:
			vlSurfaces.append(vSent.getCorrectedParseSurface(pflgRemSuffixs = pflgRemSuffixs, pflgDelDNodes = pflgDelDNodes, pflgDelXNodes = pflgDelXNodes, pflgDelPRONodes = pflgDelPRONodes, pflgMergeBrokenNodes = pflgMergeBrokenNodes))
		
		return vlSurfaces
		
	
	
	def getCorrectForms(self):
		'''
		Returns the correct surface forms of the sentences
		
		Correct surface form is the corrected sentence if any and the sentence
		itself otherwise.
		'''
		
		vlSurfaces = []
		
		for vSent in self.sentences:
			vlSurfaces.append(vSent.getCorrectForm())
		
		return vlSurfaces
		
	
	
	def getCorrectGrammars(self):
		'''
		Returns the grammatical form of the sentences
		
		If the grammar is not corrected, the sentence itself is returned.
		'''
		
		vlCorrectGrams = []
		
		for vSent in self.sentences:
			vlCorrectGrams.append(vSent.getCorrectGrammar())
		
		return vlCorrectGrams
		
	
	
	def addSentence(self, pFASentence):
		'''
		Adds a ForeeAnn sentence
		'''
		
		self.sentences.append(pFASentence)
		
	
	
	def extractSubset(self, plIndexes):
		'''
		Extract a subset of dataset containing sentences in the given
		indexes in the data set and returns as a ForeeAnn object
		'''
		
		vFASubset = ForeeAnn()
		
		for i in plIndexes:
			vFASubset.addSentence(self.sentences[i - 1])
		
		return vFASubset
		
	
	
	def replaceSentences(self, plRepIndexes, plRepSentences):
		'''
		Replaces sentences at the given positions with the provided new 
		sentences
		'''
		
		# sanity check
		if len(plRepIndexes) != len(plRepSentences):
			raise Exception("The number of indexes does not match the number of sentences!")
		
		for vIdx, vRepSent in zip(plRepIndexes, plRepSentences):
			self.sentences[vIdx - 1] = vRepSent
		
	
	
	def replaceAnnotations(self, pFieldName, plRepIndexes, plRepAnnotations):
		'''
		Replaces the values of a given annotation field for the sentences 
		at the given positions with the given new values
		'''
		
		# sanity check
		if len(plRepIndexes) != len(plRepAnnotations):
			raise Exception("The number of indexes does not match the number of sentences!")
		
		for vIdx, vValue in zip(plRepIndexes, plRepAnnotations):
			if pFieldName.lower().startswith("sentence"):
				self.sentences[vIdx - 1].setForm(vValue)
			elif pFieldName.lower().startswith("original"):
				self.sentences[vIdx - 1].setOrgForm(vValue)
			elif pFieldName.lower().startswith("context"):
				self.sentences[vIdx - 1].setContextIdx(vValue)
			elif pFieldName.lower().startswith("surface"):
				self.sentences[vIdx - 1].setSurfaceStatus(vValue)
			elif pFieldName.lower().startswith("corrected"):
				self.sentences[vIdx - 1].setCorrectedForm(vValue)
			elif pFieldName.lower().startswith("comprehensibility"):
				self.sentences[vIdx - 1].setComprehensibility(vValue)
			elif pFieldName.lower().startswith("grammatical"):
				self.sentences[vIdx - 1].setGrammaticalForm(vValue)
			elif pFieldName.lower().startswith("automatic"):
				self.sentences[vIdx - 1].setAutoParse(vValue)
			elif pFieldName.lower().startswith("manual"):
				self.sentences[vIdx - 1].setCorrectedParse(vValue)
			elif pFieldName.lower().startswith("confidence"):
				self.sentences[vIdx - 1].setConfidence(vValue)
			elif pFieldName.lower().startswith("comment"):
				self.sentences[vIdx - 1].setComment(vValue)
		
	
	
	def applySurfaceCorrections(self, pflgChangeStatus = False):
		'''
		Replaces the surface forms of the sentences with their corrected 
		forms and deletes the corrected forms and returns the IDs of the
		sentences affected
		
		It applies the correction if the corrected surface field is not 
		empty.
		
		If pflgChangeStatus is set to true, the value of the surface status
		will be changed to OK.
		'''
		
		vlCorrectedIds = []
		
		for vSent in self.sentences:
			if vSent.getCorrectedForm() != '':
				vSent.applySurfaceCorrection(pflgChangeStatus = pflgChangeStatus)
				vlCorrectedIds.append(vSent.id)
		
		return vlCorrectedIds
		
	
	
	def checkSanity(self):
		'''
		Performs some sanity tests on the annotation
		'''
		
		vlProblems = []                          # list of list problems
		
		# finding problems
		
		for i, vSent in enumerate(self.getSentences(), start = 1):
			vlProblems.append(vSent.checkSanity())
		
		# reporting problems
		
		# problem code 111
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 111 in p]
		open("sanity.111", 'w').write("Sentence is corrected but surface status is marked OK: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 112
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 112 in p]
		open("sanity.112", 'w').write("Corrected sentence is the same as the original sentence: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 113
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 113 in p]
		open("sanity.113", 'w').write("Corrected sentence does not match the automatically-parsed sentence: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 121
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 121 in p]
		open("sanity.121", 'w').write("Tokenization issue is marked but the sentence is not corrected: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 211
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 211 in p]
		open("sanity.211", 'w').write("Corrected grammar is the same as the original or corrected sentence: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 212
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 212 in p]
		open("sanity.212", 'w').write("Grammar is corrected but the changes are not marked on the corrected parse tree: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 213
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 213 in p]
		open("sanity.213", 'w').write("The length of the corrected grammar does not match the corrected parse tree (with removed _X): %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 223
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 223 in p]
		open("sanity.223", 'w').write("Tree contain suffixes but the grammar is not corrected: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 231
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 231 in p]
		open("sanity.231", 'w').write("corrected parse's surface does not match the corrected sentence: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 232
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 232 in p]
		open("sanity.232", 'w').write("corrected parse's surface does not match the original sentence: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 241
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 241 in p]
		open("sanity.241", 'w').write("The sentence starts with a lowercase letter but has not been corrected: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 242
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 242 in p]
		open("sanity.242", 'w').write("The sentence does not end with a punctuation mark but has not been corrected: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 251
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 251 in p]
		open("sanity.251", 'w').write("There are unary nodes in: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 252
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 252 in p]
		open("sanity.252", 'w').write("There are problematic NP structures in: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
		# problem code 253
		vlSentIdxs = [i for i, p in enumerate(vlProblems, start = 1) if 253 in p]
		open("sanity.253", 'w').write("There are syntactic labels with lowercase letters in: %s\n\n%s" % (','.join([str(idx) for idx in vlSentIdxs]), self.extractSubset(vlSentIdxs).getStr(pflgGenerateNewIds = False)))
		
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# ForeeAnnSentence
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class ForeeAnnSentence:
	'''
	Class for Foreebank annotation sentence
	
	A Foreebank annotation sentence object represents a sentence with all
	its annotation components.
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		self.id = 0                               # sentence ID number
		self.surface = ''                         # surface form of the sentence
		self.orgForm = ''                         # surface form of the sentence as appears in the forum
		self.contextIdx = 0                       # line position in the original corpus to access the context
		self.surfaceStatus = ''                   ## one of the following values to describe the corrections to surface form:
		                                          ##  (ok): unchanged
		                                          ##  (unc)lean containing HTML tags and alike
		                                          ##  (tok)enization or masking problem corrected
		                                          ##  (seg)mentation problem corrected
		                                          ##  (rem)oved to be replaced with a new sentence
		self.correctedForm = ''                   # corrected surface form of the sentence (not grammatically)
		self.comprehensibility = ''               ## Possible values:
		                                          ##  (comp)rehensible
		                                          ##  (cont)ext required
		                                          ##  (inc)omprehensible
		self.grammaticalForm = ''                 # closest grammatical form if is was ungrammatical
		self.automaticParse = ''                  # automatic parse tree to be manually corrected
		self.correctedParse = ''                  # manually corrected parse tree
		self.confidence = ''                      ## confidence of manual parse tree correction:
		                                          ##  sure:                             100%
		                                          ##  fairly sure:                       75%
		                                          ##  half sure:                         50%
		                                          ##  highly uncertain:                  25%
		                                          ##  impossible (needs to be dropped):   0%
		self.comment = ''                         # any comment
		
	
	
	@property
	def size(self):
		'''
		Returns the size of the sentence which is the number of tokens in
		the sentence (surface field)
		'''
		
		# The surface form is assumed to be tokenized.
		return len(self.surface.split(' '))
		
	
	
	def loadData(self, pData, pNewID = None):
		'''
		Loads sentence data
		
		If a new ID is given, it will be used instead of what is in the 
		data parameter.
		'''
		
		if pNewID != None:
			self.setID(pNewID)
		else:
			self.setID(re.search("^No[\t]+(.*)$", pData, re.MULTILINE).group(1))
		self.setSurfaceForm(re.search("^Sentence[\t]+(.*)$", pData, re.MULTILINE).group(1))
		self.setOrgForm(re.search("^Original Form[\t]+(.*)$", pData, re.MULTILINE).group(1))
		self.setContextIdx(int(re.search("^Context Index[\t]+(.*)$", pData, re.MULTILINE).group(1)))
		self.setSurfaceStatus(re.search("^Surface Status[\t]+(.*)$", pData, re.MULTILINE).group(1))
		self.setCorrectedForm(re.search("^Corrected Sentence[\t]+(.*)$", pData, re.MULTILINE).group(1))
		self.setComprehensibility(re.search("^Comprehensibility[\t]+(.*)$", pData, re.MULTILINE).group(1))
		self.setGrammaticalForm(re.search("^Grammatical Form[\t]+(.*)$", pData, re.MULTILINE).group(1))
		self.setAutoParse(re.search("^Automatic Parse[\t]+(.*)$", pData, re.MULTILINE).group(1))
		self.setCorrectedParse(re.search("^Manual Parse[\t]+(.*)$", pData, re.MULTILINE).group(1))
		self.setConfidence(re.search("^Confidence[\t]+(.*)$", pData, re.MULTILINE).group(1))
		self.setComment(re.search("^Comment[\t]+((?:.*\n)*)", pData, re.MULTILINE).group(1).strip())
		
	
	
	def __str__(self):
		'''
		String representation of the sentence
		'''
		
		return self.getStr()
		
	
	
	def getStr(self, pflgWriteID = False):
		'''
		Writes the sentence in the Foreebank annotation block format into 
		a string and returns
		
		If pflgWriteID is set to true, the sentence ID (no) is also written.
		For backward-compatibility, the default is false. 
		'''
		if pflgWriteID:
			vOut = "No\t\t\t%s\n" % self.id
		else:
			vOut = ''
		vOut += "Sentence\t\t%s\n" % self.surface
		vOut += "Original Form\t\t%s\n" % self.orgForm
		vOut += "Context Index\t\t%s\n" % self.contextIdx
		vOut += "Surface Status\t\t%s\n" % self.getSurfaceStatusStr()
		vOut += "Corrected Sentence\t%s\n" % self.correctedForm
		vOut += "Comprehensibility\t%s\n" % self.getComprehensibilityStr()
		vOut += "Grammatical Form\t%s\n" % self.grammaticalForm
		vOut += "Automatic Parse\t\t%s\n" % self.automaticParse
		vOut += "Manual Parse\t\t%s\n" % self.correctedParse
		vOut += "Confidence\t\t%s\n" % self.getConfidenceStr()
		vOut += "Comment\t\t\t%s\n" % self.comment.replace('\n', ' \\n ')
		
		return vOut
		
	
	
	# sets
	
	def setID(self, pValue):
		'''
		Sets the value of sentence ID
		'''
		
		self.id = pValue
		
	
	
	def setSurfaceForm(self, pValue):
		'''
		Sets the value of sentence surface form
		'''
		
		self.surface = pValue
		
	
	
	def setOrgForm(self, pValue):
		'''
		Sets the value of original sentence surface form
		'''
		
		self.orgForm = pValue
		
	
	
	def setContextIdx(self, pValue):
		'''
		Sets the value of context index
		'''
		
		self.contextIdx = pValue
		
	
	
	def setSurfaceStatus(self, pSurfaceStatus):
		'''
		Sets the value of status of the sentence surface form
		'''
		
		if pSurfaceStatus.lower().startswith("ok"):       # unchanged
			self.surfaceStatus = 0
		elif pSurfaceStatus.lower().startswith("unc"):    # unclean containing HTML tags and alike
			self.surfaceStatus =  1
		elif pSurfaceStatus.lower().startswith("tok"):    # tokenization or masking problem corrected
			self.surfaceStatus =  2
		elif pSurfaceStatus.lower().startswith("seg"):    # segmentation problem corrected
			self.surfaceStatus =  3
		elif pSurfaceStatus.lower().startswith("rem"):    # removed to be replaced with a new sentence
			self.surfaceStatus =  4
		elif pSurfaceStatus.strip() != "":
			raise Exception("Unknown surface status: %s" % pSurfaceStatus)
		
	
	
	def setCorrectedForm(self, pValue):
		'''
		Sets the value of corrected sentence surface form
		'''
		
		self.correctedForm = pValue
		
	
	
	def setComprehensibility(self, pComprehensibility):
		'''
		Sets the value of sentence comprehensibility
		'''
		
		if pComprehensibility.lower().startswith("comp"):      #  comprehensible
			self.comprehensibility = 0
		elif pComprehensibility.lower().startswith("cont"):    #  context required
			self.comprehensibility = 1
		elif pComprehensibility.lower().startswith("inc"):     #  incomprehensible
			self.comprehensibility = 2
		elif pComprehensibility.strip() != "":
			raise Exception("Unknown comprehensibility value: %s" % pComprehensibility)
		
	
	
	def setGrammaticalForm(self, pValue):
		'''
		Sets the value of sentence grammatical form
		'''
		
		self.grammaticalForm = pValue
		
	
	
	def setAutoParse(self, pValue):
		'''
		Sets the value of automatic parse
		'''
		
		self.automaticParse = pValue
		
	
	
	def setCorrectedParse(self, pValue):
		'''
		Sets the value of corrected parse
		'''
		
		self.correctedParse = pValue
		
	
	
	def setConfidence(self, pConfidence):
		'''
		Sets the value of parse correction confidence
		'''
		
		if pConfidence.lower().startswith("sure"):       # sure:                             100%
			self.confidence = 1.0
		elif pConfidence.lower().startswith("fair"):     # fairly sure:                       75%
			self.confidence = 0.75
		elif pConfidence.lower().startswith("half"):     # half sure:                         50%
			self.confidence = 0.5
		elif pConfidence.lower().startswith("high"):     # highly uncertain:                  25%
			self.confidence = 0.25
		elif pConfidence.lower().startswith("imp"):      # impossible (needs to be dropped):   0%
			self.confidence = 0.0
		elif pConfidence.strip() != "":
			raise Exception("Unknown confidence value: %s" % pConfidence)
		
	
	
	def setComment(self, pValue):
		'''
		Sets the value of comment field
		'''
		
		self.comment = pValue
		
	
	
	# gets
	
	def getSurfaceForm(self):
		'''
		Returns the surface form of the sentence
		'''
		
		return self.surface
		
	
	
	def getCorrectedForm(self):
		'''
		Returns the value of corrected form
		'''
		
		return self.correctedForm
		
	
	
	def getGrammaticalForm(self):
		'''
		Returns the value of grammatical form
		'''
		
		return self.grammaticalForm
		
	
	
	def getSurfaceStatusStr(self):
		'''
		Returns the string value of status of the sentence surface form
		'''
		
		if self.surfaceStatus == 0:
			return "ok"
		elif self.surfaceStatus == 1:
			return "unclean"
		elif self.surfaceStatus == 2:
			return "tokenisation"
		elif self.surfaceStatus == 3:
			return "segmentation"
		elif self.surfaceStatus == 4:
			return "removed"
		else:
			return ""
		
	
	
	def getComprehensibilityStr(self):
		'''
		Returns the string value of sentence comprehensibility
		'''
		
		if self.comprehensibility == 0:
			return "comprehensible"
		elif self.comprehensibility == 1:
			return "context required"
		elif self.comprehensibility == 2:
			return "incomprehensible"
		else:
			return ""
		
	
	
	def getConfidenceStr(self):
		'''
		Returns the string value of parse correction confidence
		'''
		
		if self.confidence == 1.0:
			return "sure"
		elif self.confidence == 0.75:
			return "fairly sure"
		elif self.confidence == 0.5:
			return "half sure"
		elif self.confidence == 0.25:
			return "highly uncertain"
		elif self.confidence == 0.0:
			return "impossible"
		else:
			return ""
		
	
	
	def getAutoParseTagset(self):
		'''
		Returns the tag set of the automatic parse of the sentence
		'''
		
		vConstTree = ForeeConstTree()
		
		vConstTree.loadPTBTree(self.automaticParse, pflgExpandTerminal = True)
		
		return vConstTree.getTagset()
		
	
	
	def getCorrectedParseTagset(self, pflgRemSuffixs = False):
		'''
		Returns the tag set of the corrected parse of the sentence
		'''
		
		vConstTree = ForeeConstTree()
		
		vConstTree.loadPTBTree(self.getCorrectedParse(pflgRemSuffixs = pflgRemSuffixs), pflgExpandTerminal = True)
		
		return vConstTree.getTagset()
		
	
	
	def getSuffixTags(self):
		'''
		Returns the tags in corrected parses containing suffixes
		'''
		
		vdTagset = {}
		
		for vTag, vCount in self.getCorrectedParseTagset().iteritems():
			if '_' in vTag:
				vdTagset[vTag] = vCount
		
		return vdTagset
		
	
	
	def getSuffixSeqList(self):
		'''
		Returns the sequence of tag suffixes in a list (empty strings for
		unsuffixed tags)
		'''
		
		vPOSTagging = pos.POSTagging()
		vPOSTagging.loadPTBPOSTagging(pPTBTree = self.getCorrectedParse(pflgRemSuffixs = False))
		
		vlSuffixSeq = []
		for vToken in vPOSTagging:
			vlSuffixSeq.append(vToken.getSuffix())
		
		return vlSuffixSeq
		
	
	
	def extractSuffixStat(self):
		'''
		Extracts statistics about suffix tags
		'''
		
		vdSuffixCounts = {}
		vdSuffixedTagCounts = {}
		
		# extracting counts
		
		for vTag, vCount in self.getSuffixTags().iteritems():
			vlSplit = vTag.split('_')
			
			if vlSplit[0] in vdSuffixedTagCounts:
				vdSuffixedTagCounts[vlSplit[0]] += vCount
			else:
				vdSuffixedTagCounts[vlSplit[0]] = vCount
			
			for vSuffix in vlSplit[1: ]:
				if vSuffix in vdSuffixCounts:
					vdSuffixCounts[vSuffix] += vCount
				else:
					vdSuffixCounts[vSuffix] = vCount
		
		# computing ratios
		
		vdSuffixs = {}
		vdSuffixedTags = {}
		
		vTagset = self.getCorrectedParseTagset(pflgRemSuffixs = True)
		
		for vTag, vCount in vdSuffixedTagCounts.iteritems():
			if vTag in vTagset:
				vdSuffixedTags[vTag] = (vCount, vCount * 1.0 / vTagset[vTag])
			else:
				vdSuffixedTags[vTag] = (vCount, 1.0)
		
		for vSuffix, vCount in vdSuffixCounts.iteritems():
			if vSuffix.endswith('M'):
				vdSuffixs[vSuffix] = (vCount, 1)
			else:
				vdSuffixs[vSuffix] = (vCount, vCount * 1.0 / self.size)
		
		return vdSuffixs, vdSuffixedTags
		
	
	
	def getAutoParse(self):
		'''
		Returns the automatic parse in the string format
		'''
		
		return self.automaticParse
		
	
	
	def getAutoConstTree(self):
		'''
		Returns the automatic parse as ConstTree object
		'''
		
		vConstTree = ForeeConstTree()
		vConstTree.loadPTBTree(self.getAutoParse(), pflgExpandTerminal = True)
		
		return vConstTree
		
	
	
	def getCorrectedParse(self, pflgRemSuffixs = False, pflgDelDNodes = False, pflgDelXNodes = False, pflgDelPRONodes = False, pflgMergeBrokenNodes = False):
		'''
		Returns the corrected parses in a list
		
		The tag suffixes can optionally be removed. The nodes inserted 
		manually to account for dropped tokens (_D suffixes), nodes marked
		as extra (_X) and also pro-drop nodes (*PRO* in PTB) can also 
		optionally be deleted.
		'''
		
		if pflgRemSuffixs or pflgDelDNodes or pflgDelXNodes or pflgDelPRONodes:
			vConstTree = ForeeConstTree()
			vConstTree.loadPTBTree(self.correctedParse, pflgExpandTerminal = True)
			
			if pflgDelDNodes:
				vConstTree.delNodes(pLabelPattern = ".*_D.*")
			
			if pflgDelXNodes:
				vConstTree.delNodes(pLabelPattern = ".*_X.*")
			
			if pflgMergeBrokenNodes:
				vConstTree.mergeBrokenTokenNodes()
			
			if pflgRemSuffixs:
				vConstTree.removeSuffix(pGlue = '_')
			
			if pflgDelPRONodes:
				vConstTree.delPRO()
			
			return '( %s )' % vConstTree.getPTBFormat()
		else:
			return self.correctedParse
		
	
	
	def getCorrectedParseSurface(self, pflgRemSuffixs = False, pflgDelDNodes = False, pflgDelXNodes = False, pflgDelPRONodes = False, pflgMergeBrokenNodes = False):
		'''
		Returns the surface form of the corrected parses in a list
		
		The tag suffixes can optionally be removed. The nodes inserted 
		manually to account for dropped tokens (_D suffixes), nodes marked
		as extra (_X) and also pro-drop nodes (*PRO* in PTB) can also 
		optionally be deleted.
		'''
		
		return self.getCorrectedConstTree(pflgRemSuffixs = pflgRemSuffixs, pflgDelDNodes = pflgDelDNodes, pflgDelXNodes = pflgDelXNodes, pflgDelPRONodes = pflgDelPRONodes, pflgMergeBrokenNodes = pflgMergeBrokenNodes).surface
		
	
	
	def getCorrectForm(self):
		'''
		Returns the correct surface form of the sentence
		
		Correct surface form is the corrected sentence if any and the sentence
		itself otherwise.
		'''
		
		vCorrectedForm = self.getCorrectedForm()
		
		if vCorrectedForm != '':
			return vCorrectedForm
		else:
			return self.getSurfaceForm()
		
	
	
	def getCorrectGrammar(self):
		'''
		Returns the grammatical form of the sentence
		
		If the grammar is not corrected, the sentence itself is returned.
		'''
		
		vGrammForm = self.getGrammaticalForm()
		
		if vGrammForm != '':
			return vGrammForm
		else:
			return self.getSurfaceForm()
		
	
	
	def getCorrectedConstTree(self, pflgRemSuffixs = False, pflgDelDNodes = False, pflgDelXNodes = False, pflgDelPRONodes = False, pflgMergeBrokenNodes = False):
		'''
		Returns the corrected parse as ConstTree object
		'''
		
		vConstTree = ForeeConstTree()
		vConstTree.loadPTBTree(self.getCorrectedParse(pflgRemSuffixs = pflgRemSuffixs, pflgDelDNodes = pflgDelDNodes, pflgDelXNodes = pflgDelXNodes, pflgDelPRONodes = pflgDelPRONodes, pflgMergeBrokenNodes = pflgMergeBrokenNodes), 
		                       pflgExpandTerminal = True)
		
		return vConstTree
		
	
	
	def getSurfaceLen(self):
		'''
		Returns the word length of the surface form
		'''
		
		return len(self.surface.split(' '))
		
	
	
	def getGrammaticalFormLen(self):
		'''
		Returns the word length of the grammatical form
		'''
		
		return len(self.grammaticalForm.split(' '))
		
	
	
	# other
	
	def isGramCorrected(self):
		'''
		Returns ture if the grammtical form of the sentence is corrected
		(not empty)
		'''
		
		return self.getGrammaticalForm()
		
	
	
	def applySurfaceCorrection(self, pflgChangeStatus = False):
		'''
		Replaces the surface form with corrected form and deletes the corrected
		form
		
		It applies the correction if the corrected surface field is not 
		empty.
		
		If pflgChangeStatus is set to true, the value of the surface status
		will be changed to OK.
		'''
		
		self.setSurfaceForm(self.getCorrectedForm())
		self.setCorrectedForm(pValue = '')
		
		if pflgChangeStatus:
			self.setSurfaceStatus("ok")
		
	
	
	def checkSanity(self):
		'''
		Performs some sanity tests on the annotation
		'''
		
		vlProblems = []
		
		vAutoConstTree = self.getAutoConstTree()
		vCorrectedConstParse = self.getCorrectedParse()
		vCorrectedConstTree = self.getCorrectedConstTree()
		vCorrectedConstTreeNoAddedNode = self.getCorrectedConstTree(pflgDelDNodes = True, pflgDelPRONodes = True)
		
		# surface
		
		if self.correctedForm.strip() != '':
			if self.getSurfaceStatusStr().lower() == "ok":
				vlProblems.append(111)
			
			if self.correctedForm.strip() == self.surface:
				vlProblems.append(112)
			
			if self.correctedForm.strip() != vAutoConstTree.surface:
				vlProblems.append(113)
		else:
			if self.getSurfaceStatusStr().lower() == "tokenisation":
				if not re.search('_T[^ ]* ', vCorrectedConstParse):
					vlProblems.append(121)
		
		# grammar and parse
		
		if self.grammaticalForm.strip() != '':
			if self.grammaticalForm.strip() == self.surface or self.grammaticalForm.strip() == self.correctedForm.strip():
				vlProblems.append(211)
			
			if not vCorrectedConstTree.hasSuffixedLabel():
				vlProblems.append(212)
			
			if self.getGrammaticalFormLen() != self.getCorrectedConstTree(pflgDelDNodes = False, pflgDelXNodes = True, pflgDelPRONodes = True, pflgMergeBrokenNodes = True).getSentLength():
				vlProblems.append(213)
		else:
			# _M and _I and _E are do not need grammar correction
			if len([s for s in vCorrectedConstTree.extractSuffixes().iterkeys() if s not in ['M', 'I', 'E']]) != 0:
				vlProblems.append(223)
		
		if self.correctedForm.strip() != '':
			if vCorrectedConstTreeNoAddedNode.surface != self.correctedForm.strip():
				#print "\t%s" % vCorrectedConstTree.surface
				#print "\t%s" % self.correctedForm.strip()
				vlProblems.append(231)
		else:
			if self.grammaticalForm.strip() == '':
				if vCorrectedConstTreeNoAddedNode.surface != self.surface:
					#print "\t%s" % vCorrectedConstTree.surface
					#print "\t%s" % self.surface
					vlProblems.append(232)
		
		if re.match("^[a-z]", self.surface):
			if self.grammaticalForm.strip() == '' or re.match("^[a-z]", self.grammaticalForm):
				vlProblems.append(241)
		
		if vCorrectedConstTree.getRootLabel() in ["S", "SQ", "SBARQ"]:
			vlChildTagSeq = vCorrectedConstTree.getPreTerminals()
			if (vlChildTagSeq[-1][0] not in ['.', ':']) and \
			   (vlChildTagSeq[-1][:3] != "NFP"):
				vlProblems.append(242)
		
		if len(vCorrectedConstTree.extractUnaryNodes()) > 0:
			vlProblems.append(251)
		
		if len(vCorrectedConstTree.extractBadNPStucts()) > 0:
			vlProblems.append(252)
		
		if len([c for c in ''.join(vCorrectedConstTree.getSynLabels()) if c.islower()]) > 0:
			vlProblems.append(253)
		
		return vlProblems
		
	
	
	def ed(self):
		'''
		Calculates the edit distance between the sentence and its grammatical 
		correction based on the Foreebank suffixes on the parse tree
		
		The edit distance is calculated by summing the inserted (or extra; _X),
		deleted (_D) and substituted (_S, _C, _W) and dividing the outcome
		by the maximum of the lengths of the sentence from and grammatical 
		correction.
		'''
		
		# extracting number of each suffix in the sentence
		vdSuffixStat, dummy = self.extractSuffixStat()
		
		vInsertions = 0
		vDeletion = 0
		vSubstitution = 0
		
		# insertions are _X nodes
		if "X" in vdSuffixStat:
			vInsertions = vdSuffixStat["X"][0]
		
		if "D" in vdSuffixStat:
			vDeletion = vdSuffixStat["D"][0]
		
		if "S" in vdSuffixStat:
			vSubstitution = vdSuffixStat["S"][0]
		if "W" in vdSuffixStat:
			vSubstitution += vdSuffixStat["W"][0]
		if "C" in vdSuffixStat:
			vSubstitution += vdSuffixStat["C"][0]
		
		vED = 1.0 * (vInsertions + vDeletion + vSubstitution) / max(self.getSurfaceLen(), self.getGrammaticalFormLen())
		
		return vInsertions, vDeletion, vSubstitution, vED
		
	
	
	def replaceSurfaceTokens(self, pOldToken, pNewToken):
		'''
		Replaces all occurrences of a given token in the surface form if
		found and returns the number of replacements
		
		It assumes that the form is already tokenized.
		'''
		
		vlTokens = self.getSurfaceForm().split(' ')
		
		vCnt = 0
		for i in range(len(vlTokens)):
			if vlTokens[i - 1] == pOldToken:
				vlTokens[i - 1] = pNewToken
				vCnt += 1
		
		self.setSurfaceForm(' '.join(vlTokens))
		
		return vCnt
		
	
	
	def replaceGramFormTokens(self, pOldToken, pNewToken):
		'''
		Replaces all occurrences of a given token in the grammatical form 
		if found and returns the number of replacements
		
		It assumes that the form is already tokenized.
		'''
		
		vlTokens = self.getGrammaticalForm().split(' ')
		
		vCnt = 0
		for i in range(len(vlTokens)):
			if vlTokens[i - 1] == pOldToken:
				vlTokens[i - 1] = pNewToken
				vCnt += 1
		
		self.setGrammaticalForm(' '.join(vlTokens))
		
		return vCnt
		
	
	
	def replaceCrrParseTerminalTokens(self, pOldToken, pNewToken):
		'''
		Replaces all occurrences of a given token in the corrected parse
		tree terminals
		'''
		
		vConstTree = ForeeConstTree()
		vConstTree.loadPTBTree(self.getCorrectedParse(pflgRemSuffixs = False, pflgDelDNodes = False, pflgDelXNodes = False, pflgDelPRONodes = False), pflgExpandTerminal = True)
		vlTerminals = vConstTree.getTerminalNodes()
		
		vCnt = 0
		for i in range(len(vlTerminals)):
			if vlTerminals[i - 1].surface == pOldToken:
				vlTerminals[i - 1].setTerminalForm(pNewToken)
				vCnt += 1
		
		self.setCorrectedParse("( %s )" % vConstTree.getPTBFormat())
		
		return vCnt
		
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# ForeeAnnIAA
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class ForeeAnnIAA:
	'''
	Class for Foreebank annotation inter-annotator agreements
	'''
	
	
	def __init__(self, p1stAnnFilename, p2ndAnnFilename):
		'''
		Constructor
		'''
		
		self.fstAnn = ForeeAnn()
		self.fstAnn.loadForeeAnn(pAnnFilename = p1stAnnFilename)
		
		self.scdAnn = ForeeAnn()
		self.scdAnn.loadForeeAnn(pAnnFilename = p2ndAnnFilename)
		
		
		# list of positions of sentences with various field disagreement
		
		self.surfaceStatusIADs = []
		self.correctedFormIADs = []
		self.comprehensIADs = []
		self.grammaticalFormIADs = []
		self.correctedParseIADs = []
		self.confidenceIADs = []
		
	
	
	def iad(self):
		'''
		Extracts inter-annotator disagreements on various annotation fields
		'''
		
		self.surfaceStatusIADs = []
		self.correctedFormIADs = []
		self.comprehensIADs = []
		self.grammaticalFormIADs = []
		self.correctedParseIADs = []
		self.confidenceIADs = []
		
		# extracting disagreement sentences
		
		for vSent1, vSent2 in zip(self.fstAnn.getSentences(), self.scdAnn.getSentences()):
			# sanity check
			if vSent1.id != vSent2.id:
				raise Exception("Sentences numbers do not match: %s vs. %s" % (vSent1.id, vSent2.id))
			
			if vSent1.surfaceStatus != vSent2.surfaceStatus:
				self.surfaceStatusIADs.append(int(vSent1.id))
			
			if vSent1.correctedForm != vSent2.correctedForm:
				self.correctedFormIADs.append(int(vSent1.id))
			
			if vSent1.comprehensibility != vSent2.comprehensibility:
				self.comprehensIADs.append(int(vSent1.id))
			
			if vSent1.grammaticalForm != vSent2.grammaticalForm:
				self.grammaticalFormIADs.append(int(vSent1.id))
			
			if vSent1.correctedParse != vSent2.correctedParse:
				self.correctedParseIADs.append(int(vSent1.id))
			
			if vSent1.confidence != vSent2.confidence:
				self.confidenceIADs.append(int(vSent1.id))
		
		# calculating agreement percentages
		
		print "%s%% agreement on surface status"    % round((self.fstAnn.size - len(self.surfaceStatusIADs))   * 100.0 / self.fstAnn.size, 2)
		print "%s%% agreement on corrected form"    % round((self.fstAnn.size - len(self.correctedFormIADs))   * 100.0 / self.fstAnn.size, 2)
		print "%s%% agreement on comprehensibility" % round((self.fstAnn.size - len(self.comprehensIADs))      * 100.0 / self.fstAnn.size, 2)
		print "%s%% agreement on grammatical form"  % round((self.fstAnn.size - len(self.grammaticalFormIADs)) * 100.0 / self.fstAnn.size, 2)
		print "%s%% agreement on corrected parse"   % round((self.fstAnn.size - len(self.correctedParseIADs))  * 100.0 / self.fstAnn.size, 2)
		print "%s%% agreement on confidence"        % round((self.fstAnn.size - len(self.confidenceIADs))      * 100.0 / self.fstAnn.size, 2)
		
	
	
	def getSurfaceStatusIADs(self):
		'''
		Return surface status disagreement list (of sentence IDs)
		'''
		
		return self.surfaceStatusIADs
		
	
	
	def getCorrectedFormIADs(self):
		'''
		Return corrected form disagreement list (of sentence IDs)
		'''
		
		return self.correctedFormIADs
		
	
	
	def getComprehensIADs(self):
		'''
		Return comprehensibility disagreement list (of sentence IDs)
		'''
		
		return self.comprehensIADs
		
	
	
	def getGrammaticalFormIADs(self):
		'''
		Return grammatical form disagreement list (of sentence IDs)
		'''
		
		return self.grammaticalFormIADs
		
	
	
	def getCorrectedParseIADs(self):
		'''
		Return corrected parse disagreement list (of sentence IDs)
		'''
		
		return self.correctedParseIADs
		
	
	
	def getConfidenceIADs(self):
		'''
		Return confidence disagreement list (of sentence IDs)
		'''
		
		return self.confidenceIADs
		
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# ForeePOSTagging
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class ForeePOSTagging(pos.POSTagging):
	'''
	Class for Foreebank-style POS tagging of a sentence
	
	The main characteristic of Foreebank POS tagging is the suffixes attached
	to the POS taggs (e.g. NNP_C)
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		pos.POSTagging.__init__(self)
		
	
	
	def _createToken(self, pForm = None, pPOS = None):
		'''
		Creates and returns a POS tagged token
		'''
		
		return ForeePOSTaggedToken(pForm = pForm, pPOS = pPOS)
		
	
	
	def removeSuffix(self, pGlue = '_'):
		'''
		Removes suffixes from POS tags.
		
		The glue parameter specifies the character used to attach the suffix.
		'''
		
		for vToken in self:
			vToken.removeSuffix(pGlue = pGlue)
		
	
	
	def getSuffixSeqList(self):
		'''
		Returns the sequence of tag suffixes in a list (empty strings for
		unsuffixed tags)
		'''
		
		vlSuffixSeq = []
		for vToken in self:
			vlSuffixSeq.append(vToken.getSuffix())
		
		return vlSuffixSeq
		
	
	
	def getSuffixSeqStr(self):
		'''
		Returns the sequence of tag suffixes in a string 
		
		In the output format, each token suffix is wrapped inside '' and
		separated by a space from neighbouring tokens. Tokens without 
		suffix are represented by '', e.g. '' '' '_W' '' '' '_X' (sentence
		with 6 tokens where only 3rd and 6th tokens have suffixes.
		'''
		
		vlSuffixStrs = []
		for vToken in self:
			vlSuffixStrs.append("'%s'" % vToken.getSuffix())
			
		
		return ' '.join(vlSuffixStrs)
		
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# ForeePOSTaggedToken
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class ForeePOSTaggedToken(pos.POSTaggedToken):
	'''
	Class for Foreebank-style POS tagged token
	
	The main characteristic of Foreebank POS tagging is the suffixes attached
	to the POS taggs (e.g. NNP_C)
	'''
	
	def __init__(self, pForm = None, pPOS = None):
		'''
		Constructor
		'''
		
		pos.POSTaggedToken.__init__(self, pForm = pForm, pPOS = pPOS)
		
	
	
	def removeSuffix(self, pGlue = '_'):
		'''
		Removes suffixes from the POS tag.
		
		The glue parameter specifies the character used to attach the suffix.
		'''
		
		vGluePos = self.tag.find(pGlue)
		
		if vGluePos != -1:
			self.tag = self.tag[ : vGluePos]
		
	
	
	def getSuffix(self, pGlue = '_'):
		'''
		Returns the POS tag suffix
		
		The glue parameter specifies the character used to attach the suffix.
		'''
		
		vGluePos = self.tag.find(pGlue)
		
		if vGluePos != -1:
			return self.tag[vGluePos : ]
		else:
			return ''
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# ForeeConstTree
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class ForeeConstTree(constparse.ConstTree):
	'''
	Class for Foreebank-style constituency tree
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		constparse.ConstTree.__init__(self)
		
	
	
	def _createRoot(self):
		'''
		Creates and returns the root node
		'''
		
		return ForeeConstNode()
		
	
	
	def extractSuffixes(self, pGlue = '_'):
		'''
		Extracts and returns suffixes attached to the syntactic labels in
		a dictionary with suffixes as it keys and their count as its values
		'''
		
		return self.root.extractSuffixes(pGlue = pGlue)
		
	
	
	def removeSuffix(self, pGlue = '_'):
		'''
		Removes suffixes from phrase labels and POS tags.
		
		The glue parameter specifies the character used to attach the suffix.
		'''
		
		self.root.removeSuffix(pGlue = pGlue, pflgAllSubtree = True)
		
	
	
	def hasSuffixedLabel(self):
		'''
		Returns true if any node in tree is labeled with Foreebank suffix
		'''
		
		return self.root.hasSuffixedLabel()
		
	
	
	def mergeBrokenTokenNodes(self):
		'''
		Merges the broken token nodes (_B) in the tree into one if their
		labels are the same
		
		Otherwise, an exception will be raised. According to the guidelines
		the broken token parts should all be assigned the same POS tag.
		'''
		
		self.root.mergeBrokenTokenNodes()
		
	
	
	def extractUnaryNodes(self):
		'''
		Extracts and returns all unary nodes in the tree, if any
		
		An example unary node is (NP (NP (NN example))).
		'''
		
		return self.root.extractUnaryNodes()
		
	
	
	def extractBadNPStucts(self):
		'''
		Extracts and returns problematic NP structures in the tree
		
		The problematic structures covered are:
		- (NP (DT X) (NN X) (PP X)) instead of (NP (NP (DT X) (NN X)) (PP X))
		- (NP (DT X) (NN X) (SBAR X)) instead of (NP (NP (DT X) (NN X)) (SBAR X))
		'''
		
		return self.root.extractBadNPStucts()
		
	
	



#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# ForeeConstNode
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class ForeeConstNode(constparse.ConstNode):
	'''
	Class for Foreebank-style constituency tree node
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		constparse.ConstNode.__init__(self)
		
	
	
	def _getNewNode(self):
		'''
		Creates and returns a node
		'''
		
		return ForeeConstNode()
		
	
	
	def getSuffixes(self, pGlue = '_'):
		'''
		Returns the suffixes attached to the node label in a list
		'''
		
		vGluePos = self.getLabel().find(pGlue)
		
		if vGluePos != -1:
			return self.getLabel()[vGluePos + 1 : ].split(pGlue)
		else:
			return []
		
	
	
	def extractSuffixes(self, pGlue = '_'):
		'''
		Extracts and returns suffixes attached to the syntactic labels in
		the whole subtree in a dictionary with suffixes as it keys and 
		their count as its values
		'''
		
		vlSuffixes = self.getSuffixes(pGlue = pGlue)
		
		for vChild in self.getChildren():
			vlSuffixes += vChild.getSuffixes(pGlue = pGlue)
		
		return util.groupBy(vlSuffixes)
		
	
	
	def removeSuffix(self, pGlue = '_', pflgAllSubtree = False):
		'''
		Removes suffixes from the node label.
		
		The glue parameter specifies the character used to attach the suffix.
		The removal can be done to only the current node or the entire subtree
		under the node.
		'''
		
		vGluePos = self.getLabel().find(pGlue)
		
		if vGluePos != -1:
			self.synTag = self.getLabel()[ : vGluePos]
		
		if pflgAllSubtree:
			for vChild in self.getChildren():
				vChild.removeSuffix(pGlue = pGlue, pflgAllSubtree = pflgAllSubtree)
		
	
	
	def hasSuffixedLabel(self):
		'''
		Returns true if any node in tree is labeled with Foreebank suffix
		'''
		
		if '_' in self.getLabel():
			return True
		else:
			for vChild in self.children:
				if vChild.hasSuffixedLabel():
					return True
			
			return False
		
	
	
	def mergeBrokenTokenNodes(self):
		'''
		Merges the broken token nodes (_B) in the subtree (or in the adjacent
		nodes if this is a _B token node too) into one if they are all 
		preterminal nodes.
		
		If their labels are not the same, the first label will be used and
		a warning message will be shown.
		'''
		
		if "B" in self.getSuffixes():
			vlBrokenNodes = [self]
			
			# gathering broken token nodes in the left
			for vNode in reversed(self.getLeftSibling()):
				if "B" in vNode.getSuffixes():
					vlBrokenNodes.append(vNode)
				else:
					break
			
			vlBrokenNodes.reverse()
			
			# gathering broken token nodes in the right
			for vNode in self.getRightSibling():
				if "B" in vNode.getSuffixes():
					vlBrokenNodes.append(vNode)
				else:
					break
			
			# checking if the broken node labels are the same
			for i in range(1, len(vlBrokenNodes)):
				if vlBrokenNodes[i].getLabel() != vlBrokenNodes[0].getLabel():
					print "Warning: broken token nodes with different labels: %s and %s; using the first label in the merging" % (vlBrokenNodes[0].getLabel(), vlBrokenNodes[1].getLabel())
			
			# checking if the gathered nodes are pre-terminals
			for vNode in vlBrokenNodes:
				if not vNode.isPreTerminal():
					raise Exception("Cannot merge broken token nodes if they are not pre-terminals!")
			
			# now merging the broken token nodes
			
			# 1. updating the first node to reflect the merged node
			vlBrokenNodes[0].getChildren()[0].setTerminalForm(''.join([n.surface for n in vlBrokenNodes]))
			
			# 2. deleting the rest of the broken token parts
			for vNode in vlBrokenNodes[1:]:
				vNode.getParent().delChild(vNode)
			
		else:
			for vChild in self.getChildren():
				# the child may have already been deleted in the merging process
				if vChild.getParent() != None:
					vChild.mergeBrokenTokenNodes()
		
	
	
	def extractUnaryNodes(self):
		'''
		Returns all unary nodes in the subtree, if any
		
		An example unary node is (NP (NP (NN example))).
		'''
		
		vlUnaryNodes = []
		
		vlChildren = self.getChildren()
		
		if len(vlChildren) == 1:
			if self.getLabel() == vlChildren[0].getLabel():
				vlUnaryNodes.append(self)
		
		for vChild in vlChildren:
			vlUnaryNodes += vChild.extractUnaryNodes()
		
		return vlUnaryNodes
		
	
	
	def extractBadNPStucts(self):
		'''
		Extracts and returns problematic NP structures in the subtree
		
		The problematic structures covered are:
		- (NP (DT X) (NN X) (PP X)) instead of (NP (NP (DT X) (NN X)) (PP X))
		- (NP (DT X) (NN X) (SBAR X)) instead of (NP (NP (DT X) (NN X)) (SBAR X))
		'''
		
		vlBadNPStructs = []
		
		if self.getSynTag() == "NP":
			vChildLabelSeq = ' '.join(self.getChildrenTags())
			
			if re.search("^(DT|PRP\$) NN[A-Z]* (PP|SBAR)$", vChildLabelSeq):
				vlBadNPStructs.append(self)
		
		for vChild in self.getChildren():
			vChild.extractBadNPStucts()
		
		return vlBadNPStructs
		
	
	

