#! /usr/bin/python
# -*- coding: utf-8 -*-

"""	
	SPLIT TO MODULES
	
	This module defines classes to load data of different formats.
	
	Version 1.0								(02-Jun-2015 to 04-Jun-2015)
	- PARC700 classes are moved to parse/parc700.py.
"""

import re
from conll import *


class PARC700DataLoader:
	'''
	Class for loading data in PARC 700 format.
	
	This class is developed mainly because the output of Const2Dep.py script
	used to convert FTB constituency to dependency is in this format (the
	modified version of the format). It is also very similar to the default
	format of Stanford typed dependency output.
	'''
	
	def __init__(self, pDataFilename, pflgVerbos = True):
		'''
		Creates object and loads data from pDataFilename
		'''
		
		# final data storage: a list of PARC700Sentence
		self.parcSentences = []
		
		try:
			vData = open(pDataFilename).read()
		except IOError:
			sys.exit("Cannot open data file: %s" % pDataFilename)
		
		vlSentLines = []
		for vLine in vData.split("\n"):
			if vLine.strip() == "":
				if len(vlSentLines) > 0:       # sentence end signal
					self.parcSentences.append(PARC700Sentence(vlSentLines))
					vlSentLines = []
					
					if pflgVerbos:
						sys.stdout.write('.')
						sys.stdout.flush()
			else:
				vlSentLines.append(vLine.strip())
		
		# just in case the file does not end with a blank line
		if len(vlSentLines) > 0:
			self.parcSentences.append(PARC700Sentence(vlSentLines))
			vlSentLines = []
			
			if pflgVerbos:
				sys.stdout.write('.')
				sys.stdout.flush()
		
		if pflgVerbos:
			print "\n%s sentences loaded!" % len(self.parcSentences)
		
	
	
	def getText(self):
		'''
		Returns the surface forms of the loaded sentences
		'''
		
		return '\n'.join([sent.surface for sent in self.parcSentences]) + '\n'
		
	


class PARC700Sentence:
	'''
	Class for sentence in PARC 700 format
	
	A sentence is a list of tokens but there is not a 1-to-1 correspondence
	between this token and sentence surface token which is a word or 
	punctuation mark. The reason is that Const2Dep.py adds a token named
	MISSINGHEAD at the beginning of the sentence to head orphan tokens. 
	Also, some tokens are duplicated to reflect their multiple heads.
	
	NOTE: There are two odd situations observed so far as of 31-Oct-2013:
	 1- There are two MISSINGHEAD tokens at the beginning of the sentence;
	    we keep the dummy root one (head) if any, and the first one otherwise.
	 2- The first line is MISSINGHEAD (MISSINGHEAD~MISSINGHEAD~-1 , MISSINGHEAD~MISSINGHEAD~-1);
	    we remove this line. When this happens, the last token modifies 
	    itself, e.g. PONCT (PONCT~?~12 , PONCT~?~12). This is handled by 
	    changing the head to MISSINGHEAD in PARC700Token creation. 
	
	See PARC700DataLoader for more information on PARC 700.
	'''
	
	def __init__(self, plLines, pLanguage = "fr"):
		'''
		Creates a sentence object from pLines
		'''
		
		self.language = pLanguage
		
		## A sentence is a list of tokens of PARC700Token type
		self.parcTokens = []
		
		self._depTree = None
		
		vMHSeen = -1
		vMHWasRoot = False
		for vLine in plLines:
			## Removing the odd MISSINGHEAD (MISSINGHEAD~MISSINGHEAD~-1 , MISSINGHEAD~MISSINGHEAD~-1)
			## This only happens occasionally and make problem when the 
			## input to Const2Dep script has TOP nodes. So, it's deactivated
			## here, because TOP should not be present in the input due to 
			## some other problems. 
			#if re.match(r"^\d+\tMISSINGHEAD\t\(\tMISSINGHEAD\~MISSINGHEAD\~-1\t,\tMISSINGHEAD\~MISSINGHEAD\~-1\t\)", vLine):
			#	continue
			
			## handling double-MISSINGHEAD situation by keeping only the
			## dummy root one (head) if any, and the first one otherwise
			
			vlFound = re.findall(r"^\d+\t([^\t]+)\t\(\t[^\t]+\t,\tMISSINGHEAD\~MISSINGHEAD\~-1\t\)", vLine)
			if len(vlFound) != 0:
				if vMHSeen > -1:
					if vMHWasRoot or vlFound[0] != "head":
						continue
					else:
						self.parcTokens.pop(vMHSeen)
						vMHWasRoot = True
					
				else:
					vMHSeen = len(self.parcTokens)
					if vlFound[0] == "head":
						vMHWasRoot = True
			
			# finally generating the token
			self.parcTokens.append(PARC700Token(vLine))
		
	
	
	@property
	def surfaceTokens(self):
		'''
		Surface tokens of the sentence (see class documentation above)
		'''
		
		## filtering MISSINGHEAD tokens to get standard tokens (see class
		## documentation above).
		vlSurfaceTokens = [(token.position, token.form) for token in self.parcTokens if token.form != "MISSINGHEAD"]
		
		## discounting duplicate PARC700 tokens with the help of unique 
		## sentence position in vlSurfaceTokens tuple
		vsDistinctTokens = sorted(set(vlSurfaceTokens), key = lambda token: token[0])
		
		return [token[1] for token in vsDistinctTokens]
		
	
	
	@property
	def length(self):
		'''
		Length of the sentence in standard sense (number of words and punctuations)
		
		Sentence length is not the number of tokens (see class documentation
		above).
		'''
		
		return len(self.surfaceTokens)
		
	
	
	@property
	def surface(self):
		'''
		Surface form of the sentence
		'''
		return ' '.join(self.surfaceTokens)
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		return self.surface
		
	
	
	@property
	def depTree(self):
		'''
		Returns the dependency tree of the sentence as a DepTree object
		in the pTreeFormat representation format.
		'''
		
		if self._depTree == None:
			self._depTree = parse.DepTree()
			self._depTree.loadPARC700Dependency(self)
		
		return self._depTree
		
	
	
	def hasMISSINGHEAD(self):
		'''
		Returns true if the sentence contains a MISSINGHEAD token.
		
		(See class documentation above for ISSINGHEAD.)
		'''
		
		if len([token for token in self.parcTokens if token.form == "MISSINGHEAD"]):
			return True
		else:
			return False
		
	
	
	def toMaltInput(self):
		'''
		Converts to Maltparser input format and returns the sentence lines.
		
		See http://www.maltparser.org/mco/french_parser/fremalt.html for
		the format.
		'''
		
		return '\n'.join([str(i) + '\t' + token.toMaltInput() for i, token in enumerate(self.parcTokens, start = 1)])
		
	
	
	def toCoNLL2009(self, pflgRemMISSINGHEAD = False):
		'''
		Converts to CoNLL 2009 format.
		
		It takes care of duplicate tokens (used to represent mutiple head)
		by removing the second one.
		
		It can remove the MISSINGHEAD token if asked. All tokens headed
		by MISSINGHEAD will be headed by the root which is headed by the
		dummy root.
		'''
		
		v2009Sentence = CoNLLSentence([], pFormatVersion = "2009")
		vRootPos = 0                             # to store the root (head) position
		
		vcntrToken = 0
		for vToken in self.parcTokens:
			if pflgRemMISSINGHEAD and vToken.form == "MISSINGHEAD":
				continue
			
			if vcntrToken == vToken.position:    # sign of duplicate token
				continue
			
			vcntrToken += 1
			
			if vToken.head == 0:
				vRootPos = vcntrToken            # will be used later 
			
			v2009Token = vToken.toCoNLL2009()
			v2009Sentence.conllTokens.append(v2009Token)
		
		# fixing the tokens headed by MISSINGHEAD
		if pflgRemMISSINGHEAD:
			for vToken in v2009Sentence.conllTokens:
				if vToken.head == -1:
					vToken.head = vRootPos
					vToken.pHead = vRootPos
		
		return v2009Sentence
		
	
	
	def toCoNLL2006(self, pflgRemMISSINGHEAD = False):
		'''
		Converts to CoNLL 2006 format.
		
		It takes care of duplicate tokens (used to represent mutiple head)
		by removing the second one.
		
		It can remove the MISSINGHEAD token if asked. All tokens headed
		by MISSINGHEAD will be headed by the root which is headed by the
		dummy root.
		'''
		
		v2006Sentence = CoNLL2006Sentence([], pLanguage = self.language)
		vRootPos = 0                             # to store the root (head) position
		
		vcntrToken = 0
		for vToken in self.parcTokens:
			if pflgRemMISSINGHEAD and vToken.form == "MISSINGHEAD":
				continue
			
			if vcntrToken == vToken.position:    # sign of duplicate token
				continue
			
			vcntrToken += 1
			
			if vToken.head == 0:
				vRootPos = vcntrToken            # will be used later 
			
			v2006Token = vToken.toCoNLL2006()
			v2006Sentence.conllTokens.append(v2006Token)
		
		# fixing the tokens headed by MISSINGHEAD
		if pflgRemMISSINGHEAD:
			for vToken in v2006Sentence.conllTokens:
				if vToken.head == -1:
					vToken.head = vRootPos
		
		return v2006Sentence
		
	


class PARC700Token:
	'''
	Class for token in PARC 700 sentence
	
	A token corresponds to a line in PARC700Sentence with attributes like
	form, head, dependency relation, etc. 
	
	Note that there is NOT a 1-to-1 correspondence between this token and
	standard sentence token which is a word or punctuation mark. See 
	PARC700Sentence for explanation.
	
	When there is a MISSINGHEAD token at the beginning of the sentence,
	both root head (dummy root represented by MISSINGPOS~null~-1) and 
	MISSINGHEAD are located at position -1 (MISSINGHEAD~MISSINGHEAD~-1). 
	This conflict is fixed here by keeping MISSINGHEAD at position -1 and
	moving root head to 0 (the latter in usual way by increasing positions 
	by 1 to starts the sentence at 1).
	
	There is an odd situation where the last token modifies itself, e.g. 
	PONCT (PONCT~?~12 , PONCT~?~12) (see PARC700Sentence). This is handled
	by changing the head to MISSINGHEAD.
	
	See PARC700DataLoader for more information on PARC 700.
	'''
	
	def __init__(self, pPARC700Line):
		'''
		Creates a token object
		'''
		
		vlTokenSplit = pPARC700Line.split()
		
		self.depRel = vlTokenSplit[1]
		
		## Cannot simply split on ~ because of confusion possibility with
		## content ~.
		# refer to the class documentation above
		v1stTildePos = vlTokenSplit[3].find('~')
		v2ndTildePos = vlTokenSplit[3].rfind('~')
		if vlTokenSplit[3][v1stTildePos + 1 : v2ndTildePos] == "MISSINGHEAD":
			self.head = int(vlTokenSplit[3][v2ndTildePos + 1 : ])
		else:
			self.head = int(vlTokenSplit[3][v2ndTildePos + 1 : ]) + 1
		
		v1stTildePos = vlTokenSplit[5].find('~')
		v2ndTildePos = vlTokenSplit[5].rfind('~')
		self.posTag = vlTokenSplit[5][ : v1stTildePos]
		self.form = vlTokenSplit[5][v1stTildePos + 1 : v2ndTildePos]
		# refer to the class documentation above
		if self.form == "MISSINGHEAD":
			self.position = int(vlTokenSplit[5][v2ndTildePos + 1 : ])
		else:
			self.position = int(vlTokenSplit[5][v2ndTildePos + 1 : ]) + 1     # position in the sentence
		
		## handling self-modifying situation by changing the head to MISSINGHEAD
		## Note that this assumes that this situation happens only when 
		## there is a MISSINGHEAD in the parse.
		if self.head == self.position:
			self.head = -1
		
		# Const2Dep masks "," with "COMMA", which is unmasked here
		self.form = self.form.replace("COMMA", ',')
		
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		return self.form
		
	
	
	def toMaltInput(self):
		'''
		Converts to Maltparser input format and returns the token line
		
		See http://www.maltparser.org/mco/french_parser/fremalt.html for
		the format.
		'''
		
		## The ID column does not exist in PARC700 and should be generated
		## by PARC700Sentence.
		
		return '\t'.join([self.form, '_', self.posTag, self.posTag, '_'])
		
	
	
	def toCoNLL2009(self):
		'''
		Converts to CoNLL 2009 format
		'''
		
		v2009Token = CoNLL2009Token()
		
		v2009Token.id = self.position
		v2009Token.form = self.form
		v2009Token.lemma = '_'
		v2009Token.pLemma = '_'
		v2009Token.POS = self.posTag
		v2009Token.pPOS = self.posTag
		v2009Token.features = '_'
		v2009Token.pFeatures = '_'
		v2009Token.head = self.head
		v2009Token.pHead = self.head
		v2009Token.depRel = self.depRel
		v2009Token.pDepRel = self.depRel
		v2009Token.fillPred = False
		v2009Token.pred = '_'
		v2009Token.args = []
		
		v2009Token.fieldNum = 14
		
		return v2009Token
		
	
	
	def toCoNLL2006(self):
		'''
		Converts to CoNLL 2006 format
		'''
		
		v2006Token = CoNLL2006Token()
		
		v2006Token.id = self.position
		v2006Token.form = self.form
		v2006Token.lemma = '_'
		v2006Token.coarsePOS = self.posTag
		v2006Token.finePOS = self.posTag
		v2006Token.features = '_'
		v2006Token.head = self.head
		v2006Token.depRel = self.depRel
		v2006Token.projHead = '_'
		v2006Token.projDepRel = '_'
		
		return v2006Token
		
	


