#! /usr/bin/python
# -*- coding: utf-8 -*-

"""	
	This module defines functions and data structures related to pos tagging.
	
	Version 1.1								(12-May-2015 to 14-May-2015)
	- length is added to POSTagging.
	- removeSuffix(), getSuffixSeqList() and getSuffix() from POSTagging
	  and POSTaggedToken are moved to ForeePOSTagging and ForeePOSTaggedToken 
	  in classes foreebank.
	- Creating a POSTaggedToken in POSTagging methods is now done in a new 
	  method called _createToken() to abstract it away.
	- Creating a POSTagging in POSTagLoader methods is now done in a new 
	  method called _createPOSTagging() to abstract it away.
	- pPOSTaggingType is added as argument to POSTagLoader constructor to
	  customize the type of POS tagged sentences loaded.
	- getPOSTags() is added to POSTagging.
	
	Version 1.0												(07-May-2015)
	- POSTagging and POSTaggedToken are moved to pos.py.
	- getSuffix() is added to POSTaggedToken.
	- getSuffixSeqList() is added to POSTagging.
	- POSTagLoader is moved from dloader/dloader.py.
	- loadFromColumnar() is added to POSTagLoader, POSTagging and POSTaggedToken.
"""

import re
import parse


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# POSTagging
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class POSTagging:
	'''
	POS tagging of a sentence
	
	POSTagging is a list of POSTagToken. 
	
	POS tagging of a sentence can be loaded from various input format. 
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		# list of POSTaggedToken
		self.taggedTokens = []
		self._index = -1
		
	
	
	@property
	def length(self):
		'''
		Returns the length of the underlying sentence
		'''
		
		return len(self.taggedTokens)
		
	
	
	def _createToken(self, pForm = None, pPOS = None):
		'''
		Creates and returns a POS tagged token
		'''
		
		return POSTaggedToken(pForm = pForm, pPOS = pPOS)
		
	
	
	def loadPTBPOSTagging(self, pPTBTree):
		'''
		Loads POS tagging of a sentence from its PTB parse.
		'''
		
		vPTBTree = parse.ConstTree()
		vPTBTree.loadPTBTree(pPTBTree = pPTBTree, pflgExpandTerminal = True)
		
		vlTokens = vPTBTree.getTerminals()
		vlPOSs = vPTBTree.getPOSs()
		
		# sanity check
		if len(vlTokens) != len(vlPOSs):
			raise Exception("Number of terminals and pre-terminals extracted from PTB tree do not match!")
		else:
			for vToken, vPOS in zip(vlTokens, vlPOSs):
				self.taggedTokens.append(self._createToken(vToken, vPOS))
		
	
	
	def loadFromSlashTag(self, pSlashTagLine, pTagSeparator):
		'''
		Loads sentence POS tagging from a given line in SlashTag format
		'''
		
		vlSlashTags = pSlashTagLine.split(' ')
		
		self.taggedTokens = []
		
		for vSlashTagToken in vlSlashTags:
			vPOSTaggedToken = self._createToken()
			vPOSTaggedToken.loadFromSlashTag(vSlashTagToken, pTagSeparator)
			self.taggedTokens.append(vPOSTaggedToken)
		
	
	
	def loadFromColumnar(self, pBlock, pSeparator = '\t'):
		'''
		Loads POS tagging from a block in tab-separated format (e.g. 
		Stanford TSV)
		
		This format is like CoNLL but the first column is the words and
		the second separated by a separator (default tab) is the tag.
		'''
		
		self.taggedTokens = []
		
		for vLine in pBlock.split('\n'):
			vPOSTaggedToken = self._createToken()
			vPOSTaggedToken.loadFromColumnar(vLine, pSeparator)
			self.taggedTokens.append(vPOSTaggedToken)
		
	
	
	def __iter__(self):
		'''
		Python iterator
		'''
		
		self._index = -1
		return self
		
	
	
	def __getitem__(self, pIndex):
		'''
		Python item getter
		'''
		print(self.taggedTokens)
		print(pIn)
		if pIndex == 0 or pIndex > len(self.taggedTokens):
			raise IndexError
		
		return self.taggedTokens[pIndex - 1]
		
	
	
	def __setitem__(self, pIndex, pValue):
		'''
		Python item setter
		'''
		
		if pIndex == 0 or pIndex > len(self.taggedTokens):
			raise IndexError
		
		self.taggedTokens[pIndex - 1] = pValue
		
	
	
	def next(self):
		'''
		Python iterator
		'''
		
		if self._index == len(self.taggedTokens) - 1:
			raise StopIteration
		
		self._index += 1
		return self.taggedTokens[self._index]
		
	
	
	def toCoNLLParserInput(self):
		'''
		Converts to CoNLL format required by the dependency parsers such 
		as Malt and returns the sentence lines.
		'''
		
		return '\n'.join([token.toCoNLLParserInput(str(i)) for i, token in enumerate(self.taggedTokens, start = 1)])
		
	
	
	def toLorgInput(self):
		'''
		Converts to the POS-tagged input format accepted by Lorg (when 
		--input-mode or -i switch is assigned tag)
		'''
		
		return ' '.join([token.toLorgInput() for token in self.taggedTokens])
		
	
	
	def toStanfTSVInput(self):
		'''
		Converts to the TSV input format of Stanford POS tagger
		'''
		
		return '\n'.join([token.toStanfTSVInput() for token in self.taggedTokens]) + '\n'
		
	
	
	def toSVMToolInput(self, pInputType, pflgEOS):
		'''
		Converts to the input format of SVMTool  tagger
		
		pInput type specifies the type of input which is train, test or 
		eval.
		'''
		
		if pflgEOS:
			return '\n'.join([token.toSVMToolInput(pInputType) for token in self.taggedTokens]) + "\n<s>"
		else:
			return '\n'.join([token.toSVMToolInput(pInputType) for token in self.taggedTokens])
		
	
	
	def mapTagging(self, pdTagMap):
		'''
		Maps the POS tags using the specified tag map and returns a new 
		POSTagging with mapped tags
		'''
		
		vNewPOSTagging = POSTagging()
		for vTaggedToken in self.taggedTokens:
			vNewPOSTagging.taggedTokens.append(vTaggedToken.mapTag(pdTagMap))
		
		return vNewPOSTagging
		
	
	
	def getPOSTags(self):
		'''
		Returns the list of POS tags of the sentence tokens
		'''
		
		return [t.tag for t in self]
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# POSTaggedToken
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class POSTaggedToken:
	'''
	Token and its POS tag
	'''
	
	
	def __init__(self, pForm = None, pPOS = None):
		'''
		Constructor
		'''
		
		self.form = pForm
		self.tag = pPOS
		
	
	
	def __str__(self):
		'''
		String representation of the class (SlashTag format)
		'''
		
		return self.form + '/' + self.tag
		
	
	
	def loadFromSlashTag(self, pSlashTagToken, pTagSeparator = '/'):
		'''
		Loads the token and POS tag
		
		The supported format as the moment is token/POS where / can be 
		replaced by any other character.
		'''
		
		vSepPOS = pSlashTagToken.rfind(pTagSeparator)
		
		# sanity check
		if vSepPOS == -1:
			raise Exception("Invalid token/POS fomrat: %s" % pSlashTagToken)
		
		self.form = pSlashTagToken[ : vSepPOS]
		self.tag = pSlashTagToken[vSepPOS + 1: ]
		
	
	
	def loadFromColumnar(self, pLine, pSeparator = '\t'):
		'''
		Loads the token and POS tag from a tab-separated line format (e.g. 
		Stanford TSV)
		
		This format is like CoNLL but the first column is the words and
		the second separated by a separator (default tab) is the tag.
		'''
		
		vLineSplit = pLine.split(pSeparator)
		
		self.form = vLineSplit[0]
		self.tag = vLineSplit[1]
		
	
	
	def toCoNLLParserInput(self, pID):
		'''
		Converts to CoNLL format required by the dependency parsers such 
		as Malt and returns the token line.
		'''
		
		## The ID column should be provided
		
		return '\t'.join([pID, self.form, '_', self.tag, self.tag, '_'])
		
	
	
	def toLorgInput(self):
		'''
		Converts to the POS-tagged input format accepted by Lorg (when 
		--input-mode or -i switch is assigned tag)
		'''
		
		return "%s ( %s )" % (self.form, self.tag)
		
	
	
	def toStanfTSVInput(self):
		'''
		Converts to the TSV input format of Stanford POS tagger
		'''
		
		return "%s\t%s" % (self.form, self.tag)
		
	
	
	def toSVMToolInput(self, pInputType):
		'''
		Converts to the input format of SVMTool tagger
		
		pInput type specifies the type of input which is train, test or 
		eval.
		'''
		
		if pInputType.lower() in ["train", "eval"]:
			return "%s %s" % (self.form, self.tag)
		else:
			return "%s" % self.form
		
	
	
	def mapTag(self, pdTagMap):
		'''
		Maps the POS tag using the specified tag map and returns a new 
		POSTaggedToken with mapped tag
		'''
		
		if self.tag not in pdTagMap:
			raise Exception("%s does not exists in the map!" % self.tag)
		else:
			vNewTag = pdTagMap[self.tag]
		
		return POSTaggedToken(self.form, vNewTag)
		
	
	


#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬
# POSTagLoader
#¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬¬

class POSTagLoader:
	'''
	Class for loading POS-tagged corpus
	'''
	
	
	def __init__(self, pPOSTaggingType):
		'''
		Constructor
		
		pPOSTaggingType is the name of the POSTagging class to be used
		(e.g. POSTagging or foreebank.ForeePOSTagging).
		'''
		
		# list of POS-tagged sentences of type POSTagging
		self.taggedSentences = []
		
		# the class of which type the POS tagged sentences are loaded
		self.posTaggingType = pPOSTaggingType
		
	
	
	def _createPOSTagging(self):
		'''
		Creates and returns a POSTagging instance
		'''
		
		return self.posTaggingType
		
	
	
	def loadFromSlashTag(self, pPOSTaggedFilename, pTagSeparator = '/'):
		'''
		Loads POS taggings from a given file in SlashTag format
		'''
		
		vlTaggedLines = open(pPOSTaggedFilename).read().strip().split('\n')
		
		self.taggedSentences = []
		for vTaggedLine in vlTaggedLines:
			vTaggedSent = self._createPOSTagging()
			vTaggedSent.loadFromSlashTag(vTaggedLine, pTagSeparator)
			self.taggedSentences.append(vTaggedSent)
		
	
	
	def loadFromColumnar(self, pColFilename, pSeparator = '\t'):
		'''
		Loads POS taggings from a given file in tab-separated format (e.g.
		Stanford TSV)
		
		This format is like CoNLL but the first column is the words and
		the second separated by a separator (default tab) is the tag.
		'''
		
		vlBlocks = open(pColFilename).read().strip().split('\n\n')
		
		self.taggedSentences = []
		for vBlock in vlBlocks:
			vTaggedSent = self._createPOSTagging()()
			vTaggedSent.loadFromColumnar(vBlock, pSeparator)
			self.taggedSentences.append(vTaggedSent)
		
	
	
	def loadFromPTB(self, pPTBFilename):
		'''
		Loads POS taggings from a PTB file format
		'''
		
		try:
			vlPTBTrees = open(pPTBFilename).read().strip().split('\n')
		except IOError:
			raise Exception("Cannot open PTB file in %s!" % pPTBFilename)
		
		vPattern = re.compile(r"\(([^ ]+) ([^()]+)\)")
		vlPOSLines = []
		
		self.taggedSentences = []
		for vTree in vlPTBTrees:
			vTaggedSent = self._createPOSTagging()()
			vTaggedSent.loadPTBPOSTagging(vTree)
			self.taggedSentences.append(vTaggedSent)
		
	
	
	def toCoNLLParserInput(self):
		'''
		Converts to CoNLL format required by the dependency parsers such 
		as Malt
		'''
		
		vlCoNLLSents = []
		
		for vTaggedSent in self.taggedSentences:
			vlCoNLLSents.append(vTaggedSent.toCoNLLParserInput())
		
		return "\n\n".join(vlCoNLLSents) + "\n\n"
		
	


