#! /usr/bin/python

## This module defines functions to process Berkeley parser output.
##
## Current version: 1.1
##
## Added:
## - mapToCommonTagSet(): maps the syntactic tags in Berkeley parser 
##   output to a common tag set. 
## - To be used by the above function, mappings are defined in the module
##   as lists of pairs (tuples). This mapping are specified using numbers
##   by callers and identified inside the function.
##   Currently, only one mapping has been defined, with mapping number of
##   1, with Berkeley output with a grammar defined on FTB and Stanford 
##   output with its built-in frenchFactored grammar in mind. The mapping
##   is from these two tag sets to a common one consisting of 25 tags.
##   Note that the mapping does not include top node tag. It is treated
##   inside the function.


from collections import namedtuple
import sys, shutil, os, codecs


##----------------------------------------------------------------------
## Mappings
##
## Note that we also include neutral mappings where source and target tags
## are the same. The reason is to catch unexpected tags not included in 
## the table, through checking for all encountered tags in the input.
##
## clMap1: maps from FTB-trained Berkeley tag set to a common tag set 
##         between itself and frenchFactored-trained Stanford: 43 tags 
##         to a set of 25 tags.

clMap1 = [('A', 'A'),
          ('AdP', 'AdP'),
          ('ADV', 'ADV'),
          ('AP', 'AP'),
          ('C', 'C'),
          ('CL', 'CL'),
          ('COORD', 'COORD'),
          ('D', 'D'),
          ('ET', 'ET'),
          ('I', 'I'),
          ('N', 'N'),
          ('NP', 'NP'),
          ('P', 'P'),
          ('PUNC', 'PUNC'),
          ('PP', 'PP'),
          ('PREF', 'PREF'),
          ('PRO', 'PRO'),
          ('SENT', 'SENT'),
          ('Sint', 'Sint'),
          ('Srel', 'Srel'),
          ('Ssub', 'Ssub'),
          ('V', 'V'),
          ('VN', 'VN'),
          ('VPinf', 'VPinf'),
          ('VPpart', 'VPpart'),
          ('MWA', 'AP'),
          ('MWADV', 'AdP'),
          ('MWC', 'C'),
          ('MWD', 'D'),
          ('MWN', 'NP'),
          ('MWP', 'PP'),
          ('MWPRO', 'PRO'),
          ('MWV', 'V'),
          ('MWET', 'ET'),
          ('MWCL', 'CL')]


##----------------------------------------------------------------------
## Changes the syntactic tags according to the specified mapping.
## Note that the mapping does not include top node tag. It is treated 
## based pChangeTopNode (saying that change the top node tag), and 
## pTopRootTag (specifying the top node tag to be used in the output)
def mapToCommonTagSet(pParseTree, pMapping, pflgChangeTopNode, pTopRootTag, pflgPrint):
 # selecting mapping based on pMapping (currently no selection)
 ##if pMapping == 1:
 vlMapping = clMap1
 
 ## removing current top node tagging:
 ## Anything outside second openning bracket and second last closing 
 ## bracket is considered the top node.
 ## For fun, we call top-node-less tree naked tree

 vStripTree = pParseTree.strip('\n ')

 vNakedTreeStart = vStripTree.find('(', 1)
 vNakedTreeEnd = vStripTree.rfind(')', 0, vStripTree.rfind(')'))
 
 vNakedTree = vStripTree[vNakedTreeStart: vNakedTreeEnd + 1]
 
 vOldTopNodeOpening = vStripTree[0:vNakedTreeStart]
 vOldTopNodeClosing = vStripTree[vNakedTreeEnd + 1:]
 
 
 # mapping tags
 
 vMappedTree = ''
 
 vCurrPos = 0
 while True:
  vTagStart = vNakedTree.find('(', vCurrPos)
  if vTagStart == -1:
   vMappedTree = (vMappedTree + 
                  vNakedTree[vCurrPos:]) 
   break
  else:
   vTagEnd = vNakedTree.find(' ', vTagStart) - 1
   
   vTag = vNakedTree[vTagStart + 1: vTagEnd + 1]
   ## The following acts as a saftey guard against non-recognized tag in
   ## in the input. I.e., any tag not found in the mapping table causes
   ## exception.
   vMappedTag = '(' + [mapping for mapping in vlMapping if mapping[0] == vTag][0][1]
   
   vMappedTree = (vMappedTree + 
                  vNakedTree[vCurrPos: vTagStart] + 
                  vMappedTag +
                  ' ')
   vCurrPos = vTagEnd + 2
   
 
 # adding TOP node
 
 if pflgChangeTopNode:
  vMappedTree = '(' + pTopRootTag.strip() + ' ' + vMappedTree + ')'
 else:
  vMappedTree = vOldTopNodeOpening + vMappedTree + vOldTopNodeClosing
 
 if pflgPrint:
  print vMappedTree
 
 return vMappedTree

