#! /usr/bin/python

"""
	This module provides a class to extracts semantic features for quality 
	estimation.
	
	Version 0.5												(23-May-2014)
	- extractPAMScores() is added to FESeQE.
	
	Version 0.4												(06-Mar-2014)
	- The following methods are added to FESeQE:
	   extractDepSRLWAProjStats()
	
	Version 0.3												(19-Feb-2014)
	- extractCommonPropStructCount() extractCommonPropStructScore() of 
	  FESeQE are refactored to accommodate changes in srl.compareStructWith().
	  It involves renaming compareStructWith() to isStructEqualTo() and 
	  taking predicate label comparison out of it.
	- The following methods are added:
	   extractPropLabelSeq()
	   extractPropSynTagSeq()
	   extractPropdDepRelSeq()
	   extractPropPOSTagSeq()
	
	Version 0.2												(11-Feb-2014)
	- The following methods are added to FESeQE:
	   extractCommonArgCount()
	   _extractCommonPredPropPairs()
	   _extractCommonArgPairs()
	   extractCommonPredScore()
	   extractCommonRoleScore()
	   extractCommonPropStructScore()
	   extractCommonPredRoleScore()
	   extractSrcArgSpanSizeSum()
	   extractTransArgSpanSizeSum()
	   extractRelArgSpanSizeSum() 
	   extractSrcArgSpanSizeSumRatio()
	   extractTransArgSpanSizeSumRatio() 
	   extractRelArgSpanSizeSumRatio()
	   _computeScore()
	   _computePrecision()
	   _computeRecall()
	   _computeF1()
	- extractCommonPredCount() is refactored to use _extractCommonPredPropPairs().
	
	Version 0.1												(05-Feb-2014)
	- extractPropCount(), extractRelPropCount, extractCommonPredCount(),
	  extractArgCount(), extractAvgArgCount(), extractRelAvgArgCount()
	  are added.
	
"""


from parse import parse
from utils import util
from mt import mt
from dloader import dloader
from nlp import nlp, lm
import sys, yaml, copy, pickle



class FESeQE:
	'''
	Feature Extractor for semantic-based quality estimation
	'''
	
	
	# feature extraction methods
	
	def extractPropCount(self, plSRLs):
		'''
		Extracts the number of propositions in the SRL annotation
		'''
		
		vlPropCounts = []
		for vSRL in plSRLs:
			vlPropCounts.append(vSRL.propCount)
		
		return vlPropCounts
		
	
	
	def extractRelPropCount(self, plSrcSRLs, plTransSRLs, pdParams):
		'''
		Extracts relationship between the proposition counts of the source 
		and translation side SRLs
		
		pdParams contains the following parameters:
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for proposition counts!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for proposition counts!" % vParam)
		
		vlRels = []
		for vSrcSRL, vTransSRL in zip(plSrcSRLs, plTransSRLs):
			vSrcPropCount = vSrcSRL.propCount
			vTransPropCount = vTransSRL.propCount
			
			if vRelationship == "diff":
				vlRels.append(vSrcPropCount - vTransPropCount)
			elif vRelationship == "ratio":
				if vTransPropCount == 0:
					if vSrcPropCount == 0:
						vlRels.append(0)
					else:
						# add-one to avoid zero division
						vlRels.append((vSrcPropCount + 1.0) / (vTransPropCount + 1))
				else:
					vlRels.append(vSrcPropCount * 1.0 / vTransPropCount)
		
		return vlRels
		
	
	
	def extractCommonPredCount(self, plSrcSRLs, plTransSRLs, pdParams):
		'''
		Extracts the number of common predicates in the source and translation
		SRLs
		
		The "level" parameter can be set to consider either frameset or 
		lemma. Frameset is the default.
		'''
		
		# setting default
		vLevel = "frameset"
		
		for vParam, vValue in pdParams.iteritems():
			if vParam == "level":
				if vValue != None:
					if vValue in ["frameset"]:
						vLevel = "frameset"
					elif vValue in ["lemma"]:
						vLevel = "lemma"
					else:
						raise Exception("'%s' is not a valid feature parameter for common predicate counts!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for common predicate counts!" % vParam)
		
		vlCommonCounts = []
		for vSrcSRL, vTransSRL in zip(plSrcSRLs, plTransSRLs):
			# extracting propositions with common predicates
			vlCommonCounts.append(len(self._extractCommonPredPropPairs(vSrcSRL, vTransSRL, vLevel)))
		
		return vlCommonCounts
		
	
	
	def extractArgCount(self, plSRLs):
		'''
		Extracts the number of arguments in the SRL annotation
		'''
		
		vlArgCounts = []
		for vSRL in plSRLs:
			vlArgCounts.append(vSRL.globalArgCount)
		
		return vlArgCounts
		
	
	
	def extractRelArgCount(self, plSrcSRLs, plTransSRLs, pdParams):
		'''
		Extracts relationship between the argument counts of the source
		and translation side SRLs
		
		pdParams contains the following parameters:
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for argument counts!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for argument counts!" % vParam)
		
		vlRels = []
		for vSrcSRL, vTransSRL in zip(plSrcSRLs, plTransSRLs):
			vSrcArgCount = vSrcSRL.globalArgCount
			vTransArgCount = vTransSRL.globalArgCount
			
			if vRelationship == "diff":
				vlRels.append(vSrcArgCount - vTransArgCount)
			elif vRelationship == "ratio":
				if vTransArgCount == 0:
					if vSrcArgCount == 0:
						vlRels.append(0)
					else:
						# add-one to avoid zero division
						vlRels.append((vSrcArgCount + 1.0) / (vTransArgCount + 1))
				else:
					vlRels.append(vSrcArgCount * 1.0 / vTransArgCount)
		
		return vlRels
		
	
	
	def extractAvgArgCount(self, plSRLs):
		'''
		Extracts average number of arguments per proposition in the SRL 
		annotation
		'''
		
		vlArgCounts = []
		for vSRL in plSRLs:
			vlArgCounts.append(vSRL.getAvgArgCount())
		
		return vlArgCounts
		
	
	
	def extractRelAvgArgCount(self, plSrcSRLs, plTransSRLs, pdParams):
		'''
		Extracts relationship between the average number of arguments in
		the source and translation side SRLs
		
		pdParams contains the following parameters:
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for average argument counts!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for average argument counts!" % vParam)
		
		vlRels = []
		for vSrcSRL, vTransSRL in zip(plSrcSRLs, plTransSRLs):
			vSrcAvgArgCount = vSrcSRL.getAvgArgCount()
			vTransAvgArgCount = vTransSRL.getAvgArgCount()
			
			if vRelationship == "diff":
				vlRels.append(vSrcAvgArgCount - vTransAvgArgCount)
			elif vRelationship == "ratio":
				if vTransAvgArgCount == 0:
					if vSrcAvgArgCount == 0:
						vlRels.append(0)
					else:
						# add-one to avoid zero division
						vlRels.append((vSrcAvgArgCount + 1.0) / (vTransAvgArgCount + 1))
				else:
					vlRels.append(vSrcAvgArgCount * 1.0 / vTransAvgArgCount)
		
		return vlRels
		
	
	
	def extractCommonRoleCount(self, plSrcSRLs, plTransSRLs, pdParams):
		'''
		Extracts the number of common argument roles between the matching
		predicates of the source and translation SRLs
		
		Note that if a predicate appears more than once in the sentence,
		the matching is based on the order in the sentence since we don't 
		use word alignment here.
		
		The parameter "pred-match-level" is the same as the parameter 
		"level" of common-pred-count feature (see extractCommonPredCount).
		
		If "collapse-adjunct" parameter is given and set to true, the
		adjunct argument labels will be collapsed to a single label.
		'''
		
		# setting defaults
		vLevel = "frameset"
		vflgCollapseAdjunct = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam == "collapse-adjunct":
				if vValue in [True, False]:
					vflgCollapseAdjunct = vValue
				else:
					raise Exception("'%s' is not a valid feature parameter for common argument counts!" % vValue)
			elif vParam == "level":
				if vValue != None:
					if vValue in ["frameset"]:
						vLevel = "frameset"
					elif vValue in ["lemma"]:
						vLevel = "lemma"
					else:
						raise Exception("'%s' is not a valid feature parameter for common argument counts!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for common argument counts!" % vParam)
		
		vlCommonCounts = []
		for vSrcSRL, vTransSRL in zip(plSrcSRLs, plTransSRLs):
			vlCommonCounts.append(0)
			## extracting propositions with common predicates and then
			## counting their common arguments
			for vProp1, vProp2 in self._extractCommonPredPropPairs(vSrcSRL, vTransSRL, vLevel):
				vlCommonCounts[-1] += len(self._extractCommonArgPairs(vProp1, vProp2, vflgCollapseAdjunct))
		
		return vlCommonCounts
		
	
	
	def extractCommonPropStructCount(self, plSrcSRLs, plTransSRLs, pdParams):
		'''
		Extracts the number of common proposition structures in the source
		and translation SRLs
		
		Proposition structure consists of all of its argument roles. Two
		propositions are compared for being common only if their predicates
		match according to criteria defined (i.e. frameset or lemma).
		
		The algorithm relies on the common predicate and argument extraction
		algorithm. Therefore, the parameter set is the same as in common 
		argument.
		'''
		
		# setting defaults
		vLevel = "frameset"
		vflgCollapseAdjunct = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam == "collapse-adjunct":
				if vValue in [True, False]:
					vflgCollapseAdjunct = vValue
				else:
					raise Exception("'%s' is not a valid feature parameter for common proposition counts!" % vValue)
			elif vParam == "level":
				if vValue != None:
					if vValue in ["frameset"]:
						vLevel = "frameset"
					elif vValue in ["lemma"]:
						vLevel = "lemma"
					else:
						raise Exception("'%s' is not a valid feature parameter for common proposition counts!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for common proposition counts!" % vParam)
		
		vlCommonCounts = []
		for vSrcSRL, vTransSRL in zip(plSrcSRLs, plTransSRLs):
			vlCommonCounts.append(0)
			## extracting propositions with common predicates and then 
			## checking if their structure match
			for vProp1, vProp2 in self._extractCommonPredPropPairs(vSrcSRL, vTransSRL, vLevel):
				if vProp1.isStructEqualTo(vProp2, vflgCollapseAdjunct) == True:
					vlCommonCounts[-1] += 1
		
		return vlCommonCounts
		
	
	
	def extractCommonPredScore(self, plSrcSRLs, plTransSRLs, pdParams):
		'''
		Extracts the score of common predicates in the source and translation
		SRLs
		
		The score can be precision, recall, or F1 identified by "score" 
		parameter.
		
		The feature is similar to common predicate count but with the 
		number of predicate in the source and/or translation taken into
		account by dividing the former by either of the latter (recall
		and precision respectively) or their harmonic mean (F1). 
		
		The level at which the predicates are compared is given in "level"
		parameter.
		'''
		
		# setting defaults
		vLevel = "frameset"
		vScore = "F1"
		
		for vParam, vValue in pdParams.iteritems():
			if vParam == "level":
				if vValue != None:
					if vValue in ["frameset"]:
						vLevel = "frameset"
					elif vValue in ["lemma"]:
						vLevel = "lemma"
					else:
						raise Exception("'%s' is not a valid feature parameter for common predicate scores!" % vValue)
			elif vParam == "score":
				if vValue != None:
					if vValue.lower().startswith("prec"):
						vScore = "precision"
					elif vValue.lower() == "recall":
						vScore = "recall"
					elif vValue.lower() == "f1":
						vScore = "F1"
					else:
						raise Exception("'%s' is not a valid feature parameter for common predicate scores!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for common predicate scores!" % vParam)
		
		vlScores = []
		for vSrcSRL, vTransSRL in zip(plSrcSRLs, plTransSRLs):
			# extracting number of propositions with common predicates
			vCommonPredCount = len(self._extractCommonPredPropPairs(vSrcSRL, vTransSRL, vLevel))
			
			# computing scores
			
			## If there is no predicate in the source and translation, we
			## consider all scores to be 1 mostly because it's a match 
			## after all.
			if vSrcSRL.propCount == vTransSRL.propCount == 0:
				vlScores.append(1.0)
			else:
				vlScores.append(self._computeScore(vCommonPredCount, vSrcSRL.propCount, vTransSRL.propCount, vScore))
		
		return vlScores
		
	
	
	def extractCommonRoleScore(self, plSrcSRLs, plTransSRLs, pdParams):
		'''
		Extracts the score of common argument roles between the matching
		predicates of the source and translation SRLs
		
		The score can be precision, recall, or F1 identified by "score" 
		parameter.
		
		The feature is similar to common argument role count but with the 
		number of arguemnts in the source and/or translation taken into
		account by dividing the former by either of the latter (recall
		and precision respectively) or their harmonic mean (F1).
		
		Note that the total number of argument in the sentence is used
		to compute scores rather than total number of arguments of only 
		matching predicates and rather than computing them locally for each
		predicate and averaging.
		
		Predicate match level and whether adjuncts should be collapsed 
		or not is given as parameter.
		'''
		
		# setting defaults
		vLevel = "frameset"
		vflgCollapseAdjunct = False
		vScore = "F1"
		
		for vParam, vValue in pdParams.iteritems():
			if vParam == "collapse-adjunct":
				if vValue in [True, False]:
					vflgCollapseAdjunct = vValue
				else:
					raise Exception("'%s' is not a valid feature parameter for common role scores!" % vValue)
			elif vParam == "level":
				if vValue != None:
					if vValue in ["frameset"]:
						vLevel = "frameset"
					elif vValue in ["lemma"]:
						vLevel = "lemma"
					else:
						raise Exception("'%s' is not a valid feature parameter for common role scores!" % vValue)
			elif vParam == "score":
				if vValue != None:
					if vValue.lower().startswith("prec"):
						vScore = "precision"
					elif vValue.lower() == "recall":
						vScore = "recall"
					elif vValue.lower() == "f1":
						vScore = "F1"
					else:
						raise Exception("'%s' is not a valid feature parameter for common role scores!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for common argument scores!" % vParam)
		
		vlScores = []
		for vSrcSRL, vTransSRL in zip(plSrcSRLs, plTransSRLs):
			# extracting total number of arguments
			vSrcArgCount = sum([p.getArgCount(vflgCollapseAdjunct) for p in vSrcSRL.propositions]) 
			vTransArgCount = sum([p.getArgCount(vflgCollapseAdjunct) for p in vTransSRL.propositions]) 
			
			# extracting number of common arguments
			vCommonRoleCount = 0
			for vProp1, vProp2 in self._extractCommonPredPropPairs(vSrcSRL, vTransSRL, vLevel):
				vCommonRoleCount += len(self._extractCommonArgPairs(vProp1, vProp2, vflgCollapseAdjunct))
			
			# computing scores
			
			## If there is no predicate in the source and translation, we
			## consider all scores to be 1 mostly because it's a match 
			## after all.
			if vSrcSRL.propCount == vTransSRL.propCount == 0:
				vlScores.append(1.0)
			else:
				vlScores.append(self._computeScore(vCommonRoleCount, vSrcArgCount, vTransArgCount, vScore))
		
		return vlScores
		
	
	
	def extractCommonPropStructScore(self, plSrcSRLs, plTransSRLs, pdParams):
		'''
		Extracts the score of common proposition structures in the source
		and translation SRLs
		
		Proposition structure consists of all of its argument roles. Two
		propositions are compared for being common only if their predicates
		match according to criteria defined (i.e. frameset or lemma).
		
		The algorithm relies on the common predicate and argument extraction
		algorithms. Therefore, the parameter set is the same as in common 
		argument score.
		'''
		
		# setting defaults
		vLevel = "frameset"
		vflgCollapseAdjunct = False
		vScore = "F1"
		
		for vParam, vValue in pdParams.iteritems():
			if vParam == "collapse-adjunct":
				if vValue in [True, False]:
					vflgCollapseAdjunct = vValue
				else:
					raise Exception("'%s' is not a valid feature parameter for common proposition structure scores!" % vValue)
			elif vParam == "level":
				if vValue != None:
					if vValue in ["frameset"]:
						vLevel = "frameset"
					elif vValue in ["lemma"]:
						vLevel = "lemma"
					else:
						raise Exception("'%s' is not a valid feature parameter for common proposition structure scores!" % vValue)
			elif vParam == "score":
				if vValue != None:
					if vValue.lower().startswith("prec"):
						vScore = "precision"
					elif vValue.lower() == "recall":
						vScore = "recall"
					elif vValue.lower() == "f1":
						vScore = "F1"
					else:
						raise Exception("'%s' is not a valid feature parameter for common proposition structure scores!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for common proposition structure scores!" % vParam)
		
		vlScores = []
		for vSrcSRL, vTransSRL in zip(plSrcSRLs, plTransSRLs):
			# extracting number of common proposition structures
			vCommonPropCount = 0
			for vProp1, vProp2 in self._extractCommonPredPropPairs(vSrcSRL, vTransSRL, vLevel):
				if vProp1.isStructEqualTo(vProp2, vflgCollapseAdjunct) == True:
					vCommonPropCount += 1
			
			# computing scores
			
			## If there is no predicate in the source and translation, we
			## consider all scores to be 1 mostly because it's a match 
			## after all.
			if vSrcSRL.propCount == vTransSRL.propCount == 0:
				vlScores.append(1.0)
			else:
				vlScores.append(self._computeScore(vCommonPropCount, vSrcSRL.propCount, vTransSRL.propCount, vScore))
		
		return vlScores
		
	
	
	def extractCommonPredRoleScore(self, plSrcSRLs, plTransSRLs, pdParams):
		'''
		Extracts the score of common predicates and argument roles between
		the matching predicates of the source and translation SRLs
		
		The score can be precision, recall, or F1 identified by "score" 
		parameter.
		
		The feature is the combination of common predicate score and common
		argument role score. This is the way SRL is scored in CoNLL 2009
		shared task, though without lexical match. Therefore, similar to
		SRl scores, predicate label and argument role are counted the same 
		in scoring.
		'''
		
		# setting defaults
		vLevel = "frameset"
		vflgCollapseAdjunct = False
		vScore = "F1"
		
		for vParam, vValue in pdParams.iteritems():
			if vParam == "collapse-adjunct":
				if vValue in [True, False]:
					vflgCollapseAdjunct = vValue
				else:
					raise Exception("'%s' is not a valid feature parameter for common predicate/role scores!" % vValue)
			elif vParam == "level":
				if vValue != None:
					if vValue in ["frameset"]:
						vLevel = "frameset"
					elif vValue in ["lemma"]:
						vLevel = "lemma"
					else:
						raise Exception("'%s' is not a valid feature parameter for common predicate/role scores!" % vValue)
			elif vParam == "score":
				if vValue != None:
					if vValue.lower().startswith("prec"):
						vScore = "precision"
					elif vValue.lower() == "recall":
						vScore = "recall"
					elif vValue.lower() == "f1":
						vScore = "F1"
					else:
						raise Exception("'%s' is not a valid feature parameter for common predicate/role scores!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for common predicate/role scores!" % vParam)
		
		vlScores = []
		for vSrcSRL, vTransSRL in zip(plSrcSRLs, plTransSRLs):
			# number of predicates
			vSrcPredArgCount = vSrcSRL.propCount
			vTransPredArgCount = vTransSRL.propCount
			
			# adding total number of arguments
			vSrcPredArgCount += sum([p.getArgCount(vflgCollapseAdjunct) for p in vSrcSRL.propositions]) 
			vTransPredArgCount += sum([p.getArgCount(vflgCollapseAdjunct) for p in vTransSRL.propositions]) 
			
			# extracting number of common predicates and arguments
			vCommonPredRoleCount = 0
			for vProp1, vProp2 in self._extractCommonPredPropPairs(vSrcSRL, vTransSRL, vLevel):
				# for common predicate
				vCommonPredRoleCount += 1
				# for common arguments
				vCommonPredRoleCount += len(self._extractCommonArgPairs(vProp1, vProp2, vflgCollapseAdjunct))
			
			# computing scores
			
			## If there is no predicate in the source and translation, we
			## consider all scores to be 1 mostly because it's a match 
			## after all.
			if vSrcSRL.propCount == vTransSRL.propCount == 0:
				vlScores.append(1.0)
			else:
				vlScores.append(self._computeScore(vCommonPredRoleCount, vSrcPredArgCount, vTransPredArgCount, vScore))
		
		return vlScores
		
	
	
	def extractArgSpanSizeSum(self, plSRLs):
		'''
		Extracts the sum of span sizes of arguments in given SRL
		'''
		
		vlSizeSums = []
		for vSRL in plSRLs:
			vlSizeSums.append(sum([a.size for p in vSRL.propositions for a in p.arguments]))
		
		return vlSizeSums
		
	
	
	def extractRelArgSpanSizeSum(self, plSrcSRLs, plTransSRLs, pdParams):
		'''
		Extracts the relationship between the sums of span sizes of arguments
		in the source and translation sides.
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for arguments span size!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for arguments span size!" % vParam)
		
		vlSrcSizeSums = self.extractArgSpanSizeSum(plSrcSRLs)
		vlTransSizeSums = self.extractArgSpanSizeSum(plTransSRLs)
		vlRels = []
		for vSrcSizeSum, vTransSizeSum in zip(vlSrcSizeSums, vlTransSizeSums):
			if vRelationship == "diff":
				vlRels.append(vSrcSizeSum - vTransSizeSum)
			elif vRelationship == "ratio":
				if vTransSizeSum == 0:
					if vSrcSizeSum == 0:
						vlRels.append(0)
					else:
						# add-one to avoid zero division
						vlRels.append((vSrcSizeSum + 1.0) / (vTransSizeSum + 1))
				else:
					vlRels.append(vSrcSizeSum * 1.0 / vTransSizeSum)
		
		return vlRels
		
	
	
	def extractArgSpanSizeSumRatio(self, plSRLs, plSentSurfaces):
		'''
		Extracts ratio of sum of span sizes of arguments in given SRL to 
		its sentence length
		'''
		
		vlSizeSums = []
		for vSRL, vSurface in zip(plSRLs, plSentSurfaces):
			vlSizeSums.append(sum([a.size for p in vSRL.propositions for a in p.arguments]) * 1.0 /
			                  len(vSurface.split(' ')))
		
		return vlSizeSums
		
	
	
	def extractRelArgSpanSizeSumRatio(self, plSrcSRLs, plTransSRLs, plSrcSentSurfaces, plTransSentSurfaces, pdParams):
		'''
		Extracts the relationship between the ratio of sums of span sizes
		of arguments to sentence lengths in the source and translation sides
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for arguments span size!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for arguments span size!" % vParam)
		
		vlSrcSizeSumRatios = self.extractArgSpanSizeSumRatio(plSrcSRLs, plSrcSentSurfaces)
		vlTransSizeSumRatios = self.extractArgSpanSizeSumRatio(plTransSRLs, plTransSentSurfaces)
		vlRels = []
		for vSrcSizeSumRatio, vTransSizeSumRatio in zip(vlSrcSizeSumRatios, vlTransSizeSumRatios):
			if vRelationship == "diff":
				vlRels.append(vSrcSizeSumRatio - vTransSizeSumRatio)
			elif vRelationship == "ratio":
				if vTransSizeSumRatio == 0:
					if vSrcSizeSumRatio == 0:
						vlRels.append(0)
					else:
						# add-one to avoid zero division
						vlRels.append((vSrcSizeSumRatio + 1.0) / (vTransSizeSumRatio + 1))
				else:
					vlRels.append(vSrcSizeSumRatio * 1.0 / vTransSizeSumRatio)
		
		return vlRels
		
	
	
	def extractPropLabelSeq(self, plSRLs, pdParams):
		'''
		Extracts the predicate/argument label sequences of propositions
		of given SRL
		'''
		
		# setting defaults
		vPredLabel = "frameset"
		vflgCollapseAdjunct = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam == "collapse-adjunct":
				if vValue in [True, False]:
					vflgCollapseAdjunct = vValue
				else:
					raise Exception("'%s' is not a valid feature parameter value for proposition labeling sequence!" % vValue)
			elif vParam == "predicate-label":
				if vValue != None:
					if vValue.lower() == "none":
						vPredLabel = "none"
					elif vValue in ["frameset"]:
						vPredLabel = "frameset"
					elif vValue in ["lemma"]:
						vPredLabel = "lemma"
					else:
						raise Exception("'%s' is not a valid feature parameter value for proposition labeling sequence!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for proposition labeling sequence!" % vParam)
		
		vlLabelSeqs = []
		for vSRL in plSRLs:
			vlThisLabelSeqs = []
			for vProp in vSRL.propositions:
				vlThisLabelSeqs.append(self._extractPropLabelSeq(vProp, pPredLabel = vPredLabel, pflgCollapseAdjunct = vflgCollapseAdjunct))
			
			vlLabelSeqs.append(vlThisLabelSeqs)
		
		return vlLabelSeqs
		
	
	
	def extractPropSynTagSeq(self, plSRLs, pdParams):
		'''
		Extracts the syntactic tag sequences of predicates and arguments 
		of propositions of given SRL
		'''
		
		vlTagSeqs = []
		for vSRL in plSRLs:
			vlThisTagSeqs = []
			for vProp in vSRL.propositions:
				vlThisTagSeqs.append(self._extractPropSynTagSeq(vProp))
			
			vlTagSeqs.append(vlThisTagSeqs)
		
		return vlTagSeqs
		
	
	
	def extractPropDepRelSeq(self, plSRLs, pdParams):
		'''
		Extracts the dependency relation sequences of predicates and arguments 
		of propositions of given SRL
		'''
		
		vlRelSeqs = []
		for vSRL in plSRLs:
			vlThisRelSeqs = []
			for vProp in vSRL.propositions:
				vlThisRelSeqs.append(self._extractPropDepRelSeq(vProp))
			
			vlRelSeqs.append(vlThisRelSeqs)
		
		return vlRelSeqs
		
	
	
	def extractPropPOSTagSeq(self, plSRLs, pdParams):
		'''
		Extracts the POS tag sequences of predicates and arguments of 
		propositions of given SRL
		'''
		
		vlTagSeqs = []
		for vSRL in plSRLs:
			vlThisTagSeqs = []
			for vProp in vSRL.propositions:
				vlThisTagSeqs.append(self._extractPropPOSTagSeq(vProp))
			
			vlTagSeqs.append(vlThisTagSeqs)
		
		return vlTagSeqs
		
	
	
	def extractDepSRLWAProjStats(self, pdParams):
		'''
		Extracts statistic from dependency SRL projection from source
		to translations using word alignments
		
		The projection statistics are supposed to be provided in feature
		parameters through a pickle file containing list of outputs of 
		getStat() method of srl.SRLProject.
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam.lower() == "projection statistics":
				try:
					vlProjStats = pickle.load(open(vValue))
				except IOError:
					raise Exception("Cannot open projection pickle file at '%s'!" % vValue)
			elif vParam.lower().startswith("stat"):
				vStat = vValue.lower()
			else:
				raise Exception("'%s' is not a valid feature parameter for statistics of dependency SRL projection!" % vParam)
		
		vlStats = []
		for vProj in vlProjStats:
			vdCntStat = vProj[0]
			vdAggCntStat = vProj[1]
			vdPctStat = vProj[2]
			vdAggPctStat = vProj[3]
			
			vflgStatFound = False
			for k, v in vdCntStat.items() + vdAggCntStat.items() + vdPctStat.items() + vdAggPctStat.items():
				if k.lower().strip() == vStat.lower().strip():
					vlStats.append(v)
					vflgStatFound = True
					break
			
			if not vflgStatFound:
				raise Exception("Invalid statistics: %s" % vStat)
		
		return vlStats
		
		
	
	
	def extractPAMScores(self, pdParams):
		'''
		Extracts PAM (predicate-argument match) scores
		
		The PAM scores are provided in a file via feature parameters. The
		file contains a score per line.
		'''
		
		vlScores = []
		
		for vParam, vValue in pdParams.iteritems():
			if vParam.lower() == "scores":
				vlScores = [float(s) for s in open(vValue).read().strip().split('\n')]
			else:
				raise Exception("'%s' is not a valid feature parameter for PAM features!" % vParam)
		
		return vlScores
		
	
	
	
	# internal methods
	
	def _extractCommonPredPropPairs(self, pSRL1, pSRL2, pPredMatchLevel = "frameset"):
		'''
		Extracts from two given SRLs pairs of propositions whose predicate
		match based on the given level
		
		The predicate matching level can be frameset or lemma
		'''
		
		## coping list of propositions of second SRL into a new list to 
		## be able to delete form it later
		vlProps2 = pSRL2.propositions[ : ]
		
		# extracting propositions with matching predicates (replicates counted)
		vlCommonPredPropPairs = []
		for vProp1 in pSRL1.propositions:
			vProp2Itr = 0
			for vProp2 in vlProps2:
				if vProp1.predicate.compareFrameWith(vProp2.predicate, pPredMatchLevel) == True:
					del vlProps2[vProp2Itr]
					vlCommonPredPropPairs.append((vProp1, vProp2))
					break
				else:
					vProp2Itr += 1
					continue
		
		return vlCommonPredPropPairs
		
	
	
	def _extractCommonArgPairs(self, pProp1, pProp2, pflgCollapsAdjunct = False):
		'''
		Extracts from given two propositions, pairs of arguments with the 
		same role labels
		
		It takes into account the replicated arguments, i.e. treat them
		separately. Therefore, if an argument role appears twice in each
		proposition, both of them will be returned.
		
		If pflgCollapsAdjunct is set to true, the adjunct argument labels
		will be collapsed to a single label.
		
		Note that the method still does not collapse R-Args (relative pronoun
		arguments) and C-Args (argument continuation).
		'''
		
		## coping list of arguments of second proposition into a new list 
		## to be able to delete form it later
		vlArgs2 = pProp2.arguments[ : ]
		
		## extracting common arguments (arguments with same role labels; 
		## replicates counted)
		vlCommonArgPairs = []
		for vArg1 in pProp1.arguments:
			vArg2Itr = 0
			for vArg2 in vlArgs2:
				if vArg1.compareRoleWith(vArg2, pflgCollapsAdjunct) == True:
					del vlArgs2[vArg2Itr]
					vlCommonArgPairs.append((vArg1, vArg2))
					break
				else:
					vArg2Itr += 1
					continue
		
		return vlCommonArgPairs
		
	
	
	def _computeScore(self, pCorrectCount, pGoldCount, pPredCount, pScore):
		'''
		Computes either of precision, recall or F1 of the predictions
		
		Score type is identified by pScore.
		'''
		
		if pScore == "precision":
			return self._computePrecision(pCorrectCount, pPredCount)
		elif pScore == "recall":
			return self._computeRecall(pCorrectCount, pGoldCount)
		elif pScore == "F1":
			return self._computeF1(pCorrectCount, pGoldCount, pPredCount)
		
	
	
	def _computePrecision(self, pCorrectCount, pPredCount):
		'''
		Computes either of precision the predictions
		'''
		
		if pPredCount == 0:
			return 0
		else:
			return pCorrectCount * 1.0 / pPredCount
		
	
	
	def _computeRecall(self, pCorrectCount, pGoldCount):
		'''
		Computes either of recall the predictions
		'''
		
		if pGoldCount == 0:
			return 0
		else:
			return pCorrectCount * 1.0 / pGoldCount
		
	
	
	def _computeF1(self, pCorrectCount, pGoldCount, pPredCount):
		'''
		Computes either of F1 the predictions
		'''
		
		vPrec = self._computePrecision(pCorrectCount, pPredCount)
		vRecall = self._computeRecall(pCorrectCount, pGoldCount)
		if vPrec + vRecall == 0:
			return 0
		else:
			return 2 * vPrec * vRecall / (vPrec + vRecall)
		
	
	
	def _extractPropLabelSeq(self, pProp, pPredLabel = "frameset", pflgCollapseAdjunct = False):
		'''
		Extracts and returns the labeling sequence of given proposition
		'''
		
		if pPredLabel == "none":
			return '_'.join([a.getLabel(pflgCollapseAdjunct) for a in pProp.arguments])
		elif pPredLabel == "frameset":
			return pProp.predicate.frameset + '_' + '_'.join([a.getLabel(pflgCollapseAdjunct) for a in pProp.arguments])
		elif pPredLabel == "lemma":
			return pProp.predicate.lemma + '_' + '_'.join([a.getLabel(pflgCollapseAdjunct) for a in pProp.arguments])
		
	
	
	def _extractPropSynTagSeq(self, pProp):
		'''
		Extracts and returns the syntactic tag sequence of given proposition
		
		Note that the underlying SRL needs to be of constituency type.
		'''
		
		return pProp.predicate.getSynTag() + '_' + '_'.join([a.getSynTag() for a in pProp.arguments])
		
	
	
	def _extractPropDepRelSeq(self, pProp):
		'''
		Extracts and returns the dependency relation sequence of given 
		proposition
		
		Note that the underlying SRL needs to be of dependency type.
		'''
		
		return pProp.predicate.getDepRel() + '_' + '_'.join([a.getDepRel() for a in pProp.arguments])
		
	
	
	def _extractPropPOSTagSeq(self, pProp):
		'''
		Extracts and returns the POS tag sequence of given proposition
		
		Note that the underlying SRL needs to be of dependency type.
		'''
		
		return pProp.predicate.getPOSTag() + '_' + '_'.join([a.getPOSTag() for a in pProp.arguments])
		
	
	
