#! /usr/bin/python

"""
	This module provides a class to extracts syntax and semantics features
	for quality estimation.
	
	Version 1.4												(11-May-2015)
	- Minor changes are made to comply with the changes in svmfg module.
	
	Version 1.3												(27-May-2014)
	- extractPAMScores() is added to FESSQE.
	- Pre-extracted feature is supported. extractPreex() is added for this
	  purpose. It is used to load individual features already extracted
	  into a file (one column of feature values). 
	
	Version 1.2												(06-Mar-2014)
	- Semantic features are added
	
	Version 1.1												(19-Feb-2014)
	- This module previously contained fesyqe.py content. Now, fesyqe.py
	  contains only syntactic features and feseqe.py semantic ones, but 
	  both are interfaced with this module. So, codes importing fessqe.py
	  will still work fine.
	- loadSrcDepTrees(), loadTransDepTrees() and _loadDepTrees() are renamed
	  to loadSrcDependencies(), loadTransDependencies(), _loadDependencies()
	  respectively to mean to load both syntactic and semantic dependencies.
	- _loadDependencies() is upgraded to load semantic dependencies.
	- loadSrcConstSRLsFromDepSRLs(), loadTransConstSRLsFromDepSRLs(), 
	  _loadConstSRLsFromDepSRLs() are added to FESSQE.
	- fesyqe, feseqe, srcDepSRL, transDepSRL, srcConstSRL, transConstSRL, 
	  srcLanguage, transLanguage are added to FESSQE.
	- pflgExpandTerminal option is added to constituency parse loaders.
	- srcDepSRL, transDepSRL, srcConstSRL and transConstSRL in FESSQE are
	  renamed to srcDepSRLs, transDepSRLs, srcConstSRLs and transConstSRLs.
	- Semantic features are added: 
	   source-proposition-count
	   translation-proposition-count
	   proposition-count-relationship
	   source-translation-common-predicate-count
	   source-argument-count
	   translation-argument-count
	   argument-count-relationship
	   source-average-argument-count
	   translation-average-argument-count
	   average-argument-count-relationship
	   source-translation-common-role-count
	   source-translation-common-proposition-structure-count
	   source-translation-common-predicate-score
	   source-translation-common-role-score
	   source-translation-common-proposition-structure-score
	   source-translation-common-predicate-role-score
	   source-argument-span-size-sum
	   translation-argument-span-size-sum
	   argument-span-size-sum-relationship
	   source-argument-span-size-ratio-sum
	   translation-argument-span-size-ratio-sum
	   argument-span-size-sum-ratio-relationship
	   source-proposition-label-sequence
	   translation-proposition-label-sequence
	   source-proposition-syntax-tag-sequence
	   translation-proposition-syntax-tag-sequence
	   source-proposition-dependency-relation-sequence
	   translation-proposition-dependency-relation-sequence
	   source-proposition-pos-tag-sequence
	   translation-proposition-pos-tag-sequence
"""

import fesyqe, feseqe
from ml import svmfg
from dloader import dloader


class FESSQE:
	'''
	Feature Extractor for Syntax and Semantic-based quality estimation
	'''
	
	
	def __init__(self, pSrcLanguage, pTransLanguage):
		'''
		Constructor
		'''
		
		self.srcLanguage = pSrcLanguage
		self.transLanguage = pTransLanguage
		
		self.fesyqe = fesyqe.FESyQE()
		self.feseqe = feseqe.FESeQE()
		
		# source side constituency parses (list of ConstParses)
		self.srcConstParses = []
		# translation side constituency parses (list of ConstParses)
		self.transConstParses = []
		# source side dependency parses (list of DepTree)
		self.srcDepTrees = []
		# translation side dependency parses (list of DepTree)
		self.transDepTrees = []
		# source side dependency-based semantic role labeling
		self.srcDepSRLs = []
		# translation side dependency-based semantic role labeling
		self.transDepSRLs = []
		# source side constituency-based semantic role labeling
		self.srcConstSRLs = []
		# translation side constituency-based semantic role labeling
		self.transConstSRLs = []
		
	
	
	@property
	def srcConstTrees(self):
		'''
		Returns the source side parse trees (the topmost of n-best)
		'''
		
		return [parses.tree for parses in self.srcConstParses]
		
	
	
	@property
	def transConstTrees(self):
		'''
		Returns the source side parse trees (the topmost of n-best)
		'''
		
		return [parses.tree for parses in self.transConstParses]
		
	
	
	@property
	def srcSentSurfaces(self):
		'''
		Returns the surface forms of the source side sentences
		'''
		
		## It first checks dependency trees since they are cheaper to 
		## extract surface form from. Note that if none of the trees are
		## loaded it returns empty list.
		
		if self.isSrcDepTreesLoaded():
			return [dt.surface for dt in self.srcDepTrees]
		else:
			return [ct.surface for ct in self.srcConstTrees]
		
	
	
	@property
	def transSentSurfaces(self):
		'''
		Returns the surface forms of the translation side sentences
		'''
		
		## It first checks dependency trees since they are cheaper to 
		## extract surface form from. Note that if none of the trees are
		## loaded it returns empty list.
		
		if self.isTransDepTreesLoaded():
			return [dt.surface for dt in self.transDepTrees]
		else:
			return [ct.surface for ct in self.transConstTrees]
		
	
	
	def isSrcDepTreesLoaded(self):
		'''
		Returns true is the source side dependency trees are loaded
		'''
		
		return len(self.srcDepTrees) != 0
		
	
	
	def isTransDepTreesLoaded(self):
		'''
		Returns true is the translation side dependency trees are loaded
		'''
		
		return len(self.transDepTrees) != 0
		
	
	
	## Data preparation #####
	
	def loadSrcConstParses(self, pParseFile, pInputFormat = "ptb", pflgExpandTerminal = True):
		'''
		Loads source side constituency parses.
		'''
		
		self.srcConstParses = self._loadConstParses(pParseFile, pInputFormat, pflgExpandTerminal)
		
	
	
	def loadTransConstParses(self, pParseFile, pInputFormat = "ptb", pflgExpandTerminal = True):
		'''
		Loads translation side constituency parses.
		'''
		
		self.transConstParses = self._loadConstParses(pParseFile, pInputFormat, pflgExpandTerminal)
		
	
	
	def _loadConstParses(self, pParseFile, pInputFormat, pflgExpandTerminal):
		'''
		Loads constituency parses.
		
		Parse tree file can be in any of the following formats:
		- ptb or penn: Penn Treebank bracketing format: e.g. (S (NP (NNP ...
		- lorg: the verbose output of LORG (use ptb if it is not verbose)
		'''
		
		vFormat = pInputFormat.lower()
		vParseLoader = dloader.ConstParseLoader()
		
		if vFormat in ["ptb", "penn"]:
			vParseLoader.loadPTBTrees(pParseFile, pflgExpandTerminal = pflgExpandTerminal)
		elif vFormat == "lorg":
			vParseLoader.loadLorgParses(pParseFile, pflgExpandTerminal = pflgExpandTerminal)
		elif vFormat == "log+tree":
			vParseLoader.loadParses(pParseFile, pflgExpandTerminal = pflgExpandTerminal)
		
		return vParseLoader.parses
		
	
	
	def loadSrcDependencies(self, pDepFile, pdAnnotationOptions, pInputFormat = "conll2006"):
		'''
		Loads source side dependency parses
		'''
		
		self.srcDepTrees, self.srcDepSRLs = self._loadDependencies(pDepFile, self.srcLanguage, pdAnnotationOptions, pInputFormat)
		
	
	
	def loadTransDependencies(self, pDepFile, pdAnnotationOptions, pInputFormat = "conll2006"):
		'''
		Loads translation side dependency parses
		'''
		
		self.transDepTrees, self.transDepSRLs = self._loadDependencies(pDepFile, self.transLanguage, pdAnnotationOptions, pInputFormat)
		
	
	
	def _loadDependencies(self, pDepFile, pLanguage, pdAnnotationOptions, pInputFormat = "conll2006"):
		'''
		Loads dependency parses
		
		Dependency parse file can be in any of the following formats:
		- conllxxxx: CoNLL format where xxxx is the optional version
		- parc700: the format of const2dep converter for French
		'''
		
		vFormat = pInputFormat.lower()
		
		if vFormat.startswith("conll"):
			vCoNLLLoader = dloader.CoNLLDataLoader(pDepFile, pLanguage = pLanguage, pFormatVersion = pInputFormat, pflgVerbos = False)
			
			if len(vFormat) == 9:
				vVersion = vFormat[5:]
			
			vlDepTrees = []
			vlSRLs = []
			for vCoNLLSent in vCoNLLLoader.conllSentences:
				vlDepTrees.append(vCoNLLSent.depTree(pdAnnotationOptions))
				## for CoNLL 2008 and 2009, SRL is also loaded both into 
				## self.SRL and into dependency tree itself
				if vCoNLLSent.version in ["2008", "2009"]:
					vlSRLs.append(vCoNLLSent.srl)
					vlDepTrees[-1].loadSRL(vCoNLLSent.srl)
			
			return vlDepTrees, vlSRLs
		elif vFormat == "parc700":
			vDepLoader = dloader.DepParseLoader()
			vDepLoader.loadPARC700Deps(pDepFile)
			
			return vDepLoader.depTrees, []
		
	
	
	def loadSrcConstSRLsFromDepSRLs(self, pConversionMethod = "elevate"):
		'''
		Loads the constituency SRL annotation of source side from alread-
		loaded dependency SRL
		'''
		
		self.srcConstSRLs = self._loadConstSRLsFromDepSRLs(self.srcConstTrees, self.srcDepSRLs, pConversionMethod)
		
	
	
	def loadTransConstSRLsFromDepSRLs(self, pConversionMethod = "elevate"):
		'''
		Loads the constituency SRL annotation of translation side from 
		alread-loaded dependency SRL
		'''
		
		self.transConstSRLs = self._loadConstSRLsFromDepSRLs(self.transConstTrees, self.transDepSRLs, pConversionMethod)
		
	
	
	def _loadConstSRLsFromDepSRLs(self, plConstTrees, plDepSRLs, pConversionMethod):
		'''
		Loads the constituency SRL annotations from dependency SRLs
		'''
		
		vlConstSRLs = []
		for vConstTree, vDepSRL in zip(plConstTrees, plDepSRLs):
			vlConstSRLs.append(vDepSRL.convertToConst(vConstTree, pConversionMethod))
		
		return vlConstSRLs
		
	
	
	def mapConstTreeTags(self, plTrees, pMapFile):
		'''
		OBSOLETE: embed into the caller functions for efficiency
		
		Maps syntactic tags of constituency trees to new tags based on map
		in pMapFile and returns the new trees.
		
		Syntactic tags include phrase categories and POS tags.
		'''
		
		try:
			vdMap = parse.loadTagMap(open(pMapFile).read())
		except IOError:
			raise Exception("Cannot open tag map file: %s" % pMapFile)
		
		vlMappedTrees = []
		for vTree in plTrees:
			vNewTree = copy.deepcopy(vTree)
			vNewTree.mapTags(vdMap)
			vlMappedTrees.append(vNewTree)
		
		return vlMappedTrees
		
	
	
	## Generating outputs #####
	
	def genSVMFVectors(self, pConfig = '', pflgRawVectors = False):
		'''
		Generates SVM feature vectors based on the specified configuration
		in YAML-compatible format.
		
		Feature vectors are returned in list of strings, one per instance,
		in the classic SVM input format, e.g.:
		    1:0.01 2:0.875 3:0.2
		    
		It can optionally output raw feature and value pairs in the following 
		format:
		    <FEATURE 1 NAME=FEATURE 1 VALUE> <FEATURE 2 NAME=FEATURE 2 VALUE> ...
		
		The pFConfig file contains the feature configuration. See fvg.FVGen
		for the description of its format.
		'''
		
		vFVG = svmfg.SVMFVGen(pConfig)
		vlFeatures = vFVG.getFeatures()
		
		## TODO: use vFVG.getRawFVectors() as in pos.posamender
		## Each element of raw feature vector list is a dictionary containing
		## feature names and their value for one instance.
		vlRawFVectors = []
		
		for vFeature in vlFeatures:
			vlFeatureValues = self.extractFeature(vFeature.name, vFeature.featureParams)
			vFVG.loadFeatureValues(vFeature.name, vlFeatureValues)
			
			if pflgRawVectors:
				if len(vlRawFVectors) == 0:
					vlRawFVectors = [{} for i in vlFeatureValues]
				for i in range(len(vlFeatureValues)):
					vlRawFVectors[i][vFeature.name] = vlFeatureValues[i]
		
		vlFVectors, vConfig = vFVG.genFeatureVectors()
		
		if pflgRawVectors:
			return vlFVectors, vConfig, vlRawFVectors
		else:
			return vlFVectors, vConfig
		
	
	
	## Feature extraction #####
	
	def extractFeature(self, pFeatureName, pdFeatureParams):
		'''
		Extracts values for pFeatureName 
		
		pFeatureName is the one specified in confiquration file. In order
		to be able to have multiple settings of a single feature type, an 
		entry for each setting is created in the configuration file with 
		the feature name suffixed by any unique string. The specific feature
		parameters then goes under each entry. The reason is that configuration
		file is in YAML format which is dictionary-like and requires unique 
		keys. So, the feature name can appear only once or only one of the
		appearances would be considered. For example, src-pos-ngram can
		be extracted for either original or universal POS tags but not 
		both at the same time since they both go under the same feature 
		name in the configuration file. To handle this, one can have two
		entries in the configuration and suffix the name in one of them 
		with any string (e.g. -universal). The specific setting then goes
		under each entry (unirversal tag map under src-pos-ngram-universal
		for this example).
		'''
		
		if pFeatureName.startswith("source-constituency-tree-root") or pFeatureName.startswith("src-const-root"):
			vlValues = self.extractSrcConstRoot()
		elif pFeatureName.startswith("translation-constituency-tree-root") or pFeatureName.startswith("trans-const-root"):
			vlValues = self.extractTransConstRoot()
		elif pFeatureName.startswith("source-constituency-tree-height") or pFeatureName.startswith("src-const-tree-height"):
			vlValues = self.extractSrcConstTreeHeight()
		elif pFeatureName.startswith("translation-constituency-tree-height") or pFeatureName.startswith("trans-const-tree-height"):
			vlValues = self.extractTransConstTreeHeight()
		elif pFeatureName.startswith("constituency-tree-height-relationship") or pFeatureName.startswith("const-tree-height-rel"):
			vlValues = self.extractRelConstTreeHeight(pdFeatureParams)
		elif pFeatureName.startswith("source-constituency-tree-size") or pFeatureName.startswith("src-const-tree-size"):
			vlValues = self.extractSrcConstTreeSize()
		elif pFeatureName.startswith("translation-constituency-tree-size") or pFeatureName.startswith("trans-const-tree-size"):
			vlValues = self.extractTransConstTreeSize()
		elif pFeatureName.startswith("constituency-tree-size-relationship") or pFeatureName.startswith("const-tree-size-rel"):
			vlValues = self.extractRelConstTreeSize(pdFeatureParams)
		elif pFeatureName.startswith("source-coarse-constituency-chunks") or pFeatureName.startswith("src-coarse-const-chunks"):
			vlValues = self.extractSrcCoarseConstChunks(pdFeatureParams)
		elif pFeatureName.startswith("translation-coarse-constituency-chunks") or pFeatureName.startswith("trans-coarse-const-chunks"):
			vlValues = self.extractTransCoarseConstChunks(pdFeatureParams)
		elif pFeatureName.startswith("source-const-tree-width") or pFeatureName.startswith("src-const-width") or \
			 pFeatureName.startswith("source-sentence-length") or pFeatureName.startswith("src-sent-length"):
			vlValues = self.extractSrcSentLength()
		elif pFeatureName.startswith("translation-const-tree-width") or pFeatureName.startswith("trans-const-width") or \
			 pFeatureName.startswith("translation-sentence-length") or pFeatureName.startswith("trans-sent-length"):
			vlValues = self.extractTransSentLength()
		elif pFeatureName.startswith("const-tree-width-relationship") or pFeatureName.startswith("const-width-rel") or \
			 pFeatureName.startswith("sentence-length-relationship") or pFeatureName.startswith("sent-length-rel"):
			vlValues = self.extractRelSentLength(pdFeatureParams)
		elif pFeatureName.startswith("source-parse-probability") or pFeatureName.startswith("src-parse-prob"):
			vlValues = self.extractSrcParseProb(pdFeatureParams)
		elif pFeatureName.startswith("translation-parse-probability") or pFeatureName.startswith("trans-parse-prob"):
			vlValues = self.extractTransParseProb(pdFeatureParams)
		elif pFeatureName.startswith("parse-probability-relationship") or pFeatureName.startswith("parse-prob-rel"):
			vlValues = self.extractRelParseProb(pdFeatureParams)
		elif pFeatureName.startswith("source-pos-ngrams") or pFeatureName.startswith("src-pos-ngrams"):
			vlValues = self.extractSrcPOSngrams(pdFeatureParams)
		elif pFeatureName.startswith("translation-pos-ngrams") or pFeatureName.startswith("trans-pos-ngrams"):
			vlValues = self.extractTransPOSngrams(pdFeatureParams)
		elif pFeatureName.startswith("source-pos-lm-score") or pFeatureName.startswith("src-pos-lm-score"):
			vlValues = self.extractSrcPOSLMScore(pdFeatureParams)
		elif pFeatureName.startswith("translation-pos-lm-score") or pFeatureName.startswith("trans-pos-lm-score"):
			vlValues = self.extractTransPOSLMScore(pdFeatureParams)
		elif pFeatureName.startswith("pos-lm-score-relationship") or pFeatureName.startswith("pos-lm-score-rel"):
			vlValues = self.extractRelPOSLMScore(pdFeatureParams)
		elif pFeatureName.startswith("source-specific-pos-count") or pFeatureName.startswith("src-pos-count"):
			vlValues = self.extractSrcPOSCount(pdFeatureParams)
		elif pFeatureName.startswith("translation-specific-pos-count") or pFeatureName.startswith("trans-pos-count"):
			vlValues = self.extractTransPOSCount(pdFeatureParams)
		elif pFeatureName.startswith("specific-pos-count-relationship") or pFeatureName.startswith("pos-count-rel"):
			vlValues = self.extractRelPOSCount(pdFeatureParams)
		elif pFeatureName.startswith("source-specific-phrase-tag-count") or pFeatureName.startswith("src-phrase-tag-count"):
			vlValues = self.extractSrcPhraseTagCount(pdFeatureParams)
		elif pFeatureName.startswith("translation-specific-phrase-tag-count") or pFeatureName.startswith("trans-phrase-tag-count"):
			vlValues = self.extractTransPhraseTagCount(pdFeatureParams)
		elif pFeatureName.startswith("specific-phrase-tag-count-rel") or pFeatureName.startswith("phrase-tag-count-rel"):
			vlValues = self.extractRelPhraseTagCount(pdFeatureParams)
		elif pFeatureName.startswith("source-specific-pos-location") or pFeatureName.startswith("src-pos-loc"):
			vlValues = self.extractSrcPOSLoc(pdFeatureParams)
		elif pFeatureName.startswith("translation-specific-pos-location") or pFeatureName.startswith("trans-pos-loc"):
			vlValues = self.extractTransPOSLoc(pdFeatureParams)
		elif pFeatureName.startswith("specific-pos-location-relationship") or pFeatureName.startswith("pos-loc-rel"):
			vlValues = self.extractRelPOSLoc(pdFeatureParams)
		elif pFeatureName.startswith("source-average-constituency-tree-arity") or pFeatureName.startswith("src-avg-const-arity"):
			vlValues = self.extractSrcAvgConstArity(pdFeatureParams)
		elif pFeatureName.startswith("translation-average-constituency-tree-arity") or pFeatureName.startswith("trans-avg-const-arity"):
			vlValues = self.extractTransAvgConstArity(pdFeatureParams)
		elif pFeatureName.startswith("average-constituency-tree-arity-relationship") or pFeatureName.startswith("avg-const-arity-rel"):
			vlValues = self.extractRelAvgConstArity(pdFeatureParams)
		elif pFeatureName.startswith("source-constituency-cfg-rules") or pFeatureName.startswith("src-const-rules"):
			vlValues = self.extractSrcConstRules(pdFeatureParams)
		elif pFeatureName.startswith("translation-constituency-cfg-rules") or pFeatureName.startswith("trans-const-rules"):
			vlValues = self.extractTransConstRules(pdFeatureParams)
		elif pFeatureName.startswith("source-constituency-pseudo-reference-score") or pFeatureName.startswith("src-const-pseudo-score"):
			vlValues = self.extractSrcConstPseudoScore(pdFeatureParams)
		elif pFeatureName.startswith("translation-constituency-pseudo-reference-score") or pFeatureName.startswith("trans-const-pseudo-score"):
			vlValues = self.extractTransConstPseudoScore(pdFeatureParams)
		elif pFeatureName.startswith("constituency-pseudo-reference-score-relationship") or pFeatureName.startswith("const-pseudo-score-rel"):
			vlValues = self.extractRelConstPseudoScore(pdFeatureParams)
		elif pFeatureName.startswith("source-dependency-top-node-pos-tag") or pFeatureName.startswith("src-dep-top-pos"):
			vlValues = self.extractSrcDepTopPOS(pdFeatureParams)
		elif pFeatureName.startswith("translation-dependency-top-node-pos-tag") or pFeatureName.startswith("trans-dep-top-pos"):
			vlValues = self.extractTransDepTopPOS(pdFeatureParams)
		elif pFeatureName.startswith("source-dependency-top-node-dependent-count") or pFeatureName.startswith("src-dep-top-dep-count"):
			vlValues = self.extractSrcDepTopDepCount(pdFeatureParams)
		elif pFeatureName.startswith("translation-dependency-top-node-dependent-count") or pFeatureName.startswith("trans-dep-top-dep-count"):
			vlValues = self.extractTransDepTopDepCount(pdFeatureParams)
		elif pFeatureName.startswith("dependency-top-node-dependent-count-relationship") or pFeatureName.startswith("dep-top-dep-count-rel"):
			vlValues = self.extractRelDepTopDepCount(pdFeatureParams)
		elif pFeatureName.startswith("source-dependency-top-node-dependency-relations-sequence") or pFeatureName.startswith("src-dep-top-dep-rels"):
			vlValues = self.extractSrcDepTopDepRels(pdFeatureParams)
		elif pFeatureName.startswith("translation-dependency-top-node-dependency-relations-sequence") or pFeatureName.startswith("trans-dep-top-dep-rels"):
			vlValues = self.extractTransDepTopDepRels(pdFeatureParams)
		elif pFeatureName.startswith("source-dependency-top-node-dependents-pos-tag-sequence") or pFeatureName.startswith("src-dep-top-deps-poss"):
			vlValues = self.extractSrcDepTopDepsPOSs(pdFeatureParams)
		elif pFeatureName.startswith("translation-dependency-top-node-dependents-pos-tag-sequence") or pFeatureName.startswith("trans-dep-top-deps-poss"):
			vlValues = self.extractTransDepTopDepsPOSs(pdFeatureParams)
		elif pFeatureName.startswith("source-dependency-average-dependent-count") or pFeatureName.startswith("src-dep-avg-dep-count"):
			vlValues = self.extractSrcDepAvgDepCount(pdFeatureParams)
		elif pFeatureName.startswith("translation-dependency-average-dependent-count") or pFeatureName.startswith("trans-dep-avg-dep-count"):
			vlValues = self.extractTransDepAvgDepCount(pdFeatureParams)
		elif pFeatureName.startswith("dependency-average-dependent-count-relationship") or pFeatureName.startswith("dep-avg-dep-count-rel"):
			vlValues = self.extractRelDepAvgDepCount(pdFeatureParams)
		elif pFeatureName.startswith("source-dependency-tree-height") or pFeatureName.startswith("src-dep-tree-height"):
			vlValues = self.extractSrcDepTreeHeight(pdFeatureParams)
		elif pFeatureName.startswith("translation-dependency-tree-height") or pFeatureName.startswith("trans-dep-tree-height"):
			vlValues = self.extractTransDepTreeHeight(pdFeatureParams)
		elif pFeatureName.startswith("dependency-tree-height-relationship") or pFeatureName.startswith("dep-tree-height-rel"):
			vlValues = self.extractRelDepTreeHeight(pdFeatureParams)
		elif pFeatureName.startswith("source-dependency-relation-ngrams") or pFeatureName.startswith("src-dep-rel-ngrams"):
			vlValues = self.extractSrcDepRelngrams(pdFeatureParams)
		elif pFeatureName.startswith("translation-dependency-relation-ngrams") or pFeatureName.startswith("trans-dep-rel-ngrams"):
			vlValues = self.extractTransDepRelngrams(pdFeatureParams)
		elif pFeatureName.startswith("source-specific-dependency-relation-count") or pFeatureName.startswith("src-dep-rel-count"):
			vlValues = self.extractSrcDepRelCount(pdFeatureParams)
		elif pFeatureName.startswith("translation-specific-dependency-relation-count") or pFeatureName.startswith("trans-dep-rel-count"):
			vlValues = self.extractTransDepRelCount(pdFeatureParams)
		elif pFeatureName.startswith("specific-dependency-relation-count-relationship") or pFeatureName.startswith("dep-rel-count-rel"):
			vlValues = self.extractRelDepRelCount(pdFeatureParams)
		elif pFeatureName.startswith("source-dependency-relation-lm-score") or pFeatureName.startswith("src-dep-rel-lm-score"):
			vlValues = self.extractSrcDepRelLMScore(pdFeatureParams)
		elif pFeatureName.startswith("translation-dependency-relation-lm-score") or pFeatureName.startswith("trans-dep-rel-lm-score"):
			vlValues = self.extractTransDepRelLMScore(pdFeatureParams)
		elif pFeatureName.startswith("dependency-relation-lm-score-relationship") or pFeatureName.startswith("dep-rel-lm-score-rel"):
			vlValues = self.extractRelDepRelLMScore(pdFeatureParams)
		elif pFeatureName.startswith("source-dependency-relation-ngram-quartile-frequency") or pFeatureName.startswith("src-dep-rel-ngram-quart-freq"):
			vlValues = self.extractSrcDepRelngramQFreq(pdFeatureParams)
		elif pFeatureName.startswith("translation-dependency-relation-ngram-quartile-frequency") or pFeatureName.startswith("trans-dep-rel-ngram-quart-freq"):
			vlValues = self.extractTransDepRelngramQFreq(pdFeatureParams)
		elif pFeatureName.startswith("dependency-relation-ngram-quartile-frequency-relationship") or pFeatureName.startswith("dep-rel-ngram-quart-freq-rel"):
			vlValues = self.extractRelDepRelngramQFreq(pdFeatureParams)
		elif pFeatureName.startswith("source-pos-tag-ngram-quartile-frequency") or pFeatureName.startswith("src-pos-ngram-quart-freq"):
			vlValues = self.extractSrcPOSngramQFreq(pdFeatureParams)
		elif pFeatureName.startswith("translation-pos-tag-ngram-quartile-frequency") or pFeatureName.startswith("trans-pos-ngram-quart-freq"):
			vlValues = self.extractTransPOSngramQFreq(pdFeatureParams)
		elif pFeatureName.startswith("pos-tag-ngram-quartile-frequency-relationship") or pFeatureName.startswith("pos-ngram-quart-freq-rel"):
			vlValues = self.extractRelPOSngramQFreq(pdFeatureParams)
		elif pFeatureName.startswith("source-dependency-relation-token-pairs") or pFeatureName.startswith("src-dep-rel-token-pairs"):
			vlValues = self.extractSrcDepRelTokenPairs(pdFeatureParams)
		elif pFeatureName.startswith("translation-dependency-relation-token-pairs") or pFeatureName.startswith("trans-dep-rel-token-pairs"):
			vlValues = self.extractTransDepRelTokenPairs(pdFeatureParams)
		elif pFeatureName.startswith("source-proposition-count") or pFeatureName.startswith("src-prop-count"):
			vlValues = self.extractSrcPropCount()
		elif pFeatureName.startswith("translation-proposition-count") or pFeatureName.startswith("trans-prop-count"):
			vlValues = self.extractTransPropCount()
		elif pFeatureName.startswith("proposition-count-relationship") or pFeatureName.startswith("prop-count-rel"):
			vlValues = self.extractRelPropCount(pdFeatureParams)
		elif pFeatureName.startswith("source-translation-common-predicate-count") or pFeatureName.startswith("common-pred-count"):
			vlValues = self.extractCommonPredCount(pdFeatureParams)
		elif pFeatureName.startswith("source-argument-count") or pFeatureName.startswith("src-arg-count"):
			vlValues = self.extractSrcArgCount()
		elif pFeatureName.startswith("translation-argument-count") or pFeatureName.startswith("trans-arg-count"):
			vlValues = self.extractTransArgCount()
		elif pFeatureName.startswith("argument-count-relationship") or pFeatureName.startswith("arg-count-rel"):
			vlValues = self.extractRelArgCount(pdFeatureParams)
		elif pFeatureName.startswith("source-average-argument-count") or pFeatureName.startswith("src-avg-arg-count"):
			vlValues = self.extractSrcAvgArgCount()
		elif pFeatureName.startswith("translation-average-argument-count") or pFeatureName.startswith("trans-avg-arg-count"):
			vlValues = self.extractTransAvgArgCount()
		elif pFeatureName.startswith("average-argument-count-relationship") or pFeatureName.startswith("avg-arg-count-rel"):
			vlValues = self.extractRelAvgArgCount(pdFeatureParams)
		elif pFeatureName.startswith("source-translation-common-role-count") or pFeatureName.startswith("common-role-count"):
			vlValues = self.extractCommonRoleCount(pdFeatureParams)
		elif pFeatureName.startswith("source-translation-common-proposition-structure-count") or pFeatureName.startswith("common-prop-struct-count"):
			vlValues = self.extractCommonPropStructCount(pdFeatureParams)
		elif pFeatureName.startswith("source-translation-common-predicate-score") or pFeatureName.startswith("common-pred-score"):
			vlValues = self.extractCommonPredScore(pdFeatureParams)
		elif pFeatureName.startswith("source-translation-common-role-score") or pFeatureName.startswith("common-role-score"):
			vlValues = self.extractCommonRoleScore(pdFeatureParams)
		elif pFeatureName.startswith("source-translation-common-proposition-structure-score") or pFeatureName.startswith("common-prop-struct-score"):
			vlValues = self.extractCommonPropStructScore(pdFeatureParams)
		elif pFeatureName.startswith("source-translation-common-predicate-role-score") or pFeatureName.startswith("common-pred-role-score"):
			vlValues = self.extractCommonPredRoleScore(pdFeatureParams)
		elif pFeatureName.startswith("source-argument-span-size-sum") or pFeatureName.startswith("src-arg-span-size-sum"):
			vlValues = self.extractSrcArgSpanSizeSum()
		elif pFeatureName.startswith("translation-argument-span-size-sum") or pFeatureName.startswith("trans-arg-span-size-sum"):
			vlValues = self.extractTransArgSpanSizeSum()
		elif pFeatureName.startswith("argument-span-size-sum-relationship") or pFeatureName.startswith("arg-span-size-sum-rel"):
			vlValues = self.extractRelArgSpanSizeSum(pdFeatureParams)
		elif pFeatureName.startswith("source-argument-span-size-ratio") or pFeatureName.startswith("src-arg-span-size-ratio"):
			vlValues = self.extractSrcArgSpanSizeSumRatio()
		elif pFeatureName.startswith("translation-argument-span-size-ratio") or pFeatureName.startswith("trans-arg-span-size-ratio"):
			vlValues = self.extractTransArgSpanSizeSumRatio()
		elif pFeatureName.startswith("argument-span-size-ratio-relationship") or pFeatureName.startswith("arg-span-size-ratio-rel"):
			vlValues = self.extractRelArgSpanSizeSumRatio(pdFeatureParams)
		elif pFeatureName.startswith("source-proposition-label-sequence") or pFeatureName.startswith("src-prop-label-seq"):
			vlValues = self.extractSrcPropLabelSeq(pdFeatureParams)
		elif pFeatureName.startswith("translation-proposition-label-sequence") or pFeatureName.startswith("trans-prop-label-seq"):
			vlValues = self.extractTransPropLabelSeq(pdFeatureParams)
		elif pFeatureName.startswith("source-proposition-syntax-tag-sequence") or pFeatureName.startswith("src-prop-syn-tag-seq"):
			vlValues = self.extractSrcPropSynTagSeq(pdFeatureParams)
		elif pFeatureName.startswith("translation-proposition-syntax-tag-sequence") or pFeatureName.startswith("trans-prop-syn-tag-seq"):
			vlValues = self.extractTransPropSynTagSeq(pdFeatureParams)
		elif pFeatureName.startswith("source-proposition-dependency-relation-sequence") or pFeatureName.startswith("src-prop-dep-rel-seq"):
			vlValues = self.extractSrcPropDepRelSeq(pdFeatureParams)
		elif pFeatureName.startswith("translation-proposition-dependency-relation-sequence") or pFeatureName.startswith("trans-prop-dep-rel-seq"):
			vlValues = self.extractTransPropDepRelSeq(pdFeatureParams)
		elif pFeatureName.startswith("source-proposition-pos-tag-sequence") or pFeatureName.startswith("src-prop-pos-tag-seq"):
			vlValues = self.extractSrcPropPOSTagSeq(pdFeatureParams)
		elif pFeatureName.startswith("translation-proposition-pos-tag-sequence") or pFeatureName.startswith("trans-prop-pos-tag-seq"):
			vlValues = self.extractTransPropPOSTagSeq(pdFeatureParams)
		elif pFeatureName.startswith("dependency-srl-word-alignment-projection-statistics") or pFeatureName.startswith("dep-srl-wa-proj-stat"):
			vlValues = self.extractDepSRLWAProjStats(pdFeatureParams)
		elif pFeatureName.startswith("predicate-argument-match-scores") or pFeatureName.startswith("pam"):
			vlValues = self.extractPAMScores(pdFeatureParams)
		elif pFeatureName.startswith("pre-extracted") or pFeatureName.startswith("preex"):
			vlValues = self.extractPreex(pdFeatureParams)
		else:
			raise Exception("Feature %s is unknown! Check the spelling." % pFeatureName)
		
		return vlValues
		
	
	
	# pre-extracted features
	
	def extractPreex(self, pdFeatureParams):
		'''
		Extracts a previously-extracted feature from a file
		
		The file contains a column of values for the feature.
		'''
		
		vlValues = []
		
		for vParam, vValue in pdFeatureParams.iteritems():
			if vParam.lower() in ["source", "file", "input"]:
				vlValues = [float(s) for s in open(vValue).read().strip().split('\n')]
			else:
				raise Exception("'%s' is not a valid feature parameter for pre-extracted features!" % vParam)
		
		return vlValues
		
		  
	
	
	# syntactic features
	
	def extractSrcConstRoot(self):
		'''
		Extracts root node tags of constituency trees of source side
		
		Feature name: source-constituency-tree-root or src-const-root
		'''
		
		return self.fesyqe.extractConstRoot(self.srcConstTrees)
		
	
	
	def extractTransConstRoot(self):
		'''
		Extracts root node tags of constituency trees of translation
		
		Feature name: translation-constituency-tree-root or trans-const-root
		'''
		
		return self.fesyqe.extractConstRoot(self.transConstTrees)
		
	
	
	def extractSrcConstTreeHeight(self):
		'''
		Extracts the height of source side constituency trees.
		
		Feature name: source-constituency-tree-height or src-const-tree-height
		'''
		
		return self.fesyqe.extractConstTreeHeight(self.srcConstTrees)
		
	
	
	def extractTransConstTreeHeight(self):
		'''
		Extracts the height of translation side constituency trees.
		
		Feature name: translation-constituency-tree-height or trans-const-tree-height
		'''
		
		return self.fesyqe.extractConstTreeHeight(self.transConstTrees)
		
	
	
	def extractRelConstTreeHeight(self, pdFeatureParams):
		'''
		Extracts the relationship between the source and translation 
		constituency tree heights.
		
		Relationship is given in feature parameters.
		
		Feature name: constituency-tree-height-relationship or const-tree-height-rel
		'''
		
		return self.fesyqe.extractRelConstTreeHeight(self.srcConstTrees, self.transConstTrees, pdFeatureParams)
		
	
	
	def extractSrcConstTreeSize(self):
		'''
		Extracts the size of source side constituency trees.
		
		Feature name: source-constituency-tree-size or src-const-tree-size
		'''
		
		return self.fesyqe.extractConstTreeSize(self.srcConstTrees)
		
	
	
	def extractTransConstTreeSize(self):
		'''
		Extracts the size of translation side constituency trees.
		
		Feature name: translation-constituency-tree-size or trans-const-tree-size
		'''
		
		return self.fesyqe.extractConstTreeSize(self.transConstTrees)
		
	
	
	def extractRelConstTreeSize(self, pdFeatureParams):
		'''
		Extracts the relationship between the source and translation 
		constituency tree sizes.
		
		Relationship is given in feature parameters.
		
		Feature name: constituency-tree-size-relationship or const-tree-size-rel
		'''
		
		return self.fesyqe.extractRelConstTreeSize(self.srcConstTrees, self.transConstTrees, pdFeatureParams)
		
	
	
	def extractSrcCoarseConstChunks(self, pdFeatureParams):
		'''
		Extracts coarse-grained chunks from source constituency trees.
		
		Feature name: source-coarse-constituency-chunks or src-coarse-const-chunks
		'''
		
		return self.fesyqe.extractCoarseConstChunks(self.srcConstTrees, pdFeatureParams)
		
	
	
	def extractTransCoarseConstChunks(self, pdFeatureParams):
		'''
		Extracts coarse-grained chunks from translation constituency trees.
		
		Feature name: translation-coarse-constituency-chunks or trans-coarse-const-chunks
		'''
		
		return self.fesyqe.extractCoarseConstChunks(self.transConstTrees, pdFeatureParams)
		
	
	
	def extractSrcSentLength(self):
		'''
		Extracts source sentence length (constituency tree width).
		
		Feature name: source-sentence-length, src-sent-length
		'''
		
		return self.fesyqe.extractSentLength(self.srcConstTrees)
		
	
	
	def extractTransSentLength(self):
		'''
		Extracts translation sentence length (constituency tree width).
		
		Feature name: translation-sentence-length, trans-sent-length
		'''
		
		return self.fesyqe.extractSentLength(self.transConstTrees)
		
	
	
	def extractRelSentLength(self, pdFeatureParams):
		'''
		Extracts the relationship between the source and translation 
		sentence length (constituency tree width).
		
		Relationship is given in feature parameters.
		
		Feature name: sentence-length-relationship or sent-length-rel
		'''
		
		return self.fesyqe.extractRelSentLength(self.srcConstTrees, self.transConstTrees, pdFeatureParams)
		
	
	
	def extractSrcParseProb(self, pdFeatureParams):
		'''
		Extracts the source side parse probabilities or log probabilities
		depending on what is specified in feature parameters.
		
		Feature name: source-parse-probability or src-parse-prob
		'''
		
		return self.fesyqe.extractParseProb(self.srcConstParses, pdFeatureParams)
		
	
	
	def extractTransParseProb(self, pdFeatureParams):
		'''
		Extracts the translation side parse probabilities or log probabilities
		depending on what is specified in feature parameters.
		
		Feature name: translation-parse-probability or trans-parse-prob
		'''
		
		return self.fesyqe.extractParseProb(self.transConstParses, pdFeatureParams)
		
	
	
	def extractRelParseProb(self, pdFeatureParams):
		'''
		Extracts the relationship between the source and translation 
		parse probabilities or log probabilities depending on what is 
		specified in feature parameters.
		
		Relationship is given in feature parameters.
		
		Feature name: parse-probability-relanship or parse-prob-rel
		'''
		
		return self.fesyqe.extractRelParseProb(self.srcConstParses, self.transConstParses, pdFeatureParams)
		
	
	
	def extractSrcPOSngrams(self, pdFeatureParams):
		'''
		Extracts POS n-grams from source side, where n is specified in 
		pdFeatureParams by order or n.
		
		POS tags can be mapped into a different tag set.
		
		Feature name: source-pos-ngrams or src-pos-ngrams
		'''
		
		return self.fesyqe.extractPOSngrams(self.srcConstTrees, pdFeatureParams)
		
	
	
	def extractTransPOSngrams(self, pdFeatureParams):
		'''
		Extracts POS n-grams from translation side, where n is specified
		in pdFeatureParams by order or n.
		
		POS tags can be mapped into a different tag set.
		
		Feature name: translation-pos-ngrams or trans-pos-ngrams
		'''
		
		return self.fesyqe.extractPOSngrams(self.transConstTrees, pdFeatureParams)
		
	
	
	def extractSrcPOSLMScore(self, pdFeatureParams):
		'''
		Extracts POS language model scores (prob, log prob or perplexity)
		of the source side from the output of the LM tool provided in feature 
		parameters.
		
		SRILM output is supported.
		
		Feature name: source-pos-lm-score or src-pos-lm-score
		'''
		
		return self.fesyqe.extractPOSLMScore(pdFeatureParams)
		
	
	
	def extractTransPOSLMScore(self, pdFeatureParams):
		'''
		Extracts POS language model scores (prob, log prob or perplexity)
		of the translation side from the output of the LM tool provided in
		feature parameters.
		
		SRILM output is supported.
		
		Feature name: translation-pos-lm-score or trans-pos-lm-score
		'''
		
		return self.fesyqe.extractPOSLMScore(pdFeatureParams)
		
	
	
	def extractRelPOSLMScore(self, pdFeatureParams):
		'''
		Extracts the relationship between the source and translation POS
		language model scores (prob, log prob or perplexity).
		
		Relationship is given in feature parameters.
		
		Feature name: pos-lm-score-relationship or pos-lm-score-rel
		'''
		
		return self.fesyqe.extractRelPOSLMScore(pdFeatureParams)
		
	
	
	def extractSrcPOSCount(self, pdFeatureParams):
		'''
		Extracts the number of times POS tags in a given set appear in 
		the source side trees.
		
		The set of POS tags is given in the feature parameters.
		
		POS tags can be mapped into a different tag set.
		
		Feature name: source-specific-pos-count or src-pos-count
		'''
		
		return self.fesyqe.extractPOSCount(self.srcConstTrees, pdFeatureParams)
		
	
	
	def extractTransPOSCount(self, pdFeatureParams):
		'''
		Extracts the number of times POS tags in a given set appear in 
		the translation side trees.
		
		The set of POS tags is given in the feature parameters.
		
		POS tags can be mapped into a different tag set.
		
		Feature name: translation-specific-pos-count or trans-pos-count
		'''
		
		return self.fesyqe.extractPOSCount(self.transConstTrees, pdFeatureParams)
		
	
	
	def extractRelPOSCount(self, pdFeatureParams):
		'''
		Extracts the relationship between the number of times POS tags 
		in a given set appear in the source and translation side.
		
		The set of POS tags is given in the feature parameters.
		
		POS tags can be mapped into a different tag set.
		
		Relationship is given in feature parameters.
		
		Feature name: pos-count-relationship or pos-count-rel
		'''
		
		return self.fesyqe.extractRelPOSCount(self.srcConstTrees, self.transConstTrees, pdFeatureParams)
		
	
	
	def extractSrcPhraseTagCount(self, pdFeatureParams):
		'''
		Extracts the number of times phrase tags in a given set appear in 
		the source side trees.
		
		The set of phrase tags is given in the feature parameters.
		
		Phrase tags can be mapped into a different tag set.
		
		Feature name: source-specific-phrase-tag-count or src-phrase-tag-count
		'''
		
		return self.fesyqe.extractPhraseTagCount(self.srcConstTrees, pdFeatureParams)
		
	
	
	def extractTransPhraseTagCount(self, pdFeatureParams):
		'''
		Extracts the number of times phrase tags in a given set appear in 
		the translation side trees.
		
		The set of phrase tags is given in the feature parameters.
		
		Phrase tags can be mapped into a different tag set.
		
		Feature name: translation-specific-phrase-tag-count or trans-phrase-tag-count
		'''
		
		return self.fesyqe.extractPhraseTagCount(self.transConstTrees, pdFeatureParams)
		
	
	
	def extractRelPhraseTagCount(self, pdFeatureParams):
		'''
		Extracts the relationship (e.g. ratio) between the number of times
		phrase tags in a given set appear in the source and translation 
		side, where the relationship is specified in feature parameters.
		
		The set of phrase tags is given in the feature parameters.
		
		Phrase tags tags can be mapped into a different tag set.
		
		Relationship is given in feature parameters.
		
		Feature name: specific-phrase-tag-count-rel or phrase-tag-count-rel
		'''
		
		return self.fesyqe.extractRelPhraseTagCount(self.srcConstTrees, self.transConstTrees, pdFeatureParams)
		
	
	
	def extractSrcPOSLoc(self, pdFeatureParams):
		'''
		Extracts the location of the first occurrence of a specific POS 
		tag from the beginning of the source sentence.
		
		The POS tag is given in the feature parameters.
		
		POS tags can be mapped into a different tag set.
		
		Feature name: source-specific-pos-location or src-pos-loc
		'''
		
		return self.fesyqe.extractPOSLoc(self.srcConstTrees, pdFeatureParams)
		
	
	
	def extractTransPOSLoc(self, pdFeatureParams):
		'''
		Extracts the location of the first occurrence of a specific POS 
		tag from the beginning of the source sentence.
		
		The POS tag is given in the feature parameters.
		
		POS tags can be mapped into a different tag set.
		
		Feature name: translation-specific-pos-location or trans-pos-loc
		'''
		
		return self.fesyqe.extractPOSLoc(self.transConstTrees, pdFeatureParams)
		
	
	
	def extractRelPOSLoc(self, pdFeatureParams):
		'''
		Extracts the relatioship between the location of the first occurrence
		of a specific POS tag from the beginning of the source sentence 
		and from the beginning of the translation sentence.
		
		The POS tag is given in the feature parameters.
		
		POS tags can be mapped into a different tag set.
		
		Feature name: specific-pos-location-relationship or pos-loc-rel
		'''
		
		return self.fesyqe.extractRelPOSLoc(self.srcConstTrees, self.transConstTrees, pdFeatureParams)
		
	
	
	def extractSrcAvgConstArity(self, pdFeatureParams):
		'''
		Extracts the average arity of the production rules of the nodes
		of the source constituency tree.
		
		The arity is in fact the number of children of that node. The 
		terminal nodes are not counted in averaging.
		
		Feature name: source-average-constituency-tree-arity or src-avg-const-arity
		'''
		
		return self.fesyqe.extractAvgConstArity(self.srcConstTrees, pdFeatureParams)
		
	
	
	def extractTransAvgConstArity(self, pdFeatureParams):
		'''
		Extracts the average arity of the production rules of the nodes
		of the translation constituency tree.
		
		The arity is in fact the number of children of that node. The 
		terminal nodes are not counted in averaging.
		
		Feature name: translation-average-constituency-tree-arity or 
		              trans-avg-const-arity
		'''
		
		return self.fesyqe.extractAvgConstArity(self.transConstTrees, pdFeatureParams)
		
	
	
	def extractRelAvgConstArity(self, pdFeatureParams):
		'''
		Extracts the relationship between the average arity of the production
		rules of the nodes of the source and of the translation constituency
		trees.
		
		Relationship is given in feature parameters.
		
		Feature name: average-constituency-tree-arity-relationship or 
		              avg-const-arity-rel
		'''
		
		return self.fesyqe.extractRelAvgConstArity(self.srcConstTrees, self.transConstTrees, pdFeatureParams)
		
	
	
	def extractSrcConstRules(self, pdFeatureParams):
		'''
		Extracts the CFG production rules expanding each constituency tree
		node of the source side.
		
		Certain tags can be ignored and excluded from the rules.
		
		Lexical, non-lexical or both types of rules can be extracted.
		
		Tags can be mapped into a different tag set.
		
		Feature name: source-constituency-cfg-rules or src-const-rules
		'''
		
		return self.fesyqe.extractConstRules(self.srcConstTrees, pdFeatureParams)
		
	
	
	def extractTransConstRules(self, pdFeatureParams):
		'''
		Extracts the CFG production rules expanding each constituency tree
		node of the translation side. 
		
		Certain tags can be ignored and excluded from the rules.
		
		Lexical, non-lexical or both types of rules can be extracted.
		
		Tags can be mapped into a different tag set.
		
		Feature name: translation-constituency-cfg-rules or trans-const-rules
		'''
		
		return self.fesyqe.extractConstRules(self.transConstTrees, pdFeatureParams)
		
	
	
	def extractSrcConstPseudoScore(self, pdFeatureParams):
		'''
		Extracts scores of the constituency parses of the source side 
		evaluated against the output of a pseudo-reference parser. 
		
		Scores are in a file (one per line) given in feature parameters.
		
		Feature name: source-constituency-pseudo-reference-score or 
		              src-const-pseudo-score
		'''
		
		return self.fesyqe.extractConstPseudoScore(pdFeatureParams)
		
	
	
	def extractTransConstPseudoScore(self, pdFeatureParams):
		'''
		Extracts scores of the constituency parses of the translation side 
		evaluated against the output of a pseudo-reference parser. 
		
		Scores are in a file (one per line) given in feature parameters.
		
		Feature name: translation-constituency-pseudo-reference-score or 
		              trans-const-pseudo-score
		'''
		
		return self.fesyqe.extractConstPseudoScore(pdFeatureParams)
		
	
	
	def extractRelConstPseudoScore(self, pdFeatureParams):
		'''
		Extracts the relationship bescores of the constituency parses of the translation side 
		evaluated against the output of a pseudo-reference parser. 
		
		Scores are in files (one per line) given in feature parameters.
		
		Relationship is given in feature parameters.
		
		Feature name: constituency-pseudo-reference-score-relationship or 
		              const-pseudo-score-rel
		'''
		
		return self.fesyqe.extractRelConstPseudoScore(pdFeatureParams)
		
	
	
	def extractSrcDepTopPOS(self, pdFeatureParams):
		'''
		Extracts the POS tag of the top node (the dependent of the dummy 
		root node) of the dependency trees of the source side.
		
		Tags can be mapped into a different tag set.
		
		Feature name: source-dependency-root-node-pos-tag or src-dep-root-pos
		'''
		
		return self.fesyqe.extractDepTopPOS(self.srcDepTrees, pdFeatureParams)
		
	
	
	def extractTransDepTopPOS(self, pdFeatureParams):
		'''
		Extracts the POS tag of the root node (the dependent of the dummy 
		root node) of the dependency trees of the translation side.
		
		Tags can be mapped into a different tag set.
		
		Feature name: translation-dependency-root-node-pos-tag or trans-dep-root-pos
		'''
		
		return self.fesyqe.extractDepTopPOS(self.transDepTrees, pdFeatureParams)
		
	
	
	def extractSrcDepTopDepCount(self, pdFeatureParams):
		'''
		Extracts number of dependents of the top node (the dependent of 
		the dummy root node) of the dependency trees of the source side.
		
		Feature name: source-dependency-top-node-dependent-count or
		              src-dep-top-dep-count
		'''
		
		return self.fesyqe.extractDepTopDepCount(self.srcDepTrees, pdFeatureParams)
		
	
	
	def extractTransDepTopDepCount(self, pdFeatureParams):
		'''
		Extracts number of dependents of the top node (the dependent of 
		the dummy root node) of the dependency trees of the translation
		side.
		
		Feature name: translation-dependency-top-node-dependent-count or
		              trans-dep-top-dep-count
		'''
		
		return self.fesyqe.extractDepTopDepCount(self.transDepTrees, pdFeatureParams)
		
	
	
	def extractRelDepTopDepCount(self, pdFeatureParams):
		'''
		Extracts the relatinoship between the number of dependents of the
		top node (the dependent of the dummy root node) of the dependency
		trees of the source and translation side.
		
		Relationship is given in feature parameters.
		
		Feature name: dependency-top-node-dependent-count-relationship or
		              dep-top-dep-count-rel
		'''
		
		return self.fesyqe.extractRelDepTopDepCount(self.srcDepTrees, self.transDepTrees, pdFeatureParams)
		
	
	
	def extractSrcDepTopDepRels(self, pdFeatureParams):
		'''
		Extracts the sequence of dependency relations of top nodes of the
		source side trees to their dependents.
		
		Feature name: source-dependency-top-node-dependency-relations-sequence
		             or src-dep-top-dep-rels
		'''
		
		return self.fesyqe.extractDepTopDepRels(self.srcDepTrees, pdFeatureParams)
		
	
	
	def extractTransDepTopDepRels(self, pdFeatureParams):
		'''
		Extracts the sequence of dependency relations of top nodes of the
		translation side trees to their dependents.
		
		Feature name: translation-dependency-top-node-dependency-relations-sequence
		              or trans-dep-top-dep-rels
		'''
		
		return self.fesyqe.extractDepTopDepRels(self.transDepTrees, pdFeatureParams)
		
	
	
	def extractSrcDepTopDepsPOSs(self, pdFeatureParams):
		'''
		Extracts the sequence of POS tags of the dependents of the top
		nodes of the source side dependency trees.
		
		Tags can be mapped into a different tag set.
		
		Feature name: source-dependency-top-node-dependents-pos-tag-sequence
		              or src-dep-top-deps-poss
		'''
		
		return self.fesyqe.extractDepTopDepsPOSs(self.srcDepTrees, pdFeatureParams)
		
	
	
	def extractTransDepTopDepsPOSs(self, pdFeatureParams):
		'''
		Extracts the sequence of POS tags of the dependents of the top
		nodes of the source side dependency trees.
		
		Tags can be mapped into a different tag set.
		
		Feature name: translation-dependency-top-node-dependents-pos-tag-sequence
		              or pFeatureName.startswith("trans-dep-top-deps-poss
		'''
		
		return self.fesyqe.extractDepTopDepsPOSs(self.transDepTrees, pdFeatureParams)
		
	
	
	def extractSrcDepAvgDepCount(self, pdFeatureParams):
		'''
		Extracts average number of dependents of the source side dependency
		trees.
		
		Feature name: source-dependency-average-dependent-count or
		              src-dep-avg-dep-count
		'''
		
		return self.fesyqe.extractDepAvgDepCount(self.srcDepTrees, pdFeatureParams)
		
	
	
	def extractTransDepAvgDepCount(self, pdFeatureParams):
		'''
		Extracts average number of dependents of the translation side 
		dependency trees.
		
		Feature name: translation-dependency-average-dependent-count or
		              trans-dep-avg-dep-count
		'''
		
		return self.fesyqe.extractDepAvgDepCount(self.transDepTrees, pdFeatureParams)
		
	
	
	def extractRelDepAvgDepCount(self, pdFeatureParams):
		'''
		Extracts the relationship between the average number of dependents
		in the source and the translation side dependency trees.
		
		Relationship is given in feature parameters.
		
		Feature name: dependency-average-dependent-count-relationship or
		              dep-avg-dep-count-rel
		'''
		
		return self.fesyqe.extractRelDepAvgDepCount(self.srcDepTrees, self.transDepTrees, pdFeatureParams)
		
	
	
	def extractSrcDepTreeHeight(self, pdFeatureParams):
		'''
		Extracts the height of the source side dependency tree.
		
		Feature name: source-dependency-tree-height, src-dep-tree-height
		'''
		
		return self.fesyqe.extractDepTreeHeight(self.srcDepTrees, pdFeatureParams)
		
	
	
	def extractTransDepTreeHeight(self, pdFeatureParams):
		'''
		Extracts the height of the translation side dependency tree.
		
		Feature name: translation-dependency-tree-height, trans-dep-tree-height
		'''
		
		return self.fesyqe.extractDepTreeHeight(self.transDepTrees, pdFeatureParams)
		
	
	
	def extractRelDepTreeHeight(self, pdFeatureParams):
		'''
		Extracts the relationship between the heights of the source and
		the translation dependency trees.
		
		Feature name: dependency-tree-height-rel, dep-tree-height-rel
		'''
		
		return self.fesyqe.extractRelDepTreeHeight(self.srcDepTrees, self.transDepTrees, pdFeatureParams)
		
	
	
	def extractSrcDepRelngrams(self, pdFeatureParams):
		'''
		Extracts dependency relation n-grams from the source side dependency 
		trees.
		
		Order is of n-gram is given in feature parameters.
		
		Feature name: source-dependency-relation-ngrams or src-dep-rel-ngrams
		'''
		
		return self.fesyqe.extractDepRelngrams(self.srcDepTrees, pdFeatureParams)
		
	
	
	def extractTransDepRelngrams(self, pdFeatureParams):
		'''
		Extracts dependency relation n-grams from the translation side 
		dependency trees.
		
		Order is of n-gram is given in feature parameters.
		
		Feature name: source-dependency-relation-ngrams or src-dep-rel-ngrams
		'''
		
		return self.fesyqe.extractDepRelngrams(self.transDepTrees, pdFeatureParams)
		
	
	
	def extractSrcDepRelCount(self, pdFeatureParams):
		'''
		Extracts the number of times dependency relations in a given set
		appear in the source side trees.
		
		The set of dependency relations is given in the feature parameters.
		
		Dependency relation tags can be mapped into a different tag set.
		
		Feature name: source-specific-dependency-relation-count or 
		              src-dep-rel-count
		'''
		
		return self.fesyqe.extractDepRelCount(self.srcDepTrees, pdFeatureParams)
		
	
	
	def extractTransDepRelCount(self, pdFeatureParams):
		'''
		Extracts the number of times dependency relations in a given set
		appear in the translation side trees.
		
		The set of dependency relations is given in the feature parameters.
		
		Dependency relation tags can be mapped into a different tag set.
		
		Feature name: transaltion-specific-dependency-relation-count or 
		              trans-dep-rel-count
		'''
		
		return self.fesyqe.extractDepRelCount(self.transDepTrees, pdFeatureParams)
		
	
	
	def extractRelDepRelCount(self, pdFeatureParams):
		'''
		Extracts the relationship between the number of times dependency
		relations in a given set appear in the source side trees and the 
		number of times dependency relations in another given set (possibly 
		the same) appear in the translation side trees.
		
		The set of dependency relations is given in the feature parameters.
		
		Dependency relation tags can be mapped into a different tag set.
		
		Feature name: specific-dependency-relation-count-relationship or 
		              dep-rel-count-rel
		'''
		
		return self.fesyqe.extractRelDepRelCount(self.srcDepTrees, self.transDepTrees, pdFeatureParams)
		
	
	
	def extractSrcDepRelLMScore(self, pdFeatureParams):
		'''
		Extracts dependency relation language model scores (prob, log prob
		or perplexity) of the source side from the output of the LM tool
		provided in feature parameters.
		
		SRILM output is supported.
		
		Dependency relation tags can be mapped into a different tag set.
		
		Feature name: source-dependency-relation-lm-score or 
		              src-dep-rel-lm-score
		'''
		
		return self.fesyqe.extractDepRelLMScore(pdFeatureParams)
		
	
	
	def extractTransDepRelLMScore(self, pdFeatureParams):
		'''
		Extracts dependency relation language model scores (prob, log prob
		or perplexity) of the translation side from the output of the LM
		tool provided in feature parameters.
		
		SRILM output is supported.
		
		Feature name: translation-dependency-relation-lm-score or 
		              trans-dep-rel-lm-score
		'''
		
		return self.fesyqe.extractDepRelLMScore(pdFeatureParams)
		
	
	
	def extractRelDepRelLMScore(self, pdFeatureParams):
		'''
		Extracts the relationship between the LM scores (prob, log prob
		or perplexity) of dependency relations of the source and translation 
		sides from the output of the LM tool provided in feature parameters.
		
		SRILM output is supported.
		
		Feature name: dependency-relation-lm-score-relationship or 
		              dep-rel-lm-score-rel
		'''
		
		return self.fesyqe.extractRelDepRelLMScore(pdFeatureParams)
		
	
	
	def extractSrcDepRelngramQFreq(self, pdFeatureParams):
		'''
		Extracts the dependency relation n-gram frequency of the source 
		side n-grams (average number of n-grams) in n-gram frequency 
		quarters of a corpus.
		
		The quarter and n-grams counts of the corpus are given in feature 
		parameters.
		
		N-gram counts should be in the format output by SRILM.
		
		Feature name: source-dependency-relation-ngram-quartile-frequency
		              or src-dep-rel-ngram-quart-freq
		'''
		
		return self.fesyqe.extractDepRelngramQFreq(self.srcDepTrees, pdFeatureParams)
		
	
	
	def extractTransDepRelngramQFreq(self, pdFeatureParams):
		'''
		Extracts the dependency relation n-gram frequency of the translation 
		side n-grams (average number of n-grams) in n-gram frequency 
		quarters of a corpus.
		
		The quarter and n-grams counts of the corpus are given in feature 
		parameters.
		
		N-gram counts should be in the format output by SRILM.
		
		Feature name: translation-dependency-relation-ngram-quartile-frequency
		              or trans-dep-rel-ngram-quart-freq
		'''
		
		return self.fesyqe.extractDepRelngramQFreq(self.transDepTrees, pdFeatureParams)
		
	
	
	def extractRelDepRelngramQFreq(self, pdFeatureParams):
		'''
		Extracts the relationship between the dependency relation n-gram 
		frequency of the source side in n-gram frequency quarters of a 
		corpus with those of the translation side.
		
		The quarter and n-grams counts of the corpus are given in feature 
		parameters.
		
		N-gram counts should be in the format output by SRILM.
		
		Feature name: dependency-relation-ngram-quartile-frequency-relationship
		              or dep-rel-ngram-quart-freq-rel
		'''
		
		return self.fesyqe.extractRelDepRelngramQFreq(self.srcDepTrees, self.transDepTrees, pdFeatureParams)
		
	
	
	def extractSrcPOSngramQFreq(self, pdFeatureParams):
		'''
		Extracts the POS n-gram frequency of the source side n-grams (average
		number of n-grams) in n-gram frequency  quarters of a corpus.
		
		The quarter and n-grams counts of the corpus are given in feature 
		parameters.
		
		N-gram counts should be in the format output by SRILM.
		
		POS tags can be mapped into a different tag set.
		
		Feature name: source-pos-tag-ngram-quartile-frequency
		              or src-pos-ngram-quart-freq
		'''
		
		return self.fesyqe.extractPOSngramQFreq(self.srcConstTrees, pdFeatureParams)
		
	
	
	def extractTransPOSngramQFreq(self, pdFeatureParams):
		'''
		Extracts the POS n-gram frequency of the translation side n-grams
		(average number of n-grams) in n-gram frequency  quarters of a corpus.
		
		The quarter and n-grams counts of the corpus are given in feature 
		parameters.
		
		N-gram counts should be in the format output by SRILM.
		
		POS tags can be mapped into a different tag set.
		
		Feature name: translation-pos-tag-ngram-quartile-frequency
		              or trans-pos-ngram-quart-freq
		'''
		
		return self.fesyqe.extractPOSngramQFreq(self.transConstTrees, pdFeatureParams)
		
	
	
	def extractRelPOSngramQFreq(self, pdFeatureParams):
		'''
		Extracts the relationship between the POS n-gram frequency of the
		source side in n-gram frequency quarters of a corpus with those 
		of the translation side.
		
		The quarter and n-grams counts of the corpus are given in feature 
		parameters.
		
		N-gram counts should be in the format output by SRILM.
		
		POS tags can be mapped into a different tag set.
		
		Feature name: pos-tag-ngram-quartile-frequency-relationship
		              or pos-ngram-quart-freq-rel
		'''
		
		return self.fesyqe.extractRelPOSngramQFreq(self.srcConstTrees, self.transConstTrees, pdFeatureParams)
		
	
	
	def extractSrcDepRelTokenPairs(self, pdFeatureParams):
		'''
		Extracts the dependency relation-token pairs from source dependency
		trees.
		
		This feature resembles lexical rules of constituency features.
		
		Feature name: source-dependency-relation-token-pairs
		              or src-dep-rel-token-pairs
		'''
		
		return self.fesyqe.extractDepRelTokenPairs(self.srcDepTrees, pdFeatureParams)
		
	
	
	def extractTransDepRelTokenPairs(self, pdFeatureParams):
		'''
		Extracts the dependency relation-token pairs from translation 
		dependency trees.
		
		This feature resembles lexical rules of constituency features.
		
		Feature name: translation-dependency-relation-token-pairs
		              or trans-dep-rel-token-pairs
		'''
		
		return self.fesyqe.extractDepRelTokenPairs(self.transDepTrees, pdFeatureParams)
		
	
	
	# semantic features
	
	def extractSrcPropCount(self):
		'''
		Extracts source side proposition count
		'''
		
		return self.feseqe.extractPropCount(self.srcDepSRLs)
		
	
	
	def extractTransPropCount(self):
		'''
		Extracts translations side proposition count
		'''
		
		return self.feseqe.extractPropCount(self.transDepSRLs)
		
	
	
	def extractRelPropCount(self, pdFeatureParams):
		'''
		Extracts the relationship between proposition count of the source
		and translation sides.
		'''
		
		return self.feseqe.extractRelPropCount(self.srcDepSRLs, self.transDepSRLs, pdFeatureParams)
		
	
	
	def extractCommonPredCount(self, pdFeatureParams):
		'''
		Extracts the number of common predicates in the source and translation
		SRLs
		
		The "level" parameter can be set to consider either frameset or 
		lemma. Frameset is the default.
		'''
		
		return self.feseqe.extractCommonPredCount(self.srcDepSRLs, self.transDepSRLs, pdFeatureParams)
		
	
	
	def extractSrcArgCount(self):
		'''
		Extracts source side argument count
		'''
		
		return self.feseqe.extractArgCount(self.srcDepSRLs)
		
	
	
	def extractTransArgCount(self):
		'''
		Extracts translation side proposition count
		'''
		
		return self.feseqe.extractArgCount(self.transDepSRLs)
		
	
	
	def extractRelArgCount(self, pdFeatureParams):
		'''
		Extracts the relationship between argument counts of the source
		and translation sides.
		'''
		
		return self.feseqe.extractRelArgCount(self.srcDepSRLs, self.transDepSRLs, pdFeatureParams)
		
	
	
	def extractSrcAvgArgCount(self):
		'''
		Extracts average number of arguments per proposition in the source
		side
		'''
		
		return self.feseqe.extractAvgArgCount(self.srcDepSRLs)
		
	
	
	def extractTransAvgArgCount(self):
		'''
		Extracts average number of arguments per proposition in the translation
		side
		'''
		
		return self.feseqe.extractAvgArgCount(self.transDepSRLs)
		
	
	
	def extractRelAvgArgCount(self, pdFeatureParams):
		'''
		Extracts the relationship between average argument counts of the
		source and translation sides.
		'''
		
		return self.feseqe.extractRelAvgArgCount(self.srcDepSRLs, self.transDepSRLs, pdFeatureParams)
		
	
	
	def extractCommonRoleCount(self, pdFeatureParams):
		'''
		Extracts the number of common argument roles between the matching
		predicates of the source and translation SRLs
		
		Note that if a predicate appears more than once in the sentence,
		the matching is based on the order in the sentence since we don't 
		use word alignment here.
		
		The parameter "pred-match-level" is the same as the parameter 
		"level" of common-pred-count feature (see extractCommonPredCount).
		
		If "collapse-adjunct" parameter is given and set to true, the
		adjunct argument labels will be collapsed to a single label.
		'''
		
		return self.feseqe.extractCommonRoleCount(self.srcDepSRLs, self.transDepSRLs, pdFeatureParams)
		
	
	
	def extractCommonPropStructCount(self, pdFeatureParams):
		'''
		Extracts the number of common proposition structures in the source
		and translation SRLs
		
		Proposition structure consists of all of its argument roles. Two
		propositions are compared for being common only if their predicates
		match according to criteria defined (i.e. frameset or lemma).
		
		The algorithm relies on the common predicate and argument extraction
		algorithm. Therefore, the parameter set is the same as in common 
		argument.
		'''
		
		return self.feseqe.extractCommonPropStructCount(self.srcDepSRLs, self.transDepSRLs, pdFeatureParams)
		
	
	
	def extractCommonPredScore(self, pdFeatureParams):
		'''
		Extracts the score of common predicates in the source and translation
		SRLs
		
		The score can be precision, recall, or F1 identified by "score" 
		parameter.
		
		The feature is similar to common predicate count but with the 
		number of predicate in the source and/or translation taken into
		account by dividing the former by either of the latter (recall
		and precision respectively) or their harmonic mean (F1).
		
		The level at which the predicates are compared is given in "level"
		parameter.
		'''
		
		return self.feseqe.extractCommonPredScore(self.srcDepSRLs, self.transDepSRLs, pdFeatureParams)
		
	
	
	def extractCommonRoleScore(self, pdFeatureParams):
		'''
		Extracts the score of common argument roles between the matching
		predicates of the source and translation SRLs
		
		The score can be precision, recall, or F1 identified by "score" 
		parameter.
		
		The feature is similar to common argument role count but with the 
		number of arguemnts in the source and/or translation taken into
		account by dividing the former by either of the latter (recall
		and precision respectively) or their harmonic mean (F1).
		
		Predicate match level and whether adjuncts should be collapsed 
		or not is given as parameter.
		'''
		
		return self.feseqe.extractCommonRoleScore(self.srcDepSRLs, self.transDepSRLs, pdFeatureParams)
		
	
	
	def extractCommonPropStructScore(self, pdFeatureParams):
		'''
		Extracts the score of common proposition structures in the source
		and translation SRLs
		
		Proposition structure consists of all of its argument roles. Two
		propositions are compared for being common only if their predicates
		match according to criteria defined (i.e. frameset or lemma).
		
		The algorithm relies on the common predicate and argument extraction
		algorithms. Therefore, the parameter set is the same as in common 
		argument score.
		'''
		
		return self.feseqe.extractCommonPropStructScore(self.srcDepSRLs, self.transDepSRLs, pdFeatureParams)
		
	
	
	def extractCommonPredRoleScore(self, pdFeatureParams):
		'''
		Extracts the score of common predicates and argument roles between
		the matching predicates of the source and translation SRLs
		
		The score can be precision, recall, or F1 identified by "score" 
		parameter.
		
		The feature is the combination of common predicate score and common
		argument role score. This is the way SRL is scored in CoNLL 2009
		shared task, though without lexical match. Therefore, similar to
		SRl scores, predicate label and argument role are counted the same 
		in scoring.
		'''
		
		return self.feseqe.extractCommonPredRoleScore(self.srcDepSRLs, self.transDepSRLs, pdFeatureParams)
		
	
	
	def extractSrcArgSpanSizeSum(self):
		'''
		Extracts the sum of span sizes of arguments in the source side
		'''
		
		return self.feseqe.extractArgSpanSizeSum(self.srcConstSRLs)
		
	
	
	def extractTransArgSpanSizeSum(self):
		'''
		Extracts the sum of span sizes of arguments in the translation side
		'''
		
		return self.feseqe.extractArgSpanSizeSum(self.transConstSRLs)
		
	
	
	def extractRelArgSpanSizeSum(self, pdFeatureParams):
		'''
		Extracts the relationship between the sums of span sizes of arguments
		in the source and translation sides.
		'''
		
		return self.feseqe.extractRelArgSpanSizeSum(self.srcConstSRLs, self.transConstSRLs, pdFeatureParams)
		
	
	
	def extractSrcArgSpanSizeSumRatio(self):
		'''
		Extracts the ratio of sum of span sizes of arguments in the source
		sentence to its length
		'''
		
		return self.feseqe.extractArgSpanSizeSumRatio(self.srcConstSRLs, self.srcSentSurfaces)
		
	
	
	def extractTransArgSpanSizeSumRatio(self):
		'''
		Extracts the ratio of sum of span sizes of arguments in the translation
		sentence to its length
		'''
		
		return self.feseqe.extractArgSpanSizeSumRatio(self.transConstSRLs, self.transSentSurfaces)
		
	
	
	def extractRelArgSpanSizeSumRatio(self, pdFeatureParams):
		'''
		Extracts the relationship between the ratio of sums of span sizes
		of arguments to sentence lengths in the source and translation sides
		'''
		
		return self.feseqe.extractRelArgSpanSizeSumRatio(self.srcConstSRLs, self.transConstSRLs, self.srcSentSurfaces, self.transSentSurfaces, pdFeatureParams)
		
	
	
	def extractSrcPropLabelSeq(self, pdFeatureParams):
		'''
		Extracts the predicate/argument label sequences of source side 
		propositions
		'''
		
		return self.feseqe.extractPropLabelSeq(self.srcConstSRLs, pdFeatureParams)
		
	
	
	def extractTransPropLabelSeq(self, pdFeatureParams):
		'''
		Extracts the predicate/argument label sequences of translation 
		side propositions
		'''
		
		return self.feseqe.extractPropLabelSeq(self.transConstSRLs, pdFeatureParams)
		
	
	
	def extractSrcPropSynTagSeq(self, pdFeatureParams):
		'''
		Extracts the syntactic tag sequences of predicates and arguments 
		of source side propositions
		'''
		
		return self.feseqe.extractPropSynTagSeq(self.srcConstSRLs, pdFeatureParams)
		
	
	
	def extractTransPropSynTagSeq(self, pdFeatureParams):
		'''
		Extracts the syntactic tag sequences of predicates and arguments 
		of translation side propositions
		'''
		
		return self.feseqe.extractPropSynTagSeq(self.transConstSRLs, pdFeatureParams)
		
	
	
	def extractSrcPropDepRelSeq(self, pdFeatureParams):
		'''
		Extracts the dependency relation sequences of predicates and arguments 
		of source side propositions
		'''
		
		return self.feseqe.extractPropDepRelSeq(self.srcDepSRLs, pdFeatureParams)
		
	
	
	def extractTransPropDepRelSeq(self, pdFeatureParams):
		'''
		Extracts the dependency relation sequences of predicates and arguments 
		of translation side propositions
		'''
		
		return self.feseqe.extractPropDepRelSeq(self.transDepSRLs, pdFeatureParams)
		
	
	
	def extractSrcPropPOSTagSeq(self, pdFeatureParams):
		'''
		Extracts the POS tag sequences of predicates and arguments of source
		side propositions
		'''
		
		return self.feseqe.extractPropPOSTagSeq(self.srcDepSRLs, pdFeatureParams)
		
	
	
	def extractTransPropPOSTagSeq(self, pdFeatureParams):
		'''
		Extracts the POS tag sequences of predicates and arguments of 
		translation side propositions
		'''
		
		return self.feseqe.extractPropPOSTagSeq(self.transDepSRLs, pdFeatureParams)
		
	
	
	def extractDepSRLWAProjStats(self, pdFeatureParams):
		'''
		Extracts statistic from dependency SRL projection from source
		to translations using word alignments
		'''
		
		return self.feseqe.extractDepSRLWAProjStats(pdParams = pdFeatureParams)
		
	
	
	def extractPAMScores(self, pdFeatureParams):
		'''
		Extracts PAM (predicate-argument match) scores
		'''
		
		return self.feseqe.extractPAMScores(pdParams = pdFeatureParams)
		
	
	
	## Statistics #####
	
	def getFeatureValueFreqs(self, pFeatureName):
		'''
		Extracts and returns values for pFeature name with their frequency 
		in the data set.
		'''
		
		return util.groupBy(self.extractFeature(pFeatureName))
		
	
	
	def featureCorrelation(self, pFeatureName, pdFeatureParams, pTargetValues):
		'''
		Computes the correlation between a feature and target values (QE
		scores).
		'''
		
		vlTargeValues = [float(v) for v in pTargetValues.strip().split('\n')]
		vlFeatureValues = self.extractFeature(pFeatureName, pdFeatureParams)
		
		vEval = util.PredEval(vlFeatureValues, vlTargeValues)
		
		return vEval.r()
		
	
