#! /usr/bin/python

"""
	This module provides a class to extracts syntax features for quality 
	estimation.
	
	TODO: VERIFY AND CHANGE extractSentLength() TO MAKE USE OF NEW SURFACE
	FORM ATTRIBUTES OF FESSQE INSTEAD OF CONSTITUENCY TREES DIRECTLY (TO
	BE CHEAPER). 
	
	Version 1.4												(05-Jun-2014)
	- extractDepTopDepRels() is edited to comply with changes in 
	  DepNode.getDepDepRels().
	
	Version 1.3												(03-Feb-2014)
	- The module is renamed from fessqe.py to fesyqe.py. fessqe.py is the 
	  interface module for both syntax (fesyqe.py) and semantic-based 
	  features (feseqe.py). So, codes importing fessqe.py will still work
	  fine.
	
	Version 1.2												(02-Aug-2013)
	- _loadConstParses() is upgraded to support log+tree (log probability
	  followed by the tree) format.
	
	Version 1.1												(25-Jul-2013)
	- src-dep-rel-token-pairs and trans-dep-rel-token-pairs are extracted.
	- Division by zero bug in _extractAvgConstArity() is fixed.
	- Bug in _extractParseProb() is fixed. 
	
	Version 1.0												(08-Jul-2013)
	- const-tree-size-ratio is renamed to const-tree-size-rel.
	- extractConstTreeSizeRatio() is renamed to extractRelConstTreeSize()
	  and _extractConstTreeSizeRatio() to _extractRelConstTreeSize()
	- Sentence length can also be called constituency tree width with new
	  additional feature names src-const-width and trans-const-width.
	- extractSentLengthRatio() is renamed to extractRelSentLength()
	  and _extractSentLengthRatio() to _extractRelSentLength()
	- extractParseProbRatio() is renamed to extractRelParseProb()
	  and _extractParseProbRatio() to _extractRelParseProb()
	- extractPOSLMScoreRatio() is renamed to extractRelPOSLMScore()
	  and _extractPOSLMScoreRatio() to _extractRelPOSLMScore()
	- extractPOSCountDiff() is renamed to extractRelPOSCount()
	  and _extractPOSCountRatio() to _extractRelPOSCount()
	- _extractConstRules() is upgraded to choose for lexical rules.
	
	Version 0.9												(28-Jun-2013)
	- src-dep-rel-count, trans-dep-rel-count and dep-rel-count-rel are 
	  extracted.
	- src-dep-rel-lm-score, trans-dep-rel-lm-score and dep-rel-lm-score-rel
	  are extracted.
	- src-dep-rel-ngram-quart-freq, trans-dep-rel-ngram-quart-freq and
	  dep-rel-ngram-quart-freq-rel are extracted.
	- Methods using SRILM are edited to comply with the new version of SRILM.
	- src-pos-ngram-quart-freq, trans-pos-ngram-quart-freq and
	  pos-ngram-quart-freq-rel are extracted.
	
	Version 0.8												(21-Jun-2013)
	- Dependency trees are loaded.
	- src-dep-top-pos, trans-dep-top-pos are extracted.
	- src-dep-top-dep-count, trans-dep-top-dep-count, dep-top-dep-count-rel
	  are extracted.
	- src-dep-top-dep-rels and trans-dep-top-dep-rels are extracted.
	- src-dep-top-deps-poss and trans-dep-top-deps-poss are extracted.
	- src-dep-avg-dep-count and trans-avg-dep-count are extracted.
	- src-dep-tree-height, trans-dep-tree-height, dep-tree-height-rel are
	  extracted.
	- const-tree-height-ratio is renamed to const-tree-height-rel.
	- extractConstTreeHeightRatio() is renamed to extractRelConstTreeHeight()
	  and _extractConstTreeHeightRatio() to _extractRelConstTreeHeight()
	- src-dep-rel-ngrams and trans-dep-rel-ngrams are extracted.
	
	
	Version 0.7												(15-Jun-2013)
	- _extractPOSngrams() improved to used _catchall tag map.
	- src-pos-lm-score, trans-pos-lm-score and pos-lm-score-ratio are 
	  extracted.
	- extractFeature() is improved to support suffixes in feature names
	  in the configuration file to be able to have various settings of
	  a feature type in.
	- src-pos-count, trans-pos-count and pos-count-diff are extracted.
	- src-phrase-tag-count, trans-phrase-tag-count and phrase-tag-count-rel
	  are extracted.
	- src-pos-loc, trans-pos-loc, pos-loc-rel are extracted.
	- src-avg-const-arty, trans-avg-const-arity, const-arity-rel are
	  extracted.
	- src-const-rules, trans-const-rules are extracted.
	- src-const-pseudo-score, trans-const-pseudo-score, const-pseudo-score-rel
	  are extracted.
	
	Version 0.6												(30-May-2013)
	- loadSrcConstTrees(), loadSrcConstTrees() and related functions were
	  upgraded to load other parsing details as well (e.g. parse probs).
	  They are renamed to ...ConstParses().
	- Parse probability is extracted.
	- featureCorrelation() is added.
	- POS n-grams are extracted.
	
	Version 0.5												(13-May-2013)
	- Feature extraction is now customizable through newly-implemented 
	  feature parameters in svmfg. This helps prevent having one feature
	  per variation of a general feature.
	- Syntactic tags of constituency trees can be mapped to a new tag set
	  before extracting features.
	- Ratio of source and translation constituency tree sizes is extracted.
	- Sentence lengths and their ratio are extracted. Though this is not
	  a syntactic feature, this is extracted for comparison purposes.
	
	Version 0.4												(09-May-2013)
	- Coarse-grained constituency chunks extraction supports filering
	  punctuation.
	- Source and translation constituency tree sizes are extracted.
	- Ratio of source and translation constituency tree heights is extracted.
	
	Version 0.3												(08-May-2013)
	- In all namings target is changed to translation.
	- Source and translation constituency tree heights are extracted.
	- Source and translation coarse-grained constituency chuncks are extracted.
	- extractFeature() is added.
	- getFeatureValueFreqs() is added.
	
	Version 0.2												(07-May-2013)
	- Source and target constituency root tags are extracted.
	
	Version 0.1												(01-May-2013)
	- FESSQE is added.
	- loadSrcConstTrees() and loadTrgConstTrees() were added to FESSQE.
	
"""

from parse import parse
from utils import util
from nlp import nlp, lm
import sys, yaml, copy

class FESyQE:
	'''
	Feature Extractor for syntax-based quality estimation
	'''
	
	
	# feature extraction methods
	
	def extractConstRoot(self, pConstTrees):
		'''
		Extracts root node tags of constituency trees
		'''
		
		vlRootTags = []
		for vTree in pConstTrees:
			vlRootTags.append(vTree.root.synTag)
		
		return vlRootTags
		
	
	
	def extractConstTreeHeight(self, pConstTrees):
		'''
		Extracts the height of constituency trees.
		'''
		
		vlHeights = []
		for vTree in pConstTrees:
			vlHeights.append(vTree.getHeight())
		
		return vlHeights
		
	
	
	def extractRelConstTreeHeight(self, pConstTrees1, pConstTrees2, pdParams):
		'''
		Extracts the relationship between the constituency trees heights
		in pConstTrees1 and pConstTrees2.
		
		pdParams contains the following parameters:
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for constituency tree height relationship!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for constituency tree height relationship!" % vParam)
		
		vlRels = []
		for vTree1, vTree2 in zip(pConstTrees1, pConstTrees2):
			vTreeHeight1 = vTree1.getHeight()
			vTreeHeight2 = vTree2.getHeight()
			
			if vRelationship == "diff":
				vlRels.append(vTreeHeight1 - vTreeHeight2)
			elif vRelationship == "ratio":
				if vTreeHeight2 == 0:
					vlRels.append(0)
				else:
					vlRels.append(vTreeHeight1 * 1.0 / vTreeHeight2)
		
		return vlRels
		
	
	
	def extractConstTreeSize(self, pConstTrees):
		'''
		Extracts the size of constituency trees.
		'''
		
		vlSizes = []
		for vTree in pConstTrees:
			vlSizes.append(vTree.getSize())
		
		return vlSizes
		
	
	
	def extractRelConstTreeSize(self, pConstTrees1, pConstTrees2, pdParams):
		'''
		Extracts the relationship between the constituency trees sizes
		in pConstTrees1 and pConstTrees2.
		
		pdParams contains the following parameters:
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for constituency tree size relationship!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for constituency tree size relationship!" % vParam)
		
		vlRels = []
		for vTree1, vTree2 in zip(pConstTrees1, pConstTrees2):
			vTreeHSize1 = vTree1.getSize()
			vTreeHSize2 = vTree2.getSize()
			
			if vRelationship == "diff":
				vlRels.append(vTreeHSize1 - vTreeHSize2)
			elif vRelationship == "ratio":
				if vTreeHSize2 == 0:
					vlRels.append(0)
				else:
					vlRels.append(vTreeHSize1 * 1.0 / vTreeHSize2)
		
		return vlRels
		
	
	
	def extractCoarseConstChunks(self, pConstTrees, pdParams):
		'''
		Extracts coarse-grained chunks from constituency trees.
		
		pdParams is a dictionary specifying following options:
		  - "remove punctuation": if set to true, punctuation tags will 
		                          be removed from the sequence.
		  - "tag map": if set to a file containing the map to a new syntactic
		               tag set, the tags in the tree will be mapped. If 
		               None or false, nothing will be done.
		'''
		
		# extracting parameters
		
		vflgRemPunct = False
		vflgMaptags = False
		vTagMapFile = ""
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["remove punctuation", "rem punct"]:
				vflgRemPunct = vValue
			elif vParam == "tag map":
				if vValue != None and vValue != False:
					vflgMaptags = True
					vTagMapFile = vValue
			else:
				raise Exception("'%s' is not a valid feature parameter for source-coarse-constituency-chunks" % vParam)
		
		# mapping tags if requested
		if vflgMaptags:
			vlTrees = self.mapConstTreeTags(pConstTrees, vTagMapFile)
		else:
			vlTrees = pConstTrees
		
		vlChucks = []
		if vflgRemPunct:
			for vTree in vlTrees:
				vChunks = vTree.getCoarseChunks()
				vFilteredChunck = []
				for i in range(len(vChunks)):
					if vChunks[i] not in ['.', ':', ',', '``', "''", "PONCT"]:
						vFilteredChunck.append(vChunks[i])
				vlChucks.append('-'.join(vFilteredChunck))
		else:
			for vTree in vlTrees:
				vlChucks.append('-'.join(vTree.getCoarseChunks()))
		
		return vlChucks
		
	
	
	def extractSentLength(self, pConstTrees):
		'''
		Extract the sentence lengths (constituency tree widths) from the
		trees.
		'''
		
		vlLens = []
		for vTree in pConstTrees:
			vlLens.append(vTree.getSentLength())
		
		return vlLens
		
	
	
	def extractRelSentLength(self, pConstTrees1, pConstTrees2, pdParams):
		'''
		Extracts the relationship between the sentence lengths (constituency
		trees widths) of pConstTrees1 and pConstTrees2.
		
		pdParams contains the following parameters:
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for sentence length relationship!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for sentence length relationship!" % vParam)
		
		vlRels = []
		for vTree1, vTree2 in zip(pConstTrees1, pConstTrees2):
			vSentLen1 = vTree1.getSentLength()
			vSentLen2 = vTree2.getSentLength()
			
			if vRelationship == "diff":
				vlRels.append(vSentLen1 - vSentLen2)
			elif vRelationship == "ratio":
				if vSentLen2 == 0:
					vlRels.append(0)
				else:
					vlRels.append(vSentLen1 * 1.0 / vSentLen2)
		
		return vlRels
		
	
	
	def extractParseProb(self, pConstParses, pdParams):
		'''
		Extracts parse probabilities from pConstParses.
		
		pdParams contains the following parameters:
		- "type": type of probability which can either be "prob" for standard
		          probability and "log" for log probability. The default is
		          normal probability.
		'''
		
		vProbType = "prob"
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["type", "Type"]:
				vProbType = vValue
			else:
				raise Exception("'%s' is not a valid feature parameter for parse probability" % vParam)
		
		## Some sentences may not have parses thus no parse probabilities.
		## We assign them a low probability here.
		
		vlProbs = []
		for vParse in pConstParses:
			if vProbType == "prob":
				if vParse.prob == None:
					vlProbs.append(0)
				else:
					vlProbs.append(vParse.prob)
			elif vProbType == "log":
				if vParse.logProb == None:
					vlProbs.append(-100)
				else:
					vlProbs.append(vParse.logProb)
		
		return vlProbs
		
	
	
	def extractRelParseProb(self, pConstParses1, pConstParses2, pdParams):
		'''
		Extracts the relationship between the parse probabilities of 
		pConstTrees1 and pConstTrees2.
		
		pdParams contains the following parameters:
		- "relatioship": type of relationship: difference, ratio
		- "type": type of probability which can either be "prob" for standard
		          probability and "lo" for log probability. The default is
		          normal probability.
		'''
		
		vProbType = "prob"
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["type", "Type"]:
				vProbType = vValue
			elif vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for parse probability relationship!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for parse probability relationship" % vParam)
		
		vlRels = []
		for vParse1, vParse2 in zip(pConstParses1, pConstParses2):
			if vProbType == "prob":
				vValue1 = vParse1.prob
				vValue2 = vParse2.prob
			elif vProbType == "log":
				vValue1 = vParse1.logProb
				vValue2 = vParse2.logProb
			
			if vRelationship == "diff":
				vlRels.append(vValue1 - vValue2)
			elif vRelationship == "ratio":
				if vValue2 == 0:
					vlRels.append(0)
				else:
					vlRels.append(vValue1 * 1.0 / vValue2)
		
		return vlRels
		
	
	
	def extractPOSngrams(self, pConstTrees, pdParams):
		'''
		Extracts POS n-grams, where n is specified in pdFeatureParams by 
		order or n.
		
		pdParams contains the following parameters:
		  - "n" or "order": order of n-gram
		  - "tag map": if set to a file containing the map to a new syntactic
		               tag set, the tags in the tree will be mapped. If 
		               None or false, nothing will be done.
		'''
		
		vN = 0
		vflgMaptags = False
		vdTagMap = ""
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["n", "order"]:
				vN = vValue
			elif vParam == "tag map":
				if vValue != None and vValue != False:
					vflgMaptags = True
					try:
						vdTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open tag map file: %s" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for POS n-grams" % vParam)
		
		vlNGrams = []
		for vTree in pConstTrees:
			vlPOSTags = []
			# mapping POS tags if requested
			if vflgMaptags:
				for vTag in vTree.getPOSs():
					try:
						vlPOSTags.append(vdTagMap[vTag])
					except KeyError:
						if "_catchall" in vdTagMap:
							vlPOSTags.append(vdTagMap["_catchall"])
						else:
							vlPOSTags.append(vTag)
			else:
				vlPOSTags = vTree.getPOSs()
			
			# extracting n-grams
			vlInstanceNGrams = nlp.extractNGrams(vlPOSTags, vN)
			vlNGrams.append(['-'.join(ngram) for ngram in vlInstanceNGrams])
		
		return vlNGrams
		
	
	
	def extractPOSLMScore(self, pdParams):
		'''
		Extracts POS language model scores (prob, log prob or perplexity) 
		from the ouput of the LM tool provided in feature parameters.
		
		SRILM output is supported.
		
		pdParams contains the following parameters:
		  - "srilm output": output file of SRILM containing segment-level 
		                    scores.
		  - "type": score type (prob, log, ppl, ppl1)
		            The default is prob (normal probability).
		'''
		
		vScoreType = 'prob'
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["srilm output", "SRILM output"]:
				if vValue != None and vValue != False:
					try:
						vSRILMOutput = open(vValue).read()
					except IOError:
						raise Exception("Cannot open SRILM output file: %s" % vValue)
			elif vParam in ["type", "Type", "score type", "Score Type"]:
				vScoreType = vValue
			else:
				raise Exception("'%s' is not a valid feature parameter for POS LM score" % vParam)
		
		vSRILM = lm.SRILM()
		vSRILM.loadEval(vSRILMOutput, 1)
		vlScores = []
		if vScoreType == "prob":
			vlScores = [segscores.prob for segscores in vSRILM.segmentEvals];
		elif vScoreType == "log":
			vlScores = [segscores.logprob for segscores in vSRILM.segmentEvals];
		elif vScoreType == "ppl":
			vlScores = [segscores.ppl for segscores in vSRILM.segmentEvals];
		elif vScoreType == "ppl1":
			vlScores = [segscores.ppl1 for segscores in vSRILM.segmentEvals];
		else:
			raise Exception("Unknown LM score type: %s" % vScoreType)
		
		return vlScores
		
	
	
	def extractRelPOSLMScore(self, pdParams):
		'''
		Extracts the relationship between the source and translation POS
		language model scores (prob, log prob or perplexity).
		
		pdParams contains the following parameters:
		  - "src srilm output": output file of SRILM containing segment-level 
		                        scores for the source side.
		  - "trans srilm output": output file of SRILM containing segment-level 
		                          scores for the translation side.
		  - "type": score type (prob, log, ppl, ppl1)
		            The default is prob (normal probability).
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		vScoreType = "prob"
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["src srilm output", "src SRILM output"]:
				if vValue != None and vValue != False:
					try:
						vSrcSRILMOutput = open(vValue).read()
					except IOError:
						raise Exception("Cannot open SRILM output file for source side: %s" % vValue)
			elif vParam in ["trans srilm output", "trans SRILM output"]:
				if vValue != None and vValue != False:
					try:
						vTransSRILMOutput = open(vValue).read()
					except IOError:
						raise Exception("Cannot open SRILM output file for translation side: %s" % vValue)
			elif vParam in ["type", "Type", "score type", "Score Type"]:
				vScoreType = vValue
			elif vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for POS LM score relationship!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for POS LM score relationship!" % vParam)
		
		vSrcSRILM = lm.SRILM()
		vSrcSRILM.loadEval(vSrcSRILMOutput,  1)
		vTransSRILM = lm.SRILM()
		vTransSRILM.loadEval(vTransSRILMOutput, 1)
		
		vlRels = []
		for srcSegScores, transSegScores in zip(vSrcSRILM.segmentEvals, vTransSRILM.segmentEvals):
			if vScoreType == "prob":
				vSrcScore = srcSegScores.prob
				vTransScore = transSegScores.prob;
			elif vScoreType == "log":
				vSrcScore = srcSegScores.logprob
				vTransScore = transSegScores.logprob
			elif vScoreType == "ppl":
				vSrcScore = srcSegScores.ppl
				vTransScore = transSegScores.ppl
			elif vScoreType == "ppl1":
				vSrcScore = srcSegScores.ppl1
				vTransScore = transSegScores.ppl1
			else:
				raise Exception("Unknown LM score type: %s" % vScoreType)
			
			if vRelationship == "diff":
				vlRels.append(vSrcScore - vTransScore)
			elif vRelationship == "ratio":
				if vTransScore == 0:
					vlRels.append(0)
				else:
					vlRels.append(vSrcScore * 1.0 / vTransScore)
		
		return vlRels
		
	
	
	def extractPOSCount(self, pConstTrees, pdParams):
		'''
		Extracts the number of times POS tags in a given set appear in 
		the trees.
		
		The set of POS tags is given in the feature parameters.
		
		pdParams contains the following parameters:
		  - "POS tags": list of POS tags  to be counted
		  - "tag map": if set to a file containing the map to a new syntactic
		               tag set, the tags in the tree will be mapped. If 
		               None or false, nothing will be done.
		'''
		
		vlPOSTagList = []
		vflgMapTags = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["pos tags", "POS tags"]:
				if vValue != None and isinstance(vValue, list):
					vlPOSTagList = vValue
			elif vParam == "tag map":
				if vValue != None and vValue != False:
					vflgMapTags = True
					try:
						vdTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open tag map file: %s" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for POS count" % vParam)
		
		if len(vlPOSTagList) == 0:
			raise Exception("No POS tag list is provided in the feature parameters!")
		
		vlCounts = []
		for vTree in pConstTrees:
			vlPOSTags = []
			# mapping POS tags if requested
			if vflgMapTags:
				for vTag in vTree.getPOSs():
					try:
						vlPOSTags.append(vdTagMap[vTag])
					except KeyError:
						if "_catchall" in vdTagMap:
							vlPOSTags.append(vdTagMap["_catchall"])
						else:
							vlPOSTags.append(vTag)
			else:
				vlPOSTags = vTree.getPOSs()
			
			# extracting counts
			vlCounts.append(len([pos for pos in vlPOSTags if pos in vlPOSTagList]))
		
		return vlCounts
		
	
	
	def extractRelPOSCount(self, pSrcConstTrees, pTransConstTrees, pdParams):
		'''
		Extracts the relationship between the number of times POS tags 
		in a given set appear in the input constituency trees.
		
		The set of POS tags is given in the feature parameters.
		
		POS tags can be mapped into a different tag set.
		
		Relationship is given in feature parameters.
		
		pdParams contains the following parameters:
		  - "POS tags": list of POS tags  to be counted
		  - "src tag map": source side tag map
		  - "trans tag map": translation side tag map
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		vlPOSTagList = []
		vflgMaptagsSrc = False
		vflgMaptagsTrans = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["pos tags", "POS tags"]:
				if vValue != None and isinstance(vValue, list):
					vlPOSTagList = vValue
			elif vParam == "src tag map":
				if vValue != None and vValue != False:
					vflgMaptagsSrc = True
					try:
						vdSrcTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open source side tag map file: %s" % vValue)
			elif vParam == "trans tag map":
				if vValue != None and vValue != False:
					vflgMaptagsTrans = True
					try:
						vdTransTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open translation side tag map file: %s" % vValue)
			elif vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for POS count relationship!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for POS count relationship!" % vParam)
		
		if len(vlPOSTagList) == 0:
			raise Exception("No POS tag list is provided in the feature parameters!")
		
		vlRels = []
		for vSrcTree, vTransTree in zip(pSrcConstTrees, pTransConstTrees):
			vlSrcPOSTags = []
			# mapping POS tags if requested
			if vflgMaptagsSrc:
				for vTag in vSrcTree.getPOSs():
					try:
						vlSrcPOSTags.append(vdSrcTagMap[vTag])
					except KeyError:
						if "_catchall" in vdSrcTagMap:
							vlSrcPOSTags.append(vdSrcTagMap["_catchall"])
						else:
							vlSrcPOSTags.append(vTag)
			else:
				vlSrcPOSTags = vSrcTree.getPOSs()
			
			vlTransPOSTags = []
			# mapping POS tags if requested
			if vflgMaptagsTrans:
				for vTag in vTransTree.getPOSs():
					try:
						vlTransPOSTags.append(vdTransTagMap[vTag])
					except KeyError:
						if "_catchall" in vdTransTagMap:
							vlTransPOSTags.append(vdTransTagMap["_catchall"])
						else:
							vlTransPOSTags.append(vTag)
			else:
				vlTransPOSTags = vTransTree.getPOSs()
			
			# extracting counts and computing difference
			vSrcPOSCount = len([pos for pos in vlSrcPOSTags if pos in vlPOSTagList])
			vTransPOSCount = len([pos for pos in vlTransPOSTags if pos in vlPOSTagList])
			
			if vRelationship == "diff":
				vlRels.append(vSrcPOSCount - vTransPOSCount)
			elif vRelationship == "ratio":
				if vTransPOSCount == 0:
					vlRels.append(0)
				else:
					vlRels.append(vSrcPOSCount * 1.0 / vTransPOSCount)
		
		return vlRels
		
	
	
	def extractPhraseTagCount(self, pConstTrees, pdParams):
		'''
		Extracts the number of times phrase tags in a given set appear in 
		the trees.
		
		The set of phrase tags is given in the feature parameters.
		
		Phrase tags can be mapped into a different tag set.
		
		pdParams contains the following parameters:
		  - "phrase tags": list of phrase tags  to be counted
		  - "tag map": if set to a file containing the map to a new syntactic
		               tag set, the tags in the tree will be mapped. If 
		               None or false, nothing will be done.
		'''
		
		vlPhraseTagList = []
		vflgMaptags = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["phrase tags", "tags"]:
				if vValue != None and isinstance(vValue, list):
					vlPhraseTagList = vValue
			elif vParam == "tag map":
				if vValue != None and vValue != False:
					vflgMaptags = True
					try:
						vdTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open tag map file: %s" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for phrase tag count" % vParam)
		
		if len(vlPhraseTagList) == 0:
			raise Exception("No phrase tag list is provided in the feature parameters!")
		
		vlCounts = []
		for vTree in pConstTrees:
			vNewTree = copy.deepcopy(vTree)
			
			# mapping phrase tags if requested
			if vflgMaptags:
				vNewTree.mapTags(vdMap)
			
			# extracting counts
			vdTagCounts = vNewTree.getTagset()
			vlCounts.append(sum([count for tag, count in vdTagCounts.iteritems() if tag in vlPhraseTagList]))
		
		return vlCounts
		
	
	
	def extractRelPhraseTagCount(self, pSrcConstTrees, pTransConstTrees, pdParams):
		'''
		Extracts the relationship (e.g. ratio) between the number of times
		phrase tags in a given set appear in the source and translation 
		side, where the relationship is specified in feature parameters.
		
		The set of phrase tags is given in the feature parameters.
		
		Phrase tags tags can be mapped into a different tag set.
		
		pdParams contains the following parameters:
		  - "src tags": list of phrase tags of source side to be counted
		  - "trans tags": list of phrase tags of translation side to be 
		                  counted
		  - "src tag map": source side tag map
		  - "trans tag map": translation side tag map
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		vlSrcTagList = []
		vlTransTagList = []
		vflgMaptagsSrc = False
		vflgMaptagsTrans = False
		vRelationship = ""
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["src tags", "source tags"]:
				if vValue != None and isinstance(vValue, list):
					vlSrcTagList = vValue
			elif vParam in ["trans tags", "translation tags"]:
				if vValue != None and isinstance(vValue, list):
					vlTransTagList = vValue
			elif vParam == "src tag map":
				if vValue != None and vValue != False:
					vflgMaptagsSrc = True
					try:
						vdSrcTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open source side tag map file: %s" % vValue)
			elif vParam == "trans tag map":
				if vValue != None and vValue != False:
					vflgMaptagsTrans = True
					try:
						vdTransTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open translation side tag map file: %s" % vValue)
			elif vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for phrase tag count!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for phrase tag count relationship!" % vParam)
		
		if len(vlSrcTagList) == 0:
			raise Exception("No source side phrase tag list is provided in the feature parameters!")
		if len(vlTransTagList) == 0:
			raise Exception("No translation side phrase tag list is provided in the feature parameters!")
		if vRelationship == "":
			raise Exception("Relatioship operator for phrase tag count is not specified!")
		
		vlRels = []
		for vSrcTree, vTransTree in zip(pSrcConstTrees, pTransConstTrees):
			vlSrcTags = []
			vNewSrcTree = copy.deepcopy(vSrcTree)
			
			# mapping source phrase tags if requested
			if vflgMaptagsSrc:
				vNewSrcTree.mapTags(vdSrcTagMap)
			
			# extracting source tag counts
			vdSrcTagCounts = vNewSrcTree.getTagset()
			vSrcTagCount = sum([count for tag, count in vdSrcTagCounts.iteritems() if tag in vlSrcTagList])
			
			vlTransTags = []
			vNewTransTree = copy.deepcopy(vTransTree)
			
			# mapping translation phrase tags if requested
			if vflgMaptagsTrans:
				vNewTransTree.mapTags(vdTransTagMap)
			
			# extracting translation tag counts
			vdTransTagCounts = vNewTransTree.getTagset()
			vTransTagCount = sum([count for tag, count in vdTransTagCounts.iteritems() if tag in vlTransTagList])
			
			# computing relationship
			if vRelationship == "diff":
				vlRels.append(vSrcTagCount - vTransTagCount)
			elif vRelationship == "ratio":
				if vTransTagCount == 0:
					vlRels.append(vSrcTagCount / 0.1)
				else:
					vlRels.append(vSrcTagCount * 1.0 / vTransTagCount)
		
		return vlRels
		
	
	
	def extractPOSLoc(self, pConstTrees, pdParams):
		'''
		Extracts the location of the first occurrence of a specific POS 
		tag from the beginning of the sentence.
		
		The POS tag is given in the feature parameters.
		
		POS tags can be mapped into a different tag set.
		
		pdParams contains the following parameters:
		  - "POS tag": or "tag" is the tag the location of which to be
		               extracted. 
		  - "tag map": if set to a file containing the map to a new syntactic
		               tag set, the tags in the tree will be mapped. If 
		               None or false, nothing will be done.
		'''
		
		vPOSTag = ''
		vflgMaptags = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["pos tag", "POS tag", "tag"]:
				if vValue != None:
					vPOSTag = vValue
			elif vParam == "tag map":
				if vValue != None and vValue != False:
					vflgMaptags = True
					try:
						vdTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open tag map file: %s" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for POS location" % vParam)
		
		if vPOSTag == '':
			raise Exception("No POS tag is provided in the feature parameters!")
		
		vlLocs = []
		for vTree in pConstTrees:
			vlPOSTags = []
			# mapping POS tags if requested
			if vflgMaptags:
				for vTag in vTree.getPOSs():
					try:
						vlPOSTags.append(vdTagMap[vTag])
					except KeyError:
						if "_catchall" in vdTagMap:
							vlPOSTags.append(vdTagMap["_catchall"])
						else:
							vlPOSTags.append(vTag)
			else:
				vlPOSTags = vTree.getPOSs()
			
			# extracting locations
			vLoc = 0
			for i, vPOS in enumerate(vlPOSTags):
				if vPOS == vPOSTag:
					vLoc = i + 1
					break
			
			vlLocs.append(vLoc)
		
		return vlLocs
		
	
	
	def extractRelPOSLoc(self, pSrcConstTrees, pTransConstTrees, pdParams):
		'''
		Extracts the relatioship between the location of the first occurrence
		of a specific POS tag from the beginning of the source sentence 
		and from the beginning of the translation sentence.
		
		The POS tag is given in the feature parameters.
		
		POS tags can be mapped into a different tag set.
		
		pdParams contains the following parameters:
		  - "POS tag": or "tag" is the tag the location of which to be
		               extracted.
		  - "src tag map": source side tag map
		  - "trans tag map": translation side tag map
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		vPOSTag = ''
		vflgMaptagsSrc = False
		vflgMaptagsTrans = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["pos tag", "POS tag", "tag"]:
				if vValue != None:
					vPOSTag = vValue
			elif vParam in ["src tag map", "source tag map"]:
				if vValue != None and vValue != False:
					vflgMaptagsSrc = True
					try:
						vdSrcTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open source side tag map file: %s" % vValue)
			elif vParam in ["trans tag map", "translation tag map"]:
				if vValue != None and vValue != False:
					vflgMaptagsTrans = True
					try:
						vdTransTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open translation side tag map file: %s" % vValue)
			elif vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relatioship operator for phrase tag count!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for POS location relationship" % vParam)
		
		if vPOSTag == '':
			raise Exception("No POS tag is provided in the feature parameters!")
		
		vlRels = []
		for vSrcTree, vTransTree in zip(pSrcConstTrees, pTransConstTrees):
			
			## source side
			
			vlPOSTags = []
			# mapping POS tags if requested
			if vflgMaptagsSrc:
				for vTag in vSrcTree.getPOSs():
					try:
						vlPOSTags.append(vdSrcTagMap[vTag])
					except KeyError:
						if "_catchall" in vdSrcTagMap:
							vlPOSTags.append(vdSrcTagMap["_catchall"])
						else:
							vlPOSTags.append(vTag)
			else:
				vlPOSTags = vTree.getPOSs()
			
			# extracting locations
			vSrcLoc = 0
			for i, vPOS in enumerate(vlPOSTags):
				if vPOS == vPOSTag:
					vSrcLoc = i + 1
					break
			
			## translation side
			
			vlPOSTags = []
			# mapping POS tags if requested
			if vflgMaptagsTrans:
				for vTag in vTransTree.getPOSs():
					try:
						vlPOSTags.append(vdTransTagMap[vTag])
					except KeyError:
						if "_catchall" in vdTransTagMap:
							vlPOSTags.append(vdTransTagMap["_catchall"])
						else:
							vlPOSTags.append(vTag)
			else:
				vlPOSTags = vTree.getPOSs()
			
			# extracting locations
			vTransLoc = 0
			for i, vPOS in enumerate(vlPOSTags):
				if vPOS == vPOSTag:
					vTransLoc = i + 1
					break
			
			## computing relationship
			
			if vRelationship == "diff":
				vlRels.append(vSrcLoc - vTransLoc)
			elif vRelationship == "ratio":
				if vTransLoc == 0:
					vlRels.append(0)
				else:
					vlRels.append(vSrcLoc * 1.0 / vTransLoc)
		
		
		return vlRels
		
	
	
	def extractAvgConstArity(self, pConstTrees, pdParams):
		'''
		Extracts the average arity of the production rules of the nodes
		of the translation constituency tree.
		
		The arity is in fact the number of children of that node. The 
		terminal nodes are not counted in averaging.
		'''
		
		## The arity of a production rule expanding a constituent is the
		## number of its children (both terminal and non-terminal). In 
		## fact other than the root node, all the other nodes are children,
		## so their count is the total number of children in the tree
		## (tree size - 1). This number should be divided by the number of
		## parents. Again, all tree nodes except terminals (or pre-terminals
		## because we consider them in one node) are parents (tree size -
		## sentence length)
		
		vlValues = []
		for vTree in pConstTrees:
			vSize = vTree.getSize()
			vLen = vTree.getSentLength()
			vDenom = vSize - vLen
			if vDenom == 0:
				vlValues.append(0)
			else:
				vlValues.append((vSize - 1) * 1.0 / vDenom)
		
		return vlValues
		
	
	
	def extractRelAvgConstArity(self, pSrcConstTrees, pTransConstTrees, pdParams):
		'''
		Extracts the relationship between the average arity of the production
		rules of the nodes of the source and of the translation constituency
		trees.
		
		pdParams contains the following parameters:
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for average constituency arity!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for average constituency arity!" % vParam)
		
		vlRels = []
		for vSrcTree, vTransTree in zip(pSrcConstTrees, pTransConstTrees):
			
			## source side
			vSize = vSrcTree.getSize()
			vLen = vSrcTree.getSentLength()
			vSrcAvg = (vSize - 1) * 1.0 / (vSize - vLen)
			
			## translation side
			vSize = vTransTree.getSize()
			vLen = vTransTree.getSentLength()
			vTransAvg = (vSize - 1) * 1.0 / (vSize - vLen)
			
			## computing relationship
			
			if vRelationship == "diff":
				vlRels.append(vSrcAvg - vTransAvg)
			elif vRelationship == "ratio":
				if vTransAvg == 0:
					vlRels.append(0)
				else:
					vlRels.append(vSrcAvg * 1.0 / vTransAvg)
		
		
		return vlRels
		
	
	
	def extractConstRules(self, pConstTrees, pdParams):
		'''
		Extracts the CFG production rules expanding each constituency tree
		node of the tree.
		
		Certain tags can be ignored and excluded from the rules.
		
		Lexical, non-lexical or both types of rules can be extracted.
		
		Tags can be mapped into a different tag set.
		
		pdParams is a dictionary specifying following options:
		  - "ignore list": list of tags to be ignored and not included.
		  - "rule type": can be "lexical" (or "lex"), "non-lexical" or "nolex"
		                 or "all" (or "both"). Default is "all"
		  - "tag map": if set to a file containing the map to a new syntactic
		               tag set, the tags in the tree will be mapped. If 
		               None or false, nothing will be done.
		'''
		
		vIgnoreList = []
		vRuleType = "all"
		vflgMaptags = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam == "ignore list":
				if vValue != None and isinstance(vValue, list):
					vIgnoreList = vValue
			elif vParam == "tag map":
				if vValue != None and vValue != False:
					vflgMaptags = True
					try:
						vdTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open tag map file: %s" % vValue)
			elif vParam == "rule type":
				if vValue != None and vValue != False:
					vRuleType = vValue.lower()
			else:
				raise Exception("'%s' is not a valid feature parameter for CFG production rules!" % vParam)
		
		vlRules = []
		for vTree in pConstTrees:
			vNewTree = copy.deepcopy(vTree)
			
			# mapping phrase tags if requested
			if vflgMaptags:
				vNewTree.mapTags(vdTagMap)
			
			# extracting rules
			if vRuleType == "all":
				vlTreeRule = vNewTree.extractCFGRules(pflgNoLexRule = False)
			elif vRuleType in ["nolex", "non-lexical"]:
				vlTreeRule = vNewTree.extractCFGRules(pflgNoLexRule = True)
			elif vRuleType in ["lex", "lexical"]:
				vlTreeRule = vNewTree.extractLexRules()
			
			# formatting rules
			vlTreeRuleStrs = ["%s>%s" % (rule["lhs"], '-'.join([tag for tag in rule["rhs"] if tag not in vIgnoreList])) for rule in vlTreeRule]
			
			vlRules.append(vlTreeRuleStrs)
		
		return vlRules
		
	
	
	def extractConstPseudoScore(self, pdParams):
		'''
		Extracts scores of the constituency parses evaluated against the 
		output of a pseudo-reference parser. 
		
		Scores are in a file (one per line) given in feature parameters.
		
		pdParams is a dictionary specifying following options:
		  - "scores" or "pseudo scores": score file name which contains 
		                                 one score per line.
		'''
		
		vlScores = []
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["scores", "pseudo scores"]:
				if vValue != None:
					vlScores = [float(score) for score in open(vValue).read().strip().split('\n')]
			else:
				raise Exception("'%s' is not a valid feature parameter for pseuso-reference constituency parse scores!" % vParam)
		
		return vlScores
		
	
	
	def extractRelConstPseudoScore(self, pdParams):
		'''
		Extracts the relationship between the pseudo reference evaluation
		scores of the source and translation constituency parses against 
		the output of their corresponding pseudo-reference parsers. 
		
		Scores are in files (one per line) given in feature parameters.
		
		Relationship is given in feature parameters.
		
		pdParams is a dictionary specifying following options:
		  - "src scores" or "src pseudo scores": source side score file 
		                                         name which contains one 
		                                         score per line.
		  - "trans scores" or "trans pseudo scores": source side score file 
		                                             name which contains 
		                                             one score per line.
		'''
		
		vlSrcScores = []
		vlTransScores = []
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["src scores", "src pseudo scores"]:
				if vValue != None:
					vlSrcScores = [float(score) for score in open(vValue).read().strip().split('\n')]
			elif vParam in ["trans scores", "trans pseudo scores"]:
				if vValue != None:
					vlTransScores = [float(score) for score in open(vValue).read().strip().split('\n')]
			elif vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship pseudo-reference constituency parse scores!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for pseudo-reference constituency parse scores!" % vParam)
		
		vlRels = []
		for vSrcScore, vTransScore in zip(vlSrcScores, vlTransScores):
			if vRelationship == "diff":
				vlRels.append(vSrcScore - vTransScore)
			elif vRelationship == "ratio":
				if vTransScore == 0:
					vlRels.append(0)
				else:
					vlRels.append(vSrcScore * 1.0 / vTransScore)
		
		return vlRels
		
	
	
	def extractDepTopPOS(self, pDepTrees, pdParams):
		'''
		Extracts the POS tag os the root node (the dependent of the dummy 
		root node) of the dependency trees.
		
		Tags can be mapped into a different tag set.
		
		pdParams is a dictionary specifying following options:
		  - "tag map": if set to a file containing the map to a new syntactic
		               tag set, the tags in the tree will be mapped. If 
		               None or false, nothing will be done.
		'''
		
		vflgMaptags = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam == "tag map":
				if vValue != None and vValue != False:
					vflgMaptags = True
					try:
						vdTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open tag map file: %s" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for dependency top node POS tag!" % vParam)
		
		vlPOSTags = []
		for vTree in pDepTrees:
			vTag = vTree.getTopNodePOS()
			
			# mapping POS tag if requested
			if vflgMaptags:
				try:
					vlPOSTags.append(vdTagMap[vTag])
				except KeyError:
					if "_catchall" in vdTagMap:
						vlPOSTags.append(vdTagMap["_catchall"])
					else:
						vlPOSTags.append(vTag)
			else:
				vlPOSTags.append(vTag)
		
		return vlPOSTags
		
	
	
	def extractDepTopDepCount(self, pDepTrees, pdParams):
		'''
		Extracts number of dependents of the top node (the dependent of 
		the dummy root node) of the dependency trees
		'''
		
		vlCounts = []
		for vTree in pDepTrees:
			vlCounts.append(vTree.topNode.getDepCount())
		
		return vlCounts
		
	
	
	def extractRelDepTopDepCount(self, pDepTrees1, pDepTrees2, pdParams):
		'''
		Extracts the relatinoship between the number of dependents of the
		top node (the dependent of the dummy root node) of the dependency
		trees 1 and 2.
		
		pdParams contains the following parameters:
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for top node dependent counts!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for top node dependent counts relationship!" % vParam)
		
		vlRels = []
		for vTree1, vTree2 in zip(pDepTrees1, pDepTrees2):
			vDepCount1 = vTree1.topNode.getDepCount()
			vDepCount2 = vTree2.topNode.getDepCount()
			
			if vRelationship == "diff":
				vlRels.append(vDepCount1 - vDepCount2)
			elif vRelationship == "ratio":
				if vDepCount2 == 0:
					vlRels.append(0)
				else:
					vlRels.append(vDepCount1 * 1.0 / vDepCount2)
		
		return vlRels
		
	
	
	def extractDepTopDepRels(self, pDepTrees, pdParams):
		'''
		Extracts the sequence of dependency relations of top nodes of the
		dependency trees to their dependents.
		'''
		
		vlDepRels = []
		for vTree in pDepTrees:
			vlDepRels.append('-'.join(vTree.topNode.getDepDepRels()))
		
		return vlDepRels
		
	
	
	def extractDepTopDepsPOSs(self, pDepTrees, pdParams):
		'''
		Extracts the sequence of POS tags of the dependents of the top
		nodes of the dependency trees.
		
		Tags can be mapped into a different tag set.
		
		pdParams is a dictionary specifying following options:
		  - "tag map": if set to a file containing the map to a new syntactic
		               tag set, the tags in the tree will be mapped. If 
		               None or false, nothing will be done.
		'''
		
		vflgMaptags = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam == "tag map":
				if vValue != None and vValue != False:
					vflgMaptags = True
					try:
						vdTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open tag map file: %s" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for dependency top node dependents' POS tag sequence!" % vParam)
		
		vlPOSSeqences = []
		for vTree in pDepTrees:
			vlPOSTags = [node.posTag for node in vTree.topNode.getDependentNodes()]
			
			# mapping POS tag if requested
			if vflgMaptags:
				for i in range(len(vlPOSTags)):
					try:
						vlPOSTags[i] = vdTagMap[vlPOSTags[i]]
					except KeyError:
						if "_catchall" in vdTagMap:
							vlPOSTags[i] = vdTagMap["_catchall"]
			
			vlPOSSeqences.append('-'.join(vlPOSTags))
		
		return vlPOSSeqences
		
	
	
	def extractDepAvgDepCount(self, pDepTrees, pdParams):
		'''
		Extracts average number of dependents of the dependency trees.
		'''
		
		vlAvgs = []
		for vTree in pDepTrees:
			vlDepCounts = [node.getDepCount() for node in vTree.nodes[1: ]]
			vlAvgs.append(sum(vlDepCounts) * 1.0 / len(vlDepCounts))
		
		return vlAvgs
		
	
	
	def extractRelDepAvgDepCount(self, pDepTrees1, pDepTrees2, pdParams):
		'''
		Extracts the relationship between the average number of dependents
		in the dependency trees 1 and 2.
		
		pdParams contains the following parameters:
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for average dependent counts!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for average dependent counts relationship!" % vParam)
		
		vlRels = []
		for vTree1, vTree2 in zip(pDepTrees1, pDepTrees2):
			vlDepCounts = [node.getDepCount() for node in vTree1.nodes[1: ]]
			vAvgDepCount1 = sum(vlDepCounts) * 1.0 / len(vlDepCounts)
			vlDepCounts = [node.getDepCount() for node in vTree2.nodes[1: ]]
			vAvgDepCount2 = sum(vlDepCounts) * 1.0 / len(vlDepCounts)
			
			if vRelationship == "diff":
				vlRels.append(vAvgDepCount1 - vAvgDepCount2)
			elif vRelationship == "ratio":
				if vAvgDepCount2 == 0:
					vlRels.append(0)
				else:
					vlRels.append(vAvgDepCount1 * 1.0 / vAvgDepCount2)
		
		return vlRels
		
	
	
	def extractDepTreeHeight(self, pDepTrees, pdParams):
		'''
		Extracts the height of the dependency trees.
		'''
		
		return [tree.height for tree in pDepTrees]
		
	
	
	def extractRelDepTreeHeight(self, pDepTrees1, pDepTrees2, pdParams):
		'''
		Extracts the relationship between the heights of the dependency 
		trees 1 and 2.
		
		pdParams contains the following parameters:
		  - "rel" or "relatioship": type of relationship: difference, ratio
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for dependency tree heights!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for dependency tree heights!" % vParam)
		
		vlRels = []
		for vTree1, vTree2 in zip(pDepTrees1, pDepTrees2):
			vTreeHeight1 = vTree1.height
			vTreeHeight2 = vTree2.height
			
			if vRelationship == "diff":
				vlRels.append(vTreeHeight1 - vTreeHeight2)
			elif vRelationship == "ratio":
				if vTreeHeight2 == 0:
					vlRels.append(0)
				else:
					vlRels.append(vTreeHeight1 * 1.0 / vTreeHeight2)
		
		return vlRels
		
	
	
	def extractDepRelngrams(self, pDepTrees, pdParams):
		'''
		Extracts dependency relation n-grams from the dependency trees.
		
		Order is of n-gram is given in feature parameters.
		
		pdParams contains the following parameters:
		  - "n" or "order": order of n-gram
		'''
		
		vN = 0
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["n", "order"]:
				vN = vValue
			else:
				raise Exception("'%s' is not a valid feature parameter for dependency relation n-grams" % vParam)
		
		vlNGrams = []
		for vTree in pDepTrees:
			vlDepRels = [rel for rels in map(lambda node: [headDeps[1] for headDeps in node.headDeps], vTree.nodes[1: ]) for rel in rels]
			
			# extracting n-grams
			vlInstanceNGrams = nlp.extractNGrams(vlDepRels, vN)
			vlNGrams.append(['-'.join(ngram) for ngram in vlInstanceNGrams])
		
		return vlNGrams
		
	
	
	def extractDepRelCount(self, pDepTrees, pdParams):
		'''
		Extracts the number of times dependency relations in a given set
		appear in the trees.
		
		The set of dependency relations is given in the feature parameters.
		
		Dependency relation tags can be mapped into a different tag set.
		
		pdParams contains the following parameters:
		  - "rels" or "dep rels" or "relations" or "tags": list of dependency
		    relations to be counted
		  - "tag map": if set to a file containing the map to a new tag
		               set, the tags in the tree will be mapped. If 
		               None or false, nothing will be done.
		'''
		
		vlDepList = []
		vflgMapTags = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["rels", "dep rels", "relations", "tags"]:
				if vValue != None and isinstance(vValue, list):
					vlDepList = vValue
			elif vParam == "tag map":
				if vValue != None and vValue != False:
					vflgMapTags = True
					try:
						vdTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open tag map file: %s" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for dependency relation count!" % vParam)
		
		if len(vlDepList) == 0:
			raise Exception("No dependency relation list is provided in the feature parameters!")
		
		vlCounts = []
		for vTree in pDepTrees:
			vdDepRels = {}
			# mapping dependency relation tags if requested
			if vflgMapTags:
				for vTag, vCount in vTree.getDepRelTagset().iteritems():
					try:
						vdDepRels[vdTagMap[vTag]] = vCount
					except KeyError:
						if "_catchall" in vdTagMap:
							vdDepRels[vdTagMap["_catchall"]] = vCount
						else:
							vdDepRels[vTag] = vCount
			else:
				vdDepRels = vTree.getDepRelTagset()
			
			# extracting counts
			vlCounts.append(sum([count for rel, count in vdDepRels.iteritems() if rel in vlDepList]))
		
		return vlCounts
		
	
	
	def extractRelDepRelCount(self, pSrcDepTrees, pTransDepTrees, pdParams):
		'''
		Extracts the relationship between the number of times dependency
		relations in a given set appear in the source side trees and the 
		number of times dependency relations in another given set (possibly 
		the same) appear in the translation side trees.
		
		Dependency relation tags can be mapped into a different tag set.
		
		pdParams contains the following parameters:
		  - "rels" or "dep rels" or "relations" or "tags": list of dependency
		    relations to be counted
		  - "src tag map": source side tag map
		  - "trans tag map": translation side tag map
		'''
		
		vlDepList = []
		vflgMapTagsSrc = False
		vflgMapTagsTrans = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["rels", "dep rels", "relations", "tags"]:
				if vValue != None and isinstance(vValue, list):
					vlDepList = vValue
			elif vParam == "src tag map":
				if vValue != None and vValue != False:
					vflgMapTagsSrc = True
					try:
						vdSrcTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open source side tag map file: %s" % vValue)
			elif vParam == "trans tag map":
				if vValue != None and vValue != False:
					vflgMapTagsTrans = True
					try:
						vdTransTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open translation side tag map file: %s" % vValue)
			elif vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for dependency tree heights!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for dependency relation count relationship" % vParam)
		
		if len(vlDepList) == 0:
			raise Exception("No dependency relation list is provided in the feature parameters!")
		
		vlRels = []
		for vSrcTree, vTransTree in zip(pSrcDepTrees, pTransDepTrees):
			vlSrcDepRels = []
			# mapping dependency relation tags if requested
			if vflgMapTagsSrc:
				for vTag, vCount in vSrcTree.getDepRelTagset().iteritems():
					try:
						vdSrcDepRels[vdSrcTagMap[vTag]] = vCount
					except KeyError:
						if "_catchall" in vdSrcTagMap:
							vdSrcDepRels[vdSrcTagMap["_catchall"]] = vCount
						else:
							vdSrcDepRels[vTag] = vCount
			else:
				vdSrcDepRels = vSrcTree.getDepRelTagset()
			
			vlTransDepRels = []
			# mapping dependency relation tags if requested
			if vflgMapTagsTrans:
				for vTag, vCount in vTransTree.getDepRelTagset().iteritems():
					try:
						vdTransDepRels[vdTransTagMap[vTag]] = vCount
					except KeyError:
						if "_catchall" in vdTransTagMap:
							vdTransDepRels[vdTransTagMap["_catchall"]] = vCount
						else:
							vdTransDepRels[vTag] = vCount
			else:
				vdTransDepRels = vTransTree.getDepRelTagset()
			
			# extracting counts
			vSrcCount = sum([count for rel, count in vdSrcDepRels.iteritems() if rel in vlDepList])
			vTransCount = sum([count for rel, count in vdTransDepRels.iteritems() if rel in vlDepList])
			
			# computing relationships
			if vRelationship == "diff":
				vlRels.append(vSrcCount - vTransCount)
			elif vRelationship == "ratio":
				if vTransCount == 0:
					vlRels.append(0)
				else:
					vlRels.append(vSrcCount * 1.0 / vTransCount)
		
		return vlRels
		
	
	
	def extractDepRelLMScore(self, pdParams):
		'''
		Extracts POS language model scores (prob, log prob or perplexity) 
		from the ouput of the LM tool provided in feature parameters.
		
		SRILM output is supported.
		
		pdParams contains the following parameters:
		  - "srilm output": output file of SRILM containing segment-level 
		                    scores.
		  - "type": score type (prob, log, ppl, ppl1)
		            The default is prob (normal probability).
		'''
		
		vScoreType = 'prob'
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["srilm output", "SRILM output"]:
				if vValue != None:
					try:
						vSRILMOutput = open(vValue).read()
					except IOError:
						raise Exception("Cannot open SRILM output file: %s" % vValue)
			elif vParam in ["type", "Type", "score type", "Score Type"]:
				vScoreType = vValue
			else:
				raise Exception("'%s' is not a valid feature parameter for dependency relation LM score" % vParam)
		
		vSRILM = lm.SRILM()
		vSRILM.loadEval(vSRILMOutput, 1)
		vlScores = []
		if vScoreType == "prob":
			vlScores = [segscores.prob for segscores in vSRILM.segmentEvals];
		elif vScoreType == "log":
			vlScores = [segscores.logprob for segscores in vSRILM.segmentEvals];
		elif vScoreType == "ppl":
			vlScores = [segscores.ppl for segscores in vSRILM.segmentEvals];
		elif vScoreType == "ppl1":
			vlScores = [segscores.ppl1 for segscores in vSRILM.segmentEvals];
		else:
			raise Exception("Unknown LM score type: %s" % vScoreType)
		
		return vlScores
		
	
	
	def extractRelDepRelLMScore(self, pdParams):
		'''
		Extracts the relationship between the LM scores (prob, log prob
		or perplexity) of dependency relations of the source and translation 
		sides from the output of the LM tool provided in feature parameters.
		
		SRILM output is supported.
		
		pdParams contains the following parameters:
		  - "src srilm output": output file of SRILM containing segment-level 
		                        scores for the source side.
		  - "trans srilm output": output file of SRILM containing segment-level 
		                          scores for the translation side.
		  - "type": score type (prob, log, ppl, ppl1)
		            The default is prob (normal probability).
		'''
		
		vScoreType = "prob"
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["src srilm output", "src SRILM output"]:
				if vValue != None and vValue != False:
					try:
						vSrcSRILMOutput = open(vValue).read()
					except IOError:
						raise Exception("Cannot open SRILM output file for source side: %s" % vValue)
			elif vParam in ["trans srilm output", "trans SRILM output"]:
				if vValue != None and vValue != False:
					try:
						vTransSRILMOutput = open(vValue).read()
					except IOError:
						raise Exception("Cannot open SRILM output file for translation side: %s" % vValue)
			elif vParam in ["type", "Type", "score type", "Score Type"]:
				vScoreType = vValue
			elif vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for dependency relation LM probability relationship!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for dependency relation LM probability relationship!" % vParam)
		
		vSrcSRILM = lm.SRILM()
		vSrcSRILM.loadEval(vSrcSRILMOutput, 1)
		vTransSRILM = lm.SRILM()
		vTransSRILM.loadEval(vTransSRILMOutput, 1)
		
		vlRels = []
		for vSrcScores, vTransScores in zip(vSrcSRILM.segmentEvals, vTransSRILM.segmentEvals):
			# extracting scores
			if vScoreType == "prob":
				vSrcScore = vSrcScores.prob
				vTransScore = vTransScores.prob;
			elif vScoreType == "log":
				vSrcScore = vSrcScores.logprob
				vTransScore = vTransScores.logprob;
			elif vScoreType == "ppl":
				vSrcScore = vSrcScores.ppl
				vTransScore = vTransScores.ppl;
			elif vScoreType == "ppl1":
				vSrcScore = vSrcScores.ppl1
				vTransScore = vTransScores.ppl1;
			else:
				raise Exception("Unknown LM score type: %s" % vScoreType)
			
			# computing relationships
			if vRelationship == "diff":
				vlRels.append(vSrcScore - vTransScore)
			elif vRelationship == "ratio":
				if vTransScore == 0:
					vlRels.append(0)
				else:
					vlRels.append(vSrcScore * 1.0 / vTransScore)
		
		return vlRels
		
	
	
	def extractDepRelngramQFreq(self, pDepTrees, pdParams):
		'''
		Extracts the dependency relation n-gram frequency (average number 
		of n-grams) in n-gram frequency quarters of a corpus.
		
		The quarter and n-grams counts of the corpus are given in feature 
		parameters.
		
		N-gram counts should be in the format output by SRILM.
		
		pdParams contains the following parameters:
		  - "n" or "order": order of n-gram
		  - "q" or "quarter": the frequency quarter
		  - "n-gram counts": a file containing n-grams counts of a corpus
		                     used for language modeling (in SRILM count 
		                     file format)
		'''
		
		vN = 0
		vQ = 0
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["n", "order"]:
				vN = vValue
			elif vParam in ["q", "quarter"]:
				vQ = vValue
			elif vParam == "n-gram counts":
				if vValue != None:
					try:
						vCounts = open(vValue).read()
					except IOError:
						raise Exception("Cannot open count file: %s" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for dependency relation frequency quartile feature!" % vParam)
		
		vNGramCounts = lm.NGramCounts(vN)
		vNGramCounts.loadNGramCounts(vCounts, "SRILM")
		vlQuarterNgrams = [i[0] for i in vNGramCounts.getQuantileSubset(4, vQ)]
		
		vlFreqs = []
		for vTree in pDepTrees:
			vlDepRels = [rel for rels in map(lambda node: [headDeps[1] for headDeps in node.headDeps], vTree.nodes[1: ]) for rel in rels]
			
			# extracting n-grams
			vlInstanceNGrams = nlp.extractNGrams(vlDepRels, vN)
			
			if len(vlInstanceNGrams) == 0:
				vlFreqs.append(0.0)
			else:
				# calculating frequency of n-grams in the quarter 
				vlFreqs.append(len([ngram for ngram in vlInstanceNGrams if ' '.join(ngram) in vlQuarterNgrams]) * 1.0 / len(vlInstanceNGrams))
		
		return vlFreqs
		
	
	
	def extractRelDepRelngramQFreq(self, pSrcDepTrees, pTransDepTrees, pdParams):
		'''
		Extracts the relationship between the dependency relation n-gram 
		frequency of the source side in n-gram frequency quarters of a 
		corpus with those of the translation side.
		
		The quarter and n-grams counts of the corpus are given in feature 
		parameters.
		
		N-gram counts should be in the format output by SRILM.
		
		pdParams contains the following parameters:
		  - "n" or "order": order of n-gram
		  - "q" or "quarter": the frequency quarter
		  - "rel" or "relationship": type of relationship: difference, ratio
		  - "src n-gram counts": a file containing n-grams counts of a corpus
		                         used for language modeling for the source
		                         side (in SRILM count file format)
		  - "trans n-gram counts": a file containing n-grams counts of a corpus
		                         used for language modeling for the translation
		                         side (in SRILM count file format)
		'''
		
		vN = 0
		vQ = 0
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["n", "order"]:
				vN = vValue
			elif vParam in ["q", "quarter"]:
				vQ = vValue
			elif vParam == "src n-gram counts":
				if vValue != None:
					try:
						vSrcCounts = open(vValue).read()
					except IOError:
						raise Exception("Cannot open source count file: %s" % vValue)
			elif vParam == "trans n-gram counts":
				if vValue != None:
					try:
						vTransCounts = open(vValue).read()
					except IOError:
						raise Exception("Cannot open translation count file: %s" % vValue)
			elif vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for dependency relation frequency quartile!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for dependency relation frequency quartile feature!" % vParam)
		
		vSrcNGramCounts = lm.NGramCounts(vN)
		vSrcNGramCounts.loadNGramCounts(vSrcCounts, "SRILM")
		vlSrcQuarterNgrams = [i[0] for i in vSrcNGramCounts.getQuantileSubset(4, vQ)]
		vTransNGramCounts = lm.NGramCounts(vN)
		vTransNGramCounts.loadNGramCounts(vTransCounts, "SRILM")
		vlTransQuarterNgrams = [i[0] for i in vTransNGramCounts.getQuantileSubset(4, vQ)]
		
		vlRels = []
		for vSrcTree, vTransTree in zip(pSrcDepTrees, pTransDepTrees):
			vlSrcDepRels = [rel for rels in map(lambda node: [headDeps[1] for headDeps in node.headDeps], vSrcTree.nodes[1: ]) for rel in rels]
			vlTransDepRels = [rel for rels in map(lambda node: [headDeps[1] for headDeps in node.headDeps], vTransTree.nodes[1: ]) for rel in rels]
			
			# extracting instance n-grams
			vlSrcInstanceNGrams = nlp.extractNGrams(vlSrcDepRels, vN)
			vlTransInstanceNGrams = nlp.extractNGrams(vlTransDepRels, vN)
			
			# calculating frequency of n-grams in the quarter 
			
			if len(vlSrcInstanceNGrams) == 0:
				vSrcFreq = 0.0
			else:
				vSrcFreq = len([ngram for ngram in vlSrcInstanceNGrams if ' '.join(ngram) in vlSrcQuarterNgrams]) * 1.0 / len(vlSrcInstanceNGrams)
				
			if len(vlTransInstanceNGrams) == 0:
				vTransFreq = 0.0
			else:
				vTransFreq = len([ngram for ngram in vlTransInstanceNGrams if ' '.join(ngram) in vlTransQuarterNgrams]) * 1.0 / len(vlTransInstanceNGrams)
			
			# computing relationships
			if vRelationship == "diff":
				vlRels.append(vSrcFreq - vTransFreq)
			elif vRelationship == "ratio":
				if vTransFreq == 0:
					vlRels.append(0)
				else:
					vlRels.append(vSrcFreq * 1.0 / vTransFreq)
		
		return vlRels
		
	
	
	def extractPOSngramQFreq(self, pConstTrees, pdParams):
		'''
		Extracts the POS n-gram frequency (average number of n-grams) in
		n-gram frequency quarters of a corpus.
		
		The quarter and n-grams counts of the corpus are given in feature 
		parameters.
		
		N-gram counts should be in the format output by SRILM.
		
		POS tags can be mapped to a different tag set.
		
		pdParams contains the following parameters:
		  - "n" or "order": order of n-gram
		  - "q" or "quarter": the frequency quarter
		  - "n-gram counts": a file containing n-grams counts of a corpus
		                     used for language modeling (in SRILM count 
		                     file format)
		  - "tag map": if set to a file containing the map to a new tag
		               set, the tags in the tree will be mapped. If None
		               or false, nothing will be done.
		'''
		
		vN = 0
		vQ = 0
		vflgMaptags = False
		vdTagMap = ""
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["n", "order"]:
				vN = vValue
			elif vParam in ["q", "quarter"]:
				vQ = vValue
			elif vParam == "n-gram counts":
				if vValue != None:
					try:
						vCounts = open(vValue).read()
					except IOError:
						raise Exception("Cannot open count file: %s" % vValue)
			elif vParam == "tag map":
				if vValue != None and vValue != False:
					vflgMaptags = True
					try:
						vdTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open tag map file: %s" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for POS frequency quartile feature!" % vParam)
		
		vNGramCounts = lm.NGramCounts(vN)
		vNGramCounts.loadNGramCounts(vCounts, "SRILM")
		vlQuarterNgrams = [i[0] for i in vNGramCounts.getQuantileSubset(4, vQ)]
		
		vlFreqs = []
		for vTree in pConstTrees:
			vlPOSTags = []
			# mapping POS tags if requested
			if vflgMaptags:
				for vTag in vTree.getPOSs():
					try:
						vlPOSTags.append(vdTagMap[vTag])
					except KeyError:
						if "_catchall" in vdTagMap:
							vlPOSTags.append(vdTagMap["_catchall"])
						else:
							vlPOSTags.append(vTag)
			else:
				vlPOSTags = vTree.getPOSs()
			
			# extracting n-grams
			vlInstanceNGrams = nlp.extractNGrams(vlPOSTags, vN)
			
			if len(vlInstanceNGrams) == 0:
				vlFreqs.append(0.0)
			else:
				# calculating frequency of n-grams in the quarter 
				vlFreqs.append(len([ngram for ngram in vlInstanceNGrams if ' '.join(ngram) in vlQuarterNgrams]) * 1.0 / len(vlInstanceNGrams))
		
		return vlFreqs
		
	
	
	def extractRelPOSngramQFreq(self, pSrcConstTrees, pTransConstTrees, pdParams):
		'''
		Extracts the relationship between the POS n-gram frequency of the
		source side in n-gram frequency quarters of a corpus with those 
		of the translation side.
		
		The quarter and n-grams counts of the corpus are given in feature 
		parameters.
		
		N-gram counts should be in the format output by SRILM.
		
		POS tags can be mapped to a different tag set.
		
		pdParams contains the following parameters:
		  - "n" or "order": order of n-gram
		  - "q" or "quarter": the frequency quarter
		  - "rel" or "relationship": type of relationship: difference, ratio
		  - "src n-gram counts": a file containing n-grams counts of a corpus
		                         used for language modeling for the source
		                         side (in SRILM count file format)
		  - "trans n-gram counts": a file containing n-grams counts of a corpus
		                         used for language modeling for the translation
		                         side (in SRILM count file format)
		  - "src tag map": source side tag map
		  - "trans tag map": translation side tag map
		'''
		
		vN = 0
		vQ = 0
		vflgMaptagsSrc = False
		vflgMaptagsTrans = False
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["n", "order"]:
				vN = vValue
			elif vParam in ["q", "quarter"]:
				vQ = vValue
			elif vParam == "src n-gram counts":
				if vValue != None:
					try:
						vSrcCounts = open(vValue).read()
					except IOError:
						raise Exception("Cannot open source count file: %s" % vValue)
			elif vParam == "trans n-gram counts":
				if vValue != None:
					try:
						vTransCounts = open(vValue).read()
					except IOError:
						raise Exception("Cannot open translation count file: %s" % vValue)
			elif vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for dependency relation frequency quartile!" % vValue)
			elif vParam == "src tag map":
				if vValue != None and vValue != False:
					vflgMaptagsSrc = True
					try:
						vdSrcTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open source side tag map file: %s" % vValue)
			elif vParam == "trans tag map":
				if vValue != None and vValue != False:
					vflgMaptagsTrans = True
					try:
						vdTransTagMap = parse.loadTagMap(open(vValue).read(), pflgSelfMapIn = True)
					except IOError:
						raise Exception("Cannot open translation side tag map file: %s" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for dependency relation frequency quartile feature!" % vParam)
		
		vSrcNGramCounts = lm.NGramCounts(vN)
		vSrcNGramCounts.loadNGramCounts(vSrcCounts, "SRILM")
		vlSrcQuarterNgrams = [i[0] for i in vSrcNGramCounts.getQuantileSubset(4, vQ)]
		vTransNGramCounts = lm.NGramCounts(vN)
		vTransNGramCounts.loadNGramCounts(vTransCounts, "SRILM")
		vlTransQuarterNgrams = [i[0] for i in vTransNGramCounts.getQuantileSubset(4, vQ)]
		
		vlRels = []
		for vSrcTree, vTransTree in zip(pSrcConstTrees, pTransConstTrees):
			vlSrcPOSTags = []
			# mapping POS tags if requested
			if vflgMaptagsSrc:
				for vTag in vSrcTree.getPOSs():
					try:
						vlSrcPOSTags.append(vdSrcTagMap[vTag])
					except KeyError:
						if "_catchall" in vdSrcTagMap:
							vlSrcPOSTags.append(vdSrcTagMap["_catchall"])
						else:
							vlSrcPOSTags.append(vTag)
			else:
				vlSrcPOSTags = vSrcTree.getPOSs()
			
			vlTransPOSTags = []
			# mapping POS tags if requested
			if vflgMaptagsTrans:
				for vTag in vTransTree.getPOSs():
					try:
						vlTransPOSTags.append(vdTransTagMap[vTag])
					except KeyError:
						if "_catchall" in vdTransTagMap:
							vlTransPOSTags.append(vdTransTagMap["_catchall"])
						else:
							vlTransPOSTags.append(vTag)
			else:
				vlTransPOSTags = vTransTree.getPOSs()
			
			# extracting instance n-grams
			vlSrcInstanceNGrams = nlp.extractNGrams(vlSrcPOSTags, vN)
			vlTransInstanceNGrams = nlp.extractNGrams(vlTransPOSTags, vN)
			
			# calculating frequency of n-grams in the quarter 
			
			if len(vlSrcInstanceNGrams) == 0:
				vSrcFreq = 0.0
			else:
				vSrcFreq = len([ngram for ngram in vlSrcInstanceNGrams if ' '.join(ngram) in vlSrcQuarterNgrams]) * 1.0 / len(vlSrcInstanceNGrams)
				
			if len(vlTransInstanceNGrams) == 0:
				vTransFreq = 0.0
			else:
				vTransFreq = len([ngram for ngram in vlTransInstanceNGrams if ' '.join(ngram) in vlTransQuarterNgrams]) * 1.0 / len(vlTransInstanceNGrams)
			
			# computing relationships
			if vRelationship == "diff":
				vlRels.append(vSrcFreq - vTransFreq)
			elif vRelationship == "ratio":
				if vTransFreq == 0:
					vlRels.append(0)
				else:
					vlRels.append(vSrcFreq * 1.0 / vTransFreq)
		
		return vlRels
		
	
	
	def extractDepRelTokenPairs(self, pDepTrees, pdParams):
		'''
		Extracts the dependency relation-token pairs from dependency trees.
		'''
		
		vlRelTokenPairs = []
		for vTree in pDepTrees:
			vlRelTokenPairs.append(["%s-%s" % (rel, token) for token, rel in vTree.getFormRelPairs()])
		
		return vlRelTokenPairs
		
	
	
	def extractTemp(self, pConstTrees, pdParams):
		'''
		'''
		
		vlValues = []
		for vTree in pConstTrees:
			vSize = vTree.getSize()
			vlValues.append((vSize - 1) * 1.0 / (vSize - vTree.getSentLength()))
		
		return vlValues
		
	
	
	def extractRelTemp(self, pSrcConstTrees, pTransConstTrees, pdParams):
		'''
		pdParams contains the following parameters:
		  - "relatioship": type of relationship: difference, ratio
		'''
		
		for vParam, vValue in pdParams.iteritems():
			if vParam in ["relationship", "rel"]:
				if vValue != None:
					if vValue in ["diff", "difference"]:
						vRelationship = "diff"
					elif vValue == "ratio":
						vRelationship = "ratio"
					else:
						raise Exception("'%s' is not a valid relationship operator for TEMP!" % vValue)
			else:
				raise Exception("'%s' is not a valid feature parameter for TEMP" % vParam)
		
		vlValues = []
		for vSrcTree, vTransTree in zip(pSrcConstTrees, pTransConstTrees):
			
			## source side
			vSrcValue = vSrcTree.root.getChildrenCount()
			
			## translation side
			vTransValue = vTransTree.root.getChildrenCount()
			
			## computing relationship
			
			if vRelationship == "diff":
				vlRels.append(vSrcValue - vTransValue)
			elif vRelationship == "ratio":
				if vTransLoc == 0:
					vlRels.append(0)
				else:
					vlRels.append(vSrcValue * 1.0 / vTransValue)
		
		return vlRels
		
	
	
