#! /usr/bin/env python

"""	
	MT Quality Estimation Metric based on Predicate-Argument Structure Match
	
	Version 1.1													(25-Jul-2014)
	- extractAnalysis() is added to PAM analysis.
	
	Version 1.0													(03-Jul-2014)
	- PAMAnalyze is added.
	- Scores are now in decimal format and stored in PAM after scoring.
	- PAM.__str__() and getScoresStr() are added.
	
	Version: 0.3												(10-Jun-2014)
	- LTPAM is added.
	
	Version: 0.2												(21-May-2014)
	- WAPAM.score() is upgraded to support argument boundary option. 
	
	Version: 0.1												(13-May-2014)
	- PAM is added.
	- WAPAM is added.
"""


import srl, copy, util, re
from decimal import Decimal


class PAM:
	'''
	Abstract class for PAM QE metic
	'''
	
	def __init__(self, pSrcSRL, pTrgSRL, pNoSrcPropScores, pNoTrgPropScores):
		'''
		Constructor
		'''
		
		self.srcSRL = pSrcSRL
		self.trgSRL = pTrgSRL
		
		# scores to be returned if no proposition in source
		self.noSrcPropScores = pNoSrcPropScores
		# scores to be returned if no proposition in target but in source
		self.noTrgPropScores = pNoTrgPropScores
		
		# dictionary of scores customized in inherited classes
		self.scores = {}
		
	
	
	def _toDecimal(self, pdScores):
		'''
		Converts float scores to decimal
		
		Decimal format is more suitable for exact calculation. Float numbers
		exhibit strange behaviours.
		'''
		
		vdScores = {} 
		for vScore in pdScores.iterkeys():
			vdScores[vScore] = Decimal(str(pdScores[vScore]))
		
		return vdScores
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		return self.getScoresStr()
		
	
	
	def getScoresStr(self):
		'''
		Returns string representation of PAM scores
		'''
		
		return '\t'.join(["%s: %s" % (n, v) for n, v in self.scores.iteritems()])
		
	


class WAPAM(PAM):
	'''
	Class for PAM QE metic using word alignment and dependency SRL
	'''
	
	def __init__(self, pSrcCoNLLSentence, pTrgCoNLLSentence, pWordAlignments, pNoSrcPropScores = 1, pNoTrgPropScores = 0, pOnSRLConstraintViolation = "exception", pPredicateType = "all"):
		'''
		Constructor
		'''
		
		PAM.__init__(self, 
		             pSrcCoNLLSentence.getSRL(pdAnnotationOptions = {"on-srl-constraint-violation": pOnSRLConstraintViolation, "gold-or-predicted": "pred", "predicate-type": pPredicateType}), 
		             pTrgCoNLLSentence.getSRL(pdAnnotationOptions = {"on-srl-constraint-violation": pOnSRLConstraintViolation, "gold-or-predicted": "pred"}),
		             pNoSrcPropScores = pNoSrcPropScores, 
		             pNoTrgPropScores = pNoTrgPropScores)
		
		self.srcCoNLLSentence = pSrcCoNLLSentence
		self.trgCoNLLSentence = pTrgCoNLLSentence
		self.wa = pWordAlignments
		
	
	
	def score(self, pMethod = "project", pArgMatchBoundary = "word", pTrgConstTree = None, pPredLabelMatchLevel = "frameset", pflgCollapseAdjunct = "False"):
		'''
		Scores the translation and returns the scores
		
		pMethods specifies the scoring method. Currently only one method
		is implemented as follows:
		- project(ion): projecting the source SRL to the target and treating
		                it as SRL scoring
		pArgMatchBoundary specifes the boundary to compare the argument
		role fillers. It can be:
		- word
		- phrase or constituent
		'''
		
		if pMethod.lower().startswith("project"):
			vdScores = self._scoreProj(pArgMatchBoundary = pArgMatchBoundary, pTrgConstTree = pTrgConstTree, pPredLabelMatchLevel = pPredLabelMatchLevel, pflgCollapseAdjunct = pflgCollapseAdjunct)
			self.scores = vdScores
			return vdScores
		else:
			raise Exception("%s is not a supported method" % pMethod)
		
	
	
	def _scoreProj(self, pArgMatchBoundary = "word", pTrgConstTree = None, pPredLabelMatchLevel = "frameset", pflgCollapseAdjunct = "False"):
		'''
		Scores the translation and returns the scores
		
		The method is to first project the source SRL to the target using
		word-alignments and then treat it as SRL scoring, where the projected
		SRL resembles gold SRL and the translation SRL resembles predicted
		SRL.
		
		The scoring is similar to CoNLL 2009 but the predicate label
		comparison can be customized. In other word, unlike in CoNLL 2009
		where only frameset sense is compared, the whole frameset label
		can also be compared as well as ignoring label and taking only 
		boundary into account (even for labeled scores).
		
		Two types of boundaries can be used when assessing argument role
		filler match: word boundary and phrase boundary. In the former,
		only the aligned words are compared. For the latter, the target
		and projected SRLs which are dependency-based are convereted to 
		constituency-based SRL and then the phrase boundaries are compared.
		For this case, target constituency tree must be provided.
		
		The scores are unlabeled and labeled precision, recall and F1 which 
		are returned in a dictionary.
		'''
		
		# doing projection
		vProj = srl.DepSRLProject(self.srcCoNLLSentence.surface, self.srcSRL, self.srcCoNLLSentence.depTree(pdAnnotationOptions = {"gold-or-predicted": "pred"}), "en")
		vNewTrgDepTree = copy.deepcopy(self.trgCoNLLSentence.depTree(pdAnnotationOptions = {"gold-or-predicted": "pred"}))
		vProjSRL = vProj.waProject(self.trgCoNLLSentence.surface, vNewTrgDepTree, self.wa, "fr",
		                           pPredOneToManyMethod = 1, pPredOverlapMethod = 0, pArgOneToManyMethod = 0, pArgOverlapMethod = 0, pflgClassicDRule = True)
		
		# evaluating the projected SRL
		if pArgMatchBoundary.lower() in ["const", "constituency", "phrase"]:
			vConstProjSRL = vProjSRL.convertToConst(pTrgConstTree, pConversionMethod = "elevate", pflgLoadIntoTree = True)
			## Check if we cannot use self.trgSRL instead of getting SRL again
			vConstTrgSRL = self.trgCoNLLSentence.getSRL(pdAnnotationOptions = {"gold-or-predicted": "pred"}).convertToConst(pTrgConstTree, pConversionMethod = "elevate", pflgLoadIntoTree = True)
			vSRLComp = srl.SRLComp(pPredLabelMatchLevel = pPredLabelMatchLevel, pflgCollapseAdjunct = pflgCollapseAdjunct)
			vSRLComp.compare(vConstTrgSRL, vConstProjSRL)
		else:
			## Check if we cannot use self.trgSRL instead of getting SRL again
			vTrgSRL = self.trgCoNLLSentence.getSRL(pdAnnotationOptions = {"gold-or-predicted": "pred"})
			vSRLComp = srl.SRLComp(pPredLabelMatchLevel = pPredLabelMatchLevel, pflgCollapseAdjunct = pflgCollapseAdjunct)
			vSRLComp.compare(vTrgSRL, vProjSRL)
		
		# scoring
		vdScores = {}
		if self.srcSRL.propCount == 0:
			vdScores["UPrecision"] = self.noSrcPropScores
			vdScores["URecall"] = self.noSrcPropScores
			vdScores["UF1"] = self.noSrcPropScores
			
			vdScores["LPrecision"] = self.noSrcPropScores
			vdScores["LRecall"] = self.noSrcPropScores
			vdScores["LF1"] = self.noSrcPropScores
		elif self.trgSRL.propCount == 0:
			vdScores["UPrecision"] = self.noTrgPropScores
			vdScores["URecall"] = self.noTrgPropScores
			vdScores["UF1"] = self.noTrgPropScores
			
			vdScores["LPrecision"] = self.noTrgPropScores
			vdScores["LRecall"] = self.noTrgPropScores
			vdScores["LF1"] = self.noTrgPropScores
		else:
			vdScores["UPrecision"] = 1.0 * (vSRLComp.alignedPredCount + vSRLComp.alignedArgCount) / (self.trgSRL.propCount + self.trgSRL.globalArgCount)
			vdScores["URecall"] = 1.0 * (vSRLComp.alignedPredCount + vSRLComp.alignedArgCount) / (self.srcSRL.propCount + self.srcSRL.globalArgCount)
			if vdScores["UPrecision"] + vdScores["URecall"] == 0:
				vdScores["UF1"] = 0
			else:
				vdScores["UF1"] = (2 * vdScores["UPrecision"] * vdScores["URecall"]) / (vdScores["UPrecision"] + vdScores["URecall"])
			
			vdScores["LPrecision"] = 1.0 * (vSRLComp.frameMatchPredCount + vSRLComp.roleMatchArgCount) / (self.trgSRL.propCount + self.trgSRL.globalArgCount)
			vdScores["LRecall"] = 1.0 * (vSRLComp.frameMatchPredCount + vSRLComp.roleMatchArgCount) / (self.srcSRL.propCount + self.srcSRL.globalArgCount)
			if vdScores["LPrecision"] + vdScores["LRecall"] == 0:
				vdScores["LF1"] = 0
			else:
				vdScores["LF1"] = (2 * vdScores["LPrecision"] * vdScores["LRecall"]) / (vdScores["LPrecision"] + vdScores["LRecall"])
		
		return self._toDecimal(vdScores)
		
	


class LTPAM(PAM):
	'''
	Class for PAM QE metic using lexical translation table
	'''
	
	
	def __init__(self, pSrcCoNLLSentence, pTrgCoNLLSentence, pLTTable, pNoSrcPropScores = 1, pNoTrgPropScores = 0, pOnSRLConstraintViolation = "exception", pPredicateType = "all"):
		'''
		Constructor
		'''
		
		PAM.__init__(self, 
		             pSrcCoNLLSentence.getSRL(pdAnnotationOptions = {"on-srl-constraint-violation": pOnSRLConstraintViolation, "gold-or-predicted": "pred", "predicate-type": pPredicateType}), 
		             pTrgCoNLLSentence.getSRL(pdAnnotationOptions = {"on-srl-constraint-violation": pOnSRLConstraintViolation, "gold-or-predicted": "pred"}),
		             pNoSrcPropScores = pNoSrcPropScores, 
		             pNoTrgPropScores = pNoTrgPropScores)
		
		self.srcCoNLLSentence = pSrcCoNLLSentence
		self.trgCoNLLSentence = pTrgCoNLLSentence
		self.ltt = pLTTable
		
	
	
	def score(self, pPredLabelMatchLevel = "frameset", pflgCollapseAdjunct = "False"):
		'''
		Scores the translation and returns the scores
		
		pPredLabelMatchLevel determines whether the frameset or only the
		lemma of the predicates should be compared in finding predicate
		match.
		'''
		
		# finding aligned predicates
		
		vlPropAlignment = self._alignPredicates(pPredLabelMatchLevel)
		vlAlignedProps = [pa for pa in vlPropAlignment if pa[1] != None]
		
		# finding aligned arguments of aligned predicates
		
		vlAlignedArgs = []
		for vSrcProp, vTrgProp in vlAlignedProps:
			vlAlignedArgs += self._alignArgs(vSrcProp, vTrgProp, pflgCollapseAdjunct = pflgCollapseAdjunct)
		
		# scoring
		
		vAlignedPredCount = len(vlAlignedProps)
		vAlignedArgCount = len(vlAlignedArgs)
		vdScores = {}
		if self.srcSRL.propCount == 0:
			vdScores["Precision"] = self.noSrcPropScores
			vdScores["Recall"] = self.noSrcPropScores
			vdScores["F1"] = self.noSrcPropScores
		elif self.trgSRL.propCount == 0:
			vdScores["Precision"] = self.noTrgPropScores
			vdScores["Recall"] = self.noTrgPropScores
			vdScores["F1"] = self.noTrgPropScores
		else:
			vdScores["Precision"] = 1.0 * (vAlignedPredCount + vAlignedArgCount) / (self.trgSRL.propCount + self.trgSRL.globalArgCount)
			vdScores["Recall"] = 1.0 * (vAlignedPredCount + vAlignedArgCount) / (self.srcSRL.propCount + self.srcSRL.globalArgCount)
			if vdScores["Precision"] + vdScores["Recall"] == 0:
				vdScores["F1"] = 0
			else:
				vdScores["F1"] = (2 * vdScores["Precision"] * vdScores["Recall"]) / (vdScores["Precision"] + vdScores["Recall"])
		
		self.scores = self._toDecimal(vdScores)
		
		return self.scores
		
	
	
	def _alignPredicates(self, pPredLabelMatchLevel = "frameset"):
		'''
		Finds the target predicate aligned to each source predicate and
		returns list of tuples the first element of which is the proposition
		of the source predicate and the second its aligned target proposition
		if found and None otherwise.
		'''
		
		vlTrgProps = self.trgSRL.getPropositions()[ : ]
		
		vlPropAlignment = []
		
		## The method is to first look at the predicate labels assuming 
		## that both sides have the same frameset labels. Note that the
		## label can be both frameset or lemma only. If there is a match
		## it is assumed as an alignment and the search is stopped for 
		## that predicate. Otherwise, all the translations of the predicate 
		## in the lexical translation table will be compared to the target 
		## predicates in their translation score order until a match is 
		## found. Finally, in case of no match found, the predicate None
		## will be paired with the source predicates in the returned list.
		## To avoid aligning a source predicate to a target predicate which
		## is already aligned to another source predicate, the target 
		## prediates are pools first and then they are removed from the 
		## pool once they are aligned.
		
		for vSrcProp in self.srcSRL.getPropositions():
			vSrcPred = vSrcProp.getPredicate()
			vflgAligned = False
			
			# 1. comparing labels
			for i, vTrgProp in enumerate(vlTrgProps):
				vTrgPred = vTrgProp.getPredicate()
				if vSrcPred.getLabel(pLevel = pPredLabelMatchLevel) == vTrgPred.getLabel(pLevel = pPredLabelMatchLevel):
					vlPropAlignment.append((vSrcProp, vTrgProp))
					vflgAligned = True
					del vlTrgProps[i]
					break
			
			# 2. looking for translation 
			if not vflgAligned:
				for vSrcPredTrans in self.ltt.getLTrans(vSrcPred.getForm()):
					for i, vTrgProp in enumerate(vlTrgProps):
						vTrgPred = vTrgProp.getPredicate()
						if vSrcPredTrans == vTrgPred.getForm():
							vlPropAlignment.append((vSrcProp, vTrgProp))
							vflgAligned = True
							del vlTrgProps[i]
							break
					
					if vflgAligned:
						break
			
			if not vflgAligned:
				vlPropAlignment.append((vSrcProp, None))
		
 		return vlPropAlignment
		
	
	
	def _alignArgs(self, pSrcProp, pTrgProp, pflgCollapseAdjunct = False):
		'''
		Finds aligned arguments between source and target propositions and
		returns list of tuples of source/target aligned arguments
		'''
		
		vlAlignedArgs = []
		
		## The method is to first pair source/target arguments having the
		## same role label. Note that adjunct labels can be collapsed. 
		## The translations of the filler of each source argument found 
		## in the lexical translation table is then compared to the filler
		## of its target pair. If a match is found, the arguments are added
		## to the list of aligned arguments.
		
		for vSrcArg in pSrcProp.getArguments():
			vflgAligned = False
			for vTrgArg in pTrgProp.getArguments():
				# 1. comparing labels
				if vSrcArg.getLabel(pflgCollapseAdjunct = pflgCollapseAdjunct) == vTrgArg.getLabel(pflgCollapseAdjunct = pflgCollapseAdjunct):
					# 2. comparing fillers
					for vSrcArgTrans in self.ltt.getLTrans(vSrcArg.getForm()):
						if vSrcArgTrans == vTrgArg.getForm():
							vlAlignedArgs.append((vSrcArg, vTrgArg))
							vflgAligned = True
							break
					
				if vflgAligned:
					break
		
 		return vlAlignedArgs
		
	
	


class PTPAM(PAM):
	'''
	Class for PAM QE metic using phrase table
	'''
	
	
	def __init__(self, pSrcCoNLLSentence, pTrgCoNLLSentence, pPTTable, pSrcConstTree, pTrgConstTree, pNoSrcPropScores = 1, pNoTrgPropScores = 0,  pOnSRLConstraintViolation = "exception", pPredicateType = "all"):
		'''
		Constructor
		'''
		
		PAM.__init__(self, 
		             pSrcCoNLLSentence.getSRL(pdAnnotationOptions = {"on-srl-constraint-violation": pOnSRLConstraintViolation, "gold-or-predicted": "pred", "predicate-type": pPredicateType}).convertToConst(pSrcConstTree, pConversionMethod = "elevate", pflgLoadIntoTree = True), 
		             pTrgCoNLLSentence.getSRL(pdAnnotationOptions = {"on-srl-constraint-violation": pOnSRLConstraintViolation, "gold-or-predicted": "pred"}).convertToConst(pTrgConstTree, pConversionMethod = "elevate", pflgLoadIntoTree = True),
		             pNoSrcPropScores = pNoSrcPropScores, 
		             pNoTrgPropScores = pNoTrgPropScores)
		
		self.srcCoNLLSentence = pSrcCoNLLSentence
		self.trgCoNLLSentence = pTrgCoNLLSentence
		self.ptt = pPTTable
		
	
	
	def score(self, pPredLabelMatchLevel = "frameset", pflgCollapseAdjunct = "False"):
		'''
		Scores the translation and returns the scores
		
		pPredLabelMatchLevel determines whether the frameset or only the
		lemma of the predicates should be compared in finding predicate
		match.
		'''
		
		# finding aligned predicates
		
		vlPropAlignment = self._alignPredicates(pPredLabelMatchLevel)
		vlAlignedProps = [pa for pa in vlPropAlignment if pa[1] != None]
		
		# finding aligned arguments of aligned predicates
		
		vlAlignedArgs = []
		for vSrcProp, vTrgProp in vlAlignedProps:
			vlAlignedArgs += self._alignArgs(vSrcProp, vTrgProp, pflgCollapseAdjunct = pflgCollapseAdjunct)
		
		# scoring
		
		vAlignedPredCount = len(vlAlignedProps)
		vAlignedArgCount = len(vlAlignedArgs)
		vdScores = {}
		if self.srcSRL.propCount == 0:
			vdScores["Precision"] = self.noSrcPropScores
			vdScores["Recall"] = self.noSrcPropScores
			vdScores["F1"] = self.noSrcPropScores
		elif self.trgSRL.propCount == 0:
			vdScores["Precision"] = self.noTrgPropScores
			vdScores["Recall"] = self.noTrgPropScores
			vdScores["F1"] = self.noTrgPropScores
		else:
			vdScores["Precision"] = 1.0 * (vAlignedPredCount + vAlignedArgCount) / (self.trgSRL.propCount + self.trgSRL.globalArgCount)
			vdScores["Recall"] = 1.0 * (vAlignedPredCount + vAlignedArgCount) / (self.srcSRL.propCount + self.srcSRL.globalArgCount)
			if vdScores["Precision"] + vdScores["Recall"] == 0:
				vdScores["F1"] = 0
			else:
				vdScores["F1"] = (2 * vdScores["Precision"] * vdScores["Recall"]) / (vdScores["Precision"] + vdScores["Recall"])
		
		self.scores = self._toDecimal(vdScores)
		
		return self.scores
		
	
	
	def _alignPredicates(self, pPredLabelMatchLevel = "frameset"):
		'''
		Finds the target predicate aligned to each source predicate and
		returns list of tuples the first element of which is the proposition
		of the source predicate and the second its aligned target proposition
		if found and None otherwise.
		'''
		
		vlTrgProps = self.trgSRL.getPropositions()[ : ]
		
		vlPropAlignment = []
		
		## The method is to first look at the predicate labels assuming 
		## that both sides have the same frameset labels. Note that the
		## label can be both frameset or lemma only. If there is a match
		## it is assumed as an alignment and the search is stopped for 
		## that predicate. Otherwise, all the translations of the predicate 
		## in the lexical translation table will be compared to the target 
		## predicates in their translation score order until a match is 
		## found. Finally, in case of no match found, the predicate None
		## will be paired with the source predicates in the returned list.
		## To avoid aligning a source predicate to a target predicate which
		## is already aligned to another source predicate, the target 
		## predicates are pooled first and then removed from the  pool once
		## they are aligned.
		
		for vSrcProp in self.srcSRL.getPropositions():
			vSrcPred = vSrcProp.getPredicate()
			vflgAligned = False
			
			# 1. comparing labels
			for i, vTrgProp in enumerate(vlTrgProps):
				vTrgPred = vTrgProp.getPredicate()
				if vSrcPred.getLabel(pLevel = pPredLabelMatchLevel) == vTrgPred.getLabel(pLevel = pPredLabelMatchLevel):
					vlPropAlignment.append((vSrcProp, vTrgProp))
					vflgAligned = True
					del vlTrgProps[i]
					break
			
			# 2. looking for translation 
			if not vflgAligned:
				for vSrcPredTrans in self.ptt.getPTrans(vSrcPred.getForm()):
					for i, vTrgProp in enumerate(vlTrgProps):
						vTrgPred = vTrgProp.getPredicate()
						if vSrcPredTrans == vTrgPred.getForm():
							vlPropAlignment.append((vSrcProp, vTrgProp))
							vflgAligned = True
							del vlTrgProps[i]
							break
					
					if vflgAligned:
						break
			
			if not vflgAligned:
				vlPropAlignment.append((vSrcProp, None))
		
 		return vlPropAlignment
		
	
	
	def _alignArgs(self, pSrcProp, pTrgProp, pflgCollapseAdjunct = False):
		'''
		Finds aligned arguments between source and target propositions and
		returns list of tuples of source/target aligned arguments
		'''
		
		vlAlignedArgs = []
		
		## The method is to first pair source/target arguments having the
		## same role label. Note that adjunct labels can be collapsed. 
		## The translations of the filler of each source argument found 
		## in the lexical translation table is then compared to the filler
		## of its target pair. If a match is found, the arguments are added
		## to the list of aligned arguments.
		
		for vSrcArg in pSrcProp.getArguments():
			vflgAligned = False
			for vTrgArg in pTrgProp.getArguments():
				# 1. comparing labels
				if vSrcArg.getLabel(pflgCollapseAdjunct = pflgCollapseAdjunct) == vTrgArg.getLabel(pflgCollapseAdjunct = pflgCollapseAdjunct):
					# 2. comparing fillers
					for vSrcArgTrans in self.ptt.getPTrans(vSrcArg.getForm()):
						if vSrcArgTrans == vTrgArg.getForm():
							vlAlignedArgs.append((vSrcArg, vTrgArg))
							vflgAligned = True
							break
					
				if vflgAligned:
					break
		
 		return vlAlignedArgs
		
	
	


class PAMAnalyze():
	'''
	Class for analyzing PAM scores
	'''
	
	def __init__(self, plPAMs):
		'''
		Constructor
		'''
		
		self.PAMs = plPAMs
		self.dataSize = len(self.PAMs)
		
		## Reference scores are stored in a dictionary with its keys as
		## score name and values as list of scores. This makes it customizable 
		## for the user.
		self.refScores = {}
		
		# dictionary to store various statistics
		self.stat = {}
		
		# list of manual analysis results
		self.manalysis = []
		
		# dictionary to store manual analysis statistics
		self.manalStat = {}
		
	
	
	def addRefScores(self, pScoreName, pScores):
		'''
		Adds a set of reference scores to the reference scores dictionary
		
		Score are given in a  list
		'''
		
		# sanity check
		vRefScoreLen = len(pScores)
		if vRefScoreLen != self.dataSize:
			raise Exception("PAM and reference score set length mismatch: %s vs. %s!" % (self.dataSize, vRefScoreLen))
		
		## Float scores are converted to decimal for the later needs for
		## exact values.
		if pScoreName not in self.refScores:
			self.refScores[pScoreName] = [Decimal(str(s)) for s in pScores]
		
	
	
	def getDiffs(self, pPAMScoreName, pRefScoreName, pflgSorted = False):
		'''
		Generates and returns the differences of the given PAM and reference
		scores in a list
		
		It sorts the difference values based on their absolute value if 
		pflgSorted is set to true. In this case the list elements are
		tuples the first element of which is the position of score in the 
		original list and the second the difference value. 
		'''
		
		vlDiffs = []
		for vPAM, vRefScore in zip(self.PAMs, self.refScores[pRefScoreName]):
			vlDiffs.append(Decimal(str(vPAM.scores[pPAMScoreName])) - vRefScore)
		
		if pflgSorted:
			return sorted([(i, d) for i, d in enumerate(vlDiffs, start = 1)], key = lambda x: abs(x[1]))
		else:
			return vlDiffs
		
	
	
	def getDiffGroups(self, pPAMScoreName, pRefScoreName, pGroupDiffRange = 0, pflgSeparateExactMathes = False):
		'''
		Groups the differences of the given PAM and reference scores based 
		on a difference range and returns them in a dictionary with keys
		as group values and values as dictionaries, in turn, containing
		members of each group and their positions in the score list.
		
		It is similar to SQL group by.
		
		Group difference range value of, say, 0.1 means that difference 
		values in (0, 0.1], (0.1, 0.2], (0.2, 0.3] and so on  will be grouped
		together. The inclusion of the 0 depends on the value of pflgSeparateExactMathes;
		If exact matches are requested to be a separate group, then the
		first group is (0, 0.1], [0, 0.1] otherwise. A value of 0 means 
		that the difference values are grouped individually.
		
		The keys of the returned dictionary are the upper bounds of the
		ranges, or the difference values themselves if the range of 0 is 
		chosen.
		'''
		
		vGroupDiffRange = Decimal(str(pGroupDiffRange))
		
		# sanity check 
		if 0 > vGroupDiffRange > 1:
			raise Exception("Difference range %s is not valid!" % vGroupDiffRange)
		
		vlDiffs = self.getDiffs(pPAMScoreName = pPAMScoreName, pRefScoreName = pRefScoreName)
		
		vdGroups = {}
		for vPos, vDiff in enumerate(vlDiffs, start = 1):
			# we use absolute value of difference for group calculation
			vABSDiff = abs(vDiff)
			
			# in both cases the difference value is the key of the group
			if vGroupDiffRange == 0 or (pflgSeparateExactMathes and vDiff == 0):
				vKey = vABSDiff
			else:
				vDiv = int(vABSDiff / vGroupDiffRange)
				vMod = vABSDiff % vGroupDiffRange
				
				if vMod == 0 and vDiv != 0:
					vKey = vDiv * vGroupDiffRange
				else:
					vKey = (vDiv + 1) * vGroupDiffRange
			
			if vKey in vdGroups:
				vdGroups[vKey]["members"].append(vDiff)
				vdGroups[vKey]["positions"].append(vPos)
			else:
				vdGroups[vKey] = {}
				vdGroups[vKey]["members"] = [vDiff]
				vdGroups[vKey]["positions"] = [vPos]
		
		return vdGroups
		
	
	
	def getDiffGroupsExt(self, pPAMScoreName, pRefScoreName, pGroupDiffRange = 0, pflgSeparateExactMathes = False):
		'''
		Extended version of getDiffGroups which also returns the size and
		ratio of the group sizes to the whole data set size, the average
		of difference values in the group, etc.
		'''
		
		vdOrgGroups = self.getDiffGroups(pPAMScoreName = pPAMScoreName, pRefScoreName = pRefScoreName, pGroupDiffRange = pGroupDiffRange, pflgSeparateExactMathes = pflgSeparateExactMathes)
		
		vdExtGroups = {}
		
		for vKey in vdOrgGroups.iterkeys():
			vdExtGroups[vKey] = {}
			vdExtGroups[vKey]["members"] = vdOrgGroups[vKey]["members"]
			vdExtGroups[vKey]["positions"] = vdOrgGroups[vKey]["positions"]
			vdExtGroups[vKey]["size"] = len(vdOrgGroups[vKey]["members"])
			vdExtGroups[vKey]["ratio"] = 1.0 * vdExtGroups[vKey]["size"] / self.dataSize
			vdExtGroups[vKey]["average"] = sum(vdOrgGroups[vKey]["members"]) / vdExtGroups[vKey]["size"]
		
		return vdExtGroups
		
	
	
	def evalPAMScore(self, pPAMScoreName, pRefScoreName, pScorePositions, pAEMMAccEM = 0.05):
		'''
		Evaluates a selected subset of a given PAM score against a given
		reference scores.
		
		Subset selection is made by pScorePositions which is a list of 
		positions of the scores in the subset.
		'''
		
		vlPreds = [float(self.PAMs[p - 1].scores[pPAMScoreName]) for p in pScorePositions]
		vlRef = [float(self.refScores[pRefScoreName][p - 1]) for p in pScorePositions]
		
		vSampleLen = len(vlRef)
		
		vdEvalScores = {}
		vPredEval = util.PredEval(plReference = vlRef)
		vdEvalScores["AEMM"] = round(vPredEval.acc(plPrediction = vlPreds, pType = "abs", pErrMargin = pAEMMAccEM), 2)
		vdEvalScores["MAE"] = round(vPredEval.mae(plPrediction = vlPreds), 4)
		vdEvalScores["RMSE"] = round(vPredEval.rmse(plPrediction = vlPreds), 4)
		vdEvalScores["r"] = round(vPredEval.r(plPrediction = vlPreds), 4)
		
		return vdEvalScores
		
	
	
	def extractStat(self):
		'''
		Extract various statistics
		'''
		
		_vPAMs = [i for i, pam in enumerate(self.PAMs, start = 1) if pam.srcSRL.propCount == 0]
		_vLen = len(_vPAMs)
		self.stat["Source sentences without verb predicates"] = (_vLen, _vLen * 100.0 / self.dataSize, _vPAMs)
		_vPAMs = [i for i, pam in enumerate(self.PAMs, start = 1) if pam.trgSRL.propCount == 0]
		_vLen = len(_vPAMs)
		self.stat["Target sentences without verb predicates"] = (_vLen, _vLen * 100.0 / self.dataSize, _vPAMs)
		_vPAMs = [i for i, pam in enumerate(self.PAMs, start = 1) if pam.srcSRL.propCount == pam.trgSRL.propCount == 0]
		_vLen = len(_vPAMs)
		self.stat["Sentence pairs without verb predicates"] = (_vLen, _vLen * 100.0 / self.dataSize, _vPAMs)
		 
	
	
	def extractAnalysis(self, pAnalysisFilename):
		'''
		Extracts manual analysis results from the CSV file
		
		Note that it overwrites current statistics in self.manalysis and 
		self.manalStat.
		'''
		
		vfAnalysis = open(pAnalysisFilename)
		
		self.manalysis = []
		vdAnalysisRec = {"Alignment Problem": 0, "Source PI Problem": 0, "Source PL Problem": 0, "Source AI Problem": 0, "Source AL Problem": 0,
		                                        "Target PI Problem": 0, "Target PL Problem": 0, "Target AI Problem": 0, "Target AL Problem": 0,
		                 "Translation Divergence": 0, "Evaluation Problem": 0,
		                 "Source SRL PI Error": 0, "Source Sentence Structure": 0 , "Source UGC Error": 0, "Preprocessing Problem": 0,
		                 "Target SRL PI Error": 0, "Target Sentence Structure": 0 , "Untranslated UGC Error": 0, "Preprocessing Problem": 0, "Translation Error": 0}
		self.manalStat = {"Alignment Problems": 0, "Source PI Problems": 0, "Source PL Problems": 0, "Source AI Problems": 0, "Source AL Problems": 0,
		                                        "Target PI Problems": 0, "Target PL Problems": 0, "Target AI Problems": 0, "Target AL Problems": 0,
		                  "Translation Divergences": 0, "Evaluation Problems": 0,
		                  "Source SRL PI Errors": 0, "Source Sentence Structures": 0 , "Source UGC Errors": 0, "Preprocessing Problems": 0,
		                  "Target SRL PI Errors": 0, "Target Sentence Structures": 0 , "Untranslated UGC Errors": 0, "Preprocessing Problems": 0, "Translation Errors": 0}
		
		for vLine in vfAnalysis:
			vlLineSplit = vLine.split('\t')
			
			if re.match("^\d+\t", vLine):
				vNo = int(vlLineSplit[0])
			elif vLine.lower().startswith("alignment problem"):
				vdAnalysisRec["Alignment Problem"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("source pi problem"):
				vdAnalysisRec["Source PI Problem"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("source pl problem"):
				vdAnalysisRec["Source PL Problem"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("source ai problem"):
				vdAnalysisRec["Source AI Problem"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("source al problem"):
				vdAnalysisRec["Source AL Problem"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("target pi problem"):
				vdAnalysisRec["Target PI Problem"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("target pl problem"):
				vdAnalysisRec["Target PL Problem"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("target ai problem"):
				vdAnalysisRec["Target AI Problem"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("target al problem"):
				vdAnalysisRec["Target AL Problem"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("translation divergence"):
				vdAnalysisRec["Translation Divergence"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("evaluation problem"):
				vdAnalysisRec["Evaluation Problem"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("comments"):
				vdAnalysisRec["Comments"] = vlLineSplit[1]
			elif vLine.lower().startswith("source srl pi error"): 
				vdAnalysisRec["Source SRL PI Error"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("source sentence structure"): 
				vdAnalysisRec["Source Sentence Structure"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("source ugc error"): 
				vdAnalysisRec["Source UGC Error"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("preprocessing problem"): 
				vdAnalysisRec["Preprocessing Problem"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("target srl pi error"): 
				vdAnalysisRec["Target SRL PI Error"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("target sentence structure"): 
				vdAnalysisRec["Target Sentence Structure"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("untranslated ugc error"): 
				vdAnalysisRec["Untranslated UGC Error"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			elif vLine.lower().startswith("translation error"): 
				vdAnalysisRec["Translation Error"] = int(vlLineSplit[1]) if vlLineSplit[1] != '' else 0
			
			# end of block signals
			if vLine.lower().startswith("comments") or vLine.lower().startswith("source ugc error") or vLine.lower().startswith("translation error"): 
				
				# calculating statistics
				
				self.manalStat["Alignment Problems"] += vdAnalysisRec["Alignment Problem"]
				self.manalStat["Source PI Problems"] += vdAnalysisRec["Source PI Problem"]
				self.manalStat["Source PL Problems"] += vdAnalysisRec["Source PL Problem"]
				self.manalStat["Source AI Problems"] += vdAnalysisRec["Source AI Problem"]
				self.manalStat["Source AL Problems"] += vdAnalysisRec["Source AL Problem"]
				self.manalStat["Target PI Problems"] += vdAnalysisRec["Target PI Problem"]
				self.manalStat["Target PL Problems"] += vdAnalysisRec["Target PL Problem"]
				self.manalStat["Target AI Problems"] += vdAnalysisRec["Target AI Problem"]
				self.manalStat["Target AL Problems"] += vdAnalysisRec["Target AL Problem"]
				self.manalStat["Translation Divergences"] += vdAnalysisRec["Translation Divergence"]
				self.manalStat["Evaluation Problems"] += vdAnalysisRec["Evaluation Problem"]
				
				self.manalStat["Source SRL PI Errors"] += vdAnalysisRec["Source SRL PI Error"]
				self.manalStat["Source Sentence Structures"] += vdAnalysisRec["Source Sentence Structure"]
				self.manalStat["Source UGC Errors"] += vdAnalysisRec["Source UGC Error"]
				self.manalStat["Preprocessing Problems"] += vdAnalysisRec["Preprocessing Problem"]
				self.manalStat["Target SRL PI Errors"] += vdAnalysisRec["Target SRL PI Error"]
				self.manalStat["Target Sentence Structures"] += vdAnalysisRec["Target Sentence Structure"]
				self.manalStat["Untranslated UGC Errors"] += vdAnalysisRec["Untranslated UGC Error"]
				self.manalStat["Translation Errors"] += vdAnalysisRec["Translation Error"]
				
				self.manalysis.append((vNo, vdAnalysisRec))
				for vKey in vdAnalysisRec.iterkeys():
					vdAnalysisRec[vKey] = 0
		
		
	
	
