#! /usr/bin/python
# -*- coding: utf-8 -*-

"""	
	This module defines classes to for Aspect Based Sentiment Analysis (ABSA).
	
	NOTE: bugs identified by PyCharm should be fixed (e.g. _attachATSuffix() 
	      in ABSADNode). It seems they have never been tested. 
	
	
	Version 3.0											   (30-Apr-2018)
	- getTokens() and getOverallPolarity() are added to ABSAContext.
	
	Version 2.9											   (03-Jan-2018 to 05-Jan-2017)
	- mergeWith()is added to ABSASet.
	- pickle() and loadFromPickle() are added to ABSASet.
	- getOEsIO() is added to AspectTerm.
	
	Version 2.8											   (17-Jul-2017 to 18-Jul-2017)
	- ABSAContext is added to support SemEval 2016 data. The code is refactored so that
	  it still works seamlessly with 2014 data. 
	
	Version 2.7											   (17-May-2017)
	- Unary tree kernels in ABSANGramKTree is upgraded.
	
	Version 2.6											   (24-Mar-2017)
	- AspectTerm.getSentenceMarking() is added. 
	
	Version 2.5											   (23-Jan-2017)
	- Bug in ABSASet.loadBratOE() fixed to account for cases where SE is annotated
	  from/to the middle of a token due to tokenization problem.
	- ABSASet.oeToBIO() is upgraded to include POS tags in the output.
	
	Version 2.4								(01-Dec-2016 to 08-Dec-2016)
	- oeToBIO() is added to ABSASent.
	
	Version 2.3								(11-Oct-2016 to 13-Oct-2016)
	- loadBratOEs() is added to ABSASent to load opinion/sentiment expressions
	  annotated in Brat annotator.
	- oes is added to AspectTemrm to keep opinion/sentiment expressions.
	- vABSASent.oe is renamed to oes.
	- Type is added to OpinionExpression.
	
	Version 2.2								(22-Jun-2016)
	- writeBratInput()is added to ABSASet.
	
	Version 2.1								(02-May-2016 to 05-May-2016)
	- ABSANGramKTree is upgraded to use customized unigram auxiliary nodes
	  instead of X for binary format.
	- extractAT2RootPathTree() is added to ABSADTree.
	- ABSANGramKTree is upgraded to produce lower case word tree kernels.
	- New format is added to ABSANGramKTree kernels for word vector similarity.
	
	Version 2.0								(12-Apr-2016 to 26-Apr-2016)
	- ABSANGramKTree is upgraded to produce POS tag and sentiment score 
	  tree kernels similar to word tree kernels.
	- getPOSTags() is added to ABSASent.
	- New formats are added to ABSADKTree and ABSACKTree.
	- getVocabulary() and loadWordVectors() are added to the ABSASet.
	- getWordVectors() is added to ABSASent.
	- A new attribute is added to ABSASent to store the ABSASet it belongs to.
	  Its value is set during loading the data.
	
	Version 1.9								(04-Apr-2016 to 11-Apr-2016)
	- New formats are added to ABSANGramKTree and getNGramKTreeEmbeding()
	  of AspectTerm is modified to handle various n-gram tree formats.
	- getPOSTags() is added to ABSASent.
	
	Version 1.8								(10-Mar-2016 to 29-Mar-2016)
	- extractAT2OEPath() of ABSADepTree is renamed to  extractAT2OEDepRelPath().
	- getPolarScores() is added to ABSASent.
	
	Version 1.7								(17-Feb-2016 to 09-Mar-2016)
	- Constituency and dependency path between aspect term and opinion 
	  expression is extracted.
	- Bug in ABSADNode.loadFromDepNode() is fixed which assigned DepTree
	  object instead of ABSADepTree to depTree attribute.
	- Average sentiment score of the sentence and opinion expressions can
	  be extracted using newly added methods to ABSASent and OpinionExpression.
	- getTokens() is added to OpinionExpression and AspectTerm.
	- The aspect length groups are now extracted as part of the aspect term
	  statistics by ABSASet.extractATStat().
	
	Version 1.6								(28-Jan-2016 to 02-Feb-2016)
	- Opinion expression is introduced and OpinionExpression is added.
	- loadBIOOpinionExpressions() is added to ABSASet and ABSASent.
	- Methods are added to embed opinion expressions in the trees.
	- ABSADNode.decorate() is renamed to ABSADNode.decorateAT(). 
	- New formats are added to ABSADKTree.
	
	Version 1.5								(19-Nov-2015 to 26-Nov-2015)
	- Naming of internal methods is ABSACKTree and ABSADKTree are modified
	  to fix the bug causing conflict between the methods of these classes
	  and their parent classes.
	- New formats are added to ABSACKTree.
	- shallowCopy() is added to ABSAConstTree.
	- deep copying method the trees in tree embedding generations in aspec 
	  term class is changed to use python deepcopy as it performed better
	  with ConstTree in an experiment. However, it has not been tested for
	  the dependency tree despite applying the idea. In both cases, this 
	  still needs to be further confirmed in the future experiments.
	
	Version 1.4								(13-Nov-2015 to 16-Nov-2015)
	- New format is added to ABSADKTree to handle semantic roles.
	
	Version 1.3								(22-Oct-2015 to 02-Nov-2015)
	- ABSACTree.loadSentimentScore() is changed and ABSACNode.loadSentScores()
	  is added to enable optionally propagating sentiment scores in the 
	  tree nodes.
	- pNeutralScore is added to loadSentimentScores in ABSASet and ABSASent
	  to assign a score for neutral words other than None.
	- ABSACKTree is added with two new formats.
	- ABSASent.getSentimentScores() is changed to return the scores in a 
	  list instead of a dictionary.
	
	Version 1.2								(12-Oct-2015 to 15-Oct-2015)
	- ABSADKTree.generateDepKTree() supports subtrees under a node provided
	  as a parameter in addition to the whole tree under the root.
	- ABSACDKTree is added to support the integration of constituency and 
	  dependency trees.
	- ABSANGramKTree is added to construct a tree from text token forms.
	- getNGramKTreeEmbeding() is added to AspectTerm.
	- extractCaraFXInput() is added to ABSASent.
	
	Version 1.1								(28-Sep-2015 to 02-Oct-2015)
	- getSentimentScores(), length() are added to ABSASent.
	- tokenLength() and getSentimentScores() are added to ABSASet.
	- New formats are added to generateDepKTree() of ABSADKTree.
	- Sentiment score representation is added to constituency tree decoration.
	- loadSSInCTree() is added to ABSASet and ABSASent.
	
	Version 1.0								(03-Sep-2015 to 08-Sep-2015)
	- loadSentimenScores() is added to ABSASet and ABSASent.
	- sentScores is added to ABSASent to represent sentiment scores of the
	  sentence words.
	- sentScore is added to ABSADNode to represent sentiment score of the
	  associated node.
	- loadSSInDTrees() is added to ABSASet to load sentiment scores into
	  the dependency trees.
	- loadSSInDTree() is added to ABSASent to load sentiment scores into
	  the dependency tree of the sentence.
	- loadSentScores() is added to ABSADTree.
	- setSentScore() is added to ABSADNode.
	- deepCopy() is added to ABSADNode and the constructor and node creation
	  methods were accordingly changed.
	
	Version 0.9								(05-Aug-2015 to 26-Aug-2015)
	- ABSADTree and ABSADNode are added, and all the corresponding updates
	  are done in ABSASet, ABSASent, AspectTerm.
	
	Version 0.8								(30-Jul-2015 to 04-Aug-2015)
	- The representation of aspect terms in the constituency tree is decoupled
	  from embedding them in the tree. Once an aspect term is embedded in
	  the tree (either in pre-terminals in its span or in the constituent
	  node mapped to it), it can be represented in the tree by inserting
	  a new node, attaching suffix and attaching polarity in various formats.
	  This is done in ABSACNode.decorateAT(). AspectTerm.getCTreeEmbeding()
	  and ABSACNode.embedAspectTerm() were changed accordingly. Also,
	  ABSACNode.getPTBFormat() is not overridden any more since polarity 
	  attachment is done in decorateAT(). Finally, inserting node during 
	  embedding aspect terms only happens optionally when there is a mismatch.
	  This has been implemented by changing pInsertATNode to pOnMismatch
	  which specifies the method for handling aspect term/constituent 
	  mismatch, on of which being inserting new AT node.
	
	Version 0.7								(27-Jul-2015 to 29-Jul-2015)
	- Another option (pAttachATSuffix) is added to embedding aspect term 
	  in constituency tree to mark nodes in the aspect term subtree with 
	  a specific suffix tag (_AT). 
	- ConstNode.embedAspectTerm() is edited to support embedding the aspect 
	  term in pre-terminals in addition to constituent nodes spanning the
	  aspect term. _embedATINSpanConst() and _embedATINPreterminals() are 
	  added to ConstNode for this purpose and extractAspectTerms() is edited
	  to account for repetitions stemming from embedding aspect terms in
	  pre-terminals.
	- ABSADTree and ABSADNode are implemented.
	
	Version 0.6								(22-Jun-2015 to 23-Jun-2015)
	- aspectTerm in ABSACNode is changed to aspectTerms which is a list of
	  aspect terms ambeded into the node to allow a node to carry more than
	  one aspect terms. The methods handling this attribute have consequently
	  been updated.
	- extractATsInCTree() is added to ABSASent.
	- getSentencesWithNoAT() is added to ABSASet.
	- ABSASet.extractATStat() is update to extract more statistics.
	
	Version 0.5													18-May-2015
	- ABSACNode.embedAspectTerm() is edited to better handle the situation 
	  where pInsertATNode is set to mismatch. ABSACNode._plugAspectTerm
	  has accordingly been updated.
	- getAspectTermCount() is added to ABSASet and ABSASent.
	
	Version 0.4													08-May-2015
	- loadConstTrees() in ABSASet and ABSASent are changed not to embed 
	  the aspect terms. Embeding aspect terms is done instead by new methods
	  ABSASet.embedATsInCTrees() and ABSASent.embedATsInCTree().
	- embedAspectTerm() is added to ABSASent to be able to embed individual
	  aspect terms in the tree.
	- sentence (attr), embedInCtree(), getCTreeEmbeding() are added to 
	  AspectTerm.
	- The default value for pflgAttachATPolarity argument of getPTBFormat()
	  method in ABSACTree and ABSACNode is changed to False. 
	
	Version 0.3													06-May-2015
	- Embedding aspect terms is changed. Accordingly, ABSACNode.embedAspectTerms()
	  is renamed to embedAspectTerm().
	- addSentences() is added to ABSASet.
	
	Version 0.2													18-Feb-2015
	- The NP structures in ABSACTree can be modified to avoid term/constituent
	  mismatch.
	
	Version 0.1													17-Feb-2015
	- ABSA, ABSASent, AspectTerm, ABSACTree, ABSACNode are added.
	
"""


from collections import namedtuple
from parse import constparse, depparse, pos
from ml import tk
from utils import util
import sentexicon
import re, copy
import pickle


class ABSASet:
	'''
	Class for aspect-based sentiment analysis data set
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		self.contexts = []
		
		# WordVector object containing the word vectors of the vocabulary in the dataset
		self.wv = None
		
	
	
	@property
	def size(self):
		'''
		Returns the size of the data set which is the number of its sentences
		'''
		
		return len(self.getSentences())
		
	
	
	def getContextCount(self):
		'''
		Returns the number of contexts i the data set
		'''
		
		return len(self.contexts)
		
	
	
	@property
	def tokenLength(self):
		'''
		Returns the number of tokens in the data set
		'''
		
		return sum([s.length for s in self.getSentences()])
		
	
	
	def getContexts(self):
		'''
		Returns ABSA contexts of the data set 
		'''
		
		return self.contexts
		
	
	
	def getSentences(self, pSort = ''):
		'''
		Returns the ABSA sentences
		
		The sort options are:
		- None: in document order
		- text: in sentence text order
		'''
		
		vlSentences = [s for c in self.getContexts() for s in c.getSentences()]
		
		if pSort.lower() == "text":
			return [s for s in sorted(vlSentences, key = lambda x: x.getText())]
		else:
			return vlSentences
		
	
	
	def getSentencesWithNoAT(self):
		'''
		Returns the ABSA sentences which do not have any aspect terms
		'''
		
		return [s for s in self.getSentences() if s.getAspectTermCount() == 0]
		
	
	
	def addContext(self, pContext):
		'''
		Adds ABSA context to the data set 
		'''
		
		if pContext not in self.getContexts():
			pContext.dataset = self
			self.contexts.append(pContext)
		
	
	
	def addContexts(self, plContexts):
		'''
		Adds ABSA contexts to existing contexts
		'''
		
		for vContext in plContexts:
			self.addContext(vContext)
		
	
	
	def delContext(self, pContextIdx):
		'''
		Deletes context at the given index from the dataset 
		'''
		
		del self.contexts[pContextIdx]
		
	
	
	def mergeWith(self, pABSASet):
		'''
		Merges this data set with the given ABSA data set 
		
		It first checks if both sets to be merged are of the same class derived from ABSASet) 
		'''
		
		if self.__class__ != pABSASet.__class__:
			raise Exception("Two data sets must be of the same ABSA type: %s vs. %s" %(self.__class__ , pABSASet.__class__))
		
		# merging contexts
		self.addContexts(pABSASet.getContexts())
		
		# NOTE: word embeddings should also be merged
		print("Word embeddings were not merged. Reload them for the merged data set.")
		
		# NOTE: take care of other newly added attributes if any
		
	
	
	def addSentences(self, plSentences):
		'''
		Adds ABSA sentences to existing sentences
		
		Currently, no care is taken regarding ID duplication.
		'''
		
		for vSent in plSentences:
			if vSent.context not in self.getContexts():
				self.addContext(vSent.context)
			
			vSent.context.addSentence(vSent)
		
	
	
	def getVocabulary(self):
		'''
		Extracts and returns the vocabulary of the dataset 
		'''
		
		return sorted(set([t for s in self.getSentences() for t in s.getTokens()])) 
		
	
	
	def getAspectTerms(self):
		'''
		Returns all aspect terms objects
		'''
		
		vlAspectTerms = []
		
		for vSent in self.getSentences():
			vlAspectTerms += vSent.getAspectTerms()
		
		return vlAspectTerms
		
	
	
	def getAspectTermCount(self):
		'''
		Returns the total number of aspect terms in the dataset
		'''
		
		vTotalCount = 0
		
		for vSent in self.getSentences():
			vTotalCount += vSent.getAspectTermCount()
		
		return vTotalCount
		
	
	
	def extractSentenceForms(self, pSort = None):
		'''
		Returns the surface form of the sentences
		
		The sort options are:
		- None: in document order
		- id:   in sentence ID order
		- text: in sentence text order
		- at: in aspect term order
		'''
		
		if pSort.lower() == "at":
			return [at.sentence.getText() for at in self.getAspectTerms()]
		else:
			return [s.getText() for s in self.getSentences(pSort = pSort)]
		
	
	
	def extractAspectTerms(self):
		'''
		Extracts all aspect terms (forms) in the data and their counts
		'''
		
		vlAspectTerms = []
		
		for vSent in self.getSentences():
			vlAspectTerms += vSent.getAspectTermForms()
		
		return util.groupBy(vlAspectTerms)
		
	
	
	def extractATStat(self):
		'''
		Extract various aspect term statistics
		'''
		
		# data structure to store statistics
		
		ATStat = namedtuple('ATStat', "sentsWithAnAT, sentsWithNoAT, allATCount distATCount hapaxATCount polCounts CTreeATCount, ATConstMismatch, ATLenGroups")
		
		vSentsWithNoAT = len(self.getSentencesWithNoAT())
		vSentsWithAnAT = self.size - vSentsWithNoAT
		
		vdATs = self.extractAspectTerms()
		
		# allATCount: number of all aspect terms
		vAllATCount = sum([at for at in vdATs.itervalues()])
		
		# distATCount: number of distinct aspect terms
		vDistATCount = len(vdATs)
		
		# hapaxATCount: number of hapax aspect terms (terms appearing just once)
		vHapaxATCount = sum([at for at in vdATs.itervalues() if at == 1])
		
		# polCounts: number of each polarity type
		vdPolCounts = {"positive": 0, "negative": 0, "neutral": 0, "conflict": 0}
		for vAT in self.getAspectTerms():
			vdPolCounts[vAT.getPolarity()] += 1
		
		# CTreeATCount: number of aspect terms transferred to the contituency tree
		vCTreeATCount = sum([len(s.extractATsInCTree()) for s in self.getSentences()])
		
		# ATConstMismatch: number of mismatched aspect terms and constituents
		vATConstMismatch = sum([len(s.extractATermConstMismatch()) for s in self.getSentences()])
		
		# ATLenGroups: grouped AT token lengths in a dictionary
		vdATLenGroups = util.groupBy([len(at.getTokens()) for at in self.getAspectTerms()])
		
		return ATStat(vSentsWithAnAT, vSentsWithNoAT, vAllATCount, vDistATCount, vHapaxATCount, vdPolCounts, vCTreeATCount, vATConstMismatch, vdATLenGroups)
		
	
	
	def loadConstTrees(self, plConstTrees):
		'''
		Loads the constituency parse trees of the sentences
		
		It assumes that the provided constituency trees are in the order
		in which the sentences are loaded.
		
		The constituency trees can be provided in bracketing format or as
		constparse.ConstTree objects (in a list).
		'''
		
		for vSent, pCTree in zip(self.getSentences(), plConstTrees):
			vSent.loadConstTree(pCTree)
		
	
	
	def loadDepTrees(self, plDepTrees):
		'''
		Loads the dependency parse trees of the sentences
		
		It assumes that the provided dependency trees are in the order in
		which the sentences are loaded.
		
		The dependency trees are assumed to be provided a list of 
		depparse.DepTree objects.
		'''
		
		for vSent, pDTree in zip(self.getSentences(), plDepTrees):
			vSent.loadDepTree(pDTree)
		
	
	
	def loadPOSTaggings(self, pPOSTaggingFilename, pSort = ''):
		'''
		Loads POS taggings of the sentences in the dataset
		 
		POS tagging file should be in a columnar format and the order of the sentences should be given in pSort.
		'''
		
		for vSent, vSentPOS in zip(self.getSentences(pSort = pSort), open(pPOSTaggingFilename).read().strip().split('\n\n')):
			vPOSTagging = pos.POSTagging()
			vPOSTagging.loadFromColumnar(vSentPOS)
			
			if vPOSTagging.length != vSent.length:
				raise Exception("Length of sentence and POS tags don't match:\n\n%s\n%s" % (vSent.getTokens(), vPOSTagging.toLorgInput()))
			else:
				vSent.loadPOSTagging(vPOSTagging)
		
	
	
	def loadBIOOpinionExpressions(self, pllBIO):
		'''
		Loads opinion expression annotations based on BIO tagging
		
		The BIO tagging is provided in a 2D list the first dimension of which is supposed to match sentences in the 
		dataset and the second the BIO labels of the tokens in each sentence.
		'''
		
		for vSent, vlOEBIO in zip(self.getSentences(), pllBIO):
			vSent.loadBIOOpinionExpressions(vlOEBIO)
		
	
	
	def embedATsInCTrees(self, pflgModNPStruct = False, pEmbedPosition = "spanning-constituent", pOnMismatch = 'nothing', pflgExtendATSpanToDT = False):
		'''
		Embeds the aspect terms in the constituency trees
		
		pflgModNPStruct specifies whether the NP structures should be edited
		in order to avoid term/constituent mismatch.
		
		pEmbedPosition specifies where the aspect term should be embedded
		in the subtree. The possible values are:
		- (span)ning-constituent: embedding into a node in subtree which 
		                          spans the aspect term tokens. This may 
		                          cause mismatches where the aspect term 
		                          span is not fully covered by the node
		                          span. pOnMismatch can be set to fix the
		                          issue.
		- (pre)-terminals: embedding into each and every pre-terminal node
		                   in the subtree falling in the aspect term span.
		
		pOnMismatch specifies the method for handling aspect term/constituent
		mismatch. The possible values are:
		- nothing (or none): do not handle the mismatches
		- node (or insert): insert a new node covering the aspect term span
		'''
		
		for vSent in self.getSentences():
			vSent.embedATsInCTree(pflgModNPStruct = pflgModNPStruct, pEmbedPosition = pEmbedPosition, pOnMismatch = pOnMismatch, pflgExtendATSpanToDT = pflgExtendATSpanToDT)
		
	
	
	def embedOEsInCTrees(self, pflgModNPStruct = False, pEmbedPosition = "spanning-constituent", pOnMismatch = 'nothing'):
		'''
		Embeds the opinion expressions in the constituency trees
		
		pflgModNPStruct specifies whether the NP structures should be edited in order to avoid expression/constituent 
		mismatch.
		
		pEmbedPosition specifies where the opinion expression should be	embedded in the subtree. The possible values are:
		- (span)ning-constituent: embedding into a node in subtree which spans the OE tokens. This may cause mismatches 
		                          where the OE span is not fully covered by the node span. pOnMismatch can be set to fix
		                          the issue.
		- (pre)-terminals: embedding into each and every pre-terminal node in the subtree falling in the OE span.
		
		pOnMismatch specifies the method for handling expression/constituent mismatch. The possible values are:
		- nothing (or none): do not handle the mismatches
		- node (or insert): insert a new node covering the aspect term span
		'''
		
		for vSent in self.getSentences():
			vSent.embedOEsInCTree(pflgModNPStruct = pflgModNPStruct, pEmbedPosition = pEmbedPosition, pOnMismatch = pOnMismatch)
		
	
	
	def embedATsInDTrees(self):
		'''
		Embeds the aspect terms in the dependency trees
		'''
		
		for vSent in self.getSentences():
			vSent.embedATsInDTree()
		
	
	
	def embedOEsInDTrees(self):
		'''
		Embeds the opinion expressions in the dependency trees
		'''
		
		for vSent in self.getSentences():
			vSent.embedOEsInDTree()
		
	
	
	def loadSentimentScores(self, pSentexicon, pNeutralScore = None):
		'''
		Loads polarity scores to sentences from a sentiment lexicon which
		is a Sentexicon object
		
		For details about Sentexicon object, see sentexicon.py
		'''
		
		vTotalWordNum = 0          # total number of words in the data set
		vTotalEntryWordNum = 0     # number of words found in the lexicon
		
		for vSent in self.getSentences():
			vWordNum, vEntryWordNum = vSent.loadSentimentScores(pSentexicon, pNeutralScore)
			
			vTotalWordNum += vWordNum
			vTotalEntryWordNum += vEntryWordNum
		
		return vTotalWordNum, vTotalEntryWordNum
		
	
	
	def getSentimentScores(self):
		'''
		Returns a dictionary of words in the data set and the sentiment 
		scores attached to them 
		'''
		
		vdResult = {}
		
		for vSent in self.getSentences():
			for vWord, vScore in zip(vSent.getTokens(), vSent.getSentimentScores()):
				vdResult[vWord] = vScore
		
		return vdResult
		
	
	
	def loadSSInDTrees(self):
		'''
		Loads sentiment scores into dependency tree nodes
		'''
		
		for vSent in self.getSentences():
			vSent.loadSSInDTree()
		
	
	
	def loadSSInCTrees(self, pPropagation = None):
		'''
		Loads sentiment scores into constituency tree nodes
		'''
		
		for vSent in self.getSentences():
			vSent.loadSSInCTree(pPropagation = pPropagation)
		
	
	
	def loadWordVectors(self, pWordVectors, pflgFilter = True):
		''' 
		Loads word vectors from a file or WordVector object, whichever is given
		
		By default, it filters out the words not in the data vocabulary, which can be changed to not filter (e.g. when 
		the input is already filtered).   
		'''
		
		from ml import wv
		
		self.wv = wv.WordVector()
		
		if type(pWordVectors) == str:
			vWV.load(pWVFilename = vWVFile, plFilterVocab = self.getVocabualry())
		else:
			self.wv = pWordVectors
		
	
	
	def toBratInput(self):
		'''
		Generates the dataset in Brat annotation input format
		
		There are two input types: a text input containing the raw sentences and a annotation input containing the 
		annotation.
		'''
		
		
		vSentDocOffset = 0                                          # document offset of the current sentence
		vlTxt = []
		vlAnn = []
		
		for i, vAT in enumerate(self.getAspectTerms(), start = 1):
			vlTxt.append(vAT.sentence.getText())
			
			# setting the aspect term's character offset in the document
			
			# part of the sentence before the aspect term token
			vSentUpToAT = ' '.join(vAT.sentence.getTokens()[ : (vAT.getTokenSpan()[0] - 1)])
			
			# character offset of the aspect term in token(s) containing the aspect term (e.g. 5 for built in well-built)
			vATPosInToken = ' '.join(vAT.getTokens()).find(vAT.getForm())
			
			if (vSentUpToAT == ''):
				vATDocOffset = vSentDocOffset + len(vSentUpToAT) + vATPosInToken
			else:
				vATDocOffset = vSentDocOffset + len(vSentUpToAT) + vATPosInToken + 1
			
			vlAnn.append("T%s\tAT-%s %s %s\t%s" % (i, vAT.getPolarity()[:3], vATDocOffset, vATDocOffset + len(vAT.getForm()), vAT.getForm()))
			
			# setting the next sentence's character offset in the document 
			vSentDocOffset += len(vAT.sentence.getText()) + 1       # +1 for newline 
		
		return vlTxt, vlAnn
		
	
	
	def loadBratOE(self, pBratAnnFilename, pflgVerbos = False, pdNewAnnTypes = None):
		'''
		Loads opinion or sentiment expression annotation from Brat annotator output format
		
		It assumes that the Brat input was provided using toBratInput() method, so the order 
		of the sentences are retained.
		
		pdNewAnnNames is a dictionary which translates the annotation type names with new ones. These is primarily added 
		to rename OE (opinion exression) to SE (sentiment expression) in the SE annotation internship project.
		'''
		
		# first loading OE from Brat ann file
		
		vlOEs = []
		
		for l in open(pBratAnnFilename).read().strip().split('\n'):
			vlSplit = l.split('\t')
			if vlSplit[1].startswith("OE"):
				vlOESplit = vlSplit[1].split()
				
				# renaming the annotation types if new names are given
				if pdNewAnnTypes is not None and pdNewAnnTypes != {}:
					vAnnType = pdNewAnnTypes[vlOESplit[0]]
				else:
					vAnnType = vlOESplit[0]
				
				vlOEs.append({"type": vAnnType, "from": int(vlOESplit[1]), "to": int(vlOESplit[2]), "surface": vlSplit[2]})
		
		
		# sorting OE list based on their document offset
		vlOEs.sort(key = lambda x: x["from"])
		
		
		# extracting OEs for each aspect term: assumes that the Brat input was provided using toBratInput() method, so the order 
		# of the sentences are retained.
		
		vSentDocOffset = 0
		l = []
		for i, vAT in enumerate(self.getAspectTerms(), start = 1):
			if pflgVerbos:
				print "%s) %s" % (i, vAT.sentence.getText())
				print "\nAT: " + vAT.getForm()
				print "    Polairty: " + vAT.getPolarity()
				print "OEs:"
			
			l.append(vAT.sentence.getText())
			
			# calculating document character span of the aspect term's sentence
			vSentDocSpan = (vSentDocOffset, vSentDocOffset + len(vAT.sentence.getText()))
			
			# collecting OEs of the current AT based on character offsets: not very efficient
			vlATOEs = [oe for oe in vlOEs if oe["from"] >= vSentDocSpan[0] and oe["to"] <= vSentDocSpan[1]]
			
			if pflgVerbos:
				if len(vlATOEs) == 0:
					print "    No OE annotated for this aspect term (polarity is %s)" % vAT.getPolarity()
			
			# adding the OEs to the aspect term 
			for oe in vlATOEs:
				# sanity check
				if oe["to"] > vSentDocSpan[1]:
					raise Exception("Invalid OE span: %s > %s; the end falls out of the sentence" % (oe["to"], vSentDocSpan[1]))
				
				# finding token span of the OE
				vTokenOffsetStart = vSentDocOffset
				vOETokenSpanStart = 0
				vOETokenSpanEnd = 0
				for i, token in enumerate(vAT.sentence.getTokens(), start = 1):
					vTokenOffsetEnd = vTokenOffsetStart + len(token)         # end of this token
					
					# start token: sometimes the OE is annotated from the middle of token when the tokenization has problem (e.g. -when) 
					if oe["from"] >= vTokenOffsetStart and oe["from"] <= vTokenOffsetEnd:
						vOETokenSpanStart = i
					
					# end token: sometimes the OE is annotated until the middle of token when the tokenization has problem (e.g. headphones/mic) 
					if oe["to"] >= vTokenOffsetStart and oe["to"] <= vTokenOffsetEnd:
						vOETokenSpanEnd = i
					
					vTokenOffsetStart += len(token) + 1  # start of next token
				
				vOE = OpinionExpression()
				vOE.span = (vOETokenSpanStart, vOETokenSpanEnd)
				vOE.type = oe["type"]
				
				if vOE.span == (0, 0):
					raise Exception("No token match was found!\n%s" % oe) 
				else:
					vAT.addOE(vOE)
					
					if pflgVerbos:
						print "    %s" % vOE.type
						print "        Original: %s" % oe["surface"]
						print "        Extracted: %s" % vOE.getForm()
			
			if pflgVerbos:
				print "\n............................................."
			
			# setting the next sentence's character offset in the document 
			vSentDocOffset += len(vAT.sentence.getText()) + 1       # +1 for newline 
		
	
	
	def oeToBIO(self, pflgPOSTags = False):
		'''
		Converts and returns opinion/sentiment expressions of each aspect term to BIO format
		
		The returned output is a 2D list of aspect terms sentences and their tokens in "token\ttag" format. Optionally,
		the POS tags can also be included in the output making the format "token\tPOS\ttag".
		'''
		
		vllOutput = []
		
		for vAT in self.getAspectTerms():
			vllOutput.append([])
			
			vlAnn = ['O' for i in range(vAT.sentence.length)]
			
			if len(vAT.oes) > 0:
				for vOE in vAT.oes:
					vlAnn[vOE.getTokenSpan()[0] - 1] = 'B'
					
					for i in range(vOE.getTokenSpan()[0], vOE.getTokenSpan()[1]):
						vlAnn[i] = 'I'
			
			if pflgPOSTags:
				for vAnn, vPOS, vTok in zip(vlAnn, vAT.sentence.getPOSTags(), vAT.sentence.getTokens()):
					vllOutput[-1].append("%s\t%s\t%s" % (vTok, vPOS, vAnn))
			else:
				for vAnn, vTok in zip(vlAnn, vAT.sentence.getTokens()):
					vllOutput[-1].append("%s\t%s" % (vTok, vAnn))
		
		return vllOutput
		
	
	
	def oeToIO(self):
		'''
		Converts and returns opinion/sentiment expressions of each aspect term to binary IO format 
		
		Binary IO tags can be used when there is only one type of sentiment expression annotated and one sentiment expression
		per sentence is possible.
		'''
		
		return [at.getOEsIO() for at in self.getAspectTerms()]
		
	
	
	def pickle(self, pFilename):
		'''
		Pickles the object into the give file name 
		'''
		
		pickle.dump(self.__dict__, open(pFilename, "wb"), protocol = 2)
		
	
	
	def loadFromPickle(self, pFilename):
		'''
		Loads the pickled ABSASet object to this object 
		'''
		
		self.__dict__.update(pickle.load(open(pFilename)))
		
	
	
	def getPOSTagSet(self):
		'''
		Retunrs the POS tag set of the sentences of the dataset 
		'''
		
		vlPOSTags = []
		
		for vSent in self.getSentences():
			vlPOSTags += vSent.getPOSTags()
		
		return set(vlPOSTags)
		
	
	


class ABSAContext:
	"""
	Class for aspect-based sentiment analysis context (set of sentences) 
	"""
	
	
	def __init__(self, pABSASet):
		'''
		Constructor 
		'''
		
		self.dataset = pABSASet
		
		self.sentences = []
		
	
	
	def getSentences(self, pSort=''):
		'''
		Returns the ABSA sentences
		
		The sort options are:
		- None: in document order
		- text: in sentence text order
		'''
		
		if pSort.lower() == "text":
			return [s for s in sorted(self.getSentences(), key=lambda x: x.getText())]
		else:
			return self.sentences
		
	
	
	def addSentence(self, pSentence):
		'''
		Adds ABSA sentence to the context 
		'''
		
		if pSentence not in self.sentences:
			pSentence.context = self
			self.sentences.append(pSentence)
		
	
	
	def getTokens(self):
		'''
		Returns token list of all sentences in the context
		
		For tokens to make sense, make sure the data is tokenized.
		'''
		
		return [t for s in self.getSentences() for t in s.getTokens()]
		
	
	
	def getATPolarities(self):
		'''
		Returns list of polarities of all aspect temrs in the context 
		'''
		
		return [at.getPolarity() for s in self.getSentences() for at in s.getAspectTerms()]
		
	
	
	def getOverallPolarity(self):
		'''
		Returns the overall polarity of the context based on the polarity of its aspect terms
		
		See the code for how the overall is calculated. Basically, number positive or negative polarities in the context
		should be at least twice as many as the opposite polarity, or otherwise the polarity will be considered neutral.
		
		ToDo: polarity values are fixated here. They should be variable based on the data.
		'''
		vdPolarities = util.groupBy(self.getATPolarities())
		
		if "negative" not in vdPolarities and "positive" not in vdPolarities:
			return "neutral"
		elif "negative" not in vdPolarities and "positive" in vdPolarities:
			return "positive"
		elif "negative" in vdPolarities and "positive" not in vdPolarities:
			return "negative"
		elif "negative" in vdPolarities and "positive" in vdPolarities:
			if vdPolarities["negative"] >= (2 * vdPolarities["positive"]):
				return "negative"
			elif vdPolarities["positive"] >= (2 * vdPolarities["negative"]):
				return "positive"
			else:
				return "neutral"
		else:
			raise Exception("Strange situation: %s" % vdPolarities)
		
	
	


class ABSASent:
	'''
	Class for aspect-based sentiment analysis sentence
	'''
	
	
	def __init__(self, pABSAContext):
		'''
		Constructor
		'''
		
		# ABSAContext the sentence belongs to
		self.context = pABSAContext
		
		self.text = None
		self.aspectTerms = []
		self.cTree = None
		self.dTree = None
		self.posTagging = None
		self.sentScores = []     # sentiment score, one per word in the tokenized self.text
		self.oes = []             # opinion expressions in the sentence (not those of aspect terms)
		
	
	
	def getText(self):
		'''
		Returns sentence text (form)
		'''
		
		return self.text
		
	
	
	def getTokens(self):
		'''
		Returns the tokenization of the sentence
		
		The sentence text is assumed to be in tokenized format and only 
		splits on space.
		'''
		
		return self.getText().split()
		
	
	
	@property
	def length(self):
		'''
		Returns the sentence length
		'''
		
		return len(self.getTokens())
		
	
	
	def getAspectTerms(self):
		'''
		Returns aspect terms (objects) of the sentence 
		'''
		
		return self.aspectTerms
		
	
	
	def getOEs(self):
		'''
		Returns opinion expressions (objects) of the sentence 
		'''
		
		return self.oes
		
	
	
	def getAspectTermCount(self):
		'''
		Returns the number of aspect terms in the sentence 
		'''
		
		return len(self.aspectTerms)
		
	
	
	def getAspectTermForms(self):
		'''
		Returns aspect term forms (term attributes) of the sentence 
		'''
		
		return [t.getForm() for t in self.getAspectTerms()]
		
	
	
	def getConstTree(self):
		'''
		Returns the constituency parse tree of the sentence
		
		The returned object is of type constparse.ConstTree
		'''
		
		return self.cTree
		
	
	
	def getPOSTags(self):
		'''
		Returns the list of POS tags which matches the token list
		
		The POS tags are extracted from the constituency tree or dependency tree
		'''
		
		if self.cTree is not None:
			return self.cTree.getPOSs()
		elif self.dTree is not None:
			return self.dTree.getPOSs()
		elif self.posTagging is not None:
			return self.posTagging.getPOSTags()
		else:
			return []
		
	
	
	def getDepTree(self):
		'''
		Returns the dependency parse tree of the sentence
		
		The returned object is of type depparse.DepTree
		'''
		
		return self.dTree
		
	
	
	def loadConstTree(self, pConstTree):
		'''
		Loads the constituency parse tree of the sentence
		
		The contituency tree can be provided in bracketing format or as
		constparse.ConstTree object.
		'''
		
		# loading the tree
		
		if isinstance(pConstTree, constparse.ConstTree):
			vConstTree = pConstTree.getPTBFormat()
		else:
			vConstTree = pConstTree
		
		self.cTree = ABSACTree()
		self.cTree.loadPTBTree(vConstTree, pflgExpandTerminal = True)
		
		# sanity check; comment out
		#if self.cTree.surface != self.getText():
		#	print "Sentence and tree mismatch:\nSentence: %s\nTree: %s\n" % (self.getText(), self.cTree.getPTBFormat())
		
	
	
	def loadDepTree(self, pDepTree):
		'''
		Loads the dependency parse tree of the sentence
		
		The dependency tree is assumed to be depparse.DepTree object.
		'''
		
		# loading the tree
		
		if not isinstance(pDepTree, depparse.DepTree):
			raise Exception("A DepTree object is expected!")
		
		self.dTree = ABSADTree()
		self.dTree.loadFromDepTree(pDepTree = pDepTree)
		
		# sanity check; comment out
		#if self.dTree.surface != self.getText():
		#	print "Sentence and tree mismatch:\nSentence: %s\nTree:%s\n" % (self.getText(), self.dTree.surface)
		
	
	
	def loadPOSTagging(self, pPOSTagging):
		'''
		Loads POS tagging of the sentence
		
		POS tagging is a pos.POSTagging object.
		'''
		
		self.posTagging = pPOSTagging
		
	
	
	def loadBIOOpinionExpressions(self, plBIO):
		'''
		Loads opinion expression annotations of the sentence based on BIO tagging
		
		The BIO labels are provided in a list which is supposed to match the tokens in the sentence.
		'''
		
		# start and end of 1-based span 
		vSpanStart = 0
		vSpanEnd = 0
		
		for i, vLabel in enumerate(plBIO, start = 1):
			if vLabel.lower() == 'b':
				vSpanStart = i
				vSpanEnd = i
			elif vLabel.lower() == 'i':
				vSpanEnd += 1
			elif vLabel.lower() == 'o':
				if vSpanStart != 0:                            # means the first token after span
					vOE = OpinionExpression()
					vOE.span = (vSpanStart, vSpanEnd)
					vOE.sentence = self
					self.oes.append(vOE)
					vSpanStart = 0
					vSpanEnd = 0
		
	
	
	def embedATsInCTree(self, pflgModNPStruct = False, pEmbedPosition = "spanning-constituent", pOnMismatch = 'no', pflgExtendATSpanToDT = False):
		'''
		Embeds the aspect terms in the constituency tree
		
		pflgModNPStruct specifies whether the NP structures should be edited
		in order to avoid term/constituent mismatch.
		
		pEmbedPosition specifies where the aspect term should be embedded
		in the subtree. The possible values are:
		- (span)ning-constituent: embedding into a node in subtree which 
		                          spans the aspect term tokens. This may 
		                          cause mismatches where the aspect term 
		                          span is not fully covered by the node
		                          span. pOnMismatch can be set to fix the
		                          issue.
		- (pre)-terminals: embedding into each and every pre-terminal node
		                   in the subtree falling in the aspect term span.
		
		pOnMismatch specifies the method for handling aspect term/constituent
		mismatch. The possible values are:
		- nothing (or none): do not handle the mismatches
		- node (or insert): insert a new node covering the aspect term span
		
		If pflgExtendATSpanToDT is set to true, the span of aspect terms which 
		only exclude the determiner of the NP is extended to cover the 
		determiner to reduce the number of mismatches.
		'''
		
		# modifying NP structures
		
		if pflgModNPStruct:
			self.cTree.modifyNPStruct()
		
		# embeding the aspect terms
		
		self.cTree.embedAspectTerms(plAspectTerm = self.getAspectTerms(), pEmbedPosition = pEmbedPosition, pOnMismatch = pOnMismatch, pflgExtendToDT = pflgExtendATSpanToDT)
		
	
	
	def embedOEsInCTree(self, pflgModNPStruct = False, pEmbedPosition = "spanning-constituent", pOnMismatch = 'no'):
		'''
		Embeds the aspect terms in the constituency tree
		
		pflgModNPStruct specifies whether the NP structures should be edited in order to avoid expression/constituent 
		mismatch.
		
		pEmbedPosition specifies where the opinion expression should be	embedded in the subtree. The possible values are:
		- (span)ning-constituent: embedding into a node in subtree which spans the OE tokens. This may cause mismatches 
		                          where the OE span is not fully covered by the node span. pOnMismatch can be set to fix
		                          the issue.
		- (pre)-terminals: embedding into each and every pre-terminal node in the subtree falling in the OE span.
		
		pOnMismatch specifies the method for handling expression/constituent mismatch. The possible values are:
		- nothing (or none): do not handle the mismatches
		- node (or insert): insert a new node covering the aspect term span
		'''
		
		# modifying NP structures
		
		if pflgModNPStruct:
			self.cTree.modifyNPStruct()
		
		# embeding the opinion expressions
		
		self.cTree.embedOpinionExpressions(plOEs = self.getOEs(), pEmbedPosition = pEmbedPosition, pOnMismatch = pOnMismatch)
		
	
	
	def embedATsInDTree(self):
		'''
		Embeds the aspect terms in the dependency tree
		'''
		
		# embeding the aspect terms
		
		self.dTree.embedAspectTerms(plAspectTerms = self.getAspectTerms())
		
	
	
	def embedOEsInDTree(self):
		'''
		Embeds the opinion expressions in the dependency tree
		'''
		
		self.dTree.embedOpinionExpressions(plOEs = self.getOEs())
		
	
	
	def getPTBConstTree(self):
		'''
		Returns the constituency tree of the sentence in PTB bracketing format
		'''
		
		return self.cTree.getPTBFormat()
		
	
	
	def extractATsInCTree(self):
		'''
		Extracts and returns the list of aspect terms embedded into the
		constituency tree if the tree is already loaded and an empty list
		otherwise
		'''
		
		if self.cTree == None:
			return []
		else:
			return self.cTree.extractAspectTerms()
		
	
	
	def extractATermConstMismatch(self):
		'''
		Extracts the aspect terms which do not match a constituent node 
		in the tree thus missing in the constituency tree
		
		Aspect term/constituent node mismatches happen because of not being 
		embeded in the tree which occur due to the inconsistency between 
		syntactic phrases and the phrases to which these terms are originally 
		assigned. Parsing errors can be one reason for this but also the 
		annotation scheme, such as flat noun phrase annotation, can also 
		cause this problem.
		'''
		
		if self.cTree == None:
			return []
		
		vlSentATerms = self.getAspectTerms()
		vlCTreeATerms = self.cTree.extractAspectTerms()
		
		vlMismatches = []
		
		if len(vlCTreeATerms) != len(vlSentATerms):
			for vSentAT in vlSentATerms:
				if vSentAT not in vlCTreeATerms:
					vlMismatches.append(vSentAT)
		
		return vlMismatches
		
	
	
	def loadSentimentScores(self, pSentexicon, pNeutralScore = None):
		'''
		Loads sentiment scores to words in the sentence from a sentiment
		lexicon which is a Sentexicon object
		
		It returns a tuple of the total number of words in the sentence 
		and the number of words found in the lexicon.
		
		If the word is not found in the lexicon, None will be used.
		
		For details about Sentexicon object, see sentexicon.py
		'''
		
		vlWords = self.getTokens()
		self.sentScores = []
		
		for vWord in vlWords:
			vScore = pSentexicon.getScore(vWord)
			if vScore is None and pNeutralScore is not None:
				self.sentScores.append(pNeutralScore)
			else:
				self.sentScores.append(vScore)
		
		return self.length, len([s for s in self.sentScores if s is not None])
		
	
	
	def getSentimentScores(self):
		'''
		Returns the sentiment scores in a list corresponding to the token list
		'''
		
		return self.sentScores
		
	
	
	def loadSSInDTree(self):
		'''
		Loads sentiment scores into the dependency tree of the sentence
		'''
		
		self.dTree.loadSentScores(self.sentScores)
		
	
	
	def loadSSInCTree(self, pPropagation = None):
		'''
		Loads sentiment scores into the constituency tree of the sentence
		'''
		
		self.cTree.loadSentScores(self.sentScores, pPropagation = pPropagation)
		
	
	
	def generateNGramKTree(self):
		'''
		Generates and returns the tree representation of the surface form
		of the sentence
		'''
		
		vABSANGramKTree = ABSANGramKTree(pABSASent = self)
		
		return vABSANGramKTree.generateNGramKTree()
		
	
	
	def extractCaraFXInput(self):
		'''
		Extracts the data in the format required for Cara feature extractor
		 
		Cara feature extractor requires three files: normalized text, POS tagged sentences and aspect term indexes. See
		https://github.com/CNGL-repo/Cara/wiki/Cara pipeline.
		'''
		
		# The data must be tokenized as loading time.
		vNormalizedTxt = self.getText()
		vNormalizedTxt = vNormalizedTxt.replace("(", "-LRB-")
		vNormalizedTxt = vNormalizedTxt.replace(")", "-RRB-")
		
		# format: (POS_1 token_1)(POS_2 token_2)...(POS_n token_n)
		vPOSTagged = ''.join(["(%s %s)" % (p, t) for p, t in zip(self.cTree.getPOSs(), self.getTokens())])
		vPOSTagged = vPOSTagged.replace(" (", " -LRB-")
		vPOSTagged = vPOSTagged.replace(" )", " -RRB-")
		
		
		return vNormalizedTxt, vPOSTagged
		
	
	
	def getAvgOESentScore(self):
		'''
		Calculates and returns the average sentiment score of the opinion expression tokens of the sentence 
		'''
		
		vlOEs = self.getOEs()
		
		if len(vlOEs) == 0:
			return 0
		else:
			return sum([oe.getAvgSentScore() for oe in vlOEs]) / len(vlOEs)
		
	
	
	def getAvgSentScore(self):
		'''
		Calculates and returns the average sentiment score sentence tokens
		'''
		
		vlSentScores = self.getSentimentScores()
		
		return sum(vlSentScores) / len(vlSentScores)
		
	
	
	def getPolarScores(self, pNeutralScore = None):
		'''
		Returns sentiment scores with non-neutral polarity
		
		Neutral polarity score can be set as parameter. It is None by default meaning that no polarity score is assigned
		to neutral words.
		'''
		
		return [s for s in self.sentScores if s != pNeutralScore]
		
	
	
	def getWordVectors(self):
		'''
		Returns the word vectors of the sentence tokens
		'''
		
		return [self.dataset.wv.getVector(t) for t in self.getTokens()]
		
	
	



class AspectTerm:
	'''
	Class for aspect-based sentiment analysis aspect term
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		self.term = None
		self.polarity = None
		
		self.sentence = None
		
		# the token span of the term in the text (1-base indexes)
		self.span = None
		
		# opinion/sentiment expressions towards the aspect term (not those of the sentence, i.e. ABSASent.oe
		self.oes = []
		
	
	
	def getForm(self):
		'''
		Returns the form of the aspect term
		'''
		
		return self.term
		
	
	
	def getTokens(self):
		'''
		Returns list of tokens of the aspect term
		'''
		
		if self.span is not None:
			return self.sentence.getTokens()[self.span[0]-1 : self.span[1]]
		else:
			return ''
		
	
	
	def getPolarity(self):
		'''
		Returns the polarity of the aspect term
		'''
		
		return self.polarity
		
	
	
	def getTokenSpan(self):
		'''
		Returns the token span of the term
		'''
		
		return self.span
		
	
	
	def getSentenceMarking(self):
		'''
		Returns the marking of the aspect terms on its sentence 
		
		The marking is a list corresponding to the aspect term's sentence tokens, every element of which is either 0 or 1
		depending on if the corresponding token is in the aspect term span or not. 
		'''
		
		return [1 if self.span[0] <= (i + 1) <= self.span[1] else 0 for i in range(self.sentence.length)] 
		
	
	
	def embedInCTree(self, pEmbedPosition, pOnMismatch = "nothing", pflgExtendToDT = False):
		'''
		Embeds the aspect term in the constituency tree of the sentence
		
		pEmbedPosition specifies where the aspect term should be embedded
		in the subtree. The possible values are:
		- (span)ning-constituent: embedding into a node in subtree which 
		                          spans the aspect term tokens. This may 
		                          cause mismatches where the aspect term 
		                          span is not fully covered by the node
		                          span. pAspectTerm can be set to fix the
		                          issue.
		- (pre)-terminals: embedding into each and every pre-terminal node
		                   in the subtree falling in the aspect term span.
		
		pOnMismatch specifies the method for handling aspect term/constituent
		mismacth. The possible values are:
		each aspect term covering the term span. The values are:
		- nothing (or none): do not handle the mismatches
		- node (or insert): insert a new node covering the aspect term span
		
		If pflgExtendATSpanToDT is set to true, the span of aspect terms which 
		only exclude the determiner of the NP is extended to cover the 
		determiner to reduce the number of mismatches.
		'''
		
		self.sentence.getConstTree().embedAspectTerm(pAspectTerm = self, pEmbedPosition = pEmbedPosition, pOnMismatch = pOnMismatch, pflgExtendToDT = pflgExtendToDT)
		
	
	
	def getCTreeEmbeding(self, pEmbedPosition, pOnMismatch = "nothing", pflgExtendToDT = False, pdATReprOptions = {}):
		'''
		Embeds the aspect terms in a copy of the constituency tree of the
		sentence and returns the resulting tree
		
		The representation of aspect term in the tree is based on the options
		specified in a dictionary (pdATReprOptions). The options involve
		attaching aspect term suffix, inserting aspect term node and attaching
		aspect term polarity, each with several possibilities:
		- {suffix:   [node/subtree/pre-terminal/parents/parents-partial]}
		- {node:     [parent/sister]}
		- {polarity: [node/subtree]}
		
		For details of the options, see ABSACNode.decorateAT().
		'''
		
		# copying the tree to keep the original tree intact
		# python deepcopy is used as it did a better job in an experiment than the ConstTree.deepCopy().
		vCTreeCopy = copy.deepcopy(self.sentence.getConstTree())
		
		if vCTreeCopy.embedAspectTerm(pAspectTerm = self, pEmbedPosition = pEmbedPosition, pOnMismatch = pOnMismatch, pflgExtendToDT = pflgExtendToDT):
			vCTreeCopy.decorateAT(pdATReprOptions = pdATReprOptions)
			return vCTreeCopy
		else:
			return False
		
	
	
	def getDTreeEmbeding(self, pdATReprOptions = {}):
		'''
		Embeds the aspect terms in a copy of the dependency tree of the
		sentence and returns the resulting tree
		
		The representation of aspect term in the tree is based on the options
		specified in a dictionary (pdATReprOptions). The options involve
		attaching aspect term suffix and attaching aspect term polarity, 
		each with several possibilities.
		
		NOTE: unlike in constituency tree decoration, AT node insertion 
		is not done for the dependency tree. The reason is that in dependency
		tree, there is a one to one relation between the nodes and sentence
		tokens. Inserting a node causes a mismatch between the surface 
		form of the sentence and its dependency tree. Also, concepts like
		word form, POS tag and dependency relation are not meaningful in
		an inserted AT node. Instead, the aspect term representation using
		inserted AT nodes can be done at the PTB bracketing format representation
		level of the tree. See ABSADTree.generateDepKTree(). Note that
		the current design of this method in terms of the argument passed
		to it may not be coherent and meaningful, because what is expected
		from an aspect term representation parameter (pdATReprOptions) is
		to carry all the setting required for this purpose, not only the 
		part concerned with suffix and polarity attachement. This may be 
		handled in the future.
		
		For details of the options, see ABSADNode.decorateAT().
		'''
		
		# copying the tree to keep the original tree intact
		# NOTE: consider testing with python deepcopy (Python deepcopy did a 
		# better job in an experiment than the ConstTree.deepCopy() in the
		# getCTreeEmbeding())
		vDTreeCopy = self.sentence.getDepTree().deepCopy()
		#vDTreeCopy = copy.deepcopy(self.sentence.getDepTree())
		
		vDTreeCopy.embedAspectTerm(pAspectTerm = self)
		vDTreeCopy.decorateAT(pdATReprOptions = pdATReprOptions)
		
		return vDTreeCopy
		
	
	
	def getNGramKTreeEmbeding(self, pFormat = "unary", pNodeContentType = "word", pdNGramKTreeOptions = None, pdATReprOptions = None, pdOptions = None):
		'''
		Generates and returns the tree representation of the surface form
		of the corresponding sentence in various formats with aspect term 
		optionally embedded in it
		
		Formats include:
		- unary: each word is the child of its previous word 
		- bigram: each bigram forms a parent/child subtree and all of these subtrees
		          are dominated by a root node at the top. These captures unigrams as
		          well if subset tree kernels are used.
		- binary: each node has two children which are both nodes representing the word next to the current node's word.
		          The first children has a dummy terminal child X which helps capture unigrams via subset tree fragments.
		          The second child recursively continues the format by having the next word as its children in the same 
		          way.
		
		pNodeContentTypes specifies the content of the n-gram tree nodes, such as word forms or POS tags. For details, 
		see the ABSANGramKTree constructor.
		
		Aspect term representation options include:
		- "node": inserts AT node at:
		    - "x": replacing X node under/above the aspect term token  
		- "suffix": attached AT suffix to:
		    - "token": to aspect term token
		'''
		
		vABSANGramKTree = ABSANGramKTree(pABSASent = self.sentence, pNodeContentType = pNodeContentType, pAspectTerm = self, pdOptions = pdNGramKTreeOptions)
		
		return vABSANGramKTree.generateNGramKTree(pFormat, pdATReprOptions, pdOptions)
		
	
	
	def addOE(self, pOE):
		'''
		Adds opinion exression object to the aspect term
		'''
		
		pOE.sentence = self.sentence
		pOE.aspectTerm = self
		
		self.oes.append(pOE)
		
	
	
	def getOEsBIO(self, pflgPOSTags=False):
		'''
		Returns BIO tagging of the opinion/sentiment expressions of the aspect term
	
		The returned output is a list of BIO tags corresponding to aspect term sentence tokens.
		'''
		
		vlAnn = ['O' for i in range(self.sentence.length)]
		
		if len(self.oes) > 0:
			for vOE in self.oes:
				vlAnn[vOE.getTokenSpan()[0] - 1] = 'B'
				
				for i in range(vOE.getTokenSpan()[0], vOE.getTokenSpan()[1]):
					vlAnn[i] = 'I'
		
		return vlAnn
		
	
	
	def getOEsIO(self):
		'''
		Returns the binary IO tagging of the opinion/sentiment expression boundaries of the aspect term
		
		Binary IO tags can be used when there is only one type of sentiment expression annotated and one sentiment expression
		per sentence is possible.
		'''
		
		vlAnn = ['O'] * self.sentence.length
		
		if len(self.oes) > 0:
			for vOE in self.oes:
				for i in range(vOE.getTokenSpan()[0] - 1, vOE.getTokenSpan()[1]):
					vlAnn[i] = 'I'
		
		return vlAnn
		
	
	



class OpinionExpression():
	'''
	Class for opinion expression annotations
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		self.type = None       # type of the opinion expression if there is a categorization 
		                       # e.g. SE (sentiment expression), SE-pcomp (preceeding complement of SE)
		
		self.sentence = None
		
		# the aspect term towards which the opinion is expressed.
		self.aspectTerm = None
		
		self.span = None
		
	
	
	def getTokens(self):
		'''
		Returns list of tokens of the opinion expression
		'''
		
		return self.sentence.getTokens()[self.span[0]-1 : self.span[1]]
		
	
	
	def getForm(self):
		'''
		Returns the form of the opinion expression
		'''
		
		return ' '.join(self.getTokens())
		
	
	
	def getTokenSpan(self):
		'''
		Returns the token span of the term
		'''
		
		return self.span
		
	
	
	@property
	def length(self):
		'''
		Return the token length of the opinion expression  
		'''
		
		return self.span[1] - self.span[0] + 1
		
	
	
	def embedInCTree(self, pEmbedPosition, pOnMismatch = "nothing"):
		'''
		Embeds the opinion expression in the constituency tree of the sentence
		
		pEmbedPosition specifies where the opinion expression should be	embedded in the subtree. The possible values are:
		- (span)ning-constituent: embedding into a node in subtree which spans the OE tokens. This may cause mismatches 
		                          where the OE span is not fully covered by the node span. pOnMismatch can be set to fix
		                          the issue.
		- (pre)-terminals: embedding into each and every pre-terminal node in the subtree falling in the OE span.
		
		pOnMismatch specifies the method for handling expression/constituent mismatch. The possible values are:
		- nothing (or none): do not handle the mismatches
		- node (or insert): insert a new node covering the aspect term span
		'''
		
		self.sentence.getConstTree().embedOpinionExpression(pOE = self, pEmbedPosition = pEmbedPosition, pOnMismatch = pOnMismatch)
		
	
	
	def getCTreeEmbeding(self, pEmbedPosition, pOnMismatch = "nothing", pdOEReprOptions = {}):
		'''
		Embeds the opinion expression in a copy of the constituency tree of the sentence and returns the resulting tree
		
		The representation of OE in the tree is based on the options specified in a dictionary (pdOEReprOptions). The 
		options involve attaching OE suffix, inserting OE node and attaching OE polarity, each with several possibilities:
		- {suffix:   [node/subtree/pre-terminal/parents/parents-partial]}
		- {node:     [parent/sister]}
		- {polarity: [node/subtree]}
		
		For details of the options, see ABSACNode.decorateOE().
		'''
		
		# copying the tree to keep the original tree intact
		# python deepcopy is used as it did a better job in an experiment than the ConstTree.deepCopy().
		vCTreeCopy = copy.deepcopy(self.sentence.getConstTree())
		
		if vCTreeCopy.embedOpinionExpression(pOE = self, pEmbedPosition = pEmbedPosition, pOnMismatch = pOnMismatch):
			vCTreeCopy.decorateOE(pdOEReprOptions = pdOEReprOptions)
			return vCTreeCopy
		else:
			return False
		
	
	
	def getDTreeEmbeding(self, pdOEReprOptions = {}):
		'''
		Embeds the opinion expression in a copy of the dependency tree of the
		sentence and returns the resulting tree
		
		The representation of OE in the tree is based on the options specified 
		in a dictionary (pdOEReprOptions). The options involve attaching OE
		suffix and attaching its polarity, each with several possibilities.
		
		NOTE: see the same method for aspect term for further documentation. 
		
		For details of the options, see ABSADNode.decorateOE().
		'''
		
		# copying the tree to keep the original tree intact
		# NOTE: consider testing with python deepcopy (Python deepcopy did a 
		# better job in an experiment than the ConstTree.deepCopy() in the
		# getCTreeEmbeding())
		vDTreeCopy = self.sentence.getDepTree().deepCopy()
		#vDTreeCopy = copy.deepcopy(self.sentence.getDepTree())
		
		vDTreeCopy.embedOpinionExpression(pOE = self)
		vDTreeCopy.decorateOE(pdOEReprOptions = pdOEReprOptions)
		
		return vDTreeCopy
		
	
	
	def getAvgSentScore(self):
		'''
		Calculates and returns the average sentiment score of the opinion expression tokens 
		'''
		
		return sum(self.sentence.getSentimentScores()[self.span[0] - 1 : self.span[1]]) / self.length
		
	
	


class ABSACTree(constparse.ConstTree):
	'''
	Class for constituency parse tree of a ABSA sentence
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		constparse.ConstTree.__init__(self)
	
	
	
	def _createNewTree(self):
		'''
		Creates and returns a new ABSACTree
		 
		This method is useful in class inheriting.
		'''
		
		return ABSACTree()
	
	
	
	def embedAspectTerms(self, plAspectTerms, pEmbedPosition, pOnMismatch = "nothing", pflgExtendToDT = False):
		'''
		Embeds the aspect terms in the tree nodes
		
		pEmbedPosition specifies where the aspect term should be embedded
		in the subtree. The possible values are:
		- (span)ning-constituent: embedding into a node in subtree which 
		                          spans the aspect term tokens. This may 
		                          cause mismatches where the aspect term 
		                          span is not fully covered by the node
		                          span. pAspectTerm can be set to fix the
		                          issue.
		- (pre)-terminals: embedding into each and every pre-terminal node
		                   in the subtree falling in the aspect term span.
		
		pOnMismatch specifies the method for handling aspect term/constituent
		mismacth. The possible values are:
		each aspect term covering the term span. The values are:
		- nothing (or none): do not handle the mismatches
		- node (or insert): insert a new node covering the aspect term span
		
		If pflgExtendATSpanToDT is set to true, the span of aspect terms which 
		only exclude the determiner of the NP is extended to cover the 
		determiner to reduce the number of mismatches.
		'''
		
		for vAT in plAspectTerms:
			self.embedAspectTerm(pAspectTerm = vAT, pEmbedPosition = pEmbedPosition, pOnMismatch = pOnMismatch, pflgExtendToDT = pflgExtendToDT)
	
	
	
	def embedAspectTerm(self, pAspectTerm, pEmbedPosition, pOnMismatch = "nothing", pflgExtendToDT = False):
		'''
		Embeds the aspect term in the tree and returns true if the embedding 
		happened
		
		pEmbedPosition specifies where the aspect term should be embedded
		in the subtree. The possible values are:
		- (span)ning-constituent: embedding into a node in subtree which 
		                          spans the aspect term tokens. This may 
		                          cause mismatches where the aspect term 
		                          span is not fully covered by the node
		                          span. pAspectTerm can be set to fix the
		                          issue.
		- (pre)-terminals: embedding into each and every pre-terminal node
		                   in the subtree falling in the aspect term span.
		
		pOnMismatch specifies the method for handling aspect term/constituent
		mismacth. The possible values are:
		each aspect term covering the term span. The values are:
		- nothing (or none): do not handle the mismatches
		- node (or insert): insert a new node covering the aspect term span
		
		If pflgExtendATSpanToDT is set to true, the span of aspect terms which 
		only exclude the determiner of the NP is extended to cover the 
		determiner to reduce the number of mismatches.
		
		NOTE: This method is not redundant to embedAspectTerms. Its main
		purpose is to be called by AspectTerm.
		'''
		
		return self.root.embedAspectTerm(pAspectTerm = pAspectTerm, pEmbedPosition = pEmbedPosition, pOnMismatch = pOnMismatch, pflgExtendToDT = pflgExtendToDT)
	
	
	
	def embedOpinionExpressions(self, plOEs, pEmbedPosition, pOnMismatch = "nothing"):
		'''
		Embeds the opinion expressions in the tree nodes
		
		pEmbedPosition specifies where the opinion expression should be	embedded in the subtree. The possible values are:
		- (span)ning-constituent: embedding into a node in subtree which spans the OE tokens. This may cause mismatches 
		                          where the OE span is not fully covered by the node span. pOnMismatch can be set to fix
		                          the issue.
		- (pre)-terminals: embedding into each and every pre-terminal node in the subtree falling in the OE span.
		
		pOnMismatch specifies the method for handling expression/constituent mismatch. The possible values are:
		- nothing (or none): do not handle the mismatches
		- node (or insert): insert a new node covering the aspect term span
		'''
		
		for vOE in plOEs:
			self.embedOpinionExpression(pOE = vOE, pEmbedPosition = pEmbedPosition, pOnMismatch = pOnMismatch)
	
	
	
	def embedOpinionExpression(self, pOE, pEmbedPosition, pOnMismatch = "nothing"):
		'''
		Embeds the opinion expressions in the tree and returns true if the embedding happened
		
		pEmbedPosition specifies where the opinion expression should be	embedded in the subtree. The possible values are:
		- (span)ning-constituent: embedding into a node in subtree which spans the OE tokens. This may cause mismatches 
		                          where the OE span is not fully covered by the node span. pOnMismatch can be set to fix
		                          the issue.
		- (pre)-terminals: embedding into each and every pre-terminal node in the subtree falling in the OE span.
		
		pOnMismatch specifies the method for handling expression/constituent mismatch. The possible values are:
		- nothing (or none): do not handle the mismatches
		- node (or insert): insert a new node covering the aspect term span
		
		NOTE: This method is not redundant to embedOpinionExpressions. Its main	purpose is to be called by OpinionExpression.
		'''
		
		return self.root.embedOpinionExpression(pOE = pOE, pEmbedPosition = pEmbedPosition, pOnMismatch = pOnMismatch)
		
	
	
	def decorateAT(self, pdATReprOptions = {}):
		'''
		Decorates the tree with aspect term representation according to 
		the options
		
		pdATReprOptions specifies the representation details of the aspect
		terms in the tree. See ABSACNode.decorateAT() for details.
		'''
		
		self.root.decorateAT(pdATReprOptions = pdATReprOptions)
		
	
	
	def decorateOE(self, pdOEReprOptions = {}):
		'''
		Decorates the tree with opinion expression representation according to the options
		
		pdOEReprOptions specifies the representation details of the opinion expression in the tree. See ABSACNode.decorateOE()
		for details.
		'''
		
		self.root.decorateOE(pdOEReprOptions = pdOEReprOptions)
		
	
	
	def decorateSS(self, pdSSReprOptions = {}):
		'''
		Decorates the tree with sentiment score representation according to 
		the options
		
		pdSSReprOptions specifies the representation details of the sentiment
		scores in the tree. See ABSACNode.decorateSS() for details.
		'''
		
		self.root.decorateSS(pdSSReprOptions = pdSSReprOptions)
		
	
	
	def _createRoot(self):
		'''
		Creates and returns the root node
		'''
		
		return ABSACNode()
		
	
	
	def extractAspectTerms(self):
		'''
		Extracts the aspect terms from nodes
		'''
		
		return self.root.extractAspectTerms()
		
	
	
	def modifyNPStruct(self):
		'''
		Modifies the structure of the noun phrases in order to avoid
		term/constituent mismatch caused by a flat NP structure
		'''
		
		self.root.modifyNPStruct()
		
	
	
	def loadSentScores(self, plSentScores, pPropagation = None):
		'''
		Loads sentiment scores to tree nodes
		
		The sentiment scores are given in a list which corresponds to the
		terminal (i.e. token) list.
		
		Propagation argument specifies the method by which the score are 
		propagated from terminals up to the root node. None means scores
		are only assigned to the terminal nodes, thus they are 0 for the 
		phrase nodes.
		'''
		
		self.root.loadSentScores(plSentScores, pPropagation = pPropagation)
		
	
	
	def extractAT2OEPath(self, pAT, pOE):
		'''
		Extracts and returns the constituency path from the given aspect term to the given opinion expression on the tree
		
		The path is a string containing the syntactic labels and > and < representing up move and down move in the tree
		respectively.
		'''
		
		## NOTE: to speed things up, this method does not use/need aspect terms and opinion expressions embedded in the
		## tree beforehand. It only heuristically finds a node for them based on their span.
		
		vATSpan = pAT.getTokenSpan()
		vOESpan = pOE.getTokenSpan()
		
		# finding AT node: if only one token, use the pre-terminal, otherwise the parent of the first node (heuristic)
		if (vATSpan[1] - vATSpan[0]) == 0:
			vATNode = self.getPreTerminalNodeAt(vATSpan[0])
		else:
			vATNode = self.getPreTerminalNodeAt(vATSpan[0]).getParent()
		
		# finding OE node: if only one token, use the pre-terminal, otherwise the parent of the first node (heuristic)
		if (vOESpan[1] - vOESpan[0]) == 0:
			vOENode = self.getPreTerminalNodeAt(vOESpan[0])
		else:
			vOENode = self.getPreTerminalNodeAt(vOESpan[0]).getParent()
		
		return vATNode.getPathTo(vOENode)
		
	
	
	def extractTopmostVP(self, pSpan):
		'''
		Extracts and returns the topmost verb phrase node above the given span in the tree
		'''
		
		return self.root.extractTopmostVP(pSpan)
		
	
	


class ABSACNode(constparse.ConstNode):
	'''
	Class for constituency tree node of a ABSA sentence
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		constparse.ConstNode.__init__(self)
		
		# list of type AspectTerm
		self.aspectTerms = []
		
		# sentiment score
		self.sentScore = None
		
		# list of type OpinionExpression
		self.oe = []
		
	
	
	def deepCopy(self, pflgCopyTree = False):
		'''
		NOTE: it seems python deepcopy() works better. Try the idea before
		using this method.
		
		Creates and returns a deep copy of the node optionally including 
		the sub tree under it
		'''
		
		vNodeCopy = constparse.ConstNode.deepCopy(self, pflgCopyTree)
		
		# copying the aspect terms
		vNodeCopy.aspectTerms = self.aspectTerms[:]
		
		# copying the sentiment score
		vNodeCopy.sentScore = self.sentScore
		
		# copying the opinion expressions
		vNodeCopy.oe = self.oe[:]
		
		return vNodeCopy
		
	
	
	def shallowCopy(self, pflgCopyTree = False):
		'''
		NOTE: before using, check if python shallowcopy() does not work
		as expected.
		
		Creates and returns a shallow copy of the node
		
		The shallow copy does not have a parent and children.
		'''
		
		vNodeCopy = constparse.ConstNode.shallowCopy(self, pflgCopyTree)
		
		# copying the aspect terms
		#vNodeCopy.aspectTerms = self.aspectTerms[:]
		
		# copying the sentiment score
		vNodeCopy.sentScore = self.sentScore
		
		return vNodeCopy
		
	
	
	def _getNewNode(self):
		'''
		Creates and returns a node
		'''
		
		return ABSACNode()
		
	
	
	def setSentScore(self, pScore):
		'''
		Sets the value of sentiment score of the node's word
		'''
		
		self.sentScore = pScore
		
	
	
	def getSentScore(self):
		'''
		Returns the sentiment score of the node's word
		'''
		
		return self.sentScore
		
	
	
	def getAspectTerms(self):
		'''
		Returns the aspect terms associated with this node
		
		Note the difference to extractAspectTerms().
		'''
		
		return self.aspectTerms
		
	
	
	def getOEs(self):
		'''
		Returns the opinion expressions associated with this node
		
		Note the difference to extractOEs().
		'''
		
		return self.oe
		
	
	
	def embedAspectTerm(self, pAspectTerm, pEmbedPosition, pOnMismatch = "nothing", pflgExtendToDT = False):
		'''
		Embeds the given aspect term in the proper node in the subtree and
		returns true if the embedding happened
		
		pEmbedPosition specifies where the aspect term should be embedded
		in the subtree. The possible values are:
		- (span)ning-constituent: embedding into a node in subtree which 
		                          spans the aspect term tokens. This may 
		                          cause mismatches where the aspect term 
		                          span is not fully covered by the node
		                          span. pAspectTerm can be set to fix the
		                          issue.
		- (pre)-terminals: embedding into each and every pre-terminal node
		                   in the subtree falling in the aspect term span.
		
		Aspect term is not assigned to terminal nodes.
		
		pOnMismatch specifies the method for handling aspect term/constituent
		mismacth. The possible values are:
		each aspect term covering the term span. The values are:
		- nothing (or none): do not handle the mismatches
		- node (or insert): insert a new node covering the aspect term span
		
		If pflgExtendToDT is set to true, the span of aspect terms which 
		only exclude the determiner of the NP is extended to cover the 
		determiner to reduce the number of mismatches.
		'''
		
		if self.isTerminal():
			return False
		
		if pEmbedPosition.lower().startswith("span") or pEmbedPosition.lower().startswith("const"):
			return self._embedATINSpanConst(pAspectTerm = pAspectTerm, pOnMismatch = pOnMismatch, pflgExtendToDT = pflgExtendToDT)
		elif pEmbedPosition.lower().startswith("pre"):
			self._embedATINPreterminals(pAspectTerm = pAspectTerm, pflgExtendToDT = pflgExtendToDT)
			return True
		else:
			raise Exception("Embedding position option is not recognized: %s" % pEmbedPosition)
		
	
	
	def _embedATINSpanConst(self, pAspectTerm, pOnMismatch, pflgExtendToDT):
		'''
		Embeds the given aspect term in the constituent spanning the aspect
		term tokens
		
		For details, see embedAspectTerm().
		'''
		
		vATSpan = pAspectTerm.getTokenSpan()
		vNodeSpan = self.getTokenSpan()
		vTokenSpanRel = self.getTokenSpanRelation(vATSpan)
		
		# node span and AT span are equal
		if vTokenSpanRel == 0:
			self._plugAspectTerm(pAspectTerm, pInsertATNode = False)
			return True
		
		# AT span is inside the node span
		elif vTokenSpanRel == 1:
			## extending the ST span to determiner if possible (can be done
			## after scanning the children as well; depends on priority
			## given to the fixing method, i.e. inserting AT node or extending 
			## span)
			if pflgExtendToDT:
				if (self.getPreTerminalNodes()[0].getSynTag() == "DT") and (vNodeSpan[0] + 1 == vATSpan[0] and vNodeSpan[1] == vATSpan[1]):
					self._plugAspectTerm(pAspectTerm, pInsertATNode = True if pOnMismatch.lower() in ["node", "insert"] else False)
					return True
			
			for vChild in self.getChildren():
				if vChild._embedATINSpanConst(pAspectTerm, pOnMismatch = pOnMismatch, pflgExtendToDT = pflgExtendToDT):
					return True
			
			## At this stage, there should have been a mismatch between 
			## aspect term span and constituent nodes of the subtree. If
			## inserting an aspect term node is not requested, the end is 
			## here. Otherwise, consecutive children FULLY covering the 
			## aspect term span will be searched for, to insert an aspect
			## term node above. If such full coverage is not found, the 
			## aspect term will not be embedded.
			if pOnMismatch.lower() not in ["node", "insert"]:
				return False
			else:
				vlCoverdChildPos = []
				for i, vChild in enumerate(self.getChildren(), start = 1):
					vChildSpanRel = vChild.getTokenSpanRelation(vATSpan)
					if vChildSpanRel in [-2, -3]:
						## The AT span cuts through a child node span. So
						## the full constituent node coverage cannot be 
						## found; give up.
						return False
					elif vChildSpanRel == -1:
						## The AT span includes the child node span, so
						## the new node should dominate it.
						vlCoverdChildPos.append(i)
				
				# now inserting the new node above children found above 
				if len(vlCoverdChildPos) > 0:
					vNewATNode = self.insertIntermChild(pLabel = "AT", ptDomChildSpan = (vlCoverdChildPos[0], vlCoverdChildPos[-1]))
					vNewATNode._plugAspectTerm(pAspectTerm, pInsertATNode = False)
					return True
		else:
			return False
		
	
	
	def _embedATINPreterminals(self, pAspectTerm, pflgExtendToDT):
		'''
		Embeds the given aspect term in all pre-terminal nodes in this 
		subtree covering the aspect term span
		
		For details, see embedAspectTerm().
		'''
		
		## This algorithm may be inefficient (should be tested) but it is
		## far less complex.
		
		vATSpan = pAspectTerm.getTokenSpan()
		vNodeSpan = self.getTokenSpan()
		
		## embeding aspect term in all pre-terminal nodes in the subtree
		## falling in aspect term span
		for vATTokenPos in range(vATSpan[0], vATSpan[1] + 1):
			if vATTokenPos in range(vNodeSpan[0], vNodeSpan[1] + 1):
				vPreTerminal = self.getPreTerminalNodeAt(vATTokenPos)
				vPreTerminal._plugAspectTerm(pAspectTerm, pInsertATNode = False)
		
		## expanding aspect term span to DT if there is a DT immediately
		## before the span starts AND in the subtree
		if pflgExtendToDT:
			if (self.getTokenSpanRelation((vATSpan[0]-1, vATSpan[0]-1)) == 1):
				vNode = self.getPreTerminalNodeAt(vATSpan[0] - 1)
				if vNode.getLabel() == "DT":
					vNode._plugAspectTerm(pAspectTerm, pInsertATNode = False)
		
	
	
	def _plugAspectTerm(self, pAspectTerm, pInsertATNode):
		'''
		Plugs aspect term to the node
		'''
		
		### modify to first check if the aspect term is already embedded.
		
		if pInsertATNode:
			self.insertParent(pLabel = "AT")
			self.parent.aspectTerms = [pAspectTerm]
		else:
			self.aspectTerms.append(pAspectTerm)
		
	
	
	def _attachATSuffix(self, pSuffix = "_AT", pPosition = "node"):
		'''
		Attaches _AT suffix to the node label and optionally to all nodes
		in the subtree 
		
		pPosition specifies where in the subtree the suffix is to be
		attached:
		- node: only the node itself
		- subtree: all nodes in the subtree
		- pre-terminals: only pre-terminals in the subtree
		- parents: all parents of the node up to the root node (exclusive)
		'''
		
		vPosition = pPosition.lower()
		vNodeLabel = self.getLabel()
		
		if vPosition in ["node", "subtree"]:
			if vNodeLabel != "AT":
				self.setLabel(vNodeLabel + pSuffix)
		
		if vPosition == "subtree":
			for vChild in self.getChildren():
				vChild._attachATSuffix(pSuffix = pSuffix, pPosition = "subtree")
		
		if vPosition.startswith("pre"):
			for vNode in self.getPreTerminalNodes():
				vNode._attachATSuffix(pSuffix = pSuffix, pPosition = "node")
		
		if vPosition == "parents" and not self.isRoot():
			vParent = self.getParent()
			vParentLabel = vParent.getLabel()
			
			if not vParent.isRoot() and vParentLabel != "AT":
				vParent.setLabel(vParentLabel + pSuffix)
				vParent._attachATSuffix(pSuffix = pSuffix, pPosition = "parents")
		
	
	
	def _attachATPolarity(self, pPolarity = '', pPosition = "node"):
		'''
		Attaches aspect term polarity label as suffix to the node label and optionally
		to all nodes in the subtree 
		
		pPosition specifies where in the subtree the suffix is to be
		attached:
		- node: only the node itself
		- subtree: all nodes in the subtree
		- pre-terminals: only pre-terminals in the subtree
		- parents: all parents of the node up to the root node (exclusive)
		'''
		
		vPosition = pPosition.lower()
		vNodeLabel = self.getLabel()
		if pPolarity == '':
			vPolarity = ''.join(["_%s" % at.getPolarity() for at in self.getAspectTerms()])
		else:
			vPolarity = pPolarity
		
		if vPosition in ["node", "subtree"]:
			self.setLabel(vNodeLabel + vPolarity)
		
		if vPosition == "subtree":
			for vChild in self.getChildren():
				vChild._attachATPolarity(pPolarity = vPolarity, pPosition = "subtree")
		
		if vPosition.startswith("pre"):
			for vNode in self.getPreTerminalNodes():
				vNode._attachATPolarity(pPolarity = vPolarity, pPosition = "node")
		
		if vPosition == "parents" and not self.isRoot():
			vParent = self.getParent()
			vParentLabel = vParent.getLabel()
			
			vParent.setLabel(vParentLabel + vPolarity)
			vParent._attachATPolarity(pPolarity = vPolarity, pPosition = "parents")
		
	
	
	def extractAspectTerms(self):
		'''
		Extracts the aspect terms from nodes in the subtree
		
		Note the difference to getAspectTerms().
		'''
		
		vlAspectTerms = self.getAspectTerms()
		
		for vChild in self.getChildren():
			vlAspectTerms += vChild.extractAspectTerms()
		
		## When an aspect term is embedded in all pre-terminals in its span
		## the extract list must be edited to remove duplicates.
		return list(set(vlAspectTerms))
		
	
	
	def decorateAT(self, pdATReprOptions):
		'''
		Decorates the node with aspect term representation according to 
		the options
		
		pdATReprOptions specifies the representation details of the aspect
		terms in the tree. The options involve attaching aspect term suffix,
		inserting apect term node and attaching aspect term polarity, each 
		with several possibilities. The following list the dictionary keys
		with the description of their possible values:
		- suffix: list of following values
		   - node: attach to aspect term node itself
		   - subtree: attach to all nodes in the aspect term subtree
		   - pre-terminal: attach to all pre-terminals in aspect term 
		                    subtree
		   - parents: attach to all parents of the aspect term node up to 
		              the root (exclusive)
		   - parents-partial: attach to all parents of the aspect term 
		                      node up to the root but consider them as 
		                      partial AT nodes and use _ATP instead of _AT
		- node: list of following values
		   - parent: insert AT node as parent of the aspect term node
		   - child: insert AT node as a child of the aspect term node
		- polarity: list of following values
		   - node: attach polarity to the aspect term node
		   - subtree: attach polarity to all nodes in the aspect term subtree
		   - parents: attach to all parents of the aspect term node up to 
		              the root (inclusive)
		
		An empty list for an option means that it must be ignore. For
		example, suffix: [] means that no suffix must be attached. Also
		note that combining some of the values in the option lists does 
		not make an extra effect. For example, polarity: [node, subtree]
		is equal to polarity [subtree].
		'''
		
		vSuffix = "_AT"
		
		if len(self.getAspectTerms()) != 0:
			for vAT in self.getAspectTerms():
				# handling suffix
				if "suffix" in pdATReprOptions:
					if "node" in pdATReprOptions["suffix"] and "subtree" not in pdATReprOptions["suffix"]:
						self._attachATSuffix(pSuffix = vSuffix, pPosition = "node")
					
					if "subtree" in pdATReprOptions["suffix"]:
						self._attachATSuffix(pSuffix = vSuffix, pPosition = "subtree")
					
					if "pre-terminal" in pdATReprOptions["suffix"] and "subtree" not in pdATReprOptions["suffix"]:
						self._attachATSuffix(pSuffix = vSuffix, pPosition = "pre-terminals")
					
					if "parents" in pdATReprOptions["suffix"]:
						self._attachATSuffix(pSuffix = vSuffix, pPosition = "parents")
					
					if "parents-partial" in pdATReprOptions["suffix"]:
						self._attachATSuffix(pSuffix = vSuffix + 'P', pPosition = "parents")
				
				# handling node insertion: only if the node is not AT node itself
				if "node" in pdATReprOptions:
					if "parent" in pdATReprOptions["node"] and self.getLabel() != "AT":
						self.insertParent(pLabel = "AT")
					
					if "child" in pdATReprOptions["node"] and self.getLabel() != "AT":
						self.insertChild(pLabel = "AT", pPosition = self.getChildrenCount() + 1)
				
				# handling polarity
				if "polarity" in pdATReprOptions:
					if "node" in pdATReprOptions["polarity"] and "subtree" not in pdATReprOptions["polarity"]:
						self._attachATPolarity(pPosition = "node")
					
					if "subtree" in pdATReprOptions["polarity"]:
						self._attachATPolarity(pPosition = "subtree")
					
					if "pre-terminals" in pdATReprOptions["polarity"] and "subtree" not in pdATReprOptions["polarity"]:
						self._attachATPolarity(pPosition = "pre-terminals")
					
					if "parents" in pdATReprOptions["polarity"]:
						self._attachATPolarity(pPosition = "parents")
		
		# now the recursive part
		
		for vChild in self.getChildren():
			vChild.decorateAT(pdATReprOptions = pdATReprOptions)
		
	
	
	def embedOpinionExpression(self, pOE, pEmbedPosition, pOnMismatch = "nothing"):
		'''
		Embeds the given opinion expression in the proper node in the subtree and returns true if the embedding happened
		
		pEmbedPosition specifies where the opinion expression should be	embedded in the subtree. The possible values are:
		- (span)ning-constituent: embedding into a node in subtree which spans the OE tokens. This may cause mismatches 
		                          where the OE span is not fully covered by the node span. pOnMismatch can be set to fix
		                          the issue.
		- (pre)-terminals: embedding into each and every pre-terminal node in the subtree falling in the OE span.
		
		pOnMismatch specifies the method for handling expression/constituent mismatch. The possible values are:
		- nothing (or none): do not handle the mismatches
		- node (or insert): insert a new node covering the aspect term span
		'''
		
		if self.isTerminal():
			return False
		
		if pEmbedPosition.lower().startswith("span") or pEmbedPosition.lower().startswith("const"):
			return self._embedOEINSpanConst(pOE = pOE, pOnMismatch = pOnMismatch)
		elif pEmbedPosition.lower().startswith("pre"):
			self._embedOEInPreterminals(pOE = pOE)
			return True
		else:
			raise Exception("Embedding position option is not recognized: %s" % pEmbedPosition)
	
	
	
	def _embedOEINSpanConst(self, pOE, pOnMismatch):
		'''
		Embeds the given opinion expression in the constituent spanning the expression tokens
		
		For details, see embedOE().
		'''
		
		vOESpan = pOE.getTokenSpan()
		vNodeSpan = self.getTokenSpan()
		vTokenSpanRel = self.getTokenSpanRelation(vOESpan)
		
		# node span and OE span are equal
		if vTokenSpanRel == 0:
			self._plugOE(pOE, pInsertOENode = False)
			return True
		
		# OE span is inside the node span
		elif vTokenSpanRel == 1:
			for vChild in self.getChildren():
				if vChild._embedOEINSpanConst(pOE, pOnMismatch = pOnMismatch):
					return True
			
			## At this stage, there should have been a mismatch between OE span and constituent nodes of the subtree. If
			## inserting an OE node is not requested, the end is here. Otherwise, consecutive children FULLY covering the 
			## OE span will be searched for, to insert an OE node above. If such full coverage is not found, the OE will
			## not be embedded.
			if pOnMismatch.lower() not in ["node", "insert"]:
				return False
			else:
				vlCoverdChildPos = []
				for i, vChild in enumerate(self.getChildren(), start = 1):
					vChildSpanRel = vChild.getTokenSpanRelation(vOESpan)
					if vChildSpanRel in [-2, -3]:
						## The OE span cuts through a child node span. So the full constituent node coverage cannot be 
						## found; give up.
						return False
					elif vChildSpanRel == -1:
						## The OE span includes the child node span, so the new node should dominate it.
						vlCoverdChildPos.append(i)
				
				# now inserting the new node above children found above 
				if len(vlCoverdChildPos) > 0:
					vNewOENode = self.insertIntermChild(pLabel = "OE", ptDomChildSpan = (vlCoverdChildPos[0], vlCoverdChildPos[-1]))
					vNewOENode._plugOE(pOE, pInsertOENode = False)
					return True
		else:
			return False
	
	
	
	def _embedOEInPreterminals(self, pOE):
		'''
		Embeds the given opinin expression in all pre-terminal nodes in this subtree covering the opinion expression span
		
		For details, see embedOE().
		'''
		
		## This algorithm may be inefficient (should be tested) but it is far less complex.
		
		vOESpan = pOE.getTokenSpan()
		vNodeSpan = self.getTokenSpan()
		
		## embedding OE in all pre-terminal nodes in the subtree falling in OE span 
		for vOETokenPos in range(vOESpan[0], vOESpan[1] + 1):
			if vOETokenPos in range(vNodeSpan[0], vNodeSpan[1] + 1):
				vPreTerminal = self.getPreTerminalNodeAt(vOETokenPos)
				vPreTerminal._plugOE(pOE, pInsertOENode = False)
		
	
	
	def _plugOE(self, pOE, pInsertOENode):
		'''
		Plugs opinion expression to the node
		'''
		
		### modify to first check if the OE is already embedded.
		
		if pInsertOENode:
			self.insertParent(pLabel = "OE")
			self.parent.oe = [pOE]
		else:
			self.oe.append(pOE)
	
	
	
	def _attachOESuffix(self, pSuffix = "_OE", pPosition = "node"):
		'''
		Attaches _OE suffix to the node label and optionally to all nodes in the subtree 
		
		pPosition specifies where in the subtree the suffix is to be
		attached:
		- node: only the node itself
		- subtree: all nodes in the subtree
		- pre-terminals: only pre-terminals in the subtree
		- parents: all parents of the node up to the root node (exclusive)
		'''
		
		vPosition = pPosition.lower()
		vNodeLabel = self.getLabel()
		
		if vPosition in ["node", "subtree"]:
			if vNodeLabel != "OE":
				self.setLabel(vNodeLabel + pSuffix)
		
		if vPosition == "subtree":
			for vChild in self.getChildren():
				vChild._attachOESuffix(pSuffix = pSuffix, pPosition = "subtree")
		
		if vPosition.startswith("pre"):
			for vNode in self.getPreTerminalNodes():
				vNode._attachOESuffix(pSuffix = pSuffix, pPosition = "node")
		
		if vPosition == "parents" and not self.isRoot():
			vParent = self.getParent()
			vParentLabel = vParent.getLabel()
			
			if not vParent.isRoot() and vParentLabel != "OE" and not vParentLabel.endswith(pSuffix):
				vParent.setLabel(vParentLabel + pSuffix)
				vParent._attachOESuffix(pSuffix = pSuffix, pPosition = "parents")
	
	
	
	def _attachOEPolarity(self, pPolarity = '', pPosition = "node"):
		'''
		Attaches opinion expression polarity label as suffix to the node label and optionally to all nodes in the subtree 
		
		pPosition specifies where in the subtree the suffix is to be attached:
		- node: only the node itself
		- subtree: all nodes in the subtree
		- pre-terminals: only pre-terminals in the subtree
		- parents: all parents of the node up to the root node (exclusive)
		'''
		
		vPosition = pPosition.lower()
		vNodeLabel = self.getLabel()
		if pPolarity == '':
			vPolarity = ''.join(["_%s" % oe.getPolarity() for oe in self.getOEs()])
		else:
			vPolarity = pPolarity
		
		if vPosition in ["node", "subtree"]:
			self.setLabel(vNodeLabel + vPolarity)
		
		if vPosition == "subtree":
			for vChild in self.getChildren():
				vChild._attachOEPolarity(pPolarity = vPolarity, pPosition = "subtree")
		
		if vPosition.startswith("pre"):
			for vNode in self.getPreTerminalNodes():
				vNode._attachOEPolarity(pPolarity = vPolarity, pPosition = "node")
		
		if vPosition == "parents" and not self.isRoot():
			vParent = self.getParent()
			vParentLabel = vParent.getLabel()
			
			vParent.setLabel(vParentLabel + vPolarity)
			vParent._attachOEPolarity(pPolarity = vPolarity, pPosition = "parents")
	
	
	
	def extractOEs(self):
		'''
		Extracts the opinion expressions from nodes in the subtree
		
		Note the difference to getOEs().
		'''
		
		vlOEs = self.getOEs()
		
		for vChild in self.getChildren():
			vlOEs += vChild.extractOEs()
		
		## When an OE is embedded in all pre-terminals in its span the extract list must be edited to remove duplicates.
		return list(set(vlOEs))
	
	
	
	def decorateOE(self, pdOEReprOptions):
		'''
		Decorates the node with opinion expression representation according to the options
		
		pdOEReprOptions specifies the representation details of the OEs in the tree. The options involve attaching OE
		suffix,	inserting OE node and attaching OE polarity, each with several possibilities. The following list the 
		dictionary keys with the description of their possible values:
		- suffix: list of following values
		   - node: attach to OE node itself
		   - subtree: attach to all nodes in the OE subtree
		   - pre-terminal: attach to all pre-terminals in OE subtree
		   - parents: attach to all parents of the OE node up to the root (exclusive)
		   - parents-partial: attach to all parents of the OE node up to the root but consider them as partial OE nodes
		                      and use _OEP instead of _OE
		- node: list of following values
		   - parent: insert OE node as parent of the opinion expression node
		   - child: insert OE node as a child of the opinion expression node
		- polarity: list of following values
		   - node: attach polarity to the opinion expression node
		   - subtree: attach polarity to all nodes in the opinion expression subtree
		   - parents: attach to all parents of the opinion expression node up to the root (inclusive)
		
		An empty list for an option means that it must be ignore. For example, suffix: [] means that no suffix must be 
		attached. Also note that combining some of the values in the option lists does not make an extra effect. For 
		example, polarity: [node, subtree] is equal to polarity [subtree].
		'''
		
		vSuffix = "_OE"
		
		if len(self.getOEs()) != 0:
			for vAT in self.getOEs():
				# handling suffix
				if "suffix" in pdOEReprOptions:
					if "node" in pdOEReprOptions["suffix"] and "subtree" not in pdOEReprOptions["suffix"]:
						self._attachOESuffix(pSuffix = vSuffix, pPosition = "node")
					
					if "subtree" in pdOEReprOptions["suffix"]:
						self._attachOESuffix(pSuffix = vSuffix, pPosition = "subtree")
					
					if "pre-terminal" in pdOEReprOptions["suffix"] and "subtree" not in pdOEReprOptions["suffix"]:
						self._attachOESuffix(pSuffix = vSuffix, pPosition = "pre-terminals")
					
					if "parents" in pdOEReprOptions["suffix"]:
						self._attachOESuffix(pSuffix = vSuffix, pPosition = "parents")
					
					if "parents-partial" in pdOEReprOptions["suffix"]:
						self._attachOESuffix(pSuffix = vSuffix + 'P', pPosition = "parents")
				
				# handling node insertion: only if the node is not OE node itself
				if "node" in pdOEReprOptions:
					if "parent" in pdOEReprOptions["node"] and self.getLabel() != "OE":
						self.insertParent(pLabel = "OE")
					
					if "child" in pdOEReprOptions["node"] and self.getLabel() != "OE":
						self.insertChild(pLabel = "OE", pPosition = self.getChildrenCount() + 1)
				
				# handling polarity
				if "polarity" in pdOEReprOptions:
					if "node" in pdOEReprOptions["polarity"] and "subtree" not in pdOEReprOptions["polarity"]:
						self._attachOEPolarity(pPosition = "node")
					
					if "subtree" in pdOEReprOptions["polarity"]:
						self._attachOEPolarity(pPosition = "subtree")
					
					if "pre-terminals" in pdOEReprOptions["polarity"] and "subtree" not in pdOEReprOptions["polarity"]:
						self._attachOEPolarity(pPosition = "pre-terminals")
					
					if "parents" in pdOEReprOptions["polarity"]:
						self._attachOEPolarity(pPosition = "parents")
		
		# now the recursive part
		
		for vChild in self.getChildren():
			vChild.decorateOE(pdOEReprOptions = pdOEReprOptions)
		
	
	
	def decorateSS(self, pdSSReprOptions):
		'''
		NOTE: this method is obsolete and kept for backward compatibility 
		plus those features which are not implemented in the tree kernel
		alternative (ABSACKTree). Use ABSACKTree and related formats instead.
		
		Decorates the node with sentiment score representation according 
		to the options
		
		pdSSReprOptions specifies the representation details of the sentiment 
		scores and is a dictionary with the following as possible keys:
		- node: inserting a new node for the score in:
		   - parent: insert score node as parent of the terminal
		   - child: insert score node as child of the terminal
		   - grand-parent: insert score node as grand parent of the terminal,
		                   i.e. as parent of pre-terminal (POS tag)
		- replace: replacing the label of a node with the sentiment score:
		   - terminal: replacing terminal word with sentiment score
		   - 
		- neutral: score used for neutral nodes:
		   - '': use nothing 
		   - 0 or anything else
		
		NOTE: unlike as usual, this function is not recursive and works
		by first collecting a list of terminals under the node.
		'''
		
		for vTerminal in self.getTerminalNodes():
			vScore = vTerminal.getSentScore()
			
			if vScore is None:
				if "neutral" in pdSSReprOptions:
					if pdSSReprOptions["neutral"].strip() == '':
						continue
					else:
						vScore = pdSSReprOptions["neutral"]
				else:
					vScore = 0
			
			if "node" in pdSSReprOptions:
				if pdSSReprOptions["node"].lower().startswith("parent"):
					vTerminal.insertParent(pLabel = vScore)
				elif pdSSReprOptions["node"].lower().startswith("child"):
					vTerminal.insertChild(pLabel = vScore, pPosition = 1)
				elif pdSSReprOptions["node"].lower().startswith("grand"):
					vTerminal.getParent().insertParent(pLabel = vScore)
			
			if "replace" in pdSSReprOptions:
				if pdSSReprOptions["replace"].lower().startswith("terminal"):
					vTerminal.setTerminalForm(str(vScore))
	
	
	
	def modifyNPStruct(self):
		'''
		Modifies the structure of the noun phrases in order to avoid
		term/constituent mismatch caused by a flat NP structure
		'''
		
		if self.getSynTag() == "NP":
			vChildLabelSeq = ' '.join(self.getChildrenTags())
			
			if re.search("^(DT|PRP\$) NN[A-Z]* NN[A-Z]*$", vChildLabelSeq):
				self.insertIntermChild("NP", (2, 3))
				return
			
			if re.search("^(DT|PRP\$) JJ[A-Z]* NN[A-Z]*$", vChildLabelSeq):
				self.insertIntermChild("ADJP", (2, 3))
				return
			
			if re.search("^(DT|PRP\$) VBG NN[A-Z]*$", vChildLabelSeq):
				self.insertIntermChild("NP", (2, 3))
				return
			
			if re.search("^(DT|PRP\$) NN[A-Z]* NN[A-Z]* NN[A-Z]*$", vChildLabelSeq):
				self.insertIntermChild("NP", (2, 4))
				return
			
			if re.search("^(DT|PRP\$) JJ[A-Z]* NN[A-Z]* NN[A-Z]*$", vChildLabelSeq):
				self.insertIntermChild("NP", (3, 4))
				self.insertIntermChild("NP", (2, 3))
				return
			
			if re.search("^(DT|PRP\$) ADJP NN[A-Z]*$", vChildLabelSeq):
				self.insertIntermChild("NP", (2, 3))
				return
			
			if re.search("^(DT|PRP\$) CD NN[A-Z]* NN[A-Z]*$", vChildLabelSeq):
				self.insertIntermChild("NP", (2, 4))
				return
			
			if re.search("^(DT|PRP\$) JJ[A-Z]* JJ[A-Z]* NN[A-Z]*$", vChildLabelSeq):
				self.insertIntermChild("NP", (3, 4))
				self.insertIntermChild("NP", (2, 3))
				return
		
		for vChild in self.getChildren():
			vChild.modifyNPStruct()
	
	
	
	def loadSentScores(self, plSentScores, pPropagation = None):
		'''
		Loads sentiment scores to nodes in the subtree
		
		The sentiment scores are given in a list which corresponds to the
		terminal (i.e. token) list.
		
		Neutral score will be used instead of None for words (and nodes when
		propagating) without a sentiment score (None). 
		
		Propagation argument specifies the method by which the score are 
		propagated from terminals up to this node. None means scores
		are only assigned to the terminal nodes, thus they are 0 for the 
		phrase nodes. The following are the possible methods:
		- sum: nodes score is the sum of its children score
		- vote: node score is the dominant positive or negative score in the 
		        children nodes (i.e. more +1: score is +1, more -1: score -1) 
		'''
		
		vTokenSpan = self.getTokenSpan()
		
		if self.isTerminal():
			# sanity check
			if vTokenSpan[0] != vTokenSpan[1]:
				raise Exception("Either the node is not terminal or its span is wrong: %s" % self)
			else:
				self.setSentScore(plSentScores[vTokenSpan[0] - 1])
		else:
			vlChildrenSSores = []
			for vChild in self.getChildren():
				vlChildrenSSores.append(vChild.loadSentScores(plSentScores, pPropagation))
			
			# calculating the sentiment score of the node based on its children's (propagation)
			if pPropagation is not None:
				if pPropagation.lower() == "sum":
					self.setSentScore(sum(vlChildrenSSores))
				elif pPropagation.lower() == "vote":
					self.setSentScore(self._getDominantSentiment(vlChildrenSSores))
		
		return self.getSentScore()
	
	
	
	def _getDominantSentiment(self, plScores):
		'''
		Returns +1 or -1 whichever is dominant in the given list of sentiment 
		scores
		
		If the same number of both sentiment scores exist, 0 is returned.
		'''
		
		vPosCount = 0
		vNegCount = 0
		
		for vScore in plScores:
			if vScore == 1:
				vPosCount += 1
			elif vScore == -1:
				vNegCount += 1
		
		if vPosCount > vNegCount:
			return 1
		elif vPosCount < vNegCount:
			return -1
		else:
			return 0
		
	
	
	def extractTopmostVP(self, pSpan):
		'''
		Extracts and returns the topmost verb phrase node which overlaps the given span in the node subtree
		
		Overlap means that the given span and the span of the VP must not ne disjoint. So, left and right crossing will
		also be considered.
		'''
		
		if self.getSynTag() == 'VP':
			vSpanRel = self.getTokenSpanRelation(pSpan)
			if vSpanRel != -4:
				return self
			else:
				return None 
		else:
			for vChild in self.getChildren():
				vTopVP =  vChild.extractTopmostVP(pSpan)
				if vTopVP is not None:
					return vTopVP
		
		return None
		
	
	


class ABSADTree(depparse.DepTree):
	'''
	Class for dependency parse tree of a ABSA sentence
	'''
	
	
	def __init__(self, pLanguage = ''):
		'''
		Constructor
		'''
		
		depparse.DepTree.__init__(self, pLanguage = pLanguage)
		
	
	
	def loadFromDepTree(self, pDepTree):
		'''
		Loads the tree from  DepTree object
		'''
		
		# 1. nodes
		
		self.nodes = []
		
		for vNode in pDepTree.nodes:
			self.nodes.append(ABSADNode())
			self.nodes[-1].loadFromDepNode(pABSADTree = self, pDepNode = vNode)
		
		# 2. SRL
		
		self.srl = pDepTree.srl
		
		# 3. language
		
		self.language = pDepTree.language
		
	
	
	def _createNewTree(self, pLanguage = ''):
		'''
		Creates and returns a new tree
		'''
		
		return ABSADTree(pLanguage = pLanguage)
		
	
	
	def embedAspectTerms(self, plAspectTerms):
		'''
		Embeds the aspect terms in the tree
		'''
		
		for vAT in plAspectTerms:
			self.embedAspectTerm(pAspectTerm = vAT)
		
	
	
	def embedAspectTerm(self, pAspectTerm):
		'''
		Embeds the aspect term in the tree
		
		NOTE: This method is not redundant to embedAspectTerms. Its main
		purpose is to be called by AspectTerm.
		'''
		
		vATSpan = pAspectTerm.getTokenSpan()
		
		for vNode in self.getNodes():
			vNodePos = vNode.getPosition()
			
			if vATSpan[0] <= vNodePos <= vATSpan[1]:
				vNode.embedAspectTerm(pAspectTerm = pAspectTerm)
		
	
	
	def decorateAT(self, pdATReprOptions = {}):
		'''
		Decorates the node with aspect term representation according to 
		the options
		
		pdATReprOptions specifies the representation details of the aspect
		terms in the tree. See ABSADNode.decorateAT() for details.
		
		NOTE: unlike in constituency tree decoration, AT node insertion 
		is not done for the dependency tree. The reason is that in dependency
		tree, there is a one to one relation between the nodes and sentence
		tokens. Inserting a node causes a mismatch between the surface 
		form of the sentence and its dependency tree. Also, concepts like
		word form, POS tag and dependency relation are not meaningful in
		an inserted AT node. Instead, the aspect term representation using
		inserted AT nodes can be done at the PTB bracketing format representation
		level of the tree. See ABSADTree.generateDepKTree(). Note that
		the current design of this method in terms of the argument passed
		to it may not be coherent and meaningful, because what is expected
		from an aspect term representation parameter (pdATReprOptions) is
		to carry all the setting required for this purpose, not only the 
		part concerned with suffix and polarity attachement. This may be 
		handled in the future.
		'''
		
		for vNode in self.getNodes():
			vNode.decorateAT(pdATReprOptions = pdATReprOptions)
		
	
	
	def embedOpinionExpressions(self, plOEs):
		'''
		Embeds the opinion expressions in the tree
		'''
		
		for vOE in plOEs:
			self.embedOpinionExpression(pOE = vOE)
		
	
	
	def embedOpinionExpression(self, pOE):
		'''
		Embeds the opinion expression in the tree
		
		NOTE: This method is not redundant to embedOpinionExpressions. Its main
		purpose is to be called by OpinionExpression.
		'''
		
		vOESpan = pOE.getTokenSpan()
		
		for vNode in self.getNodes():
			vNodePos = vNode.getPosition()
			
			if vOESpan[0] <= vNodePos <= vOESpan[1]:
				vNode.embedOpinionExpression(pOE = pOE)
		
	
	
	def decorateOE(self, pdOEReprOptions = {}):
		'''
		Decorates the node with opinion expression representation according 
		to the options
		
		pdOEReprOptions specifies the representation details of the OE in 
		the tree. See ABSADNode.decorateOE() for details.
		
		NOTE: see the same method for aspect term for further documentation. 
		'''
		
		for vNode in self.getNodes():
			vNode.decorateOE(pdOEReprOptions = pdOEReprOptions)
		
	
	
	def loadSentScores(self, plSentScores):
		'''
		Loads sentiment scores to tree nodes
		
		The sentiment scores are given in a list which corresponds to the
		tree node list.
		'''
		
		for vNode, vSentScore in zip(self.getNodes(), plSentScores):
			vNode.setSentScore(vSentScore)
		
	
	
	def generateDepKTree(self, pFormat = "(rel form)", pdOptions = {}):
		'''
		Generates the dependency tree representation in PTB bracketing 
		for tree kernels
		
		pdOptions provides options specific to each format.
		'''
		
		vDepKTree = ABSADKTree(pDepTree = self)
		
		return vDepKTree.generateDepKTree(pNode = "root", pFormat = pFormat, pdOptions = pdOptions)
		
	
	
	def extractAT2OEDepRelPath(self, pAT, pOE):
		'''
		Extracts and returns the dependency path from the given aspect term to the given opinion expression on the tree
		
		The path is a string containing the dependency labels and > and < representing up move and down move in the tree
		respectively.
		'''
		
		## NOTE: to speed things up, this method does not use/need aspect terms and opinion expressions embedded in the
		## tree beforehand. It only heuristically finds a node for them based on their span.
		
		vATSpan = pAT.getTokenSpan()
		vOESpan = pOE.getTokenSpan()
		
		# finding AT node: the first node in the span (heuristic)
		vATNode = self.getNode(vATSpan[0])
		
		# finding OE node: the first node in the span (heuristic)
		vOENode = self.getNode(vOESpan[0])
		
		return vATNode.getDepRelPathTo(vOENode)
		
	
	
	def extractAT2OEDepPathWords(self, pAT, pOE):
		'''
		Extracts and returns list of the words in the dependency path from the given aspect term to the given opinion 
		expression on the tree
		'''
		
		## NOTE: to speed things up, this method does not use/need aspect terms and opinion expressions embedded in the
		## tree beforehand. It only heuristically finds a node for them based on their span.
		
		vATSpan = pAT.getTokenSpan()
		vOESpan = pOE.getTokenSpan()
		
		# finding AT node: the first node in the span (heuristic)
		vATNode = self.getNode(vATSpan[0])
		
		# finding OE node: the first node in the span (heuristic)
		vOENode = self.getNode(vOESpan[0])
		
		return vATNode.getWordsInPathTo(vOENode)
		
	
	
	def extractAT2RootPathTree(self, pAT, pNodeContent = "rel"):
		'''
		Extracts and returns the dependency path from the given aspect term to the tree root in the form of a unary tree
		
		The output is a string containing the unary tree with nodes being determined by the pNodeContent option as follows:
		- "rel": dependency relation
		- "word": word form
		- "POS": POS tag
		- "polarity": polarity scores
		'''
		
		## NOTE: to speed things up, this method does not use/need aspect terms embedded in the tree beforehand. It only 
		## heuristically finds a node for them based on their span.
		
		vATSpan = pAT.getTokenSpan()
		
		# finding AT node: the first node in the span (heuristic)
		vATNode = self.getNode(vATSpan[0])
		
		if pNodeContent.lower().startswith("rel"):		
			vlTreeNodeContent = [n.getDepRel() for n in vATNode.getPathTo(self.root)[0]]
		elif pNodeContent.lower().startswith("word"):
			vlTreeNodeContent = [n.getForm() for n in vATNode.getPathTo(self.root)[0]]
		elif pNodeContent.lower().startswith("pos"):
			vlTreeNodeContent = [n.getPOSTag() for n in vATNode.getPathTo(self.root)[0]]
		elif pNodeContent.lower().startswith("polarity"):
			vlTreeNodeContent = [str(n.getSentScore()) if n.getSentScore() is not None else '0' for n in vATNode.getPathTo(self.root)[0]]
		else:
			raise Exception("%s is not a valid node content type!" % pNodeContent)
		
		vTree = tk.NGramKTree(plTokens = vlTreeNodeContent)
		return vTree.generateNGramKTree(pFormat = "binary")
		
	
	


class ABSADNode(depparse.DepNode):
	'''
	Class for dependency parse node of a ABSA sentence
	'''
	
	
	def __init__(self, pABSADTree = None, pForm = "", pPosition = 0, plHeadDeps = None, plDependents = None, pPOSTag = "", plPredRoles = None, plAspectTerms = [], plOE = [], pSentScore = None):
		'''
		Constructor
		'''
		
		depparse.DepNode.__init__(self, pDepTree = pABSADTree, pForm = pForm, pPosition = pPosition, plHeadDeps = plHeadDeps, plDependents = plDependents, pPOSTag = pPOSTag, plPredRoles = plPredRoles)
		
		# list of type AspectTerm
		self.aspectTerms = plAspectTerms
		
		# sentiment score
		self.sentScore = pSentScore
		
		# list of type OpinionExpression
		self.oe = plOE
		
	
	
	def loadFromDepNode(self, pABSADTree, pDepNode):
		'''
		Loads the node data from DepNode object
		'''
		
		self.depTree = pABSADTree                   # ABSA dependency tree the node belongs to
		self.form = pDepNode.form                   # token surface form
		self.position = pDepNode.position           # token position in the sentence
		self.headDeps = pDepNode.headDeps           # list of head and dependency tuples
		self.dependents = pDepNode.dependents       # children
		self.posTag = pDepNode.posTag               # POS tag
		self.predRoles = pDepNode.predRoles         # list of (predicate position, semantic role) tuples
		
	
	
	def _createNewNode(self, pDepTree = None, pForm = "", pPosition = 0, plHeadDeps = None, plDependents = None, pPOSTag = "", plPredRoles = None, plAspectTerms = [], plOE = [], pSentScore = None):
		'''
		Creates and returns an new node
		'''
		
		return ABSADNode(pABSADTree = pDepTree,
		                 pForm = pForm,
		                 pPosition = pPosition,
		                 plHeadDeps = plHeadDeps[:],
		                 plDependents = plDependents[:],
		                 pPOSTag = pPOSTag,
		                 plPredRoles = plPredRoles[:],
		                 plAspectTerms = plAspectTerms[:],
		                 plOE = self.oe[:],
		                 pSentScore = pSentScore)
		
	
	
	def deepCopy(self, pDepTree):
		'''
		NOTE: it seems python deepcopy() works better. Try the idea before
		using this method.
		
		Creates and returns a new dependency node which is a deep copy of 
		the current node
		'''
		
		return self._createNewNode(pDepTree = pDepTree,
		                           pForm = self.form,
		                           pPosition = self.position,
		                           plHeadDeps = self.headDeps[:],
		                           plDependents = self.dependents[:],
		                           pPOSTag = self.posTag,
		                           plPredRoles = self.predRoles[:],
		                           plAspectTerms = self.aspectTerms[:],
		                           plOE = self.oe[:],
		                           pSentScore = self.sentScore)
		
	
	
	def getAspectTerms(self):
		'''
		Returns the aspect terms associated with this node
		'''
		
		return self.aspectTerms
		
	
	
	def getOEs(self):
		'''
		Returns the opinion expressions associated with this node
		'''
		
		return self.oe
		
	
	
	def embedAspectTerm(self, pAspectTerm):
		'''
		Embeds the given aspect term into the node
		
		The aspect term is not assigned to the root node.
		'''
		
		if not self.isRoot():
			self._plugAspectTerm(pAspectTerm)
		
	
	
	def _plugAspectTerm(self, pAspectTerm):
		'''
		Plugs aspect term to the node
		'''
		
		self.aspectTerms.append(pAspectTerm)
		
	
	
	def decorateAT(self, pdATReprOptions = {}):
		'''
		Decorates the node with aspect term representation according to 
		the options
		
		pdATReprOptions specifies the representation details of the aspect
		terms in the tree. The options involve attaching aspect term suffix
		and attaching aspect term polarity, each with several possibilities. 
		The following list the dictionary keys with the description of 
		their possible values:
		- suffix: list of (attach_in, attach_to) where:
		   - attach_in can be:
		      - node: attach only in this node
		      - subtree: attach in all nodes in the aspect term subtree of 
		                 dependents
		      - parents: IMPLEMENT: attach in all parents (heads, heads of heads, and 
		              so on) of the aspect term node up to the root (inclusive)
		   - attach_to can be:
		      - form: attach to node form
		      - rel: attach to node's dependency relation(s) to head(s)
		      - pos: attach to node pos tag
		- polarity: list of (attach_in, attach_to) where:
		   - attach_in can be:
		      - node: attach only in this node
		      - subtree: attach in all nodes in the aspect term subtree of 
		                 dependents
		      - parents: IMPLEMENT: attach in all parents (heads, heads of heads, and 
		              so on) of the aspect term node up to the root (inclusive)
		   - attach_to can be:
		      - form: attach to node form
		      - rel: attach to node's dependency relation(s) to head(s)
		      - pos: attach to node pos tag
		
		An empty list for an option means that it must be ignored. For
		example, suffix: [] means that no suffix must be attached. 
		
		NOTE: unlike in constituency tree decoration, AT node insertion 
		is not done for the dependency tree. The reason is that in dependency
		tree, there is a one to one relation between the nodes and sentence
		tokens. Inserting a node causes a mismatch between the surface 
		form of the sentence and its dependency tree. Also, concepts like
		word form, POS tag and dependency relation are not meaningful in
		an inserted AT node. Instead, the aspect term representation using
		inserted AT nodes can be done at the PTB bracketing format representation
		level of the tree. See ABSADTree.generateDepKTree(). Note that
		the current design of this method in terms of the argument passed
		to it may not be coherent and meaningful, because what is expected
		from an aspect term representation parameter (pdATReprOptions) is
		to carry all the setting required for this purpose, not only the 
		part concerned with suffix and polarity attachement. This may be 
		handled in the future.
		
		'''
		
		vSuffix = "_AT"
		
		if len(self.getAspectTerms()) != 0:
			for vAT in self.getAspectTerms():
				# handling suffix
				if "suffix" in pdATReprOptions:
					# summarizing the options for suffix
					vdSuffixOptions = {}
					for vIn, vTo in pdATReprOptions["suffix"]:
						if vIn not in vdSuffixOptions:
							vdSuffixOptions[vIn] = [vTo]
						elif vTo not in vdSuffixOptions[vIn]:
							vdSuffixOptions[vIn].append(vTo)
					
					if "node" in vdSuffixOptions:
						self._attachATSuffix(pSuffix = vSuffix, pAttachIn = "node", plAttachTo = vdSuffixOptions["node"])
					
					if "subtree" in vdSuffixOptions:
						self._attachATSuffix(pSuffix = vSuffix, pAttachIn = "subtree", plAttachTo = vdSuffixOptions["subtree"])
				
				# handling polarity
				if "polarity" in pdATReprOptions:
					# summarizing the options for polarity
					vdPolarityOptions = {}
					for vIn, vTo in pdATReprOptions["polarity"]:
						if vIn not in vdPolarityOptions:
							vdPolarityOptions[vIn] = [vTo]
						elif vTo not in vdPolarityOptions[vIn]:
							vdPolarityOptions[vIn].append(vTo)
					
					if "node" in vdPolarityOptions:
						self._attachATPolarity(pAttachIn = "node", plAttachTo = vdPolarityOptions["node"])
					
					if "subtree" in vdPolarityOptions:
						self._attachATPolarity(pAttachIn = "subtree", plAttachTo = vdPolarityOptions["subtree"])
		
	
	
	def _attachATSuffix(self, pSuffix = "_AT", pAttachIn = "node", plAttachTo = "rel"):
		'''
		Attaches aspect term suffix to the specified locations in the tree
		to which the node belongs
		
		pAttachIn can be:
		 - node: attach only in this node
		 - subtree: attach in all nodes in the aspect term subtree of 
		            dependents
		 - parents: IMPLEMENT: attach in all parents (heads, heads of 
		            heads, and so on) of the aspect term node up to the
		            root (inclusive)
		plAttachTo can be:
		 - form: attach to node form
		 - rel: attach to node's dependency relation(s) to head(s)
		 - pos: attach to node pos tag
		'''
		
		vAttachIn = pAttachIn.lower()
		
		if vAttachIn in ["node", "subtree"]:
			self._attachATSuffixTo(pSuffix = pSuffix, plAttachTo = plAttachTo)
		
		if vAttachIn == "subtree":
			for vDependent in self.getDependentNodes():
				vDependent._attachATSuffix(pSuffix = pSuffix, pAttachIn = "subtree", plAttachTo = plAttachTo)
		
		if vAttachIn == "parents" and not self.isRoot():
			# NOTE: only the first head is used for multihead cases
			vHead = self.getHeadNode()
			
			if not vHead.isRoot():
				vHead._attachATSuffixTo(pSuffix = pSuffix, pAttachTo = vAttachTo)
				vHead._attachATSuffix(pSuffix = pSuffix, pAttachIn = "parents", plAttachTo = plAttachTo)
		
	
	
	def _attachATSuffixTo(self, pSuffix, plAttachTo):
		'''
		Attaches aspect term suffix to the specified locations in the node
		
		plAttachTo can be a list of:
		 - form: attach to node form
		 - rel: attach to node's dependency relation(s) to head(s)
		 - pos: attach to node pos tag
		'''
		
		vlPositions = [p.lower() for p in plAttachTo]
		
		if "form" in vlPositions:
			vForm = self.getForm()
			if vForm != "AT":
				self.setForm(vForm + pSuffix)
		
		if "rel" in vlPositions:
			for vHead in self.getHeadPos():
				vRel = self.getDepRel(pHead = vHead)
				self.setDepRel(pRel = vRel + pSuffix, pHead = vHead)
		
		if "pos" in vlPositions:
			vPOS = self.getPOSTag()
			self.setPOSTag(vPOS + pSuffix)
		
	
	
	def _attachATPolarity(self, pPolarity = '', pAttachIn = "node", plAttachTo = "rel"):
		'''
		Attaches polarity label as suffix to the node label and optionally
		to all nodes in the subtree 
		
		- pAttachIn can be:
		   - node: attach only in this node
		   - subtree: attach in all nodes in the aspect term subtree of 
		              dependents
		   - parents: IMPLEMENT: attach in all parents (heads, heads of heads, and 
		              so on) of the aspect term node up to the root (inclusive)
		- plAttachTo can be:
		   - form: attach to node form
		   - rel: attach to node's dependency relation(s) to head(s)
		   - pos: attach to node pos tag
		'''
		
		# implement
		
		pass
		
	
	
	def embedOpinionExpression(self, pOE):
		'''
		Embeds the given opinion expression into the node
		
		The OE is not assigned to the root node.
		'''
		
		if not self.isRoot():
			self._plugOE(pOE)
		
	
	
	def _plugOE(self, pOE):
		'''
		Plugs opinion expression to the node
		'''
		
		self.oe.append(pOE)
		
	
	
	def decorateOE(self, pdOEReprOptions = {}):
		'''
		Decorates the node with opinion expression representation according to 
		the options
		
		pdOEReprOptions specifies the representation details of the OE in the 
		tree. The options involve attaching OE suffix and attaching its polarity,
		each with several possibilities. The following list the dictionary keys 
		with the description of their possible values:
		- suffix: list of (attach_in, attach_to) where:
		   - attach_in can be:
		      - node: attach only in this node
		      - subtree: attach in all nodes in the OE subtree of dependents
		      - parents: IMPLEMENT: attach in all parents (heads, heads of heads,
		                 and so on) of the aspect term node up to the root (inclusive)
		   - attach_to can be:
		      - form: attach to node form
		      - rel: attach to node's dependency relation(s) to head(s)
		      - pos: attach to node pos tag
		- polarity: list of (attach_in, attach_to) where:
		   - attach_in can be:
		      - node: attach only in this node
		      - subtree: attach in all nodes in the OE subtree of dependents
		      - parents: IMPLEMENT: attach in all parents (heads, heads of heads,
		                and so on) of the aspect term node up to the root (inclusive)
		   - attach_to can be:
		      - form: attach to node form
		      - rel: attach to node's dependency relation(s) to head(s)
		      - pos: attach to node pos tag
		
		An empty list for an option means that it must be ignored. For
		example, suffix: [] means that no suffix must be attached. 
		
		NOTE: see the same method for aspect term for further documentation. 
		'''
		
		vSuffix = "_OE"
		
		if len(self.getOEs()) != 0:
			for vAT in self.getOEs():
				# handling suffix
				if "suffix" in pdOEReprOptions:
					# summarizing the options for suffix
					vdSuffixOptions = {}
					for vIn, vTo in pdOEReprOptions["suffix"]:
						if vIn not in vdSuffixOptions:
							vdSuffixOptions[vIn] = [vTo]
						elif vTo not in vdSuffixOptions[vIn]:
							vdSuffixOptions[vIn].append(vTo)
					
					if "node" in vdSuffixOptions:
						self._attachOESuffix(pSuffix = vSuffix, pAttachIn = "node", plAttachTo = vdSuffixOptions["node"])
					
					if "subtree" in vdSuffixOptions:
						self._attachOESuffix(pSuffix = vSuffix, pAttachIn = "subtree", plAttachTo = vdSuffixOptions["subtree"])
				
				# handling polarity
				if "polarity" in pdOEReprOptions:
					# summarizing the options for polarity 
					vdPolarityOptions = {}
					for vIn, vTo in pdOEReprOptions["polarity"]:
						if vIn not in vdPolarityOptions:
							vdPolarityOptions[vIn] = [vTo]
						elif vTo not in vdPolarityOptions[vIn]:
							vdPolarityOptions[vIn].append(vTo)
					
					if "node" in vdPolarityOptions:
						self._attachOEPolarity(pAttachIn = "node", plAttachTo = vdPolarityOptions["node"])
					
					if "subtree" in vdPolarityOptions:
						self._attachOEPolarity(pAttachIn = "subtree", plAttachTo = vdPolarityOptions["subtree"])
		
	
	
	def _attachOESuffix(self, pSuffix = "_OE", pAttachIn = "node", plAttachTo = "rel"):
		'''
		Attaches opinion expression suffix to the specified locations 
		in the tree to which the node belongs
		
		pAttachIn can be:
		 - node: attach only in this node
		 - subtree: attach in all nodes in the OE subtree of dependents
		 - parents: IMPLEMENT: attach in all parents (heads, heads of 
		            heads, and so on) of the OE node up to the root (inclusive)
		plAttachTo can be:
		 - form: attach to node form
		 - rel: attach to node's dependency relation(s) to head(s)
		 - pos: attach to node pos tag
		'''
		
		vAttachIn = pAttachIn.lower()
		
		if vAttachIn in ["node", "subtree"]:
			self._attachOESuffixTo(pSuffix = pSuffix, plAttachTo = plAttachTo)
		
		if vAttachIn == "subtree":
			for vDependent in self.getDependentNodes():
				vDependent._attachOESuffix(pSuffix = pSuffix, pAttachIn = "subtree", plAttachTo = plAttachTo)
		
		if vAttachIn == "parents" and not self.isRoot():
			# NOTE: only the first head is used for multihead cases
			vHead = self.getHeadNode()
			
			if not vHead.isRoot():
				vHead._attachOESuffixTo(pSuffix = pSuffix, pAttachTo = vAttachTo)
				vHead._attachOESuffix(pSuffix = pSuffix, pAttachIn = "parents", plAttachTo = plAttachTo)
		
	
	
	def _attachOESuffixTo(self, pSuffix, plAttachTo):
		'''
		Attaches opinion expression suffix to the specified locations in the 
		node
		
		plAttachTo can be a list of:
		 - form: attach to node form
		 - rel: attach to node's dependency relation(s) to head(s)
		 - pos: attach to node pos tag
		'''
		
		vlPositions = [p.lower() for p in plAttachTo]
		
		if "form" in vlPositions:
			vForm = self.getForm()
			if vForm != "OE":
				self.setForm(vForm + pSuffix)
		
		if "rel" in vlPositions:
			for vHead in self.getHeadPos():
				vRel = self.getDepRel(pHead = vHead)
				self.setDepRel(pRel = vRel + pSuffix, pHead = vHead)
		
		if "pos" in vlPositions:
			vPOS = self.getPOSTag()
			self.setPOSTag(vPOS + pSuffix)
		
	
	
	def _attachOEPolarity(self, pPolarity = '', pAttachIn = "node", plAttachTo = "rel"):
		'''
		Attaches polarity label as suffix to the node label and optionally
		to all nodes in the subtree 
		
		- pAttachIn can be:
		   - node: attach only in this node
		   - subtree: attach in all nodes in the aspect term subtree of 
		              dependents
		   - parents: IMPLEMENT: attach in all parents (heads, heads of heads, and 
		              so on) of the aspect term node up to the root (inclusive)
		- plAttachTo can be:
		   - form: attach to node form
		   - rel: attach to node's dependency relation(s) to head(s)
		   - pos: attach to node pos tag
		'''
		
		# implement
		
		pass
		
	
	
	def setSentScore(self, pScore):
		'''
		Sets the value of sentiment score of the node's word
		'''
		
		self.sentScore = pScore
		
	
	
	def getSentScore(self):
		'''
		Returns the sentiment score of the node's word
		'''
		
		return self.sentScore
		
	
	


class ABSADKTree(tk.DepKTree):
	'''
	The class for ABSA dependency tree for use in tree kernels.
	
	For use in tree kernels, the tree is represented in PTB bracketing 
	format.
	'''
	
	
	def generateDepKTree(self, pNode = "root", pFormat = "(rel form)", pdOptions = {}):
		'''
		Generates dependency kernel tree or subtree under a given node in
		the required format 
		
		NOTE: The subtree here should not be confused with the notion of
		subtree as a tree kernel variation used in parallel to subset tree
		kernel.
		
		The default format is (rel form) which is the pure dependency tree
		with only dependency relations and token forms as node labels. In
		general, the format string is the representation of the innermost
		treelet in the format. See each format-specific method for exact
		details.
		
		NOTE: if the insertion of an AT node is required for the aspect 
		term representation in the dependency tree, this method should be
		used.
		
		pdOptions provides options specific to each format.
		
		See the documentation of the parent class.
		'''
		
		if type(pNode) is str and pNode.lower() == "root":
			pNode = self.depTree.root
		
		if pFormat == "(at (rel (pos form)))":
			self.kTree = "( %s)" % ''.join([self._generateABSADKSubtree1(self.depTree.getNode(d), 0) for d in pNode.dependents])
		elif pFormat == "(at (rel (at (pos (at form)))))":
			self.kTree = "( %s)" % ''.join([self._generateABSADKSubtree2(self.depTree.getNode(d), 0) for d in pNode.dependents])
		elif pFormat == "(at (rel )(pos )(form ()))":
			self.kTree = "( %s)" % ''.join([self._generateABSADKSubtree3(self.depTree.getNode(d), 0) for d in pNode.dependents])
		elif pFormat == "(score (rel (pos form)))":
			self.kTree = "( %s)" % ''.join([self._generateABSADKSubtree4(self.depTree.getNode(d), 0, pdOptions) for d in pNode.dependents])
		elif pFormat == "(rel (pos score))":
			self.kTree = "( %s)" % ''.join([self._generateABSADKSubtree5(self.depTree.getNode(d), 0, pdOptions) for d in pNode.dependents])
		elif pFormat == "(score (rel (roles (pos form))))":
			self.kTree = "( %s)" % ''.join([self._generateABSADKSubtree6(self.depTree.getNode(d), 0, pdOptions) for d in pNode.dependents])
		elif pFormat == "(score (rel (pos_roles form)))":
			self.kTree = "( %s)" % ''.join([self._generateABSADKSubtree7(self.depTree.getNode(d), 0, pdOptions) for d in pNode.dependents])
		elif pFormat == "(oe (rel (pos form)))":
			self.kTree = "( %s)" % ''.join([self._generateABSADKSubtree8(self.depTree.getNode(d), 0, pdOptions) for d in pNode.dependents])
		elif pFormat == "(rel_oe (pos form))":
			self.kTree = "( %s)" % ''.join([self._generateABSADKSubtree9(self.depTree.getNode(d), 0, pdOptions) for d in pNode.dependents])
		elif pFormat == "(oe_score (rel (pos form)))":
			self.kTree = "( %s)" % ''.join([self._generateABSADKSubtree10(self.depTree.getNode(d), 0, pdOptions) for d in pNode.dependents])
		elif pFormat == "(rel_score (pos form))":
			self.kTree = "( %s)" % ''.join([self._generateABSADKSubtree11(self.depTree.getNode(d), 0, pdOptions) for d in pNode.dependents])
		elif pFormat == "(rel (pos wve))":
			self.kTree = "( %s)" % ''.join([self._generateABSADKSubtree12(self.depTree.getNode(d), 0, pdOptions) for d in pNode.dependents])
		else:
			self.kTree = tk.DepKTree.generateDepKTree(self, pNode = pNode, pFormat = pFormat)
		
		return self.kTree
		
	
	
	# (at (rel (pos form))) -> no equivalent number format in version 0.3
	def _generateABSADKSubtree1(self, pNode, pCurrentHead):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (at (rel (pos form))) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format, the nodes bearing an aspect term are represented
		as follows. The other nodes are represented in (rel (pos form))
		format.
		leaf node    : (AT (<dependency relation> (<POS tag> <token form>)))
		internal node: (AT (<dependency relation> (<POS tag> (<token form> <dependent 1> .. <dependent n>))))
		'''
		
		vDependents = ''.join([self._generateABSADKSubtree1(self.depTree.getNode(d), pNode.position) for d in pNode.dependents])
		
		if len(pNode.getAspectTerms()) > 0:
			if len(pNode.dependents) == 0:
				vKSubtree = "(AT (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
				                                   pNode.getPOSTag(),
				                                   pNode.form)
			else:
				vKSubtree = "(AT (%s (%s (%s %s))))" % (pNode.getDepRel(pCurrentHead),
				                                        pNode.getPOSTag(),
				                                        pNode.form,
				                                       vDependents)
		else:
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
				                              pNode.getPOSTag(),
				                              pNode.form)
			else:
				vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
				                                   pNode.getPOSTag(),
				                                   pNode.form,
				                                   vDependents)
		
		return vKSubtree
		
	
	
	# (at (rel (at (pos (at form))))) -> no equivalent number format in version 0.3
	def _generateABSADKSubtree2(self, pNode, pCurrentHead):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (at (rel (at (pos (at form))))) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format, the nodes bearing an aspect term are represented
		as follows. The other nodes are represented in (rel (pos form))
		format.
		leaf node    : (AT (<dependency relation> (AT (<POS tag> (AT <token form>)))))
		internal node: (AT (<dependency relation> (AT (<POS tag> (AT (<token form> <dependent 1> .. <dependent n>))))))
		'''
		
		vDependents = ''.join([self._generateABSADKSubtree2(self.depTree.getNode(d), pNode.position) for d in pNode.dependents])
		
		if len(pNode.getAspectTerms()) > 0:
			if len(pNode.dependents) == 0:
				vKSubtree = "(AT (%s (AT (%s (AT %s)))))" % (pNode.getDepRel(pCurrentHead),
				                                           pNode.getPOSTag(),
				                                           pNode.form)
			else:
				vKSubtree = "(AT (%s (AT (%s (AT (%s %s))))))" % (pNode.getDepRel(pCurrentHead),
				                                                pNode.getPOSTag(),
				                                                pNode.form,
				                                                vDependents)
		else:
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
				                              pNode.getPOSTag(),
				                              pNode.form)
			else:
				vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
				                                   pNode.getPOSTag(),
				                                   pNode.form,
				                                   vDependents)
		
		return vKSubtree
		
	
	
	# (at (rel )(pos )(form ())) -> no equivalent number format in version 0.3
	def _generateABSADKSubtree3(self, pNode, pCurrentHead):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (at (rel )(pos )(form ())) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format, the nodes bearing an aspect term are represented
		as follows. The other nodes are represented in (rel (pos form))
		format.
		leaf node    : (AT (<dependency relation> )(<POS tag> )(<token form> ))
		internal node: (AT (<dependency relation> )(<POS tag> )(<token form> (<dependent 1> .. <dependent n>)))
		'''
		
		vDependents = ''.join([self._generateABSADKSubtree3(self.depTree.getNode(d), pNode.position) for d in pNode.dependents])
		
		if len(pNode.getAspectTerms()) > 0:
			if len(pNode.dependents) == 0:
				vKSubtree = "(AT (%s )(%s )(%s ))" % (pNode.getDepRel(pCurrentHead),
				                                      pNode.getPOSTag(),
				                                      pNode.form)
			else:
				vKSubtree = "(AT (%s )(%s )(%s %s))" % (pNode.getDepRel(pCurrentHead),
				                                                pNode.getPOSTag(),
				                                                pNode.form,
				                                                vDependents)
		else:
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s (%s )(%s ))" % (pNode.getDepRel(pCurrentHead),
				                              pNode.getPOSTag(),
				                              pNode.form)
			else:
				vKSubtree = "(%s (%s )(%s %s))" % (pNode.getDepRel(pCurrentHead),
				                                   pNode.getPOSTag(),
				                                   pNode.form,
				                                   vDependents)
		
		return vKSubtree
		
	
	
	# (score (rel (pos form))) -> no equivalent number format in version 0.3
	def _generateABSADKSubtree4(self, pNode, pCurrentHead, pdOptions = {}):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (score (rel (pos form))) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format, the nodes bearing an aspect term are represented
		as follows. The other nodes are represented in (rel (pos form))
		format.
		leaf node    : (<Sentiment score> (<dependency relation> (<POS tag> <token form>)))
		internal node: (<Sentiment score> (<dependency relation> (<POS tag> (<token form> <dependent 1> .. <dependent n>))))
		
		pdOptions contains the following options to be used in formatting:
		- neutral: the way the neutral words, i.e. those without a sentiment 
		           score should be treated. The possible values include
		           an empty string which means do not add any node for such
		           words, and a string value which will be used as a node
		           to be inserted in the same way the scores are.
		'''
		
		vDependents = ''.join([self._generateABSADKSubtree4(self.depTree.getNode(d), pNode.position, pdOptions) for d in pNode.dependents])
		
		if pNode.getSentScore() is not None:
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s (%s (%s %s)))" % (pNode.getSentScore(),
				                                   pNode.getDepRel(pCurrentHead),
				                                   pNode.getPOSTag(),
				                                   pNode.form)
			else:
				vKSubtree = "(%s (%s (%s (%s %s))))" % (pNode.getSentScore(),
				                                        pNode.getDepRel(pCurrentHead),
				                                        pNode.getPOSTag(),
				                                        pNode.form,
				                                        vDependents)
		else:
			if "neutral" in pdOptions and pdOptions["neutral"].strip() != '':
				if len(pNode.dependents) == 0:
					vKSubtree = "(%s (%s (%s %s)))" % (pdOptions["neutral"],
					                                   pNode.getDepRel(pCurrentHead),
					                                   pNode.getPOSTag(),
					                                   pNode.form)
				else:
					vKSubtree = "(%s (%s (%s (%s %s))))" % (pdOptions["neutral"],
					                                        pNode.getDepRel(pCurrentHead),
					                                        pNode.getPOSTag(),
					                                        pNode.form,
					                                        vDependents)
			else:
				if len(pNode.dependents) == 0:
					vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
					                              pNode.getPOSTag(),
					                              pNode.form)
				else:
					vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
					                                   pNode.getPOSTag(),
					                                   pNode.form,
					                                   vDependents)
		
		return vKSubtree
		
	
	
	# (rel (pos score)) -> no equivalent number format in version 0.3
	def _generateABSADKSubtree5(self, pNode, pCurrentHead, pdOptions = {}):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (rel (pos score)) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format, the nodes bearing an aspect term are represented
		as follows. The other nodes are represented in (rel (pos form))
		format.
		leaf node    : (<dependency relation> (<POS tag> <Sentiment score>))
		internal node: (<dependency relation> (<POS tag> (<Sentiment score> <dependent 1> .. <dependent n>))))
		'''
		
		vDependents = ''.join([self._generateABSADKSubtree5(self.depTree.getNode(d), pNode.position, pdOptions) for d in pNode.dependents])
		
		if pNode.getSentScore() is not None:
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
				                              pNode.getPOSTag(),
				                              pNode.getSentScore())
			else:
				vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
				                                   pNode.getPOSTag(),
				                                   pNode.getSentScore(),
				                                   vDependents)
		else:
			if "neutral" in pdOptions and pdOptions["neutral"].strip() != '':
				if len(pNode.dependents) == 0:
					vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
					                              pNode.getPOSTag(),
					                              pdOptions["neutral"])
				else:
					vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
					                                   pNode.getPOSTag(),
					                                   pdOptions["neutral"],
					                                   vDependents)
			else:
				if len(pNode.dependents) == 0:
					vKSubtree = "(%s %s)" % (pNode.getDepRel(pCurrentHead),
					                         pNode.getPOSTag())
				else:
					vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
					                                   pNode.getPOSTag(),
					                                   vDependents)
		
		return vKSubtree
		
	
	
	# (score (rel (roles (pos form)))) -> no equivalent number format in version 0.3
	def _generateABSADKSubtree6(self, pNode, pCurrentHead, pdOptions = {}):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (score (rel (roles (pos form)))) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format, the nodes bearing an aspect term are represented
		as follows. The other nodes are represented in (rel (pos form))
		format.
		leaf node    : (<Sentiment score> (<dependency relation> (<Role labels> (<POS tag> <token form>))))
		internal node: (<Sentiment score> (<dependency relation> (<Role labels> (<POS tag> (<token form> <dependent 1> .. <dependent n>)))))
		
		pdOptions contains the following options to be used in formatting:
		- neutral: the way the neutral words, i.e. those without a sentiment 
		           score should be treated. The possible values include
		           an empty string which means do not add any node for such
		           words, and a string value which will be used as a node
		           to be inserted in the same way the scores are.
		- no-arg: the way non-argument nodes are represented. The possible
		          values include an empty string which means the node will 
		          be represented as in non-semantic format, and a string 
		          value (e.g. null) which will be used as the label for 
		          representing semantic role of such nodes.  
		'''
		
		vDependents = ''.join([self._generateABSADKSubtree6(self.depTree.getNode(d), pNode.position, pdOptions) for d in pNode.dependents])
		
		
		if pNode.isArgument():
			if pNode.getSentScore() is not None:
				if len(pNode.dependents) == 0:
					vKSubtree = "(%s (%s (%s (%s %s))))" % (pNode.getSentScore(),
					                                        pNode.getDepRel(pCurrentHead),
					                                        '_'.join(pNode.getArgRoles()),
					                                        pNode.getPOSTag(),
					                                        pNode.form)
				else:
					vKSubtree = "(%s (%s (%s (%s (%s %s)))))" % (pNode.getSentScore(),
					                                             pNode.getDepRel(pCurrentHead),
					                                             '_'.join(pNode.getArgRoles()),
					                                             pNode.getPOSTag(),
					                                             pNode.form,
					                                             vDependents)
			else:
				if "neutral" in pdOptions and pdOptions["neutral"].strip() != '':
					if len(pNode.dependents) == 0:
						vKSubtree = "(%s (%s (%s (%s %s))))" % (pdOptions["neutral"],
						                                        pNode.getDepRel(pCurrentHead),
						                                        '_'.join(pNode.getArgRoles()),
						                                        pNode.getPOSTag(),
						                                        pNode.form)
					else:
						vKSubtree = "(%s (%s (%s (%s (%s %s)))))" % (pdOptions["neutral"],
						                                             pNode.getDepRel(pCurrentHead),
						                                             '_'.join(pNode.getArgRoles()),
						                                             pNode.getPOSTag(),
						                                             pNode.form,
						                                             vDependents)
				else:
					if len(pNode.dependents) == 0:
						vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
						                                   '_'.join(pNode.getArgRoles()),
						                                   pNode.getPOSTag(),
						                                   pNode.form)
					else:
						vKSubtree = "(%s (%s (%s (%s %s))))" % (pNode.getDepRel(pCurrentHead),
					                                            '_'.join(pNode.getArgRoles()),
						                                        pNode.getPOSTag(),
						                                        pNode.form,
						                                        vDependents)
		elif "no-arg" in pdOptions and pdOptions["no-arg"].strip() != '':
			if pNode.getSentScore() is not None:
				if len(pNode.dependents) == 0:
					vKSubtree = "(%s (%s (%s (%s %s))))" % (pNode.getSentScore(),
					                                        pNode.getDepRel(pCurrentHead),
					                                        pdOptions["no-arg"],
					                                        pNode.getPOSTag(),
					                                        pNode.form)
				else:
					vKSubtree = "(%s (%s (%s (%s (%s %s)))))" % (pNode.getSentScore(),
					                                             pNode.getDepRel(pCurrentHead),
					                                             pdOptions["no-arg"],
					                                             pNode.getPOSTag(),
					                                             pNode.form,
					                                             vDependents)
			else:
				if "neutral" in pdOptions and pdOptions["neutral"].strip() != '':
					if len(pNode.dependents) == 0:
						vKSubtree = "(%s (%s (%s (%s %s))))" % (pdOptions["neutral"],
						                                        pNode.getDepRel(pCurrentHead),
						                                        pdOptions["no-arg"],
						                                        pNode.getPOSTag(),
						                                        pNode.form)
					else:
						vKSubtree = "(%s (%s (%s (%s (%s %s)))))" % (pdOptions["neutral"],
						                                             pNode.getDepRel(pCurrentHead),
						                                             pdOptions["no-arg"],
						                                             pNode.getPOSTag(),
						                                             pNode.form,
						                                             vDependents)
				else:
					if len(pNode.dependents) == 0:
						vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
						                                   pdOptions["no-arg"],
						                                   pNode.getPOSTag(),
						                                   pNode.form)
					else:
						vKSubtree = "(%s (%s (%s (%s %s))))" % (pNode.getDepRel(pCurrentHead),
					                                            pdOptions["no-arg"],
						                                        pNode.getPOSTag(),
						                                        pNode.form,
						                                        vDependents)
		else:
			if pNode.getSentScore() is not None:
				if len(pNode.dependents) == 0:
					vKSubtree = "(%s (%s (%s %s)))" % (pNode.getSentScore(),
					                                   pNode.getDepRel(pCurrentHead),
					                                   pNode.getPOSTag(),
					                                   pNode.form)
				else:
					vKSubtree = "(%s (%s (%s (%s %s))))" % (pNode.getSentScore(),
					                                        pNode.getDepRel(pCurrentHead),
					                                        pNode.getPOSTag(),
					                                        pNode.form,
					                                        vDependents)
			else:
				if "neutral" in pdOptions and pdOptions["neutral"].strip() != '':
					if len(pNode.dependents) == 0:
						vKSubtree = "(%s (%s (%s %s)))" % (pdOptions["neutral"],
						                                   pNode.getDepRel(pCurrentHead),
						                                   pNode.getPOSTag(),
						                                   pNode.form)
					else:
						vKSubtree = "(%s (%s (%s (%s %s))))" % (pdOptions["neutral"],
						                                        pNode.getDepRel(pCurrentHead),
						                                        pNode.getPOSTag(),
						                                        pNode.form,
						                                        vDependents)
				else:
					if len(pNode.dependents) == 0:
						vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
						                              pNode.getPOSTag(),
						                              pNode.form)
					else:
						vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
						                                   pNode.getPOSTag(),
						                                   pNode.form,
						                                   vDependents)
		
		return vKSubtree
		
	
	
	# (score (rel (pos_roles form))) -> no equivalent number format in version 0.3
	def _generateABSADKSubtree7(self, pNode, pCurrentHead, pdOptions = {}):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (score (rel (pos_roles form))) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format, the nodes bearing an aspect term are represented
		as follows. The other nodes are represented in (rel (pos form))
		format.
		leaf node    : (<Sentiment score> (<dependency relation> (<POS tag>_<Role labels> <token form>))))
		internal node: (<Sentiment score> (<dependency relation> (<POS tag>_<Role labels> (<token form> <dependent 1> .. <dependent n>)))))
		
		pdOptions contains the following options to be used in formatting:
		- neutral: the way the neutral words, i.e. those without a sentiment 
		           score should be treated. The possible values include
		           an empty string which means do not add any node for such
		           words, and a string value which will be used as a node
		           to be inserted in the same way the scores are.
		'''
		
		vDependents = ''.join([self._generateABSADKSubtree7(self.depTree.getNode(d), pNode.position, pdOptions) for d in pNode.dependents])
		
		
		if pNode.isArgument():
			if pNode.getSentScore() is not None:
				if len(pNode.dependents) == 0:
					vKSubtree = "(%s (%s (%s_%s %s)))" % (pNode.getSentScore(),
					                                      pNode.getDepRel(pCurrentHead),
					                                      pNode.getPOSTag(),
					                                      '_'.join(pNode.getArgRoles()),
					                                      pNode.form)
				else:
					vKSubtree = "(%s (%s (%s_%s (%s %s))))" % (pNode.getSentScore(),
					                                           pNode.getDepRel(pCurrentHead),
					                                           pNode.getPOSTag(),
					                                           '_'.join(pNode.getArgRoles()),
					                                           pNode.form,
					                                           vDependents)
			else:
				if "neutral" in pdOptions and pdOptions["neutral"].strip() != '':
					if len(pNode.dependents) == 0:
						vKSubtree = "(%s (%s (%s_%s %s)))" % (pdOptions["neutral"],
						                                      pNode.getDepRel(pCurrentHead),
						                                      pNode.getPOSTag(),
						                                      '_'.join(pNode.getArgRoles()),
						                                      pNode.form)
					else:
						vKSubtree = "(%s (%s (%s_%s (%s %s))))" % (pdOptions["neutral"],
						                                           pNode.getDepRel(pCurrentHead),
						                                           pNode.getPOSTag(),
						                                           '_'.join(pNode.getArgRoles()),
						                                           pNode.form,
						                                           vDependents)
				else:
					if len(pNode.dependents) == 0:
						vKSubtree = "(%s (%s_%s %s))" % (pNode.getDepRel(pCurrentHead),
						                                 pNode.getPOSTag(),
						                                 '_'.join(pNode.getArgRoles()),
						                                 pNode.form)
					else:
						vKSubtree = "(%s (%s_%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
						                                      pNode.getPOSTag(),
					                                          '_'.join(pNode.getArgRoles()),
						                                      pNode.form,
						                                      vDependents)
		else:
			if pNode.getSentScore() is not None:
				if len(pNode.dependents) == 0:
					vKSubtree = "(%s (%s (%s %s)))" % (pNode.getSentScore(),
					                                   pNode.getDepRel(pCurrentHead),
					                                   pNode.getPOSTag(),
					                                   pNode.form)
				else:
					vKSubtree = "(%s (%s (%s (%s %s))))" % (pNode.getSentScore(),
					                                        pNode.getDepRel(pCurrentHead),
					                                        pNode.getPOSTag(),
					                                        pNode.form,
					                                        vDependents)
			else:
				if "neutral" in pdOptions and pdOptions["neutral"].strip() != '':
					if len(pNode.dependents) == 0:
						vKSubtree = "(%s (%s (%s %s)))" % (pdOptions["neutral"],
						                                   pNode.getDepRel(pCurrentHead),
						                                   pNode.getPOSTag(),
						                                   pNode.form)
					else:
						vKSubtree = "(%s (%s (%s (%s %s))))" % (pdOptions["neutral"],
						                                        pNode.getDepRel(pCurrentHead),
						                                        pNode.getPOSTag(),
						                                        pNode.form,
						                                        vDependents)
				else:
					if len(pNode.dependents) == 0:
						vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
						                              pNode.getPOSTag(),
						                              pNode.form)
					else:
						vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
						                                   pNode.getPOSTag(),
						                                   pNode.form,
						                                   vDependents)
		
		return vKSubtree
		
	
	
	# (oe (rel (pos form))) -> no equivalent number format in version 0.3
	def _generateABSADKSubtree8(self, pNode, pCurrentHead, pdOptions = {}):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (oe (rel (pos form))) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format, the nodes bearing an aspect term are represented
		as follows. The other nodes are represented in (rel (pos form))
		format.
		leaf node    : (OE (<dependency relation> (<POS tag> <token form>)))
		internal node: (OE (<dependency relation> (<POS tag> (<token form> <dependent 1> .. <dependent n>))))
		
		pdOptions contains the following options to be used in formatting:
		'''
		
		vDependents = ''.join([self._generateABSADKSubtree8(self.depTree.getNode(d), pNode.position, pdOptions) for d in pNode.dependents])
		
		if len(pNode.getOEs()) > 0:
			if len(pNode.dependents) == 0:
				vKSubtree = "(OE (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
				                                   pNode.getPOSTag(),
				                                   pNode.form)
			else:
				vKSubtree = "(OE (%s (%s (%s %s))))" % (pNode.getDepRel(pCurrentHead),
				                                        pNode.getPOSTag(),
				                                        pNode.form,
				                                        vDependents)
		else:
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
				                              pNode.getPOSTag(),
				                              pNode.form)
			else:
				vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
				                                   pNode.getPOSTag(),
				                                   pNode.form,
				                                   vDependents)
		
		return vKSubtree
		
	
	
	# (rel_oe (pos form)) -> no equivalent number format in version 0.3
	def _generateABSADKSubtree9(self, pNode, pCurrentHead, pdOptions = {}):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (rel_oe (pos form)) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format, the nodes bearing an aspect term are represented
		as follows. The other nodes are represented in (rel (pos form))
		format.
		leaf node    : (<dependency relation>_OE (<POS tag> <token form>))
		internal node: (<dependency relation>_OE (<POS tag> (<token form> <dependent 1> .. <dependent n>)))
		
		pdOptions contains the following options to be used in formatting:
		'''
		
		vDependents = ''.join([self._generateABSADKSubtree9(self.depTree.getNode(d), pNode.position, pdOptions) for d in pNode.dependents])
		
		if len(pNode.getOEs()) > 0:
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s_OE (%s %s))" % (pNode.getDepRel(pCurrentHead),
				                                 pNode.getPOSTag(),
				                                 pNode.form)
			else:
				vKSubtree = "(%s_OE (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
				                                      pNode.getPOSTag(),
				                                      pNode.form,
				                                      vDependents)
		else:
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
				                              pNode.getPOSTag(),
				                              pNode.form)
			else:
				vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
				                                   pNode.getPOSTag(),
				                                   pNode.form,
				                                   vDependents)
		
		return vKSubtree
		
	
	
	# (oe_score (rel (pos form))) -> no equivalent number format in version 0.3
	def _generateABSADKSubtree10(self, pNode, pCurrentHead, pdOptions = {}):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (oe_score (rel (pos form))) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format, the nodes bearing an aspect term are represented
		as follows. The other nodes are represented in (rel (pos form))
		format.
		leaf node    : (OE_<Sentiment score> (<dependency relation> (<POS tag> <token form>)))
		internal node: (OE_<Sentiment score> (<dependency relation> (<POS tag> (<token form> <dependent 1> .. <dependent n>))))
		
		pdOptions contains the following options to be used in formatting:
		- neutral: the way the neutral words, i.e. those without a sentiment 
		           score should be treated. The possible values include
		           an empty string which means do not add any node for such
		           words, and a string value which will be used as a node
		           to be inserted in the same way the scores are.
		'''
		
		vDependents = ''.join([self._generateABSADKSubtree10(self.depTree.getNode(d), pNode.position, pdOptions) for d in pNode.dependents])
		
		if len(pNode.getOEs()) > 0:
			if pNode.getSentScore() is not None:
				if len(pNode.dependents) == 0:
					vKSubtree = "(OE_%s (%s (%s %s)))" % (pNode.getSentScore(),
													      pNode.getDepRel(pCurrentHead),
													      pNode.getPOSTag(),
													      pNode.form)
				else:
					vKSubtree = "(OE_%s (%s (%s (%s %s))))" % (pNode.getSentScore(),
															   pNode.getDepRel(pCurrentHead),
															   pNode.getPOSTag(),
															   pNode.form,
															   vDependents)
			else:
				if "neutral" in pdOptions and pdOptions["neutral"].strip() != '':
					if len(pNode.dependents) == 0:
						vKSubtree = "(OE_%s (%s (%s %s)))" % (pdOptions["neutral"],
														      pNode.getDepRel(pCurrentHead),
														      pNode.getPOSTag(),
														      pNode.form)
					else:
						vKSubtree = "(OE_%s (%s (%s (%s %s))))" % (pdOptions["neutral"],
																   pNode.getDepRel(pCurrentHead),
																   pNode.getPOSTag(),
																   pNode.form,
																   vDependents)
				else:
					if len(pNode.dependents) == 0:
						vKSubtree = "(OE (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
													       pNode.getPOSTag(),
													       pNode.form)
					else:
						vKSubtree = "(OE (%s (%s (%s %s))))" % (pNode.getDepRel(pCurrentHead),
														        pNode.getPOSTag(),
														        pNode.form,
														        vDependents)
		else:
			if pNode.getSentScore() is not None:
				if len(pNode.dependents) == 0:
					vKSubtree = "(%s (%s (%s %s)))" % (pNode.getSentScore(),
													   pNode.getDepRel(pCurrentHead),
													   pNode.getPOSTag(),
													   pNode.form)
				else:
					vKSubtree = "(%s (%s (%s (%s %s))))" % (pNode.getSentScore(),
															pNode.getDepRel(pCurrentHead),
															pNode.getPOSTag(),
															pNode.form,
															vDependents)
			else:
				if "neutral" in pdOptions and pdOptions["neutral"].strip() != '':
					if len(pNode.dependents) == 0:
						vKSubtree = "(%s (%s (%s %s)))" % (pdOptions["neutral"],
														   pNode.getDepRel(pCurrentHead),
														   pNode.getPOSTag(),
														   pNode.form)
					else:
						vKSubtree = "(%s (%s (%s (%s %s))))" % (pdOptions["neutral"],
																pNode.getDepRel(pCurrentHead),
																pNode.getPOSTag(),
																pNode.form,
																vDependents)
				else:
					if len(pNode.dependents) == 0:
						vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
													  pNode.getPOSTag(),
													  pNode.form)
					else:
						vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
														   pNode.getPOSTag(),
														   pNode.form,
														   vDependents)
		
		return vKSubtree
		
	
	
	# (rel_score (pos form)) -> no equivalent number format in version 0.3
	def _generateABSADKSubtree11(self, pNode, pCurrentHead, pdOptions = {}):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (score (rel (pos form))) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format, the nodes bearing an aspect term are represented
		as follows. The other nodes are represented in (rel (pos form))
		format.
		leaf node    : (<dependency relation>_<Sentiment score> (<POS tag> <token form>))
		internal node: (<dependency relation>_<Sentiment score> (<POS tag> (<token form> <dependent 1> .. <dependent n>)))
		
		pdOptions contains the following options to be used in formatting:
		- neutral: the way the neutral words, i.e. those without a sentiment 
		           score should be treated. The possible values include
		           an empty string which means do not add any node for such
		           words, and a string value which will be used as a node
		           to be inserted in the same way the scores are.
		'''
		
		vDependents = ''.join([self._generateABSADKSubtree11(self.depTree.getNode(d), pNode.position, pdOptions) for d in pNode.dependents])
		
		if pNode.getSentScore() is not None:
			if len(pNode.dependents) == 0:
				vKSubtree = "(%s_%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
				                                 pNode.getSentScore(),
				                                 pNode.getPOSTag(),
				                                 pNode.form)
			else:
				vKSubtree = "(%s_%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
				                                      pNode.getSentScore(),
				                                      pNode.getPOSTag(),
				                                      pNode.form,
				                                      vDependents)
		else:
			if "neutral" in pdOptions and pdOptions["neutral"].strip() != '':
				if len(pNode.dependents) == 0:
					vKSubtree = "(%s_%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
					                                 pdOptions["neutral"],
					                                 pNode.getPOSTag(),
					                                 pNode.form)
				else:
					vKSubtree = "(%s_%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
					                                      pdOptions["neutral"],
					                                      pNode.getPOSTag(),
					                                      pNode.form,
					                                      vDependents)
			else:
				if len(pNode.dependents) == 0:
					vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
					                              pNode.getPOSTag(),
					                              pNode.form)
				else:
					vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
					                                   pNode.getPOSTag(),
					                                   pNode.form,
					                                   vDependents)
		
		return vKSubtree
		
	
	
	# (rel (pos wve)) -> no equivalent number format in version 0.3
	def _generateABSADKSubtree12(self, pNode, pCurrentHead, pdOptions = {}):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (rel (pos score)) format
		
		pCurrentHead identifies which head is calling this method in case
		the node has multiple heads.
		
		In this format, the nodes bearing an aspect term are represented
		as follows. The other nodes are represented in (rel (pos form))
		format.
		leaf node    : (<dependency relation> (<POS tag> <Sentiment score>))
		internal node: (<dependency relation> (<POS tag> (<Sentiment score> <dependent 1> .. <dependent n>))))
		'''
		
		vDependents = ''.join([self._generateABSADKSubtree12(self.depTree.getNode(d), pNode.position, pdOptions) for d in pNode.dependents])
		
		# word vector object to use (compulsory option)
		vWV = pdOptions["wv"]
		
		# dimension to use, e.g , the 5th dimension of the vector 
		if "dimension" in pdOptions:
			vDim = int(pdOptions["dimension"])
		else:
			vDim = 1
		
		# number of digits after decimal point to round 
		if "round-digits" in pdOptions:
			vRoundDigits = int(pdOptions["round-digits"])
		else:
			vRoundDigits = 0
		
		# some unmaskings (write a function in parse.py for this or get the tokens from sentence or ...)
		vToken = pNode.getForm().replace("-LRB-", '(').replace("-RRB-", ')').replace("-LSB-", '[').replace("-RSB-", ']')
		vToken = vToken.replace("\/", "/")
		vToken = vToken.replace("``", '"').replace("''", '"').replace("`", "'")
		vToken = vToken.replace("-XCL-", "!").replace("-QSN-", "?").replace("-HSH-", "#")
		
		if len(pNode.dependents) == 0:
			vKSubtree = "(%s (%s %s))" % (pNode.getDepRel(pCurrentHead),
										  pNode.getPOSTag(),
										  str(round(vWV.getVector(vToken)[vDim - 1], vRoundDigits)))
		else:
			vKSubtree = "(%s (%s (%s %s)))" % (pNode.getDepRel(pCurrentHead),
											   pNode.getPOSTag(),
											   str(round(vWV.getVector(vToken)[vDim - 1], vRoundDigits)),
											   vDependents)
		
		return vKSubtree
		
	
	


class ABSACKTree(tk.ConstKTree):
	'''
	The class for ABSA constituency tree for use in tree kernels
	
	For use in tree kernels, the tree is represented in PTB bracketing 
	format.
	'''
	
	
	
	def generateConstKTree(self, pFormat = "(phrase (pos form))", pdOptions = None):
		'''
		Generates constituency kernel tree or subtree under a given node in
		the required format 
		
		NOTE: The subtree here should not be confused with the notion of
		subtree as a tree kernel variation used in parallel to subset tree
		kernel.
		
		The default format is (phrase (pos form)) which is the pure constituency
		tree in bracketing (s-expression) format. In general, the format string 
		is the representation of the adequately innermost treelet in the format.
		See each format-specific method for exact details.
		
		NOTE: if the insertion of an AT node is required for the aspect term 
		representation, use decoration method provided with the constituency 
		tree class.
		
		See the documentation of the parent class.
		
		pdOptions contains specific options to each format.
		'''
		
		if pFormat.lower() == "(score (score score))":
			self.kTree = "( %s)" % self._generateABSACKSubtree1(self.constTree.root, pdOptions = pdOptions)
		elif pFormat.lower() == "(score (phrase (score (pos form))))":
			self.kTree = "( %s)" % self._generateABSACKSubtree2(self.constTree.root, pdOptions = pdOptions)
		elif pFormat.lower() == "(phrase_score (pos_score form))":
			self.kTree = "( %s)" % self._generateABSACKSubtree3(self.constTree.root, pdOptions = pdOptions)
		elif pFormat.lower() == "(score (phrase_args (score (pos_args form))))":
			self.kTree = "( %s)" % self._generateABSACKSubtree4(self.constTree.root, pdOptions = pdOptions)
		elif pFormat.lower() == "(phrase (score )(pos (score )(form )))":
			self.kTree = "( %s)" % self._generateABSACKSubtree5(self.constTree.root, pdOptions = pdOptions)
		elif pFormat.lower() == "(phrase (pos wve))":
			self.kTree = "( %s)" % self._generateABSACKSubtree6(self.constTree.root, pdOptions = pdOptions)
		else:
			self.kTree = tk.ConstKTree.generateConstKTree(self, pFormat = pFormat)
		
		return self.kTree
		
	
	
	def _generateABSACKSubtree1(self, pNode, pdOptions = None):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (score (score score)) format
		
		In this format, all the nodes represent the sentiment score 
		associated with them.
		
		pdOptions contains the following options:
		- keep-at: if true will keep the AT label and wont use sentiment 
		           scores to replace it (default false) 
		'''
		
		if pNode.isPreTerminal() or pNode.isTerminal():
			vKSubtree = "(%s %s)" % (pNode.getSentScore(),
			                         pNode.getTerminalNodes()[0].getSentScore())
		elif pdOptions is not None and "keep-at" in pdOptions and pdOptions["keep-at"] == True and pNode.getSynTag().upper() == "AT":     
			vKSubtree = "(AT %s)" % (''.join([self._generateABSACKSubtree1(n, pdOptions) for n in pNode.children]))
		else:
			vKSubtree = "(%s %s)" % (pNode.getSentScore(),
			                         ''.join([self._generateABSACKSubtree1(n, pdOptions) for n in pNode.children]))
		
		return vKSubtree
		
	
	
	def _generateABSACKSubtree2(self, pNode, pdOptions = None):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (score (phrase (score (pos form)))) 
		format
		
		In this format, all the nodes represent the sentiment score 
		associated with them.
		
		pdOptions contains the following options:
		- keep-at: if true will keep the AT label and wont use sentiment 
		           scores to replace it (default false) 
		'''
		
		if pNode.isPreTerminal():
			vKSubtree = "(%s (%s %s))" % (pNode.getSentScore(),
										  pNode.getSynTag(),
			                              pNode.getTerminal())
		elif pdOptions is not None and "keep-at" in pdOptions and pdOptions["keep-at"] == True and pNode.getSynTag().upper() == "AT":     
			vKSubtree = "(AT %s)" % (''.join([self._generateABSACKSubtree2(n, pdOptions) for n in pNode.children]))
		else:
			vKSubtree = "(%s (%s %s))" % (pNode.getSentScore(),
			                              pNode.getSynTag(),
			                              ''.join([self._generateABSACKSubtree2(n, pdOptions) for n in pNode.children]))
		
		return vKSubtree
		
	
	
	def _generateABSACKSubtree3(self, pNode, pdOptions = None):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (phrase_score (pos_score form))
		format
		
		In this format, all the nodes represent the sentiment score 
		associated with them.
		
		pdOptions contains the following options:
		- keep-at: if true will keep the AT label and wont use sentiment 
		           scores to replace it (default false) 
		'''
		
		if pNode.isPreTerminal():
			vKSubtree = "(%s_%s %s)" % (pNode.getSynTag(),
			                            pNode.getSentScore(),
			                            pNode.getTerminal())
		elif pdOptions is not None and "keep-at" in pdOptions and pdOptions["keep-at"] == True and pNode.getSynTag().upper() == "AT":     
			vKSubtree = "(AT %s)" % (''.join([self._generateABSACKSubtree3(n, pdOptions) for n in pNode.children]))
		else:
			vKSubtree = "(%s_%s %s)" % (pNode.getSynTag(),
			                            pNode.getSentScore(),
			                            ''.join([self._generateABSACKSubtree3(n, pdOptions) for n in pNode.children]))
		
		return vKSubtree
		
	
	
	def _generateABSACKSubtree4(self, pNode, pdOptions = None):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (score (phrase_args (score (pos_args form)))) 
		format
		
		In this format, all the nodes represent the sentiment score 
		associated with them.
		
		pdOptions contains the following options:
		- keep-at: if true will keep the AT label and wont use sentiment 
		           scores to replace it (default false) 
		'''
		
		if pNode.isPreTerminal():
			vKSubtree = "(%s (%s %s))" % (pNode.getSentScore(),
										  '_'.join([pNode.getSynTag()] + pNode.getArgRoles()),
			                              pNode.getTerminal())
		elif pdOptions is not None and "keep-at" in pdOptions and pdOptions["keep-at"] == True and pNode.getSynTag().upper() == "AT":     
			vKSubtree = "(AT %s)" % (''.join([self._generateABSACKSubtree4(n, pdOptions) for n in pNode.children]))
		else:
			vKSubtree = "(%s (%s %s))" % (pNode.getSentScore(),
			                              '_'.join([pNode.getSynTag()] + pNode.getArgRoles()),
			                              ''.join([self._generateABSACKSubtree4(n, pdOptions) for n in pNode.children]))
		
		return vKSubtree
		
	
	
	def _generateABSACKSubtree5(self, pNode, pdOptions = None):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (phrase (score )(pos (score )(form ))) 
		format
		
		In this format, all the nodes represent the sentiment score 
		associated with them.
		
		pdOptions contains the following options:
		- keep-at: if true will keep the AT label and wont use sentiment 
		           scores to replace it (default false) 
		'''
		
		if pNode.isPreTerminal():
			vKSubtree = "(%s (%s )(%s ))" % (pNode.getSynTag(),
										     pNode.getSentScore(),
			                                 pNode.getTerminal())
		elif pdOptions is not None and "keep-at" in pdOptions and pdOptions["keep-at"] == True and pNode.getSynTag().upper() == "AT":     
			vKSubtree = "(AT %s)" % (''.join([self._generateABSACKSubtree5(n, pdOptions) for n in pNode.children]))
		else:
			vKSubtree = "(%s (%s )%s)" % (pNode.getSynTag(),
			                              pNode.getSentScore(),
			                              ''.join([self._generateABSACKSubtree5(n, pdOptions) for n in pNode.children]))
		
		return vKSubtree
		
	
	
	def _generateABSACKSubtree6(self, pNode, pdOptions = None):
		'''
		Recursively generates the kernel subtree of the given node in
		bracketing representation in (phrase (pos wve)) format
		
		In this format, all the nodes represent the sentiment score 
		associated with them.
		
		pdOptions contains the following options:
		- keep-at: if true will keep the AT label and wont use sentiment 
		           scores to replace it (default false) 
		'''
		
		# word vector object to use (compulsory option)
		vWV = pdOptions["wv"]
		
		# dimension to use, e.g , the 5th dimension of the vector 
		if "dimension" in pdOptions:
			vDim = int(pdOptions["dimension"])
		else:
			vDim = 1
		
		# number of digits after decimal point to round 
		if "round-digits" in pdOptions:
			vRoundDigits = int(pdOptions["round-digits"])
		else:
			vRoundDigits = 0
		
		if pNode.isPreTerminal() or pNode.isTerminal():
			# some unmaskings (write a function in parse.py for this or get the tokens from sentence or ...)
			vToken = pNode.getTerminal().replace("-LRB-", '(').replace("-RRB-", ')').replace("-LSB-", '[').replace("-RSB-", ']')
			vToken = vToken.replace("\/", "/")
			vToken = vToken.replace("``", '"').replace("''", '"').replace("`", "'")
			vToken = vToken.replace("-XCL-", "!").replace("-QSN-", "?").replace("-HSH-", "#")
			vKSubtree = "(%s %s)" % (pNode.getSynTag(),
			                         str(round(vWV.getVector(vToken)[vDim - 1], vRoundDigits)))
		elif pdOptions is not None and "keep-at" in pdOptions and pdOptions["keep-at"] == True and pNode.getSynTag().upper() == "AT":     
			vKSubtree = "(AT %s)" % ''.join([self._generateABSACKSubtree6(n, pdOptions) for n in pNode.children])
		else:
			vKSubtree = "(%s %s)" % (pNode.getSynTag(),
			                         ''.join([self._generateABSACKSubtree6(n, pdOptions) for n in pNode.children]))
		
		return vKSubtree
		
	
	


class ABSACDKTree(tk.ConstDepKTree):
	'''
	Class for integrating ABSA dependency subtrees in ABSA constituency 
	trees for tree kernel use
	'''
	
	
	def _createDepKTree(self):
		'''
		Creates and returns a new ABSADKTree
		'''
		
		return ABSADKTree(pDepTree = self.depTree)
		
	
	
	def generateConstDepKTree(self, pFormat = "(phrase (pos (form (rel head))))", pdOptions = {}):
		'''
		Generates the tree representation in the required format
		
		See the parent class for more details.
		'''
		
		
		self.kTree = tk.ConstDepKTree.generateConstDepKTree(self, pFormat = pFormat, pdOptions = pdOptions)
		
		return self.kTree
		
	
	


class ABSANGramKTree(tk.NGramKTree):
	'''
	Class for implementing n-gram tree for ABSA sentences
	'''
	
	
	def __init__(self, pABSASent, pNodeContentType = "word", pAspectTerm = None, pdOptions = None):
		'''
		Constructor
		
		pNodeContentTypes can take:
		- word: word n-gram trees are produced, i.e. nodes are word forms
		- POS: POS n-gram trees are produced, i.e. nodes are POS tags
		- sentiment: sentiment score n-gram trees are produced, i.e. nodes are sentiment polarity scores
		- wve: word vector element n-gram trees are produced, i.e. nodes are i'th elements of word vectors 
		- wvp: prefixed words for word vector similarity computation are produced (e.g. with svmlight-tk-we) 
		
		pdOptions contains specific options to each format.
		'''
		
		if pdOptions is None:
			pdOptions = {}
		
		if pNodeContentType.lower() == "word":
			tk.NGramKTree.__init__(self, plTokens = pABSASent.getTokens())
		elif pNodeContentType.lower() == "word-lower":
			tk.NGramKTree.__init__(self, plTokens = [t.lower() for t in pABSASent.getTokens()])
		elif pNodeContentType.lower() == "pos":
			tk.NGramKTree.__init__(self, plTokens = pABSASent.getPOSTags())
		elif pNodeContentType.lower() == "sentiment":
			tk.NGramKTree.__init__(self, plTokens = [str(s) for s in pABSASent.getSentimentScores()])
		elif pNodeContentType.lower() == "wve":
			# the dimension to use, e.g , the 5th dimension of the vector 
			if "dimension" in pdOptions:
				vDim = int(pdOptions["dimension"])
			else:
				vDim = 1
			# the number of digits after decimal point to round 
			if "round-digits" in pdOptions:
				vRoundDigits = int(pdOptions["round-digits"])
			else:
				vRoundDigits = 0
			# now generating the object
			tk.NGramKTree.__init__(self, plTokens = [str(round(v[vDim - 1], vRoundDigits)) for v in pABSASent.getWordVectors()])
		elif pNodeContentType.lower() == "wvp":
			if "prefix" in pdOptions:
				vPrefix= pdOptions["prefix"]
			else:
				vPrefix= "___"
			tk.NGramKTree.__init__(self, plTokens = [vPrefix + t for t in pABSASent.getTokens()])
		else:
			raise Exception("%s is an invalid node content type!" % pNodeContentType)
		
		self.absaSent = pABSASent
		
		self.aspectTermMap = self._mapAspectTermToTokens(pAspectTerm)
		
	
	
	def _mapAspectTermToTokens(self, pAspectTerm):
		'''
		Maps the aspect term to sentence by marking its span in a parallel list to sentence token list 
		'''
		
		vSpan = pAspectTerm.getTokenSpan()
		vlAspectTermMap = []
		
		for i, vToken in enumerate(self.tokens, start = 1):
			if i in range(vSpan[0], vSpan[1] + 1):
				vlAspectTermMap.append(True)
			else:
				vlAspectTermMap.append(False)
		
		return vlAspectTermMap
		
	
	
	def generateNGramKTree(self, pFormat = "binary", pdATReprOptions = None, pdOptions = None):
		'''
		Generates and returns a tree representation of the sentence tokens
		
		pdOptions contains specific options to each format.
		'''
		
		if pdOptions is None:
			pdOptions = {}
		
		if pFormat == "unary":
			if pdATReprOptions is None:
				self.kTree = tk.NGramKTree.generateNGramKTree(self, pFormat = "unary")
			elif "node" in pdATReprOptions and pdATReprOptions["node"].lower() == "parent":
				self.kTree = "( %s)" % self._generateABSANGramKTree1(self.tokens, self.aspectTermMap)
			elif "suffix" in pdATReprOptions and pdATReprOptions["suffix"].lower() == "token":
				self.kTree = "( %s)" % self._generateABSANGramKTree5(self.tokens + ["</s>"], self.aspectTermMap + [False])
		elif pFormat == "bigram":
			if pdATReprOptions is None:
				self.kTree = tk.NGramKTree.generateNGramKTree(self, pFormat = "bigram")
			elif "node" in pdATReprOptions and pdATReprOptions["node"].lower() == "parent":
				self.kTree = "( %s)" % self._generateABSANGramKTree2(self.tokens, self.aspectTermMap)
		elif pFormat == "binary":
			# processing format options
			if "auxiliary-node" in pdOptions:            # content of auxiliary node for unigram extraction (default: X)
				if pdOptions["auxiliary-node"].lower() == "pos":
					plUniAuxTokens = self.absaSent.getPOSTags()
			else:
				plUniAuxTokens = ['X']*len(self.tokens)
			
			if pdATReprOptions is None:
				self.kTree = tk.NGramKTree.generateNGramKTree(self, pFormat = "binary")
			elif "node" in pdATReprOptions and pdATReprOptions["node"].lower() == "x":
				self.kTree = "( %s)" % self._generateABSANGramKTree3(["<s>"] + self.tokens, [False] + self.aspectTermMap)
			elif "suffix" in pdATReprOptions and pdATReprOptions["suffix"].lower() == "token":
				self.kTree = "( %s)" % self._generateABSANGramKTree4(["<s>"] + self.tokens, [False] + self.aspectTermMap, [None] + plUniAuxTokens)
			elif "node" in pdATReprOptions and pdATReprOptions["node"].lower() == "parent":
				raise Exception("Change parent to X in pdATReprOptions and run again")   # temporary
		else:
			self.kTree = tk.NGramKTree.generateNGramKTree(self)
		
		return self.kTree
		
	
	
	def _generateABSANGramKTree1(self, plTokens, plAspectTermMap):
		'''
		Recursively generates the n-gram tree in the unary format with AT node representated as parent node to the aspect
		term token node 
		
		The aspect term map determines the aspect term tokens. 
		'''
		
		if len(plTokens) == 1:
			if plAspectTermMap[0]:
				return "(AT (%s ))" % plTokens[0]
			else:
				return "(%s )" % plTokens[0]
		elif len(plTokens) == 2:
			if plAspectTermMap[0] and plAspectTermMap[1]:
				return "(AT (%s (AT %s)))" % (plTokens[0], plTokens[1])
			elif plAspectTermMap[0] and not plAspectTermMap[1]:
				return "(AT (%s %s))" % (plTokens[0], plTokens[1])
			elif not plAspectTermMap[0] and plAspectTermMap[1]:
				return "(%s (AT %s))" % (plTokens[0], plTokens[1])
			else:
				return "(%s %s)" % (plTokens[0], plTokens[1])
		else:
			if plAspectTermMap[0]:
				return "(AT (%s %s))" % (plTokens[0], self._generateABSANGramKTree1(plTokens[1:], plAspectTermMap[1:]))
			else:
				return "(%s %s)" % (plTokens[0], self._generateABSANGramKTree1(plTokens[1:], plAspectTermMap[1:]))
		
	
	
	def _generateABSANGramKTree2(self, plTokens, plAspectTermMap):
		'''
		Generates the n-gram tree in the bigram format with AT node inserted above each aspect term token
		
		The aspect term map determines the aspect term tokens. 
		'''
		
		if len(plTokens) == 1:
			if plAspectTermMap[0]:
				return "(AT (%s %s)" % (plTokens[0], plTokens[0])
			else:
				return "(%s %s)" % (plTokens[0], plTokens[0])
		else:
			vTree = ''
			
			for i in range(len(plTokens) - 1):
				if plAspectTermMap[i] and plAspectTermMap[i+1]:
					vTree += "(AT (%s (%s (AT %s))))" % (plTokens[i], plTokens[i], plTokens[i+1])
				elif plAspectTermMap[i]:
					vTree += "(AT (%s (%s %s)))" % (plTokens[i], plTokens[i], plTokens[i+1])
				elif plAspectTermMap[i+1]:
					vTree += "(%s (%s (AT %s)))" % (plTokens[i], plTokens[i], plTokens[i+1])
				else:
					vTree += "(%s (%s %s))" % (plTokens[i], plTokens[i], plTokens[i+1])
			
			return vTree
		
	
	
	def _generateABSANGramKTree3(self, plTokens, plAspectTermMap):
		'''
		Recursively generates the tree in the binary format with AT node representation
		
		Example: The food was good.
		         (<s> (The X)(The (food AT)(food (was X)(was (good X)(good (. X)(. )))))) )
		
		The aspect term map determines the aspect term tokens. 
		'''
		
		## Note: <s> is supposed to be added at the beginning of tokens, so len(plTokens)=1 is no possible.
		## Note: When len(plTokens)=2 and the first token is an AT, it is already looked after in the previous step, or
		## it is <s> and cannot be an aspect term.
		
		if len(plTokens) == 2:
			if plAspectTermMap[1]:
				return "(%s (%s AT)(%s ))" % (plTokens[0], plTokens[1], plTokens[1])
			else:
				return "(%s (%s X)(%s ))" % (plTokens[0], plTokens[1], plTokens[1])
		else:
			if plAspectTermMap[1]:
				return "(%s (%s AT)%s)" % (plTokens[0], plTokens[1], self._generateABSANGramKTree3(plTokens[1:], plAspectTermMap[1:]))
			else:
				return "(%s (%s X)%s)" % (plTokens[0], plTokens[1], self._generateABSANGramKTree3(plTokens[1:], plAspectTermMap[1:]))
		
	
	
	def _generateABSANGramKTree4(self, plTokens, plAspectTermMap, plUniAuxTokens):
		'''
		Recursively generates the tree in the binary format with AT node representation
		
		Example: The food was good.
		         (<s> (The X)(The (food_AT X)(food_AT (was X)(was (good X)(good (. X)(. )))))) )
		
		The aspect term map determines the aspect term tokens. 
		
		plUniAuxTokens is the list of tokens to be used under left child node's to help extract unigrams. Default is a 
		list of 'X'.
		'''
		
		## Note: <s> is supposed to be added at the beginning of tokens, so len(plTokens)=1 is no possible.
		## Note: When len(plTokens)=2 and the first token is an AT, it is already looked after in the previous step, or
		## it is <s> and cannot be an aspect term.
		
		if len(plTokens) == 2:
			if plAspectTermMap[0] and plAspectTermMap[1]:
				return "(%s_AT (%s_AT %s)(%s_AT ))" % (plTokens[0], plTokens[1], plUniAuxTokens[1], plTokens[1])
			elif plAspectTermMap[0]:
				return "(%s_AT (%s %s)(%s ))" % (plTokens[0], plTokens[1], plUniAuxTokens[1], plTokens[1])
			elif plAspectTermMap[1]:
				return "(%s (%s_AT %s)(%s_AT ))" % (plTokens[0], plTokens[1], plUniAuxTokens[1], plTokens[1])
			else:
				return "(%s (%s %s)(%s ))" % (plTokens[0], plTokens[1], plUniAuxTokens[1], plTokens[1])
		else:
			if plAspectTermMap[0] and plAspectTermMap[1]:
				return "(%s_AT (%s_AT %s)%s)" % (plTokens[0], plTokens[1], plUniAuxTokens[1], self._generateABSANGramKTree4(plTokens[1:], plAspectTermMap[1:], plUniAuxTokens[1:]))
			elif plAspectTermMap[0]:
				return "(%s_AT (%s %s)%s)" % (plTokens[0], plTokens[1], plUniAuxTokens[1], self._generateABSANGramKTree4(plTokens[1:], plAspectTermMap[1:], plUniAuxTokens[1:]))
			elif plAspectTermMap[1]:
				return "(%s (%s_AT %s)%s)" % (plTokens[0], plTokens[1], plUniAuxTokens[1], self._generateABSANGramKTree4(plTokens[1:], plAspectTermMap[1:], plUniAuxTokens[1:]))
			else:
				return "(%s (%s %s)%s)" % (plTokens[0], plTokens[1], plUniAuxTokens[1], self._generateABSANGramKTree4(plTokens[1:], plAspectTermMap[1:], plUniAuxTokens[1:]))
		
		
	
	
	def _generateABSANGramKTree5(self, plTokens, plAspectTermMap):
		'''
		Recursively generates the n-gram tree in the unary format with AT node representated by attaching a suffic to 
		the aspect term token node 
		
		The aspect term map determines the aspect term tokens. 
		'''
		
		if len(plTokens) == 1:
			if plAspectTermMap[0]:
				return "(%s_AT )" % plTokens[0]
			else:
				return "(%s )" % plTokens[0]
		elif len(plTokens) == 2:
			if plAspectTermMap[0] and plAspectTermMap[1]:
				return "(%s_AT %s_AT)" % (plTokens[0], plTokens[1])
			elif plAspectTermMap[0] and not plAspectTermMap[1]:
				return "(%s_AT %s)" % (plTokens[0], plTokens[1])
			elif not plAspectTermMap[0] and plAspectTermMap[1]:
				return "(%s_AT %s)" % (plTokens[0], plTokens[1])
			else:
				return "(%s %s)" % (plTokens[0], plTokens[1])
		else:
			if plAspectTermMap[0]:
				return "(%s_AT %s)" % (plTokens[0], self._generateABSANGramKTree5(plTokens[1:], plAspectTermMap[1:]))
			else:
				return "(%s %s)" % (plTokens[0], self._generateABSANGramKTree5(plTokens[1:], plAspectTermMap[1:]))
		
		
		
	
	