#! /usr/bin/python
# -*- coding: utf-8 -*-

"""	
	This module defines classes for document level sentiment analysis (DLSA) with Amazon review data.


	Version 0.1										22-Feb-2018
	- AmazonSet is added
"""

from dlsa import *
from nlp import nlp
from utils import util

import random


class AmazonSet(DLSASet):
	'''
	Class for document level sentiment analysis (DLSA) with Amazon review data 
	'''
	
	
	def __str__(self):
		'''
		String representation
		'''
		
		vStr = ''
		
		for vReview in self.getReviews():
			vStr += vReview.getStr() + '\n.............................................\n'
		
		return vStr
		
	
	
	def getIDs(self):
		'''
		Return unique IDs of the reviews in a list
		'''
		
		return [r.getID() for r in self.getReviews()]
		
	
	
	def getReviews(self):
		'''
		Alising documents
		'''
		
		return self.documents
		
	
	
	def addReviews(self, plReviews):
		'''
		Adds a list of reviews to the dataset
		'''
		
		self.documents += plReviews
		
	
	
	def load(self, plFilenames, plPolarities, pFormat = "ziser", pflgTokenize = False):
		'''
		Loads data from a list of Amazon review files each associated with
		a sentiment polarity value
		
		Each review file is expected to contain reviews all with the same
		sentiment polarity. The sentiment polarity values are expected to 
		be given in a list, in the same order as the review file list. 
		
		Amazon review files should be in either "blitzer" format similar
		to those in https://www.cs.jhu.edu/~mdredze/datasets/sentiment/
		or "yftah" format similar to those in 
		https://github.com/yftah89/structural-correspondence-learning-SCL/tree/master/data.
		'''
		
		for vFilename, vPol in zip(plFilenames, plPolarities):
			if pFormat.lower() == "blitzer":
				self.addReviews(self._loadFromBlitzer(vFilename, vPol, pflgTokenize))
			elif pFormat.lower() == "ziser":
				self.addReviews(self._loadFromZiser(vFilename, vPol, pflgTokenize))
		
	
	
	def _loadFromBlitzer(self, pFilename, pPolarity, pflgTokenize):
		'''
		Loads reviews from Amazon review file in format used by Blitzer 
		at al. (2007) together with its sentiment polarity 
		'''
		
		vlReviews = []
		
		for vLine in open(pFilename).readlines():
			if vLine.strip() == "<review>":
				vlReviewLines = []
			elif vLine.strip() == "</review>":
				vReview = AmazonReview(self)
				vReview.loadReviewFromLines(vlReviewLines, "blitzer")
				vReview.setPolarity(pPolarity)
				
				if pflgTokenize:
					vReview.tozenizeText()
				
				vlReviews.append(vReview)
			else:
				vlReviewLines.append(vLine)
		
		return vlReviews
		
	
	
	def _loadFromZiser(self, pFilename, pPolarity, pflgTokenize):
		'''
		Loads reviews from Amazon review file in format used by Ziser and 
		Reichart (2017) together with its sentiment polarity
		
		Only tab characters are removed.
		'''
		
		vlReviews = []
		
		for vLine in open(pFilename).readlines():
			if "</review>" in vLine:                                    # end of review
				vIdx = vLine.find("</review>")
				if vIdx > 0:
					vlReviewLines.append(vLine[ : vIdx].replace('\t', ''))
				
				vReview = AmazonReview(self)
				vReview.loadReviewFromLines(vlReviewLines, "ziser")
				vReview.setPolarity(pPolarity)
				
				if pflgTokenize:
					vReview.tokenizeText()
				
				vlReviews.append(vReview)
			
			if "<review>" in vLine:                                     # start of review
				vIdx = vLine.find("<review>")
				vlReviewLines = [vLine[vIdx + 8 : ].replace('\t', '')]
			
			if "<reviews>" not in vLine  and "<review>" not in vLine and "</review>" not in vLine:    # middle review lines
				vlReviewLines.append(vLine.replace('\t', ''))
		
		return vlReviews
		
	
	
	def subSet(self, pMethod = "random", pSize = 0, pIDs = '', pIDDelim = ',', pflgComplement = False):
		'''
		Selects a subset of reviews from the data based on the specified
		method and returns them in a set.
		
		pMethod is:
		- random: pSize sentences are selected randomly
		- list: selection is done based on a list of review IDs
		
		If pflgComplement is set to true and the method is "list", the 
		complement of specified indexes is selected.
		'''
		
		# preparing IDs of selected reviews
		if pMethod.lower() in ["r", "rand", "random"]:
			vlSelIDs = random.sample(self.getIDs(), pSize)
		elif pMethod.lower() in ["l", "list"]:
			vlSelIDs = pIDs.split(pIDDelim)
		
		# extracting selected reviews: a bit expensive a search
		vlSelReviews = []
		for vReview in self.getReviews():
			if (not pflgComplement and vReview.getID() in vlSelIDs) or (pflgComplement and vReview.getID() not in vlSelIDs):
				vlSelReviews.append(vReview)
		
		# returning selection as a set
		vNewSet = AmazonSet()
		vNewSet.addReviews(vlSelReviews)
		
		return vNewSet
		
	
	
	def alternatePolarities(self, plPolarityOrder = None):
		'''
		Reorders the reviews so that the polarities alternate
		
		The alternation will be based on the given order, if any, or ad-hoc
		otherwise. 
		'''
		
		vlReorderedReviews = []
		vdPolBins = {}
		
		# putting reviews in polarity bins
		for vReview in self.getReviews():
			vPol = vReview.getPolarity()
			if vPol in vdPolBins:
				vdPolBins[vPol].append(vReview)
			else:
				vdPolBins[vPol] = [vReview] 
		
		if plPolarityOrder:
			vlZip = zip(*[vdPolBins[p] for p in plPolarityOrder])
		else:
			vlZip = zip(*vdPolBins.itervalues())
		
		for vtReviews in vlZip:
			for i in range(len(vdPolBins)):
				vlReorderedReviews.append(vtReviews[i])
		
		# returning selection as a set
		vNewSet = AmazonSet()
		vNewSet.addReviews(vlReorderedReviews)
		
		return vNewSet
		
	
	



class AmazonReview(DLSADoc):
	'''
	Class for Amazon review units
	'''
	
	
	def __init__(self, pAmazaonSet):
		'''
		Constructor
		'''
		
		DLSADoc.__init__(self, pAmazaonSet)
		
		self.strID = None
		self.intID = None
		self.asin = None
		self.product_name = None
		self.product_type = None
		self.helpful = None
		self.rating = None
		self.title = None
		self.date = None
		self.reviewer = None
		self.reviewer_location = None
		
		# in DLSADoc
		#self.text = None
		#self.polarity = None
		
	
	
	def __str__(self):
		'''
		String representation
		'''
		
		return self.getStr()
		
	
	
	def getStr(self):
		'''
		Returns string representation of the review
		'''
		
		vStr  = "Unique ID        : (%s, %s)\n" % self.getID()
		vStr += "ASIN             : %s\n" % self.asin
		vStr += "Product Name     : %s\n" % self.product_name
		vStr += "Product Type     : %s\n" % self.product_type
		vStr += "Helpful          : %s\n" % self.helpful
		vStr += "Rating           : %s\n" % self.rating
		vStr += "Title            : %s\n" % self.title
		vStr += "Date             : %s\n" % self.date
		vStr += "Reviewer Name    : %s\n" % self.reviewer
		vStr += "Reviewer Location: %s\n" % self.reviewer_location
		vStr += "Review Text      : \n\n%s\n" % self.text
		vStr += "Polarity         : %s"   % self.polarity
		
		return vStr
		
	
	
	def getID(self):
		'''
		Returns the unique id of the review which is the combination of
		the two unique IDs in the review data
		'''
		
		return (self.strID, self.intID)
		
	
	
	def loadReviewFromLines(self, plReviewLines, pFormat):
		'''
		Loads review from lines read from Amazon review in either Blitzer
		or Ziser format
		'''
		
		if pFormat.lower() == "blitzer":
			self._loadBlitzerLines(plReviewLines)
		elif pFormat.lower() == "ziser":
			self._loadZiserLines(plReviewLines)
		
	
	
	def _loadBlitzerLines(self, plReviewLines):
		'''
		Loads review from lines read from Amazon review in Blitzer format
		'''
		
		i = 0
		
		while i < len(plReviewLines):
			if plReviewLines[i].strip().startswith('</') and plReviewLines[i].strip().endswith('>'):
				self._loadTagData(vTag, vContent)
			elif plReviewLines[i].strip().startswith('<') and plReviewLines[i].strip().endswith('>'):
				vTag = plReviewLines[i][1:-2]
				vContent = ''
			else:
				vContent += plReviewLines[i]
			
			i += 1
		
	
	
	def _loadZiserLines(self, plReviewLines):
		'''
		Loads review from lines read from Amazon review in Ziser format
		'''
		
		# Ziser format only contain review text. Other fields will remain 
		# None.
		
		self.text = ''.join(plReviewLines)
		
	
	
	def _loadTagData(self, pTag, pContent):
		'''
		Loads data from given content for the given tag
		'''
		
		if pTag == "unique_id":
			if util.isNumeric(pContent):
				self.intID = int(pContent.strip())
			else:
				self.strID = pContent.strip()
		elif pTag == "asin":
			self.asin = pContent.strip()
		elif pTag == "product_name":
			self.product_name = pContent.strip()
		elif pTag == "product_type":
			self.product_type = pContent.strip()
		elif pTag == "helpful":
			self.helpful = pContent.strip()
		elif pTag == "rating":
			self.rating = pContent.strip()
		elif pTag == "title":
			self.title = pContent.strip()
		elif pTag == "date":
			self.date = pContent.strip()
		elif pTag == "reviewer":
			self.reviewer = pContent.strip()
		elif pTag == "reviewer_location":
			self.reviewer_location = pContent.strip()
		elif pTag == "review_text":
			self.text = pContent.strip()
		else:
			raise Exception("Unknown tag: %s" % pTag)
		
	
	
	def setPolarity(self, pPolarity):
		'''
		Sets the sentiment polarity of the review
		'''
		
		self.polarity = pPolarity
		
	
	
	def getPolarity(self):
		'''
		Returns the sentiment polarity of the review
		'''
		
		return self.polarity
		
	
	
	def getReviewText(self):
		'''
		Returns review text
		'''
		
		return self.text
		
	
	
	def getReviewTokens(self):
		'''
		Returns review token list
		
		For tokens to make sense, make sure you use the tokenize option when loading the data.
		'''
		
		return self.text.split(' ')
		
	
	
	def tokenizeText(self):
		'''
		Tokenizes review text
		'''
		
		self.text = nlp.tokenizeSegment(self.text, pLang = "en", pflgTokenizeFSlash = False, pflgUnicodeErrs = "replace")
		
	
	
