#! /usr/bin/python
# -*- coding: utf-8 -*-

"""	
	This module defines classes for document-level sentiment analysis (DLSA).
	
	Version 0.1										22-Feb-2018 to 06-Mar-2018
	- DLSASet, DLSADoc are added.
"""


class DLSASet:
	'''
	Abstract class for document level sentiment analysis dataset.
	
	The class is abstract because there is no method for loading data. It
	should be implemented in classes inherited to implement a specific
	format.
	'''
	
	
	def __init__(self):
		'''
		Constructor 
		'''
		
		self.documents = []
		
		# WordVector object containing the word vectors of the vocabulary in the dataset
		self.wv = None
		
	
	
	@property
	def size(self):
		'''
		Returns the size of the data set which is the number of its documents
		'''
		
		return len(self.getDocuments())
		
	
	
	@property
	def tokenLength(self):
		'''
		Returns the number of tokens in the data set
		'''
		
		return sum([d.tokenLength for d in self.getDocuments()])
		
	
	
	def getDocuments(self):
		'''
		Returns the DLSA documents
		'''
		
		return self.documents
		
	
	
	def addDocuments(self, plDocuments):
		'''
		Adds DLSA documents to existing documents
		'''
		
		for vDoc in plDocuments:
			vDoc.dataset = self
			self.documents.append(vDoc)
		
	
	
	def getVocabulary(self):
		'''
		Extracts and returns the vocabulary of the dataset 
		'''
		
		return sorted(set([t for d in self.getDocuments() for t in d.getTokens()]))
		
	
	
	def loadSentimentScores(self, pSentexicon, pNeutralScore = None):
		'''
		Loads polarity scores to documents, from a sentiment lexicon which
		is a Sentexicon object
		
		For details about Sentexicon object, see sentexicon.py
		'''
		
		vTotalWordNum = 0          # total number of words in the data set
		vTotalEntryWordNum = 0     # number of words found in the lexicon
		
		for vDoc in self.getDocuments():
			vWordNum, vEntryWordNum = vDoc.loadSentimentScores(pSentexicon, pNeutralScore)
			
			vTotalWordNum += vWordNum
			vTotalEntryWordNum += vEntryWordNum
		
		return vTotalWordNum, vTotalEntryWordNum
		
	
	
	def getSentimentScores(self):
		'''
		Returns a dictionary of words in the dataset and the sentiment 
		scores attached to them 
		'''
		
		vdResult = {}
		
		for vSent in self.getDocuments():
			for vWord, vScore in zip(vDoc.getTokens(), vDoc.getSentimentScores()):
				vdResult[vWord] = vScore
		
		return vdResult
		
	
	
	def loadWordVectors(self, pWordVectors, pflgFilter = True):
		''' 
		Loads word vectors from a file or WordVector object, whichever is given
		
		By default, it filters out the words not in the data vocabulary, which can be changed to not filter (e.g. when 
		the input is already filtered).   
		'''
		
		from ml import wv
		
		self.wv = wv.WordVector()
		
		if type(pWordVectors) == str:
			vWV.load(pWVFilename = vWVFile, plFilterVocab = self.getVocabualry())
		else:
			self.wv = pWordVectors
		
	
	



class DLSADoc:
	'''
	Abstract class for DLSA document
	
	The class is abstract because there is no method for loading data. It
	should be implemented in classes inherited to implement a specific
	format.
	'''
	
	
	def __init__(self, pDLSASet):
		'''
		Constructor
		'''
		
		# DLSASet the document belongs to
		self.document = pDLSASet
		
		self.text = None
		self.polarity = None 
		
		self.sentScores = []     # sentiment score, one per word in the tokenized self.text
		
	
	
	def getText(self):
		'''
		Returns document text (form)
		'''
		
		return self.text
		
	
	
	def getTokens(self):
		'''
		Returns the tokenization of the document
		
		The document text is assumed to be in tokenized format and only 
		splits on space.
		'''
		
		return self.getText().split()
		
	
	
	def getPolarity(self):
		'''
		Returns the polarity of the document
		'''
		
		return self.polarity
		
	
	
	@property
	def length(self):
		'''
		Returns the document length
		'''
		
		return len(self.getTokens())
		
	
	
	def loadSentimentScores(self, pSentexicon, pNeutralScore = None):
		'''
		Loads sentiment scores to words in the document from a sentiment
		lexicon which is a Sentexicon object
		
		It returns a tuple of the total number of words in the document 
		and the number of words found in the lexicon.
		
		If the word is not found in the lexicon, None will be used.
		
		For details about Sentexicon object, see sentexicon.py
		'''
		
		vlWords = self.getTokens()
		self.sentScores = []
		
		for vWord in vlWords:
			vScore = pSentexicon.getScore(vWord)
			if vScore is None and pNeutralScore is not None:
				self.sentScores.append(pNeutralScore)
			else:
				self.sentScores.append(vScore)
		
		return self.length, len([s for s in self.sentScores if s is not None])
		
	
	
	def getSentimentScores(self):
		'''
		Returns the sentiment scores in a list corresponding to the token list
		'''
		
		return self.sentScores
		
	
	
	def getAvgSentScore(self):
		'''
		Calculates and returns the average sentiment score document tokens
		'''
		
		vlSentScores = self.getSentimentScores()
		
		return sum(vlSentScores) / len(vlSentScores)
		
	
	
	def getPolarScores(self, pNeutralScore = None):
		'''
		Returns sentiment scores with non-neutral polarity
		
		Neutral polarity score can be set as parameter. It is None by default meaning that no polarity score is assigned
		to neutral words.
		'''
		
		return [s for s in self.sentScores if s != pNeutralScore]
		
	
	
	def getWordVectors(self):
		'''
		Returns the word vectors of the document tokens
		'''
		
		return [self.dataset.wv.getVector(t) for t in self.getTokens()]
		
	
	
