# -*- coding: utf-8 -*-
__author__ = "rszk"


"""	
	This module defines classes for loading MPQA corpus.
	
	Version 1.1                            													(14-Oct-2016 to 21-Oct-2016)
	- Some problems in annotations are rectified using code.
	- Document loaders can now take care of the version-specific document file name.
	- A new flag parameter is used to specifiy whether the program should stop and raise an error when facing an unknown
	  annotation type during document loading.
	- getBIOTagging() is added to MPQADoc and MPQASentence.
	- getDocument() is added to MPQA.
	- extractOrgSentForms() and extractTokSentForms() are added to MPQA.
	
	Version 1.0                            																	(02-Feb-2015)
	- Constituency parse trees are now embedded in the data. 
	
	Version 0.1                            													(04-Dec-2015 to 19-Jan-2015)
	- MPQA, MPQADoc, MPQASentence, MPQAAnn, ExpressiveSubjective, DirectSubjective, ObjSpeechEvent, Agent, Target, Inside,
	  Attitude, TargetSpeech, ETarget and TargetFrame are added.
"""


import re
from collections import namedtuple
from nlp import nlp
from utils import util


class MPQA(object):
	'''
	Class for MPQA v2 and v3 corpus
	'''
	
	
	def __init__(self, pDatabase, pVersion = 2.0):
		'''
		Constructor
		
		pDatabase is the path to MPQA document database directory. The  directory is assumed to have the structure 
		described in 
		'''
		
		self.database = pDatabase
		
		self.version = pVersion
		
		# non-specific documents also containing version 3 documents (MPQADoc objects) 
		self.docs = []
		
		# original MPQA documents (MPQADoc objects)
		self.docsMPQAOrg = []
		
		# OpQA subset documents (MPQADoc objects)
		self.docsOpQA = []
		
		# XBank subset documents (MPQADoc objects)
		self.docsXBank = []
		
		# ULA subset documents (MPQADoc objects)
		self.docsULA = []
		
		# ULA-LU subset documents (MPQADoc objects)
		self.docsULALU = []
		
	
	
	@property
	def size(self):
		'''
		Returns database size as the number of documents
		'''
		
		return len(self.allDocs)
		
	
	
	@property
	def usize(self):
		'''
		Returns database size as the number of documents with duplicates removed
		'''
		
		return len(self.uallDocs)
		
	
	
	@property
	def allDocs(self):
		'''
		Returns the concatenation of lists of all documents 
		
		NOTE: there are duplicate documents across various document types (e.g. non_fbis/08.47.00-17401). To get the 
		documents with this duplicates removed, use uallDocs.
		'''
		
		return self.docs + self.docsMPQAOrg + self.docsOpQA + self.docsXBank + self.docsULA + self.docsULALU
		
	
	
	@property
	def uallDocs(self):
		'''
		Returns the concatenation of lists of all documents with duplicates removed
		
		NOTE: can be made more efficient by storing the value in a class attribute the value of which is set at first 
		access.
		'''
		
		vlAllDocFilenames = []
		vlAllDocsUnique = []
		
		for vDoc in self.allDocs:
			if vDoc.filename not in vlAllDocFilenames:
				vlAllDocsUnique.append(vDoc)
				vlAllDocFilenames.append(vDoc.filename)
		
		return vlAllDocsUnique
		
	
	
	@property
	def sentences(self):
		'''
		Returns all sentence objects in the corpus
		
		NOTE: there are duplicate documents across various document types (e.g. non_fbis/08.47.00-17401). To get the 
		sentences with this duplicates removed, use uSentences.
		'''
		
		return [s for d in self.uallDocs for s in d.sentences]
		
	     
	
	@property
	def uSentences(self):
		'''
		Returns all sentence objects from unique documents
		
		NOTE: there are duplicate documents across various document types. The outpus of 
		this method is sentences from deduplicated documents.
		'''
		
		return [s for d in self.uallDocs for s in d.sentences]
		
	     
	
	def loadDocs(self, plDocFileNames, pflgStopUnknownAnn = True):
		'''
		Loads documents whose names (including path) are given in a list
		 
		pflgStopUnknownAnn specified whether the program should stop and raise an error when facing an unknown annotation
		type.
		'''
		
		self.docs = []
		
		for vFile in plDocFileNames:
			# skipping over some known problematic documents
			self.docs.append(MPQADoc(pCorpus = self,
									 pDocFilename = vFile.strip(),
									 pflgStopUnknownAnn = pflgStopUnknownAnn))
		
	
	
	def loadAll(self):
		'''
		Loads all subsets from database
		'''
		
		if self.version == 2:
			# loading original
			self.loadMPQAOrg()
		
			# loading OpQA
			self.loadOpQA()
		
			# loading XBank        ignored at the moment due to several sentence segmentation problems (some already handled)
			#self.loadXBank()
		
			#loading ULA
			self.loadULA()
		
		# loading ULA-LU
		#self.loadULALU()      ignored at the moment because the text looks messy and not well cleaned
		elif self.version == 3:
			self.loadV3()
		
	
	
	def loadMPQAOrg(self):
		'''
		Loads the original MPQA corpus
		'''

		self.docsMPQAOrg = []

		# loading documents by topic
		for vTopicFile in open("doclist.mpqaOriginalByTopic").read().strip().split('\n'):
			vlTFSplit = vTopicFile.split()

			# skipping over some known problematic documents
			if vlTFSplit[1][5:] in ["non_fbis/12.14.07-203"]:
				continue

			self.docsMPQAOrg.append(MPQADoc(pCorpus = self,
											pDocFilename = vlTFSplit[1][5:],
											pTopic = vlTFSplit[0][6:]))



	def loadOpQA(self):
		'''
		Loads the OpQA subset
		'''

		self.docsOpQA = []

		for vFile in open("doclist.opqaSubset").read().strip().split('\n'):
			self.docsOpQA.append(MPQADoc(pCorpus = self,
										 pDocFilename = vFile.strip()))



	def loadXBank(self):
		'''
		Loads the XBank subset
		'''

		self.docsXBank = []

		for vFile in open("doclist.xbankSubset").read().strip().split('\n'):
			# skipping over some known problematic documents
			if vFile.strip() in ["xbank/wsj_0551"]:                       # short doc with many sentence split problem
				continue

			self.docsXBank.append(MPQADoc(pCorpus = self,
										  pDocFilename = vFile.strip()))



	def loadULA(self):
		'''
		Loads the ULA subset (48 texts from the American National Corpus)
		'''

		self.docsULA = []

		for vFile in open("doclist.ulaSubset").read().strip().split('\n'):
			# skipping over some known problematic documents
			if vFile.strip() == "ula/NapierDianne":
				continue

			self.docsULA.append(MPQADoc(pCorpus = self,
										pDocFilename = vFile.strip()))



	def loadULALU(self):
		'''
		Loads the ULA-LU subset (24 texts from the ULA language understanding subcorpus)
		'''

		self.docsULALU= []

		for vFile in open("doclist.ula-luSubset").read().strip().split('\n'):
			self.docsULALU.append(MPQADoc(pCorpus = self,
										  pDocFilename = vFile.strip()))



	def loadV3(self):
		'''
		Loads version 3 documents
		'''
		
		self.docs = []
		
		for vFile in open("doclist").read().strip().split('\n'):
			self.docs.append(MPQADoc(pCorpus = self,
									 pDocFilename = vFile.strip()))
		
	
	
	def getDocument(self, pDocFilename, pflgErrIfNotFound = True):
		'''
		Returns the document object corresponding to the given document file name
		'''
		
		for doc in self.docs:
			if doc.filename == pDocFilename:
				return doc
		
		if pflgErrIfNotFound:
			raise Exception("Document %s not found!" % pDocFilename)
		else:
			return None
		
	
	
	def loadPOSTagging(self):
		'''
		Loads the POS tagging of the sentences of the corpus
		
		The POS tag files are supposed to be stored in the pos directory in the database path under the same directory
		structure and with the same file names suffixed with .pos as the documents themselves.
		'''
		
		for vDoc in self.allDocs:
			vDoc.loadPOSTagging()
		
	
	
	def loadConstParses(self, pFileSuffix):
		'''
		Loads the constituency parses of the sentences of the corpus
		
		The constituency parse files are supposed to be stored in the "parse" directory in the database path under the 
		same directory structure and with the same file names suffixed with pFileSuffix as the documents themselves.
		'''
		
		for vDoc in self.allDocs:
			vDoc.loadConstParses(pFileSuffix)
		
	
	
	def tokenizeDocs(self, pflgSave = False):
		'''
		Generates a sentence-segmented and tokenized version of the documents
		
		The results are returned in a dictionary with the document file names as keys and the tokenized content in a 
		string as values. Each line in the output content contains a tokenized sentence.
		
		If the save flag parameter is set, the output will be saved in a file with the same name as the corresponding 
		original document file name suffixes with .tok.
		'''
		
		vdOutput = {}
		
		for vDoc in self.allDocs:
			print vDoc.filename
			vdOutput[vDoc.filename] = vDoc.tokenize(pflgSave)
		
		return vdOutput
		
	
	
	def getSentNum(self):
		'''
		Returns the number of sentences in the database
		'''
		
		return sum([d.size for d in self.uallDocs])
		
	
	
	def getAnnNum(self):
		'''
		Returns the number of annotations in the database
		'''
		
		return sum([d.getAnnNum() for d in self.uallDocs])
		
	
	
	def getDS(self, pdFilter = {}):
		'''
		Returns the direct subjective annotations of the corpus
		'''
		
		return [a for d in self.uallDocs for a in d.getDS(pdFilter)]
		
	
	
	def getES(self, pdFilter = {}):
		'''
		Returns the expressive subjective annotations of the corpus
		'''
		
		return [a for d in self.uallDocs for a in d.getES(pdFilter)]
		
	
	
	def getOSE(self, pdFilter = {}):
		'''
		Returns the objective speech event annotations of the corpus
		'''
		
		return [a for d in self.uallDocs for a in d.getOSE(pdFilter)]
		
	
	
	def getAttitude(self, pdFilter = {}):
		'''
		Returns the attitude annotations of the corpus
		'''
		
		return [a for d in self.uallDocs for a in d.getAttitude(pdFilter)]
		
	
	
	def getSTarget(self):
		'''
		Returns the (span-based) target annotations of the corpus
		'''
		
		return [a for d in self.uallDocs for a in d.getSTarget()]
		
	
	
	def getETarget(self):
		'''
		Returns the entity/event target annotations of the corpus
		'''
		
		return [a for d in self.uallDocs for a in d.getETarget()]
		
	
	
	def getTF(self):
		'''
		Returns the target frame annotations of the corpus
		'''
		
		return [a for d in self.uallDocs for a in d.getTF()]
		
	
	
	def getAgent(self):
		'''
		Returns the agent annotations of the corpus
		'''
		
		return [a for d in self.uallDocs for a in d.getAgent()]
		
	
	
	def getTS(self):
		'''
		Returns the target speech annotations of the corpus
		'''
		
		return [a for d in self.uallDocs for a in d.getTS()]
		
	
	
	def getInside(self):
		'''
		Returns the inside annotations of the corpus
		'''
		
		return [a for d in self.uallDocs for a in d.getInside()]
		
	
	
	def getDSNum(self, pdFilter = {}):
		'''
		Returns the number of direct subjective annotations in the corpus
		'''
		
		return len(self.getDS(pdFilter))
		
	
	
	def getESNum(self, pdFilter = {}):
		'''
		Returns the number of expressive subjective annotations in the corpus
		'''
		
		return len(self.getES(pdFilter))
		
	
	
	def getOSENum(self, pdFilter = {}):
		'''
		Returns the number of objective speech event annotations in the corpus
		'''
		
		return len(self.getOSE(pdFilter))
		
	
	
	def getAttitudeNum(self, pdFilter = {}):
		'''
		Returns the number of attitude annotations in the corpus
		'''
		
		return len(self.getAttitude(pdFilter))
		
	
	
	def getSTargetNum(self):
		'''
		Returns the number of (span-based) target annotations in the corpus
		'''
		
		return len(self.getSTarget())
		
	
	
	def getETargetNum(self):
		'''
		Returns the number of entity/event target annotations in the corpus
		'''
		
		return len(self.getETarget())
		
	
	
	def getTFNum(self):
		'''
		Returns the number of target frame annotations in the corpus
		'''
		
		return len(self.getTF())
		
	
	
	def getAgentNum(self):
		'''
		Returns the number of agent annotations in the corpus
		'''
		
		return len(self.getAgent())
		
	
	
	def getTSNum(self):
		'''
		Returns the number of target speech annotations in the corpus
		'''
		
		return len(self.getTS())
		
	
	
	def getInsideNum(self):
		'''
		Returns the number of inside annotations in the corpus
		'''
		
		return len(self.getInside())
		
	
	
	def getUnmatchedAnn(self):
		'''
		Returns list of annotations not matching a sentence boundary
		'''
		
		vlUnmatched = []
		
		for vDoc in self.uallDocs:
			vlUnmatched += vDoc.unmatchedAnn
		
		return vlUnmatched
		
	
	
	def dump(self, pFilename):
		'''
		Pickles the object into the given file
		'''
		
		import pickle
		
		pickle.dump(self, open(pFilename, 'w'))
		
	
	
	def getDSOverlap(self):
		'''
		Returns the overlap between direct subjectives of the sentences in the corpus
		'''
		
		vlOverlap = []
		
		for vDoc in self.uallDocs:
			vlOverlap += vDoc.getDSOverlap()
		
		return vlOverlap
		
	
	
	def getDSWithSTargetOverlap(self):
		'''
		Returns the direct subjectives the span-based targets of which overlap
		'''
		
		vlOverlap = []
		
		for vDoc in self.uallDocs:
			vlOverlap += vDoc.getDSWithSTargetOverlap()
		
		return vlOverlap
		
	
	
	def getDSAttOverlap(self):
		'''
		Returns the overlap between direct subjectives and attitudes of the sentences in the corpus
		'''
		
		vlOverlap = []
		
		for vDoc in self.uallDocs:
			vlOverlap += vDoc.getDSAttOverlap()
		
		return vlOverlap
		
	
	
	def getAvgDSLength(self, pdFilter = {}):
		'''
		Returns the average token length of the direct subjective annotations in the corpus
		'''
		
		vlDS = self.getDS(pdFilter)
		
		if len(vlDS) != 0:
			return sum([a.length for a in vlDS]) * 1.0 / len(vlDS)
		else:
			return 0
		
	
	
	def getAvgESLength(self, pdFilter = {}):
		'''
		Returns the average token length of the expressive subjective annotations in the corpus
		'''
		
		vlES = self.getES(pdFilter)
		
		if len(vlES) != 0:
			return sum([a.length for a in vlES]) * 1.0 / len(vlES)
		else:
			return 0
		
	
	
	def getAvgAttitudeLength(self, pdFilter = {}):
		'''
		Returns the average token length of the attitude annotations in the corpus
		'''
		
		vlAttitude = self.getAttitude(pdFilter)
		
		if len(vlAttitude) != 0:
			return sum([a.length for a in vlAttitude]) * 1.0 / len(vlAttitude)
		else:
			return 0
		
	
	
	def getAttOverlap(self):
		'''
		Returns the overlap between attitudes of the sentences in the corpus
		'''
		
		vlOverlap = []
		
		for vDoc in self.uallDocs:
			vlOverlap += vDoc.getAttOverlap()
		
		return vlOverlap
		
	
	
	def extractOrgSentForms(self):
		'''
		Extract sentence forms of the entire corpus in a list
		'''
		
		return [sent.getOrgForm() for doc in self.docs for sent in doc.sentences]
	
	
	def extractTokSentForms(self):
		'''
		Extract tokenized sentence forms of the entire corpus in a list
		'''
		
		return [sent.getTokenForm() for doc in self.docs for sent in doc.sentences]
		
	
	


class MPQADoc(object):
	'''
	Class for MPQA documents
	'''
	
	
	def __init__(self, pCorpus, pDocFilename, pTopic = None, pflgStopUnknownAnn = True):
		'''
		Constructor
		'''
		
		self.corpus = pCorpus
		self.filename = pDocFilename
		self.topic = pTopic
		
		self.text = None
		self.meta = {}                   # keys: meta attributes (e.g. region for meta_region), values: list of meta values
		self.sentences = []              # list of sentences (MPQASentence objects)
		
		self.unmatchedAnn = []           # annotation objects not matching to a sentence boundary
		
		self._stopUnknownAnn = pflgStopUnknownAnn
		                                 # whether to stop and raise an error when facing an unknown annotation type
		
		self._load()
		
	
	
	@property
	def version(self):
		'''
		Returns the document version which is the version of the corpus it belongs to
		'''
		
		return self.corpus.version
		
	
	
	def _load(self):
		'''
		Loads the document and annotation
		
		It assumes that all files are 
		'''
		
		print self.filename
		
		# loading the document text as is
		self._loadText()
		
		# loading meta-data
		self._loadMeta()
		
		# loading sentences
		self._loadSentences()
		
		# loading annotations
		self.unmatchedAnn = self._loadAnnotation()
		
	
	
	def _loadText(self):
		'''
		Loads document's text
		'''
		
		vText = open("%s/docs/%s" % (self.corpus.database, self.filename)).read()
		
		if self.filename == "ula/NapierDianne":              
			self.text = vText.replace('�', "H")         ## replacing the unicode character used as placeholder causing 
			                                            ## sentence segmentation problem
		else:
			self.text = vText
		
	
	
	def _loadMeta(self):
		'''
		Loads document's meta data
		'''
		
		try:
			vlMetaLines = open("%s/meta_anns/%s" % (self.corpus.database, self.filename)).read().strip().split('\n')
		except IOError:                                 # some documents lack a meta data
			return
		
		for vLine in vlMetaLines:
			if not vLine.startswith('#'):
				vlLSplit = vLine.strip().split()
				vMetaAttr = vlLSplit[3][5:]             # e.g. region from meta_region
				if len(vlLSplit) > 4:                   # some meta attributes lack a value
					vMetaValue = ' '.join(vlLSplit[4:])
				else:
					vMetaValue = ''
				
				if not vMetaAttr in self.meta:
					self.meta[vMetaAttr] = [vMetaValue]
				else:
					self.meta[vMetaAttr].append(vMetaValue)
		
	
	
	def _loadSentences(self):
		'''
		Loads document's sentences
		
		In MPQA, a sentence is not a processing unit. Instead, the corpus is organised as a set of documents. However, a part
		of manual annotation is the sentence segmentation, where sentences in each document are identified by the document
		character offset. We exploit this information and create sentence class.
		
		NOTE: not all sentences in the document are marked. For example, "Copyright 2002 The New York Times Company" in
		non_fbis/08.14.55-18533 is not marked, apparently due to being considered as non-content. 
		'''
		
		# extracting sentence offsets
		vlSentOffsets = []
		for vSentAnn in open("%s/man_anns/%s/gatesentences.mpqa.%s" % (self.corpus.database, self.filename, self.version)).read().strip().split('\n'):
			vlSentAnnSplit = vSentAnn.split()
			vOffset = tuple([int(o) for o in vlSentAnnSplit[1].split(',')])
			
			# handling some known bugs in sentence boundaries
			if self.filename == "non_fbis/06.29.16-13721":
				if vOffset == (4492,4582):
					vOffset = (4492,4698)
				elif vOffset == (4483,4698):
					continue                        # the above two offsets actually belong to one sentence
			elif self.filename == "temp_fbis/21.37.33-17834":
				if vOffset == (2316,2437):
					vOffset = (2316,2552)
				elif vOffset == (2438,2552):
					continue                        # the above two offsets actually belong to one sentence
			elif self.filename == "xbank/wsj_0709":
				if vOffset == (9,22):
					vOffset = (9,116)
				elif vOffset == (23,116):
					continue                        # the above two offsets actually belong to one sentence
			elif self.filename == "xbank/wsj_0541":
				if vOffset == (9,39):
					vOffset = (9,146)
				elif vOffset == (40,63) or vOffset == (67,146):
					continue                        # the above offsets actually belong to one sentence
			elif self.filename == "xbank/wsj_0316":
				if vOffset == (9,29):
					vOffset = (9,123)
				elif vOffset == (30,84) or vOffset == (85,123):
					continue                        # the above offsets actually belong to one sentence
			
			vlSentOffsets.append(vOffset)
		
		# handling some known missing sentences
		if self.filename == "20020415/21.38.12-11637":
			vlSentOffsets.append((79,200))
		elif self.filename == "20020427/22.07.25-26605":
			vlSentOffsets.append((2,109))
		
		# extracting text based on sorted offsets
		for vOffset in sorted(vlSentOffsets):
			self.sentences.append(MPQASentence(pDocument = self, pOffset= vOffset))
		
	
	
	def _getAnnLineComps(self, pAnnLine):
		'''
		Extracts the annotation components from the given annotation line
		'''
		
		vntAnnComps = namedtuple("AnnotationComponents", "id, span, type, attributes")
		
		vlAnnSplit = pAnnLine.split('\t')
		
		vntAnnComps.id = vlAnnSplit[0]
		vntAnnComps.span = vlAnnSplit[1]
		if self.version in [1.2, 2]:
			vntAnnComps.type = vlAnnSplit[3][5:]
			if len(vlAnnSplit) > 4:
				vntAnnComps.attributes = vlAnnSplit[4]
			else:
				vntAnnComps.attributes = ''
		elif self.version == 3:
			vntAnnComps.type = vlAnnSplit[2]
			vntAnnComps.attributes = vlAnnSplit[3]
		
		return vntAnnComps
		
	
	
	def _loadAnnotation(self):
		'''
		Loads document's manual annotation and returns those not matching a sentence boundary
		
		According to the corpus documentation, each document has two agent annotation with offset span 0,0, one 
		representing the writer and the other an implicit agent. In addition to these, which account for 1376 
		annotations (probably for only 688 documents out of total 692) there are 292 such spans for other annotation 
		types. All of these annotations are ignored here.
		'''
		
		# annotation objects not matching to a sentence boundary
		vlUnmatched= []
		
		# extracting annotations
		for vAnn in open("%s/man_anns/%s/gateman.mpqa.lre.%s" % (self.corpus.database, self.filename, self.version)).read().strip().split('\n'):
			if vAnn.startswith('#'):
				continue
			
			vntAnnComps = self._getAnnLineComps(vAnn)
			
			# skipping over some known problematic annotations which could not be solved
			if self.filename == "non_fbis/08.46.28-13637":
				if vntAnnComps.id in ["347", "686", "746"]:      # problem with 746's span but the others linked to it
					continue
			elif self.filename == "ula/Article247_3500":
				if vntAnnComps.id == "1820":                     # seems to be confused with direct-subjective
					continue
			
			if vntAnnComps.type == "direct-subjective":
				vAnnotation = DirectSubjective(pDocument = self,
				                               pAnnID = vntAnnComps.id,
				                               pOffset = self._setAnnOffset(vntAnnComps.id, vntAnnComps.span)) 
				if vntAnnComps.attributes != '':
					vAnnotation.loadAnnAttr(vntAnnComps.attributes.strip(), self._stopUnknownAnn)
				
				if vAnnotation.offset != (0,0):
					vSent = self._findSentenceMatch(vAnnotation)
					if vSent != None:
						vAnnotation.assignSentence(vSent)
						vSent.directSubjective.append(vAnnotation)
					else:
						vlUnmatched.append(vAnnotation)
			elif vntAnnComps.type == "expressive-subjectivity":
				vAnnotation = ExpressiveSubjective(pDocument = self,
				                                   pAnnID = vntAnnComps.id,
				                                   pOffset = self._setAnnOffset(vntAnnComps.id, vntAnnComps.span)) 
				if vntAnnComps.attributes != '':
					vAnnotation.loadAnnAttr(vntAnnComps.attributes.strip(), self._stopUnknownAnn)
				
				if vAnnotation.offset != (0,0):
					vSent = self._findSentenceMatch(vAnnotation)
					if vSent != None:
						vAnnotation.assignSentence(vSent)
						vSent.expressiveSubjective.append(vAnnotation)
					else:
						vlUnmatched.append(vAnnotation)
			elif vntAnnComps.type == "objective-speech-event":
				vAnnotation = ObjSpeechEvent(pDocument = self,
				                             pAnnID = vntAnnComps.id,
				                             pOffset = self._setAnnOffset(vntAnnComps.id, vntAnnComps.span)) 
				if vntAnnComps.attributes != '':
					vAnnotation.loadAnnAttr(vntAnnComps.attributes.strip(), self._stopUnknownAnn)
				
				if vAnnotation.offset != (0,0):
					vSent = self._findSentenceMatch(vAnnotation)
					if vSent != None:
						vAnnotation.assignSentence(vSent)
						vSent.objectiveSpeechEvent.append(vAnnotation)
					else:
						vlUnmatched.append(vAnnotation)
			elif vntAnnComps.type == "agent":
				vAnnotation = Agent(pDocument = self,
				                    pAnnID = vntAnnComps.id,
				                    pOffset = self._setAnnOffset(vntAnnComps.id, vntAnnComps.span)) 
				if vntAnnComps.attributes != '':
					vAnnotation.loadAnnAttr(vntAnnComps.attributes.strip(), self._stopUnknownAnn)
				
				if vAnnotation.offset != (0,0):
					vSent = self._findSentenceMatch(vAnnotation)
					if vSent != None:
						vAnnotation.assignSentence(vSent)
						vSent.agent.append(vAnnotation)
					else:
						vlUnmatched.append(vAnnotation)
			elif vntAnnComps.type in ["target", "sTarget"]:                   # target used in v2 and sTarget in v3
				vAnnotation = Target(pDocument = self,
				                     pAnnID = vntAnnComps.id,
				                     pOffset = self._setAnnOffset(vntAnnComps.id, vntAnnComps.span)) 
				if vntAnnComps.attributes != '':
					vAnnotation.loadAnnAttr(vntAnnComps.attributes.strip(), self._stopUnknownAnn)
				
				if vAnnotation.offset != (0,0):
					vSent = self._findSentenceMatch(vAnnotation)
					if vSent != None:
						vAnnotation.assignSentence(vSent)
						vSent.target.append(vAnnotation)
					else:
						vlUnmatched.append(vAnnotation)
			elif vntAnnComps.type in ["inside", "sentence"]:                   # inside used in v2 and sentence in v3
				vAnnotation = Inside(pDocument = self,
				                     pAnnID = vntAnnComps.id,
				                     pOffset = self._setAnnOffset(vntAnnComps.id, vntAnnComps.span)) 
				if vntAnnComps.attributes != '':
					vAnnotation.loadAnnAttr(vntAnnComps.attributes.strip(), self._stopUnknownAnn)
				
				if vAnnotation.offset != (0,0):
					vSent = self._findSentenceMatch(vAnnotation)
					if vSent != None:
						vAnnotation.assignSentence(vSent)
						vSent.inside.append(vAnnotation)
					else:
						vlUnmatched.append(vAnnotation)
			elif vntAnnComps.type == "attitude":
				vAnnotation = Attitude(pDocument = self,
				                       pAnnID = vntAnnComps.id,
				                       pOffset = self._setAnnOffset(vntAnnComps.id, vntAnnComps.span)) 
				if vntAnnComps.attributes != '':
					vAnnotation.loadAnnAttr(vntAnnComps.attributes.strip(), self._stopUnknownAnn)
				
				if vAnnotation.offset != (0,0):
					vSent = self._findSentenceMatch(vAnnotation)
					if vSent != None:
						vAnnotation.assignSentence(vSent)
						vSent.attitude.append(vAnnotation)
					else:
						vlUnmatched.append(vAnnotation)
			elif vntAnnComps.type == "target-speech":
				vAnnotation = TargetSpeech(pDocument = self,
				                           pAnnID = vntAnnComps.id,
				                           pOffset = self._setAnnOffset(vntAnnComps.id, vntAnnComps.span)) 
				if vntAnnComps.attributes != '':
					vAnnotation.loadAnnAttr(vntAnnComps.attributes.strip(), self._stopUnknownAnn)
				
				if vAnnotation.offset != (0,0):
					vSent = self._findSentenceMatch(vAnnotation)
					if vSent != None:
						vAnnotation.assignSentence(vSent)
						vSent.targetSpeech.append(vAnnotation)
					else:
						vlUnmatched.append(vAnnotation)
			elif vntAnnComps.type == "eTarget":
				vAnnotation = ETarget(pDocument = self,
				                      pAnnID = vntAnnComps.id,
				                      pOffset = self._setAnnOffset(vntAnnComps.id, vntAnnComps.span)) 
				if vntAnnComps.attributes != '':
					vAnnotation.loadAnnAttr(vntAnnComps.attributes.strip(), self._stopUnknownAnn)
				
				if vAnnotation.offset != (0,0):
					vSent = self._findSentenceMatch(vAnnotation)
					if vSent != None:
						vAnnotation.assignSentence(vSent)
						vSent.eTarget.append(vAnnotation)
					else:
						vlUnmatched.append(vAnnotation)
			elif vntAnnComps.type == "targetFrame":
				vAnnotation = TargetFrame(pDocument = self,
				                          pAnnID = vntAnnComps.id,
				                          pOffset = self._setAnnOffset(vntAnnComps.id, vntAnnComps.span)) 
				if vntAnnComps.attributes != '':
					vAnnotation.loadAnnAttr(vntAnnComps.attributes.strip(), self._stopUnknownAnn)
				
				if vAnnotation.offset != (0,0):
					vSent = self._findSentenceMatch(vAnnotation)
					if vSent != None:
						vAnnotation.assignSentence(vSent)
						vSent.targetFrame.append(vAnnotation)
					else:
						vlUnmatched.append(vAnnotation)
			elif vntAnnComps.type in ["split", "on", "intention-pos", "supplementaryExpressive-subjectivity", "supplementaryAttitude"]:
				continue
			else:
				if self._stopUnknownAnn:
					raise Exception("Unknown annotation type: %s" % vntAnnComps.type)
		
		return vlUnmatched
		
	
	
	def _setAnnOffset(self, pAnnID, pAnnOffsetText):
		'''
		extracts the annotation offset span while fixing some know issues
		'''
		
		if self.filename == "non_fbis/06.25.38-22700":
			if pAnnID == "17":
				vOffset = (2696,2778)
			elif pAnnID == "179":
				vOffset = (811,892)
			elif pAnnID == "10":
				vOffset = (1996,2097)
			elif pAnnID == "67":
				vOffset = (4819,4868)
			else:
				vOffset = tuple([int(o) for o in pAnnOffsetText.split(',')])
		elif self.filename == "20020124/20.36.11-21342":
			if pAnnID == "2":
				vOffset = (503,587)
			else:
				vOffset = tuple([int(o) for o in pAnnOffsetText.split(',')])
		elif self.filename == "20010818/01.09.50-14694":
			if pAnnID == "165":
				vOffset = (1318,1481)
			else:
				vOffset = tuple([int(o) for o in pAnnOffsetText.split(',')])
		elif self.filename == "non_fbis/08.15.19-23507":
			if pAnnID == "72":
				vOffset = (1059,1061)
			else:
				vOffset = tuple([int(o) for o in pAnnOffsetText.split(',')])
		elif self.filename == "non_fbis/12.19.35-18077":
			if pAnnID == "95":
				vOffset = (105,117)
			else:
				vOffset = tuple([int(o) for o in pAnnOffsetText.split(',')])
		elif self.filename == "20020131/20.58.51-26741":
			if pAnnID == "100":
				vOffset = (1987,2002)
			else:
				vOffset = tuple([int(o) for o in pAnnOffsetText.split(',')])
		elif self.filename == "20011218/21.00.22-21144":
			if pAnnID == "268":
				vOffset = (722,744)
			else:
				vOffset = tuple([int(o) for o in pAnnOffsetText.split(',')])
		elif self.filename == "20020320/11.58.59-16532":
			if pAnnID == "80":
				vOffset = (1284,1351)
			else:
				vOffset = tuple([int(o) for o in pAnnOffsetText.split(',')])
		elif self.filename == "non_fbis/09.53.15-23595":
			if pAnnID == "226":
				vOffset = (16,24)
			else:
				vOffset = tuple([int(o) for o in pAnnOffsetText.split(',')])
		elif self.filename == "non_fbis/13.21.23-8227":
			if pAnnID == "1":
				vOffset = (999,1009)
			else:
				vOffset = tuple([int(o) for o in pAnnOffsetText.split(',')])
		elif self.filename == "20020123/21.21.45-6259":
			if pAnnID == "22":
				vOffset = (1132,1218)
			else:
				vOffset = tuple([int(o) for o in pAnnOffsetText.split(',')])
		else:
			vOffset = tuple([int(o) for o in pAnnOffsetText.split(',')])
		
		return vOffset 
		
	
	
	def _findSentenceMatch(self, pAnnotation):
		'''
		Return the sentence the document character span of which matches that of the given annotation
		
		None is returned in case no sentence is found. 
		'''
		
		for vSent in self.sentences:
			if pAnnotation.offset[0] >= vSent.offset[0] and pAnnotation.offset[1] <= vSent.offset[1]:
				return vSent
		
		# at this point, no sentence match has been found; raise an exception or return None depending on use
		return None
		#raise Exception("No sentence matches the annotation boundary of (%s,%s)!" % pAnnotation.offset)
		
	
	
	def tokenize(self, pflgSave = False):
		'''
		Generates a sentence-segmented and tokenized version of the document
		
		Each line in the output contains a tokenized sentence.
		
		If the save flag parameter is set, the output will be saved in a file with the same name as the original document
		file name suffixes with .tok.
		'''
		
		vOutput = ''
		
		for vSent in self.sentences:
			vOutput += vSent.getTokenForm() + '\n'
		
		if pflgSave:
			open("%s/docs/%s.tok" % (self.corpus.database, self.filename), 'w').write(vOutput)
		
		return vOutput
		
	
	
	def loadPOSTagging(self):
		'''
		Loads the POS tagging of the sentences of the document
		
		The POS tag files are supposed to be stored in the pos directory in the same relative path to document file itself
		and with the same file names suffixed with .pos.
		'''
		
		from parse import pos 
		
		vPTL = pos.POSTagLoader(pos.POSTagging)
		vPTL.loadFromColumnar("%s/pos/%s.pos" % (self.corpus.database, self.filename))
		
		# sanity check: the number of sentences in the document and the POS tagging must be equal
		if self.size != len(vPTL.taggedSentences):
			raise Exception("Number of sentences in the document and the POS tagging must be equal: %s vs. %s" % (self.size, len(vPTL.taggedSentences)))
		else:
			for vDocSent, vTaggetSent in zip(self.sentences, vPTL.taggedSentences):
				vDocSent.loadPOSTagging(vTaggetSent)
		
	
	
	def loadConstParses(self, pFileSuffix):
		'''
		Loads the constituency parses of the sentences of the document
		
		The constituency parse files are supposed to be stored in the "parse" directory in the database path under the 
		same directory structure and with the same file names suffixed with pFileSuffix as the documents themselves.
		'''
		
		from parse import constparse
		
		vPL = constparse.ConstParseLoader()
		vPL.loadParses("%s/parse/%s.%s" % (self.corpus.database, self.filename, pFileSuffix))
		
		# sanity check: the number of sentences in the document and parse trees must be equal
		if self.size != len(vPL.parses):
			raise Exception("Number of sentences in the document and parse trees must be equal: %s vs. %s" % (self.size, len(vPL.parses)))
		else:
			for vDocSent, vParse in zip(self.sentences, vPL.parses):
				vDocSent.loadConstTree(vParse.tree)
		
	
	
	@property
	def size(self):
		'''
		Returns the document size as the number of sentences
		'''
		
		return len(self.sentences)
		
	
	
	def getAnnNum(self):
		'''
		Returns the number of annotations in the document
		'''
		
		return sum([s.getAnnNum() for s in self.sentences])
		
	
	
	def getDS(self, pdFilter = {}):
		'''
		Returns the direct subjective annotations of the document
		'''
		
		return [a for s in self.sentences for a in s.getDS(pdFilter)]
		
	
	
	def getES(self, pdFilter = {}):
		'''
		Returns the expressive subjective annotations of the document
		'''
		
		return [a for s in self.sentences for a in s.getES(pdFilter)]
		
	
	
	def getOSE(self, pdFilter = {}):
		'''
		Returns the objective speech event annotations of the document
		'''
		
		return [a for s in self.sentences for a in s.getOSE(pdFilter)]
		
	
	
	def getAttitude(self, pdFilter = {}):
		'''
		Returns the attitude annotations of the document
		'''
		
		return [a for s in self.sentences for a in s.getAttitude(pdFilter)]
		
	
	
	def getSTarget(self):
		'''
		Returns the (span-based) target annotations of the document
		'''
		
		return [a for s in self.sentences for a in s.getSTarget()]
		
	
	
	def getETarget(self):
		'''
		Returns the entity/event target annotations of the document
		'''
		
		return [a for s in self.sentences for a in s.getETarget()]
		
	
	
	def getTF(self):
		'''
		Returns the target frame annotations of the document
		'''
		
		return [a for s in self.sentences for a in s.getTF()]
		
	
	
	def getAgent(self):
		'''
		Returns the agent annotations of the document
		'''
		
		return [a for s in self.sentences for a in s.getAgent()]
		
	
	
	def getTS(self):
		'''
		Returns the target speech annotations of the document
		'''
		
		return [a for s in self.sentences for a in s.getTS()]
		
	
	
	def getInside(self):
		'''
		Returns the inside annotations of the document
		'''
		
		return [a for s in self.sentences for a in s.getInside()]
		
	
	
	def getDSNum(self, pdFilter = {}):
		'''
		Returns the number of direct subjective annotations in the document
		'''
		
		return len(self.getDS(pdFilter))
		
	
	
	def getESNum(self, pdFilter = {}):
		'''
		Returns the number of expressive subjective annotations in the document
		'''
		
		return len(self.getES(pdFilter))
		
	
	
	def getOSENum(self, pdFilter = {}):
		'''
		Returns the number of objective speech event annotations in the document
		'''
		
		return len(self.getOSE(pdFilter))
		
	
	
	def getAttitudeNum(self, pdFilter = {}):
		'''
		Returns the number of attitude annotations in the document
		'''
		
		return len(self.getAttitude(pdFilter))
		
	
	
	def getSTargetNum(self):
		'''
		Returns the number of (span-based) target annotations in the document
		'''
		
		return len(self.getSTarget())
		
	
	
	def getETargetNum(self):
		'''
		Returns the number of entity/event target annotations in the document
		'''
		
		return len(self.getETarget())
		
	
	
	def getTFNum(self):
		'''
		Returns the number of target frame annotations in the document
		'''
		
		return len(self.getTF())
		
	
	
	def getAgentNum(self):
		'''
		Returns the number of agent annotations in the document
		'''
		
		return len(self.getAgent())
		
	
	
	def getTSNum(self):
		'''
		Returns the number of target speech annotations in the document
		'''
		
		return len(self.getTS())
		
	
	
	def getInsideNum(self):
		'''
		Returns the number of inside annotations in the document
		'''
		
		return len(self.getInside())
		
	
	
	def getDSOverlap(self):
		'''
		Returns the overlap between direct subjectives of the sentences in the document
		'''
		
		vlOverlap = []
		
		for vSent in self.sentences:
			vlOverlap += vSent.getDSOverlap()
		
		return vlOverlap
		
	
	
	def getDSWithSTargetOverlap(self):
		'''
		Returns the direct subjectives the span-based targets of which overlap
		'''
		
		vlOverlap = []
		
		for vSent in self.sentences:
			vlOverlap += vSent.getDSWithSTargetOverlap()
		
		return vlOverlap
		
	
	
	def getDSAttOverlap(self):
		'''
		Returns the overlap between direct subjectives and attitudes of the sentences in the document
		'''
		
		vlOverlap = []
		
		for vSent in self.sentences:
			vlOverlap += vSent.getDSAttOverlap()
		
		return vlOverlap
		
	
	
	def getAvgDSLength(self, pdFilter = {}):
		'''
		Returns the average token length of the direct subjective annotations in the document
		'''
		
		vlDS = self.getDS(pdFilter)
		
		if len(vlDS) != 0:
			return sum([a.length for a in vlDS]) * 1.0 / len(vlDS)
		else:
			return 0
		
	
	
	def getAvgESLength(self, pdFilter = {}):
		'''
		Returns the average token length of the expressive subjective annotations in the document
		'''
		
		vlES = self.getES(pdFilter)
		
		if len(vlES) != 0:
			return sum([a.length for a in vlES]) * 1.0 / len(vlES)
		else:
			return 0
		
	
	
	def getAvgAttitudeLength(self, pdFilter = {}):
		'''
		Returns the average token length of the attitude annotations in the document
		'''
		
		vlAttitude = self.getAttitude(pdFilter)
		
		if len(vlAttitude) != 0:
			return sum([a.length for a in vlAttitude]) * 1.0 / len(vlAttitude)
		else:
			return 0
		
	
	
	def getAttOverlap(self):
		'''
		Returns the overlap between attitudes of the sentences in the document
		'''
		
		vlOverlap = []
		
		for vSent in self.sentences:
			vlOverlap += vSent.getAttOverlap()
		
		return vlOverlap
		
	
	
	def getBIOTagging(self, plAnnotationTypes = None, pflgIncludeTokens = True):
		'''
		Returns BIO tagging of the specified annotations in the document sentences
	
		A 2D list is returned, the rows of which represent sentences and the columns sentence tokens, each element of 
		which is a dictionary with one (only "tag") or two keys ("token", "tag") depending on the value of pflgIncludeTokens.
	
		Annotation types to be tagged in the sentence should be specified in a list which the following key words:
		- "DS": direct subjective
		- "ES": expressive subjective
		- TODO: add more annotations
		'''
		
		return [s.getBIOTagging(plAnnotationTypes = plAnnotationTypes, pflgIncludeTokens = pflgIncludeTokens) for s in self.sentences]
		
	
	


class MPQASentence(object):
	'''
	Class for MPQA sentence
	
	In MPQA, a sentence is not a processing unit. Instead, the corpus is organised as a set of documents. However, a part
	of manual annotation is the sentence segmentation, where sentences in each document are identified by the document
	character offset. We exploit this information and create sentence class. For specific annotation types, see the section
	4 of the MPQA corpus readme.
	'''
	
	
	def __init__(self, pDocument, pOffset):
		'''
		Constructor
		'''
		
		self.document = pDocument
		
		# document character offset of the sentence (as extracted) 
		self.offset = pOffset
		
		# original text which is not tokenized
		self._text = re.sub(r"[\s]+", ' ', self.document.text[self.offset[0] : self.offset[1]])
		
		# sentence POS tagging (pos.POSTagging object)
		self.posTagging = None
		
		# sentence constitunecy parse tree (constparse.ConstTree object)
		self.cTree = None
		
		# annotations: each MPQA annotation type has a list containing its instances in the document.
		
		self.directSubjective = []
		self.expressiveSubjective = []
		self.objectiveSpeechEvent = []
		self.attitude = []
		self.target = []
		self.agent = []
		self.targetSpeech = []                        # ?? this is no documented
		self.inside = []
		
		# new in version 3
		self.eTarget = []
		self.targetFrame = []
		
	
	
	def getOrgForm(self):
		'''
		Returns the original untokenized form of the sentence
		'''
		
		return self._text
		
	
	
	def getTokenForm(self):
		'''
		Returns the tokenized form of the sentence
		'''
		
		return nlp.tokenizeSegment(self._text, pLang = "en", pflgTokenizeFSlash = False)
		
	
	
	@property
	def tokens(self):
		'''
		Returns sentence tokens in a list
		'''
		
		return self.getTokenForm().split()
		
	
	
	@property
	def length(self):
		'''
		Returns the number of tokens in the tokenized form of the sentence as its length
		'''
		
		return len(self.tokens)
		
	
	
	def loadPOSTagging(self, pPOSTagging):
		'''
		Loads the POS tagging of the sentences which is in pos.POSTagging format
		
		The input parameter is a POSTagging object which is a pos tagged sentence and is supposed to have the same number 
		of tokens as the sentence itself.
		'''
		
		# sanity check: the number of token in the sentence and the POS tagging must be equal
		if self.length != pPOSTagging.length:
			raise Exception("Number of tokens in the sentence and the POS tagging must be equal: %s vs. %s" % (self.length, pPOSTagging.length))
		else:
			self.posTagging = pPOSTagging
		
	
	
	def loadConstTree(self, pConstTree):
		'''
		Loads the constituency trees of the sentences
		
		The input tree is an constparse.ConstTree object.
		'''
		
		# sanity check: the number of token in the sentence and the length of const tree sentence must match
		if self.length != pConstTree.getSentLength():
			raise Exception("Number of tokens in the sentence and the POS tagging must be equal: %s vs. %s" % (self.length, pConstTree.getSentLength()))
		else:
			self.cTree = pConstTree
		
	
	
	def getAnnNum(self):
		'''
		Returns the number of annotations in the sentence
		'''
		
		return self.getDSNum() + \
			   self.getESNum() + \
			   self.getOSENum() + \
			   self.getAttitudeNum() + \
			   self.getSTargetNum() + \
			   self.getETargetNum() + \
			   self.getTFNum() + \
			   self.getAgentNum() + \
			   self.getTSNum() + \
			   self.getInsideNum()
		
	
	
	def getDS(self, pdFilter = {}):
		'''
		Returns the direct subjective annotations of the sentence
		
		pdFilter is a filter dictionary to be applied to the annotations.  
		'''
		
		vlAnns = self.directSubjective
		
		if "implicit" in pdFilter:
			if pdFilter["implicit"] == True:
				vlAnns = [a for a in vlAnns if a.implicit]
			elif pdFilter["implicit"] == False:
				vlAnns = [a for a in vlAnns if not a.implicit]
		
		if "polarity" in pdFilter:
			if pdFilter["polarity"] == None:
				vlAnns = [a for a in vlAnns if a.polarity is None or a.polarity in ['-', '']]
			else:
				vlAnns = [a for a in vlAnns if a.polarity in pdFilter["polarity"]]
		
		if "with-s-target" in pdFilter:
			if pdFilter["with-s-target"] == True:
				vlAnns = [a for a in vlAnns if len(a.getSTargets()) > 0]
			elif pdFilter["with-s-target"] == False:
				vlAnns = [a for a in vlAnns if len(a.getSTargets()) == 0]
		
		if "with-e-target" in pdFilter:
			if pdFilter["with-e-target"] == True:
				vlAnns = [a for a in vlAnns if len(a.getETargets()) > 0]
			elif pdFilter["with-e-target"] == False:
				vlAnns = [a for a in vlAnns if len(a.getETargets()) == 0]
		
		if "with-attitude" in pdFilter:
			if pdFilter["with-attitude"] == True:
				vlAnns = [a for a in vlAnns if len(a.getAttitudes()) > 0]
			elif pdFilter["with-attitude"] == False:
				vlAnns = [a for a in vlAnns if len(a.getAttitudes()) == 0]
		
		return vlAnns
		
	
	
	def getES(self, pdFilter = {}):
		'''
		Returns the expressive subjective annotations of the sentence
		
		pdFilter is a filter dictionary to be applied to the annotations.  
		'''
		
		vlAnns = self.expressiveSubjective
		
		if "with-s-target" in pdFilter:
			if pdFilter["with-s-target"] == True:
				vlAnns = [a for a in vlAnns if len(a.getSTargets()) > 0]
			elif pdFilter["with-s-target"] == False:
				vlAnns = [a for a in vlAnns if len(a.getSTargets()) == 0]
		
		if "with-e-target" in pdFilter:
			if pdFilter["with-e-target"] == True:
				vlAnns = [a for a in vlAnns if len(a.getETargets()) > 0]
			elif pdFilter["with-e-target"] == False:
				vlAnns = [a for a in vlAnns if len(a.getETargets()) == 0]
		
		return vlAnns
		
	
	
	def getOSE(self, pdFilter = {}):
		'''
		Returns the objective speech event annotations of the sentence
		
		pdFilter is a filter dictionary to be applied to the annotations.  
		'''
		
		vlAnns = self.objectiveSpeechEvent
		
		if "with-s-target" in pdFilter:
			if pdFilter["with-s-target"] == True:
				vlAnns = [a for a in vlAnns if len(a.getSTargets()) > 0]
			elif pdFilter["with-s-target"] == False:
				vlAnns = [a for a in vlAnns if len(a.getSTargets()) == 0]
		
		if "with-e-target" in pdFilter:
			if pdFilter["with-e-target"] == True:
				vlAnns = [a for a in vlAnns if len(a.getETargets()) > 0]
			elif pdFilter["with-e-target"] == False:
				vlAnns = [a for a in vlAnns if len(a.getETargets()) == 0]
		
		if "with-attitude" in pdFilter:
			if pdFilter["with-attitude"] == True:
				vlAnns = [a for a in vlAnns if len(a.getAttitudes()) > 0]
			elif pdFilter["with-attitude"] == False:
				vlAnns = [a for a in vlAnns if len(a.getAttitudes()) == 0]
		
		return vlAnns
		
	
	
	def getAttitude(self, pdFilter = {}):
		'''
		Returns the attitude annotations of the sentence
		
		pdFilter is a filter dictionary to be applied to the annotations.  
		'''
		
		vlAnns = self.attitude
		
		if "with-s-target" in pdFilter:
			if pdFilter["with-s-target"] == True:
				vlAnns = [a for a in vlAnns if len(a.getSTargets()) > 0]
			elif pdFilter["with-s-target"] == False:
				vlAnns = [a for a in vlAnns if len(a.getSTargets()) == 0]
		
		if "with-e-target" in pdFilter:
			if pdFilter["with-e-target"] == True:
				vlAnns = [a for a in vlAnns if len(a.getETargets()) > 0]
			elif pdFilter["with-e-target"] == False:
				vlAnns = [a for a in vlAnns if len(a.getETargets()) == 0]
		
		return vlAnns
		
	
	
	def getSTarget(self):
		'''
		Returns the (span-based) target annotations of the sentence
		'''
		
		return self.sTarget
		
	
	
	def getETarget(self):
		'''
		Returns the entity/event target annotations of the sentence
		'''
		
		return self.eTarget
		
	
	
	def getTF(self):
		'''
		Returns the target frame annotations of the sentence
		'''
		
		return self.targetFrame
		
	
	
	def getAgent(self):
		'''
		Returns the agent annotations of the sentence
		'''
		
		return self.agent
		
	
	
	def getTS(self):
		'''
		Returns the target speech annotations of the sentence
		'''
		
		return self.targetSpeech
		
	
	
	def getInside(self):
		'''
		Returns the inside annotations of the sentence
		'''
		
		return self.inside
		
	
	
	def getDSNum(self, pdFilter = {}):
		'''
		Returns the number of direct subjective annotations in the sentence
		
		pdFilter is a filter dictionary to be applied to the annotations.  
		'''
		
		return len(self.getDS(pdFilter))
		
	
	
	def getESNum(self, pdFilter = {}):
		'''
		Returns the number of expressive subjective annotations in the sentence
		
		pdFilter is a filter dictionary to be applied to the annotations.  
		'''
		
		return len(self.getES(pdFilter))
		
	
	
	def getOSENum(self, pdFilter = {}):
		'''
		Returns the number of objective speech event annotations in the sentence
		
		pdFilter is a filter dictionary to be applied to the annotations.  
		'''
		
		return len(self.getOSE(pdFilter))
		
	
	
	def getAttitudeNum(self, pdFilter = {}):
		'''
		Returns the number of attitude annotations in the sentence
		
		pdFilter is a filter dictionary to be applied to the annotations.  
		'''
		
		return len(self.getAttitude(pdFilter))
		
	
	
	def getSTargetNum(self):
		'''
		Returns the number of (span-based) target annotations in the sentence
		'''
		
		return len(self.getSTarget())
		
	
	
	def getETargetNum(self):
		'''
		Returns the number of entity/event target annotations in the sentence
		'''
		
		return len(self.getETarget())
		
	
	
	def getTFNum(self):
		'''
		Returns the number of target frame annotations in the sentence
		'''
		
		return len(self.getTF())
		
	
	
	def getAgentNum(self):
		'''
		Returns the number of agent annotations in the sentence
		'''
		
		return len(self.getAgent())
		
	
	
	def getTSNum(self):
		'''
		Returns the number of target speech annotations in the sentence
		'''
		
		return len(self.getTS())
		
	
	
	def getInsideNum(self):
		'''
		Returns the number of inside annotations in the sentence
		'''
		
		return len(self.getInside())
		
	
	
	@property 
	def sTarget(self):
		'''
		Aliases Target annotations as sTarget for version 3
		'''
		
		return self.target
		
	
	
	@property 
	def sentence(self):
		'''
		Aliases Inside annotations as Sentence for version 3
		'''
		
		return self.inside
		
	
	
	def findAttitude(self, pAID):
		'''
		Finds and returns the attitude annotation with the given attitude ID
		
		Note that attitude ID is different from annotation ID assigned to every single annotation.
		'''
		
		for vAttitude in self.attitude:
			if vAttitude.id == pAID:
				return vAttitude
		
		return None
		
	
	
	def findSTarget(self, pTID):
		'''
		Finds and returns span-based target annotation with the given target ID
		
		In both version 2 and 3, span-based target annotation is stored in Target object. 
		
		Note that target ID is different from annotation ID assigned to every single annotation.
		'''
		
		# Note: target speech is not documented so its relation to other versions of target is not clear. 
		
		for vTarget in self.target:
			if vTarget.id == pTID:
				return vTarget
		
		return None
		
	
	
	def findSTargets(self, pTFID):
		'''
		Finds and returns span-based target annotations linked with the given target frame ID (version 3 only)
		
		A target frame may link to more than one sTarget. 
		
		Note that target ID is different from annotation ID assigned to every single annotation.
		'''
		
		# Note: target speech is not documented so its relation to other versions of target is not clear. 
		
		for vTF in self.targetFrame:
			if vTF.id == pTFID:
				return vTF.getSTargets()
		
		return []
		
	
	
	def findETarget(self, pTID):
		'''
		Finds and returns entity/event target annotation with the given target ID (version 3 only)
		
		Note that target ID is different from annotation ID assigned to every single annotation.
		'''
		
		for vTarget in self.eTarget:
			if vTarget.id == pTID:
				return vTarget
		
		return None
		
	
	
	def findETargets(self, pTFID):
		'''
		Finds and returns entity/event target annotations linked with the given target frame ID (version 3 only)
		
		A target frame may link to more than one eTarget. 
		
		Note that target ID is different from annotation ID assigned to every single annotation.
		'''
		
		for vTF in self.targetFrame:
			if vTF.id == pTFID:
				return vTF.getETargets()
		
		return []
		
	
	
	def findTargetSpeech(self, pTSID):
		'''
		Finds and returns the target speech annotation with the given ID
		
		Note that target speech ID is different from annotation ID assigned to every single annotation.
		'''
		
		for vTS in self.targetSpeech:
			if vTS.id == pTSID:
				return vTS
		
		return None
		
	
	
	def getDSOverlap(self):
		'''
		Returns the overlap between direct subjectives of the sentence
		'''
		
		vlOverlaps = []
		
		for vDS1 in self.directSubjective:
			for vDS2 in self.directSubjective:
				if vDS1 != vDS2:
					if util.compareRanges(vDS1.offset, vDS2.offset) != 0:
						vlOverlaps.append((vDS1, vDS2))
		
		return vlOverlaps
		
	
	
	def getDSWithSTargetOverlap(self):
		'''
		Returns the direct subjectives the span-based targets of which overlap
		'''
		
		vlOverlaps = []
		
		for vDS in self.directSubjective:
			vlOT = vDS.getOverlapSTargets()
			
			if len(vlOT) > 0:
				vlOverlaps.append((vDS, vlOT))
		
		return vlOverlaps
		
	
	
	def getDSAttOverlap(self):
		'''
		Returns the overlap between direct subjectives and attitudes of the sentence
		'''
		
		vlOverlaps = []
		
		for vDS in self.directSubjective:
			for vAtt in self.attitude:
				if util.compareRanges(vDS.offset, vAtt.offset) != 0:
					vlOverlaps.append((vDS, vAtt))
		
		return vlOverlaps
		
	
	
	def getAvgDSLength(self, pdFilter = {}):
		'''
		Returns the average token length of the direct subjective annotations in the sentence
		'''
		
		vlDS = self.getDS(pdFilter)
		
		if len(vlDS) != 0:
			return sum([a.length for a in vlDS]) * 1.0 / len(vlDS)
		else:
			return 0
		
	
	
	def getAvgESLength(self, pdFilter = {}):
		'''
		Returns the average token length of the expressive subjective annotations in the sentence
		'''
		
		vlES = self.getES(pdFilter)
		
		if len(vlES) != 0:
			return sum([a.length for a in vlES]) * 1.0 / len(vlES)
		else:
			return 0
		
	
	
	def getAttOverlap(self):
		'''
		Returns the overlap between attitudes of the sentence
		'''
		
		vlOverlaps = []
		
		for vAtt1 in self.attitude:
			for vAtt2 in self.attitude:
				if vAtt1 != vAtt2:
					if util.compareRanges(vAtt1.offset, vAtt2.offset) != 0:
						vlOverlaps.append((vAtt1, vAtt2))
		
		return vlOverlaps
		
	
	
	def getAvgAttitudeLength(self, pdFilter = {}):
		'''
		Returns the average token length of the attitude annotations in the sentence
		'''
		
		vlAttitude = self.getAttitude(pdFilter)
		
		if len(vlAttitude) != 0:
			return sum([a.length for a in vlAttitude]) * 1.0 / len(vlAttitude)
		else:
			return 0
		
	
	
	def getBIOTagging(self, plAnnotationTypes = None, pflgIncludeTokens = True):
		'''
		Returns BIO tagging of the specified annotations in the sentence
		
		A list is returned representing sentence tokens, each element of which is a dictionary with one (only "tag") or 
		two keys ("token", "tag") depending on the value of pflgIncludeTokens.
		
		Annotation types to be tagged in the sentence should be specified in a list which the following key words:
		- "DS": direct subjective
		- "ES": expressive subjective
		- TODO: add more annotations
		'''
		
		# extracting the annotations
		
		vlAnn = []
		
		for ann in plAnnotationTypes:
			if ann == "DS":
				vlAnn += self.getDS()
			elif ann == "ES":
				vlAnn += self.getES()
		
		# tagging
		
		vlResult = []
		
		for i in range(1, self.length + 1):
			if pflgIncludeTokens:
				vlResult.append({"token": self.tokens[i - 1]})
			else:
				vlResult.append({})
			
			if i in [a.tokenSpan[0] for a in vlAnn]:
				vlResult[-1]["tag"] = 'B'
			elif i in [idx for a in vlAnn for idx in range(a.tokenSpan[0] + 1, a.tokenSpan[1] + 1)]:
				vlResult[-1]["tag"] = 'I'
			else:
				vlResult[-1]["tag"] = 'O'
		
		return vlResult
		
	
	


class MPQAAnn(object):
	'''
	Abstract class for MPQA annotation
	
	An Annotation object is the smallest unit of annotation used in MPQA associated with an ID, offset, annotation type
	and annotations themselves. This class abstracts over all annotation types (8 in version 2). Each annotation type 
	then has it own class and annotation details.
	
	NOTE: there is an annotation type Split which is used to mark sentence segmentation. This is ignored here. There is
	also some annotations which are not mentioned in the documentation but found in the data. All these annotations are
	ignored except "target-speech" which is addressed here.
	'''
	
	
	def __init__(self, pDocument, pAnnID, pOffset):
		'''
		Constructor
		'''
		
		self.document = pDocument
		
		self.annID = pAnnID
		self.offset = pOffset
		
		# NOTE: both document and sentence are maintained so that an annotation can belong to a document without being
		# associated with any sentence
		self.sentence = None
		self.tokenSpan = None              # unlike python list span: start from 1 and ends in the last token's position
		                                   ## Note: token span (0,0) is used for implicit annotations which are those 
		                                   ## whose character offset points to the beginning of the sentence.
		
	
	
	def __str__(self):
		'''
		String representation of the annotation
		'''
		
		return self.getOrgForm()
		
	
	
	def assignSentence(self, pSentence):
		'''
		Assigns the sentence to which the annotation belongs and sets its token span in the sentence
		'''
		
		self.sentence = pSentence
		self.tokenSpan = self._docOffsetToTokenSpan()
		
	
	
	def _docOffsetToTokenSpan(self):
		'''
		Maps the original document character offset of the annotation to sentence token span
		'''
		
		## The main problem of mapping is the tokenization as the original 
		## offset values are based on untokenized version of the text.
		## The problem can be alleviated by starting the search for the 
		## term in the (probably tokenized) sentence from the offset start,
		## because the tokenization tends to expand the text, so the term
		## will be pushed farther in the sentence. However, there can be 
		## a problem because our tokenizer trims trailing or surrounding
		## spaces, so in rare cases term can be pushed ahead of the original
		## offset. Therefore, we start N characters early but a sanity 
		## check should be performed, when using this function.
		
		
		def _calcSearchStart(pStepBack):
			'''
			Calculates the position in sentence from which the search for annotation starts
			
			pStepBack specifies how many characters back the search should start from to account for possible tokenization
			and cleaning problems.
			'''
			
			if self.offset[0] - self.sentence.offset[0] > pStepBack:
				return self.offset[0] - self.sentence.offset[0] - pStepBack
			else:
				return 0
			
		
		## handling cases where the annotation offset does not point to a token
		## These are in general cases in which the offset points to the beginning of the sentence to represent implicit
		## expressions, i.e. annotations which do not have a texual realization inside the sentence.
		if self.offset[0] == self.offset[1]:
			return (0, 0)
		
		vAnnTokenForm =  nlp.tokenizeSegment(self.getOrgForm(), pLang = "en", pflgTokenizeFSlash = False)
		
		# handling known tokenization issues: e.g. when the annotation offset points to 't in cann't
		vAnnTokenForm = re.sub(r"^' (t|s|m|re)", r"'\1", vAnnTokenForm)
		
		# default step back characters
		vStepBack = 5
		
		# locating the annotation in the sentence (character position)
		while True:
			# translating document character offset to sentence character offset while taking the cautionary step back
			vSearchStart = _calcSearchStart(vStepBack)
			
			vPosition = self.sentence.getTokenForm().find(vAnnTokenForm, vSearchStart)
			
			# repeating the search by increasing the step back until it is found or the search starts at sentence start
			if vPosition == -1 and vSearchStart > 0:
				vStepBack += 5
			else:
				break
		
		vSpanStart = 1
		
		# calculating the token span if the annotation is found and legitimate
		if vPosition == -1:
			print "Annotation: %s %s cannot be located in: %s" % (vAnnTokenForm, self.offset, self.sentence.getTokenForm())
		elif vPosition > self.offset[0] + len(self.sentence.getTokenForm()[:vPosition].split()):
			print "Annotation: %s %s is located at %s in: %s, but the position seems suspicious!" % (vAnnTokenForm, self.offset, vPosition, self.sentence.getTokenForm())
		else:
			# The number of spaces before the position is where the span starts.
			for i in range(vPosition):
				if self.sentence.getTokenForm()[i] == ' ':
					vSpanStart += 1
		
		return (vSpanStart, vSpanStart + len(vAnnTokenForm.split()) - 1)
		
	
	
	def getOrgForm(self):
		'''
		Returns the original untokenized form of the annotation text
		'''
		
		return re.sub(r"[\s]+", ' ', self.document.text[self.offset[0] : self.offset[1]])
		
	
	
	def getTokenForm(self):
		'''
		Returns the tokenized form of the sentence
		'''
		
		if self.tokenSpan == (0,0):
			return ''
		else:
			return ' '.join(self.sentence.tokens[self.tokenSpan[0] - 1: self.tokenSpan[1]])
		
	
	
	@property
	def tokens(self):
		'''
		Returns sentence tokens in a list
		'''
		
		return self.getTokenForm().split()
		
	
	
	@property
	def length(self):
		'''
		Returns the number of tokens the annotation spans as its length
		'''
		
		return len(self.tokens)
		
	
	


class DirectSubjective(MPQAAnn):
	'''
	Class for MPQA direct subjectives
	'''
	
	
	def __init__(self, pDocument, pAnnID, pOffset):
		'''
		Constructors
		'''
		
		MPQAAnn.__init__(self, pDocument, pAnnID, pOffset)
		        
		# annotation attributes
		self.nestedSource = None
		self.intensity = None                      # low, medium, high, extreme
		self.exprIntensity = None                  # low, medium, high, extreme
		self.polarity = None                       # neutral, positive, negative, both, neutral, uncertain-positive, 
		                                           ## uncertain-negative, uncertain-both, uncertain-neutral
		self.implicit = None                       # presence-based attribute (i.e. not None means implicit)
		self.insubstantial = None                  # c1, c2, c3
		self.attitudeLink = None                   
		self.annUncertain = None                   # somewhat-uncertain, very-uncertain
		self.subjUncertain = None                  # somewhat-uncertain, very-uncertain
		self.targetSpeechLink = None               # not given in the documentation
		
		# version 3 additions
		self.id = None
		self.attitudeType = None                   # agree-neg, agree-pos, arguing-neg, arguing-pos, both, intention-neg, 
		                                           # intention-pos, interior-sentiment-neg, interior-sentiment-pos, 
		                                           # negative, other, other-attitude, positive, sentiment-neg, 
		                                           # sentiment-pos, speculation
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		if self.implicit:
			# more of a sanity check
			if self.getOrgForm() != '':
				vForm = "[IMPLICIT] Warning: span is not empty!"
			else:
				vForm = "[IMPLICIT]"
		else:
			vForm = self.getOrgForm()
		
		return "Direct Subjective           : %s\n" % vForm + \
		       "  Polarity                  : %s\n" % ('-' if self.polarity is None else self.polarity) + \
		       "  Intensity                 : %s\n" % ('-' if self.intensity is None else self.intensity) + \
		       "  Expression Intensity      : %s\n" % ('-' if self.exprIntensity is None else self.exprIntensity) + \
		       "  Insubstantial             : %s\n" % ('-' if self.insubstantial is None else self.insubstantial) + \
		       "  Attitude                  : %s\n" % self._extractAttitudes() + \
		       "  Attitude Type (v3)        : %s\n" % ('N/A' if self.document.version != 3 else self.attitudeType) + \
		       "  Nested Source             : %s\n" % self.nestedSource + \
		       "  Span-based targets (v2)   : %s\n" % self._extractSTargets() + \
		       "  Event/Entity targets (v3) : %s\n" % ('N/A' if self.document.version != 3 else self._extractETargets()) + \
		       "  Target Speech             : %s\n" % ('-' if self.targetSpeechLink is None else self._extractTargetSpeech()) + \
		       "  Annotation Uncertain      : %s\n" % ('-' if self.annUncertain is None else self.annUncertain) + \
		       "  Subjective Uncertain      : %s\n" % ('-' if self.subjUncertain is None else self.subjUncertain) + \
		       "  Tokenized Form            : %s\n" % self.getTokenForm()
		
	
	
	def loadAnnAttr(self, pAnnAttrSegment, pflgStopUnknownAnn = True):
		'''
		Loads annotation attributes from the text segment extracted from MPQA annotation
		'''
		
		if pAnnAttrSegment.strip() == '':
			return
		#print pAnnAttrSegment
		
		# rectifying some known annotation issues before hand (e.g. when space is not only used as attribute delimiter) 
		vAnnAttrSegment = re.sub(r'(?<=[^"])[ ]+', '', pAnnAttrSegment)
		
		vlAttrsSplit = vAnnAttrSegment.split()
		
		for vAttr in vlAttrsSplit:
			vAttrSplit = vAttr.split('=')
			
			if vAttrSplit[0] == "nested-source":
				self.nestedSource = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.nestedSource
			elif vAttrSplit[0] == "intensity":
				self.intensity = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.intensity
			elif vAttrSplit[0] == "expression-intensity":
				self.exprIntensity = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.exprIntensity
			elif vAttrSplit[0] == "polarity":
				self.polarity = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.polarity
			elif vAttrSplit[0] in ["implicit", "is-implicit"]:           # implicit in v2 and is-implicit in v3
				self.implicit = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.implicit
			elif vAttrSplit[0] == "insubstantial":
				self.insubstantial = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.insubstantial
			elif vAttrSplit[0] == "attitude-link":
				self.attitudeLink = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.attitudeLink
			elif vAttrSplit[0] == "annotation-uncertain":
				self.annUncertain = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.annUncertain
			elif vAttrSplit[0] == "subjective-uncertain":
				self.subjUncertain = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.subjUncertain
			elif vAttrSplit[0] == "target-speech-link":
				self.targetSpeechLink = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.targetSpeechLink
			elif vAttrSplit[0] == "id":
				self.id = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.id
			elif vAttrSplit[0] == "attitude-type":
				self.attitudeType = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.attitudeType
			elif vAttrSplit[0] in ["attitude-uncertain", "objective-uncertain", "attitude-towards", "tone", "attitude-toward"]:       # ignores as infrequent
				continue
			else:
				if pflgStopUnknownAnn:
					raise Exception("Unknown attribute: %s" % vAttrSplit[0])
		
	
	
	def getAttitudes(self):
		'''
		Returns the attitude (Attitude objects) annotations linked to this annotation
		'''
		
		if self.attitudeLink is None or self.attitudeLink == '':
			return []
		
		vlAttitudes = []
		
		for vLink in self.attitudeLink.split(','):
			vAttitude = self.sentence.findAttitude(pAID = vLink)
			if vAttitude is None:
				vlAttitudes.append(None)
			else:
				vlAttitudes.append(vAttitude)
		
		return vlAttitudes
		
	
	
	def _extractAttitudes(self):
		'''
		Extracts the surface form of the attitude linked to this annotation
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getAttitudes()])
		
	
	
	def getSTargets(self):
		'''
		Returns the span-based target annotations (Target objects) linked to this annotation (via attitudes)
		'''
		
		vlSTargets = []
		
		for vAttitude in self.getAttitudes():
			if vAttitude is not None:
				vlSTargets += vAttitude.getSTargets()
		
		return vlSTargets
		
	
	
	def _extractSTargets(self):
		'''
		Extracts the surface form of the span-based targets linked to this annotation via (attitudes)
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getSTargets()])
		
	
	
	def getETargets(self):
		'''
		Returns the eTarget annotations (ETarget objects) linked to this annotation (via attitudes)
		'''
		
		vlETargets = []
		
		for vAttitude in self.getAttitudes():
			if vAttitude is not None:
				vlETargets += vAttitude.getETargets()
		
		return vlETargets
		
	
	
	def _extractETargets(self):
		'''
		Extracts the surface form of the eTargets linked to this annotation via (attitudes)
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getETargets()])
		
	
	
	def getTargetSpeech(self):
		'''
		Returns the target speech (TargetSpeech objects) annotations linked to this annotation
		
		Target speeches are not documented but can be found in the corpus (for just a few documents). They are linked to
		direct subjectives and objective speech events.
		'''
		
		if self.targetSpeechLink is None or self.targetSpeechLink == '':
			return []
		
		vlTargetSpeeches = []
		
		for vLink in self.targetSpeechLink.split(','):
			vTargetSpeech = self.sentence.findTargetSpeech(pTSID = vLink)
			if vTargetSpeech is None:
				vlTargetSpeeches.append(None)
			else:
				vlTargetSpeeches.append(vTargetSpeech)
		
		return vlTargetSpeeches
		
	
	
	def _extractTargetSpeech(self):
		'''
		Extracts the surface form of the target speeches linked to this annotation via (attitudes)
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getTargetSpeech()])
		
	
	
	def getOverlapSTargets(self):
		'''
		Returns the overlapping span-based targets of the direct subjective 
		'''
		
		vlOverlaps = []
		vlTargets = self.getSTargets()
		
		for vTarget1 in vlTargets:
			if vTarget1 is None:
				continue
			for vTarget2 in vlTargets:
				if vTarget2 is None or vTarget1 == vTarget2:
					continue
				if util.compareRanges(vTarget1.offset, vTarget2.offset) != 0:
					vlOverlaps.append((vTarget1, vTarget2))
		
		return vlOverlaps
		
	
	
	def getSTargetTokenCoverage(self):
		'''
		Returns the token indexes (1-based) of all span-based targets of the annotation merged in a sorted list
		'''
		
		vlTokenIdxs = []
		
		for vTarget in self.getSTargets():
			if vTarget is not None:
				vlTokenIdxs += range(vTarget.tokenSpan[0], vTarget.tokenSpan[1] + 1)
		
		return sorted(vlTokenIdxs)
		
	
	


class ExpressiveSubjective(MPQAAnn):
	'''
	Class for MPQA expressive subjectives
	'''
	
	
	def __init__(self, pDocument, pAnnID, pOffset):
		'''
		Constructors
		'''
		
		MPQAAnn.__init__(self, pDocument, pAnnID, pOffset)
		
		# annotation attributes
		self.nestedSource = None
		self.intensity = None                      # low, medium, high, extreme
		self.polarity = None                       ## positive, negative, both, neutral, uncertain-positive, 
		                                           ## uncertain-negative, uncertain-both, uncertain-neutral
		self.esUncertain = None                    # annotation uncertain: somewhat-uncertain, very-uncertain
		self.nsUncertain = None                    # nested source uncertain: somewhat-uncertain, very-uncertain
		
		# version 3 additions
		self.id = None                             # also existed in version 2 but ignored as infrequent or optional
		self.targetFrameLink = None                # mentioned as targetFrame in their README
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		vForm = self.getOrgForm()
		
		return "Expressive Subjective       : %s\n" % vForm + \
		       "  Polarity                  : %s\n" % ('-' if self.polarity is None else self.polarity) + \
		       "  Intensity                 : %s\n" % ('-' if self.intensity is None else self.intensity) + \
		       "  Nested Source             : %s\n" % self.nestedSource + \
		       "  Span-based targets (v2)   : %s\n" % self._extractSTargets() + \
		       "  Event/Entity targets (v3) : %s\n" % ('N/A' if self.document.version != 3 else self._extractETargets()) + \
		       "  Annotation Uncertain      : %s\n" % ('-' if self.esUncertain is None else self.esUncertain) + \
		       "  Nested Source  Uncertain  : %s\n" % ('-' if self.nsUncertain is None else self.nsUncertain) + \
		       "  Tokenized Form            : %s\n" % self.getTokenForm()
		
	
	
	def loadAnnAttr(self, pAnnAttrSegment, pflgStopUnknownAnn = True):
		'''
		Loads annotation attributes from the text segment extracted from MPQA annotation
		'''
		
		if pAnnAttrSegment.strip() == '':
			return
		# print pAnnAttrSegment
		
		# rectifying some known annotation issues before hand (e.g. when space is not only used as attribute delimiter) 
		vAnnAttrSegment = re.sub(r'(?<=[^"])[ ]+', '', pAnnAttrSegment)
		vAnnAttrSegment = vAnnAttrSegment.replace("nexted-source", "nested-source")
		
		vlAttrsSplit = vAnnAttrSegment.split()
		
		for vAttr in vlAttrsSplit:
			vAttrSplit = vAttr.split('=')
			
			if vAttrSplit[0] == "nested-source":
				self.nestedSource = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.nestedSource
			elif vAttrSplit[0] in ["intensity", "strength"]:                                # strength seems to be a typo
				self.intensity = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.intensity
			elif vAttrSplit[0] == "polarity":
				self.polarity = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.polarity
			elif vAttrSplit[0] == "es-uncertain":
				self.esUncertain = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.esUncertain
			elif vAttrSplit[0] == "nested-source-uncertain":
				self.nsUncertain = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.nsUncertain
			elif vAttrSplit[0] == "targetFrame-link":                                       # version 3
				self.targetFrameLink = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.targetFrameLink
			elif vAttrSplit[0] == "id":                                                     # version 3
				self.id = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.id
			elif vAttrSplit[0] in ["ese-type", "attitude-type", "sarcasm", "disagreement", "diagreement", "uncertain"] or \
			     vAttrSplit[0].endswith("-eTarget-link") or \
			     vAttrSplit[0].endswith("-eTarget-lnk") or \
				 vAttrSplit[0].endswith("-isReferredInSpan"):     # ignores as infrequent or wrong
				continue
			else:
				if pflgStopUnknownAnn:
					raise Exception("Unknown attribute: %s" % vAttrSplit[0])
		
	
	
	def getSTargets(self):
		'''
		Returns the span-based targets (Target objects) linked to this annotation via target frame link (version 3)
		
		Note: in version 2, expressive subjectives are not linked to any target.
		'''
		
		if self.targetFrameLink is None or self.targetFrameLink == '':
			return []
		
		vlTargets = []
		
		for vLink in self.targetFrameLink.split(','):
			vlTargets += self.sentence.findSTargets(pTFID = vLink)
		
		return vlTargets
		
	
	
	def _extractSTargets(self):
		'''
		Extracts the surface form of the span-based targets linked to this annotation via target frame link (version 3)
		
		Note: in version 2, expressive subjectives are not linked to any target.
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getSTargets()])
		
	
	
	def getETargets(self):
		'''
		Returns the eTargets linked to this annotation via target frame link (version 3)
		
		Note: in version 2, expressive subjectives are not linked to any target.
		'''
		
		if self.targetFrameLink is None or self.targetFrameLink == '':
			return []
		
		vlTargets = []
		
		for vLink in self.targetFrameLink.split(','):
			vlTargets += self.sentence.findETargets(pTFID = vLink)
		
		return vlTargets
		
	
	
	def _extractETargets(self):
		'''
		Extracts the surface form of the eTargets linked to this annotation via target frame link (version 3)
		
		Note: in version 2, expressive subjectives are not linked to any target.
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getETargets()])
		
	
	


class ObjSpeechEvent(MPQAAnn):
	'''
	Class for MPQA objective speech events
	'''
	
	
	def __init__(self, pDocument, pAnnID, pOffset):
		'''
		Constructors
		'''
		
		MPQAAnn.__init__(self, pDocument, pAnnID, pOffset)
		
		# annotation attributes
		self.nestedSource = None
		self.implicit = None                       # presence-based attribute (i.e. not None means implicit)
		self.insubstantial = None                  # c1, c2, c3
		self.annUncertain = None                   # somewhat-uncertain, very-uncertain
		self.objUncertain = None                   # somewhat-uncertain, very-uncertain
		self.targetSpeechLink = None               # not given in the documentation
		self.attitudeLink = None                   # not given in the documentation
		self.overallStrength = None                # not given in the documentation
		self.onStrength = None                     # not given in the documentation
		
		# version 3 additions
		self.id = None                             # also existed in version 2 but ignored as infrequent or optional
		self.targetFrameLink = None
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		if self.implicit:
			# more of a sanity check
			if self.getOrgForm() != '':
				vForm = "[IMPLICIT] Warning: span is not empty!"
			else:
				vForm = "[IMPLICIT]"
		else:
			vForm = self.getOrgForm()
		
		return "Objective Speech Event      : %s\n" % vForm + \
		       "  Insubstantial             : %s\n" % ('-' if self.insubstantial is None else self.insubstantial) + \
		       "  Attitude                  : %s\n" % self._extractAttitudes() + \
		       "  Nested Source             : %s\n" % self.nestedSource + \
		       "  Span-based targets (v2)   : %s\n" % self._extractSTargets() + \
		       "  Event/Entity targets (v3) : %s\n" % ('N/A' if self.document.version != 3 else self._extractETargets()) + \
		       "  Target Speech             : %s\n" % ('-' if self.targetSpeechLink is None else self._extractTargetSpeech()) + \
		       "  Overall Strength          : %s\n" % self.overallStrength + \
		       "  On Strength               : %s\n" % self.onStrength + \
		       "  Annotation Uncertain      : %s\n" % ('-' if self.annUncertain is None else self.annUncertain) + \
		       "  Objective Uncertain       : %s\n" % ('-' if self.objUncertain is None else self.objUncertain) + \
		       "  Tokenized Form            : %s\n" % self.getTokenForm()
		
	
	
	def loadAnnAttr(self, pAnnAttrSegment, pflgStopUnknownAnn = True):
		'''
		Loads annotation attributes from the text segment extracted from MPQA annotation
		'''
		
		if pAnnAttrSegment.strip() == '':
			return
		#print pAnnAttrSegment
		
		# rectifying some known annotation issues before hand (e.g. when space is not only used as attribute delimiter) 
		vAnnAttrSegment = re.sub(r'(?<=[^"])[ ]+', '', pAnnAttrSegment)
		
		vlAttrsSplit = vAnnAttrSegment.split()
		
		for vAttr in vlAttrsSplit:
			vAttrSplit = vAttr.split('=')
			
			if vAttrSplit[0] == "nested-source":
				self.nestedSource = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.nestedSource
			elif vAttrSplit[0] == "implicit":
				self.implicit = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.implicit
			elif vAttrSplit[0] == "insubstantial":
				self.insubstantial = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.insubstantial
			elif vAttrSplit[0] == "annotation-uncertain":
				self.annUncertain = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.annUncertain
			elif vAttrSplit[0] == "objective-uncertain":
				self.subjUncertain = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.subjUncertain
			elif vAttrSplit[0] == "target-speech-link":
				self.targetSpeechLink = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.targetSpeechLink
			elif vAttrSplit[0] == "attitude-link":
				self.attitudeLink = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.attitudeLink
			elif vAttrSplit[0] == "overall-strength":
				self.overallStrength = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.overallStrength
			elif vAttrSplit[0] == "on-strength":
				self.onStrength = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.onStrength
			elif vAttrSplit[0] == "targetFrame-link":                                   # version 3
				self.targetFrameLink = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.targetFrameLink
			elif vAttrSplit[0] == "id":                                                 # version 3
				self.id = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.id
			elif vAttrSplit[0] in ["polarity", "id", "expression-intensity"]:           # ignores as infrequent or wrong
				continue
			else:
				if pflgStopUnknownAnn:
					raise Exception("Unknown attribute: %s" % vAttrSplit[0])
		
	
	
	def getAttitudes(self):
		'''
		Returns the attitude (Attitude objects) annotations linked to this annotation
		'''
		
		if self.attitudeLink is None or self.attitudeLink == '':
			return []
		
		vlAttitudes = []
		
		for vLink in self.attitudeLink.split(','):
			vAttitude = self.sentence.findAttitude(pAID = vLink)
			if vAttitude is None:
				vlAttitudes.append(None)
			else:
				vlAttitudes.append(vAttitude)
		
		return vlAttitudes
		
	
	
	def _extractAttitudes(self):
		'''
		Extracts the surface form of the attitude linked to this annotation
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getAttitudes()])
		
	
	
	def getSTargets(self):
		'''
		Returns the span-based target annotations (Target objects) linked to this annotation (via attitudes)
		'''
		
		vlSTargets = []
		
		for vAttitude in self.getAttitudes():
			if vAttitude is not None:
				vlSTargets += vAttitude.getSTargets()
		
		return vlSTargets
		
	
	
	def _extractSTargets(self):
		'''
		Extracts the surface form of the span-based targets linked to this annotation via (attitudes)
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getSTargets()])
		
	
	
	def getETargets(self):
		'''
		Returns the eTarget annotations (ETarget objects) linked to this annotation (via attitudes)
		'''
		
		vlETargets = []
		
		for vAttitude in self.getAttitudes():
			if vAttitude is not None:
				vlETargets += vAttitude.getETargets()
		
		return vlETargets
		
	
	
	def _extractETargets(self):
		'''
		Extracts the surface form of the eTargets linked to this annotation via (attitudes)
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getETargets()])
		
	
	
	def getTargetSpeech(self):
		'''
		Returns the target speech (TargetSpeech objects) annotations linked to this annotation
		
		Target speeches are not documented but can be found in the corpus (for just a few documents). They are linked to
		direct subjectives and objective speech events.
		'''
		
		if self.targetSpeechLink is None or self.targetSpeechLink == '':
			return []
		
		vlTargetSpeeches = []
		
		for vLink in self.targetSpeechLink.split(','):
			vTargetSpeech = self.sentence.findTargetSpeech(pTSID = vLink)
			if vTargetSpeech is None:
				vlTargetSpeeches.append(None)
			else:
				vlTargetSpeeches.append(vTargetSpeech)
		
		return vlTargetSpeeches
		
	
	
	def _extractTargetSpeech(self):
		'''
		Extracts the surface form of the target speeches linked to this annotation via (attitudes)
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getTargetSpeech()])
		
	
	


class Attitude(MPQAAnn):
	'''
	Class for MPQA attitude annotation
	'''
	
	
	def __init__(self, pDocument, pAnnID, pOffset):
		'''
		Constructors
		'''
		
		MPQAAnn.__init__(self, pDocument, pAnnID, pOffset)
		
		# annotation attributes
		self.id = None
		self.attitudeType = None                     ##	positive sentiment, negative sentiment, positive arguing
		                                             ## negative arguing, positive agreement, negative agreement (disagreement)
		                                             ## positive intention, negative intention, speculation, other-attitude
		self.attitudeUncertain = None                # somewhat-uncertain, very-uncertain
		self.targetLink = None                       # version 2 only
		self.inferred = None                         # seems to be presence-based
		self.intensity = None                        # low, low-medium, medium, medium-high, high, high-extreme, extreme
		self.repetition = None                       # seems to be presence-based
		self.contrast = None                         # seems to be presence-based
		self.sarcastic = None                        # seems to be presence-based
		
		# version 3 additions
		self.targetFrameLink = None                  # mentioned as targetFrame in their README
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		vForm = self.getOrgForm()
		
		return "Attitude                    : %s\n" % vForm + \
		       "  Attitude Type             : %s\n" % ('-' if self.attitudeType is None else self.attitudeType) + \
		       "  Span-based targets (v2)   : %s\n" % self._extractSTargets() + \
		       "  Event/Entity targets (v3) : %s\n" % ('N/A' if self.document.version != 3 else self._extractETargets()) + \
		       "  Inferred                  : %s\n" % ('-' if self.inferred is None else self.inferred) + \
		       "  Intensity                 : %s\n" % ('-' if self.intensity is None else self.intensity) + \
		       "  Repetition                : %s\n" % ('-' if self.repetition is None else self.repetition) + \
		       "  Contrast                  : %s\n" % ('-' if self.contrast is None else self.contrast) + \
		       "  Sarcastic                 : %s\n" % ('-' if self.sarcastic is None else self.sarcastic) + \
		       "  Attitude Uncertain        : %s\n" % ('-' if self.attitudeUncertain is None else self.attitudeUncertain) + \
		       "  Tokenized Form            : %s\n" % self.getTokenForm()
		
	
	
	def loadAnnAttr(self, pAnnAttrSegment, pflgStopUnknownAnn = True):
		'''
		Loads annotation attributes from the text segment extracted from MPQA annotation
		'''
		
		if pAnnAttrSegment.strip() == '':
			return
		#print pAnnAttrSegment
		
		# rectifying some known annotation issues before hand (e.g. when space is not only used as attribute delimiter) 
		vAnnAttrSegment = re.sub(r'(?<=[^"])[ ]+', '', pAnnAttrSegment)
		vAnnAttrSegment = vAnnAttrSegment.replace(" t59", "t59")
		
		vlAttrsSplit = vAnnAttrSegment.split()
		
		for vAttr in vlAttrsSplit:
			vAttrSplit = vAttr.split('=')
			
			if vAttrSplit[0] == "id":
				self.id = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.id
			elif vAttrSplit[0] == "attitude-type":
				self.attitudeType = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.attitudeType
			elif vAttrSplit[0] in ["attitude-uncertain", "annotation-uncertain"]:     # found to be interchangeably used
				self.attitudeUncertain = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.attitudeUncertain
			elif vAttrSplit[0] == "target-link":
				self.targetLink = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.targetLink
			elif vAttrSplit[0] == "inferred":
				self.inferred = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.inferred
			elif vAttrSplit[0] == "intensity":
				self.intensity = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.intensity
			elif vAttrSplit[0] == "repetition":
				self.repetition = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.repetition
			elif vAttrSplit[0] == "contrast":
				self.contrast = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.contrast
			elif vAttrSplit[0] == "sarcastic":
				self.contrast = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.sarcastic
			elif vAttrSplit[0] == "targetFrame-link":                                       # version 3
				self.targetFrameLink = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.targetFrameLink
			elif vAttrSplit[0] in ["negative-toward", "positive-toward", "negative-intensity", "positive-intensity", 
			                       "arguing-about", "arguing-intensity", "intention-pos", "intention-ps", "intention-pos-intensity",
			                       "attitude-neg-toward", "attitude-neg-intensity", "attitude-pos-toward", "attitude-pos-intensity",
			                       "arguing-pos-about", "arguing-pos-intensity", "attitude-link", "comment", "disagreement",
								   "diagreement", "2-targetSpan"]:
				continue
			else:
				if pflgStopUnknownAnn:
					raise Exception("Unknown attribute: %s" % vAttrSplit[0])
		
	
	
	def getSTargets(self):
		'''
		Returns the span-based targets (Target objects) linked to this annotation
		
		Note: in version 2, the targets are directly linked to attitudes, while in version 3, they are linked via target 
		frame link.
		'''
		
		vlTargets = []
		
		if self.document.version == 2:
			if self.targetLink is not None and self.targetLink != '':
				for vLink in self.targetLink.split(','):
					vlTargets.append(self.sentence.findSTarget(pTID = vLink))
		elif self.document.version == 3:
			if self.targetFrameLink is not None and self.targetFrameLink != '':
				for vLink in self.targetFrameLink.split(','):
					vlTargets += self.sentence.findSTargets(pTFID = vLink)
		
		return vlTargets
		
	
	
	def _extractSTargets(self):
		'''
		Extracts the surface form of the span-based targets linked to this annotation
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getSTargets()])
		
	
	
	def getETargets(self):
		'''
		Returns the eTargets linked to this annotation via target frame link (version 3)
		'''
		
		if self.targetFrameLink is None or self.targetFrameLink == '':
			return []
		
		vlTargets = []
		
		for vLink in self.targetFrameLink.split(','):
			vlTargets += self.sentence.findETargets(pTFID = vLink)
		
		return vlTargets
		
	
	
	def _extractETargets(self):
		'''
		Extracts the surface form of the eTargets linked to this annotation via target frame link (version 3)
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getETargets()])
		
	
	
	def getSTargetTokenCoverage(self):
		'''
		Returns the token indexes (1-based) of all span-based targets of the annotation merged in a sorted list
		'''
		
		vlTokenIdxs = []
		
		for vTarget in self.getSTargets():
			if vTarget is not None:
				vlTokenIdxs += range(vTarget.tokenSpan[0], vTarget.tokenSpan[1] + 1)
		
		return sorted(vlTokenIdxs)
		
	
	


class TargetFrame(MPQAAnn):
	'''
	Class for MPQA v3 target frame annotation
	
	According to the README, the targetFrame annotations are automatically created to organize the eTarget annotations in
	the version 3 and the span-based target annotations in version 2 in the same framework.
	'''
	
	
	def __init__(self, pDocument, pAnnID, pOffset):
		'''
		Constructors
		'''
		
		MPQAAnn.__init__(self, pDocument, pAnnID, pOffset)
		
		# annotation attributes
		self.id = None
		self.sTargetLink = None
		self.newETargetLink = None
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		vForm = self.getOrgForm()
		
		return "Target Frame  (v3)          : %s\n" % vForm + \
		       "  Span-based targets        : %s\n" % ('-' if self.sTargetLink is None else self._extractSTargets()) + \
		       "  Event/Entity targets      : %s\n" % ('-' if self.newETargetLink is None else self._extractETargets()) + \
		       "  Tokenized Form            : %s\n" % self.getTokenForm()
		
	
	
	def loadAnnAttr(self, pAnnAttrSegment, pflgStopUnknownAnn = True):
		'''
		Loads annotation attributes from the text segment extracted from MPQA annotation
		'''
		
		if pAnnAttrSegment.strip() == '':
			return
		#print '\n', pAnnAttrSegment
		
		# rectifying some known annotation issues before hand (e.g. when space is not only used as attribute delimiter) 
		vAnnAttrSegment = re.sub(r'(?<=[^"])[ ]+', '', pAnnAttrSegment)
		
		vlAttrsSplit = vAnnAttrSegment.split()
		
		for vAttr in vlAttrsSplit:
			vAttrSplit = vAttr.split('=')
			
			if vAttrSplit[0] == "id":
				self.id = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.id
			elif vAttrSplit[0] == "sTarget-link":
				self.sTargetLink = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.sTargetLink
			elif vAttrSplit[0] == "newETarget-link":
				self.newETargetLink = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.newETargetLink
			elif vAttrSplit[0] in ["target-link", "isReferredInSpan"] or vAttrSplit[0].endswith("eTarget-link"):       # infrequent or wrong
				continue
			else:
				if pflgStopUnknownAnn:
					raise Exception("Unknown attribute: %s" % vAttrSplit[0])
		
	
	
	def getSTargets(self):
		'''
		Returns the span-based target annotations linked with this target frame
		'''
		
		if self.sTargetLink is None or self.sTargetLink == '':
			return []
		
		vlSTargets = []
		
		for vLink in self.sTargetLink.split(','):
			vTarget = self.sentence.findSTarget(pTID = vLink)
			if vTarget is None:
				vlSTargets.append(None)
			else:
				vlSTargets.append(vTarget)
		
		return vlSTargets
		
	
	
	def _extractSTargets(self):
		'''
		Extracts the surface form of the span-based target annotations linked with this target frame
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getSTargets()])
		
	
	
	def getETargets(self):
		'''
		Returns the eTarget annotations linked with this target frame
		'''
		
		if self.newETargetLink is None or self.newETargetLink == '':
			return []
		
		vlETargets = []
		
		for vLink in self.newETargetLink.split(','):
			vTarget = self.sentence.findETarget(pTID = vLink)
			if vTarget is None:
				vlETargets.append(None)
			else:
				vlETargets.append(vTarget)
		
		return vlETargets
		
	
	
	def _extractETargets(self):
		'''
		Extracts the surface form of the eTarget annotations linked with this target frame
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getETargets()])
		
	
	


class Target(MPQAAnn):
	'''
	Class for MPQA targets (sTarget in version 3)
	'''
	
	
	def __init__(self, pDocument, pAnnID, pOffset):
		'''
		Constructors
		'''
		
		MPQAAnn.__init__(self, pDocument, pAnnID, pOffset)
		
		# annotation attributes
		self.id = None
		self.targetUncertain = None                   # somewhat-uncertain, very-uncertain
		
		# version 3 additions
		self.eTargetLink = None
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		vForm = self.getOrgForm()
		
		return "Span-based Target           : %s\n" % vForm + \
		       "  Event/Entity targets (v3) : %s\n" % ('N/A' if self.document.version != 3 else self._extractETargets()) + \
		       "  Target Uncertain          : %s\n" % ('-' if self.targetUncertain is None else self.targetUncertain) + \
		       "  Tokenized Form            : %s\n" % self.getTokenForm()
		
	
	
	def loadAnnAttr(self, pAnnAttrSegment, pflgStopUnknownAnn = True):
		'''
		Loads annotation attributes from the text segment extracted from MPQA annotation
		'''
		
		if pAnnAttrSegment.strip() == '':
			return
		#print '\n', pAnnAttrSegment
		
		# rectifying some known annotation issues before hand (e.g. when space is not only used as attribute delimiter) 
		vAnnAttrSegment = re.sub(r'(?<=[^"])[ ]+', '', pAnnAttrSegment)
		
		vlAttrsSplit = vAnnAttrSegment.split()
		
		for vAttr in vlAttrsSplit:
			vAttrSplit = vAttr.split('=')
			
			if vAttrSplit[0] == "id":
				self.id = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.id
			elif vAttrSplit[0] == "target-uncertain":
				self.targetUncertain = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.targetUncertain
			elif vAttrSplit[0] == "eTarget-link":
				self.eTargetLink = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.eTargetLink
			elif vAttrSplit[0] in ["attitude-link", "annotation-uncertain"]:                       # infrequent or wrong
				continue
			else:
				if pflgStopUnknownAnn:
					raise Exception("Unknown attribute: %s" % vAttrSplit[0])
		
	
	
	def getETargets(self):
		'''
		Returns the eTargets linked to this annotation (version 3)
		'''
		
		if self.eTargetLink is None or self.eTargetLink == '':
			return []
		
		vlTargets = []
		
		for vLink in self.eTargetLink.split(','):
			vlTargets += self.sentence.findETargets(pTFID = vLink)
		
		return vlTargets
		
	
	
	def _extractETargets(self):
		'''
		Extracts the surface form of the eTargets linked to this annotation (version 3)
		'''
		
		return ', '.join(["<%s>" % a.getOrgForm() if a is not None else "None" for a in self.getETargets()])
		
	
	


class ETarget(MPQAAnn):
	'''
	Class for MPQA v3 eTarget annotation
	'''
	
	
	def __init__(self, pDocument, pAnnID, pOffset):
		'''
		Constructors
		'''
		
		MPQAAnn.__init__(self, pDocument, pAnnID, pOffset)
		
		# annotation attributes
		self.id = None
		self.type = None                                   # values: entity, event, other
		self.isNegated = None                              # values: yes, no
		self.isReferredInSpan = None                       # values: yes, no
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		vForm = self.getOrgForm()
		
		return "Event/Entity Target         : %s\n" % vForm + \
		       "  Target Type               : %s\n" % ('-' if self.type is None else self.type) + \
		       "  Negated                   : %s\n" % ('-' if self.isNegated is None else self.isNegated) + \
		       "  Referred in Span          : %s\n" % ('-' if self.isReferredInSpan is None else self.isReferredInSpan) + \
		       "  Tokenized Form            : %s\n" % self.getTokenForm()
		
	
	
	def loadAnnAttr(self, pAnnAttrSegment, pflgStopUnknownAnn = True):
		'''
		Loads annotation attributes from the text segment extracted from MPQA annotation
		'''
		
		if pAnnAttrSegment.strip() == '':
			return
		#print '\n', pAnnAttrSegment
		
		# rectifying some known annotation issues before hand (e.g. when space is not only used as attribute delimiter) 
		vAnnAttrSegment = re.sub(r'(?<=[^"])[ ]+', '', pAnnAttrSegment)
		
		vlAttrsSplit = vAnnAttrSegment.split()
		
		for vAttr in vlAttrsSplit:
			vAttrSplit = vAttr.split('=')
			
			if vAttrSplit[0] == "id":
				self.id = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.id
			elif vAttrSplit[0] == "type":
				self.type = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.type
			elif vAttrSplit[0] == "isNegated":
				self.isNegated = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.isNegated
			elif vAttrSplit[0] == "isReferredInSpan":
				self.isReferredInSpan = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.isReferredInSpan
			elif vAttrSplit[0] in ["isTheNonNegatedEvent"]:       # infrequent or wrong
				continue
			else:
				if pflgStopUnknownAnn:
					raise Exception("Unknown attribute: %s" % vAttrSplit[0])
		
	
	


class Agent(MPQAAnn):
	'''
	Class for MPQA agents
	'''
	
	
	def __init__(self, pDocument, pAnnID, pOffset):
		'''
		Constructors
		'''
		
		MPQAAnn.__init__(self, pDocument, pAnnID, pOffset)
		
		# annotation attributes
		self.id = None
		self.nestedSource = None
		self.agentUncertain = None                   # somewhat-uncertain, very-uncertain
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		vForm = self.getOrgForm()
		
		return "Agent                       : %s\n" % vForm + \
		       "  Nested Source             : %s\n" % self.nestedSource + \
		       "  Agent Uncertain           : %s\n" % ('-' if self.agentUncertain is None else self.agentUncertain) + \
		       "  Tokenized Form            : %s\n" % self.getTokenForm()
		
	
	
	def loadAnnAttr(self, pAnnAttrSegment, pflgStopUnknownAnn = True):
		'''
		Loads annotation attributes from the text segment extracted from MPQA annotation
		'''
		
		if pAnnAttrSegment.strip() == '':
			return
		#print pAnnAttrSegment
		
		# rectifying some known annotation issues before hand (e.g. when space is not only used as attribute delimiter) 
		vAnnAttrSegment = re.sub(r'(?<=[^"])[ ]+', '', pAnnAttrSegment)
		
		vlAttrsSplit = vAnnAttrSegment.split()
		
		for vAttr in vlAttrsSplit:
			vAttrSplit = vAttr.split('=')
			
			if vAttrSplit[0] == "id":
				self.id = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.id
			elif vAttrSplit[0] == "nested-source":
				self.nestedSource = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.nestedSource
			elif vAttrSplit[0] == "agent-uncertain":
				self.agentUncertain = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.agentUncertain
			elif vAttrSplit[0] in ["expression-intensity", "attitude-link", "polarity", "target-uncertain"]:                  # infrequent or wrong
				continue
			else:
				if pflgStopUnknownAnn:
					raise Exception("Unknown attribute: %s" % vAttrSplit[0])
		
	
	


class TargetSpeech(MPQAAnn):
	'''
	Class for MPQA target speech annotation
	'''
	
	
	def __init__(self, pDocument, pAnnID, pOffset):
		'''
		Constructors
		'''
		
		MPQAAnn.__init__(self, pDocument, pAnnID, pOffset)
		
		# annotation attributes
		self.id = None
		self.targetUncertain = None                   # somewhat-uncertain, very-uncertain
		
	
	
	def __str__(self):
		'''
		String representation of the class
		'''
		
		vForm = self.getOrgForm()
		
		return "Target Speech               : %s\n" % vForm + \
		       "  Target Uncertain          : %s\n" % ('-' if self.targetUncertain is None else self.targetUncertain) + \
		       "  Tokenized Form            : %s\n" % self.getTokenForm()
		
	
	
	def loadAnnAttr(self, pAnnAttrSegment, pflgStopUnknownAnn = True):
		'''
		Loads annotation attributes from the text segment extracted from MPQA annotation
		'''
		
		if pAnnAttrSegment.strip() == '':
			return
		#print '\n', pAnnAttrSegment
		
		# rectifying some known annotation issues before hand (e.g. when space is not only used as attribute delimiter) 
		vAnnAttrSegment = re.sub(r'(?<=[^"])[ ]+', '', pAnnAttrSegment)
		
		vlAttrsSplit = vAnnAttrSegment.split()
		
		for vAttr in vlAttrsSplit:
			vAttrSplit = vAttr.split('=')
			
			if vAttrSplit[0] == "id":
				self.id = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.id
			elif vAttrSplit[0] == "target-uncertain":
				self.targetUncertain = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.targetUncertain
			else:
				if pflgStopUnknownAnn:
					raise Exception("Unknown attribute: %s" % vAttrSplit[0])
		
	
	


class Inside(MPQAAnn):
	'''
	Class for MPQA inside annotation in version 2 and sentence annotation in version 3
	
	Inside is automatically created and seems to be equal to a sentence.
	'''
	
	
	def __init__(self, pDocument, pAnnID, pOffset):
		'''
		Constructors
		'''
		
		MPQAAnn.__init__(self, pDocument, pAnnID, pOffset)
		
		# annotation attributes
		self.nestedSource = None
		self.insideUncertain = None                   # somewhat-uncertain, very-uncertain
		
	
	
	def loadAnnAttr(self, pAnnAttrSegment, pflgStopUnknownAnn = True):
		'''
		Loads annotation attributes from the text segment extracted from MPQA annotation
		'''
		
		if pAnnAttrSegment.strip() == '':
			return
		#print pAnnAttrSegment
		
		# rectifying some known annotation issues before hand (e.g. when space is not only used as attribute delimiter) 
		vAnnAttrSegment = re.sub(r'(?<=[^"])[ ]+', '', pAnnAttrSegment)
		
		vlAttrsSplit = vAnnAttrSegment.split()
		
		for vAttr in vlAttrsSplit:
			vAttrSplit = vAttr.split('=')
			
			if vAttrSplit[0] == "nested-source":
				self.nestedSource = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.nestedSource
			elif vAttrSplit[0] == "inside-uncertain":
				self.insideUncertain = vAttrSplit[1][1 : -1]
				#print vAttrSplit[0], self.insideUncertain
			elif vAttrSplit[0] in ["comment"]:                                            # infrequent or wrong
				continue
			else:
				if pflgStopUnknownAnn:
					raise Exception("Unknown attribute: %s" % vAttrSplit[0])
		
	
	


