#! /usr/bin/python
# -*- coding: utf-8 -*-

"""	
	This module defines classes for SemEval task on Aspect Based Sentiment
	Analysis (ABSA).
	
	
	Version 1.2													30-Apr-2018
	getReviewTokens() is added to ABSA2016Review.
	
	Version 1.1													17-Jul-2017 to 18-Jul-2017
	- SemEval 2016 ABSA support is added. 
	
	Version 1.0													13-Oct-2016
	- Previous versions of loader have had "etc." wrongly tokenized (for vague
	  reasons). A flag parameter is added to the loaders to tokenized it if
	  needed to stay compatible with previous versions of data.
	
	Version 0.9													21-Apr-2016
	- A new attribute is added to ABSA2014Sent to store the ABSA2014Set it 
	  belongs to. Its value is set during loading the data.
	
	Version 0.8													03-Sep-2015
	- Language is added to data loading.
	
	Version 0.7													19-Jun-2015
	- ABSA2014Sent.loadParsedSent() and ABSA2014ATerm.loadParsedATerm() 
	  are updated with the new changes in the tokenizer.
	
	Version 0.7													08-May-2015
	- selectSentences() and getIDs() are added to ABSASet.
	- ABSA2014ATerm.loadParsedATerm() sets the new sentence to the aspect 
	  term.
	
	Version 0.6													16-Feb-2015
	- ABSA2014Loader is renamed to ABSA2014Set.
	- Sort feature is added to ABSA2014.load().
	
	Version 0.5													12-Feb-2015
	- ABSA2014 is renamed to ABSA2014Loader.
	- ABSA2014Sent and ABSA2014Term are now inherited from sa.absa.ABSASent
	  and sa.absa.AspectTerm respectively.
	
	Version 0.4													10-Feb-2015
	- _offsetToSpan() and getTokenSpan() are added to ABSA2014ATerm.
	- getTerm() in ABSA2014ATerm is renamed to getForm()
	- The sentence and term forms are optionally stored as UTF8.
	
	Version 0.3													30-Jan-2015
	- The data including both text and aspect terms can optionally be 
	  tokenized when loading to ABSA2014.
	- extractATStat() and getAspectTerms() are added to ABSA2014.
	- getSentenceForms() is renamed to extractSentenceForms() in ABSA2014. 
	- getPolarity() is added to ABSA2014ATerm.
	
	Version 0.2													21-Jan-2015
	- extractAspectTerms(), extractATCountGroups() are added to ABSA2014.
	- getAspectTerms(), getAspectTermForms(), getAspectTermCount() are 
	  added to ABSA2014Sent.
	- getTerm() is added to ABSA2014ATerm.
	
	Version 0.1													20-Jan-2015
	- ABSA2014, ABSA2014Sent, ABSA2014ATerm are added.
	
"""

import xml.etree.ElementTree as xmlet
import random
from nlp import nlp
from sa import absa
from utils import util


class ABSA2014Set(absa.ABSASet):
	'''
	Class for SemEval 2014 ABSA data set
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		absa.ABSASet.__init__(self)
		
	
	
	def load(self, pXMLFilename, pflgTokenize = False, pflgUTF8 = False, pSort = '', pLanguage = "en", pflgFixETC = False):
		'''
		Loads the data from XML file
		
		Since the original data is not tokenized, both text and aspect 
		terms can optionally be tokenized during loading.
		
		The sort options are:
		- None: in document order
		- id:   in sentence ID order
		- text: in sentence text order
		
		pflgFixETC is to whether or not fix the "etc." tokenization issue caused by the older version. If true, "etc."
		will be tokenized for backward compatibility with previous version of data. 
		'''
		
		
		try:
			vXMLTree = xmlet.parse(pXMLFilename)
		except IOError:
			raise Exception("Cannot open the XML file in %s" % pXMLFilename)
		
		vXMLRoot = vXMLTree.getroot()
		
		# SemEval 2014 has no context (i.e. review) defined in the data format. We consider every sentence belonging to 
		# one and only one context (review). 
		self.contexts = []
		
		for vSent in vXMLRoot.iter("sentence"):
			vABSA2014Review = ABSA2014Review(pABSA2014Set = self)
			
			vABSA2014Sent = ABSA2014Sent(pABSAContext = vABSA2014Review)
			vABSA2014Sent.loadParsedSent(vSent, pflgTokenize = pflgTokenize, pflgUTF8 = pflgUTF8, pLanguage = pLanguage, pflgFixETC = pflgFixETC)
			vABSA2014Review.addSentence(vABSA2014Sent)
			
			# context gets the sentence ID
			vABSA2014Review.id = vABSA2014Sent.getID()
			self.addContext(vABSA2014Review)
		
		if pSort.lower() == "id":
			# The sentence IDs are in different formats in different data sets.
			if ':' in self.contexts[0].getID():
				self.contexts.sort(key = lambda x: [int(i) for i in x.getID().split(':')])
			elif '#' in self.contexts[0].getID():
				self.contexts.sort(key = lambda x: [int(i) for i in x.getID().split('#')])
			else:
				self.contexts.sort(key = lambda x: int(x.getID()))
		elif pSort.lower() == "text":
			self.contexts.sort(key = lambda x: x.getText())
		
	
	
	def getSentences(self, pSort = ''):
		'''
		Returns the ABSA 2014 sentences
		
		The sort options are:
		- None: in document order
		- id:   in sentence ID order
		- text: in sentence text order
		'''
		
		vlSentences = [s for c in self.getContexts() for s in c.getSentences()]
		
		if pSort.lower() == "id":
			# The sentence IDs are in different formats in different data sets.
			if ':' in vlSentences[0].getID():
				return [s for s in sorted(vlSentences, key = lambda x: [int(i) for i in x.getID().split(':')])]
			elif '#' in vlSentences[0].getID():
				return [s for s in sorted(vlSentences, key = lambda x: [int(i) for i in x.getID().split('#')])]
			else:
				return [s for s in sorted(vlSentences, key = lambda x: int(x.getID()))]
		elif pSort.lower() == "text":
			return [s for s in sorted(vlSentences, key = lambda x: x.getText())]
		else:
			return vlSentences
		
	
	
	def getIDs(self):
		'''
		Returns all sentence IDs in a list
		'''
		
		return [s.getID() for s in self.getSentences(pSort = "id")]
		
	
	
	def extractATCountGroups(self):
		'''
		Extract and returns sentence groups based on the number of aspect
		terms in them
		'''
		
		vdATCountGroups = {}
		
		for vSent in self.getSentences():
			vATCount = vSent.getAspectTermCount()
			if vATCount in vdATCountGroups:
				vdATCountGroups[vATCount].append(vSent.getID())
			else:
				vdATCountGroups[vATCount] = [vSent.getID()]
		
		return vdATCountGroups
		
	
	
	def subSet(self, pMethod = "random", pSize = 0, pIDs = '', pIDDelim = ',', pflgComplement = False):
		'''
		Selects a subset of sentences from the data based on the specified
		method and returns them in a set.
		
		pMethod is:
		- random: pSize sentences are selected randomly
		- list: selection is done based on a list of sentence IDs
		
		If pflgComplement is set to true and the method is "list", the 
		complement of specified indexes is selected.
		'''
		
		# preparing IDs of selected sentences
		if pMethod.lower() in ["r", "rand", "random"]:
			vlSelIDs = random.sample(self.getIDs(), pSize)
		elif pMethod.lower() in ["l", "list"]:
			vlSelIDs = pIDs.split(pIDDelim)
		
		# extracting selected sentences: a bit expensive a search
		vlSelSentences = []
		for vSent in self.getSentences(pSort = "id"):
			if (not pflgComplement and vSent.getID() in vlSelIDs) or (pflgComplement and vSent.getID() not in vlSelIDs):
				vlSelSentences.append(vSent)
		
		# returning selection as a set
		vNewSet = ABSA2014Set()
		vNewSet.addSentences(vlSelSentences)
		return vNewSet
		
	
	


class ABSA2014Review(absa.ABSAContext):
	'''
	Class to substitute review for ABSA 2014
	
	SemEval 2014 has no context (i.e. review) defined in the data format. We consider every sentence belonging to one 
	and only one context (review). 
	'''
	
	
	def __init__(self, pABSA2014Set):
		'''
		Constructor
		'''
		
		absa.ABSAContext.__init__(self, pABSA2014Set)
		
		self.id = None
		
	
	
	def getID(self):
		'''
		Returns the id of the review
		'''
		
		return self.id
		
	
	


class ABSA2014Sent(absa.ABSASent):
	'''
	Class for SemEval 2014 ABSA sentence
	'''
	
	
	def __init__(self, pABSAContext):
		'''
		Constructor
		'''
		
		absa.ABSASent.__init__(self, pABSAContext)
		
		self.id = None
		
	
	
	def loadParsedSent(self, pXMLSent, pflgTokenize = False, pflgUTF8 = False, pLanguage = "en", pflgFixETC = False):
		'''
		Loads the sentence content from parsed XML sentence
		
		Since the original data is not tokenized, both text and aspect 
		terms can optionally be tokenized during loading.
		'''
		
		self.id = pXMLSent.get("id")
		
		if pflgUTF8:
			self.text = pXMLSent.findtext("text").strip().encode("UTF-8")
		else:
			self.text = pXMLSent.findtext("text").strip()
		
		if pflgTokenize:
			self.text = nlp.tokenizeSegment(self.text, pLang = pLanguage, pflgTokenizeFSlash = False)
			# fixing "etc." tokenization issue
			if pflgFixETC:
				self.text = self.text.replace(" etc.", " etc .")
		
		for vATerm in pXMLSent.iter("aspectTerm"):
			vABSA2014ATerm = ABSA2014ATerm()
			vABSA2014ATerm.loadParsedATerm(vATerm, pABSA2014Sent = self, pflgTokenize = pflgTokenize, pflgUTF8 = pflgUTF8, pLanguage = pLanguage, pflgFixETC = pflgFixETC)
			self.aspectTerms.append(vABSA2014ATerm)
		
	
	
	def getID(self):
		'''
		Returns sentence ID
		'''
		
		return self.id
		
	
	


class ABSA2014ATerm(absa.AspectTerm):
	'''
	Class for SemEval 2014 ABSA aspect term
	'''
	
	
	def __init__(self):
		'''
		Constructor
		'''
		
		absa.AspectTerm.__init__(self)
		
		## original from and to (0-base character offset)
		## Caution: do not match text if it's tokenized during load.
		self._from = None                 
		self._to = None
		
	
	
	def loadParsedATerm(self, pXMLATerm, pABSA2014Sent, pflgTokenize = False, pflgUTF8 = False, pLanguage = "en", pflgFixETC = False):
		'''
		Loads the aspect term details from parsed XML aspect term
		
		Since the original data is not tokenized, both text and aspect 
		terms can optionally be tokenized during loading.
		'''
		
		
		if pflgUTF8:
			self.term = pXMLATerm.get("term").encode("UTF-8")
		else:
			self.term = pXMLATerm.get("term")
		
		if pflgTokenize:
			self.term = nlp.tokenizeSegment(self.term, pLang = pLanguage, pflgTokenizeFSlash = False)
			
			# fixing "etc." tokenization issue
			if pflgFixETC:
				self.term = self.term.replace(" etc.", " etc .")
		
		self.sentence = pABSA2014Sent
		
		self.polarity = pXMLATerm.get("polarity")
		self._from = int(pXMLATerm.get("from"))
		self._to = int(pXMLATerm.get("to"))
		
		# Mapping original character offset to token span
		self.span = self._offsetToSpan(pABSA2014Sent.getText(), self.getForm(), self._from, self._to)
		
	
	
	def _offsetToSpan(self, pSentenceForm, pTermForm, pCOffsetFrom, pCOffsetTo):
		'''
		Maps the original character offset to token span
		'''
		
		## The main problem of mapping is the tokenization as the original 
		## offset values are based on untokenized version of the sentence.
		## The problem can be alleviated by starting the search for the 
		## term in the (probably tokenized) sentence from the offset start,
		## because the tokenization tends to expand the text, so the term
		## will be pushed farther in the sentence. However, there can be 
		## a problem because our tokenizer trims trailing or surrounding
		## spaces, so in rare cases term can be pushed ahead of the original
		## offset. Therefore, we start N characters early but a sanity 
		## check should be performed, when using this function.
		
		if pCOffsetFrom > 5:
			vSearchStart = pCOffsetFrom - 5
		else:
			vSearchStart = 0
		
		vPosition = pSentenceForm.find(pTermForm, vSearchStart)
		
		vSpanStart = 1
		
		if vPosition == -1:
			print "Term %s (%s, %s) can not be located in: %s" % (pTermForm, pCOffsetFrom, pCOffsetTo, pSentenceForm)
		elif vPosition > pCOffsetFrom +  len(pSentenceForm[:vPosition].split()):
			print "Term %s (%s, %s) is located at %s in: %s, but the position seems suspicious!" % (pTermForm, pCOffsetFrom, pCOffsetTo, vPosition, pSentenceForm)
		else:
			# The number of spaces before the position is where the span starts.
			
			for i in range(vPosition):
				if pSentenceForm[i] == ' ':
					vSpanStart += 1
		
		return (vSpanStart, vSpanStart + len(pTermForm.split()) - 1)
	
	


#_______________________________________________________________________________________________________________________



class ABSA2016Set(absa.ABSASet):
	'''
	Class for SemEval 2016 ABSA data set
	'''
	
	def __init__(self):
		'''
		Constructor
		'''
		
		absa.ABSASet.__init__(self)
		
	
	
	def load(self, pXMLFilename, pflgTokenize=False, pflgUTF8=False, pSort='', pLanguage="en", pflgFixETC=False):
		'''
		Loads the data from XML file
		
		Since the original data is not tokenized, both text and aspect 
		terms can optionally be tokenized during loading.
		
		NOTE: only restaurant data set has aspect term annotation. The laptop data set only has aspect categories.
		
		The sort options are:
		- None: in document order
		- id:   in review/sentence ID order
		- text: in sentence text order
		
		pflgFixETC is to whether or not fix the "etc." tokenization issue caused by the older version. If true, "etc."
		will be tokenized for backward compatibility with previous version of data. 
		'''
		
		try:
			vXMLTree = xmlet.parse(pXMLFilename)
		except IOError:
			raise Exception("Cannot open the XML file in %s" % pXMLFilename)
		
		vXMLRoot = vXMLTree.getroot()
		
		self.reviews = []
		
		for vReview in vXMLRoot.iter("Review"):
			vABSA2016Review = ABSA2016Review(pABSA2016Set=self)
			vABSA2016Review.loadParsedReview(vReview, pflgTokenize=pflgTokenize, pflgUTF8=pflgUTF8, pSort=pSort, pLanguage=pLanguage, pflgFixETC=pflgFixETC)
			self.contexts.append(vABSA2016Review)
		
		if pSort.lower() == "id":
			self.contexts.sort(key=lambda x: x.getID())
		
	
	
	def getReviews(self, pSort=''):
		'''
		Returns the ABSA 2016 reviews
		
		The sort options are:
		- None: in document order
		- id:   in sentence ID order
		'''
		
		vlReviews = self.getContexts()
		
		if pSort.lower() == "id":
			return [r for r in sorted(vlReviews, key=lambda x: x.getID())]
		else:
			return vlReviews
		
	
	
	def getSentences(self, pSort=''):
		'''
		Returns the ABSA 2016 sentences
		
		The sort options are:
		- None: in document order
		- id:   in sentence ID order
		- text: in sentence text order
		'''
		
		vlSentences = [s for c in self.getContexts() for s in c.getSentences()]
		
		if pSort.lower() == "id":
			if ':' in vlSentences[0].getID():
				return [s for s in sorted(vlSentences, key=lambda x: x.getID().split(':')[1])]
		elif pSort.lower() == "text":
			return [s for s in sorted(vlSentences, key=lambda x: x.getText())]
		else:
			return vlSentences
		
	
	
	def getReviewIDs(self):
		'''
		Returns all review IDs in a list
		'''
		
		return [r.getID() for r in self.getReviews(pSort="id")]
		
	
	
	def getSentenceIDs(self):
		'''
		Returns all sentence IDs in a list
		'''
		
		return [s.getID() for s in self.getSentences(pSort="id")]
		
	
	
	def extractATCountGroups(self):
		'''
		Extract and returns sentence groups based on the number of aspect
		terms in them
		'''
		
		vdATCountGroups = {}
		
		for vSent in self.getSentences():
			vATCount = vSent.getAspectTermCount()
			if vATCount in vdATCountGroups:
				vdATCountGroups[vATCount].append(vSent.getID())
			else:
				vdATCountGroups[vATCount] = [vSent.getID()]
		
		return vdATCountGroups
		
	
	
	def subSet(self, pMethod="random", pSize=0, pIDs='', pIDDelim=',', pflgComplement=False):
		'''
		Selects a subset of reviews from the data based on the specified method and returns them in a set.
		
		pMethod is:
		- random: pSize reviews are selected randomly
		- list: selection is done based on a list of review IDs
		
		If pflgComplement is set to true and the method is "list", the complement of specified indexes is selected.
		'''
		
		# preparing IDs of selected reviews
		if pMethod.lower() in ["r", "rand", "random"]:
			vlSelIDs = random.sample(self.getReviewIDs(), pSize)
		elif pMethod.lower() in ["l", "list"]:
			vlSelIDs = pIDs.split(pIDDelim)
		
		# extracting selected reviews: a bit expensive a search
		vlSelReviews = []
		for vRev in self.getReviews(pSort="id"):
			if (not pflgComplement and vRev.getID() in vlSelIDs) or (pflgComplement and vRev.getID() not in vlSelIDs):
				vlSelReviews.append(vRev)
		
		# returning selection as a set
		vNewSet = ABSA2016Set()
		vNewSet.addContexts(vlSelReviews)
		return vNewSet
		
	
	
	def toBratInput(self):
		'''
		Generates the dataset in Brat annotation input format
		
		There are two input types: a text input containing the raw sentences and a annotation input containing the 
		annotation.
		'''
		
		vSentDocOffset = 0  # document offset of the current sentence
		vlTxt = []
		vlAnn = []
		
		for i, vAT in enumerate(self.getAspectTerms(), start=1):
			# concatenating the aspect category at the beginning of the sentence to be shown to the annotator which is 
			# required for cases where aspect term offset is not annotated 
			vlTokens = [vAT.category + ':'] + vAT.sentence.getTokens()
			vlTxt.append(' '.join(vlTokens))
			
			# setting the aspect term's character offset in the document
			
			# part of the sentence before the aspect term token (including aspect category)
			if vAT.getTokenSpan() is not None:
				vATForm = vAT.getForm()
				
				vSentUpToAT = ' '.join(vlTokens[: (vAT.getTokenSpan()[0])])
				
				# character offset of the aspect term in token(s) containing the aspect term (e.g. 5 for built in well-built)
				vATPosInToken = ' '.join(vAT.getTokens()).find(vAT.getForm())
			else:
				vATForm = vAT.category
				
				vSentUpToAT = ''
				
				# aspect term is substitueted with the aspect category concatenated to the beginning of the sentence
				vATPosInToken = 0
			
			
			if (vSentUpToAT == ''):
				vATDocOffset = vSentDocOffset + len(vSentUpToAT.decode("utf-8")) + vATPosInToken
			else:
				vATDocOffset = vSentDocOffset + len(vSentUpToAT.decode("utf-8")) + vATPosInToken + 1
			
			vlAnn.append("T%s\tAT-%s %s %s\t%s" % (i, vAT.getPolarity()[:3], vATDocOffset, vATDocOffset + len(vATForm.decode("utf-8")), vATForm))
			
			# setting the next sentence's character offset in the document 
			vSentDocOffset += len(vlTxt[-1].decode("utf-8")) + 1  # +1 for newline
		
		return vlTxt, vlAnn
		
	
	
	def loadBratOE(self, pBratAnnFilename, pflgVerbos = False, pdNewAnnTypes = None):
		'''
		Calls loadBratSE()
		'''
		
		self.loadBratSE(pBratAnnFilename, pflgVerbos = pflgVerbos, pdNewAnnTypes = pdNewAnnTypes)
		
	
	
	def loadBratSE(self, pBratAnnFilename, pflgVerbos = False, pdNewAnnTypes = None):
		'''
		Loads sentiment or sentiment expression annotation from Brat annotator output format
		
		It assumes that the Brat input was provided using toBratInput() method, so the order 
		of the sentences are retained.
		'''
		
		# first loading SE from Brat ann file
		
		vlSEs = []
		
		for l in open(pBratAnnFilename).read().strip().split('\n'):
			vlSplit = l.split('\t')
			if vlSplit[1].startswith("SE"):
				vlSESplit = vlSplit[1].split()
				
				# renaming the annotation types if new names are given
				if pdNewAnnTypes is not None and pdNewAnnTypes != {}:
					vAnnType = pdNewAnnTypes[vlSESplit[0]]
				else:
					vAnnType = vlSESplit[0]
				
				vlSEs.append({"type": vAnnType, "from": int(vlSESplit[1]), "to": int(vlSESplit[2]), "surface": vlSplit[2]})
		
		# sorting SE list based on their document offset
		vlSEs.sort(key=lambda x: x["from"])
		
		# extracting SEs for each aspect term: assumes that the Brat input was provided using toBratInput() method, so the order 
		# of the sentences are retained.
		
		vSentDocOffset = 0
		
		vlMismatches =  []                                          # list of mismatches between original and extracted SE tokens
		
		for i, vAT in enumerate(self.getAspectTerms(), start=1):
			# NOTE: this may create problem. It is used because during loading the data, the sentence can be encoded (see the code)
			# which causes problem here. There may be a better solution but it will require revisiting all the code.
			vSentence = vAT.sentence.getText().decode("UTF-8")
			
			if pflgVerbos:
				print "%s) %s" % (i, vSentence)
				if vAT.getTokenSpan() is None:
					print "\nAT: " + vAT.category
				else:
					print "\nAT: " + vAT.getForm()
				print "    Polairty: " + vAT.getPolarity()
				print "SEs:"
			
			# calculating document character span of the aspect term's sentence
			vSentDocSpan = (vSentDocOffset, vSentDocOffset + len(vAT.category) + 1 + len(vSentence) + 1)
			
			# collecting SEs of the current AT based on character offsets: not very efficient
			vlATSEs = [se for se in vlSEs if se["from"] >= vSentDocSpan[0] and se["to"] <= vSentDocSpan[1]]
			if pflgVerbos:
				if len(vlATSEs) == 0:
					print "    No SE annotated for this aspect term (polarity is %s)" % vAT.getPolarity()
			
			# adding the SEs to the aspect term 
			for se in vlATSEs:
				# sanity check
				if se["to"] > vSentDocSpan[1]:
					raise Exception(
						"Invalid SE span: %s > %s; the end falls out of the sentence" % (se["to"], vSentDocSpan[1]))
				
				# finding token span of the SE
				vTokenOffsetStart = len(vAT.category) + 2 + vSentDocOffset
				vSETokenSpanStart = 0
				vSETokenSpanEnd = 0
				for j, token in enumerate(vAT.sentence.getTokens(), start=1):
					# NOTE: see above for sentence decoding
					vToken = token.decode("UTF-8")
					
					vTokenOffsetEnd = vTokenOffsetStart + len(vToken)          # end of this token
					
					# start token: sometimes the SE is annotated from the middle of token when the tokenization has problem (e.g. -when) 
					if se["from"] >= vTokenOffsetStart and se["from"] <= vTokenOffsetEnd:
						vSETokenSpanStart = j
					
					# end token: sometimes the SE is annotated until the middle of token when the tokenization has problem (e.g. headphones/mic) 
					if se["to"] >= vTokenOffsetStart and se["to"] <= vTokenOffsetEnd:
						vSETokenSpanEnd = j
					
					vTokenOffsetStart += len(vToken) + 1  # start of next token
				
				vSE = absa.OpinionExpression()
				vSE.span = (vSETokenSpanStart, vSETokenSpanEnd)
				vSE.type = se["type"]
				
				if vSE.span == (0, 0):
					raise Exception("No token match was found!\n%s" % se)
				else:
					vAT.addOE(vSE)
					
					if pflgVerbos:
						print "    %s" % vSE.type
						print "        Original: %s" % se["surface"]
						print "        Extracted: %s" % vSE.getForm()
				
				if se["surface"] != vSE.getForm():
					#raise Exception("original != extracted!")
					vlMismatches.append((i, vSentence, se["surface"], vSE.getForm()))
				
			if pflgVerbos:
				print "\n............................................."
			
			# setting the next sentence's character offset in the document 
			vSentDocOffset += (len(vAT.category) + 2 + len(vSentence) + 1)  # +2 for :space between category and sentene; +1 for newline 
		
		return vlMismatches
		
	
	
	def getAllAspectCategories(self):
		'''
		Extracts and returns all aspect categories in the data set
		 
		The returned structure is a dictionary with the aspect categories as keys and their cound as values 
		'''
		
		return util.groupBy([c for s in self.getSentences() for c in s.getAspectCategories()])
		
	
	
	def getAllAspectEntities(self):
		'''
		Extracts and returns all aspect entities in the data set
		
		The returned structure is a dictionary with the aspect entities as keys and their cound as values 
		'''
		
		return util.groupBy([c for s in self.getSentences() for c in s.getAspectEntities()])
		
	
	
	def getAllAspectAttributes(self):
		'''
		Extracts and returns all aspect attributes in the data set
		
		The returned structure is a dictionary with the aspect attributes as keys and their cound as values 
		'''
		
		return util.groupBy([c for s in self.getSentences() for c in s.getAspectAttributes()])
		
	
	


class ABSA2016Review(absa.ABSAContext):
	'''
	Class for ABSA 2016 reviews
	'''
	
	
	def __init__(self, pABSA2016Set):
		'''
		Constructor
		'''
		
		absa.ABSAContext.__init__(self, pABSA2016Set)
		
		self.id = None
		
	
	
	def getID(self):
		'''
		Returns the id of the review
		'''
		
		return self.id
		
	
	
	def loadParsedReview(self, pXMLReview, pflgTokenize=False, pflgUTF8=False, pSort='', pLanguage="en", pflgFixETC=False):
		'''
		Loads the review content from parsed XML review
		'''
		
		self.id = pXMLReview.get("rid")
		
		for vSent in pXMLReview.iter("sentence"):
			vABSA2016Sent = ABSA2016Sent(pABSA2016Review = self)
			vABSA2016Sent.loadParsedSent(vSent, pflgTokenize=pflgTokenize, pflgUTF8=pflgUTF8, pLanguage=pLanguage, pflgFixETC=pflgFixETC)
			self.sentences.append(vABSA2016Sent)
		
		if pSort.lower() == "id":
			self.sentences.sort(key=lambda x: int(x.getID().split(':')[1]))
		elif pSort.lower() == "text":
			self.sentences.sort(key=lambda x: x.getText())
		
	
	
	def getAspectCategories(self):
		'''
		Extracts and returns the list of aspect categories in the review
		'''
		
		return [c for s in self.getSentences() for c in s.getAspectCategories()]
		
	
	
	def getReviewTokens(self):
		'''
		Returns all tokens in the review sentences 
		'''
		
		return self.getTokens()
		
	


class ABSA2016Sent(absa.ABSASent):
	'''
	Class for SemEval 2016 ABSA sentence
	'''
	
	def __init__(self, pABSA2016Review):
		'''
		Constructor
		'''
		
		# ABSA2016Review the sentence belongs to
		self.review = pABSA2016Review
		
		absa.ABSASent.__init__(self, pABSA2016Review)
		
		self.id = None
		
	
	
	def loadParsedSent(self, pXMLSent, pflgTokenize=False, pflgUTF8=False, pLanguage="en", pflgFixETC=False):
		'''
		Loads the sentence content from parsed XML sentence
		
		Since the original data is not tokenized, both text and aspect 
		terms can optionally be tokenized during loading.
		'''
		
		self.id = pXMLSent.get("id")
		
		if pflgUTF8:
			self.text = pXMLSent.findtext("text").strip().encode("UTF-8")
		else:
			self.text = pXMLSent.findtext("text").strip()
		
		##########
		#print '--->', pflgUTF8, type(pXMLSent.findtext("text").strip().decode("UTF-8"))
		##########
		
		if pflgTokenize:
			self.text = nlp.tokenizeSegment(self.text, pLang=pLanguage, pflgTokenizeFSlash=False)
			# fixing "etc." tokenization issue
			if pflgFixETC:
				self.text = self.text.replace(" etc.", " etc .")
		
		# aspect term is equivalent to opinion in 2016 data
		for vATerm in pXMLSent.iter("Opinion"):
			vABSA2016ATerm = ABSA2016ATerm()
			vABSA2016ATerm.loadParsedATerm(vATerm, pABSA2016Sent=self, pflgTokenize=pflgTokenize, pflgUTF8=pflgUTF8, pLanguage=pLanguage, pflgFixETC=pflgFixETC)
			self.aspectTerms.append(vABSA2016ATerm)
		
	
	
	def getID(self):
		'''
		Returns sentence ID
		'''
		
		return self.id
		
	
	
	def getAspectCategories(self):
		'''
		Extracts and returns the list of aspect categories in the sentence
		'''
		
		return [at.getCategory() for at in self.getAspectTerms()]
		
	
	
	def getAspectEntities(self):
		'''
		Extracts and returns the list of aspect entities in the sentence
		'''
		
		return [at.getEntity() for at in self.getAspectTerms()]
		
	
	
	def getAspectAttributes(self):
		'''
		Extracts and returns the list of aspect attributes in the sentence
		'''
		
		return [at.getAttribute() for at in self.getAspectTerms()]
		
	


class ABSA2016ATerm(absa.AspectTerm):
	'''
	Class for SemEval 2016 ABSA aspect term
	
	Aspect terms are called opinions in 2016 data.
	'''
	
	def __init__(self):
		'''
		Constructor
		'''
		
		absa.AspectTerm.__init__(self)
		
		## original from and to (0-base character offset)
		## Caution: do not match text if it's tokenized during load.
		## NOTE: the laptop data set does not contain aspect term offset annotation
		self._from = None
		self._to = None
		
		self.category= None
		
	
	
	def loadParsedATerm(self, pXMLATerm, pABSA2016Sent, pflgTokenize=False, pflgUTF8=False, pLanguage="en", pflgFixETC=False):
		'''
		Loads the aspect term details from parsed XML aspect term
		
		Since the original data is not tokenized, both text and aspect terms can optionally be tokenized during loading.
		'''
		
		# Aspect term is called target in 2016 data and is not annotated in the laptop data set and in some restaurant
		# sentences, it's NULL meaning implicit apsect terms
		
		vTerm = pXMLATerm.get("target")
		if vTerm is not None and vTerm != "NULL":
			if pflgUTF8:
				self.term = pXMLATerm.get("target").encode("UTF-8")
			else:
				self.term = pXMLATerm.get("target")
			
			if pflgTokenize:
				self.term = nlp.tokenizeSegment(self.term, pLang=pLanguage, pflgTokenizeFSlash=False)
				
				# fixing "etc." tokenization issue
				if pflgFixETC:
					self.term = self.term.replace(" etc.", " etc .")
				
			self._from = int(pXMLATerm.get("from"))
			self._to = int(pXMLATerm.get("to"))
			
			# Mapping original character offset to token span
			self.span = self._offsetToSpan(pABSA2016Sent.getText(), self.getForm(), self._from, self._to)
		
		self.category = pXMLATerm.get("category")
		self.polarity = pXMLATerm.get("polarity")
		
		self.sentence = pABSA2016Sent
		
	
	
	def _offsetToSpan(self, pSentenceForm, pTermForm, pCOffsetFrom, pCOffsetTo):
		'''
		Maps the original character offset to token span
		'''
		
		## The main problem of mapping is the tokenization as the original 
		## offset values are based on untokenized version of the sentence.
		## The problem can be alleviated by starting the search for the 
		## term in the (probably tokenized) sentence from the offset start,
		## because the tokenization tends to expand the text, so the term
		## will be pushed farther in the sentence. However, there can be 
		## a problem because our tokenizer trims trailing or surrounding
		## spaces, so in rare cases term can be pushed ahead of the original
		## offset. Therefore, we start N characters early but a sanity 
		## check should be performed, when using this function.
		
		if pCOffsetFrom > 5:
			vSearchStart = pCOffsetFrom - 5
		else:
			vSearchStart = 0
		
		vPosition = pSentenceForm.decode("utf-8").find(pTermForm.decode("utf-8"), vSearchStart)
		#vPosition = pSentenceForm.find(pTermForm, vSearchStart)
		
		vSpanStart = 1
		
		if vPosition == -1:
			print "Term %s (%s, %s) can not be located in: %s" % (pTermForm, pCOffsetFrom, pCOffsetTo, pSentenceForm)
		elif vPosition > pCOffsetFrom + len(pSentenceForm[:vPosition].split()):
			print "Term %s (%s, %s) is located at %s in: %s, but the position seems suspicious!" % (pTermForm, pCOffsetFrom, pCOffsetTo, vPosition, pSentenceForm)
		else:
			# The number of spaces before the position is where the span starts.
			
			for i in range(vPosition):
				if pSentenceForm.decode("utf-8")[i] == ' ':
					vSpanStart += 1
		
		return (vSpanStart, vSpanStart + len(pTermForm.split()) - 1)
		
	
	
	def getCategory(self):
		'''
		Returns the aspect category
		 
		The aspect category is a string in Entity#Attribute format (e.g. FOOD#QUALITY)
		'''
		
		return self.category
		
	
	
	def getEntity(self):
		'''
		Returns the aspect entity
		
		The aspect entity is extracted from aspect category which is a string in Entity#Attribute format (e.g. FOOD#QUALITY)
		'''
		
		return self.getCategory().split('#')[0]
		
	
	
	def getAttribute(self):
		'''
		Returns the aspect attribute
		
		The aspect attribute is extracted from aspect category which is a string in Entity#Attribute format (e.g. FOOD#QUALITY)
		'''
		
		return self.getCategory().split('#')[1]
		
	
	









