#! /usr/bin/python

"""	
	This module defines general-purpose utility functions.
	
	
	Version 3.5												(11-Jun-2018)
	- getLogger() is added.
	
	Version 3.4												(27-Nov-2017)
	- createOneHot() is added.
	
	Version 3.3												(29-Sep-2017)
	- padListSequences() is added.
	
	Version 3.2								(14-Feb-2017 to 24-Feb-2017)
	- padSequences() is added.
	- prettyPrintDict is added.
	
	Version 3.1												(12-Feb-2016)
	- getSpanDistance() is added.
	- Bug in compareRages() is fixed for the condition of completely
	  separate ranges.
	
	Version 3.0												(11-Jan-2016)
	- compareRanges() is added.
	
	Version 2.9												(19-Aug-2015)
	- findDataType() is added.
	
	Version 2.8												(31-Jul-2015)
	- isNumeric() is added.
	
	Version 2.7												(15-May-2015)
	- PredEval, PredStatSig, TTest, IAA are moved to ml/eval.py.
	
	Version 2.6												(02-May-2014)
	- IAA is added.
	
	Version 2.5												(28-Feb-2014)
	- percent() is added.
	
	Version 2.4												(04-Feb-2014)
	- sed() is added.
	- rm() is added.
	
	Version 2.3												(18-Jun-2013)
	- extractBlocks() is improved and bug-fixed.
	
	Version 2.2												(07-Jun-2013)
	- range() is added to PredEval.
	- sd() is added to PredEval.
	
	Version 2.1												(26-May-2013)
	- normalizeLang() is added.
	- extractBlocks() is added.
	
	Version 2.0												(03-May-2013)
	- normalizeNumber() is renamed to scaleNumber().
	
	Version 1.9												(18-Apr-2013)
	- mae() and acc() are added to PredEval.
	- winRatioTestAcc(), winRatioTestMAE() are added to PredStatSig.
	
	Version 1.8												(11-Apr-2013)
	- TTest is added.
	- PredEval is added.
	- PredStatSig is added.
	
	Version 1.7												(24-Mar-2013)
	- padDecimalZero() is added.
	
	Version 1.6												(22-Feb-2012)
	- normalizeLangID() is added.
	
	Version 1.5												(22-Jan-2012)
	- getDicItemIdx() is added and its coding improved.
	
	Version 1.4												(30-Nov-2012)
	- matchMasks() is upgraded to include 999.
	
	Version 1.3												(29-Nov-2012)
	- groupBy() is added.
	- matchMasks() is added.
	
	Version 1.2												(11-Oct-2012)
	- csvIntersection() is added.
	
	Version 1.1												(13-Aug-2012)
	- lineDiff() is changed.
	
	Version 1.0												(07-Aug-2012)
	- normalizeNumber() is added.
	- lineDiff() is moved over from tempfuncs.
	
"""

from decimal import *
import sys, re, math, random
import numpy as np
import logging


def scaleNumber(pNumber, pSrcRange, pTrgRange):
	''' 
	Given a number and the range it belongs (pSrcRange), scales it into 
	a target range (pTrgRange).
	'''
	
	if pTrgRange[0] > pTrgRange[1]:
		raise Exception("Not a valid range: [" + str(pTrgRange[0]) + ', ' + str(pTrgRange[1]) + ']')
	
	if not pSrcRange[0] <= pNumber <= pSrcRange[1]:
		raise Exception(str(pNumber) + " is not in [" + str(pSrcRange[0]) + ', ' + str(pSrcRange[1]) + ']')
	elif pSrcRange[0] == pSrcRange[1]:
		## in this strange situation, the number is mapped to the middle 
		## of target range 
		return pTrgRange[0] + pTrgRange[1] / 2.0
	else:
		return pTrgRange[0] + (pNumber - pSrcRange[0]) * (pTrgRange[1] - pTrgRange[0]) * 1.0 / (pSrcRange[1] - pSrcRange[0])
		
	


def lineDiff(pFilename1, pFilename2):
	'''
	06-Aug-2012
	
	Given two test files with the same number of lines, finds number of 
	lines different between them. Also outputs the 1-base index of different
	lines.
	'''
	
	# openning files
	
	try:
		vf1 = open(pFilename1)
	except IOError:
		sys.exit("Cannot open " + pFilename1)
		
	try:
		vf2 = open(pFilename2)
	except IOError:
		sys.exit("Cannot open " + pFilename2)
		
	vLineCntr = 0
	vDiffLines = 0
	for vLine1, vLine2 in zip(vf1, vf2):
		if vLine1.strip() != vLine2.strip():
			vDiffLines += 1
			sys.stderr.write(str(vLineCntr + 1) + ' ')
		vLineCntr += 1
	
	print "Number of all lines in each file: " + str(vLineCntr)
	print "Number of different lines:        " + str(vDiffLines)
		
	


def csvIntersection(plCSVFiles, pOutFilename="isect", pDelim=','):
	'''
	Given a list of csv files, extracts the intersection of pDelim-separated
	items and outputs in a file.
	'''
	
	if len(plCSVFiles) < 2:
		sys.exit("At leas two CSV files are needed")
	
	# list of list of items in each file
	vlCSVItems = []
	
	# reading CSVs
	for vCSVFilename in plCSVFiles:
		try:
			vlCSVItems.append(open(vCSVFilename).read().strip().split(pDelim))
		except IOError:
			sys.exit("Cannot open CSV file: %s" % vCSVFilename)
	
	# getting intersection
	vsIsect = set(vlCSVItems[0])
	for vSet in vlCSVItems[1:]:
		vsIsect = vsIsect.intersection(set(vSet))
	
	# writing output
	
	try:
		vfOut = open(pOutFilename, 'w')
	except IOError:
		sys.exit("cannot create output file: %s" % pOutFilename)
	
	vfOut.write(pDelim.join(list(vsIsect)) + '\n')
	
	vfOut.close()
		
	


def groupBy(plList):
	'''
	Given a list, group by items into a dictionary.
	'''
	
	vdGroupBy = {}
	for vItem in plList:
		if vItem in vdGroupBy:
			vdGroupBy[vItem] += 1
		else:
			vdGroupBy[vItem] = 1
	
	return vdGroupBy
		
	


def matchMasks(pCorpus1, pCorpus2, pflgPrint=False):
	'''
	Given two parallel corpora, checks them line by line for matching 
	masks. Masks are strings in the -ABC- format created by clean.py
	script.
	'''
	
	try:
		vlCorpus1 = open(pCorpus1).readlines()
	except IOError:
		sys.exit("cannot open %s" % pCorpus1)
	
	try:
		vlCorpus2 = open(pCorpus2).readlines()
	except IOError:
		sys.exit("cannot open %s" % pCorpus12)
	
	vC1Len = len(vlCorpus1)
	vC2Len = len(vlCorpus2)
	if vC1Len != vC2Len:
		sys.exit("Mismatch in corpora length! %s <> %s" % (vC1Len, vC2Len))
	
	vMaskLineNum = 0
	vlMismatch = []
	vdMasks1 = vdMasks2 = {}
	for i, (vLine1, vLine2) in enumerate(zip(vlCorpus1, vlCorpus2), start=1):
		vLine1 = vLine1.strip('\n')
		vLine2 = vLine2.strip('\n')
		vPattern = re.compile(r"(?:-[A-Z]{3}-)|(?:\b999\b)")
		vdMasks1 = groupBy(re.findall(vPattern, vLine1))
		vdMasks2 = groupBy(re.findall(vPattern, vLine2))
		
		if len(vdMasks1) > 0 or len(vdMasks2) > 0:
			vMaskLineNum += 1
		
		if vdMasks1 != vdMasks2:
			vlMismatch.append(i)
			
			if pflgPrint:
				print "Line %s" % i
				print "[1]:\t%s" % vLine1
				print "[2]:\t%s" % vLine2
				print
	
	print "============================================================"
	print "Number of lines containing mask on either side: %s (out of %s)" % (vMaskLineNum, vC1Len)
	print "Number of lines containing mask mismatch: %s (%s%%)" % (len(vlMismatch), (Decimal(len(vlMismatch))*100/vMaskLineNum).quantize(Decimal('0.01')))
	
	return vlMismatch, vdMasks1, vdMasks2
		
	


def getDicItemIdx(pList, pKey, pValue):
	'''
	Given a list of dictionary, pList, returns the index of list element
	where the value of key pKey is pValue.
	
	If the list in not unique on pKey, the first occurrence will be returned.
	'''
	
	lItems = [item for item in pList if item[pKey] == pValue]
	
	return pList.index(lItems[0])
		
	


def normalizeLangID(pLang):
	'''
	Given a string, tries to normalize it to a two-character language 
	ID or code (e.g. en for English).
	'''
	
	if pLang.lower() in ["english", "en"]:
		return "en"
	elif pLang.lower() in ["french", "fr"]:
		return "fr"
	elif pLang.lower() in ["german", "de"]:
		return "de"
		
	


def padDecimalZero(pNum, pDecDigNum):
	'''
	Pads zeros to the end of pNum which is a float number to make the number
	of decimal digit be pDecDigNum, e.g. padDecimalZero(2.1, 3) = "2.100"
	'''
	
	vNumStr = str(pNum)
	vlParts = vNumStr.split('.')
	
	vIntPart = vlParts[0]
	if len(vlParts) > 1:
		vFracpart = vlParts[1]
	else:
		vFracpart = '0'
	
	vFracpart = vFracpart.ljust(pDecDigNum, '0')
	
	return vIntPart + '.' + vFracpart
		
	


def normalizeLang(pLang):
	'''
	Normalizes language identifier string to lowercase 2-letter codes.
	'''
	
	if pLang.lower() in ["english", "en", "eng"]:
		return "en"
	elif pLang.lower() in ["french", "fr"]:
		return "fr"
	elif pLang.lower() in ["german", "de"]:
		return "de"
	else:
		raise Exception("Unknown language %s" % pLang)
		
	


def extractBlocks(pText, pBlockStartRX = '',  pBlockEndRX = '',  pSelectionMethod = "random", pSelectionSize = 0, plPositions = None):
	'''
	Given a text, extracts blocks of text, which starts with a line matching 
	pBlockStartRX and ends with a line mathcing pBlockEndRX, according to
	pSelectionMethod.
	
	It returns the selection and the leftover.
	
	There are 4 possible combinations of start and end regex:
	 #1- None are given: A block is interpreted as a paragraph where a blank
	     line triggers the end of block. So, it is converted to #2.
	 #2- Both are given: If the start regex is met again before the current
	     block is ended by end regex, it will be ignored and lines will 
	     be added to the current block. Note that anything after the end
	     of current block and before the start of next block will be lost.
	 #3- Only start is given: A block is started once the start regex is
	     met. All lines before the first start line will be ignored.
	 #4- Only end is given: A block is ended anytime the end regex 
	     is met. All lines after the last end line will form a new block.
	
	pSelectionMethod is one of:
	- random (or rnd or rand): randomly selects pSelectionSize blocks
	- pos (or position or index): extracts block at positions specified 
	  in plPositions.
	'''
	
	# loading text into blocks
	
	# each block element is a list of lines of the block
	vlBlocks = []
	
	## If none of start and end regex are given the default values are set
	## to convert it to configuration #2
	if pBlockStartRX == '' and pBlockEndRX == '':
		vBlockStartRX = ".+"
		vBlockEndRX = "^$"
	else:
		vBlockStartRX = pBlockStartRX
		vBlockEndRX = pBlockEndRX
	
	if vBlockStartRX != '' and vBlockEndRX != '':
		vflgBlockEnded = True
		for vLine in pText.split("\n"):
			if re.match(vBlockStartRX, vLine) and vflgBlockEnded:
				vlBlocks.append([vLine])
				vflgBlockEnded = False
			else:
				vlBlocks[-1].append(vLine)
				if re.match(vBlockEndRX, vLine):
					vflgBlockEnded = True
		
	elif vBlockStartRX != '' and vBlockEndRX == '':
		for vLine in pText.split("\n"):
			if re.match(vBlockStartRX, vLine):
				vlBlocks.append([vLine])
			else:
				vlBlocks[-1].append(vLine)
		
	elif vBlockEndRX != '' and vBlockStartRX == '':
		vflgBlockEnded = True
		for vLine in pText.split("\n"):
			if vflgBlockEnded:
				vlBlocks.append([vLine])
				vflgBlockEnded = False
			else:
				vlBlocks[-1].append(vLine)
				if re.match(vBlockEndRX, vLine):
					vflgBlockEnded = True
	
	# setting selection
	
	vTextSize = len(vlBlocks)
	vSelMethod = pSelectionMethod.lower()
	
	if vSelMethod in ["random", "rand", "rnd"]:
		vPopulation = range(0, vTextSize)
		vlSelection = random.sample(vPopulation, pSelectionSize)
	elif vSelMethod in ["pos", "position", "index"]:
		vlSelection = list(set([int(idx) - 1 for idx in plPositions]))
	
	vlSelection.sort()
	
	# extracting selection
	
	vlSelBlocks = []
	vlLeftoverBlocks = []
	
	for i, vBlock in enumerate(vlBlocks):
		if i in vlSelection:
			vlSelBlocks.append(vBlock)
		else:
			vlLeftoverBlocks.append(vBlock)
	
	# putting blocks in text
	
	return '\n'.join(['\n'.join(block) for block in vlSelBlocks]) + '\n', \
	        '\n'.join(['\n'.join(block) for block in vlLeftoverBlocks]) + '\n'
		
	


def sed(pPattern, pReplacement, pInputFile, pOutputFile = ''):
	'''
	Substitutes bash sed command
	'''
	
	import re
	
	vPattern = re.compile(pPattern)
	vOutput = vPattern.sub(pReplacement, open(pInputFile).read())
	
	if pOutputFile == '':
		pOutputFile = pInputFile
	
	open(pOutputFile, 'w').write(vOutput)
		
	


def rm(pPattern):
	'''
	Substitutes bash rm command for removing files matching the given pattern
	'''
	
	import os, glob
	
	for vFilename in glob.glob(pPattern):
		if os.path.isfile(vFilename):
			os.remove(vFilename)
		
	


def percent(pPart, pWhole):
	'''
	Calculates and returns what pecentage of a given whole is a given part
	'''
	
	if pWhole == 0:
		return 0
	else:
		return pPart * 100.0 / pWhole 
		
	


def isNumeric(pInput):
	'''
	Returns true if the input is numerical (float or integer) and false
	otherwise
	'''
	
	if type(pInput) is float or type(pInput) is int:
		return True
	elif type(pInput) is str:
		try:
			float(pInput)
			return True
		except ValueError:
			return False
	else:
		return False
		
	


def findDataType(pStrValue):
	'''
	Given a string value, it finds out and returns the data type of 
	the value contained in the string
	
	If the value is not string, the existing data type is returned.
	'''
	
	if type(pStrValue) != str:
		return type(pStrValue)
	elif isNumeric(pStrValue):
		if '.' in pStrValue:
			return type(0.)
		else:
			return type(0)
	else:
		return type('')
	
	


def compareRanges(ptRange1, ptRange2):
	'''
	Compares two ranges and returns a value which determines the status of their intersection
	
	The ranges are tuples of which the boundaries are inclusive.
	
	The return values are:
	- 0: two ranges are completely separate
	- 1: two ranges are exactly equal
	- 2: range 1 is included in range 2
	- 3: range 2 is included in range 1
	- 4: range 1 left-crosses range 2
	- 5: range 1 right-crosses range 2
	'''
	
	if ptRange1 == ptRange2:
		return 1
	elif ptRange1[0] > ptRange2[1] or ptRange1[1] < ptRange2[0]:
		return 0
	elif ptRange1[0] >= ptRange2[0] and ptRange1[1] <= ptRange2[1]:
		return 2
	elif ptRange1[0] <= ptRange2[0] and ptRange1[1] >= ptRange2[1]:
		return 3
	elif ptRange1[0] < ptRange2[0] and ptRange1[1] < ptRange2[1]:
		return 4
	elif ptRange1[0] > ptRange2[0] and ptRange1[1] > ptRange2[1]:
		return 5
	
	


def getSpanDistance(ptSpan1, ptSpan2):
	'''
	Calculates and returns the distance between to spans in a virtual sequence
	
	Spans are tuples of starting and ending element position in the sequence.
	
	If span 2 is in the right of span 1, the distance will be positive and negative otherwise.
	'''
	
	if ptSpan1[1] <= ptSpan2[0]:                           # span 2 is in the right and outside of span 1
		return ptSpan2[0] - ptSpan1[1]
	elif ptSpan1[0] >= ptSpan2[1]:                         # span 2 is in the left and outside of span 1
		return ptSpan2[1] - ptSpan1[0]
	elif compareRanges(ptSpan1, ptSpan2) > 0:              # spans overlap
		return 0
	
	


def padSequences(pllSequences, pMaxSeqLen, pPadStr=''):
	'''
	Pads sequences of tokens with the given padding string, up to the given length
	
	Sequences longer than the given length will be truncated.
	'''
	
	vllPaddedSeqs = []
	
	for vlSeq in pllSequences:
		if len(vlSeq) >= pMaxSeqLen:
			vllPaddedSeqs.append(vlSeq[: pMaxSeqLen])
		else:
			vllPaddedSeqs.append(vlSeq + [pPadStr] * (pMaxSeqLen - len(vlSeq)))
	
	return vllPaddedSeqs
	


def padNumSequences(pllSequences, pMaxSeqLen, pPadNum=0):
	'''
	Pads sequences of numbers (normally indicies) with the given padding number, up to the given length
	
	Sequences longer than the given length will be truncated.
	'''
	
	vllPaddedSeqs = []
	
	for vlSeq in pllSequences:
		if len(vlSeq) >= pMaxSeqLen:
			vllPaddedSeqs.append(vlSeq[: pMaxSeqLen])
		else:
			vllPaddedSeqs.append(vlSeq + [pPadNum] * (pMaxSeqLen - len(vlSeq)))
	
	return vllPaddedSeqs
	


def padListSequences(pllSequences, pMaxSeqLen, pPadList=[0]):
	'''
	Pads sequences of lists with the given padding list, up to the given length
	
	Sequences longer than the given length will be truncated.
	'''
	
	vllPaddedSeqs = []
	
	for vlSeq in pllSequences:
		if len(vlSeq) >= pMaxSeqLen:
			vllPaddedSeqs.append(vlSeq[: pMaxSeqLen])
		else:
			vllPaddedSeqs.append(vlSeq + [pPadList] * (pMaxSeqLen - len(vlSeq)))
	
	return vllPaddedSeqs
	


def prettyPrintDict(pdInput, pSort = "asc", pIndent = 0, pDelim = '\n'):
	'''
	Pretty-prints the given dictionary into a string as the return value 
	'''
	
	if pSort.lower().startswith("asc"):
		vlList =  sorted([(k, v) for k, v in pdInput.iteritems()], key = lambda x: x[0])
	elif pSort.lower().startswith("desc"):
		vlList =  sorted([(k, v) for k, v in pdInput.iteritems()], key = lambda x: x[0], reverse = true)
	
	return pDelim.join(["%s%s: %s" % (pIndent * ' ', k, v) for k, v in vlList])
	


def createOneHot(psCategories):
	'''
	Given a python set of categories, creates and returns a dictionary which maps each category to a one-hot vector 
	'''
	
	vCatCount = len(psCategories)
	vdOneHots = {}
	
	for i, c in enumerate(sorted(psCategories)):
		vaOneHot = np.zeros(vCatCount, dtype=np.int)
		np.put(vaOneHot, i, 1)
		vdOneHots[c] = vaOneHot
	
	return vdOneHots
	


def getLogger(pFilename):
	'''
	Returns a logger instance that writes in filename
	
	Args:
		filename: (string) path to log.txt
	'''
	
	vLogger = logging.getLogger('logger')
	vLogger.setLevel(logging.DEBUG)
	logging.basicConfig(format='%(message)s', level = logging.DEBUG)
	vHandler = logging.FileHandler(pFilename)
	vHandler.setLevel(logging.DEBUG)
	vHandler.setFormatter(logging.Formatter('%(asctime)s:%(levelname)s: %(message)s'))
	vLogger.addHandler(vHandler)
	
	return vLogger
	

